/* Mapping tables for Big5-HKSCS handling.
   Copyright (C) 1997, 1998, 2000 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Ulrich Drepper <drepper@cygnus.com>, 1997.
   Modified for Big5-HKSCS by Roger So <spacehunt@e-fever.org>, 2000.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <dlfcn.h>
#include <gconv.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>


/* Table for Big5 to UCS conversion.

   With HKSCS mappings 0x8140-0xA0FE and 0xFA40-0xFEFE added; more info:
   http://www.digital21.gov.hk/eng/hkscs/index.html
     - spacehunt 07/01/2000

   We have a few problems with this
   conversion.  Cited from the file `big5.txt' from the Unicode CD:

   ``WARNING!  It is currently impossible to provide round-trip compatibility
               between BIG5 and Unicode.

     A number of characters are not currently mapped because
     of conflicts with other mappings.  They are as follows:

     BIG5        Description                    Comments

     0xA15A      SPACING UNDERSCORE             duplicates A1C4
     0xA1C3      SPACING HEAVY OVERSCORE        not in Unicode
     0xA1C5      SPACING HEAVY UNDERSCORE       not in Unicode
     0xA1FE      LT DIAG UP RIGHT TO LOW LEFT   duplicates A2AC
     0xA240      LT DIAG UP LEFT TO LOW RIGHT   duplicates A2AD
     0xA2CC      HANGZHOU NUMERAL TEN           conflicts with A451 mapping
     0xA2CE      HANGZHOU NUMERAL THIRTY        conflicts with A4CA mapping''

   To resolve this problem we don't recognize these characters as legal.

   The following table can be generated from the file
	unix/mappings/eastasia/other/big5.txt
   from the Unicode CD (also available on their FTP server) using
   the command

	egrep '^0x' \
	  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
	grep -v 'NO MAPPING' | perl ~/big5.pl

   where big5.pl contains:

  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  while (<>) {
    local($big5, $ucs, %rest) = split;
    $big5 = hex($big5) - 0xa140;
    local($ridx) = int($big5 / 256) * 157;
    if ($big5 % 256 > 0x3e) {
      $ridx += ($big5 % 256) - 0x22;
    } else {
      $ridx += ($big5 % 256);
    }
    printf "\n " if (($n % 4) == 0);
    ++$n;
    printf " [%5d] = %s,", $ridx, $ucs;
  }
  print "\n";
  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const uint16_t big5_to_ucs[] =
{
  [    0] = 0xEEB8, [    1] = 0xEEB9, [    2] = 0xEEBA, [    3] = 0xEEBB,
  [    4] = 0xEEBC, [    5] = 0xEEBD, [    6] = 0xEEBE, [    7] = 0xEEBF,
  [    8] = 0xEEC0, [    9] = 0xEEC1, [   10] = 0xEEC2, [   11] = 0xEEC3,
  [   12] = 0xEEC4, [   13] = 0xEEC5, [   14] = 0xEEC6, [   15] = 0xEEC7,
  [   16] = 0xEEC8, [   17] = 0xEEC9, [   18] = 0xEECA, [   19] = 0xEECB,
  [   20] = 0xEECC, [   21] = 0xEECD, [   22] = 0xEECE, [   23] = 0xEECF,
  [   24] = 0xEED0, [   25] = 0xEED1, [   26] = 0xEED2, [   27] = 0xEED3,
  [   28] = 0xEED4, [   29] = 0xEED5, [   30] = 0xEED6, [   31] = 0xEED7,
  [   32] = 0xEED8, [   33] = 0xEED9, [   34] = 0xEEDA, [   35] = 0xEEDB,
  [   36] = 0xEEDC, [   37] = 0xEEDD, [   38] = 0xEEDE, [   39] = 0xEEDF,
  [   40] = 0xEEE0, [   41] = 0xEEE1, [   42] = 0xEEE2, [   43] = 0xEEE3,
  [   44] = 0xEEE4, [   45] = 0xEEE5, [   46] = 0xEEE6, [   47] = 0xEEE7,
  [   48] = 0xEEE8, [   49] = 0xEEE9, [   50] = 0xEEEA, [   51] = 0xEEEB,
  [   52] = 0xEEEC, [   53] = 0xEEED, [   54] = 0xEEEE, [   55] = 0xEEEF,
  [   56] = 0xEEF0, [   57] = 0xEEF1, [   58] = 0xEEF2, [   59] = 0xEEF3,
  [   60] = 0xEEF4, [   61] = 0xEEF5, [   62] = 0xEEF6, [   63] = 0xEEF7,
  [   64] = 0xEEF8, [   65] = 0xEEF9, [   66] = 0xEEFA, [   67] = 0xEEFB,
  [   68] = 0xEEFC, [   69] = 0xEEFD, [   70] = 0xEEFE, [   71] = 0xEEFF,
  [   72] = 0xEF00, [   73] = 0xEF01, [   74] = 0xEF02, [   75] = 0xEF03,
  [   76] = 0xEF04, [   77] = 0xEF05, [   78] = 0xEF06, [   79] = 0xEF07,
  [   80] = 0xEF08, [   81] = 0xEF09, [   82] = 0xEF0A, [   83] = 0xEF0B,
  [   84] = 0xEF0C, [   85] = 0xEF0D, [   86] = 0xEF0E, [   87] = 0xEF0F,
  [   88] = 0xEF10, [   89] = 0xEF11, [   90] = 0xEF12, [   91] = 0xEF13,
  [   92] = 0xEF14, [   93] = 0xEF15, [   94] = 0xEF16, [   95] = 0xEF17,
  [   96] = 0xEF18, [   97] = 0xEF19, [   98] = 0xEF1A, [   99] = 0xEF1B,
  [  100] = 0xEF1C, [  101] = 0xEF1D, [  102] = 0xEF1E, [  103] = 0xEF1F,
  [  104] = 0xEF20, [  105] = 0xEF21, [  106] = 0xEF22, [  107] = 0xEF23,
  [  108] = 0xEF24, [  109] = 0xEF25, [  110] = 0xEF26, [  111] = 0xEF27,
  [  112] = 0xEF28, [  113] = 0xEF29, [  114] = 0xEF2A, [  115] = 0xEF2B,
  [  116] = 0xEF2C, [  117] = 0xEF2D, [  118] = 0xEF2E, [  119] = 0xEF2F,
  [  120] = 0xEF30, [  121] = 0xEF31, [  122] = 0xEF32, [  123] = 0xEF33,
  [  124] = 0xEF34, [  125] = 0xEF35, [  126] = 0xEF36, [  127] = 0xEF37,
  [  128] = 0xEF38, [  129] = 0xEF39, [  130] = 0xEF3A, [  131] = 0xEF3B,
  [  132] = 0xEF3C, [  133] = 0xEF3D, [  134] = 0xEF3E, [  135] = 0xEF3F,
  [  136] = 0xEF40, [  137] = 0xEF41, [  138] = 0xEF42, [  139] = 0xEF43,
  [  140] = 0xEF44, [  141] = 0xEF45, [  142] = 0xEF46, [  143] = 0xEF47,
  [  144] = 0xEF48, [  145] = 0xEF49, [  146] = 0xEF4A, [  147] = 0xEF4B,
  [  148] = 0xEF4C, [  149] = 0xEF4D, [  150] = 0xEF4E, [  151] = 0xEF4F,
  [  152] = 0xEF50, [  153] = 0xEF51, [  154] = 0xEF52, [  155] = 0xEF53,
  [  156] = 0xEF54, [  157] = 0xEF55, [  158] = 0xEF56, [  159] = 0xEF57,
  [  160] = 0xEF58, [  161] = 0xEF59, [  162] = 0xEF5A, [  163] = 0xEF5B,
  [  164] = 0xEF5C, [  165] = 0xEF5D, [  166] = 0xEF5E, [  167] = 0xEF5F,
  [  168] = 0xEF60, [  169] = 0xEF61, [  170] = 0xEF62, [  171] = 0xEF63,
  [  172] = 0xEF64, [  173] = 0xEF65, [  174] = 0xEF66, [  175] = 0xEF67,
  [  176] = 0xEF68, [  177] = 0xEF69, [  178] = 0xEF6A, [  179] = 0xEF6B,
  [  180] = 0xEF6C, [  181] = 0xEF6D, [  182] = 0xEF6E, [  183] = 0xEF6F,
  [  184] = 0xEF70, [  185] = 0xEF71, [  186] = 0xEF72, [  187] = 0xEF73,
  [  188] = 0xEF74, [  189] = 0xEF75, [  190] = 0xEF76, [  191] = 0xEF77,
  [  192] = 0xEF78, [  193] = 0xEF79, [  194] = 0xEF7A, [  195] = 0xEF7B,
  [  196] = 0xEF7C, [  197] = 0xEF7D, [  198] = 0xEF7E, [  199] = 0xEF7F,
  [  200] = 0xEF80, [  201] = 0xEF81, [  202] = 0xEF82, [  203] = 0xEF83,
  [  204] = 0xEF84, [  205] = 0xEF85, [  206] = 0xEF86, [  207] = 0xEF87,
  [  208] = 0xEF88, [  209] = 0xEF89, [  210] = 0xEF8A, [  211] = 0xEF8B,
  [  212] = 0xEF8C, [  213] = 0xEF8D, [  214] = 0xEF8E, [  215] = 0xEF8F,
  [  216] = 0xEF90, [  217] = 0xEF91, [  218] = 0xEF92, [  219] = 0xEF93,
  [  220] = 0xEF94, [  221] = 0xEF95, [  222] = 0xEF96, [  223] = 0xEF97,
  [  224] = 0xEF98, [  225] = 0xEF99, [  226] = 0xEF9A, [  227] = 0xEF9B,
  [  228] = 0xEF9C, [  229] = 0xEF9D, [  230] = 0xEF9E, [  231] = 0xEF9F,
  [  232] = 0xEFA0, [  233] = 0xEFA1, [  234] = 0xEFA2, [  235] = 0xEFA3,
  [  236] = 0xEFA4, [  237] = 0xEFA5, [  238] = 0xEFA6, [  239] = 0xEFA7,
  [  240] = 0xEFA8, [  241] = 0xEFA9, [  242] = 0xEFAA, [  243] = 0xEFAB,
  [  244] = 0xEFAC, [  245] = 0xEFAD, [  246] = 0xEFAE, [  247] = 0xEFAF,
  [  248] = 0xEFB0, [  249] = 0xEFB1, [  250] = 0xEFB2, [  251] = 0xEFB3,
  [  252] = 0xEFB4, [  253] = 0xEFB5, [  254] = 0xEFB6, [  255] = 0xEFB7,
  [  256] = 0xEFB8, [  257] = 0xEFB9, [  258] = 0xEFBA, [  259] = 0xEFBB,
  [  260] = 0xEFBC, [  261] = 0xEFBD, [  262] = 0xEFBE, [  263] = 0xEFBF,
  [  264] = 0xEFC0, [  265] = 0xEFC1, [  266] = 0xEFC2, [  267] = 0xEFC3,
  [  268] = 0xEFC4, [  269] = 0xEFC5, [  270] = 0xEFC6, [  271] = 0xEFC7,
  [  272] = 0xEFC8, [  273] = 0xEFC9, [  274] = 0xEFCA, [  275] = 0xEFCB,
  [  276] = 0xEFCC, [  277] = 0xEFCD, [  278] = 0xEFCE, [  279] = 0xEFCF,
  [  280] = 0xEFD0, [  281] = 0xEFD1, [  282] = 0xEFD2, [  283] = 0xEFD3,
  [  284] = 0xEFD4, [  285] = 0xEFD5, [  286] = 0xEFD6, [  287] = 0xEFD7,
  [  288] = 0xEFD8, [  289] = 0xEFD9, [  290] = 0xEFDA, [  291] = 0xEFDB,
  [  292] = 0xEFDC, [  293] = 0xEFDD, [  294] = 0xEFDE, [  295] = 0xEFDF,
  [  296] = 0xEFE0, [  297] = 0xEFE1, [  298] = 0xEFE2, [  299] = 0xEFE3,
  [  300] = 0xEFE4, [  301] = 0xEFE5, [  302] = 0xEFE6, [  303] = 0xEFE7,
  [  304] = 0xEFE8, [  305] = 0xEFE9, [  306] = 0xEFEA, [  307] = 0xEFEB,
  [  308] = 0xEFEC, [  309] = 0xEFED, [  310] = 0xEFEE, [  311] = 0xEFEF,
  [  312] = 0xEFF0, [  313] = 0xEFF1, [  314] = 0xEFF2, [  315] = 0xEFF3,
  [  316] = 0xEFF4, [  317] = 0xEFF5, [  318] = 0xEFF6, [  319] = 0xEFF7,
  [  320] = 0xEFF8, [  321] = 0xEFF9, [  322] = 0xEFFA, [  323] = 0xEFFB,
  [  324] = 0xEFFC, [  325] = 0xEFFD, [  326] = 0xEFFE, [  327] = 0xEFFF,
  [  328] = 0xF000, [  329] = 0xF001, [  330] = 0xF002, [  331] = 0xF003,
  [  332] = 0xF004, [  333] = 0xF005, [  334] = 0xF006, [  335] = 0xF007,
  [  336] = 0xF008, [  337] = 0xF009, [  338] = 0xF00A, [  339] = 0xF00B,
  [  340] = 0xF00C, [  341] = 0xF00D, [  342] = 0xF00E, [  343] = 0xF00F,
  [  344] = 0xF010, [  345] = 0xF011, [  346] = 0xF012, [  347] = 0xF013,
  [  348] = 0xF014, [  349] = 0xF015, [  350] = 0xF016, [  351] = 0xF017,
  [  352] = 0xF018, [  353] = 0xF019, [  354] = 0xF01A, [  355] = 0xF01B,
  [  356] = 0xF01C, [  357] = 0xF01D, [  358] = 0xF01E, [  359] = 0xF01F,
  [  360] = 0xF020, [  361] = 0xF021, [  362] = 0xF022, [  363] = 0xF023,
  [  364] = 0xF024, [  365] = 0xF025, [  366] = 0xF026, [  367] = 0xF027,
  [  368] = 0xF028, [  369] = 0xF029, [  370] = 0xF02A, [  371] = 0xF02B,
  [  372] = 0xF02C, [  373] = 0xF02D, [  374] = 0xF02E, [  375] = 0xF02F,
  [  376] = 0xF030, [  377] = 0xF031, [  378] = 0xF032, [  379] = 0xF033,
  [  380] = 0xF034, [  381] = 0xF035, [  382] = 0xF036, [  383] = 0xF037,
  [  384] = 0xF038, [  385] = 0xF039, [  386] = 0xF03A, [  387] = 0xF03B,
  [  388] = 0xF03C, [  389] = 0xF03D, [  390] = 0xF03E, [  391] = 0xF03F,
  [  392] = 0xF040, [  393] = 0xF041, [  394] = 0xF042, [  395] = 0xF043,
  [  396] = 0xF044, [  397] = 0xF045, [  398] = 0xF046, [  399] = 0xF047,
  [  400] = 0xF048, [  401] = 0xF049, [  402] = 0xF04A, [  403] = 0xF04B,
  [  404] = 0xF04C, [  405] = 0xF04D, [  406] = 0xF04E, [  407] = 0xF04F,
  [  408] = 0xF050, [  409] = 0xF051, [  410] = 0xF052, [  411] = 0xF053,
  [  412] = 0xF054, [  413] = 0xF055, [  414] = 0xF056, [  415] = 0xF057,
  [  416] = 0xF058, [  417] = 0xF059, [  418] = 0xF05A, [  419] = 0xF05B,
  [  420] = 0xF05C, [  421] = 0xF05D, [  422] = 0xF05E, [  423] = 0xF05F,
  [  424] = 0xF060, [  425] = 0xF061, [  426] = 0xF062, [  427] = 0xF063,
  [  428] = 0xF064, [  429] = 0xF065, [  430] = 0xF066, [  431] = 0xF067,
  [  432] = 0xF068, [  433] = 0xF069, [  434] = 0xF06A, [  435] = 0xF06B,
  [  436] = 0xF06C, [  437] = 0xF06D, [  438] = 0xF06E, [  439] = 0xF06F,
  [  440] = 0xF070, [  441] = 0xF071, [  442] = 0xF072, [  443] = 0xF073,
  [  444] = 0xF074, [  445] = 0xF075, [  446] = 0xF076, [  447] = 0xF077,
  [  448] = 0xF078, [  449] = 0xF079, [  450] = 0xF07A, [  451] = 0xF07B,
  [  452] = 0xF07C, [  453] = 0xF07D, [  454] = 0xF07E, [  455] = 0xF07F,
  [  456] = 0xF080, [  457] = 0xF081, [  458] = 0xF082, [  459] = 0xF083,
  [  460] = 0xF084, [  461] = 0xF085, [  462] = 0xF086, [  463] = 0xF087,
  [  464] = 0xF088, [  465] = 0xF089, [  466] = 0xF08A, [  467] = 0xF08B,
  [  468] = 0xF08C, [  469] = 0xF08D, [  470] = 0xF08E, [  471] = 0xF08F,
  [  472] = 0xF090, [  473] = 0xF091, [  474] = 0xF092, [  475] = 0xF093,
  [  476] = 0xF094, [  477] = 0xF095, [  478] = 0xF096, [  479] = 0xF097,
  [  480] = 0xF098, [  481] = 0xF099, [  482] = 0xF09A, [  483] = 0xF09B,
  [  484] = 0xF09C, [  485] = 0xF09D, [  486] = 0xF09E, [  487] = 0xF09F,
  [  488] = 0xF0A0, [  489] = 0xF0A1, [  490] = 0xF0A2, [  491] = 0xF0A3,
  [  492] = 0xF0A4, [  493] = 0xF0A5, [  494] = 0xF0A6, [  495] = 0xF0A7,
  [  496] = 0xF0A8, [  497] = 0xF0A9, [  498] = 0xF0AA, [  499] = 0xF0AB,
  [  500] = 0xF0AC, [  501] = 0xF0AD, [  502] = 0xF0AE, [  503] = 0xF0AF,
  [  504] = 0xF0B0, [  505] = 0xF0B1, [  506] = 0xF0B2, [  507] = 0xF0B3,
  [  508] = 0xF0B4, [  509] = 0xF0B5, [  510] = 0xF0B6, [  511] = 0xF0B7,
  [  512] = 0xF0B8, [  513] = 0xF0B9, [  514] = 0xF0BA, [  515] = 0xF0BB,
  [  516] = 0xF0BC, [  517] = 0xF0BD, [  518] = 0xF0BE, [  519] = 0xF0BF,
  [  520] = 0xF0C0, [  521] = 0xF0C1, [  522] = 0xF0C2, [  523] = 0xF0C3,
  [  524] = 0xF0C4, [  525] = 0xF0C5, [  526] = 0xF0C6, [  527] = 0xF0C7,
  [  528] = 0xF0C8, [  529] = 0xF0C9, [  530] = 0xF0CA, [  531] = 0xF0CB,
  [  532] = 0xF0CC, [  533] = 0xF0CD, [  534] = 0xF0CE, [  535] = 0xF0CF,
  [  536] = 0xF0D0, [  537] = 0xF0D1, [  538] = 0xF0D2, [  539] = 0xF0D3,
  [  540] = 0xF0D4, [  541] = 0xF0D5, [  542] = 0xF0D6, [  543] = 0xF0D7,
  [  544] = 0xF0D8, [  545] = 0xF0D9, [  546] = 0xF0DA, [  547] = 0xF0DB,
  [  548] = 0xF0DC, [  549] = 0xF0DD, [  550] = 0xF0DE, [  551] = 0xF0DF,
  [  552] = 0xF0E0, [  553] = 0xF0E1, [  554] = 0xF0E2, [  555] = 0xF0E3,
  [  556] = 0xF0E4, [  557] = 0xF0E5, [  558] = 0xF0E6, [  559] = 0xF0E7,
  [  560] = 0xF0E8, [  561] = 0xF0E9, [  562] = 0xF0EA, [  563] = 0xF0EB,
  [  564] = 0xF0EC, [  565] = 0xF0ED, [  566] = 0xF0EE, [  567] = 0xF0EF,
  [  568] = 0xF0F0, [  569] = 0xF0F1, [  570] = 0xF0F2, [  571] = 0xF0F3,
  [  572] = 0xF0F4, [  573] = 0xF0F5, [  574] = 0xF0F6, [  575] = 0xF0F7,
  [  576] = 0xF0F8, [  577] = 0xF0F9, [  578] = 0xF0FA, [  579] = 0xF0FB,
  [  580] = 0xF0FC, [  581] = 0xF0FD, [  582] = 0xF0FE, [  583] = 0xF0FF,
  [  584] = 0xF100, [  585] = 0xF101, [  586] = 0xF102, [  587] = 0xF103,
  [  588] = 0xF104, [  589] = 0xF105, [  590] = 0xF106, [  591] = 0xF107,
  [  592] = 0xF108, [  593] = 0xF109, [  594] = 0xF10A, [  595] = 0xF10B,
  [  596] = 0xF10C, [  597] = 0xF10D, [  598] = 0xF10E, [  599] = 0xF10F,
  [  600] = 0xF110, [  601] = 0xF111, [  602] = 0xF112, [  603] = 0xF113,
  [  604] = 0xF114, [  605] = 0xF115, [  606] = 0xF116, [  607] = 0xF117,
  [  608] = 0xF118, [  609] = 0xF119, [  610] = 0xF11A, [  611] = 0xF11B,
  [  612] = 0xF11C, [  613] = 0xF11D, [  614] = 0xF11E, [  615] = 0xF11F,
  [  616] = 0xF120, [  617] = 0xF121, [  618] = 0xF122, [  619] = 0xF123,
  [  620] = 0xF124, [  621] = 0xF125, [  622] = 0xF126, [  623] = 0xF127,
  [  624] = 0xF128, [  625] = 0xF129, [  626] = 0xF12A, [  627] = 0xF12B,
  [  628] = 0xF12C, [  629] = 0xF12D, [  630] = 0xF12E, [  631] = 0xF12F,
  [  632] = 0xF130, [  633] = 0xF131, [  634] = 0xF132, [  635] = 0xF133,
  [  636] = 0xF134, [  637] = 0xF135, [  638] = 0xF136, [  639] = 0xF137,
  [  640] = 0xF138, [  641] = 0xF139, [  642] = 0xF13A, [  643] = 0xF13B,
  [  644] = 0xF13C, [  645] = 0xF13D, [  646] = 0xF13E, [  647] = 0xF13F,
  [  648] = 0xF140, [  649] = 0xF141, [  650] = 0xF142, [  651] = 0xF143,
  [  652] = 0xF144, [  653] = 0xF145, [  654] = 0xF146, [  655] = 0xF147,
  [  656] = 0xF148, [  657] = 0xF149, [  658] = 0xF14A, [  659] = 0xF14B,
  [  660] = 0xF14C, [  661] = 0xF14D, [  662] = 0xF14E, [  663] = 0xF14F,
  [  664] = 0xF150, [  665] = 0xF151, [  666] = 0xF152, [  667] = 0xF153,
  [  668] = 0xF154, [  669] = 0xF155, [  670] = 0xF156, [  671] = 0xF157,
  [  672] = 0xF158, [  673] = 0xF159, [  674] = 0xF15A, [  675] = 0xF15B,
  [  676] = 0xF15C, [  677] = 0xF15D, [  678] = 0xF15E, [  679] = 0xF15F,
  [  680] = 0xF160, [  681] = 0xF161, [  682] = 0xF162, [  683] = 0xF163,
  [  684] = 0xF164, [  685] = 0xF165, [  686] = 0xF166, [  687] = 0xF167,
  [  688] = 0xF168, [  689] = 0xF169, [  690] = 0xF16A, [  691] = 0xF16B,
  [  692] = 0xF16C, [  693] = 0xF16D, [  694] = 0xF16E, [  695] = 0xF16F,
  [  696] = 0xF170, [  697] = 0xF171, [  698] = 0xF172, [  699] = 0xF173,
  [  700] = 0xF174, [  701] = 0xF175, [  702] = 0xF176, [  703] = 0xF177,
  [  704] = 0xF178, [  705] = 0xF179, [  706] = 0xF17A, [  707] = 0xF17B,
  [  708] = 0xF17C, [  709] = 0xF17D, [  710] = 0xF17E, [  711] = 0xF17F,
  [  712] = 0xF180, [  713] = 0xF181, [  714] = 0xF182, [  715] = 0xF183,
  [  716] = 0xF184, [  717] = 0xF185, [  718] = 0xF186, [  719] = 0xF187,
  [  720] = 0xF188, [  721] = 0xF189, [  722] = 0xF18A, [  723] = 0xF18B,
  [  724] = 0xF18C, [  725] = 0xF18D, [  726] = 0xF18E, [  727] = 0xF18F,
  [  728] = 0xF190, [  729] = 0xF191, [  730] = 0xF192, [  731] = 0xF193,
  [  732] = 0xF194, [  733] = 0xF195, [  734] = 0xF196, [  735] = 0xF197,
  [  736] = 0xF198, [  737] = 0xF199, [  738] = 0xF19A, [  739] = 0xF19B,
  [  740] = 0xF19C, [  741] = 0xF19D, [  742] = 0xF19E, [  743] = 0xF19F,
  [  744] = 0xF1A0, [  745] = 0xF1A1, [  746] = 0xF1A2, [  747] = 0xF1A3,
  [  748] = 0xF1A4, [  749] = 0xF1A5, [  750] = 0xF1A6, [  751] = 0xF1A7,
  [  752] = 0xF1A8, [  753] = 0xF1A9, [  754] = 0xF1AA, [  755] = 0xF1AB,
  [  756] = 0xF1AC, [  757] = 0xF1AD, [  758] = 0xF1AE, [  759] = 0xF1AF,
  [  760] = 0xF1B0, [  761] = 0xF1B1, [  762] = 0xF1B2, [  763] = 0xF1B3,
  [  764] = 0xF1B4, [  765] = 0xF1B5, [  766] = 0xF1B6, [  767] = 0xF1B7,
  [  768] = 0xF1B8, [  769] = 0xF1B9, [  770] = 0xF1BA, [  771] = 0xF1BB,
  [  772] = 0xF1BC, [  773] = 0xF1BD, [  774] = 0xF1BE, [  775] = 0xF1BF,
  [  776] = 0xF1C0, [  777] = 0xF1C1, [  778] = 0xF1C2, [  779] = 0xF1C3,
  [  780] = 0xF1C4, [  781] = 0xF1C5, [  782] = 0xF1C6, [  783] = 0xF1C7,
  [  784] = 0xF1C8, [  785] = 0xF1C9, [  786] = 0xF1CA, [  787] = 0xF1CB,
  [  788] = 0xF1CC, [  789] = 0xF1CD, [  790] = 0xF1CE, [  791] = 0xF1CF,
  [  792] = 0xF1D0, [  793] = 0xF1D1, [  794] = 0xF1D2, [  795] = 0xF1D3,
  [  796] = 0xF1D4, [  797] = 0xF1D5, [  798] = 0xF1D6, [  799] = 0xF1D7,
  [  800] = 0xF1D8, [  801] = 0xF1D9, [  802] = 0xF1DA, [  803] = 0xF1DB,
  [  804] = 0xF1DC, [  805] = 0xF1DD, [  806] = 0xF1DE, [  807] = 0xF1DF,
  [  808] = 0xF1E0, [  809] = 0xF1E1, [  810] = 0xF1E2, [  811] = 0xF1E3,
  [  812] = 0xF1E4, [  813] = 0xF1E5, [  814] = 0xF1E6, [  815] = 0xF1E7,
  [  816] = 0xF1E8, [  817] = 0xF1E9, [  818] = 0xF1EA, [  819] = 0xF1EB,
  [  820] = 0xF1EC, [  821] = 0xF1ED, [  822] = 0xF1EE, [  823] = 0xF1EF,
  [  824] = 0xF1F0, [  825] = 0xF1F1, [  826] = 0xF1F2, [  827] = 0xF1F3,
  [  828] = 0xF1F4, [  829] = 0xF1F5, [  830] = 0xF1F6, [  831] = 0xF1F7,
  [  832] = 0xF1F8, [  833] = 0xF1F9, [  834] = 0xF1FA, [  835] = 0xF1FB,
  [  836] = 0xF1FC, [  837] = 0xF1FD, [  838] = 0xF1FE, [  839] = 0xF1FF,
  [  840] = 0xF200, [  841] = 0xF201, [  842] = 0xF202, [  843] = 0xF203,
  [  844] = 0xF204, [  845] = 0xF205, [  846] = 0xF206, [  847] = 0xF207,
  [  848] = 0xF208, [  849] = 0xF209, [  850] = 0xF20A, [  851] = 0xF20B,
  [  852] = 0xF20C, [  853] = 0xF20D, [  854] = 0xF20E, [  855] = 0xF20F,
  [  856] = 0xF210, [  857] = 0xF211, [  858] = 0xF212, [  859] = 0xF213,
  [  860] = 0xF214, [  861] = 0xF215, [  862] = 0xF216, [  863] = 0xF217,
  [  864] = 0xF218, [  865] = 0xF219, [  866] = 0xF21A, [  867] = 0xF21B,
  [  868] = 0xF21C, [  869] = 0xF21D, [  870] = 0xF21E, [  871] = 0xF21F,
  [  872] = 0xF220, [  873] = 0xF221, [  874] = 0xF222, [  875] = 0xF223,
  [  876] = 0xF224, [  877] = 0xF225, [  878] = 0xF226, [  879] = 0xF227,
  [  880] = 0xF228, [  881] = 0xF229, [  882] = 0xF22A, [  883] = 0xF22B,
  [  884] = 0xF22C, [  885] = 0xF22D, [  886] = 0xF22E, [  887] = 0xF22F,
  [  888] = 0xF230, [  889] = 0xF231, [  890] = 0xF232, [  891] = 0xF233,
  [  892] = 0xF234, [  893] = 0xF235, [  894] = 0xF236, [  895] = 0xF237,
  [  896] = 0xF238, [  897] = 0xF239, [  898] = 0xF23A, [  899] = 0xF23B,
  [  900] = 0xF23C, [  901] = 0xF23D, [  902] = 0xF23E, [  903] = 0xF23F,
  [  904] = 0xF240, [  905] = 0xF241, [  906] = 0xF242, [  907] = 0xF243,
  [  908] = 0xF244, [  909] = 0xF245, [  910] = 0xF246, [  911] = 0xF247,
  [  912] = 0xF248, [  913] = 0xF249, [  914] = 0xF24A, [  915] = 0xF24B,
  [  916] = 0xF24C, [  917] = 0xF24D, [  918] = 0xF24E, [  919] = 0xF24F,
  [  920] = 0xF250, [  921] = 0xF251, [  922] = 0xF252, [  923] = 0xF253,
  [  924] = 0xF254, [  925] = 0xF255, [  926] = 0xF256, [  927] = 0xF257,
  [  928] = 0xF258, [  929] = 0xF259, [  930] = 0xF25A, [  931] = 0xF25B,
  [  932] = 0xF25C, [  933] = 0xF25D, [  934] = 0xF25E, [  935] = 0xF25F,
  [  936] = 0xF260, [  937] = 0xF261, [  938] = 0xF262, [  939] = 0xF263,
  [  940] = 0xF264, [  941] = 0xF265, [  942] = 0xF266, [  943] = 0xF267,
  [  944] = 0xF268, [  945] = 0xF269, [  946] = 0xF26A, [  947] = 0xF26B,
  [  948] = 0xF26C, [  949] = 0xF26D, [  950] = 0xF26E, [  951] = 0xF26F,
  [  952] = 0xF270, [  953] = 0xF271, [  954] = 0xF272, [  955] = 0xF273,
  [  956] = 0xF274, [  957] = 0xF275, [  958] = 0xF276, [  959] = 0xF277,
  [  960] = 0xF278, [  961] = 0xF279, [  962] = 0xF27A, [  963] = 0xF27B,
  [  964] = 0xF27C, [  965] = 0xF27D, [  966] = 0xF27E, [  967] = 0xF27F,
  [  968] = 0xF280, [  969] = 0xF281, [  970] = 0xF282, [  971] = 0xF283,
  [  972] = 0xF284, [  973] = 0xF285, [  974] = 0xF286, [  975] = 0xF287,
  [  976] = 0xF288, [  977] = 0xF289, [  978] = 0xF28A, [  979] = 0xF28B,
  [  980] = 0xF28C, [  981] = 0xF28D, [  982] = 0xF28E, [  983] = 0xF28F,
  [  984] = 0xF290, [  985] = 0xF291, [  986] = 0xF292, [  987] = 0xF293,
  [  988] = 0xF294, [  989] = 0xF295, [  990] = 0xF296, [  991] = 0xF297,
  [  992] = 0xF298, [  993] = 0xF299, [  994] = 0xF29A, [  995] = 0xF29B,
  [  996] = 0xF29C, [  997] = 0xF29D, [  998] = 0xF29E, [  999] = 0xF29F,
  [ 1000] = 0xF2A0, [ 1001] = 0xF2A1, [ 1002] = 0xF2A2, [ 1003] = 0xF2A3,
  [ 1004] = 0xF2A4, [ 1005] = 0xF2A5, [ 1006] = 0xF2A6, [ 1007] = 0xF2A7,
  [ 1008] = 0xF2A8, [ 1009] = 0xF2A9, [ 1010] = 0xF2AA, [ 1011] = 0xF2AB,
  [ 1012] = 0xF2AC, [ 1013] = 0xF2AD, [ 1014] = 0xF2AE, [ 1015] = 0xF2AF,
  [ 1016] = 0xF2B0, [ 1017] = 0xF2B1, [ 1018] = 0xF2B2, [ 1019] = 0xF2B3,
  [ 1020] = 0xF2B4, [ 1021] = 0xF2B5, [ 1022] = 0xF2B6, [ 1023] = 0xF2B7,
  [ 1024] = 0xF2B8, [ 1025] = 0xF2B9, [ 1026] = 0xF2BA, [ 1027] = 0xF2BB,
  [ 1028] = 0xF2BC, [ 1029] = 0xF2BD, [ 1030] = 0xF2BE, [ 1031] = 0xF2BF,
  [ 1032] = 0xF2C0, [ 1033] = 0xF2C1, [ 1034] = 0xF2C2, [ 1035] = 0xF2C3,
  [ 1036] = 0xF2C4, [ 1037] = 0xF2C5, [ 1038] = 0xF2C6, [ 1039] = 0xF2C7,
  [ 1040] = 0xF2C8, [ 1041] = 0xF2C9, [ 1042] = 0xF2CA, [ 1043] = 0xF2CB,
  [ 1044] = 0xF2CC, [ 1045] = 0xF2CD, [ 1046] = 0xF2CE, [ 1047] = 0xF2CF,
  [ 1048] = 0xF2D0, [ 1049] = 0xF2D1, [ 1050] = 0xF2D2, [ 1051] = 0xF2D3,
  [ 1052] = 0xF2D4, [ 1053] = 0xF2D5, [ 1054] = 0xF2D6, [ 1055] = 0xF2D7,
  [ 1056] = 0xF2D8, [ 1057] = 0xF2D9, [ 1058] = 0xF2DA, [ 1059] = 0xF2DB,
  [ 1060] = 0xF2DC, [ 1061] = 0xF2DD, [ 1062] = 0xF2DE, [ 1063] = 0xF2DF,
  [ 1064] = 0xF2E0, [ 1065] = 0xF2E1, [ 1066] = 0xF2E2, [ 1067] = 0xF2E3,
  [ 1068] = 0xF2E4, [ 1069] = 0xF2E5, [ 1070] = 0xF2E6, [ 1071] = 0xF2E7,
  [ 1072] = 0xF2E8, [ 1073] = 0xF2E9, [ 1074] = 0xF2EA, [ 1075] = 0xF2EB,
  [ 1076] = 0xF2EC, [ 1077] = 0xF2ED, [ 1078] = 0xF2EE, [ 1079] = 0xF2EF,
  [ 1080] = 0xF2F0, [ 1081] = 0xF2F1, [ 1082] = 0xF2F2, [ 1083] = 0xF2F3,
  [ 1084] = 0xF2F4, [ 1085] = 0xF2F5, [ 1086] = 0xF2F6, [ 1087] = 0xF2F7,
  [ 1088] = 0xF2F8, [ 1089] = 0xF2F9, [ 1090] = 0xF2FA, [ 1091] = 0xF2FB,
  [ 1092] = 0xF2FC, [ 1093] = 0xF2FD, [ 1094] = 0xF2FE, [ 1095] = 0xF2FF,
  [ 1096] = 0xF300, [ 1097] = 0xF301, [ 1098] = 0xF302, [ 1099] = 0xF303,
  [ 1100] = 0xF304, [ 1101] = 0xF305, [ 1102] = 0xF306, [ 1103] = 0xF307,
  [ 1104] = 0xF308, [ 1105] = 0xF309, [ 1106] = 0xF30A, [ 1107] = 0xF30B,
  [ 1108] = 0xF30C, [ 1109] = 0xF30D, [ 1110] = 0xF30E, [ 1111] = 0xF30F,
  [ 1112] = 0xF310, [ 1113] = 0xF311, [ 1114] = 0xF312, [ 1115] = 0xF313,
  [ 1116] = 0xF314, [ 1117] = 0xF315, [ 1118] = 0xF316, [ 1119] = 0xF317,
  [ 1120] = 0xF318, [ 1121] = 0x0100, [ 1122] = 0x00C1, [ 1123] = 0x01CD,
  [ 1124] = 0x00C0, [ 1125] = 0x0112, [ 1126] = 0x00C9, [ 1127] = 0x011A,
  [ 1128] = 0x00C8, [ 1129] = 0x014C, [ 1130] = 0x00D3, [ 1131] = 0x01D1,
  [ 1132] = 0x00D2, [ 1133] = 0xF325, [ 1134] = 0x1EBE, [ 1135] = 0xF327,
  [ 1136] = 0x1EC0, [ 1137] = 0x00CA, [ 1138] = 0x0101, [ 1139] = 0x00E1,
  [ 1140] = 0x01CE, [ 1141] = 0x00E0, [ 1142] = 0x0251, [ 1143] = 0x0113,
  [ 1144] = 0x00E9, [ 1145] = 0x011B, [ 1146] = 0x00E8, [ 1147] = 0x012B,
  [ 1148] = 0x00ED, [ 1149] = 0x01D0, [ 1150] = 0x00EC, [ 1151] = 0x014D,
  [ 1152] = 0x00F3, [ 1153] = 0x01D2, [ 1154] = 0x00F2, [ 1155] = 0x016B,
  [ 1156] = 0x00FA, [ 1157] = 0x01D4, [ 1158] = 0x00F9, [ 1159] = 0x01D6,
  [ 1160] = 0x01D8, [ 1161] = 0x01DA, [ 1162] = 0x01DC, [ 1163] = 0x00FC,
  [ 1164] = 0xF344, [ 1165] = 0x1EBF, [ 1166] = 0xF346, [ 1167] = 0x1EC1,
  [ 1168] = 0x00EA, [ 1169] = 0x0261, [ 1170] = 0xF34A, [ 1171] = 0xF34B,
  [ 1172] = 0xF34C, [ 1173] = 0xF34D, [ 1174] = 0xF34E, [ 1175] = 0xF34F,
  [ 1176] = 0xF350, [ 1177] = 0xF351, [ 1178] = 0xF352, [ 1179] = 0xF353,
  [ 1180] = 0xF354, [ 1181] = 0xF355, [ 1182] = 0xF356, [ 1183] = 0xF357,
  [ 1184] = 0xF358, [ 1185] = 0xF359, [ 1186] = 0xF35A, [ 1187] = 0xF35B,
  [ 1188] = 0xF35C, [ 1189] = 0xF35D, [ 1190] = 0xF35E, [ 1191] = 0xF35F,
  [ 1192] = 0xF360, [ 1193] = 0xF361, [ 1194] = 0xF362, [ 1195] = 0xF363,
  [ 1196] = 0xF364, [ 1197] = 0xF365, [ 1198] = 0xF366, [ 1199] = 0xF367,
  [ 1200] = 0xF368, [ 1201] = 0xF369, [ 1202] = 0xF36A, [ 1203] = 0xF36B,
  [ 1204] = 0xF36C, [ 1205] = 0xF36D, [ 1206] = 0xF36E, [ 1207] = 0xF36F,
  [ 1208] = 0xF370, [ 1209] = 0xF371, [ 1210] = 0xF372, [ 1211] = 0xF373,
  [ 1212] = 0xF374, [ 1213] = 0xF375, [ 1214] = 0xF376, [ 1215] = 0xF377,
  [ 1216] = 0xF378, [ 1217] = 0xF379, [ 1218] = 0xF37A, [ 1219] = 0xF37B,
  [ 1220] = 0xF37C, [ 1221] = 0xF37D, [ 1222] = 0xF37E, [ 1223] = 0xF37F,
  [ 1224] = 0xF380, [ 1225] = 0xF381, [ 1226] = 0xF382, [ 1227] = 0xF383,
  [ 1228] = 0xF384, [ 1229] = 0xF385, [ 1230] = 0xF386, [ 1231] = 0xF387,
  [ 1232] = 0xF388, [ 1233] = 0xF389, [ 1234] = 0xF38A, [ 1235] = 0xF38B,
  [ 1236] = 0xF38C, [ 1237] = 0xF38D, [ 1238] = 0xF38E, [ 1239] = 0xF38F,
  [ 1240] = 0xF390, [ 1241] = 0xF391, [ 1242] = 0xF392, [ 1243] = 0xF393,
  [ 1244] = 0xF394, [ 1245] = 0xF395, [ 1246] = 0xF396, [ 1247] = 0xF397,
  [ 1248] = 0xF398, [ 1249] = 0xF399, [ 1250] = 0xF39A, [ 1251] = 0xF39B,
  [ 1252] = 0xF39C, [ 1253] = 0xF39D, [ 1254] = 0xF39E, [ 1255] = 0xF39F,
  [ 1256] = 0xF3A0, [ 1257] = 0xF3A1, [ 1258] = 0xF3A2, [ 1259] = 0x650A,
  [ 1260] = 0xF3A4, [ 1261] = 0xF3A5, [ 1262] = 0x4E3D, [ 1263] = 0x6EDD,
  [ 1264] = 0x9D4E, [ 1265] = 0x91DF, [ 1266] = 0xF3AA, [ 1267] = 0xF3AB,
  [ 1268] = 0xF3AC, [ 1269] = 0x6491, [ 1270] = 0x4F1A, [ 1271] = 0x4F28,
  [ 1272] = 0x4FA8, [ 1273] = 0x5156, [ 1274] = 0x5174, [ 1275] = 0x519C,
  [ 1276] = 0x51E4, [ 1277] = 0x52A1, [ 1278] = 0x52A8, [ 1279] = 0x533B,
  [ 1280] = 0x534E, [ 1281] = 0x53D1, [ 1282] = 0x53D8, [ 1283] = 0x56E2,
  [ 1284] = 0x58F0, [ 1285] = 0x5904, [ 1286] = 0x5907, [ 1287] = 0x5932,
  [ 1288] = 0x5934, [ 1289] = 0x5B66, [ 1290] = 0x5B9E, [ 1291] = 0x5B9F,
  [ 1292] = 0x5C9A, [ 1293] = 0x5E86, [ 1294] = 0x603B, [ 1295] = 0x6589,
  [ 1296] = 0x67FE, [ 1297] = 0x6804, [ 1298] = 0x6865, [ 1299] = 0x6D4E,
  [ 1300] = 0x70BC, [ 1301] = 0x7535, [ 1302] = 0x7EA4, [ 1303] = 0x7EAC,
  [ 1304] = 0x7EBA, [ 1305] = 0x7EC7, [ 1306] = 0x7ECF, [ 1307] = 0x7EDF,
  [ 1308] = 0x7F06, [ 1309] = 0x7F37, [ 1310] = 0x827A, [ 1311] = 0x82CF,
  [ 1312] = 0x836F, [ 1313] = 0x89C6, [ 1314] = 0x8BBE, [ 1315] = 0x8BE2,
  [ 1316] = 0x8F66, [ 1317] = 0x8F67, [ 1318] = 0x8F6E, [ 1319] = 0x7411,
  [ 1320] = 0x7CFC, [ 1321] = 0x7DCD, [ 1322] = 0x6946, [ 1323] = 0x7AC9,
  [ 1324] = 0x5227, [ 1325] = 0xF3E5, [ 1326] = 0xF3E6, [ 1327] = 0xF3E7,
  [ 1328] = 0xF3E8, [ 1329] = 0x918C, [ 1330] = 0x78B8, [ 1331] = 0x915E,
  [ 1332] = 0x80BC, [ 1333] = 0xF3ED, [ 1334] = 0x8D0B, [ 1335] = 0x80F6,
  [ 1336] = 0xF3F0, [ 1337] = 0xF3F1, [ 1338] = 0xF3F2, [ 1339] = 0x809F,
  [ 1340] = 0x9EC7, [ 1341] = 0xF3F5, [ 1342] = 0x9DC9, [ 1343] = 0x9E0C,
  [ 1344] = 0xF3F8, [ 1345] = 0xF3F9, [ 1346] = 0xF3FA, [ 1347] = 0x9E0A,
  [ 1348] = 0xF3FC, [ 1349] = 0xF3FD, [ 1350] = 0xF3FE, [ 1351] = 0x6E9A,
  [ 1352] = 0x823E, [ 1353] = 0x7519, [ 1354] = 0xF402, [ 1355] = 0xF403,
  [ 1356] = 0x9A6C, [ 1357] = 0x9A8F, [ 1358] = 0x9F99, [ 1359] = 0x7987,
  [ 1360] = 0xF408, [ 1361] = 0xF409, [ 1362] = 0xF40A, [ 1363] = 0xF40B,
  [ 1364] = 0x4E24, [ 1365] = 0x4E81, [ 1366] = 0x4E80, [ 1367] = 0x4E87,
  [ 1368] = 0x4EBF, [ 1369] = 0x4EEB, [ 1370] = 0x4F37, [ 1371] = 0xF413,
  [ 1372] = 0x4FBD, [ 1373] = 0xF415, [ 1374] = 0x5003, [ 1375] = 0x5088,
  [ 1376] = 0xF418, [ 1377] = 0xF419, [ 1378] = 0xF41A, [ 1379] = 0x5186,
  [ 1380] = 0x5905, [ 1381] = 0x51DB, [ 1382] = 0x51FC, [ 1383] = 0x5205,
  [ 1384] = 0x4E89, [ 1385] = 0x5279, [ 1386] = 0x5290, [ 1387] = 0x5327,
  [ 1388] = 0xF424, [ 1389] = 0x53A9, [ 1390] = 0xF426, [ 1391] = 0x53B0,
  [ 1392] = 0xF428, [ 1393] = 0x53C2, [ 1394] = 0x5423, [ 1395] = 0xF42B,
  [ 1396] = 0xF42C, [ 1397] = 0xF42D, [ 1398] = 0x5493, [ 1399] = 0x54A3,
  [ 1400] = 0x54B4, [ 1401] = 0x54B9, [ 1402] = 0x54D0, [ 1403] = 0x54EF,
  [ 1404] = 0x5518, [ 1405] = 0x5523, [ 1406] = 0x5528, [ 1407] = 0xF437,
  [ 1408] = 0x553F, [ 1409] = 0xF439, [ 1410] = 0xF43A, [ 1411] = 0x55D7,
  [ 1412] = 0xF43C, [ 1413] = 0xF43D, [ 1414] = 0x5525, [ 1415] = 0xF43F,
  [ 1416] = 0xF440, [ 1417] = 0xF441, [ 1418] = 0xF442, [ 1419] = 0x5590,
  [ 1420] = 0xF444, [ 1421] = 0xF445, [ 1422] = 0xF446, [ 1423] = 0x8E46,
  [ 1424] = 0xF448, [ 1425] = 0xF449, [ 1426] = 0xF44A, [ 1427] = 0xF44B,
  [ 1428] = 0x777A, [ 1429] = 0xF44D, [ 1430] = 0xF44E, [ 1431] = 0xF44F,
  [ 1432] = 0xF450, [ 1433] = 0xF451, [ 1434] = 0xF452, [ 1435] = 0x64DD,
  [ 1436] = 0xF454, [ 1437] = 0xF455, [ 1438] = 0xF456, [ 1439] = 0xF457,
  [ 1440] = 0x648D, [ 1441] = 0x8E7E, [ 1442] = 0xF45A, [ 1443] = 0xF45B,
  [ 1444] = 0xF45C, [ 1445] = 0xF45D, [ 1446] = 0xF45E, [ 1447] = 0xF45F,
  [ 1448] = 0xF460, [ 1449] = 0xF461, [ 1450] = 0xF462, [ 1451] = 0xF463,
  [ 1452] = 0xF464, [ 1453] = 0xF465, [ 1454] = 0xF466, [ 1455] = 0x9AB2,
  [ 1456] = 0xF468, [ 1457] = 0xF469, [ 1458] = 0xF46A, [ 1459] = 0xF46B,
  [ 1460] = 0xF46C, [ 1461] = 0xF46D, [ 1462] = 0xF46E, [ 1463] = 0xF46F,
  [ 1464] = 0x9D6E, [ 1465] = 0x9815, [ 1466] = 0xF472, [ 1467] = 0xF473,
  [ 1468] = 0xF474, [ 1469] = 0x64B4, [ 1470] = 0x54E3, [ 1471] = 0xF477,
  [ 1472] = 0xF478, [ 1473] = 0xF479, [ 1474] = 0xF47A, [ 1475] = 0xF47B,
  [ 1476] = 0xF47C, [ 1477] = 0xF47D, [ 1478] = 0xF47E, [ 1479] = 0x64EA,
  [ 1480] = 0xF480, [ 1481] = 0xF481, [ 1482] = 0x8E68, [ 1483] = 0xF483,
  [ 1484] = 0xF484, [ 1485] = 0xF485, [ 1486] = 0xF486, [ 1487] = 0xF487,
  [ 1488] = 0xF488, [ 1489] = 0xF489, [ 1490] = 0x5873, [ 1491] = 0xF48B,
  [ 1492] = 0xF48C, [ 1493] = 0xF48D, [ 1494] = 0xF48E, [ 1495] = 0xF48F,
  [ 1496] = 0xF490, [ 1497] = 0xF491, [ 1498] = 0x5579, [ 1499] = 0xF493,
  [ 1500] = 0xF494, [ 1501] = 0xF495, [ 1502] = 0xF496, [ 1503] = 0xF497,
  [ 1504] = 0xF498, [ 1505] = 0x81AA, [ 1506] = 0x98F5, [ 1507] = 0xF49B,
  [ 1508] = 0x6379, [ 1509] = 0xF49D, [ 1510] = 0xF49E, [ 1511] = 0x8DC0,
  [ 1512] = 0x56A1, [ 1513] = 0x647C, [ 1514] = 0xF4A2, [ 1515] = 0xF4A3,
  [ 1516] = 0xF4A4, [ 1517] = 0xF4A5, [ 1518] = 0xF4A6, [ 1519] = 0xF4A7,
  [ 1520] = 0xF4A8, [ 1521] = 0xF4A9, [ 1522] = 0xF4AA, [ 1523] = 0xF4AB,
  [ 1524] = 0xF4AC, [ 1525] = 0xF4AD, [ 1526] = 0xF4AE, [ 1527] = 0xF4AF,
  [ 1528] = 0xF4B0, [ 1529] = 0x5652, [ 1530] = 0xF4B2, [ 1531] = 0xF4B3,
  [ 1532] = 0xF4B4, [ 1533] = 0xF4B5, [ 1534] = 0x6C3D, [ 1535] = 0xF4B7,
  [ 1536] = 0xF4B8, [ 1537] = 0xF4B9, [ 1538] = 0xF4BA, [ 1539] = 0xF4BB,
  [ 1540] = 0xF4BC, [ 1541] = 0xF4BD, [ 1542] = 0xF4BE, [ 1543] = 0xF4BF,
  [ 1544] = 0xF4C0, [ 1545] = 0xF4C1, [ 1546] = 0x7F93, [ 1547] = 0xF4C3,
  [ 1548] = 0xF4C4, [ 1549] = 0xF4C5, [ 1550] = 0xF4C6, [ 1551] = 0xF4C7,
  [ 1552] = 0xF4C8, [ 1553] = 0xF4C9, [ 1554] = 0xF4CA, [ 1555] = 0xF4CB,
  [ 1556] = 0xF4CC, [ 1557] = 0xF4CD, [ 1558] = 0xF4CE, [ 1559] = 0xF4CF,
  [ 1560] = 0xF4D0, [ 1561] = 0xF4D1, [ 1562] = 0xF4D2, [ 1563] = 0xF4D3,
  [ 1564] = 0xF4D4, [ 1565] = 0xF4D5, [ 1566] = 0xF4D6, [ 1567] = 0xF4D7,
  [ 1568] = 0x6432, [ 1569] = 0xF4D9, [ 1570] = 0xF4DA, [ 1571] = 0xF4DB,
  [ 1572] = 0xF4DC, [ 1573] = 0xF4DD, [ 1574] = 0xF4DE, [ 1575] = 0xF4DF,
  [ 1576] = 0xF4E0, [ 1577] = 0xF4E1, [ 1578] = 0xF4E2, [ 1579] = 0xF4E3,
  [ 1580] = 0xF4E4, [ 1581] = 0xF4E5, [ 1582] = 0xF4E6, [ 1583] = 0xF4E7,
  [ 1584] = 0xF4E8, [ 1585] = 0xF4E9, [ 1586] = 0x95AA, [ 1587] = 0x54CC,
  [ 1588] = 0x82C4, [ 1589] = 0x55B9, [ 1590] = 0xF4EE, [ 1591] = 0xF4EF,
  [ 1592] = 0x9C26, [ 1593] = 0x9AB6, [ 1594] = 0xF4F2, [ 1595] = 0xF4F3,
  [ 1596] = 0x7140, [ 1597] = 0x816D, [ 1598] = 0x80EC, [ 1599] = 0x5C1C,
  [ 1600] = 0xF4F8, [ 1601] = 0x8134, [ 1602] = 0xF4FA, [ 1603] = 0x535F,
  [ 1604] = 0xF4FC, [ 1605] = 0x91B6, [ 1606] = 0xF4FE, [ 1607] = 0xF4FF,
  [ 1608] = 0xF500, [ 1609] = 0xF501, [ 1610] = 0xF502, [ 1611] = 0xF503,
  [ 1612] = 0xF504, [ 1613] = 0xF505, [ 1614] = 0x56AF, [ 1615] = 0xF507,
  [ 1616] = 0xF508, [ 1617] = 0xF509, [ 1618] = 0xF50A, [ 1619] = 0xF50B,
  [ 1620] = 0xF50C, [ 1621] = 0xF50D, [ 1622] = 0xF50E, [ 1623] = 0xF50F,
  [ 1624] = 0xF510, [ 1625] = 0xF511, [ 1626] = 0x5A54, [ 1627] = 0xF513,
  [ 1628] = 0xF514, [ 1629] = 0xF515, [ 1630] = 0xF516, [ 1631] = 0x579C,
  [ 1632] = 0xF518, [ 1633] = 0xF519, [ 1634] = 0xF51A, [ 1635] = 0xF51B,
  [ 1636] = 0xF51C, [ 1637] = 0xF51D, [ 1638] = 0xF51E, [ 1639] = 0xF51F,
  [ 1640] = 0xF520, [ 1641] = 0xF521, [ 1642] = 0xF522, [ 1643] = 0xF523,
  [ 1644] = 0xF524, [ 1645] = 0xF525, [ 1646] = 0xF526, [ 1647] = 0x5899,
  [ 1648] = 0x5268, [ 1649] = 0xF529, [ 1650] = 0xF52A, [ 1651] = 0x7BB2,
  [ 1652] = 0x5B68, [ 1653] = 0xF52D, [ 1654] = 0xF52E, [ 1655] = 0x9F27,
  [ 1656] = 0xF530, [ 1657] = 0x9C1F, [ 1658] = 0x9B8D, [ 1659] = 0xF533,
  [ 1660] = 0xF534, [ 1661] = 0x55FB, [ 1662] = 0xF536, [ 1663] = 0x5689,
  [ 1664] = 0x4E28, [ 1665] = 0x5902, [ 1666] = 0xF53A, [ 1667] = 0xF53B,
  [ 1668] = 0x9751, [ 1669] = 0xF53D, [ 1670] = 0x4E5B, [ 1671] = 0x4EBB,
  [ 1672] = 0xF540, [ 1673] = 0x5C23, [ 1674] = 0x5F51, [ 1675] = 0x5FC4,
  [ 1676] = 0xF544, [ 1677] = 0x624C, [ 1678] = 0x6535, [ 1679] = 0x6B7A,
  [ 1680] = 0x6C35, [ 1681] = 0x6C3A, [ 1682] = 0x706C, [ 1683] = 0x722B,
  [ 1684] = 0x4E2C, [ 1685] = 0x72AD, [ 1686] = 0xF54E, [ 1687] = 0x7F52,
  [ 1688] = 0x793B, [ 1689] = 0x7CF9, [ 1690] = 0x7F53, [ 1691] = 0xF553,
  [ 1692] = 0xF554, [ 1693] = 0xF555, [ 1694] = 0xF556, [ 1695] = 0x8002,
  [ 1696] = 0x8080, [ 1697] = 0xF559, [ 1698] = 0xF55A, [ 1699] = 0x535D,
  [ 1700] = 0x8864, [ 1701] = 0x89C1, [ 1702] = 0xF55E, [ 1703] = 0x8BA0,
  [ 1704] = 0x8D1D, [ 1705] = 0x9485, [ 1706] = 0x9578, [ 1707] = 0x957F,
  [ 1708] = 0x95E8, [ 1709] = 0xF565, [ 1710] = 0x97E6, [ 1711] = 0x9875,
  [ 1712] = 0x98CE, [ 1713] = 0x98DE, [ 1714] = 0x9963, [ 1715] = 0xF56B,
  [ 1716] = 0x9C7C, [ 1717] = 0x9E1F, [ 1718] = 0x9EC4, [ 1719] = 0x6B6F,
  [ 1720] = 0x9F9C, [ 1721] = 0x4E37, [ 1722] = 0xF572, [ 1723] = 0x961D,
  [ 1724] = 0x6237, [ 1725] = 0x94A2, [ 1726] = 0xF576, [ 1727] = 0xF577,
  [ 1728] = 0xF578, [ 1729] = 0xF579, [ 1730] = 0xF57A, [ 1731] = 0xF57B,
  [ 1732] = 0xF57C, [ 1733] = 0xF57D, [ 1734] = 0xF57E, [ 1735] = 0xF57F,
  [ 1736] = 0xF580, [ 1737] = 0xF581, [ 1738] = 0xF582, [ 1739] = 0xF583,
  [ 1740] = 0xF584, [ 1741] = 0xF585, [ 1742] = 0xF586, [ 1743] = 0xF587,
  [ 1744] = 0xF588, [ 1745] = 0xF589, [ 1746] = 0xF58A, [ 1747] = 0xF58B,
  [ 1748] = 0xF58C, [ 1749] = 0xF58D, [ 1750] = 0xF58E, [ 1751] = 0xF58F,
  [ 1752] = 0xF590, [ 1753] = 0xF591, [ 1754] = 0xF592, [ 1755] = 0xF593,
  [ 1756] = 0xF594, [ 1757] = 0xF595, [ 1758] = 0xF596, [ 1759] = 0xF597,
  [ 1760] = 0xF598, [ 1761] = 0xF599, [ 1762] = 0xF59A, [ 1763] = 0xF59B,
  [ 1764] = 0xF59C, [ 1765] = 0xF59D, [ 1766] = 0xF59E, [ 1767] = 0xF59F,
  [ 1768] = 0xF5A0, [ 1769] = 0xF5A1, [ 1770] = 0xF5A2, [ 1771] = 0xF5A3,
  [ 1772] = 0xF5A4, [ 1773] = 0xF5A5, [ 1774] = 0xF5A6, [ 1775] = 0xF5A7,
  [ 1776] = 0xF5A8, [ 1777] = 0xF5A9, [ 1778] = 0xF5AA, [ 1779] = 0xF5AB,
  [ 1780] = 0xF5AC, [ 1781] = 0xF5AD, [ 1782] = 0xF5AE, [ 1783] = 0xF5AF,
  [ 1784] = 0xF5B0, [ 1785] = 0xF5B1, [ 1786] = 0xF5B2, [ 1787] = 0xF5B3,
  [ 1788] = 0xF5B4, [ 1789] = 0xF5B5, [ 1790] = 0xF5B6, [ 1791] = 0xF5B7,
  [ 1792] = 0xF5B8, [ 1793] = 0xF5B9, [ 1794] = 0xF5BA, [ 1795] = 0xF5BB,
  [ 1796] = 0xF5BC, [ 1797] = 0xF5BD, [ 1798] = 0xF5BE, [ 1799] = 0xF5BF,
  [ 1800] = 0xF5C0, [ 1801] = 0xF5C1, [ 1802] = 0xF5C2, [ 1803] = 0xF5C3,
  [ 1804] = 0xF5C4, [ 1805] = 0xF5C5, [ 1806] = 0xF5C6, [ 1807] = 0xF5C7,
  [ 1808] = 0xF5C8, [ 1809] = 0xF5C9, [ 1810] = 0xF5CA, [ 1811] = 0xF5CB,
  [ 1812] = 0xF5CC, [ 1813] = 0xF5CD, [ 1814] = 0xF5CE, [ 1815] = 0xF5CF,
  [ 1816] = 0xF5D0, [ 1817] = 0xF5D1, [ 1818] = 0xF5D2, [ 1819] = 0xF5D3,
  [ 1820] = 0xF5D4, [ 1821] = 0xF5D5, [ 1822] = 0xF5D6, [ 1823] = 0xF5D7,
  [ 1824] = 0xF5D8, [ 1825] = 0xF5D9, [ 1826] = 0xF5DA, [ 1827] = 0xF5DB,
  [ 1828] = 0xF5DC, [ 1829] = 0xF5DD, [ 1830] = 0xF5DE, [ 1831] = 0xF5DF,
  [ 1832] = 0xF5E0, [ 1833] = 0xF5E1, [ 1834] = 0xF5E2, [ 1835] = 0xF5E3,
  [ 1836] = 0xF5E4, [ 1837] = 0xF5E5, [ 1838] = 0xF5E6, [ 1839] = 0xF5E7,
  [ 1840] = 0xF5E8, [ 1841] = 0xF5E9, [ 1842] = 0xF5EA, [ 1843] = 0xF5EB,
  [ 1844] = 0xF5EC, [ 1845] = 0xF5ED, [ 1846] = 0xF5EE, [ 1847] = 0xF5EF,
  [ 1848] = 0xF5F0, [ 1849] = 0xF5F1, [ 1850] = 0xF5F2, [ 1851] = 0xF5F3,
  [ 1852] = 0xF5F4, [ 1853] = 0xF5F5, [ 1854] = 0xF5F6, [ 1855] = 0xF5F7,
  [ 1856] = 0xF5F8, [ 1857] = 0xF5F9, [ 1858] = 0xF5FA, [ 1859] = 0xF5FB,
  [ 1860] = 0xF5FC, [ 1861] = 0xF5FD, [ 1862] = 0xF5FE, [ 1863] = 0xF5FF,
  [ 1864] = 0xF600, [ 1865] = 0xF601, [ 1866] = 0xF602, [ 1867] = 0xF603,
  [ 1868] = 0xF604, [ 1869] = 0xF605, [ 1870] = 0xF606, [ 1871] = 0xF607,
  [ 1872] = 0xF608, [ 1873] = 0xF609, [ 1874] = 0xF60A, [ 1875] = 0xF60B,
  [ 1876] = 0xF60C, [ 1877] = 0xF60D, [ 1878] = 0xF60E, [ 1879] = 0xF60F,
  [ 1880] = 0xF610, [ 1881] = 0xF611, [ 1882] = 0xF612, [ 1883] = 0xF613,
  [ 1884] = 0xF614, [ 1885] = 0xF615, [ 1886] = 0xF616, [ 1887] = 0xF617,
  [ 1888] = 0xF618, [ 1889] = 0xF619, [ 1890] = 0xF61A, [ 1891] = 0xF61B,
  [ 1892] = 0xF61C, [ 1893] = 0xF61D, [ 1894] = 0xF61E, [ 1895] = 0xF61F,
  [ 1896] = 0xF620, [ 1897] = 0xF621, [ 1898] = 0xF622, [ 1899] = 0xF623,
  [ 1900] = 0xF624, [ 1901] = 0xF625, [ 1902] = 0xF626, [ 1903] = 0xF627,
  [ 1904] = 0xF628, [ 1905] = 0xF629, [ 1906] = 0xF62A, [ 1907] = 0xF62B,
  [ 1908] = 0xF62C, [ 1909] = 0xF62D, [ 1910] = 0xF62E, [ 1911] = 0xF62F,
  [ 1912] = 0xF630, [ 1913] = 0xF631, [ 1914] = 0xF632, [ 1915] = 0xF633,
  [ 1916] = 0x5D3E, [ 1917] = 0x5D48, [ 1918] = 0x5D56, [ 1919] = 0xF637,
  [ 1920] = 0xF638, [ 1921] = 0x5DA4, [ 1922] = 0x5DB9, [ 1923] = 0xF63B,
  [ 1924] = 0xF63C, [ 1925] = 0x5E42, [ 1926] = 0x5EBD, [ 1927] = 0x5F25,
  [ 1928] = 0x5F83, [ 1929] = 0xF641, [ 1930] = 0xF642, [ 1931] = 0xF643,
  [ 1932] = 0xF644, [ 1933] = 0x60D7, [ 1934] = 0x613D, [ 1935] = 0x5CE5,
  [ 1936] = 0xF648, [ 1937] = 0x61B7, [ 1938] = 0x61B9, [ 1939] = 0x61CF,
  [ 1940] = 0xF64C, [ 1941] = 0x622C, [ 1942] = 0x6290, [ 1943] = 0x62E5,
  [ 1944] = 0x6318, [ 1945] = 0xF651, [ 1946] = 0x56B1, [ 1947] = 0xF653,
  [ 1948] = 0x63E2, [ 1949] = 0x63FB, [ 1950] = 0x6407, [ 1951] = 0x645A,
  [ 1952] = 0xF658, [ 1953] = 0x64C0, [ 1954] = 0x5D15, [ 1955] = 0x5621,
  [ 1956] = 0x9F9F, [ 1957] = 0xF65D, [ 1958] = 0x6586, [ 1959] = 0xF65F,
  [ 1960] = 0x65FF, [ 1961] = 0x6653, [ 1962] = 0xF662, [ 1963] = 0x6692,
  [ 1964] = 0xF664, [ 1965] = 0x6716, [ 1966] = 0xF666, [ 1967] = 0x67A4,
  [ 1968] = 0x6800, [ 1969] = 0xF669, [ 1970] = 0x684A, [ 1971] = 0x6884,
  [ 1972] = 0xF66C, [ 1973] = 0xF66D, [ 1974] = 0xF66E, [ 1975] = 0x6909,
  [ 1976] = 0x6943, [ 1977] = 0x725C, [ 1978] = 0x6964, [ 1979] = 0x699F,
  [ 1980] = 0x6985, [ 1981] = 0xF675, [ 1982] = 0x69D6, [ 1983] = 0xF677,
  [ 1984] = 0x6A65, [ 1985] = 0x6A74, [ 1986] = 0x6A71, [ 1987] = 0x6A82,
  [ 1988] = 0xF67C, [ 1989] = 0x6A99, [ 1990] = 0xF67E, [ 1991] = 0x6AAB,
  [ 1992] = 0x6AB5, [ 1993] = 0x6AD4, [ 1994] = 0x6AF6, [ 1995] = 0x6B81,
  [ 1996] = 0x6BC1, [ 1997] = 0x6BEA, [ 1998] = 0x6C75, [ 1999] = 0x6CAA,
  [ 2000] = 0xF688, [ 2001] = 0x6D02, [ 2002] = 0x6D06, [ 2003] = 0x6D26,
  [ 2004] = 0x6D81, [ 2005] = 0xF68D, [ 2006] = 0x6DA4, [ 2007] = 0x6DB1,
  [ 2008] = 0x6E15, [ 2009] = 0x6E18, [ 2010] = 0x6E29, [ 2011] = 0x6E86,
  [ 2012] = 0xF694, [ 2013] = 0x6EBB, [ 2014] = 0x6EE2, [ 2015] = 0x6EDA,
  [ 2016] = 0x9F7F, [ 2017] = 0x6EE8, [ 2018] = 0x6EE9, [ 2019] = 0x6F24,
  [ 2020] = 0x6F34, [ 2021] = 0xF69D, [ 2022] = 0xF69E, [ 2023] = 0x6F81,
  [ 2024] = 0x6FBE, [ 2025] = 0xF6A1, [ 2026] = 0xF6A2, [ 2027] = 0x71B7,
  [ 2028] = 0x5C99, [ 2029] = 0xF6A5, [ 2030] = 0x702C, [ 2031] = 0xF6A7,
  [ 2032] = 0x7050, [ 2033] = 0x7054, [ 2034] = 0x706F, [ 2035] = 0x707F,
  [ 2036] = 0x7089, [ 2037] = 0xF6AD, [ 2038] = 0xF6AE, [ 2039] = 0xF6AF,
  [ 2040] = 0xF6B0, [ 2041] = 0xE311, [ 2042] = 0x57BE, [ 2043] = 0xE313,
  [ 2044] = 0x713E, [ 2045] = 0xE315, [ 2046] = 0xE316, [ 2047] = 0x69A2,
  [ 2048] = 0xE318, [ 2049] = 0x5B74, [ 2050] = 0x7A49, [ 2051] = 0xE31B,
  [ 2052] = 0xE31C, [ 2053] = 0x7A65, [ 2054] = 0x7A7D, [ 2055] = 0xE31F,
  [ 2056] = 0x7ABB, [ 2057] = 0x7AB0, [ 2058] = 0x7AC2, [ 2059] = 0x7AC3,
  [ 2060] = 0x71D1, [ 2061] = 0xE325, [ 2062] = 0xE326, [ 2063] = 0x7ADA,
  [ 2064] = 0x7ADD, [ 2065] = 0x7AEA, [ 2066] = 0xE32A, [ 2067] = 0x54B2,
  [ 2068] = 0xE32C, [ 2069] = 0x7B0B, [ 2070] = 0x7B55, [ 2071] = 0x7B29,
  [ 2072] = 0xE330, [ 2073] = 0xE331, [ 2074] = 0x7BA2, [ 2075] = 0x7B6F,
  [ 2076] = 0x839C, [ 2077] = 0xE335, [ 2078] = 0xE336, [ 2079] = 0x7BD0,
  [ 2080] = 0x8421, [ 2081] = 0x7B92, [ 2082] = 0xE33A, [ 2083] = 0xE33B,
  [ 2084] = 0xE33C, [ 2085] = 0xE33D, [ 2086] = 0x8492, [ 2087] = 0x7BFA,
  [ 2088] = 0xE340, [ 2089] = 0x7C35, [ 2090] = 0xE342, [ 2091] = 0x7C44,
  [ 2092] = 0x7C83, [ 2093] = 0xE345, [ 2094] = 0x7CA6, [ 2095] = 0x667D,
  [ 2096] = 0xE348, [ 2097] = 0x7CC9, [ 2098] = 0x7CC7, [ 2099] = 0x7CE6,
  [ 2100] = 0x7C74, [ 2101] = 0x7CF3, [ 2102] = 0x7CF5, [ 2103] = 0xE34F,
  [ 2104] = 0x7E67, [ 2105] = 0xE351, [ 2106] = 0xE352, [ 2107] = 0x7D5D,
  [ 2108] = 0xE354, [ 2109] = 0x748D, [ 2110] = 0x7D89, [ 2111] = 0x7DAB,
  [ 2112] = 0x7135, [ 2113] = 0x7DB3, [ 2114] = 0xE35A, [ 2115] = 0xE35B,
  [ 2116] = 0xE35C, [ 2117] = 0x7DE4, [ 2118] = 0xE35E, [ 2119] = 0x7DF5,
  [ 2120] = 0xE360, [ 2121] = 0x7DE5, [ 2122] = 0xE362, [ 2123] = 0xE363,
  [ 2124] = 0xE364, [ 2125] = 0xE365, [ 2126] = 0x7E6E, [ 2127] = 0x7E92,
  [ 2128] = 0xE368, [ 2129] = 0x946C, [ 2130] = 0x7E27, [ 2131] = 0x7F40,
  [ 2132] = 0x7F41, [ 2133] = 0x7F47, [ 2134] = 0x7936, [ 2135] = 0xE36F,
  [ 2136] = 0x99E1, [ 2137] = 0x7F97, [ 2138] = 0xE372, [ 2139] = 0x7FA3,
  [ 2140] = 0xE374, [ 2141] = 0xE375, [ 2142] = 0xE376, [ 2143] = 0xE377,
  [ 2144] = 0xE378, [ 2145] = 0xE379, [ 2146] = 0x7FFA, [ 2147] = 0xE37B,
  [ 2148] = 0xE37C, [ 2149] = 0x8008, [ 2150] = 0x801D, [ 2151] = 0xE37F,
  [ 2152] = 0x802F, [ 2153] = 0xE381, [ 2154] = 0xE382, [ 2155] = 0x803B,
  [ 2156] = 0x803C, [ 2157] = 0x8061, [ 2158] = 0xE386, [ 2159] = 0xE387,
  [ 2160] = 0xE388, [ 2161] = 0xE389, [ 2162] = 0xE38A, [ 2163] = 0x6725,
  [ 2164] = 0x80A7, [ 2165] = 0xE38D, [ 2166] = 0x8107, [ 2167] = 0x811A,
  [ 2168] = 0x58B0, [ 2169] = 0xE391, [ 2170] = 0x6C7F, [ 2171] = 0xE393,
  [ 2172] = 0xE394, [ 2173] = 0x64E7, [ 2174] = 0xE396, [ 2175] = 0x8218,
  [ 2176] = 0xE398, [ 2177] = 0x6A53, [ 2178] = 0xE39A, [ 2179] = 0xE39B,
  [ 2180] = 0xE39C, [ 2181] = 0x8229, [ 2182] = 0xE39E, [ 2183] = 0xE39F,
  [ 2184] = 0xE3A0, [ 2185] = 0x4FF9, [ 2186] = 0xE3A2, [ 2187] = 0x84E2,
  [ 2188] = 0x8362, [ 2189] = 0xE3A5, [ 2190] = 0xE3A6, [ 2191] = 0xE3A7,
  [ 2192] = 0xE3A8, [ 2193] = 0xE3A9, [ 2194] = 0x82AA, [ 2195] = 0x691B,
  [ 2196] = 0xE3AC, [ 2197] = 0xE3AD, [ 2198] = 0x854B, [ 2199] = 0x82D0,
  [ 2200] = 0x831A, [ 2201] = 0xE3B1, [ 2202] = 0xE3B2, [ 2203] = 0xE3B3,
  [ 2204] = 0xE3B4, [ 2205] = 0xE3B5, [ 2206] = 0x827B, [ 2207] = 0x82E2,
  [ 2208] = 0x8318, [ 2209] = 0xE3B9, [ 2210] = 0xE3BA, [ 2211] = 0xE3BB,
  [ 2212] = 0xE3BC, [ 2213] = 0xE3BD, [ 2214] = 0xE3BE, [ 2215] = 0x831D,
  [ 2216] = 0x55EC, [ 2217] = 0x8385, [ 2218] = 0xE3C2, [ 2219] = 0xE3C3,
  [ 2220] = 0x83AC, [ 2221] = 0xE3C5, [ 2222] = 0x83D3, [ 2223] = 0xE3C7,
  [ 2224] = 0xE3C8, [ 2225] = 0x6A57, [ 2226] = 0x855A, [ 2227] = 0xE3CB,
  [ 2228] = 0xE3CC, [ 2229] = 0xE3CD, [ 2230] = 0x8458, [ 2231] = 0xE3CF,
  [ 2232] = 0x8471, [ 2233] = 0xE3D1, [ 2234] = 0xE3D2, [ 2235] = 0x6AA7,
  [ 2236] = 0x844A, [ 2237] = 0xE3D5, [ 2238] = 0x7958, [ 2239] = 0xE3D7,
  [ 2240] = 0xE3D8, [ 2241] = 0xE3D9, [ 2242] = 0xE3DA, [ 2243] = 0x84DE,
  [ 2244] = 0xE3DC, [ 2245] = 0x8391, [ 2246] = 0xE3DE, [ 2247] = 0x8493,
  [ 2248] = 0x84E4, [ 2249] = 0xE3E1, [ 2250] = 0xE3E2, [ 2251] = 0xE3E3,
  [ 2252] = 0xE3E4, [ 2253] = 0x8534, [ 2254] = 0x5AF2, [ 2255] = 0xE3E7,
  [ 2256] = 0xE3E8, [ 2257] = 0x8573, [ 2258] = 0xE3EA, [ 2259] = 0x67BF,
  [ 2260] = 0x8616, [ 2261] = 0xE3ED, [ 2262] = 0xE3EE, [ 2263] = 0x85C1,
  [ 2264] = 0xE3F0, [ 2265] = 0x8602, [ 2266] = 0xE3F2, [ 2267] = 0xE3F3,
  [ 2268] = 0xE3F4, [ 2269] = 0xE3F5, [ 2270] = 0x8628, [ 2271] = 0xE3F7,
  [ 2272] = 0xE3F8, [ 2273] = 0x53F7, [ 2274] = 0xE3FA, [ 2275] = 0x867E,
  [ 2276] = 0x8771, [ 2277] = 0xE3FD, [ 2278] = 0x87EE, [ 2279] = 0xE3FF,
  [ 2280] = 0x87B1, [ 2281] = 0x87DA, [ 2282] = 0x880F, [ 2283] = 0x5661,
  [ 2284] = 0x866C, [ 2285] = 0x6856, [ 2286] = 0xE406, [ 2287] = 0x8845,
  [ 2288] = 0x8846, [ 2289] = 0xE409, [ 2290] = 0xE40A, [ 2291] = 0xE40B,
  [ 2292] = 0x885E, [ 2293] = 0x889C, [ 2294] = 0xE40E, [ 2295] = 0x88B4,
  [ 2296] = 0x88B5, [ 2297] = 0x63C1, [ 2298] = 0x88C5, [ 2299] = 0x7777,
  [ 2300] = 0xE414, [ 2301] = 0x8987, [ 2302] = 0x898A, [ 2303] = 0xE417,
  [ 2304] = 0xE418, [ 2305] = 0x89A7, [ 2306] = 0x89BC, [ 2307] = 0xE41B,
  [ 2308] = 0x89E7, [ 2309] = 0xE41D, [ 2310] = 0xE41E, [ 2311] = 0x8A9C,
  [ 2312] = 0x7793, [ 2313] = 0x91FE, [ 2314] = 0x8A90, [ 2315] = 0xE423,
  [ 2316] = 0x7AE9, [ 2317] = 0xE425, [ 2318] = 0xE426, [ 2319] = 0xE427,
  [ 2320] = 0xE428, [ 2321] = 0x717C, [ 2322] = 0x8B0C, [ 2323] = 0x8B1F,
  [ 2324] = 0xE42C, [ 2325] = 0xE42D, [ 2326] = 0x8B3F, [ 2327] = 0x8B4C,
  [ 2328] = 0x8B4D, [ 2329] = 0x8AA9, [ 2330] = 0xE432, [ 2331] = 0x8B90,
  [ 2332] = 0x8B9B, [ 2333] = 0x8AAF, [ 2334] = 0xE436, [ 2335] = 0xE437,
  [ 2336] = 0x884F, [ 2337] = 0x8C9B, [ 2338] = 0xE43A, [ 2339] = 0xE43B,
  [ 2340] = 0xE43C, [ 2341] = 0xE43D, [ 2342] = 0xE43E, [ 2343] = 0x8CD6,
  [ 2344] = 0xE440, [ 2345] = 0xE441, [ 2346] = 0x8D12, [ 2347] = 0x8D03,
  [ 2348] = 0xE444, [ 2349] = 0x8CDB, [ 2350] = 0x705C, [ 2351] = 0x8D11,
  [ 2352] = 0xE448, [ 2353] = 0xE449, [ 2354] = 0xE44A, [ 2355] = 0x8DA9,
  [ 2356] = 0xE44C, [ 2357] = 0xE44D, [ 2358] = 0xE44E, [ 2359] = 0xE44F,
  [ 2360] = 0xE450, [ 2361] = 0xE451, [ 2362] = 0x7AE7, [ 2363] = 0x8EAD,
  [ 2364] = 0x8EB6, [ 2365] = 0x8EC3, [ 2366] = 0x92D4, [ 2367] = 0x8F19,
  [ 2368] = 0x8F2D, [ 2369] = 0xE459, [ 2370] = 0xE45A, [ 2371] = 0x8FA5,
  [ 2372] = 0x9303, [ 2373] = 0xE45D, [ 2374] = 0xE45E, [ 2375] = 0x8FB3,
  [ 2376] = 0xE460, [ 2377] = 0xE461, [ 2378] = 0xE462, [ 2379] = 0xE463,
  [ 2380] = 0x5EF8, [ 2381] = 0xE465, [ 2382] = 0x8FF9, [ 2383] = 0xE467,
  [ 2384] = 0xE468, [ 2385] = 0xE469, [ 2386] = 0xE46A, [ 2387] = 0xE46B,
  [ 2388] = 0xE46C, [ 2389] = 0x9037, [ 2390] = 0xE46E, [ 2391] = 0xE46F,
  [ 2392] = 0x9061, [ 2393] = 0xE471, [ 2394] = 0xE472, [ 2395] = 0x90A8,
  [ 2396] = 0xE474, [ 2397] = 0x90C4, [ 2398] = 0xE476, [ 2399] = 0x90AE,
  [ 2400] = 0xE478, [ 2401] = 0x9167, [ 2402] = 0xE47A, [ 2403] = 0x91A9,
  [ 2404] = 0x91C4, [ 2405] = 0x7CAC, [ 2406] = 0xE47E, [ 2407] = 0xE47F,
  [ 2408] = 0x920E, [ 2409] = 0x6C9F, [ 2410] = 0x9241, [ 2411] = 0x9262,
  [ 2412] = 0xE484, [ 2413] = 0xE485, [ 2414] = 0xE486, [ 2415] = 0xE487,
  [ 2416] = 0xE488, [ 2417] = 0xE489, [ 2418] = 0xE48A, [ 2419] = 0x932C,
  [ 2420] = 0x936B, [ 2421] = 0xE48D, [ 2422] = 0xE48E, [ 2423] = 0x708F,
  [ 2424] = 0x5AC3, [ 2425] = 0xE491, [ 2426] = 0xE492, [ 2427] = 0xE493,
  [ 2428] = 0x9244, [ 2429] = 0xE495, [ 2430] = 0xE496, [ 2431] = 0xE497,
  [ 2432] = 0x9373, [ 2433] = 0x945B, [ 2434] = 0x8EBC, [ 2435] = 0x9585,
  [ 2436] = 0x95A6, [ 2437] = 0x9426, [ 2438] = 0x95A0, [ 2439] = 0x6FF6,
  [ 2440] = 0xE4A0, [ 2441] = 0xE4A1, [ 2442] = 0xE4A2, [ 2443] = 0xE4A3,
  [ 2444] = 0xE4A4, [ 2445] = 0xE4A5, [ 2446] = 0x6C1C, [ 2447] = 0x967B,
  [ 2448] = 0x9696, [ 2449] = 0xE4A9, [ 2450] = 0x96A3, [ 2451] = 0xE4AB,
  [ 2452] = 0x61DA, [ 2453] = 0x96B6, [ 2454] = 0x78F5, [ 2455] = 0xE4AF,
  [ 2456] = 0x96BD, [ 2457] = 0x53CC, [ 2458] = 0xE4B2, [ 2459] = 0xE4B3,
  [ 2460] = 0xE4B4, [ 2461] = 0xE4B5, [ 2462] = 0xE4B6, [ 2463] = 0xE4B7,
  [ 2464] = 0xE4B8, [ 2465] = 0xE4B9, [ 2466] = 0xE4BA, [ 2467] = 0x9731,
  [ 2468] = 0x8642, [ 2469] = 0x9736, [ 2470] = 0xE4BE, [ 2471] = 0xE4BF,
  [ 2472] = 0xE4C0, [ 2473] = 0xE4C1, [ 2474] = 0x7075, [ 2475] = 0x5B41,
  [ 2476] = 0x971B, [ 2477] = 0xE4C5, [ 2478] = 0xE4C6, [ 2479] = 0x9757,
  [ 2480] = 0x5B4A, [ 2481] = 0xE4C9, [ 2482] = 0x975F, [ 2483] = 0x9425,
  [ 2484] = 0x50D0, [ 2485] = 0xE4CD, [ 2486] = 0xE4CE, [ 2487] = 0x9789,
  [ 2488] = 0x979F, [ 2489] = 0x97B1, [ 2490] = 0x97BE, [ 2491] = 0x97C0,
  [ 2492] = 0x97D2, [ 2493] = 0x97E0, [ 2494] = 0xE4D6, [ 2495] = 0x97EE,
  [ 2496] = 0x741C, [ 2497] = 0xE4D9, [ 2498] = 0xE4DA, [ 2499] = 0x97F5,
  [ 2500] = 0xE4DC, [ 2501] = 0xE4DD, [ 2502] = 0xE4DE, [ 2503] = 0x9834,
  [ 2504] = 0x9833, [ 2505] = 0x984B, [ 2506] = 0x9866, [ 2507] = 0xE4E3,
  [ 2508] = 0xE4E4, [ 2509] = 0xE4E5, [ 2510] = 0xE4E6, [ 2511] = 0xE4E7,
  [ 2512] = 0xE4E8, [ 2513] = 0x98CA, [ 2514] = 0x98B7, [ 2515] = 0x98C8,
  [ 2516] = 0x98C7, [ 2517] = 0xE4ED, [ 2518] = 0xE4EE, [ 2519] = 0xE4EF,
  [ 2520] = 0x55B0, [ 2521] = 0x98E1, [ 2522] = 0x98E6, [ 2523] = 0x98EC,
  [ 2524] = 0x9378, [ 2525] = 0x9939, [ 2526] = 0xE4F6, [ 2527] = 0xE4F7,
  [ 2528] = 0xE4F8, [ 2529] = 0xE4F9, [ 2530] = 0x99F5, [ 2531] = 0x9A0C,
  [ 2532] = 0x9A3B, [ 2533] = 0x9A10, [ 2534] = 0x9A58, [ 2535] = 0xE4FF,
  [ 2536] = 0xE500, [ 2537] = 0xE501, [ 2538] = 0xE502, [ 2539] = 0x9AE0,
  [ 2540] = 0x9AE2, [ 2541] = 0xE505, [ 2542] = 0x9AF4, [ 2543] = 0xE507,
  [ 2544] = 0x9B14, [ 2545] = 0x9B2D, [ 2546] = 0xE50A, [ 2547] = 0x5034,
  [ 2548] = 0x9B34, [ 2549] = 0xE50D, [ 2550] = 0xE50E, [ 2551] = 0xE50F,
  [ 2552] = 0x9B50, [ 2553] = 0x9B40, [ 2554] = 0xE512, [ 2555] = 0x5A45,
  [ 2556] = 0xE514, [ 2557] = 0x9B8E, [ 2558] = 0xE516, [ 2559] = 0x9C02,
  [ 2560] = 0x9BFF, [ 2561] = 0x9C0C, [ 2562] = 0xE51A, [ 2563] = 0x9DD4,
  [ 2564] = 0xE51C, [ 2565] = 0xE51D, [ 2566] = 0xE51E, [ 2567] = 0xE51F,
  [ 2568] = 0xE520, [ 2569] = 0xE521, [ 2570] = 0x9D7E, [ 2571] = 0x9D83,
  [ 2572] = 0xE524, [ 2573] = 0x9E0E, [ 2574] = 0x6888, [ 2575] = 0x9DC4,
  [ 2576] = 0xE528, [ 2577] = 0xE529, [ 2578] = 0xE52A, [ 2579] = 0xE52B,
  [ 2580] = 0xE52C, [ 2581] = 0x9D39, [ 2582] = 0xE52E, [ 2583] = 0xE52F,
  [ 2584] = 0x9E90, [ 2585] = 0x9E95, [ 2586] = 0x9E9E, [ 2587] = 0x9EA2,
  [ 2588] = 0xE534, [ 2589] = 0x9EAA, [ 2590] = 0x9EAF, [ 2591] = 0xE537,
  [ 2592] = 0x9EC1, [ 2593] = 0xE539, [ 2594] = 0xE53A, [ 2595] = 0xE53B,
  [ 2596] = 0x4F32, [ 2597] = 0xE53D, [ 2598] = 0xE53E, [ 2599] = 0x9F02,
  [ 2600] = 0x9F08, [ 2601] = 0xE541, [ 2602] = 0x9424, [ 2603] = 0xE543,
  [ 2604] = 0x9F17, [ 2605] = 0xE545, [ 2606] = 0x9F39, [ 2607] = 0x569F,
  [ 2608] = 0x568A, [ 2609] = 0x9F45, [ 2610] = 0x99B8, [ 2611] = 0xE54B,
  [ 2612] = 0x97F2, [ 2613] = 0x847F, [ 2614] = 0x9F62, [ 2615] = 0x9F69,
  [ 2616] = 0x7ADC, [ 2617] = 0x9F8E, [ 2618] = 0x7216, [ 2619] = 0xE553,
  [ 2620] = 0xE554, [ 2621] = 0xE555, [ 2622] = 0x7177, [ 2623] = 0xE557,
  [ 2624] = 0xE558, [ 2625] = 0xE559, [ 2626] = 0x739E, [ 2627] = 0xE55B,
  [ 2628] = 0xE55C, [ 2629] = 0x799F, [ 2630] = 0xE55E, [ 2631] = 0xE55F,
  [ 2632] = 0x9369, [ 2633] = 0x93F3, [ 2634] = 0xE562, [ 2635] = 0x92EC,
  [ 2636] = 0x9381, [ 2637] = 0x93CB, [ 2638] = 0xE566, [ 2639] = 0xE567,
  [ 2640] = 0x7217, [ 2641] = 0xE569, [ 2642] = 0x7772, [ 2643] = 0x7A43,
  [ 2644] = 0x70D0, [ 2645] = 0xE56D, [ 2646] = 0xE56E, [ 2647] = 0x717E,
  [ 2648] = 0xE570, [ 2649] = 0x70A3, [ 2650] = 0xE572, [ 2651] = 0xE573,
  [ 2652] = 0xE574, [ 2653] = 0xE575, [ 2654] = 0xE576, [ 2655] = 0xE577,
  [ 2656] = 0xE578, [ 2657] = 0xE579, [ 2658] = 0xE57A, [ 2659] = 0xE57B,
  [ 2660] = 0xE57C, [ 2661] = 0xE57D, [ 2662] = 0xE57E, [ 2663] = 0xE57F,
  [ 2664] = 0xE580, [ 2665] = 0x575B, [ 2666] = 0xE582, [ 2667] = 0xE583,
  [ 2668] = 0xE584, [ 2669] = 0xE585, [ 2670] = 0xE586, [ 2671] = 0x8503,
  [ 2672] = 0xE588, [ 2673] = 0xE589, [ 2674] = 0x8455, [ 2675] = 0xE58B,
  [ 2676] = 0xE58C, [ 2677] = 0xE58D, [ 2678] = 0xE58E, [ 2679] = 0xE58F,
  [ 2680] = 0xE590, [ 2681] = 0xE591, [ 2682] = 0xE592, [ 2683] = 0xE593,
  [ 2684] = 0xE594, [ 2685] = 0x67F9, [ 2686] = 0xE596, [ 2687] = 0xE597,
  [ 2688] = 0xE598, [ 2689] = 0x586C, [ 2690] = 0xE59A, [ 2691] = 0x6810,
  [ 2692] = 0xE59C, [ 2693] = 0xE59D, [ 2694] = 0xE59E, [ 2695] = 0xE59F,
  [ 2696] = 0xE5A0, [ 2697] = 0xE5A1, [ 2698] = 0x54CB, [ 2699] = 0x569E,
  [ 2700] = 0xE5A4, [ 2701] = 0x5692, [ 2702] = 0xE5A6, [ 2703] = 0xE5A7,
  [ 2704] = 0xE5A8, [ 2705] = 0x93C6, [ 2706] = 0xE5AA, [ 2707] = 0x939C,
  [ 2708] = 0x4EF8, [ 2709] = 0x512B, [ 2710] = 0xE5AE, [ 2711] = 0xE5AF,
  [ 2712] = 0x4EBC, [ 2713] = 0xE5B1, [ 2714] = 0xE5B2, [ 2715] = 0x4F4B,
  [ 2716] = 0x4F8A, [ 2717] = 0xE5B5, [ 2718] = 0x5A68, [ 2719] = 0xE5B7,
  [ 2720] = 0xE5B8, [ 2721] = 0xE5B9, [ 2722] = 0xE5BA, [ 2723] = 0xE5BB,
  [ 2724] = 0xE5BC, [ 2725] = 0x4F29, [ 2726] = 0xE5BE, [ 2727] = 0xE5BF,
  [ 2728] = 0xE5C0, [ 2729] = 0x8ADA, [ 2730] = 0xE5C2, [ 2731] = 0x4E98,
  [ 2732] = 0x50CD, [ 2733] = 0x510D, [ 2734] = 0x4FA2, [ 2735] = 0x4F03,
  [ 2736] = 0xE5C8, [ 2737] = 0xE5C9, [ 2738] = 0x4F42, [ 2739] = 0x502E,
  [ 2740] = 0x506C, [ 2741] = 0x5081, [ 2742] = 0x4FCC, [ 2743] = 0x4FE5,
  [ 2744] = 0x5058, [ 2745] = 0x50FC, [ 2746] = 0xE5D2, [ 2747] = 0xE5D3,
  [ 2748] = 0xE5D4, [ 2749] = 0xE5D5, [ 2750] = 0x6E76, [ 2751] = 0xE5D7,
  [ 2752] = 0xE5D8, [ 2753] = 0xE5D9, [ 2754] = 0x6D72, [ 2755] = 0xE5DB,
  [ 2756] = 0xE5DC, [ 2757] = 0x51A8, [ 2758] = 0x51C3, [ 2759] = 0xE5DF,
  [ 2760] = 0xE5E0, [ 2761] = 0xE5E1, [ 2762] = 0xE5E2, [ 2763] = 0xE5E3,
  [ 2764] = 0x8D7A, [ 2765] = 0xE5E5, [ 2766] = 0xE5E6, [ 2767] = 0x5259,
  [ 2768] = 0x52A4, [ 2769] = 0xE5E9, [ 2770] = 0x52E1, [ 2771] = 0xE5EB,
  [ 2772] = 0xE5EC, [ 2773] = 0x718C, [ 2774] = 0xE5EE, [ 2775] = 0xE5EF,
  [ 2776] = 0xE5F0, [ 2777] = 0xE5F1, [ 2778] = 0x69D1, [ 2779] = 0xE5F3,
  [ 2780] = 0xE5F4, [ 2781] = 0xE5F5, [ 2782] = 0x7499, [ 2783] = 0x7414,
  [ 2784] = 0x7456, [ 2785] = 0x7398, [ 2786] = 0xE5FA, [ 2787] = 0xE5FB,
  [ 2788] = 0xE5FC, [ 2789] = 0x53D0, [ 2790] = 0xE5FE, [ 2791] = 0x720F,
  [ 2792] = 0xE600, [ 2793] = 0x55B4, [ 2794] = 0xE602, [ 2795] = 0x54CD,
  [ 2796] = 0xE604, [ 2797] = 0x571D, [ 2798] = 0x925D, [ 2799] = 0x96F4,
  [ 2800] = 0x9366, [ 2801] = 0x57DD, [ 2802] = 0x578D, [ 2803] = 0x577F,
  [ 2804] = 0xE60C, [ 2805] = 0x58CB, [ 2806] = 0x5A99, [ 2807] = 0xE60F,
  [ 2808] = 0xE610, [ 2809] = 0xE611, [ 2810] = 0xE612, [ 2811] = 0x5A2C,
  [ 2812] = 0x59B8, [ 2813] = 0x928F, [ 2814] = 0x5A7E, [ 2815] = 0x5ACF,
  [ 2816] = 0x5A12, [ 2817] = 0xE619, [ 2818] = 0xE61A, [ 2819] = 0xE61B,
  [ 2820] = 0xE61C, [ 2821] = 0xE61D, [ 2822] = 0x6D05, [ 2823] = 0x7443,
  [ 2824] = 0x5A21, [ 2825] = 0xE621, [ 2826] = 0x5A81, [ 2827] = 0xE623,
  [ 2828] = 0xE624, [ 2829] = 0x93E0, [ 2830] = 0x748C, [ 2831] = 0xE627,
  [ 2832] = 0x7105, [ 2833] = 0xE629, [ 2834] = 0x9408, [ 2835] = 0xE62B,
  [ 2836] = 0x93BD, [ 2837] = 0xE62D, [ 2838] = 0x5C1E, [ 2839] = 0x5C9E,
  [ 2840] = 0x5E5E, [ 2841] = 0x5E48, [ 2842] = 0xE632, [ 2843] = 0xE633,
  [ 2844] = 0xE634, [ 2845] = 0x5ECD, [ 2846] = 0x5B4F, [ 2847] = 0xE637,
  [ 2848] = 0xE638, [ 2849] = 0xE639, [ 2850] = 0xE63A, [ 2851] = 0xE63B,
  [ 2852] = 0xE63C, [ 2853] = 0xE63D, [ 2854] = 0x812A, [ 2855] = 0xE63F,
  [ 2856] = 0xE640, [ 2857] = 0xE641, [ 2858] = 0xE642, [ 2859] = 0x5F0C,
  [ 2860] = 0x5F0E, [ 2861] = 0xE645, [ 2862] = 0xE646, [ 2863] = 0x5A6B,
  [ 2864] = 0xE648, [ 2865] = 0x5B44, [ 2866] = 0x8614, [ 2867] = 0xE64B,
  [ 2868] = 0x8860, [ 2869] = 0x607E, [ 2870] = 0xE64E, [ 2871] = 0xE64F,
  [ 2872] = 0x5FDB, [ 2873] = 0xE651, [ 2874] = 0xE652, [ 2875] = 0xE653,
  [ 2876] = 0xE654, [ 2877] = 0xE655, [ 2878] = 0x61C0, [ 2879] = 0xE657,
  [ 2880] = 0xE658, [ 2881] = 0xE659, [ 2882] = 0x6199, [ 2883] = 0x6198,
  [ 2884] = 0x6075, [ 2885] = 0xE65D, [ 2886] = 0xE65E, [ 2887] = 0xE65F,
  [ 2888] = 0xE660, [ 2889] = 0x6471, [ 2890] = 0xE662, [ 2891] = 0xE663,
  [ 2892] = 0xE664, [ 2893] = 0xE665, [ 2894] = 0xE666, [ 2895] = 0xE667,
  [ 2896] = 0xE668, [ 2897] = 0x6337, [ 2898] = 0xE66A, [ 2899] = 0x64B6,
  [ 2900] = 0x6331, [ 2901] = 0x63D1, [ 2902] = 0xE66E, [ 2903] = 0xE66F,
  [ 2904] = 0x62A4, [ 2905] = 0xE671, [ 2906] = 0x643B, [ 2907] = 0x656B,
  [ 2908] = 0x6972, [ 2909] = 0xE675, [ 2910] = 0xE676, [ 2911] = 0xE677,
  [ 2912] = 0xE678, [ 2913] = 0xE679, [ 2914] = 0x550D, [ 2915] = 0xE67B,
  [ 2916] = 0xE67C, [ 2917] = 0xE67D, [ 2918] = 0x66CE, [ 2919] = 0xE67F,
  [ 2920] = 0xE680, [ 2921] = 0xE681, [ 2922] = 0xE682, [ 2923] = 0xE683,
  [ 2924] = 0xE684, [ 2925] = 0xE685, [ 2926] = 0xE686, [ 2927] = 0xE687,
  [ 2928] = 0xE688, [ 2929] = 0x78EE, [ 2930] = 0xE68A, [ 2931] = 0xE68B,
  [ 2932] = 0xE68C, [ 2933] = 0xE68D, [ 2934] = 0xE68E, [ 2935] = 0xE68F,
  [ 2936] = 0xE690, [ 2937] = 0x668E, [ 2938] = 0xE692, [ 2939] = 0x666B,
  [ 2940] = 0xE694, [ 2941] = 0x6630, [ 2942] = 0xE696, [ 2943] = 0xE697,
  [ 2944] = 0x6663, [ 2945] = 0xE699, [ 2946] = 0xE69A, [ 2947] = 0x661E,
  [ 2948] = 0xE69C, [ 2949] = 0xE69D, [ 2950] = 0xE69E, [ 2951] = 0xE69F,
  [ 2952] = 0xE6A0, [ 2953] = 0xE6A1, [ 2954] = 0xE6A2, [ 2955] = 0x74D0,
  [ 2956] = 0xE6A4, [ 2957] = 0x678F, [ 2958] = 0xE6A6, [ 2959] = 0x68B6,
  [ 2960] = 0x681E, [ 2961] = 0xE6A9, [ 2962] = 0x6ABE, [ 2963] = 0xE6AB,
  [ 2964] = 0xE6AC, [ 2965] = 0xE6AD, [ 2966] = 0x6A33, [ 2967] = 0x6A52,
  [ 2968] = 0x6AC9, [ 2969] = 0x6B05, [ 2970] = 0xE6B2, [ 2971] = 0x6511,
  [ 2972] = 0x6898, [ 2973] = 0x6A4C, [ 2974] = 0xE6B6, [ 2975] = 0x6A7A,
  [ 2976] = 0x6B57, [ 2977] = 0xE6B9, [ 2978] = 0xE6BA, [ 2979] = 0x93A0,
  [ 2980] = 0x92F2, [ 2981] = 0xE6BD, [ 2982] = 0xE6BE, [ 2983] = 0x9289,
  [ 2984] = 0xE6C0, [ 2985] = 0xE6C1, [ 2986] = 0x9467, [ 2987] = 0x6DA5,
  [ 2988] = 0x6F0B, [ 2989] = 0xE6C5, [ 2990] = 0xE6C6, [ 2991] = 0xE6C7,
  [ 2992] = 0xE6C8, [ 2993] = 0x6E04, [ 2994] = 0xE6CA, [ 2995] = 0x5A3D,
  [ 2996] = 0x6E0A, [ 2997] = 0x5847, [ 2998] = 0x6D24, [ 2999] = 0x7842,
  [ 3000] = 0x713B, [ 3001] = 0xE6D1, [ 3002] = 0xE6D2, [ 3003] = 0x70F1,
  [ 3004] = 0x7250, [ 3005] = 0x7287, [ 3006] = 0x7294, [ 3007] = 0xE6D7,
  [ 3008] = 0xE6D8, [ 3009] = 0x5179, [ 3010] = 0xE6DA, [ 3011] = 0xE6DB,
  [ 3012] = 0x747A, [ 3013] = 0xE6DD, [ 3014] = 0xE6DE, [ 3015] = 0xE6DF,
  [ 3016] = 0xE6E0, [ 3017] = 0xE6E1, [ 3018] = 0xE6E2, [ 3019] = 0xE6E3,
  [ 3020] = 0xE6E4, [ 3021] = 0xE6E5, [ 3022] = 0xE6E6, [ 3023] = 0x60A7,
  [ 3024] = 0xE6E8, [ 3025] = 0x74CC, [ 3026] = 0x743C, [ 3027] = 0x9387,
  [ 3028] = 0x7437, [ 3029] = 0xE6ED, [ 3030] = 0xE6EE, [ 3031] = 0xE6EF,
  [ 3032] = 0x7583, [ 3033] = 0xE6F1, [ 3034] = 0xE6F2, [ 3035] = 0xE6F3,
  [ 3036] = 0xE6F4, [ 3037] = 0x7555, [ 3038] = 0x7673, [ 3039] = 0xE6F7,
  [ 3040] = 0xE6F8, [ 3041] = 0x7468, [ 3042] = 0xE6FA, [ 3043] = 0xE6FB,
  [ 3044] = 0xE6FC, [ 3045] = 0xE6FD, [ 3046] = 0xE6FE, [ 3047] = 0xE6FF,
  [ 3048] = 0xE700, [ 3049] = 0xE701, [ 3050] = 0xE702, [ 3051] = 0x91FA,
  [ 3052] = 0x5732, [ 3053] = 0x9342, [ 3054] = 0xE706, [ 3055] = 0xE707,
  [ 3056] = 0x50DF, [ 3057] = 0xE709, [ 3058] = 0xE70A, [ 3059] = 0x7778,
  [ 3060] = 0xE70C, [ 3061] = 0x770E, [ 3062] = 0x770F, [ 3063] = 0x777B,
  [ 3064] = 0xE710, [ 3065] = 0xE711, [ 3066] = 0xE712, [ 3067] = 0xE713,
  [ 3068] = 0x7438, [ 3069] = 0x749B, [ 3070] = 0xE716, [ 3071] = 0xE717,
  [ 3072] = 0xE718, [ 3073] = 0xE719, [ 3074] = 0xE71A, [ 3075] = 0xE71B,
  [ 3076] = 0x9307, [ 3077] = 0xE71D, [ 3078] = 0x781E, [ 3079] = 0x788D,
  [ 3080] = 0x7888, [ 3081] = 0x78D2, [ 3082] = 0x73D0, [ 3083] = 0x7959,
  [ 3084] = 0xE724, [ 3085] = 0xE725, [ 3086] = 0xE726, [ 3087] = 0xE727,
  [ 3088] = 0x8496, [ 3089] = 0x79A5, [ 3090] = 0x6A2D, [ 3091] = 0xE72B,
  [ 3092] = 0x7A3A, [ 3093] = 0x79F4, [ 3094] = 0xE72E, [ 3095] = 0xE72F,
  [ 3096] = 0xE730, [ 3097] = 0x9235, [ 3098] = 0x79F1, [ 3099] = 0xE733,
  [ 3100] = 0xE734, [ 3101] = 0xE735, [ 3102] = 0xE736, [ 3103] = 0xE737,
  [ 3104] = 0xE738, [ 3105] = 0x556B, [ 3106] = 0xE73A, [ 3107] = 0xE73B,
  [ 3108] = 0xE73C, [ 3109] = 0xE73D, [ 3110] = 0x7AE2, [ 3111] = 0x5A59,
  [ 3112] = 0xE740, [ 3113] = 0xE741, [ 3114] = 0xE742, [ 3115] = 0x5A0D,
  [ 3116] = 0xE744, [ 3117] = 0x78F0, [ 3118] = 0x5A2A, [ 3119] = 0xE747,
  [ 3120] = 0x7AFE, [ 3121] = 0xE749, [ 3122] = 0x7C5D, [ 3123] = 0x7C6D,
  [ 3124] = 0xE74C, [ 3125] = 0xE74D, [ 3126] = 0xE74E, [ 3127] = 0xE74F,
  [ 3128] = 0x7CCD, [ 3129] = 0xE751, [ 3130] = 0xE752, [ 3131] = 0x7C8E,
  [ 3132] = 0x7C7C, [ 3133] = 0x7CAE, [ 3134] = 0x6AB2, [ 3135] = 0x7DDC,
  [ 3136] = 0x7E07, [ 3137] = 0x7DD3, [ 3138] = 0x7F4E, [ 3139] = 0xE75B,
  [ 3140] = 0xE75C, [ 3141] = 0xE75D, [ 3142] = 0x7D97, [ 3143] = 0xE75F,
  [ 3144] = 0xE760, [ 3145] = 0xE761, [ 3146] = 0xE762, [ 3147] = 0x67D6,
  [ 3148] = 0xE764, [ 3149] = 0xE765, [ 3150] = 0x57C4, [ 3151] = 0xE767,
  [ 3152] = 0xE768, [ 3153] = 0xE769, [ 3154] = 0x7FDD, [ 3155] = 0x7B27,
  [ 3156] = 0xE76C, [ 3157] = 0xE76D, [ 3158] = 0xE76E, [ 3159] = 0x7B0C,
  [ 3160] = 0xE770, [ 3161] = 0x99E6, [ 3162] = 0x8645, [ 3163] = 0x9A63,
  [ 3164] = 0x6A1C, [ 3165] = 0xE775, [ 3166] = 0xE776, [ 3167] = 0xE777,
  [ 3168] = 0xE778, [ 3169] = 0x9A1F, [ 3170] = 0xE77A, [ 3171] = 0x8480,
  [ 3172] = 0xE77C, [ 3173] = 0xE77D, [ 3174] = 0xE77E, [ 3175] = 0x8137,
  [ 3176] = 0xE780, [ 3177] = 0x80C6, [ 3178] = 0x8109, [ 3179] = 0x8142,
  [ 3180] = 0xE784, [ 3181] = 0x98C3, [ 3182] = 0xE786, [ 3183] = 0x8262,
  [ 3184] = 0x8265, [ 3185] = 0xE789, [ 3186] = 0x8453, [ 3187] = 0xE78B,
  [ 3188] = 0x8610, [ 3189] = 0xE78D, [ 3190] = 0x5A86, [ 3191] = 0xE78F,
  [ 3192] = 0xE790, [ 3193] = 0x5B2B, [ 3194] = 0xE792, [ 3195] = 0x5AE4,
  [ 3196] = 0xE794, [ 3197] = 0x86A0, [ 3198] = 0xE796, [ 3199] = 0xE797,
  [ 3200] = 0x882D, [ 3201] = 0xE799, [ 3202] = 0x5A02, [ 3203] = 0x886E,
  [ 3204] = 0x4F45, [ 3205] = 0x8887, [ 3206] = 0x88BF, [ 3207] = 0x88E6,
  [ 3208] = 0x8965, [ 3209] = 0x894D, [ 3210] = 0xE7A2, [ 3211] = 0x8954,
  [ 3212] = 0xE7A4, [ 3213] = 0xE7A5, [ 3214] = 0xE7A6, [ 3215] = 0xE7A7,
  [ 3216] = 0xE7A8, [ 3217] = 0xE7A9, [ 3218] = 0xE7AA, [ 3219] = 0x84A3,
  [ 3220] = 0xE7AC, [ 3221] = 0xE7AD, [ 3222] = 0xE7AE, [ 3223] = 0x8A3D,
  [ 3224] = 0x8A1C, [ 3225] = 0xE7B1, [ 3226] = 0x5F4D, [ 3227] = 0x922B,
  [ 3228] = 0xE7B4, [ 3229] = 0x65D4, [ 3230] = 0x7129, [ 3231] = 0x70C4,
  [ 3232] = 0xE7B8, [ 3233] = 0x9D6D, [ 3234] = 0x8C9F, [ 3235] = 0x8CE9,
  [ 3236] = 0xE7BC, [ 3237] = 0x599A, [ 3238] = 0x77C3, [ 3239] = 0x59F0,
  [ 3240] = 0xE7C0, [ 3241] = 0xE7C1, [ 3242] = 0x8E2A, [ 3243] = 0x8EA7,
  [ 3244] = 0xE7C4, [ 3245] = 0x8F30, [ 3246] = 0x8F4A, [ 3247] = 0xE7C7,
  [ 3248] = 0x6C58, [ 3249] = 0x6FBB, [ 3250] = 0xE7CA, [ 3251] = 0xE7CB,
  [ 3252] = 0x6F79, [ 3253] = 0x6E8B, [ 3254] = 0xE7CE, [ 3255] = 0x9BE9,
  [ 3256] = 0xE7D0, [ 3257] = 0xE7D1, [ 3258] = 0x90BB, [ 3259] = 0xE7D3,
  [ 3260] = 0x5571, [ 3261] = 0xE7D5, [ 3262] = 0x91BB, [ 3263] = 0x9404,
  [ 3264] = 0xE7D8, [ 3265] = 0xE7D9, [ 3266] = 0xE7DA, [ 3267] = 0x9427,
  [ 3268] = 0xE7DC, [ 3269] = 0xE7DD, [ 3270] = 0x84E5, [ 3271] = 0x8A2B,
  [ 3272] = 0x9599, [ 3273] = 0x95A7, [ 3274] = 0x9597, [ 3275] = 0x9596,
  [ 3276] = 0xE7E4, [ 3277] = 0x7445, [ 3278] = 0xE7E6, [ 3279] = 0xE7E7,
  [ 3280] = 0xE7E8, [ 3281] = 0xE7E9, [ 3282] = 0xE7EA, [ 3283] = 0xE7EB,
  [ 3284] = 0x968F, [ 3285] = 0xE7ED, [ 3286] = 0xE7EE, [ 3287] = 0xE7EF,
  [ 3288] = 0xE7F0, [ 3289] = 0xE7F1, [ 3290] = 0xE7F2, [ 3291] = 0xE7F3,
  [ 3292] = 0x7412, [ 3293] = 0x746B, [ 3294] = 0xE7F6, [ 3295] = 0x9741,
  [ 3296] = 0xE7F8, [ 3297] = 0x6847, [ 3298] = 0xE7FA, [ 3299] = 0xE7FB,
  [ 3300] = 0xE7FC, [ 3301] = 0xE7FD, [ 3302] = 0x9368, [ 3303] = 0xE7FF,
  [ 3304] = 0xE800, [ 3305] = 0xE801, [ 3306] = 0xE802, [ 3307] = 0x92BA,
  [ 3308] = 0x5B11, [ 3309] = 0x8B69, [ 3310] = 0xE806, [ 3311] = 0x73F9,
  [ 3312] = 0xE808, [ 3313] = 0x979B, [ 3314] = 0x9771, [ 3315] = 0x9938,
  [ 3316] = 0xE80C, [ 3317] = 0x5DC1, [ 3318] = 0xE80E, [ 3319] = 0xE80F,
  [ 3320] = 0x981F, [ 3321] = 0xE811, [ 3322] = 0x92F6, [ 3323] = 0xE813,
  [ 3324] = 0x91E5, [ 3325] = 0xE815, [ 3326] = 0xE816, [ 3327] = 0xE817,
  [ 3328] = 0xE818, [ 3329] = 0x98DC, [ 3330] = 0xE81A, [ 3331] = 0xE81B,
  [ 3332] = 0x922A, [ 3333] = 0xE81D, [ 3334] = 0x8414, [ 3335] = 0x993B,
  [ 3336] = 0x994D, [ 3337] = 0xE821, [ 3338] = 0xE822, [ 3339] = 0x999B,
  [ 3340] = 0xE824, [ 3341] = 0x99AA, [ 3342] = 0x9A5C, [ 3343] = 0xE827,
  [ 3344] = 0xE828, [ 3345] = 0x6A8F, [ 3346] = 0x9A21, [ 3347] = 0x5AFE,
  [ 3348] = 0x9A2F, [ 3349] = 0xE82D, [ 3350] = 0xE82E, [ 3351] = 0xE82F,
  [ 3352] = 0x99BC, [ 3353] = 0xE831, [ 3354] = 0xE832, [ 3355] = 0x937D,
  [ 3356] = 0x5872, [ 3357] = 0xE835, [ 3358] = 0x5822, [ 3359] = 0xE837,
  [ 3360] = 0xE838, [ 3361] = 0x7844, [ 3362] = 0xE83A, [ 3363] = 0xE83B,
  [ 3364] = 0x68C5, [ 3365] = 0xE83D, [ 3366] = 0x9458, [ 3367] = 0xE83F,
  [ 3368] = 0x6150, [ 3369] = 0xE841, [ 3370] = 0xE842, [ 3371] = 0x6107,
  [ 3372] = 0x9C4F, [ 3373] = 0x9C53, [ 3374] = 0x9C7B, [ 3375] = 0x9C35,
  [ 3376] = 0x9C10, [ 3377] = 0x9B7F, [ 3378] = 0x9BCF, [ 3379] = 0xE84B,
  [ 3380] = 0x9B9F, [ 3381] = 0xE84D, [ 3382] = 0xE84E, [ 3383] = 0x9D21,
  [ 3384] = 0xE850, [ 3385] = 0xE851, [ 3386] = 0x9E18, [ 3387] = 0xE853,
  [ 3388] = 0x9D0C, [ 3389] = 0xE855, [ 3390] = 0xE856, [ 3391] = 0xE857,
  [ 3392] = 0xE858, [ 3393] = 0x9DA5, [ 3394] = 0x84BD, [ 3395] = 0xE85B,
  [ 3396] = 0xE85C, [ 3397] = 0xE85D, [ 3398] = 0x85FC, [ 3399] = 0xE85F,
  [ 3400] = 0xE860, [ 3401] = 0xE861, [ 3402] = 0xE862, [ 3403] = 0x8420,
  [ 3404] = 0x85EE, [ 3405] = 0xE865, [ 3406] = 0xE866, [ 3407] = 0xE867,
  [ 3408] = 0x79E2, [ 3409] = 0xE869, [ 3410] = 0xE86A, [ 3411] = 0xE86B,
  [ 3412] = 0xE86C, [ 3413] = 0xE86D, [ 3414] = 0x93DB, [ 3415] = 0x92BE,
  [ 3416] = 0x9348, [ 3417] = 0xE871, [ 3418] = 0x78B9, [ 3419] = 0x9277,
  [ 3420] = 0x944D, [ 3421] = 0x4FE4, [ 3422] = 0xE876, [ 3423] = 0x9064,
  [ 3424] = 0xE878, [ 3425] = 0x783D, [ 3426] = 0x7854, [ 3427] = 0x78B6,
  [ 3428] = 0x784B, [ 3429] = 0xE87D, [ 3430] = 0xE87E, [ 3431] = 0xE87F,
  [ 3432] = 0xE880, [ 3433] = 0x4F72, [ 3434] = 0x6FDA, [ 3435] = 0x6FD9,
  [ 3436] = 0xE884, [ 3437] = 0x701E, [ 3438] = 0x5414, [ 3439] = 0xE887,
  [ 3440] = 0x57BB, [ 3441] = 0x58F3, [ 3442] = 0x578A, [ 3443] = 0x9D16,
  [ 3444] = 0x57D7, [ 3445] = 0x7134, [ 3446] = 0xE88E, [ 3447] = 0xE88F,
  [ 3448] = 0x71EB, [ 3449] = 0xE891, [ 3450] = 0xE892, [ 3451] = 0xE893,
  [ 3452] = 0xE894, [ 3453] = 0xE895, [ 3454] = 0x610C, [ 3455] = 0x5ACE,
  [ 3456] = 0x5A0B, [ 3457] = 0xE899, [ 3458] = 0xE89A, [ 3459] = 0xE89B,
  [ 3460] = 0xE89C, [ 3461] = 0xE89D, [ 3462] = 0x93BB, [ 3463] = 0x93B8,
  [ 3464] = 0xE8A0, [ 3465] = 0xE8A1, [ 3466] = 0x8472, [ 3467] = 0xE8A3,
  [ 3468] = 0xE8A4, [ 3469] = 0xE8A5, [ 3470] = 0xE8A6, [ 3471] = 0xE8A7,
  [ 3472] = 0x5994, [ 3473] = 0xE8A9, [ 3474] = 0xE8AA, [ 3475] = 0x7DA8,
  [ 3476] = 0xE8AC, [ 3477] = 0xE8AD, [ 3478] = 0xE8AE, [ 3479] = 0xE8AF,
  [ 3480] = 0xE8B0, [ 3481] = 0x92E5, [ 3482] = 0x73E2, [ 3483] = 0xE8B3,
  [ 3484] = 0x74B4, [ 3485] = 0xE8B5, [ 3486] = 0xE8B6, [ 3487] = 0xE8B7,
  [ 3488] = 0xE8B8, [ 3489] = 0x6AD8, [ 3490] = 0x73F3, [ 3491] = 0x73FB,
  [ 3492] = 0xE8BC, [ 3493] = 0xE8BD, [ 3494] = 0xE8BE, [ 3495] = 0xE8BF,
  [ 3496] = 0xE8C0, [ 3497] = 0xE8C1, [ 3498] = 0xE8C2, [ 3499] = 0xE8C3,
  [ 3500] = 0x7448, [ 3501] = 0xE8C5, [ 3502] = 0x70A5, [ 3503] = 0xE8C7,
  [ 3504] = 0x9284, [ 3505] = 0x73E6, [ 3506] = 0x935F, [ 3507] = 0xE8CB,
  [ 3508] = 0x9331, [ 3509] = 0xE8CD, [ 3510] = 0xE8CE, [ 3511] = 0x9386,
  [ 3512] = 0xE8D0, [ 3513] = 0xE8D1, [ 3514] = 0xE8D2, [ 3515] = 0xE8D3,
  [ 3516] = 0x716B, [ 3517] = 0xE8D5, [ 3518] = 0xE8D6, [ 3519] = 0x56A4,
  [ 3520] = 0xE8D8, [ 3521] = 0xE8D9, [ 3522] = 0xE8DA, [ 3523] = 0x5502,
  [ 3524] = 0x79C4, [ 3525] = 0xE8DD, [ 3526] = 0x7DFE, [ 3527] = 0xE8DF,
  [ 3528] = 0xE8E0, [ 3529] = 0xE8E1, [ 3530] = 0xE8E2, [ 3531] = 0x9401,
  [ 3532] = 0xE8E4, [ 3533] = 0xE8E5, [ 3534] = 0xE8E6, [ 3535] = 0x59B0,
  [ 3536] = 0xE8E8, [ 3537] = 0xE8E9, [ 3538] = 0xE8EA, [ 3539] = 0x5AA1,
  [ 3540] = 0xE8EC, [ 3541] = 0xE8ED, [ 3542] = 0xE8EE, [ 3543] = 0x925F,
  [ 3544] = 0x5A79, [ 3545] = 0xE8F1, [ 3546] = 0xE8F2, [ 3547] = 0x9374,
  [ 3548] = 0xE8F4, [ 3549] = 0xE8F5, [ 3550] = 0xE8F6, [ 3551] = 0xE8F7,
  [ 3552] = 0x50F4, [ 3553] = 0xE8F9, [ 3554] = 0xE8FA, [ 3555] = 0xE8FB,
  [ 3556] = 0x7175, [ 3557] = 0xE8FD, [ 3558] = 0xE8FE, [ 3559] = 0x6E0F,
  [ 3560] = 0xE900, [ 3561] = 0xE901, [ 3562] = 0x6D57, [ 3563] = 0xE903,
  [ 3564] = 0x7067, [ 3565] = 0x6CAF, [ 3566] = 0xE906, [ 3567] = 0xE907,
  [ 3568] = 0xE908, [ 3569] = 0x6E02, [ 3570] = 0x6F0C, [ 3571] = 0xE90B,
  [ 3572] = 0xE90C, [ 3573] = 0x7551, [ 3574] = 0xE90E, [ 3575] = 0xE90F,
  [ 3576] = 0xE910, [ 3577] = 0xE911, [ 3578] = 0xE912, [ 3579] = 0x59C4,
  [ 3580] = 0x926E, [ 3581] = 0xE915, [ 3582] = 0x8F41, [ 3583] = 0xE917,
  [ 3584] = 0xE918, [ 3585] = 0x5812, [ 3586] = 0x57C8, [ 3587] = 0xE91B,
  [ 3588] = 0xE91C, [ 3589] = 0x70FE, [ 3590] = 0xE91E, [ 3591] = 0xE91F,
  [ 3592] = 0xE920, [ 3593] = 0xE921, [ 3594] = 0xE922, [ 3595] = 0x68B9,
  [ 3596] = 0x6967, [ 3597] = 0xE925, [ 3598] = 0xE926, [ 3599] = 0xE927,
  [ 3600] = 0xE928, [ 3601] = 0xE929, [ 3602] = 0xE92A, [ 3603] = 0xE92B,
  [ 3604] = 0xE92C, [ 3605] = 0x6A1A, [ 3606] = 0xE92E, [ 3607] = 0xE92F,
  [ 3608] = 0x843E, [ 3609] = 0xE931, [ 3610] = 0xE932, [ 3611] = 0xE933,
  [ 3612] = 0xE934, [ 3613] = 0xE935, [ 3614] = 0xE936, [ 3615] = 0x6F17,
  [ 3616] = 0xE938, [ 3617] = 0x833D, [ 3618] = 0xE93A, [ 3619] = 0x83ED,
  [ 3620] = 0xE93C, [ 3621] = 0xE93D, [ 3622] = 0xE93E, [ 3623] = 0x5989,
  [ 3624] = 0x5A82, [ 3625] = 0xE941, [ 3626] = 0x5A61, [ 3627] = 0x5A71,
  [ 3628] = 0xE944, [ 3629] = 0xE945, [ 3630] = 0xE946, [ 3631] = 0x59EF,
  [ 3632] = 0xE948, [ 3633] = 0xE949, [ 3634] = 0x718E, [ 3635] = 0x9390,
  [ 3636] = 0x669A, [ 3637] = 0xE94D, [ 3638] = 0x5A6E, [ 3639] = 0x5A2B,
  [ 3640] = 0xE950, [ 3641] = 0x6A2B, [ 3642] = 0xE952, [ 3643] = 0xE953,
  [ 3644] = 0xE954, [ 3645] = 0xE955, [ 3646] = 0x711D, [ 3647] = 0xE957,
  [ 3648] = 0xE958, [ 3649] = 0x4FB0, [ 3650] = 0xE95A, [ 3651] = 0x5CC2,
  [ 3652] = 0xE95C, [ 3653] = 0xE95D, [ 3654] = 0xE95E, [ 3655] = 0x6A0C,
  [ 3656] = 0xE960, [ 3657] = 0xE961, [ 3658] = 0x70A6, [ 3659] = 0x7133,
  [ 3660] = 0xE964, [ 3661] = 0xE965, [ 3662] = 0x6CDF, [ 3663] = 0xE967,
  [ 3664] = 0xE968, [ 3665] = 0x7E65, [ 3666] = 0x59EB, [ 3667] = 0x5D2F,
  [ 3668] = 0xE96C, [ 3669] = 0x5F5C, [ 3670] = 0xE96E, [ 3671] = 0xE96F,
  [ 3672] = 0x7DA4, [ 3673] = 0x8426, [ 3674] = 0x5485, [ 3675] = 0xE973,
  [ 3676] = 0xE974, [ 3677] = 0xE975, [ 3678] = 0x577E, [ 3679] = 0xE977,
  [ 3680] = 0xE978, [ 3681] = 0xE979, [ 3682] = 0xE97A, [ 3683] = 0xE97B,
  [ 3684] = 0x7003, [ 3685] = 0xE97D, [ 3686] = 0x5D70, [ 3687] = 0x738F,
  [ 3688] = 0x7CD3, [ 3689] = 0xE981, [ 3690] = 0xE982, [ 3691] = 0x4FC8,
  [ 3692] = 0x7FE7, [ 3693] = 0x72CD, [ 3694] = 0x7310, [ 3695] = 0xE987,
  [ 3696] = 0x7338, [ 3697] = 0x7339, [ 3698] = 0xE98A, [ 3699] = 0x7341,
  [ 3700] = 0x7348, [ 3701] = 0xE98D, [ 3702] = 0xE98E, [ 3703] = 0x906C,
  [ 3704] = 0x71F5, [ 3705] = 0xE991, [ 3706] = 0x73E1, [ 3707] = 0x81F6,
  [ 3708] = 0xE994, [ 3709] = 0x770C, [ 3710] = 0xE996, [ 3711] = 0x6CA2,
  [ 3712] = 0x56FD, [ 3713] = 0x7419, [ 3714] = 0x741E, [ 3715] = 0x741F,
  [ 3716] = 0xE99C, [ 3717] = 0xE99D, [ 3718] = 0xE99E, [ 3719] = 0xE99F,
  [ 3720] = 0x74D3, [ 3721] = 0xE9A1, [ 3722] = 0xE9A2, [ 3723] = 0x7542,
  [ 3724] = 0x756D, [ 3725] = 0x7572, [ 3726] = 0x758D, [ 3727] = 0xE9A7,
  [ 3728] = 0x75C8, [ 3729] = 0x75DC, [ 3730] = 0xE9AA, [ 3731] = 0x764D,
  [ 3732] = 0xE9AC, [ 3733] = 0x7674, [ 3734] = 0xE9AE, [ 3735] = 0x767A,
  [ 3736] = 0xE9B0, [ 3737] = 0x7188, [ 3738] = 0x5623, [ 3739] = 0x8980,
  [ 3740] = 0x5869, [ 3741] = 0xE9B5, [ 3742] = 0x7743, [ 3743] = 0xE9B7,
  [ 3744] = 0x6761, [ 3745] = 0xE9B9, [ 3746] = 0xE9BA, [ 3747] = 0x7798,
  [ 3748] = 0xE9BC, [ 3749] = 0xE9BD, [ 3750] = 0x5C5E, [ 3751] = 0x77BE,
  [ 3752] = 0x77CB, [ 3753] = 0x58F2, [ 3754] = 0x7818, [ 3755] = 0x70B9,
  [ 3756] = 0x781C, [ 3757] = 0xE9C5, [ 3758] = 0x7839, [ 3759] = 0x7847,
  [ 3760] = 0x7851, [ 3761] = 0x7866, [ 3762] = 0x8448, [ 3763] = 0xE9CB,
  [ 3764] = 0x7933, [ 3765] = 0x6803, [ 3766] = 0x7932, [ 3767] = 0xE9CF,
  [ 3768] = 0xE9D0, [ 3769] = 0x7991, [ 3770] = 0x7999, [ 3771] = 0x8FBB,
  [ 3772] = 0x7A06, [ 3773] = 0x8FBC, [ 3774] = 0xE9D6, [ 3775] = 0x7A91,
  [ 3776] = 0xE9D8, [ 3777] = 0x7ABC, [ 3778] = 0x8279, [ 3779] = 0xE9DB,
  [ 3780] = 0x7ACF, [ 3781] = 0x7ADB, [ 3782] = 0xE9DE, [ 3783] = 0x4E21,
  [ 3784] = 0x7B62, [ 3785] = 0x7B6C, [ 3786] = 0x7B7B, [ 3787] = 0x7C12,
  [ 3788] = 0x7C1B, [ 3789] = 0xE9E5, [ 3790] = 0xE9E6, [ 3791] = 0x7C7B,
  [ 3792] = 0x7C9C, [ 3793] = 0xE9E9, [ 3794] = 0x7CB8, [ 3795] = 0xE9EB,
  [ 3796] = 0x7CED, [ 3797] = 0x8F93, [ 3798] = 0x70C0, [ 3799] = 0xE9EF,
  [ 3800] = 0x7DCF, [ 3801] = 0x7DD4, [ 3802] = 0x7DD0, [ 3803] = 0x7DFD,
  [ 3804] = 0x7FAE, [ 3805] = 0x7FB4, [ 3806] = 0x729F, [ 3807] = 0xE9F7,
  [ 3808] = 0x8020, [ 3809] = 0x8025, [ 3810] = 0x7B39, [ 3811] = 0x802E,
  [ 3812] = 0x8031, [ 3813] = 0x8054, [ 3814] = 0xE9FE, [ 3815] = 0x57B4,
  [ 3816] = 0x70A0, [ 3817] = 0x80B7, [ 3818] = 0x80E9, [ 3819] = 0xEA03,
  [ 3820] = 0x810C, [ 3821] = 0x732A, [ 3822] = 0x810E, [ 3823] = 0x8112,
  [ 3824] = 0x7560, [ 3825] = 0x8114, [ 3826] = 0xEA0A, [ 3827] = 0xEA0B,
  [ 3828] = 0x8156, [ 3829] = 0x8159, [ 3830] = 0x815A, [ 3831] = 0xEA0F,
  [ 3832] = 0x583A, [ 3833] = 0x817C, [ 3834] = 0x8184, [ 3835] = 0xEA13,
  [ 3836] = 0x8193, [ 3837] = 0xEA15, [ 3838] = 0x81A5, [ 3839] = 0x57EF,
  [ 3840] = 0x81C1, [ 3841] = 0x81E4, [ 3842] = 0x8254, [ 3843] = 0xEA1B,
  [ 3844] = 0x82A6, [ 3845] = 0x8276, [ 3846] = 0x82CA, [ 3847] = 0x82D8,
  [ 3848] = 0x82FF, [ 3849] = 0xEA21, [ 3850] = 0x8357, [ 3851] = 0x9669,
  [ 3852] = 0x698A, [ 3853] = 0x8405, [ 3854] = 0x70F5, [ 3855] = 0x8464,
  [ 3856] = 0x60E3, [ 3857] = 0x8488, [ 3858] = 0xEA2A, [ 3859] = 0x84BE,
  [ 3860] = 0x84E1, [ 3861] = 0x84F8, [ 3862] = 0x8510, [ 3863] = 0x8538,
  [ 3864] = 0x8552, [ 3865] = 0xEA31, [ 3866] = 0x856F, [ 3867] = 0x8570,
  [ 3868] = 0x85E0, [ 3869] = 0xEA35, [ 3870] = 0x8672, [ 3871] = 0x8692,
  [ 3872] = 0x86B2, [ 3873] = 0x86EF, [ 3874] = 0x9645, [ 3875] = 0x878B,
  [ 3876] = 0xEA3C, [ 3877] = 0xEA3D, [ 3878] = 0x88AE, [ 3879] = 0x88FF,
  [ 3880] = 0x8924, [ 3881] = 0x8947, [ 3882] = 0x8991, [ 3883] = 0xEA43,
  [ 3884] = 0x8A29, [ 3885] = 0x8A38, [ 3886] = 0x8A94, [ 3887] = 0x8AB4,
  [ 3888] = 0x8C51, [ 3889] = 0x8CD4, [ 3890] = 0x8CF2, [ 3891] = 0x8D1C,
  [ 3892] = 0xEA4C, [ 3893] = 0x585F, [ 3894] = 0x8DC3, [ 3895] = 0xEA4F,
  [ 3896] = 0x4EEE, [ 3897] = 0x8E3A, [ 3898] = 0x55D8, [ 3899] = 0x5754,
  [ 3900] = 0x8E71, [ 3901] = 0x55F5, [ 3902] = 0x8EB0, [ 3903] = 0xEA57,
  [ 3904] = 0x8ECE, [ 3905] = 0x8EE2, [ 3906] = 0x8EE4, [ 3907] = 0x8EED,
  [ 3908] = 0x8EF2, [ 3909] = 0x8FB7, [ 3910] = 0x8FC1, [ 3911] = 0x8FCA,
  [ 3912] = 0x8FCC, [ 3913] = 0x9033, [ 3914] = 0x99C4, [ 3915] = 0xEA63,
  [ 3916] = 0x98E0, [ 3917] = 0x9213, [ 3918] = 0xEA66, [ 3919] = 0x9228,
  [ 3920] = 0x9258, [ 3921] = 0x926B, [ 3922] = 0x92B1, [ 3923] = 0x92AE,
  [ 3924] = 0x92BF, [ 3925] = 0x92E3, [ 3926] = 0x92EB, [ 3927] = 0x92F3,
  [ 3928] = 0x92F4, [ 3929] = 0x92FD, [ 3930] = 0x9343, [ 3931] = 0x9384,
  [ 3932] = 0x93AD, [ 3933] = 0xEA75, [ 3934] = 0xEA76, [ 3935] = 0x9EBF,
  [ 3936] = 0x9417, [ 3937] = 0x5301, [ 3938] = 0x941D, [ 3939] = 0x942D,
  [ 3940] = 0x943E, [ 3941] = 0xEA7D, [ 3942] = 0x9454, [ 3943] = 0x9479,
  [ 3944] = 0x952D, [ 3945] = 0x95A2, [ 3946] = 0xEA82, [ 3947] = 0x95F4,
  [ 3948] = 0x9633, [ 3949] = 0xEA85, [ 3950] = 0x67A0, [ 3951] = 0xEA87,
  [ 3952] = 0x9740, [ 3953] = 0xEA89, [ 3954] = 0x97B2, [ 3955] = 0x97C2,
  [ 3956] = 0x5654, [ 3957] = 0xEA8D, [ 3958] = 0x60E8, [ 3959] = 0x98B9,
  [ 3960] = 0xEA90, [ 3961] = 0x98F1, [ 3962] = 0x5844, [ 3963] = 0x990E,
  [ 3964] = 0x9919, [ 3965] = 0x51B4, [ 3966] = 0x991C, [ 3967] = 0x9937,
  [ 3968] = 0x9942, [ 3969] = 0x995D, [ 3970] = 0x9962, [ 3971] = 0xEA9B,
  [ 3972] = 0x99C5, [ 3973] = 0xEA9D, [ 3974] = 0x9A3C, [ 3975] = 0x9B0F,
  [ 3976] = 0x7A83, [ 3977] = 0x9B69, [ 3978] = 0x9B81, [ 3979] = 0x9BDD,
  [ 3980] = 0x9BF1, [ 3981] = 0x9BF4, [ 3982] = 0xEAA6, [ 3983] = 0x9C20,
  [ 3984] = 0xEAA8, [ 3985] = 0xEAA9, [ 3986] = 0x9D49, [ 3987] = 0x9C3A,
  [ 3988] = 0x9EFE, [ 3989] = 0x5650, [ 3990] = 0x9D93, [ 3991] = 0x9DBD,
  [ 3992] = 0x9DC0, [ 3993] = 0x9DFC, [ 3994] = 0x94F6, [ 3995] = 0x8FB6,
  [ 3996] = 0x9E7B, [ 3997] = 0x9EAC, [ 3998] = 0x9EB1, [ 3999] = 0x9EBD,
  [ 4000] = 0x9EC6, [ 4001] = 0x94DC, [ 4002] = 0x9EE2, [ 4003] = 0x9EF1,
  [ 4004] = 0x9EF8, [ 4005] = 0x7AC8, [ 4006] = 0x9F44, [ 4007] = 0xEABF,
  [ 4008] = 0xEAC0, [ 4009] = 0xEAC1, [ 4010] = 0x691A, [ 4011] = 0x94C3,
  [ 4012] = 0x59AC, [ 4013] = 0xEAC5, [ 4014] = 0x5840, [ 4015] = 0x94C1,
  [ 4016] = 0xEAC8, [ 4017] = 0xEAC9, [ 4018] = 0xEACA, [ 4019] = 0xEACB,
  [ 4020] = 0xEACC, [ 4021] = 0x5757, [ 4022] = 0x7173, [ 4023] = 0xEACF,
  [ 4024] = 0xEAD0, [ 4025] = 0xEAD1, [ 4026] = 0x546A, [ 4027] = 0xEAD3,
  [ 4028] = 0xEAD4, [ 4029] = 0x549E, [ 4030] = 0xEAD6, [ 4031] = 0xEAD7,
  [ 4032] = 0xEAD8, [ 4033] = 0xEAD9, [ 4034] = 0xEADA, [ 4035] = 0x60E7,
  [ 4036] = 0xEADC, [ 4037] = 0x567A, [ 4038] = 0xEADE, [ 4039] = 0xEADF,
  [ 4040] = 0xEAE0, [ 4041] = 0xEAE1, [ 4042] = 0xEAE2, [ 4043] = 0xEAE3,
  [ 4044] = 0x6955, [ 4045] = 0x9C2F, [ 4046] = 0x87A5, [ 4047] = 0xEAE7,
  [ 4048] = 0xEAE8, [ 4049] = 0xEAE9, [ 4050] = 0xEAEA, [ 4051] = 0xEAEB,
  [ 4052] = 0xEAEC, [ 4053] = 0x5C20, [ 4054] = 0xEAEE, [ 4055] = 0x5E0B,
  [ 4056] = 0xEAF0, [ 4057] = 0xEAF1, [ 4058] = 0xEAF2, [ 4059] = 0x671E,
  [ 4060] = 0xEAF4, [ 4061] = 0xEAF5, [ 4062] = 0xEAF6, [ 4063] = 0xEAF7,
  [ 4064] = 0xEAF8, [ 4065] = 0xEAF9, [ 4066] = 0xEAFA, [ 4067] = 0xEAFB,
  [ 4068] = 0x5364, [ 4069] = 0x84AD, [ 4070] = 0xEAFE, [ 4071] = 0xEAFF,
  [ 4072] = 0xEB00, [ 4073] = 0x8B81, [ 4074] = 0xEB02, [ 4075] = 0xEB03,
  [ 4076] = 0xEB04, [ 4077] = 0xEB05, [ 4078] = 0x4E78, [ 4079] = 0x70BB,
  [ 4080] = 0xEB08, [ 4081] = 0xEB09, [ 4082] = 0xEB0A, [ 4083] = 0xEB0B,
  [ 4084] = 0xEB0C, [ 4085] = 0xEB0D, [ 4086] = 0xEB0E, [ 4087] = 0xEB0F,
  [ 4088] = 0x62C3, [ 4089] = 0xEB11, [ 4090] = 0xEB12, [ 4091] = 0x7198,
  [ 4092] = 0x6855, [ 4093] = 0xEB15, [ 4094] = 0x69E9, [ 4095] = 0xEB17,
  [ 4096] = 0xEB18, [ 4097] = 0xEB19, [ 4098] = 0xEB1A, [ 4099] = 0xEB1B,
  [ 4100] = 0xEB1C, [ 4101] = 0xEB1D, [ 4102] = 0x82FD, [ 4103] = 0xEB1F,
  [ 4104] = 0xEB20, [ 4105] = 0xEB21, [ 4106] = 0x89A5, [ 4107] = 0xEB23,
  [ 4108] = 0x8FA0, [ 4109] = 0xEB25, [ 4110] = 0x97B8, [ 4111] = 0xEB27,
  [ 4112] = 0x9847, [ 4113] = 0x9ABD, [ 4114] = 0xEB2A, [ 4115] = 0xEB2B,
  [ 4116] = 0xEB2C, [ 4117] = 0xEB2D, [ 4118] = 0xEB2E, [ 4119] = 0xEB2F,
  [ 4120] = 0xEB30, [ 4121] = 0xEB31, [ 4122] = 0xEB32, [ 4123] = 0xEB33,
  [ 4124] = 0xEB34, [ 4125] = 0xEB35, [ 4126] = 0xEB36, [ 4127] = 0xEB37,
  [ 4128] = 0xEB38, [ 4129] = 0xEB39, [ 4130] = 0x5FB1, [ 4131] = 0x6648,
  [ 4132] = 0x66BF, [ 4133] = 0xEB3D, [ 4134] = 0xEB3E, [ 4135] = 0xEB3F,
  [ 4136] = 0xEB40, [ 4137] = 0xEB41, [ 4138] = 0xEB42, [ 4139] = 0xEB43,
  [ 4140] = 0xEB44, [ 4141] = 0xEB45, [ 4142] = 0xEB46, [ 4143] = 0x58B5,
  [ 4144] = 0x670E, [ 4145] = 0x6918, [ 4146] = 0xEB4A, [ 4147] = 0xEB4B,
  [ 4148] = 0xEB4C, [ 4149] = 0xEB4D, [ 4150] = 0xEB4E, [ 4151] = 0xEB4F,
  [ 4152] = 0xEB50, [ 4153] = 0xEB51, [ 4154] = 0xEB52, [ 4155] = 0xEB53,
  [ 4156] = 0xEB54, [ 4157] = 0xEB55, [ 4158] = 0xEB56, [ 4159] = 0xEB57,
  [ 4160] = 0xEB58, [ 4161] = 0xEB59, [ 4162] = 0xEB5A, [ 4163] = 0xEB5B,
  [ 4164] = 0x51D2, [ 4165] = 0xEB5D, [ 4166] = 0x599F, [ 4167] = 0xEB5F,
  [ 4168] = 0xEB60, [ 4169] = 0xEB61, [ 4170] = 0xEB62, [ 4171] = 0xEB63,
  [ 4172] = 0x5788, [ 4173] = 0xEB65, [ 4174] = 0xEB66, [ 4175] = 0xEB67,
  [ 4176] = 0xEB68, [ 4177] = 0xEB69, [ 4178] = 0xEB6A, [ 4179] = 0xEB6B,
  [ 4180] = 0x8B5E, [ 4181] = 0xEB6D, [ 4182] = 0xEB6E, [ 4183] = 0xEB6F,
  [ 4184] = 0xEB70, [ 4185] = 0xEB71, [ 4186] = 0x7209, [ 4187] = 0xEB73,
  [ 4188] = 0xEB74, [ 4189] = 0x5965, [ 4190] = 0xEB76, [ 4191] = 0xEB77,
  [ 4192] = 0xEB78, [ 4193] = 0x8EDA, [ 4194] = 0xEB7A, [ 4195] = 0x528F,
  [ 4196] = 0x573F, [ 4197] = 0x7171, [ 4198] = 0xEB7E, [ 4199] = 0xEB7F,
  [ 4200] = 0xEB80, [ 4201] = 0xEB81, [ 4202] = 0x55BC, [ 4203] = 0xEB83,
  [ 4204] = 0xEB84, [ 4205] = 0xEB85, [ 4206] = 0xEB86, [ 4207] = 0xEB87,
  [ 4208] = 0xEB88, [ 4209] = 0xEB89, [ 4210] = 0xEB8A, [ 4211] = 0xEB8B,
  [ 4212] = 0xEB8C, [ 4213] = 0xEB8D, [ 4214] = 0xEB8E, [ 4215] = 0xEB8F,
  [ 4216] = 0xEB90, [ 4217] = 0x5066, [ 4218] = 0xEB92, [ 4219] = 0xEB93,
  [ 4220] = 0xEB94, [ 4221] = 0xEB95, [ 4222] = 0xEB96, [ 4223] = 0xEB97,
  [ 4224] = 0xEB98, [ 4225] = 0xEB99, [ 4226] = 0xEB9A, [ 4227] = 0xEB9B,
  [ 4228] = 0x57A1, [ 4229] = 0x7151, [ 4230] = 0xEB9E, [ 4231] = 0xEB9F,
  [ 4232] = 0xEBA0, [ 4233] = 0x9056, [ 4234] = 0xEBA2, [ 4235] = 0xEBA3,
  [ 4236] = 0x8B62, [ 4237] = 0xEBA5, [ 4238] = 0xEBA6, [ 4239] = 0x5D5B,
  [ 4240] = 0xEBA8, [ 4241] = 0xEBA9, [ 4242] = 0xEBAA, [ 4243] = 0xEBAB,
  [ 4244] = 0x8AEA, [ 4245] = 0xEBAD, [ 4246] = 0xEBAE, [ 4247] = 0xEBAF,
  [ 4248] = 0xEBB0, [ 4249] = 0xEBB1, [ 4250] = 0xEBB2, [ 4251] = 0xEBB3,
  [ 4252] = 0xEBB4, [ 4253] = 0x9465, [ 4254] = 0xEBB6, [ 4255] = 0x6195,
  [ 4256] = 0x5A27, [ 4257] = 0xEBB9, [ 4258] = 0xEBBA, [ 4259] = 0x56B9,
  [ 4260] = 0xEBBC, [ 4261] = 0xEBBD, [ 4262] = 0x4E6A, [ 4263] = 0xEBBF,
  [ 4264] = 0x9656, [ 4265] = 0x6D8F, [ 4266] = 0xEBC2, [ 4267] = 0xEBC3,
  [ 4268] = 0x8977, [ 4269] = 0xEBC5, [ 4270] = 0xEBC6, [ 4271] = 0xEBC7,
  [ 4272] = 0xEBC8, [ 4273] = 0xEBC9, [ 4274] = 0xEBCA, [ 4275] = 0x7B42,
  [ 4276] = 0xEBCC, [ 4277] = 0xEBCD, [ 4278] = 0xEBCE, [ 4279] = 0xEBCF,
  [ 4280] = 0xEBD0, [ 4281] = 0x7A45, [ 4282] = 0xEBD2, [ 4283] = 0xEBD3,
  [ 4284] = 0x9A26, [ 4285] = 0xEBD5, [ 4286] = 0xEBD6, [ 4287] = 0xEBD7,
  [ 4288] = 0xEBD8, [ 4289] = 0x7983, [ 4290] = 0xEBDA, [ 4291] = 0xEBDB,
  [ 4292] = 0x5D2C, [ 4293] = 0xEBDD, [ 4294] = 0xEBDE, [ 4295] = 0xEBDF,
  [ 4296] = 0xEBE0, [ 4297] = 0xEBE1, [ 4298] = 0x753B, [ 4299] = 0x8865,
  [ 4300] = 0xEBE4, [ 4301] = 0x58B6, [ 4302] = 0xEBE6, [ 4303] = 0xEBE7,
  [ 4304] = 0xEBE8, [ 4305] = 0xEBE9, [ 4306] = 0xEBEA, [ 4307] = 0xEBEB,
  [ 4308] = 0xEBEC, [ 4309] = 0x9281, [ 4310] = 0xEBEE, [ 4311] = 0xEBEF,
  [ 4312] = 0x9330, [ 4313] = 0xEBF1, [ 4314] = 0xEBF2, [ 4315] = 0x6C39,
  [ 4316] = 0x949F, [ 4317] = 0xEBF5, [ 4318] = 0xEBF6, [ 4319] = 0x8827,
  [ 4320] = 0x88F5, [ 4321] = 0xEBF9, [ 4322] = 0xEBFA, [ 4323] = 0xEBFB,
  [ 4324] = 0x6EB8, [ 4325] = 0xEBFD, [ 4326] = 0xEBFE, [ 4327] = 0xEBFF,
  [ 4328] = 0xEC00, [ 4329] = 0xEC01, [ 4330] = 0xEC02, [ 4331] = 0xEC03,
  [ 4332] = 0x66B6, [ 4333] = 0xEC05, [ 4334] = 0xEC06, [ 4335] = 0x8943,
  [ 4336] = 0xEC08, [ 4337] = 0xEC09, [ 4338] = 0x56D6, [ 4339] = 0xEC0B,
  [ 4340] = 0xEC0C, [ 4341] = 0xEC0D, [ 4342] = 0xEC0E, [ 4343] = 0xEC0F,
  [ 4344] = 0xEC10, [ 4345] = 0x71AD, [ 4346] = 0x8366, [ 4347] = 0xEC13,
  [ 4348] = 0xEC14, [ 4349] = 0xEC15, [ 4350] = 0xEC16, [ 4351] = 0xEC17,
  [ 4352] = 0xEC18, [ 4353] = 0xEC19, [ 4354] = 0xEC1A, [ 4355] = 0xEC1B,
  [ 4356] = 0xEC1C, [ 4357] = 0xEC1D, [ 4358] = 0x7B43, [ 4359] = 0x797E,
  [ 4360] = 0xEC20, [ 4361] = 0x6FB5, [ 4362] = 0xEC22, [ 4363] = 0x6A03,
  [ 4364] = 0xEC24, [ 4365] = 0x53A2, [ 4366] = 0xEC26, [ 4367] = 0x93BF,
  [ 4368] = 0x6836, [ 4369] = 0x975D, [ 4370] = 0xEC2A, [ 4371] = 0xEC2B,
  [ 4372] = 0xEC2C, [ 4373] = 0xEC2D, [ 4374] = 0xEC2E, [ 4375] = 0xEC2F,
  [ 4376] = 0x5D85, [ 4377] = 0xEC31, [ 4378] = 0xEC32, [ 4379] = 0x5715,
  [ 4380] = 0x9823, [ 4381] = 0xEC35, [ 4382] = 0x5DAB, [ 4383] = 0xEC37,
  [ 4384] = 0x65BE, [ 4385] = 0x69D5, [ 4386] = 0x53D2, [ 4387] = 0xEC3B,
  [ 4388] = 0xEC3C, [ 4389] = 0xEC3D, [ 4390] = 0x6736, [ 4391] = 0xEC3F,
  [ 4392] = 0xEC40, [ 4393] = 0xEC41, [ 4394] = 0xEC42, [ 4395] = 0xEC43,
  [ 4396] = 0xEC44, [ 4397] = 0xEC45, [ 4398] = 0xEC46, [ 4399] = 0xEC47,
  [ 4400] = 0xEC48, [ 4401] = 0xEC49, [ 4402] = 0xEC4A, [ 4403] = 0xEC4B,
  [ 4404] = 0xEC4C, [ 4405] = 0xEC4D, [ 4406] = 0x63E6, [ 4407] = 0xEC4F,
  [ 4408] = 0x7808, [ 4409] = 0x9255, [ 4410] = 0xEC52, [ 4411] = 0xEC53,
  [ 4412] = 0xEC54, [ 4413] = 0xEC55, [ 4414] = 0xEC56, [ 4415] = 0xEC57,
  [ 4416] = 0xEC58, [ 4417] = 0x59F8, [ 4418] = 0xEC5A, [ 4419] = 0xEC5B,
  [ 4420] = 0xEC5C, [ 4421] = 0xEC5D, [ 4422] = 0xEC5E, [ 4423] = 0xEC5F,
  [ 4424] = 0xEC60, [ 4425] = 0xEC61, [ 4426] = 0xEC62, [ 4427] = 0xEC63,
  [ 4428] = 0x8FD0, [ 4429] = 0x728F, [ 4430] = 0x568B, [ 4431] = 0xEC67,
  [ 4432] = 0xEC68, [ 4433] = 0xEC69, [ 4434] = 0xEC6A, [ 4435] = 0xEC6B,
  [ 4436] = 0xEC6C, [ 4437] = 0xEC6D, [ 4438] = 0xEC6E, [ 4439] = 0xEC6F,
  [ 4440] = 0xEC70, [ 4441] = 0xEC71, [ 4442] = 0xEC72, [ 4443] = 0xEC73,
  [ 4444] = 0x7E9F, [ 4445] = 0xEC75, [ 4446] = 0xEC76, [ 4447] = 0xEC77,
  [ 4448] = 0x9547, [ 4449] = 0xEC79, [ 4450] = 0x71A2, [ 4451] = 0xEC7B,
  [ 4452] = 0xEC7C, [ 4453] = 0x9012, [ 4454] = 0xEC7E, [ 4455] = 0xEC7F,
  [ 4456] = 0xEC80, [ 4457] = 0x8FBE, [ 4458] = 0x55C1, [ 4459] = 0x8FBA,
  [ 4460] = 0xEC84, [ 4461] = 0x8FB9, [ 4462] = 0xEC86, [ 4463] = 0xEC87,
  [ 4464] = 0x7E7F, [ 4465] = 0x6F56, [ 4466] = 0x6AB1, [ 4467] = 0x4EEA,
  [ 4468] = 0xEC8C, [ 4469] = 0xEC8D, [ 4470] = 0xEC8E, [ 4471] = 0xEC8F,
  [ 4472] = 0x8E80, [ 4473] = 0xEC91, [ 4474] = 0xEC92, [ 4475] = 0xEC93,
  [ 4476] = 0xEC94, [ 4477] = 0xEC95, [ 4478] = 0xEC96, [ 4479] = 0xEC97,
  [ 4480] = 0xEC98, [ 4481] = 0xEC99, [ 4482] = 0xEC9A, [ 4483] = 0xEC9B,
  [ 4484] = 0x4E9A, [ 4485] = 0xEC9D, [ 4486] = 0xEC9E, [ 4487] = 0x56BF,
  [ 4488] = 0xECA0, [ 4489] = 0x8E0E, [ 4490] = 0x5B6D, [ 4491] = 0xECA3,
  [ 4492] = 0xECA4, [ 4493] = 0x63DE, [ 4494] = 0xECA6, [ 4495] = 0xECA7,
  [ 4496] = 0xECA8, [ 4497] = 0x6530, [ 4498] = 0x562D, [ 4499] = 0xECAB,
  [ 4500] = 0x541A, [ 4501] = 0xECAD, [ 4502] = 0xECAE, [ 4503] = 0xECAF,
  [ 4504] = 0xECB0, [ 4505] = 0x5622, [ 4506] = 0x561E, [ 4507] = 0x7F49,
  [ 4508] = 0xECB4, [ 4509] = 0x5975, [ 4510] = 0xECB6, [ 4511] = 0x8770,
  [ 4512] = 0x4E1C, [ 4513] = 0xECB9, [ 4514] = 0xECBA, [ 4515] = 0xECBB,
  [ 4516] = 0x8117, [ 4517] = 0x9D5E, [ 4518] = 0x8D18, [ 4519] = 0x763B,
  [ 4520] = 0x9C45, [ 4521] = 0x764E, [ 4522] = 0x77B9, [ 4523] = 0x9345,
  [ 4524] = 0x5432, [ 4525] = 0x8148, [ 4526] = 0x82F7, [ 4527] = 0x5625,
  [ 4528] = 0x8132, [ 4529] = 0x8418, [ 4530] = 0x80BD, [ 4531] = 0x55EA,
  [ 4532] = 0x7962, [ 4533] = 0x5643, [ 4534] = 0x5416, [ 4535] = 0xECCF,
  [ 4536] = 0xECD0, [ 4537] = 0x5605, [ 4538] = 0x55F1, [ 4539] = 0x66F1,
  [ 4540] = 0xECD4, [ 4541] = 0xECD5, [ 4542] = 0x7534, [ 4543] = 0x55F0,
  [ 4544] = 0x55BA, [ 4545] = 0x5497, [ 4546] = 0x5572, [ 4547] = 0xECDB,
  [ 4548] = 0xECDC, [ 4549] = 0x5ED0, [ 4550] = 0xECDE, [ 4551] = 0xECDF,
  [ 4552] = 0xECE0, [ 4553] = 0xECE1, [ 4554] = 0x9EAB, [ 4555] = 0x7D5A,
  [ 4556] = 0x55DE, [ 4557] = 0xECE5, [ 4558] = 0x629D, [ 4559] = 0x976D,
  [ 4560] = 0x5494, [ 4561] = 0x8CCD, [ 4562] = 0x71F6, [ 4563] = 0x9176,
  [ 4564] = 0x63FC, [ 4565] = 0x63B9, [ 4566] = 0x63FE, [ 4567] = 0x5569,
  [ 4568] = 0xECF0, [ 4569] = 0x9C72, [ 4570] = 0xECF2, [ 4571] = 0x519A,
  [ 4572] = 0xECF4, [ 4573] = 0xECF5, [ 4574] = 0x51A7, [ 4575] = 0x544D,
  [ 4576] = 0x551E, [ 4577] = 0x5513, [ 4578] = 0x7666, [ 4579] = 0x8E2D,
  [ 4580] = 0xECFC, [ 4581] = 0x75B1, [ 4582] = 0x80B6, [ 4583] = 0x8804,
  [ 4584] = 0x8786, [ 4585] = 0x88C7, [ 4586] = 0x81B6, [ 4587] = 0x841C,
  [ 4588] = 0xED04, [ 4589] = 0xED05, [ 4590] = 0x7304, [ 4591] = 0xED07,
  [ 4592] = 0x5B90, [ 4593] = 0x830B, [ 4594] = 0xED0A, [ 4595] = 0x567B,
  [ 4596] = 0xED0C, [ 4597] = 0xED0D, [ 4598] = 0xED0E, [ 4599] = 0xED0F,
  [ 4600] = 0xED10, [ 4601] = 0xED11, [ 4602] = 0x9170, [ 4603] = 0xED13,
  [ 4604] = 0x9208, [ 4605] = 0xED15, [ 4606] = 0xED16, [ 4607] = 0xED17,
  [ 4608] = 0xED18, [ 4609] = 0x7266, [ 4610] = 0xED1A, [ 4611] = 0xED1B,
  [ 4612] = 0xED1C, [ 4613] = 0xED1D, [ 4614] = 0xED1E, [ 4615] = 0xED1F,
  [ 4616] = 0x9C5D, [ 4617] = 0x651F, [ 4618] = 0xED22, [ 4619] = 0xED23,
  [ 4620] = 0xED24, [ 4621] = 0xED25, [ 4622] = 0xED26, [ 4623] = 0xED27,
  [ 4624] = 0xED28, [ 4625] = 0xED29, [ 4626] = 0xED2A, [ 4627] = 0xED2B,
  [ 4628] = 0xED2C, [ 4629] = 0xED2D, [ 4630] = 0x71A3, [ 4631] = 0x7E8E,
  [ 4632] = 0x9D50, [ 4633] = 0x4E1A, [ 4634] = 0x4E04, [ 4635] = 0xED33,
  [ 4636] = 0x5B0D, [ 4637] = 0x6CB2, [ 4638] = 0x5367, [ 4639] = 0xED37,
  [ 4640] = 0xED38, [ 4641] = 0x537D, [ 4642] = 0xED3A, [ 4643] = 0xED3B,
  [ 4644] = 0x589A, [ 4645] = 0xED3D, [ 4646] = 0x822D, [ 4647] = 0x544B,
  [ 4648] = 0x57AA, [ 4649] = 0xED41, [ 4650] = 0xED42, [ 4651] = 0xED43,
  [ 4652] = 0xED44, [ 4653] = 0xED45, [ 4654] = 0x7374, [ 4655] = 0xED47,
  [ 4656] = 0xED48, [ 4657] = 0x9BED, [ 4658] = 0xED4A, [ 4659] = 0xED4B,
  [ 4660] = 0xED4C, [ 4661] = 0xED4D, [ 4662] = 0xED4E, [ 4663] = 0xED4F,
  [ 4664] = 0x845C, [ 4665] = 0xED51, [ 4666] = 0xED52, [ 4667] = 0xED53,
  [ 4668] = 0xED54, [ 4669] = 0x632E, [ 4670] = 0x7D25, [ 4671] = 0xED57,
  [ 4672] = 0xED58, [ 4673] = 0xED59, [ 4674] = 0x9008, [ 4675] = 0x52CC,
  [ 4676] = 0xED5C, [ 4677] = 0xED5D, [ 4678] = 0xED5E, [ 4679] = 0xED5F,
  [ 4680] = 0x7640, [ 4681] = 0x5AF0, [ 4682] = 0xED62, [ 4683] = 0x787A,
  [ 4684] = 0xED64, [ 4685] = 0x58A7, [ 4686] = 0xED66, [ 4687] = 0x567C,
  [ 4688] = 0x9B8B, [ 4689] = 0x5D74, [ 4690] = 0x7654, [ 4691] = 0xED6B,
  [ 4692] = 0x9E85, [ 4693] = 0xED6D, [ 4694] = 0xED6E, [ 4695] = 0xED6F,
  [ 4696] = 0x6119, [ 4697] = 0xED71, [ 4698] = 0xED72, [ 4699] = 0xED73,
  [ 4700] = 0x565D, [ 4701] = 0xED75, [ 4702] = 0x57A7, [ 4703] = 0xED77,
  [ 4704] = 0xED78, [ 4705] = 0x5234, [ 4706] = 0xED7A, [ 4707] = 0xED7B,
  [ 4708] = 0xED7C, [ 4709] = 0x9D7C, [ 4710] = 0x7C56, [ 4711] = 0x9B39,
  [ 4712] = 0x57DE, [ 4713] = 0xED81, [ 4714] = 0x5C53, [ 4715] = 0x64D3,
  [ 4716] = 0xED84, [ 4717] = 0xED85, [ 4718] = 0xED86, [ 4719] = 0x86AD,
  [ 4720] = 0xED88, [ 4721] = 0xED89, [ 4722] = 0xED8A, [ 4723] = 0xED8B,
  [ 4724] = 0xED8C, [ 4725] = 0x51FE, [ 4726] = 0xED8E, [ 4727] = 0x5D8E,
  [ 4728] = 0x9703, [ 4729] = 0xED91, [ 4730] = 0x9E81, [ 4731] = 0x904C,
  [ 4732] = 0x7B1F, [ 4733] = 0x9B02, [ 4734] = 0x5CD1, [ 4735] = 0x7BA3,
  [ 4736] = 0x6268, [ 4737] = 0x6335, [ 4738] = 0x9AFF, [ 4739] = 0x7BCF,
  [ 4740] = 0x9B2A, [ 4741] = 0x7C7E, [ 4742] = 0xED9E, [ 4743] = 0x7C42,
  [ 4744] = 0x7C86, [ 4745] = 0x9C15, [ 4746] = 0x7BFC, [ 4747] = 0x9B09,
  [ 4748] = 0xEDA4, [ 4749] = 0x9C1B, [ 4750] = 0xEDA6, [ 4751] = 0x9F5A,
  [ 4752] = 0x5573, [ 4753] = 0x5BC3, [ 4754] = 0x4FFD, [ 4755] = 0x9E98,
  [ 4756] = 0x4FF2, [ 4757] = 0x5260, [ 4758] = 0xEDAE, [ 4759] = 0x52D1,
  [ 4760] = 0x5767, [ 4761] = 0x5056, [ 4762] = 0x59B7, [ 4763] = 0x5E12,
  [ 4764] = 0x97C8, [ 4765] = 0x9DAB, [ 4766] = 0x8F5C, [ 4767] = 0x5469,
  [ 4768] = 0x97B4, [ 4769] = 0x9940, [ 4770] = 0x97BA, [ 4771] = 0x532C,
  [ 4772] = 0x6130, [ 4773] = 0x692C, [ 4774] = 0x53DA, [ 4775] = 0x9C0A,
  [ 4776] = 0x9D02, [ 4777] = 0xEDC1, [ 4778] = 0x9641, [ 4779] = 0x6980,
  [ 4780] = 0x50A6, [ 4781] = 0x7546, [ 4782] = 0xEDC6, [ 4783] = 0x99DA,
  [ 4784] = 0x5273, [ 4785] = 0xEDC9, [ 4786] = 0x9159, [ 4787] = 0x9681,
  [ 4788] = 0x915C, [ 4789] = 0xEDCD, [ 4790] = 0x9151, [ 4791] = 0xEDCF,
  [ 4792] = 0x637F, [ 4793] = 0xEDD1, [ 4794] = 0x6ACA, [ 4795] = 0x5611,
  [ 4796] = 0x918E, [ 4797] = 0x757A, [ 4798] = 0x6285, [ 4799] = 0xEDD7,
  [ 4800] = 0x734F, [ 4801] = 0x7C70, [ 4802] = 0xEDDA, [ 4803] = 0xEDDB,
  [ 4804] = 0xEDDC, [ 4805] = 0xEDDD, [ 4806] = 0x76D6, [ 4807] = 0x9B9D,
  [ 4808] = 0x4E2A, [ 4809] = 0xEDE1, [ 4810] = 0x83BE, [ 4811] = 0x8842,
  [ 4812] = 0xEDE4, [ 4813] = 0x5C4A, [ 4814] = 0x69C0, [ 4815] = 0xEDE7,
  [ 4816] = 0x577A, [ 4817] = 0x521F, [ 4818] = 0x5DF5, [ 4819] = 0x4ECE,
  [ 4820] = 0x6C31, [ 4821] = 0xEDED, [ 4822] = 0x4F39, [ 4823] = 0x549C,
  [ 4824] = 0x54DA, [ 4825] = 0x529A, [ 4826] = 0x8D82, [ 4827] = 0xEDF3,
  [ 4828] = 0xEDF4, [ 4829] = 0xEDF5, [ 4830] = 0xEDF6, [ 4831] = 0x6B52,
  [ 4832] = 0x917C, [ 4833] = 0x9FA5, [ 4834] = 0x9B97, [ 4835] = 0x982E,
  [ 4836] = 0x98B4, [ 4837] = 0x9ABA, [ 4838] = 0x9EA8, [ 4839] = 0x9E84,
  [ 4840] = 0x717A, [ 4841] = 0x7B14, [ 4842] = 0xEE02, [ 4843] = 0x6BFA,
  [ 4844] = 0x8818, [ 4845] = 0x7F78, [ 4846] = 0xEE06, [ 4847] = 0x5620,
  [ 4848] = 0xEE08, [ 4849] = 0x8E77, [ 4850] = 0x9F53, [ 4851] = 0xEE0B,
  [ 4852] = 0x8DD4, [ 4853] = 0x8E4F, [ 4854] = 0x9E1C, [ 4855] = 0x8E01,
  [ 4856] = 0x6282, [ 4857] = 0xEE11, [ 4858] = 0x8E28, [ 4859] = 0x8E75,
  [ 4860] = 0x7AD3, [ 4861] = 0xEE15, [ 4862] = 0x7A3E, [ 4863] = 0x78D8,
  [ 4864] = 0x6CEA, [ 4865] = 0x8A67, [ 4866] = 0x7607, [ 4867] = 0xEE1B,
  [ 4868] = 0x9F26, [ 4869] = 0x6CCE, [ 4870] = 0x87D6, [ 4871] = 0x75C3,
  [ 4872] = 0xEE20, [ 4873] = 0x7853, [ 4874] = 0xEE22, [ 4875] = 0x8D0C,
  [ 4876] = 0x72E2, [ 4877] = 0x7371, [ 4878] = 0x8B2D, [ 4879] = 0x7302,
  [ 4880] = 0x74F1, [ 4881] = 0x8CEB, [ 4882] = 0xEE2A, [ 4883] = 0x862F,
  [ 4884] = 0x5FBA, [ 4885] = 0x88A0, [ 4886] = 0xEE2E, [ 4887] = 0xEE2F,
  [ 4888] = 0xEE30, [ 4889] = 0xEE31, [ 4890] = 0xEE32, [ 4891] = 0x8A7E,
  [ 4892] = 0xEE34, [ 4893] = 0xEE35, [ 4894] = 0x60FD, [ 4895] = 0x7667,
  [ 4896] = 0x9AD7, [ 4897] = 0x9D44, [ 4898] = 0x936E, [ 4899] = 0x9B8F,
  [ 4900] = 0x87F5, [ 4901] = 0xEE3D, [ 4902] = 0xEE3E, [ 4903] = 0x8CF7,
  [ 4904] = 0x732C, [ 4905] = 0x9721, [ 4906] = 0x9BB0, [ 4907] = 0xEE43,
  [ 4908] = 0x72B2, [ 4909] = 0xEE45, [ 4910] = 0x7C51, [ 4911] = 0x994A,
  [ 4912] = 0xEE48, [ 4913] = 0x6159, [ 4914] = 0xEE4A, [ 4915] = 0x9E96,
  [ 4916] = 0x617D, [ 4917] = 0xEE4D, [ 4918] = 0x575F, [ 4919] = 0x616F,
  [ 4920] = 0x62A6, [ 4921] = 0x6239, [ 4922] = 0xEE52, [ 4923] = 0xEE53,
  [ 4924] = 0x61E2, [ 4925] = 0x53AA, [ 4926] = 0xEE56, [ 4927] = 0x6364,
  [ 4928] = 0x6802, [ 4929] = 0xEE59, [ 4930] = 0x5D57, [ 4931] = 0xEE5B,
  [ 4932] = 0x8FDA, [ 4933] = 0xEE5D, [ 4934] = 0xEE5E, [ 4935] = 0x50D9,
  [ 4936] = 0xEE60, [ 4937] = 0x7906, [ 4938] = 0x5332, [ 4939] = 0x9638,
  [ 4940] = 0xEE64, [ 4941] = 0xEE65, [ 4942] = 0xEE66, [ 4943] = 0x77FE,
  [ 4944] = 0xEE68, [ 4945] = 0x7CC2, [ 4946] = 0xEE6A, [ 4947] = 0x7CDA,
  [ 4948] = 0x7A2D, [ 4949] = 0x8066, [ 4950] = 0x8063, [ 4951] = 0x7D4D,
  [ 4952] = 0x7505, [ 4953] = 0x74F2, [ 4954] = 0x8994, [ 4955] = 0x821A,
  [ 4956] = 0x670C, [ 4957] = 0x8062, [ 4958] = 0xEE76, [ 4959] = 0x805B,
  [ 4960] = 0x74F0, [ 4961] = 0x8103, [ 4962] = 0x7724, [ 4963] = 0x8989,
  [ 4964] = 0xEE7C, [ 4965] = 0x7553, [ 4966] = 0xEE7E, [ 4967] = 0x87A9,
  [ 4968] = 0x87CE, [ 4969] = 0x81C8, [ 4970] = 0x878C, [ 4971] = 0x8A49,
  [ 4972] = 0x8CAD, [ 4973] = 0x8B43, [ 4974] = 0x772B, [ 4975] = 0x74F8,
  [ 4976] = 0x84DA, [ 4977] = 0xEE89, [ 4978] = 0x69B2, [ 4979] = 0x8DA6,
  [ 4980] = 0xEE8C, [ 4981] = 0x89A9, [ 4982] = 0xEE8E, [ 4983] = 0x6DB9,
  [ 4984] = 0x87C1, [ 4985] = 0xEE91, [ 4986] = 0x74E7, [ 4987] = 0xEE93,
  [ 4988] = 0x7176, [ 4989] = 0x60A4, [ 4990] = 0x619C, [ 4991] = 0xEE97,
  [ 4992] = 0xEE98, [ 4993] = 0x6077, [ 4994] = 0xEE9A, [ 4995] = 0x7F71,
  [ 4996] = 0xEE9C, [ 4997] = 0xEE9D, [ 4998] = 0x60E9, [ 4999] = 0xEE9F,
  [ 5000] = 0x5220, [ 5001] = 0xEEA1, [ 5002] = 0xEEA2, [ 5003] = 0xEEA3,
  [ 5004] = 0xEEA4, [ 5005] = 0xEEA5, [ 5006] = 0xEEA6, [ 5007] = 0xEEA7,
  [ 5008] = 0xEEA8, [ 5009] = 0xEEA9, [ 5010] = 0xEEAA, [ 5011] = 0x5CC1,
  [ 5012] = 0xEEAC, [ 5013] = 0xEEAD, [ 5014] = 0xEEAE, [ 5015] = 0xEEAF,
  [ 5016] = 0xEEB0, [ 5017] = 0xEEB1, [ 5018] = 0xEEB2, [ 5019] = 0x5B1F,
  [ 5020] = 0xEEB4, [ 5021] = 0x9F50, [ 5022] = 0x9EA6, [ 5023] = 0xEEB7,
  [ 5024] = 0x3000, [ 5025] = 0xFF0C, [ 5026] = 0x3001, [ 5027] = 0x3002,
  [ 5028] = 0xFF0E, [ 5029] = 0x2027, [ 5030] = 0xFF1B, [ 5031] = 0xFF1A,
  [ 5032] = 0xFF1F, [ 5033] = 0xFF01, [ 5034] = 0xFE30, [ 5035] = 0x2026,
  [ 5036] = 0x2025, [ 5037] = 0xFE50, [ 5038] = 0xFE51, [ 5039] = 0xFE52,
  [ 5040] = 0x00B7, [ 5041] = 0xFE54, [ 5042] = 0xFE55, [ 5043] = 0xFE56,
  [ 5044] = 0xFE57, [ 5045] = 0xFF5C, [ 5046] = 0x2013, [ 5047] = 0xFE31,
  [ 5048] = 0x2014, [ 5049] = 0xFE33, [ 5050] = 0x2574, [ 5051] = 0xFE34,
  [ 5052] = 0xFE4F, [ 5053] = 0xFF08, [ 5054] = 0xFF09, [ 5055] = 0xFE35,
  [ 5056] = 0xFE36, [ 5057] = 0xFF5B, [ 5058] = 0xFF5D, [ 5059] = 0xFE37,
  [ 5060] = 0xFE38, [ 5061] = 0x3014, [ 5062] = 0x3015, [ 5063] = 0xFE39,
  [ 5064] = 0xFE3A, [ 5065] = 0x3010, [ 5066] = 0x3011, [ 5067] = 0xFE3B,
  [ 5068] = 0xFE3C, [ 5069] = 0x300A, [ 5070] = 0x300B, [ 5071] = 0xFE3D,
  [ 5072] = 0xFE3E, [ 5073] = 0x3008, [ 5074] = 0x3009, [ 5075] = 0xFE3F,
  [ 5076] = 0xFE40, [ 5077] = 0x300C, [ 5078] = 0x300D, [ 5079] = 0xFE41,
  [ 5080] = 0xFE42, [ 5081] = 0x300E, [ 5082] = 0x300F, [ 5083] = 0xFE43,
  [ 5084] = 0xFE44, [ 5085] = 0xFE59, [ 5086] = 0xFE5A, [ 5087] = 0xFE5B,
  [ 5088] = 0xFE5C, [ 5089] = 0xFE5D, [ 5090] = 0xFE5E, [ 5091] = 0x2018,
  [ 5092] = 0x2019, [ 5093] = 0x201C, [ 5094] = 0x201D, [ 5095] = 0x301D,
  [ 5096] = 0x301E, [ 5097] = 0x2035, [ 5098] = 0x2032, [ 5099] = 0xFF03,
  [ 5100] = 0xFF06, [ 5101] = 0xFF0A, [ 5102] = 0x203B, [ 5103] = 0x00A7,
  [ 5104] = 0x3003, [ 5105] = 0x25CB, [ 5106] = 0x25CF, [ 5107] = 0x25B3,
  [ 5108] = 0x25B2, [ 5109] = 0x25CE, [ 5110] = 0x2606, [ 5111] = 0x2605,
  [ 5112] = 0x25C7, [ 5113] = 0x25C6, [ 5114] = 0x25A1, [ 5115] = 0x25A0,
  [ 5116] = 0x25BD, [ 5117] = 0x25BC, [ 5118] = 0x32A3, [ 5119] = 0x2105,
  [ 5120] = 0x00AF, [ 5121] = 0xFFE3, [ 5122] = 0xFF3F, [ 5123] = 0x02CD,
  [ 5124] = 0xFE49, [ 5125] = 0xFE4A, [ 5126] = 0xFE4D, [ 5127] = 0xFE4E,
  [ 5128] = 0xFE4B, [ 5129] = 0xFE4C, [ 5130] = 0xFE5F, [ 5131] = 0xFE60,
  [ 5132] = 0xFE61, [ 5133] = 0xFF0B, [ 5134] = 0xFF0D, [ 5135] = 0x00D7,
  [ 5136] = 0x00F7, [ 5137] = 0x00B1, [ 5138] = 0x221A, [ 5139] = 0xFF1C,
  [ 5140] = 0xFF1E, [ 5141] = 0xFF1D, [ 5142] = 0x2266, [ 5143] = 0x2267,
  [ 5144] = 0x2260, [ 5145] = 0x221E, [ 5146] = 0x2252, [ 5147] = 0x2261,
  [ 5148] = 0xFE62, [ 5149] = 0xFE63, [ 5150] = 0xFE64, [ 5151] = 0xFE65,
  [ 5152] = 0xFE66, [ 5153] = 0xFF5E, [ 5154] = 0x2229, [ 5155] = 0x222A,
  [ 5156] = 0x22A5, [ 5157] = 0x2220, [ 5158] = 0x221F, [ 5159] = 0x22BF,
  [ 5160] = 0x33D2, [ 5161] = 0x33D1, [ 5162] = 0x222B, [ 5163] = 0x222E,
  [ 5164] = 0x2235, [ 5165] = 0x2234, [ 5166] = 0x2640, [ 5167] = 0x2642,
  [ 5168] = 0x2295, [ 5169] = 0x2299, [ 5170] = 0x2191, [ 5171] = 0x2193,
  [ 5172] = 0x2190, [ 5173] = 0x2192, [ 5174] = 0x2196, [ 5175] = 0x2197,
  [ 5176] = 0x2199, [ 5177] = 0x2198, [ 5178] = 0x2225, [ 5179] = 0x2223,
  [ 5180] = 0xFF0F, [ 5181] = 0xFF3C, [ 5182] = 0x2215, [ 5183] = 0xFE68,
  [ 5184] = 0xFF04, [ 5185] = 0xFFE5, [ 5186] = 0x3012, [ 5187] = 0xFFE0,
  [ 5188] = 0xFFE1, [ 5189] = 0xFF05, [ 5190] = 0xFF20, [ 5191] = 0x2103,
  [ 5192] = 0x2109, [ 5193] = 0xFE69, [ 5194] = 0xFE6A, [ 5195] = 0xFE6B,
  [ 5196] = 0x33D5, [ 5197] = 0x339C, [ 5198] = 0x339D, [ 5199] = 0x339E,
  [ 5200] = 0x33CE, [ 5201] = 0x33A1, [ 5202] = 0x338E, [ 5203] = 0x338F,
  [ 5204] = 0x33C4, [ 5205] = 0x00B0, [ 5206] = 0x5159, [ 5207] = 0x515B,
  [ 5208] = 0x515E, [ 5209] = 0x515D, [ 5210] = 0x5161, [ 5211] = 0x5163,
  [ 5212] = 0x55E7, [ 5213] = 0x74E9, [ 5214] = 0x7CCE, [ 5215] = 0x2581,
  [ 5216] = 0x2582, [ 5217] = 0x2583, [ 5218] = 0x2584, [ 5219] = 0x2585,
  [ 5220] = 0x2586, [ 5221] = 0x2587, [ 5222] = 0x2588, [ 5223] = 0x258F,
  [ 5224] = 0x258E, [ 5225] = 0x258D, [ 5226] = 0x258C, [ 5227] = 0x258B,
  [ 5228] = 0x258A, [ 5229] = 0x2589, [ 5230] = 0x253C, [ 5231] = 0x2534,
  [ 5232] = 0x252C, [ 5233] = 0x2524, [ 5234] = 0x251C, [ 5235] = 0x2594,
  [ 5236] = 0x2500, [ 5237] = 0x2502, [ 5238] = 0x2595, [ 5239] = 0x250C,
  [ 5240] = 0x2510, [ 5241] = 0x2514, [ 5242] = 0x2518, [ 5243] = 0x256D,
  [ 5244] = 0x256E, [ 5245] = 0x2570, [ 5246] = 0x256F, [ 5247] = 0x2550,
  [ 5248] = 0x255E, [ 5249] = 0x256A, [ 5250] = 0x2561, [ 5251] = 0x25E2,
  [ 5252] = 0x25E3, [ 5253] = 0x25E5, [ 5254] = 0x25E4, [ 5255] = 0x2571,
  [ 5256] = 0x2572, [ 5257] = 0x2573, [ 5258] = 0xFF10, [ 5259] = 0xFF11,
  [ 5260] = 0xFF12, [ 5261] = 0xFF13, [ 5262] = 0xFF14, [ 5263] = 0xFF15,
  [ 5264] = 0xFF16, [ 5265] = 0xFF17, [ 5266] = 0xFF18, [ 5267] = 0xFF19,
  [ 5268] = 0x2160, [ 5269] = 0x2161, [ 5270] = 0x2162, [ 5271] = 0x2163,
  [ 5272] = 0x2164, [ 5273] = 0x2165, [ 5274] = 0x2166, [ 5275] = 0x2167,
  [ 5276] = 0x2168, [ 5277] = 0x2169, [ 5278] = 0x3021, [ 5279] = 0x3022,
  [ 5280] = 0x3023, [ 5281] = 0x3024, [ 5282] = 0x3025, [ 5283] = 0x3026,
  [ 5284] = 0x3027, [ 5285] = 0x3028, [ 5286] = 0x3029, [ 5287] = 0x5341,
  [ 5288] = 0x5344, [ 5289] = 0x5345, [ 5290] = 0xFF21, [ 5291] = 0xFF22,
  [ 5292] = 0xFF23, [ 5293] = 0xFF24, [ 5294] = 0xFF25, [ 5295] = 0xFF26,
  [ 5296] = 0xFF27, [ 5297] = 0xFF28, [ 5298] = 0xFF29, [ 5299] = 0xFF2A,
  [ 5300] = 0xFF2B, [ 5301] = 0xFF2C, [ 5302] = 0xFF2D, [ 5303] = 0xFF2E,
  [ 5304] = 0xFF2F, [ 5305] = 0xFF30, [ 5306] = 0xFF31, [ 5307] = 0xFF32,
  [ 5308] = 0xFF33, [ 5309] = 0xFF34, [ 5310] = 0xFF35, [ 5311] = 0xFF36,
  [ 5312] = 0xFF37, [ 5313] = 0xFF38, [ 5314] = 0xFF39, [ 5315] = 0xFF3A,
  [ 5316] = 0xFF41, [ 5317] = 0xFF42, [ 5318] = 0xFF43, [ 5319] = 0xFF44,
  [ 5320] = 0xFF45, [ 5321] = 0xFF46, [ 5322] = 0xFF47, [ 5323] = 0xFF48,
  [ 5324] = 0xFF49, [ 5325] = 0xFF4A, [ 5326] = 0xFF4B, [ 5327] = 0xFF4C,
  [ 5328] = 0xFF4D, [ 5329] = 0xFF4E, [ 5330] = 0xFF4F, [ 5331] = 0xFF50,
  [ 5332] = 0xFF51, [ 5333] = 0xFF52, [ 5334] = 0xFF53, [ 5335] = 0xFF54,
  [ 5336] = 0xFF55, [ 5337] = 0xFF56, [ 5338] = 0xFF57, [ 5339] = 0xFF58,
  [ 5340] = 0xFF59, [ 5341] = 0xFF5A, [ 5342] = 0x0391, [ 5343] = 0x0392,
  [ 5344] = 0x0393, [ 5345] = 0x0394, [ 5346] = 0x0395, [ 5347] = 0x0396,
  [ 5348] = 0x0397, [ 5349] = 0x0398, [ 5350] = 0x0399, [ 5351] = 0x039A,
  [ 5352] = 0x039B, [ 5353] = 0x039C, [ 5354] = 0x039D, [ 5355] = 0x039E,
  [ 5356] = 0x039F, [ 5357] = 0x03A0, [ 5358] = 0x03A1, [ 5359] = 0x03A3,
  [ 5360] = 0x03A4, [ 5361] = 0x03A5, [ 5362] = 0x03A6, [ 5363] = 0x03A7,
  [ 5364] = 0x03A8, [ 5365] = 0x03A9, [ 5366] = 0x03B1, [ 5367] = 0x03B2,
  [ 5368] = 0x03B3, [ 5369] = 0x03B4, [ 5370] = 0x03B5, [ 5371] = 0x03B6,
  [ 5372] = 0x03B7, [ 5373] = 0x03B8, [ 5374] = 0x03B9, [ 5375] = 0x03BA,
  [ 5376] = 0x03BB, [ 5377] = 0x03BC, [ 5378] = 0x03BD, [ 5379] = 0x03BE,
  [ 5380] = 0x03BF, [ 5381] = 0x03C0, [ 5382] = 0x03C1, [ 5383] = 0x03C3,
  [ 5384] = 0x03C4, [ 5385] = 0x03C5, [ 5386] = 0x03C6, [ 5387] = 0x03C7,
  [ 5388] = 0x03C8, [ 5389] = 0x03C9, [ 5390] = 0x3105, [ 5391] = 0x3106,
  [ 5392] = 0x3107, [ 5393] = 0x3108, [ 5394] = 0x3109, [ 5395] = 0x310A,
  [ 5396] = 0x310B, [ 5397] = 0x310C, [ 5398] = 0x310D, [ 5399] = 0x310E,
  [ 5400] = 0x310F, [ 5401] = 0x3110, [ 5402] = 0x3111, [ 5403] = 0x3112,
  [ 5404] = 0x3113, [ 5405] = 0x3114, [ 5406] = 0x3115, [ 5407] = 0x3116,
  [ 5408] = 0x3117, [ 5409] = 0x3118, [ 5410] = 0x3119, [ 5411] = 0x311A,
  [ 5412] = 0x311B, [ 5413] = 0x311C, [ 5414] = 0x311D, [ 5415] = 0x311E,
  [ 5416] = 0x311F, [ 5417] = 0x3120, [ 5418] = 0x3121, [ 5419] = 0x3122,
  [ 5420] = 0x3123, [ 5421] = 0x3124, [ 5422] = 0x3125, [ 5423] = 0x3126,
  [ 5424] = 0x3127, [ 5425] = 0x3128, [ 5426] = 0x3129, [ 5427] = 0x02D9,
  [ 5428] = 0x02C9, [ 5429] = 0x02CA, [ 5430] = 0x02C7, [ 5431] = 0x02CB,
  [ 5495] = 0x4E00, [ 5496] = 0x4E59, [ 5497] = 0x4E01, [ 5498] = 0x4E03,
  [ 5499] = 0x4E43, [ 5500] = 0x4E5D, [ 5501] = 0x4E86, [ 5502] = 0x4E8C,
  [ 5503] = 0x4EBA, [ 5504] = 0x513F, [ 5505] = 0x5165, [ 5506] = 0x516B,
  [ 5507] = 0x51E0, [ 5508] = 0x5200, [ 5509] = 0x5201, [ 5510] = 0x529B,
  [ 5511] = 0x5315, [ 5512] = 0x5341, [ 5513] = 0x535C, [ 5514] = 0x53C8,
  [ 5515] = 0x4E09, [ 5516] = 0x4E0B, [ 5517] = 0x4E08, [ 5518] = 0x4E0A,
  [ 5519] = 0x4E2B, [ 5520] = 0x4E38, [ 5521] = 0x51E1, [ 5522] = 0x4E45,
  [ 5523] = 0x4E48, [ 5524] = 0x4E5F, [ 5525] = 0x4E5E, [ 5526] = 0x4E8E,
  [ 5527] = 0x4EA1, [ 5528] = 0x5140, [ 5529] = 0x5203, [ 5530] = 0x52FA,
  [ 5531] = 0x5343, [ 5532] = 0x53C9, [ 5533] = 0x53E3, [ 5534] = 0x571F,
  [ 5535] = 0x58EB, [ 5536] = 0x5915, [ 5537] = 0x5927, [ 5538] = 0x5973,
  [ 5539] = 0x5B50, [ 5540] = 0x5B51, [ 5541] = 0x5B53, [ 5542] = 0x5BF8,
  [ 5543] = 0x5C0F, [ 5544] = 0x5C22, [ 5545] = 0x5C38, [ 5546] = 0x5C71,
  [ 5547] = 0x5DDD, [ 5548] = 0x5DE5, [ 5549] = 0x5DF1, [ 5550] = 0x5DF2,
  [ 5551] = 0x5DF3, [ 5552] = 0x5DFE, [ 5553] = 0x5E72, [ 5554] = 0x5EFE,
  [ 5555] = 0x5F0B, [ 5556] = 0x5F13, [ 5557] = 0x624D, [ 5558] = 0x4E11,
  [ 5559] = 0x4E10, [ 5560] = 0x4E0D, [ 5561] = 0x4E2D, [ 5562] = 0x4E30,
  [ 5563] = 0x4E39, [ 5564] = 0x4E4B, [ 5565] = 0x5C39, [ 5566] = 0x4E88,
  [ 5567] = 0x4E91, [ 5568] = 0x4E95, [ 5569] = 0x4E92, [ 5570] = 0x4E94,
  [ 5571] = 0x4EA2, [ 5572] = 0x4EC1, [ 5573] = 0x4EC0, [ 5574] = 0x4EC3,
  [ 5575] = 0x4EC6, [ 5576] = 0x4EC7, [ 5577] = 0x4ECD, [ 5578] = 0x4ECA,
  [ 5579] = 0x4ECB, [ 5580] = 0x4EC4, [ 5581] = 0x5143, [ 5582] = 0x5141,
  [ 5583] = 0x5167, [ 5584] = 0x516D, [ 5585] = 0x516E, [ 5586] = 0x516C,
  [ 5587] = 0x5197, [ 5588] = 0x51F6, [ 5589] = 0x5206, [ 5590] = 0x5207,
  [ 5591] = 0x5208, [ 5592] = 0x52FB, [ 5593] = 0x52FE, [ 5594] = 0x52FF,
  [ 5595] = 0x5316, [ 5596] = 0x5339, [ 5597] = 0x5348, [ 5598] = 0x5347,
  [ 5599] = 0x5345, [ 5600] = 0x535E, [ 5601] = 0x5384, [ 5602] = 0x53CB,
  [ 5603] = 0x53CA, [ 5604] = 0x53CD, [ 5605] = 0x58EC, [ 5606] = 0x5929,
  [ 5607] = 0x592B, [ 5608] = 0x592A, [ 5609] = 0x592D, [ 5610] = 0x5B54,
  [ 5611] = 0x5C11, [ 5612] = 0x5C24, [ 5613] = 0x5C3A, [ 5614] = 0x5C6F,
  [ 5615] = 0x5DF4, [ 5616] = 0x5E7B, [ 5617] = 0x5EFF, [ 5618] = 0x5F14,
  [ 5619] = 0x5F15, [ 5620] = 0x5FC3, [ 5621] = 0x6208, [ 5622] = 0x6236,
  [ 5623] = 0x624B, [ 5624] = 0x624E, [ 5625] = 0x652F, [ 5626] = 0x6587,
  [ 5627] = 0x6597, [ 5628] = 0x65A4, [ 5629] = 0x65B9, [ 5630] = 0x65E5,
  [ 5631] = 0x66F0, [ 5632] = 0x6708, [ 5633] = 0x6728, [ 5634] = 0x6B20,
  [ 5635] = 0x6B62, [ 5636] = 0x6B79, [ 5637] = 0x6BCB, [ 5638] = 0x6BD4,
  [ 5639] = 0x6BDB, [ 5640] = 0x6C0F, [ 5641] = 0x6C34, [ 5642] = 0x706B,
  [ 5643] = 0x722A, [ 5644] = 0x7236, [ 5645] = 0x723B, [ 5646] = 0x7247,
  [ 5647] = 0x7259, [ 5648] = 0x725B, [ 5649] = 0x72AC, [ 5650] = 0x738B,
  [ 5651] = 0x4E19, [ 5652] = 0x4E16, [ 5653] = 0x4E15, [ 5654] = 0x4E14,
  [ 5655] = 0x4E18, [ 5656] = 0x4E3B, [ 5657] = 0x4E4D, [ 5658] = 0x4E4F,
  [ 5659] = 0x4E4E, [ 5660] = 0x4EE5, [ 5661] = 0x4ED8, [ 5662] = 0x4ED4,
  [ 5663] = 0x4ED5, [ 5664] = 0x4ED6, [ 5665] = 0x4ED7, [ 5666] = 0x4EE3,
  [ 5667] = 0x4EE4, [ 5668] = 0x4ED9, [ 5669] = 0x4EDE, [ 5670] = 0x5145,
  [ 5671] = 0x5144, [ 5672] = 0x5189, [ 5673] = 0x518A, [ 5674] = 0x51AC,
  [ 5675] = 0x51F9, [ 5676] = 0x51FA, [ 5677] = 0x51F8, [ 5678] = 0x520A,
  [ 5679] = 0x52A0, [ 5680] = 0x529F, [ 5681] = 0x5305, [ 5682] = 0x5306,
  [ 5683] = 0x5317, [ 5684] = 0x531D, [ 5685] = 0x4EDF, [ 5686] = 0x534A,
  [ 5687] = 0x5349, [ 5688] = 0x5361, [ 5689] = 0x5360, [ 5690] = 0x536F,
  [ 5691] = 0x536E, [ 5692] = 0x53BB, [ 5693] = 0x53EF, [ 5694] = 0x53E4,
  [ 5695] = 0x53F3, [ 5696] = 0x53EC, [ 5697] = 0x53EE, [ 5698] = 0x53E9,
  [ 5699] = 0x53E8, [ 5700] = 0x53FC, [ 5701] = 0x53F8, [ 5702] = 0x53F5,
  [ 5703] = 0x53EB, [ 5704] = 0x53E6, [ 5705] = 0x53EA, [ 5706] = 0x53F2,
  [ 5707] = 0x53F1, [ 5708] = 0x53F0, [ 5709] = 0x53E5, [ 5710] = 0x53ED,
  [ 5711] = 0x53FB, [ 5712] = 0x56DB, [ 5713] = 0x56DA, [ 5714] = 0x5916,
  [ 5715] = 0x592E, [ 5716] = 0x5931, [ 5717] = 0x5974, [ 5718] = 0x5976,
  [ 5719] = 0x5B55, [ 5720] = 0x5B83, [ 5721] = 0x5C3C, [ 5722] = 0x5DE8,
  [ 5723] = 0x5DE7, [ 5724] = 0x5DE6, [ 5725] = 0x5E02, [ 5726] = 0x5E03,
  [ 5727] = 0x5E73, [ 5728] = 0x5E7C, [ 5729] = 0x5F01, [ 5730] = 0x5F18,
  [ 5731] = 0x5F17, [ 5732] = 0x5FC5, [ 5733] = 0x620A, [ 5734] = 0x6253,
  [ 5735] = 0x6254, [ 5736] = 0x6252, [ 5737] = 0x6251, [ 5738] = 0x65A5,
  [ 5739] = 0x65E6, [ 5740] = 0x672E, [ 5741] = 0x672C, [ 5742] = 0x672A,
  [ 5743] = 0x672B, [ 5744] = 0x672D, [ 5745] = 0x6B63, [ 5746] = 0x6BCD,
  [ 5747] = 0x6C11, [ 5748] = 0x6C10, [ 5749] = 0x6C38, [ 5750] = 0x6C41,
  [ 5751] = 0x6C40, [ 5752] = 0x6C3E, [ 5753] = 0x72AF, [ 5754] = 0x7384,
  [ 5755] = 0x7389, [ 5756] = 0x74DC, [ 5757] = 0x74E6, [ 5758] = 0x7518,
  [ 5759] = 0x751F, [ 5760] = 0x7528, [ 5761] = 0x7529, [ 5762] = 0x7530,
  [ 5763] = 0x7531, [ 5764] = 0x7532, [ 5765] = 0x7533, [ 5766] = 0x758B,
  [ 5767] = 0x767D, [ 5768] = 0x76AE, [ 5769] = 0x76BF, [ 5770] = 0x76EE,
  [ 5771] = 0x77DB, [ 5772] = 0x77E2, [ 5773] = 0x77F3, [ 5774] = 0x793A,
  [ 5775] = 0x79BE, [ 5776] = 0x7A74, [ 5777] = 0x7ACB, [ 5778] = 0x4E1E,
  [ 5779] = 0x4E1F, [ 5780] = 0x4E52, [ 5781] = 0x4E53, [ 5782] = 0x4E69,
  [ 5783] = 0x4E99, [ 5784] = 0x4EA4, [ 5785] = 0x4EA6, [ 5786] = 0x4EA5,
  [ 5787] = 0x4EFF, [ 5788] = 0x4F09, [ 5789] = 0x4F19, [ 5790] = 0x4F0A,
  [ 5791] = 0x4F15, [ 5792] = 0x4F0D, [ 5793] = 0x4F10, [ 5794] = 0x4F11,
  [ 5795] = 0x4F0F, [ 5796] = 0x4EF2, [ 5797] = 0x4EF6, [ 5798] = 0x4EFB,
  [ 5799] = 0x4EF0, [ 5800] = 0x4EF3, [ 5801] = 0x4EFD, [ 5802] = 0x4F01,
  [ 5803] = 0x4F0B, [ 5804] = 0x5149, [ 5805] = 0x5147, [ 5806] = 0x5146,
  [ 5807] = 0x5148, [ 5808] = 0x5168, [ 5809] = 0x5171, [ 5810] = 0x518D,
  [ 5811] = 0x51B0, [ 5812] = 0x5217, [ 5813] = 0x5211, [ 5814] = 0x5212,
  [ 5815] = 0x520E, [ 5816] = 0x5216, [ 5817] = 0x52A3, [ 5818] = 0x5308,
  [ 5819] = 0x5321, [ 5820] = 0x5320, [ 5821] = 0x5370, [ 5822] = 0x5371,
  [ 5823] = 0x5409, [ 5824] = 0x540F, [ 5825] = 0x540C, [ 5826] = 0x540A,
  [ 5827] = 0x5410, [ 5828] = 0x5401, [ 5829] = 0x540B, [ 5830] = 0x5404,
  [ 5831] = 0x5411, [ 5832] = 0x540D, [ 5833] = 0x5408, [ 5834] = 0x5403,
  [ 5835] = 0x540E, [ 5836] = 0x5406, [ 5837] = 0x5412, [ 5838] = 0x56E0,
  [ 5839] = 0x56DE, [ 5840] = 0x56DD, [ 5841] = 0x5733, [ 5842] = 0x5730,
  [ 5843] = 0x5728, [ 5844] = 0x572D, [ 5845] = 0x572C, [ 5846] = 0x572F,
  [ 5847] = 0x5729, [ 5848] = 0x5919, [ 5849] = 0x591A, [ 5850] = 0x5937,
  [ 5851] = 0x5938, [ 5852] = 0x5984, [ 5853] = 0x5978, [ 5854] = 0x5983,
  [ 5855] = 0x597D, [ 5856] = 0x5979, [ 5857] = 0x5982, [ 5858] = 0x5981,
  [ 5859] = 0x5B57, [ 5860] = 0x5B58, [ 5861] = 0x5B87, [ 5862] = 0x5B88,
  [ 5863] = 0x5B85, [ 5864] = 0x5B89, [ 5865] = 0x5BFA, [ 5866] = 0x5C16,
  [ 5867] = 0x5C79, [ 5868] = 0x5DDE, [ 5869] = 0x5E06, [ 5870] = 0x5E76,
  [ 5871] = 0x5E74, [ 5872] = 0x5F0F, [ 5873] = 0x5F1B, [ 5874] = 0x5FD9,
  [ 5875] = 0x5FD6, [ 5876] = 0x620E, [ 5877] = 0x620C, [ 5878] = 0x620D,
  [ 5879] = 0x6210, [ 5880] = 0x6263, [ 5881] = 0x625B, [ 5882] = 0x6258,
  [ 5883] = 0x6536, [ 5884] = 0x65E9, [ 5885] = 0x65E8, [ 5886] = 0x65EC,
  [ 5887] = 0x65ED, [ 5888] = 0x66F2, [ 5889] = 0x66F3, [ 5890] = 0x6709,
  [ 5891] = 0x673D, [ 5892] = 0x6734, [ 5893] = 0x6731, [ 5894] = 0x6735,
  [ 5895] = 0x6B21, [ 5896] = 0x6B64, [ 5897] = 0x6B7B, [ 5898] = 0x6C16,
  [ 5899] = 0x6C5D, [ 5900] = 0x6C57, [ 5901] = 0x6C59, [ 5902] = 0x6C5F,
  [ 5903] = 0x6C60, [ 5904] = 0x6C50, [ 5905] = 0x6C55, [ 5906] = 0x6C61,
  [ 5907] = 0x6C5B, [ 5908] = 0x6C4D, [ 5909] = 0x6C4E, [ 5910] = 0x7070,
  [ 5911] = 0x725F, [ 5912] = 0x725D, [ 5913] = 0x767E, [ 5914] = 0x7AF9,
  [ 5915] = 0x7C73, [ 5916] = 0x7CF8, [ 5917] = 0x7F36, [ 5918] = 0x7F8A,
  [ 5919] = 0x7FBD, [ 5920] = 0x8001, [ 5921] = 0x8003, [ 5922] = 0x800C,
  [ 5923] = 0x8012, [ 5924] = 0x8033, [ 5925] = 0x807F, [ 5926] = 0x8089,
  [ 5927] = 0x808B, [ 5928] = 0x808C, [ 5929] = 0x81E3, [ 5930] = 0x81EA,
  [ 5931] = 0x81F3, [ 5932] = 0x81FC, [ 5933] = 0x820C, [ 5934] = 0x821B,
  [ 5935] = 0x821F, [ 5936] = 0x826E, [ 5937] = 0x8272, [ 5938] = 0x827E,
  [ 5939] = 0x866B, [ 5940] = 0x8840, [ 5941] = 0x884C, [ 5942] = 0x8863,
  [ 5943] = 0x897F, [ 5944] = 0x9621, [ 5945] = 0x4E32, [ 5946] = 0x4EA8,
  [ 5947] = 0x4F4D, [ 5948] = 0x4F4F, [ 5949] = 0x4F47, [ 5950] = 0x4F57,
  [ 5951] = 0x4F5E, [ 5952] = 0x4F34, [ 5953] = 0x4F5B, [ 5954] = 0x4F55,
  [ 5955] = 0x4F30, [ 5956] = 0x4F50, [ 5957] = 0x4F51, [ 5958] = 0x4F3D,
  [ 5959] = 0x4F3A, [ 5960] = 0x4F38, [ 5961] = 0x4F43, [ 5962] = 0x4F54,
  [ 5963] = 0x4F3C, [ 5964] = 0x4F46, [ 5965] = 0x4F63, [ 5966] = 0x4F5C,
  [ 5967] = 0x4F60, [ 5968] = 0x4F2F, [ 5969] = 0x4F4E, [ 5970] = 0x4F36,
  [ 5971] = 0x4F59, [ 5972] = 0x4F5D, [ 5973] = 0x4F48, [ 5974] = 0x4F5A,
  [ 5975] = 0x514C, [ 5976] = 0x514B, [ 5977] = 0x514D, [ 5978] = 0x5175,
  [ 5979] = 0x51B6, [ 5980] = 0x51B7, [ 5981] = 0x5225, [ 5982] = 0x5224,
  [ 5983] = 0x5229, [ 5984] = 0x522A, [ 5985] = 0x5228, [ 5986] = 0x52AB,
  [ 5987] = 0x52A9, [ 5988] = 0x52AA, [ 5989] = 0x52AC, [ 5990] = 0x5323,
  [ 5991] = 0x5373, [ 5992] = 0x5375, [ 5993] = 0x541D, [ 5994] = 0x542D,
  [ 5995] = 0x541E, [ 5996] = 0x543E, [ 5997] = 0x5426, [ 5998] = 0x544E,
  [ 5999] = 0x5427, [ 6000] = 0x5446, [ 6001] = 0x5443, [ 6002] = 0x5433,
  [ 6003] = 0x5448, [ 6004] = 0x5442, [ 6005] = 0x541B, [ 6006] = 0x5429,
  [ 6007] = 0x544A, [ 6008] = 0x5439, [ 6009] = 0x543B, [ 6010] = 0x5438,
  [ 6011] = 0x542E, [ 6012] = 0x5435, [ 6013] = 0x5436, [ 6014] = 0x5420,
  [ 6015] = 0x543C, [ 6016] = 0x5440, [ 6017] = 0x5431, [ 6018] = 0x542B,
  [ 6019] = 0x541F, [ 6020] = 0x542C, [ 6021] = 0x56EA, [ 6022] = 0x56F0,
  [ 6023] = 0x56E4, [ 6024] = 0x56EB, [ 6025] = 0x574A, [ 6026] = 0x5751,
  [ 6027] = 0x5740, [ 6028] = 0x574D, [ 6029] = 0x5747, [ 6030] = 0x574E,
  [ 6031] = 0x573E, [ 6032] = 0x5750, [ 6033] = 0x574F, [ 6034] = 0x573B,
  [ 6035] = 0x58EF, [ 6036] = 0x593E, [ 6037] = 0x599D, [ 6038] = 0x5992,
  [ 6039] = 0x59A8, [ 6040] = 0x599E, [ 6041] = 0x59A3, [ 6042] = 0x5999,
  [ 6043] = 0x5996, [ 6044] = 0x598D, [ 6045] = 0x59A4, [ 6046] = 0x5993,
  [ 6047] = 0x598A, [ 6048] = 0x59A5, [ 6049] = 0x5B5D, [ 6050] = 0x5B5C,
  [ 6051] = 0x5B5A, [ 6052] = 0x5B5B, [ 6053] = 0x5B8C, [ 6054] = 0x5B8B,
  [ 6055] = 0x5B8F, [ 6056] = 0x5C2C, [ 6057] = 0x5C40, [ 6058] = 0x5C41,
  [ 6059] = 0x5C3F, [ 6060] = 0x5C3E, [ 6061] = 0x5C90, [ 6062] = 0x5C91,
  [ 6063] = 0x5C94, [ 6064] = 0x5C8C, [ 6065] = 0x5DEB, [ 6066] = 0x5E0C,
  [ 6067] = 0x5E8F, [ 6068] = 0x5E87, [ 6069] = 0x5E8A, [ 6070] = 0x5EF7,
  [ 6071] = 0x5F04, [ 6072] = 0x5F1F, [ 6073] = 0x5F64, [ 6074] = 0x5F62,
  [ 6075] = 0x5F77, [ 6076] = 0x5F79, [ 6077] = 0x5FD8, [ 6078] = 0x5FCC,
  [ 6079] = 0x5FD7, [ 6080] = 0x5FCD, [ 6081] = 0x5FF1, [ 6082] = 0x5FEB,
  [ 6083] = 0x5FF8, [ 6084] = 0x5FEA, [ 6085] = 0x6212, [ 6086] = 0x6211,
  [ 6087] = 0x6284, [ 6088] = 0x6297, [ 6089] = 0x6296, [ 6090] = 0x6280,
  [ 6091] = 0x6276, [ 6092] = 0x6289, [ 6093] = 0x626D, [ 6094] = 0x628A,
  [ 6095] = 0x627C, [ 6096] = 0x627E, [ 6097] = 0x6279, [ 6098] = 0x6273,
  [ 6099] = 0x6292, [ 6100] = 0x626F, [ 6101] = 0x6298, [ 6102] = 0x626E,
  [ 6103] = 0x6295, [ 6104] = 0x6293, [ 6105] = 0x6291, [ 6106] = 0x6286,
  [ 6107] = 0x6539, [ 6108] = 0x653B, [ 6109] = 0x6538, [ 6110] = 0x65F1,
  [ 6111] = 0x66F4, [ 6112] = 0x675F, [ 6113] = 0x674E, [ 6114] = 0x674F,
  [ 6115] = 0x6750, [ 6116] = 0x6751, [ 6117] = 0x675C, [ 6118] = 0x6756,
  [ 6119] = 0x675E, [ 6120] = 0x6749, [ 6121] = 0x6746, [ 6122] = 0x6760,
  [ 6123] = 0x6753, [ 6124] = 0x6757, [ 6125] = 0x6B65, [ 6126] = 0x6BCF,
  [ 6127] = 0x6C42, [ 6128] = 0x6C5E, [ 6129] = 0x6C99, [ 6130] = 0x6C81,
  [ 6131] = 0x6C88, [ 6132] = 0x6C89, [ 6133] = 0x6C85, [ 6134] = 0x6C9B,
  [ 6135] = 0x6C6A, [ 6136] = 0x6C7A, [ 6137] = 0x6C90, [ 6138] = 0x6C70,
  [ 6139] = 0x6C8C, [ 6140] = 0x6C68, [ 6141] = 0x6C96, [ 6142] = 0x6C92,
  [ 6143] = 0x6C7D, [ 6144] = 0x6C83, [ 6145] = 0x6C72, [ 6146] = 0x6C7E,
  [ 6147] = 0x6C74, [ 6148] = 0x6C86, [ 6149] = 0x6C76, [ 6150] = 0x6C8D,
  [ 6151] = 0x6C94, [ 6152] = 0x6C98, [ 6153] = 0x6C82, [ 6154] = 0x7076,
  [ 6155] = 0x707C, [ 6156] = 0x707D, [ 6157] = 0x7078, [ 6158] = 0x7262,
  [ 6159] = 0x7261, [ 6160] = 0x7260, [ 6161] = 0x72C4, [ 6162] = 0x72C2,
  [ 6163] = 0x7396, [ 6164] = 0x752C, [ 6165] = 0x752B, [ 6166] = 0x7537,
  [ 6167] = 0x7538, [ 6168] = 0x7682, [ 6169] = 0x76EF, [ 6170] = 0x77E3,
  [ 6171] = 0x79C1, [ 6172] = 0x79C0, [ 6173] = 0x79BF, [ 6174] = 0x7A76,
  [ 6175] = 0x7CFB, [ 6176] = 0x7F55, [ 6177] = 0x8096, [ 6178] = 0x8093,
  [ 6179] = 0x809D, [ 6180] = 0x8098, [ 6181] = 0x809B, [ 6182] = 0x809A,
  [ 6183] = 0x80B2, [ 6184] = 0x826F, [ 6185] = 0x8292, [ 6186] = 0x828B,
  [ 6187] = 0x828D, [ 6188] = 0x898B, [ 6189] = 0x89D2, [ 6190] = 0x8A00,
  [ 6191] = 0x8C37, [ 6192] = 0x8C46, [ 6193] = 0x8C55, [ 6194] = 0x8C9D,
  [ 6195] = 0x8D64, [ 6196] = 0x8D70, [ 6197] = 0x8DB3, [ 6198] = 0x8EAB,
  [ 6199] = 0x8ECA, [ 6200] = 0x8F9B, [ 6201] = 0x8FB0, [ 6202] = 0x8FC2,
  [ 6203] = 0x8FC6, [ 6204] = 0x8FC5, [ 6205] = 0x8FC4, [ 6206] = 0x5DE1,
  [ 6207] = 0x9091, [ 6208] = 0x90A2, [ 6209] = 0x90AA, [ 6210] = 0x90A6,
  [ 6211] = 0x90A3, [ 6212] = 0x9149, [ 6213] = 0x91C6, [ 6214] = 0x91CC,
  [ 6215] = 0x9632, [ 6216] = 0x962E, [ 6217] = 0x9631, [ 6218] = 0x962A,
  [ 6219] = 0x962C, [ 6220] = 0x4E26, [ 6221] = 0x4E56, [ 6222] = 0x4E73,
  [ 6223] = 0x4E8B, [ 6224] = 0x4E9B, [ 6225] = 0x4E9E, [ 6226] = 0x4EAB,
  [ 6227] = 0x4EAC, [ 6228] = 0x4F6F, [ 6229] = 0x4F9D, [ 6230] = 0x4F8D,
  [ 6231] = 0x4F73, [ 6232] = 0x4F7F, [ 6233] = 0x4F6C, [ 6234] = 0x4F9B,
  [ 6235] = 0x4F8B, [ 6236] = 0x4F86, [ 6237] = 0x4F83, [ 6238] = 0x4F70,
  [ 6239] = 0x4F75, [ 6240] = 0x4F88, [ 6241] = 0x4F69, [ 6242] = 0x4F7B,
  [ 6243] = 0x4F96, [ 6244] = 0x4F7E, [ 6245] = 0x4F8F, [ 6246] = 0x4F91,
  [ 6247] = 0x4F7A, [ 6248] = 0x5154, [ 6249] = 0x5152, [ 6250] = 0x5155,
  [ 6251] = 0x5169, [ 6252] = 0x5177, [ 6253] = 0x5176, [ 6254] = 0x5178,
  [ 6255] = 0x51BD, [ 6256] = 0x51FD, [ 6257] = 0x523B, [ 6258] = 0x5238,
  [ 6259] = 0x5237, [ 6260] = 0x523A, [ 6261] = 0x5230, [ 6262] = 0x522E,
  [ 6263] = 0x5236, [ 6264] = 0x5241, [ 6265] = 0x52BE, [ 6266] = 0x52BB,
  [ 6267] = 0x5352, [ 6268] = 0x5354, [ 6269] = 0x5353, [ 6270] = 0x5351,
  [ 6271] = 0x5366, [ 6272] = 0x5377, [ 6273] = 0x5378, [ 6274] = 0x5379,
  [ 6275] = 0x53D6, [ 6276] = 0x53D4, [ 6277] = 0x53D7, [ 6278] = 0x5473,
  [ 6279] = 0x5475, [ 6280] = 0x5496, [ 6281] = 0x5478, [ 6282] = 0x5495,
  [ 6283] = 0x5480, [ 6284] = 0x547B, [ 6285] = 0x5477, [ 6286] = 0x5484,
  [ 6287] = 0x5492, [ 6288] = 0x5486, [ 6289] = 0x547C, [ 6290] = 0x5490,
  [ 6291] = 0x5471, [ 6292] = 0x5476, [ 6293] = 0x548C, [ 6294] = 0x549A,
  [ 6295] = 0x5462, [ 6296] = 0x5468, [ 6297] = 0x548B, [ 6298] = 0x547D,
  [ 6299] = 0x548E, [ 6300] = 0x56FA, [ 6301] = 0x5783, [ 6302] = 0x5777,
  [ 6303] = 0x576A, [ 6304] = 0x5769, [ 6305] = 0x5761, [ 6306] = 0x5766,
  [ 6307] = 0x5764, [ 6308] = 0x577C, [ 6309] = 0x591C, [ 6310] = 0x5949,
  [ 6311] = 0x5947, [ 6312] = 0x5948, [ 6313] = 0x5944, [ 6314] = 0x5954,
  [ 6315] = 0x59BE, [ 6316] = 0x59BB, [ 6317] = 0x59D4, [ 6318] = 0x59B9,
  [ 6319] = 0x59AE, [ 6320] = 0x59D1, [ 6321] = 0x59C6, [ 6322] = 0x59D0,
  [ 6323] = 0x59CD, [ 6324] = 0x59CB, [ 6325] = 0x59D3, [ 6326] = 0x59CA,
  [ 6327] = 0x59AF, [ 6328] = 0x59B3, [ 6329] = 0x59D2, [ 6330] = 0x59C5,
  [ 6331] = 0x5B5F, [ 6332] = 0x5B64, [ 6333] = 0x5B63, [ 6334] = 0x5B97,
  [ 6335] = 0x5B9A, [ 6336] = 0x5B98, [ 6337] = 0x5B9C, [ 6338] = 0x5B99,
  [ 6339] = 0x5B9B, [ 6340] = 0x5C1A, [ 6341] = 0x5C48, [ 6342] = 0x5C45,
  [ 6343] = 0x5C46, [ 6344] = 0x5CB7, [ 6345] = 0x5CA1, [ 6346] = 0x5CB8,
  [ 6347] = 0x5CA9, [ 6348] = 0x5CAB, [ 6349] = 0x5CB1, [ 6350] = 0x5CB3,
  [ 6351] = 0x5E18, [ 6352] = 0x5E1A, [ 6353] = 0x5E16, [ 6354] = 0x5E15,
  [ 6355] = 0x5E1B, [ 6356] = 0x5E11, [ 6357] = 0x5E78, [ 6358] = 0x5E9A,
  [ 6359] = 0x5E97, [ 6360] = 0x5E9C, [ 6361] = 0x5E95, [ 6362] = 0x5E96,
  [ 6363] = 0x5EF6, [ 6364] = 0x5F26, [ 6365] = 0x5F27, [ 6366] = 0x5F29,
  [ 6367] = 0x5F80, [ 6368] = 0x5F81, [ 6369] = 0x5F7F, [ 6370] = 0x5F7C,
  [ 6371] = 0x5FDD, [ 6372] = 0x5FE0, [ 6373] = 0x5FFD, [ 6374] = 0x5FF5,
  [ 6375] = 0x5FFF, [ 6376] = 0x600F, [ 6377] = 0x6014, [ 6378] = 0x602F,
  [ 6379] = 0x6035, [ 6380] = 0x6016, [ 6381] = 0x602A, [ 6382] = 0x6015,
  [ 6383] = 0x6021, [ 6384] = 0x6027, [ 6385] = 0x6029, [ 6386] = 0x602B,
  [ 6387] = 0x601B, [ 6388] = 0x6216, [ 6389] = 0x6215, [ 6390] = 0x623F,
  [ 6391] = 0x623E, [ 6392] = 0x6240, [ 6393] = 0x627F, [ 6394] = 0x62C9,
  [ 6395] = 0x62CC, [ 6396] = 0x62C4, [ 6397] = 0x62BF, [ 6398] = 0x62C2,
  [ 6399] = 0x62B9, [ 6400] = 0x62D2, [ 6401] = 0x62DB, [ 6402] = 0x62AB,
  [ 6403] = 0x62D3, [ 6404] = 0x62D4, [ 6405] = 0x62CB, [ 6406] = 0x62C8,
  [ 6407] = 0x62A8, [ 6408] = 0x62BD, [ 6409] = 0x62BC, [ 6410] = 0x62D0,
  [ 6411] = 0x62D9, [ 6412] = 0x62C7, [ 6413] = 0x62CD, [ 6414] = 0x62B5,
  [ 6415] = 0x62DA, [ 6416] = 0x62B1, [ 6417] = 0x62D8, [ 6418] = 0x62D6,
  [ 6419] = 0x62D7, [ 6420] = 0x62C6, [ 6421] = 0x62AC, [ 6422] = 0x62CE,
  [ 6423] = 0x653E, [ 6424] = 0x65A7, [ 6425] = 0x65BC, [ 6426] = 0x65FA,
  [ 6427] = 0x6614, [ 6428] = 0x6613, [ 6429] = 0x660C, [ 6430] = 0x6606,
  [ 6431] = 0x6602, [ 6432] = 0x660E, [ 6433] = 0x6600, [ 6434] = 0x660F,
  [ 6435] = 0x6615, [ 6436] = 0x660A, [ 6437] = 0x6607, [ 6438] = 0x670D,
  [ 6439] = 0x670B, [ 6440] = 0x676D, [ 6441] = 0x678B, [ 6442] = 0x6795,
  [ 6443] = 0x6771, [ 6444] = 0x679C, [ 6445] = 0x6773, [ 6446] = 0x6777,
  [ 6447] = 0x6787, [ 6448] = 0x679D, [ 6449] = 0x6797, [ 6450] = 0x676F,
  [ 6451] = 0x6770, [ 6452] = 0x677F, [ 6453] = 0x6789, [ 6454] = 0x677E,
  [ 6455] = 0x6790, [ 6456] = 0x6775, [ 6457] = 0x679A, [ 6458] = 0x6793,
  [ 6459] = 0x677C, [ 6460] = 0x676A, [ 6461] = 0x6772, [ 6462] = 0x6B23,
  [ 6463] = 0x6B66, [ 6464] = 0x6B67, [ 6465] = 0x6B7F, [ 6466] = 0x6C13,
  [ 6467] = 0x6C1B, [ 6468] = 0x6CE3, [ 6469] = 0x6CE8, [ 6470] = 0x6CF3,
  [ 6471] = 0x6CB1, [ 6472] = 0x6CCC, [ 6473] = 0x6CE5, [ 6474] = 0x6CB3,
  [ 6475] = 0x6CBD, [ 6476] = 0x6CBE, [ 6477] = 0x6CBC, [ 6478] = 0x6CE2,
  [ 6479] = 0x6CAB, [ 6480] = 0x6CD5, [ 6481] = 0x6CD3, [ 6482] = 0x6CB8,
  [ 6483] = 0x6CC4, [ 6484] = 0x6CB9, [ 6485] = 0x6CC1, [ 6486] = 0x6CAE,
  [ 6487] = 0x6CD7, [ 6488] = 0x6CC5, [ 6489] = 0x6CF1, [ 6490] = 0x6CBF,
  [ 6491] = 0x6CBB, [ 6492] = 0x6CE1, [ 6493] = 0x6CDB, [ 6494] = 0x6CCA,
  [ 6495] = 0x6CAC, [ 6496] = 0x6CEF, [ 6497] = 0x6CDC, [ 6498] = 0x6CD6,
  [ 6499] = 0x6CE0, [ 6500] = 0x7095, [ 6501] = 0x708E, [ 6502] = 0x7092,
  [ 6503] = 0x708A, [ 6504] = 0x7099, [ 6505] = 0x722C, [ 6506] = 0x722D,
  [ 6507] = 0x7238, [ 6508] = 0x7248, [ 6509] = 0x7267, [ 6510] = 0x7269,
  [ 6511] = 0x72C0, [ 6512] = 0x72CE, [ 6513] = 0x72D9, [ 6514] = 0x72D7,
  [ 6515] = 0x72D0, [ 6516] = 0x73A9, [ 6517] = 0x73A8, [ 6518] = 0x739F,
  [ 6519] = 0x73AB, [ 6520] = 0x73A5, [ 6521] = 0x753D, [ 6522] = 0x759D,
  [ 6523] = 0x7599, [ 6524] = 0x759A, [ 6525] = 0x7684, [ 6526] = 0x76C2,
  [ 6527] = 0x76F2, [ 6528] = 0x76F4, [ 6529] = 0x77E5, [ 6530] = 0x77FD,
  [ 6531] = 0x793E, [ 6532] = 0x7940, [ 6533] = 0x7941, [ 6534] = 0x79C9,
  [ 6535] = 0x79C8, [ 6536] = 0x7A7A, [ 6537] = 0x7A79, [ 6538] = 0x7AFA,
  [ 6539] = 0x7CFE, [ 6540] = 0x7F54, [ 6541] = 0x7F8C, [ 6542] = 0x7F8B,
  [ 6543] = 0x8005, [ 6544] = 0x80BA, [ 6545] = 0x80A5, [ 6546] = 0x80A2,
  [ 6547] = 0x80B1, [ 6548] = 0x80A1, [ 6549] = 0x80AB, [ 6550] = 0x80A9,
  [ 6551] = 0x80B4, [ 6552] = 0x80AA, [ 6553] = 0x80AF, [ 6554] = 0x81E5,
  [ 6555] = 0x81FE, [ 6556] = 0x820D, [ 6557] = 0x82B3, [ 6558] = 0x829D,
  [ 6559] = 0x8299, [ 6560] = 0x82AD, [ 6561] = 0x82BD, [ 6562] = 0x829F,
  [ 6563] = 0x82B9, [ 6564] = 0x82B1, [ 6565] = 0x82AC, [ 6566] = 0x82A5,
  [ 6567] = 0x82AF, [ 6568] = 0x82B8, [ 6569] = 0x82A3, [ 6570] = 0x82B0,
  [ 6571] = 0x82BE, [ 6572] = 0x82B7, [ 6573] = 0x864E, [ 6574] = 0x8671,
  [ 6575] = 0x521D, [ 6576] = 0x8868, [ 6577] = 0x8ECB, [ 6578] = 0x8FCE,
  [ 6579] = 0x8FD4, [ 6580] = 0x8FD1, [ 6581] = 0x90B5, [ 6582] = 0x90B8,
  [ 6583] = 0x90B1, [ 6584] = 0x90B6, [ 6585] = 0x91C7, [ 6586] = 0x91D1,
  [ 6587] = 0x9577, [ 6588] = 0x9580, [ 6589] = 0x961C, [ 6590] = 0x9640,
  [ 6591] = 0x963F, [ 6592] = 0x963B, [ 6593] = 0x9644, [ 6594] = 0x9642,
  [ 6595] = 0x96B9, [ 6596] = 0x96E8, [ 6597] = 0x9752, [ 6598] = 0x975E,
  [ 6599] = 0x4E9F, [ 6600] = 0x4EAD, [ 6601] = 0x4EAE, [ 6602] = 0x4FE1,
  [ 6603] = 0x4FB5, [ 6604] = 0x4FAF, [ 6605] = 0x4FBF, [ 6606] = 0x4FE0,
  [ 6607] = 0x4FD1, [ 6608] = 0x4FCF, [ 6609] = 0x4FDD, [ 6610] = 0x4FC3,
  [ 6611] = 0x4FB6, [ 6612] = 0x4FD8, [ 6613] = 0x4FDF, [ 6614] = 0x4FCA,
  [ 6615] = 0x4FD7, [ 6616] = 0x4FAE, [ 6617] = 0x4FD0, [ 6618] = 0x4FC4,
  [ 6619] = 0x4FC2, [ 6620] = 0x4FDA, [ 6621] = 0x4FCE, [ 6622] = 0x4FDE,
  [ 6623] = 0x4FB7, [ 6624] = 0x5157, [ 6625] = 0x5192, [ 6626] = 0x5191,
  [ 6627] = 0x51A0, [ 6628] = 0x524E, [ 6629] = 0x5243, [ 6630] = 0x524A,
  [ 6631] = 0x524D, [ 6632] = 0x524C, [ 6633] = 0x524B, [ 6634] = 0x5247,
  [ 6635] = 0x52C7, [ 6636] = 0x52C9, [ 6637] = 0x52C3, [ 6638] = 0x52C1,
  [ 6639] = 0x530D, [ 6640] = 0x5357, [ 6641] = 0x537B, [ 6642] = 0x539A,
  [ 6643] = 0x53DB, [ 6644] = 0x54AC, [ 6645] = 0x54C0, [ 6646] = 0x54A8,
  [ 6647] = 0x54CE, [ 6648] = 0x54C9, [ 6649] = 0x54B8, [ 6650] = 0x54A6,
  [ 6651] = 0x54B3, [ 6652] = 0x54C7, [ 6653] = 0x54C2, [ 6654] = 0x54BD,
  [ 6655] = 0x54AA, [ 6656] = 0x54C1, [ 6657] = 0x54C4, [ 6658] = 0x54C8,
  [ 6659] = 0x54AF, [ 6660] = 0x54AB, [ 6661] = 0x54B1, [ 6662] = 0x54BB,
  [ 6663] = 0x54A9, [ 6664] = 0x54A7, [ 6665] = 0x54BF, [ 6666] = 0x56FF,
  [ 6667] = 0x5782, [ 6668] = 0x578B, [ 6669] = 0x57A0, [ 6670] = 0x57A3,
  [ 6671] = 0x57A2, [ 6672] = 0x57CE, [ 6673] = 0x57AE, [ 6674] = 0x5793,
  [ 6675] = 0x5955, [ 6676] = 0x5951, [ 6677] = 0x594F, [ 6678] = 0x594E,
  [ 6679] = 0x5950, [ 6680] = 0x59DC, [ 6681] = 0x59D8, [ 6682] = 0x59FF,
  [ 6683] = 0x59E3, [ 6684] = 0x59E8, [ 6685] = 0x5A03, [ 6686] = 0x59E5,
  [ 6687] = 0x59EA, [ 6688] = 0x59DA, [ 6689] = 0x59E6, [ 6690] = 0x5A01,
  [ 6691] = 0x59FB, [ 6692] = 0x5B69, [ 6693] = 0x5BA3, [ 6694] = 0x5BA6,
  [ 6695] = 0x5BA4, [ 6696] = 0x5BA2, [ 6697] = 0x5BA5, [ 6698] = 0x5C01,
  [ 6699] = 0x5C4E, [ 6700] = 0x5C4F, [ 6701] = 0x5C4D, [ 6702] = 0x5C4B,
  [ 6703] = 0x5CD9, [ 6704] = 0x5CD2, [ 6705] = 0x5DF7, [ 6706] = 0x5E1D,
  [ 6707] = 0x5E25, [ 6708] = 0x5E1F, [ 6709] = 0x5E7D, [ 6710] = 0x5EA0,
  [ 6711] = 0x5EA6, [ 6712] = 0x5EFA, [ 6713] = 0x5F08, [ 6714] = 0x5F2D,
  [ 6715] = 0x5F65, [ 6716] = 0x5F88, [ 6717] = 0x5F85, [ 6718] = 0x5F8A,
  [ 6719] = 0x5F8B, [ 6720] = 0x5F87, [ 6721] = 0x5F8C, [ 6722] = 0x5F89,
  [ 6723] = 0x6012, [ 6724] = 0x601D, [ 6725] = 0x6020, [ 6726] = 0x6025,
  [ 6727] = 0x600E, [ 6728] = 0x6028, [ 6729] = 0x604D, [ 6730] = 0x6070,
  [ 6731] = 0x6068, [ 6732] = 0x6062, [ 6733] = 0x6046, [ 6734] = 0x6043,
  [ 6735] = 0x606C, [ 6736] = 0x606B, [ 6737] = 0x606A, [ 6738] = 0x6064,
  [ 6739] = 0x6241, [ 6740] = 0x62DC, [ 6741] = 0x6316, [ 6742] = 0x6309,
  [ 6743] = 0x62FC, [ 6744] = 0x62ED, [ 6745] = 0x6301, [ 6746] = 0x62EE,
  [ 6747] = 0x62FD, [ 6748] = 0x6307, [ 6749] = 0x62F1, [ 6750] = 0x62F7,
  [ 6751] = 0x62EF, [ 6752] = 0x62EC, [ 6753] = 0x62FE, [ 6754] = 0x62F4,
  [ 6755] = 0x6311, [ 6756] = 0x6302, [ 6757] = 0x653F, [ 6758] = 0x6545,
  [ 6759] = 0x65AB, [ 6760] = 0x65BD, [ 6761] = 0x65E2, [ 6762] = 0x6625,
  [ 6763] = 0x662D, [ 6764] = 0x6620, [ 6765] = 0x6627, [ 6766] = 0x662F,
  [ 6767] = 0x661F, [ 6768] = 0x6628, [ 6769] = 0x6631, [ 6770] = 0x6624,
  [ 6771] = 0x66F7, [ 6772] = 0x67FF, [ 6773] = 0x67D3, [ 6774] = 0x67F1,
  [ 6775] = 0x67D4, [ 6776] = 0x67D0, [ 6777] = 0x67EC, [ 6778] = 0x67B6,
  [ 6779] = 0x67AF, [ 6780] = 0x67F5, [ 6781] = 0x67E9, [ 6782] = 0x67EF,
  [ 6783] = 0x67C4, [ 6784] = 0x67D1, [ 6785] = 0x67B4, [ 6786] = 0x67DA,
  [ 6787] = 0x67E5, [ 6788] = 0x67B8, [ 6789] = 0x67CF, [ 6790] = 0x67DE,
  [ 6791] = 0x67F3, [ 6792] = 0x67B0, [ 6793] = 0x67D9, [ 6794] = 0x67E2,
  [ 6795] = 0x67DD, [ 6796] = 0x67D2, [ 6797] = 0x6B6A, [ 6798] = 0x6B83,
  [ 6799] = 0x6B86, [ 6800] = 0x6BB5, [ 6801] = 0x6BD2, [ 6802] = 0x6BD7,
  [ 6803] = 0x6C1F, [ 6804] = 0x6CC9, [ 6805] = 0x6D0B, [ 6806] = 0x6D32,
  [ 6807] = 0x6D2A, [ 6808] = 0x6D41, [ 6809] = 0x6D25, [ 6810] = 0x6D0C,
  [ 6811] = 0x6D31, [ 6812] = 0x6D1E, [ 6813] = 0x6D17, [ 6814] = 0x6D3B,
  [ 6815] = 0x6D3D, [ 6816] = 0x6D3E, [ 6817] = 0x6D36, [ 6818] = 0x6D1B,
  [ 6819] = 0x6CF5, [ 6820] = 0x6D39, [ 6821] = 0x6D27, [ 6822] = 0x6D38,
  [ 6823] = 0x6D29, [ 6824] = 0x6D2E, [ 6825] = 0x6D35, [ 6826] = 0x6D0E,
  [ 6827] = 0x6D2B, [ 6828] = 0x70AB, [ 6829] = 0x70BA, [ 6830] = 0x70B3,
  [ 6831] = 0x70AC, [ 6832] = 0x70AF, [ 6833] = 0x70AD, [ 6834] = 0x70B8,
  [ 6835] = 0x70AE, [ 6836] = 0x70A4, [ 6837] = 0x7230, [ 6838] = 0x7272,
  [ 6839] = 0x726F, [ 6840] = 0x7274, [ 6841] = 0x72E9, [ 6842] = 0x72E0,
  [ 6843] = 0x72E1, [ 6844] = 0x73B7, [ 6845] = 0x73CA, [ 6846] = 0x73BB,
  [ 6847] = 0x73B2, [ 6848] = 0x73CD, [ 6849] = 0x73C0, [ 6850] = 0x73B3,
  [ 6851] = 0x751A, [ 6852] = 0x752D, [ 6853] = 0x754F, [ 6854] = 0x754C,
  [ 6855] = 0x754E, [ 6856] = 0x754B, [ 6857] = 0x75AB, [ 6858] = 0x75A4,
  [ 6859] = 0x75A5, [ 6860] = 0x75A2, [ 6861] = 0x75A3, [ 6862] = 0x7678,
  [ 6863] = 0x7686, [ 6864] = 0x7687, [ 6865] = 0x7688, [ 6866] = 0x76C8,
  [ 6867] = 0x76C6, [ 6868] = 0x76C3, [ 6869] = 0x76C5, [ 6870] = 0x7701,
  [ 6871] = 0x76F9, [ 6872] = 0x76F8, [ 6873] = 0x7709, [ 6874] = 0x770B,
  [ 6875] = 0x76FE, [ 6876] = 0x76FC, [ 6877] = 0x7707, [ 6878] = 0x77DC,
  [ 6879] = 0x7802, [ 6880] = 0x7814, [ 6881] = 0x780C, [ 6882] = 0x780D,
  [ 6883] = 0x7946, [ 6884] = 0x7949, [ 6885] = 0x7948, [ 6886] = 0x7947,
  [ 6887] = 0x79B9, [ 6888] = 0x79BA, [ 6889] = 0x79D1, [ 6890] = 0x79D2,
  [ 6891] = 0x79CB, [ 6892] = 0x7A7F, [ 6893] = 0x7A81, [ 6894] = 0x7AFF,
  [ 6895] = 0x7AFD, [ 6896] = 0x7C7D, [ 6897] = 0x7D02, [ 6898] = 0x7D05,
  [ 6899] = 0x7D00, [ 6900] = 0x7D09, [ 6901] = 0x7D07, [ 6902] = 0x7D04,
  [ 6903] = 0x7D06, [ 6904] = 0x7F38, [ 6905] = 0x7F8E, [ 6906] = 0x7FBF,
  [ 6907] = 0x8004, [ 6908] = 0x8010, [ 6909] = 0x800D, [ 6910] = 0x8011,
  [ 6911] = 0x8036, [ 6912] = 0x80D6, [ 6913] = 0x80E5, [ 6914] = 0x80DA,
  [ 6915] = 0x80C3, [ 6916] = 0x80C4, [ 6917] = 0x80CC, [ 6918] = 0x80E1,
  [ 6919] = 0x80DB, [ 6920] = 0x80CE, [ 6921] = 0x80DE, [ 6922] = 0x80E4,
  [ 6923] = 0x80DD, [ 6924] = 0x81F4, [ 6925] = 0x8222, [ 6926] = 0x82E7,
  [ 6927] = 0x8303, [ 6928] = 0x8305, [ 6929] = 0x82E3, [ 6930] = 0x82DB,
  [ 6931] = 0x82E6, [ 6932] = 0x8304, [ 6933] = 0x82E5, [ 6934] = 0x8302,
  [ 6935] = 0x8309, [ 6936] = 0x82D2, [ 6937] = 0x82D7, [ 6938] = 0x82F1,
  [ 6939] = 0x8301, [ 6940] = 0x82DC, [ 6941] = 0x82D4, [ 6942] = 0x82D1,
  [ 6943] = 0x82DE, [ 6944] = 0x82D3, [ 6945] = 0x82DF, [ 6946] = 0x82EF,
  [ 6947] = 0x8306, [ 6948] = 0x8650, [ 6949] = 0x8679, [ 6950] = 0x867B,
  [ 6951] = 0x867A, [ 6952] = 0x884D, [ 6953] = 0x886B, [ 6954] = 0x8981,
  [ 6955] = 0x89D4, [ 6956] = 0x8A08, [ 6957] = 0x8A02, [ 6958] = 0x8A03,
  [ 6959] = 0x8C9E, [ 6960] = 0x8CA0, [ 6961] = 0x8D74, [ 6962] = 0x8D73,
  [ 6963] = 0x8DB4, [ 6964] = 0x8ECD, [ 6965] = 0x8ECC, [ 6966] = 0x8FF0,
  [ 6967] = 0x8FE6, [ 6968] = 0x8FE2, [ 6969] = 0x8FEA, [ 6970] = 0x8FE5,
  [ 6971] = 0x8FED, [ 6972] = 0x8FEB, [ 6973] = 0x8FE4, [ 6974] = 0x8FE8,
  [ 6975] = 0x90CA, [ 6976] = 0x90CE, [ 6977] = 0x90C1, [ 6978] = 0x90C3,
  [ 6979] = 0x914B, [ 6980] = 0x914A, [ 6981] = 0x91CD, [ 6982] = 0x9582,
  [ 6983] = 0x9650, [ 6984] = 0x964B, [ 6985] = 0x964C, [ 6986] = 0x964D,
  [ 6987] = 0x9762, [ 6988] = 0x9769, [ 6989] = 0x97CB, [ 6990] = 0x97ED,
  [ 6991] = 0x97F3, [ 6992] = 0x9801, [ 6993] = 0x98A8, [ 6994] = 0x98DB,
  [ 6995] = 0x98DF, [ 6996] = 0x9996, [ 6997] = 0x9999, [ 6998] = 0x4E58,
  [ 6999] = 0x4EB3, [ 7000] = 0x500C, [ 7001] = 0x500D, [ 7002] = 0x5023,
  [ 7003] = 0x4FEF, [ 7004] = 0x5026, [ 7005] = 0x5025, [ 7006] = 0x4FF8,
  [ 7007] = 0x5029, [ 7008] = 0x5016, [ 7009] = 0x5006, [ 7010] = 0x503C,
  [ 7011] = 0x501F, [ 7012] = 0x501A, [ 7013] = 0x5012, [ 7014] = 0x5011,
  [ 7015] = 0x4FFA, [ 7016] = 0x5000, [ 7017] = 0x5014, [ 7018] = 0x5028,
  [ 7019] = 0x4FF1, [ 7020] = 0x5021, [ 7021] = 0x500B, [ 7022] = 0x5019,
  [ 7023] = 0x5018, [ 7024] = 0x4FF3, [ 7025] = 0x4FEE, [ 7026] = 0x502D,
  [ 7027] = 0x502A, [ 7028] = 0x4FFE, [ 7029] = 0x502B, [ 7030] = 0x5009,
  [ 7031] = 0x517C, [ 7032] = 0x51A4, [ 7033] = 0x51A5, [ 7034] = 0x51A2,
  [ 7035] = 0x51CD, [ 7036] = 0x51CC, [ 7037] = 0x51C6, [ 7038] = 0x51CB,
  [ 7039] = 0x5256, [ 7040] = 0x525C, [ 7041] = 0x5254, [ 7042] = 0x525B,
  [ 7043] = 0x525D, [ 7044] = 0x532A, [ 7045] = 0x537F, [ 7046] = 0x539F,
  [ 7047] = 0x539D, [ 7048] = 0x53DF, [ 7049] = 0x54E8, [ 7050] = 0x5510,
  [ 7051] = 0x5501, [ 7052] = 0x5537, [ 7053] = 0x54FC, [ 7054] = 0x54E5,
  [ 7055] = 0x54F2, [ 7056] = 0x5506, [ 7057] = 0x54FA, [ 7058] = 0x5514,
  [ 7059] = 0x54E9, [ 7060] = 0x54ED, [ 7061] = 0x54E1, [ 7062] = 0x5509,
  [ 7063] = 0x54EE, [ 7064] = 0x54EA, [ 7065] = 0x54E6, [ 7066] = 0x5527,
  [ 7067] = 0x5507, [ 7068] = 0x54FD, [ 7069] = 0x550F, [ 7070] = 0x5703,
  [ 7071] = 0x5704, [ 7072] = 0x57C2, [ 7073] = 0x57D4, [ 7074] = 0x57CB,
  [ 7075] = 0x57C3, [ 7076] = 0x5809, [ 7077] = 0x590F, [ 7078] = 0x5957,
  [ 7079] = 0x5958, [ 7080] = 0x595A, [ 7081] = 0x5A11, [ 7082] = 0x5A18,
  [ 7083] = 0x5A1C, [ 7084] = 0x5A1F, [ 7085] = 0x5A1B, [ 7086] = 0x5A13,
  [ 7087] = 0x59EC, [ 7088] = 0x5A20, [ 7089] = 0x5A23, [ 7090] = 0x5A29,
  [ 7091] = 0x5A25, [ 7092] = 0x5A0C, [ 7093] = 0x5A09, [ 7094] = 0x5B6B,
  [ 7095] = 0x5C58, [ 7096] = 0x5BB0, [ 7097] = 0x5BB3, [ 7098] = 0x5BB6,
  [ 7099] = 0x5BB4, [ 7100] = 0x5BAE, [ 7101] = 0x5BB5, [ 7102] = 0x5BB9,
  [ 7103] = 0x5BB8, [ 7104] = 0x5C04, [ 7105] = 0x5C51, [ 7106] = 0x5C55,
  [ 7107] = 0x5C50, [ 7108] = 0x5CED, [ 7109] = 0x5CFD, [ 7110] = 0x5CFB,
  [ 7111] = 0x5CEA, [ 7112] = 0x5CE8, [ 7113] = 0x5CF0, [ 7114] = 0x5CF6,
  [ 7115] = 0x5D01, [ 7116] = 0x5CF4, [ 7117] = 0x5DEE, [ 7118] = 0x5E2D,
  [ 7119] = 0x5E2B, [ 7120] = 0x5EAB, [ 7121] = 0x5EAD, [ 7122] = 0x5EA7,
  [ 7123] = 0x5F31, [ 7124] = 0x5F92, [ 7125] = 0x5F91, [ 7126] = 0x5F90,
  [ 7127] = 0x6059, [ 7128] = 0x6063, [ 7129] = 0x6065, [ 7130] = 0x6050,
  [ 7131] = 0x6055, [ 7132] = 0x606D, [ 7133] = 0x6069, [ 7134] = 0x606F,
  [ 7135] = 0x6084, [ 7136] = 0x609F, [ 7137] = 0x609A, [ 7138] = 0x608D,
  [ 7139] = 0x6094, [ 7140] = 0x608C, [ 7141] = 0x6085, [ 7142] = 0x6096,
  [ 7143] = 0x6247, [ 7144] = 0x62F3, [ 7145] = 0x6308, [ 7146] = 0x62FF,
  [ 7147] = 0x634E, [ 7148] = 0x633E, [ 7149] = 0x632F, [ 7150] = 0x6355,
  [ 7151] = 0x6342, [ 7152] = 0x6346, [ 7153] = 0x634F, [ 7154] = 0x6349,
  [ 7155] = 0x633A, [ 7156] = 0x6350, [ 7157] = 0x633D, [ 7158] = 0x632A,
  [ 7159] = 0x632B, [ 7160] = 0x6328, [ 7161] = 0x634D, [ 7162] = 0x634C,
  [ 7163] = 0x6548, [ 7164] = 0x6549, [ 7165] = 0x6599, [ 7166] = 0x65C1,
  [ 7167] = 0x65C5, [ 7168] = 0x6642, [ 7169] = 0x6649, [ 7170] = 0x664F,
  [ 7171] = 0x6643, [ 7172] = 0x6652, [ 7173] = 0x664C, [ 7174] = 0x6645,
  [ 7175] = 0x6641, [ 7176] = 0x66F8, [ 7177] = 0x6714, [ 7178] = 0x6715,
  [ 7179] = 0x6717, [ 7180] = 0x6821, [ 7181] = 0x6838, [ 7182] = 0x6848,
  [ 7183] = 0x6846, [ 7184] = 0x6853, [ 7185] = 0x6839, [ 7186] = 0x6842,
  [ 7187] = 0x6854, [ 7188] = 0x6829, [ 7189] = 0x68B3, [ 7190] = 0x6817,
  [ 7191] = 0x684C, [ 7192] = 0x6851, [ 7193] = 0x683D, [ 7194] = 0x67F4,
  [ 7195] = 0x6850, [ 7196] = 0x6840, [ 7197] = 0x683C, [ 7198] = 0x6843,
  [ 7199] = 0x682A, [ 7200] = 0x6845, [ 7201] = 0x6813, [ 7202] = 0x6818,
  [ 7203] = 0x6841, [ 7204] = 0x6B8A, [ 7205] = 0x6B89, [ 7206] = 0x6BB7,
  [ 7207] = 0x6C23, [ 7208] = 0x6C27, [ 7209] = 0x6C28, [ 7210] = 0x6C26,
  [ 7211] = 0x6C24, [ 7212] = 0x6CF0, [ 7213] = 0x6D6A, [ 7214] = 0x6D95,
  [ 7215] = 0x6D88, [ 7216] = 0x6D87, [ 7217] = 0x6D66, [ 7218] = 0x6D78,
  [ 7219] = 0x6D77, [ 7220] = 0x6D59, [ 7221] = 0x6D93, [ 7222] = 0x6D6C,
  [ 7223] = 0x6D89, [ 7224] = 0x6D6E, [ 7225] = 0x6D5A, [ 7226] = 0x6D74,
  [ 7227] = 0x6D69, [ 7228] = 0x6D8C, [ 7229] = 0x6D8A, [ 7230] = 0x6D79,
  [ 7231] = 0x6D85, [ 7232] = 0x6D65, [ 7233] = 0x6D94, [ 7234] = 0x70CA,
  [ 7235] = 0x70D8, [ 7236] = 0x70E4, [ 7237] = 0x70D9, [ 7238] = 0x70C8,
  [ 7239] = 0x70CF, [ 7240] = 0x7239, [ 7241] = 0x7279, [ 7242] = 0x72FC,
  [ 7243] = 0x72F9, [ 7244] = 0x72FD, [ 7245] = 0x72F8, [ 7246] = 0x72F7,
  [ 7247] = 0x7386, [ 7248] = 0x73ED, [ 7249] = 0x7409, [ 7250] = 0x73EE,
  [ 7251] = 0x73E0, [ 7252] = 0x73EA, [ 7253] = 0x73DE, [ 7254] = 0x7554,
  [ 7255] = 0x755D, [ 7256] = 0x755C, [ 7257] = 0x755A, [ 7258] = 0x7559,
  [ 7259] = 0x75BE, [ 7260] = 0x75C5, [ 7261] = 0x75C7, [ 7262] = 0x75B2,
  [ 7263] = 0x75B3, [ 7264] = 0x75BD, [ 7265] = 0x75BC, [ 7266] = 0x75B9,
  [ 7267] = 0x75C2, [ 7268] = 0x75B8, [ 7269] = 0x768B, [ 7270] = 0x76B0,
  [ 7271] = 0x76CA, [ 7272] = 0x76CD, [ 7273] = 0x76CE, [ 7274] = 0x7729,
  [ 7275] = 0x771F, [ 7276] = 0x7720, [ 7277] = 0x7728, [ 7278] = 0x77E9,
  [ 7279] = 0x7830, [ 7280] = 0x7827, [ 7281] = 0x7838, [ 7282] = 0x781D,
  [ 7283] = 0x7834, [ 7284] = 0x7837, [ 7285] = 0x7825, [ 7286] = 0x782D,
  [ 7287] = 0x7820, [ 7288] = 0x781F, [ 7289] = 0x7832, [ 7290] = 0x7955,
  [ 7291] = 0x7950, [ 7292] = 0x7960, [ 7293] = 0x795F, [ 7294] = 0x7956,
  [ 7295] = 0x795E, [ 7296] = 0x795D, [ 7297] = 0x7957, [ 7298] = 0x795A,
  [ 7299] = 0x79E4, [ 7300] = 0x79E3, [ 7301] = 0x79E7, [ 7302] = 0x79DF,
  [ 7303] = 0x79E6, [ 7304] = 0x79E9, [ 7305] = 0x79D8, [ 7306] = 0x7A84,
  [ 7307] = 0x7A88, [ 7308] = 0x7AD9, [ 7309] = 0x7B06, [ 7310] = 0x7B11,
  [ 7311] = 0x7C89, [ 7312] = 0x7D21, [ 7313] = 0x7D17, [ 7314] = 0x7D0B,
  [ 7315] = 0x7D0A, [ 7316] = 0x7D20, [ 7317] = 0x7D22, [ 7318] = 0x7D14,
  [ 7319] = 0x7D10, [ 7320] = 0x7D15, [ 7321] = 0x7D1A, [ 7322] = 0x7D1C,
  [ 7323] = 0x7D0D, [ 7324] = 0x7D19, [ 7325] = 0x7D1B, [ 7326] = 0x7F3A,
  [ 7327] = 0x7F5F, [ 7328] = 0x7F94, [ 7329] = 0x7FC5, [ 7330] = 0x7FC1,
  [ 7331] = 0x8006, [ 7332] = 0x8018, [ 7333] = 0x8015, [ 7334] = 0x8019,
  [ 7335] = 0x8017, [ 7336] = 0x803D, [ 7337] = 0x803F, [ 7338] = 0x80F1,
  [ 7339] = 0x8102, [ 7340] = 0x80F0, [ 7341] = 0x8105, [ 7342] = 0x80ED,
  [ 7343] = 0x80F4, [ 7344] = 0x8106, [ 7345] = 0x80F8, [ 7346] = 0x80F3,
  [ 7347] = 0x8108, [ 7348] = 0x80FD, [ 7349] = 0x810A, [ 7350] = 0x80FC,
  [ 7351] = 0x80EF, [ 7352] = 0x81ED, [ 7353] = 0x81EC, [ 7354] = 0x8200,
  [ 7355] = 0x8210, [ 7356] = 0x822A, [ 7357] = 0x822B, [ 7358] = 0x8228,
  [ 7359] = 0x822C, [ 7360] = 0x82BB, [ 7361] = 0x832B, [ 7362] = 0x8352,
  [ 7363] = 0x8354, [ 7364] = 0x834A, [ 7365] = 0x8338, [ 7366] = 0x8350,
  [ 7367] = 0x8349, [ 7368] = 0x8335, [ 7369] = 0x8334, [ 7370] = 0x834F,
  [ 7371] = 0x8332, [ 7372] = 0x8339, [ 7373] = 0x8336, [ 7374] = 0x8317,
  [ 7375] = 0x8340, [ 7376] = 0x8331, [ 7377] = 0x8328, [ 7378] = 0x8343,
  [ 7379] = 0x8654, [ 7380] = 0x868A, [ 7381] = 0x86AA, [ 7382] = 0x8693,
  [ 7383] = 0x86A4, [ 7384] = 0x86A9, [ 7385] = 0x868C, [ 7386] = 0x86A3,
  [ 7387] = 0x869C, [ 7388] = 0x8870, [ 7389] = 0x8877, [ 7390] = 0x8881,
  [ 7391] = 0x8882, [ 7392] = 0x887D, [ 7393] = 0x8879, [ 7394] = 0x8A18,
  [ 7395] = 0x8A10, [ 7396] = 0x8A0E, [ 7397] = 0x8A0C, [ 7398] = 0x8A15,
  [ 7399] = 0x8A0A, [ 7400] = 0x8A17, [ 7401] = 0x8A13, [ 7402] = 0x8A16,
  [ 7403] = 0x8A0F, [ 7404] = 0x8A11, [ 7405] = 0x8C48, [ 7406] = 0x8C7A,
  [ 7407] = 0x8C79, [ 7408] = 0x8CA1, [ 7409] = 0x8CA2, [ 7410] = 0x8D77,
  [ 7411] = 0x8EAC, [ 7412] = 0x8ED2, [ 7413] = 0x8ED4, [ 7414] = 0x8ECF,
  [ 7415] = 0x8FB1, [ 7416] = 0x9001, [ 7417] = 0x9006, [ 7418] = 0x8FF7,
  [ 7419] = 0x9000, [ 7420] = 0x8FFA, [ 7421] = 0x8FF4, [ 7422] = 0x9003,
  [ 7423] = 0x8FFD, [ 7424] = 0x9005, [ 7425] = 0x8FF8, [ 7426] = 0x9095,
  [ 7427] = 0x90E1, [ 7428] = 0x90DD, [ 7429] = 0x90E2, [ 7430] = 0x9152,
  [ 7431] = 0x914D, [ 7432] = 0x914C, [ 7433] = 0x91D8, [ 7434] = 0x91DD,
  [ 7435] = 0x91D7, [ 7436] = 0x91DC, [ 7437] = 0x91D9, [ 7438] = 0x9583,
  [ 7439] = 0x9662, [ 7440] = 0x9663, [ 7441] = 0x9661, [ 7442] = 0x965B,
  [ 7443] = 0x965D, [ 7444] = 0x9664, [ 7445] = 0x9658, [ 7446] = 0x965E,
  [ 7447] = 0x96BB, [ 7448] = 0x98E2, [ 7449] = 0x99AC, [ 7450] = 0x9AA8,
  [ 7451] = 0x9AD8, [ 7452] = 0x9B25, [ 7453] = 0x9B32, [ 7454] = 0x9B3C,
  [ 7455] = 0x4E7E, [ 7456] = 0x507A, [ 7457] = 0x507D, [ 7458] = 0x505C,
  [ 7459] = 0x5047, [ 7460] = 0x5043, [ 7461] = 0x504C, [ 7462] = 0x505A,
  [ 7463] = 0x5049, [ 7464] = 0x5065, [ 7465] = 0x5076, [ 7466] = 0x504E,
  [ 7467] = 0x5055, [ 7468] = 0x5075, [ 7469] = 0x5074, [ 7470] = 0x5077,
  [ 7471] = 0x504F, [ 7472] = 0x500F, [ 7473] = 0x506F, [ 7474] = 0x506D,
  [ 7475] = 0x515C, [ 7476] = 0x5195, [ 7477] = 0x51F0, [ 7478] = 0x526A,
  [ 7479] = 0x526F, [ 7480] = 0x52D2, [ 7481] = 0x52D9, [ 7482] = 0x52D8,
  [ 7483] = 0x52D5, [ 7484] = 0x5310, [ 7485] = 0x530F, [ 7486] = 0x5319,
  [ 7487] = 0x533F, [ 7488] = 0x5340, [ 7489] = 0x533E, [ 7490] = 0x53C3,
  [ 7491] = 0x66FC, [ 7492] = 0x5546, [ 7493] = 0x556A, [ 7494] = 0x5566,
  [ 7495] = 0x5544, [ 7496] = 0x555E, [ 7497] = 0x5561, [ 7498] = 0x5543,
  [ 7499] = 0x554A, [ 7500] = 0x5531, [ 7501] = 0x5556, [ 7502] = 0x554F,
  [ 7503] = 0x5555, [ 7504] = 0x552F, [ 7505] = 0x5564, [ 7506] = 0x5538,
  [ 7507] = 0x552E, [ 7508] = 0x555C, [ 7509] = 0x552C, [ 7510] = 0x5563,
  [ 7511] = 0x5533, [ 7512] = 0x5541, [ 7513] = 0x5557, [ 7514] = 0x5708,
  [ 7515] = 0x570B, [ 7516] = 0x5709, [ 7517] = 0x57DF, [ 7518] = 0x5805,
  [ 7519] = 0x580A, [ 7520] = 0x5806, [ 7521] = 0x57E0, [ 7522] = 0x57E4,
  [ 7523] = 0x57FA, [ 7524] = 0x5802, [ 7525] = 0x5835, [ 7526] = 0x57F7,
  [ 7527] = 0x57F9, [ 7528] = 0x5920, [ 7529] = 0x5962, [ 7530] = 0x5A36,
  [ 7531] = 0x5A41, [ 7532] = 0x5A49, [ 7533] = 0x5A66, [ 7534] = 0x5A6A,
  [ 7535] = 0x5A40, [ 7536] = 0x5A3C, [ 7537] = 0x5A62, [ 7538] = 0x5A5A,
  [ 7539] = 0x5A46, [ 7540] = 0x5A4A, [ 7541] = 0x5B70, [ 7542] = 0x5BC7,
  [ 7543] = 0x5BC5, [ 7544] = 0x5BC4, [ 7545] = 0x5BC2, [ 7546] = 0x5BBF,
  [ 7547] = 0x5BC6, [ 7548] = 0x5C09, [ 7549] = 0x5C08, [ 7550] = 0x5C07,
  [ 7551] = 0x5C60, [ 7552] = 0x5C5C, [ 7553] = 0x5C5D, [ 7554] = 0x5D07,
  [ 7555] = 0x5D06, [ 7556] = 0x5D0E, [ 7557] = 0x5D1B, [ 7558] = 0x5D16,
  [ 7559] = 0x5D22, [ 7560] = 0x5D11, [ 7561] = 0x5D29, [ 7562] = 0x5D14,
  [ 7563] = 0x5D19, [ 7564] = 0x5D24, [ 7565] = 0x5D27, [ 7566] = 0x5D17,
  [ 7567] = 0x5DE2, [ 7568] = 0x5E38, [ 7569] = 0x5E36, [ 7570] = 0x5E33,
  [ 7571] = 0x5E37, [ 7572] = 0x5EB7, [ 7573] = 0x5EB8, [ 7574] = 0x5EB6,
  [ 7575] = 0x5EB5, [ 7576] = 0x5EBE, [ 7577] = 0x5F35, [ 7578] = 0x5F37,
  [ 7579] = 0x5F57, [ 7580] = 0x5F6C, [ 7581] = 0x5F69, [ 7582] = 0x5F6B,
  [ 7583] = 0x5F97, [ 7584] = 0x5F99, [ 7585] = 0x5F9E, [ 7586] = 0x5F98,
  [ 7587] = 0x5FA1, [ 7588] = 0x5FA0, [ 7589] = 0x5F9C, [ 7590] = 0x607F,
  [ 7591] = 0x60A3, [ 7592] = 0x6089, [ 7593] = 0x60A0, [ 7594] = 0x60A8,
  [ 7595] = 0x60CB, [ 7596] = 0x60B4, [ 7597] = 0x60E6, [ 7598] = 0x60BD,
  [ 7599] = 0x60C5, [ 7600] = 0x60BB, [ 7601] = 0x60B5, [ 7602] = 0x60DC,
  [ 7603] = 0x60BC, [ 7604] = 0x60D8, [ 7605] = 0x60D5, [ 7606] = 0x60C6,
  [ 7607] = 0x60DF, [ 7608] = 0x60B8, [ 7609] = 0x60DA, [ 7610] = 0x60C7,
  [ 7611] = 0x621A, [ 7612] = 0x621B, [ 7613] = 0x6248, [ 7614] = 0x63A0,
  [ 7615] = 0x63A7, [ 7616] = 0x6372, [ 7617] = 0x6396, [ 7618] = 0x63A2,
  [ 7619] = 0x63A5, [ 7620] = 0x6377, [ 7621] = 0x6367, [ 7622] = 0x6398,
  [ 7623] = 0x63AA, [ 7624] = 0x6371, [ 7625] = 0x63A9, [ 7626] = 0x6389,
  [ 7627] = 0x6383, [ 7628] = 0x639B, [ 7629] = 0x636B, [ 7630] = 0x63A8,
  [ 7631] = 0x6384, [ 7632] = 0x6388, [ 7633] = 0x6399, [ 7634] = 0x63A1,
  [ 7635] = 0x63AC, [ 7636] = 0x6392, [ 7637] = 0x638F, [ 7638] = 0x6380,
  [ 7639] = 0x637B, [ 7640] = 0x6369, [ 7641] = 0x6368, [ 7642] = 0x637A,
  [ 7643] = 0x655D, [ 7644] = 0x6556, [ 7645] = 0x6551, [ 7646] = 0x6559,
  [ 7647] = 0x6557, [ 7648] = 0x555F, [ 7649] = 0x654F, [ 7650] = 0x6558,
  [ 7651] = 0x6555, [ 7652] = 0x6554, [ 7653] = 0x659C, [ 7654] = 0x659B,
  [ 7655] = 0x65AC, [ 7656] = 0x65CF, [ 7657] = 0x65CB, [ 7658] = 0x65CC,
  [ 7659] = 0x65CE, [ 7660] = 0x665D, [ 7661] = 0x665A, [ 7662] = 0x6664,
  [ 7663] = 0x6668, [ 7664] = 0x6666, [ 7665] = 0x665E, [ 7666] = 0x66F9,
  [ 7667] = 0x52D7, [ 7668] = 0x671B, [ 7669] = 0x6881, [ 7670] = 0x68AF,
  [ 7671] = 0x68A2, [ 7672] = 0x6893, [ 7673] = 0x68B5, [ 7674] = 0x687F,
  [ 7675] = 0x6876, [ 7676] = 0x68B1, [ 7677] = 0x68A7, [ 7678] = 0x6897,
  [ 7679] = 0x68B0, [ 7680] = 0x6883, [ 7681] = 0x68C4, [ 7682] = 0x68AD,
  [ 7683] = 0x6886, [ 7684] = 0x6885, [ 7685] = 0x6894, [ 7686] = 0x689D,
  [ 7687] = 0x68A8, [ 7688] = 0x689F, [ 7689] = 0x68A1, [ 7690] = 0x6882,
  [ 7691] = 0x6B32, [ 7692] = 0x6BBA, [ 7693] = 0x6BEB, [ 7694] = 0x6BEC,
  [ 7695] = 0x6C2B, [ 7696] = 0x6D8E, [ 7697] = 0x6DBC, [ 7698] = 0x6DF3,
  [ 7699] = 0x6DD9, [ 7700] = 0x6DB2, [ 7701] = 0x6DE1, [ 7702] = 0x6DCC,
  [ 7703] = 0x6DE4, [ 7704] = 0x6DFB, [ 7705] = 0x6DFA, [ 7706] = 0x6E05,
  [ 7707] = 0x6DC7, [ 7708] = 0x6DCB, [ 7709] = 0x6DAF, [ 7710] = 0x6DD1,
  [ 7711] = 0x6DAE, [ 7712] = 0x6DDE, [ 7713] = 0x6DF9, [ 7714] = 0x6DB8,
  [ 7715] = 0x6DF7, [ 7716] = 0x6DF5, [ 7717] = 0x6DC5, [ 7718] = 0x6DD2,
  [ 7719] = 0x6E1A, [ 7720] = 0x6DB5, [ 7721] = 0x6DDA, [ 7722] = 0x6DEB,
  [ 7723] = 0x6DD8, [ 7724] = 0x6DEA, [ 7725] = 0x6DF1, [ 7726] = 0x6DEE,
  [ 7727] = 0x6DE8, [ 7728] = 0x6DC6, [ 7729] = 0x6DC4, [ 7730] = 0x6DAA,
  [ 7731] = 0x6DEC, [ 7732] = 0x6DBF, [ 7733] = 0x6DE6, [ 7734] = 0x70F9,
  [ 7735] = 0x7109, [ 7736] = 0x710A, [ 7737] = 0x70FD, [ 7738] = 0x70EF,
  [ 7739] = 0x723D, [ 7740] = 0x727D, [ 7741] = 0x7281, [ 7742] = 0x731C,
  [ 7743] = 0x731B, [ 7744] = 0x7316, [ 7745] = 0x7313, [ 7746] = 0x7319,
  [ 7747] = 0x7387, [ 7748] = 0x7405, [ 7749] = 0x740A, [ 7750] = 0x7403,
  [ 7751] = 0x7406, [ 7752] = 0x73FE, [ 7753] = 0x740D, [ 7754] = 0x74E0,
  [ 7755] = 0x74F6, [ 7756] = 0x74F7, [ 7757] = 0x751C, [ 7758] = 0x7522,
  [ 7759] = 0x7565, [ 7760] = 0x7566, [ 7761] = 0x7562, [ 7762] = 0x7570,
  [ 7763] = 0x758F, [ 7764] = 0x75D4, [ 7765] = 0x75D5, [ 7766] = 0x75B5,
  [ 7767] = 0x75CA, [ 7768] = 0x75CD, [ 7769] = 0x768E, [ 7770] = 0x76D4,
  [ 7771] = 0x76D2, [ 7772] = 0x76DB, [ 7773] = 0x7737, [ 7774] = 0x773E,
  [ 7775] = 0x773C, [ 7776] = 0x7736, [ 7777] = 0x7738, [ 7778] = 0x773A,
  [ 7779] = 0x786B, [ 7780] = 0x7843, [ 7781] = 0x784E, [ 7782] = 0x7965,
  [ 7783] = 0x7968, [ 7784] = 0x796D, [ 7785] = 0x79FB, [ 7786] = 0x7A92,
  [ 7787] = 0x7A95, [ 7788] = 0x7B20, [ 7789] = 0x7B28, [ 7790] = 0x7B1B,
  [ 7791] = 0x7B2C, [ 7792] = 0x7B26, [ 7793] = 0x7B19, [ 7794] = 0x7B1E,
  [ 7795] = 0x7B2E, [ 7796] = 0x7C92, [ 7797] = 0x7C97, [ 7798] = 0x7C95,
  [ 7799] = 0x7D46, [ 7800] = 0x7D43, [ 7801] = 0x7D71, [ 7802] = 0x7D2E,
  [ 7803] = 0x7D39, [ 7804] = 0x7D3C, [ 7805] = 0x7D40, [ 7806] = 0x7D30,
  [ 7807] = 0x7D33, [ 7808] = 0x7D44, [ 7809] = 0x7D2F, [ 7810] = 0x7D42,
  [ 7811] = 0x7D32, [ 7812] = 0x7D31, [ 7813] = 0x7F3D, [ 7814] = 0x7F9E,
  [ 7815] = 0x7F9A, [ 7816] = 0x7FCC, [ 7817] = 0x7FCE, [ 7818] = 0x7FD2,
  [ 7819] = 0x801C, [ 7820] = 0x804A, [ 7821] = 0x8046, [ 7822] = 0x812F,
  [ 7823] = 0x8116, [ 7824] = 0x8123, [ 7825] = 0x812B, [ 7826] = 0x8129,
  [ 7827] = 0x8130, [ 7828] = 0x8124, [ 7829] = 0x8202, [ 7830] = 0x8235,
  [ 7831] = 0x8237, [ 7832] = 0x8236, [ 7833] = 0x8239, [ 7834] = 0x838E,
  [ 7835] = 0x839E, [ 7836] = 0x8398, [ 7837] = 0x8378, [ 7838] = 0x83A2,
  [ 7839] = 0x8396, [ 7840] = 0x83BD, [ 7841] = 0x83AB, [ 7842] = 0x8392,
  [ 7843] = 0x838A, [ 7844] = 0x8393, [ 7845] = 0x8389, [ 7846] = 0x83A0,
  [ 7847] = 0x8377, [ 7848] = 0x837B, [ 7849] = 0x837C, [ 7850] = 0x8386,
  [ 7851] = 0x83A7, [ 7852] = 0x8655, [ 7853] = 0x5F6A, [ 7854] = 0x86C7,
  [ 7855] = 0x86C0, [ 7856] = 0x86B6, [ 7857] = 0x86C4, [ 7858] = 0x86B5,
  [ 7859] = 0x86C6, [ 7860] = 0x86CB, [ 7861] = 0x86B1, [ 7862] = 0x86AF,
  [ 7863] = 0x86C9, [ 7864] = 0x8853, [ 7865] = 0x889E, [ 7866] = 0x8888,
  [ 7867] = 0x88AB, [ 7868] = 0x8892, [ 7869] = 0x8896, [ 7870] = 0x888D,
  [ 7871] = 0x888B, [ 7872] = 0x8993, [ 7873] = 0x898F, [ 7874] = 0x8A2A,
  [ 7875] = 0x8A1D, [ 7876] = 0x8A23, [ 7877] = 0x8A25, [ 7878] = 0x8A31,
  [ 7879] = 0x8A2D, [ 7880] = 0x8A1F, [ 7881] = 0x8A1B, [ 7882] = 0x8A22,
  [ 7883] = 0x8C49, [ 7884] = 0x8C5A, [ 7885] = 0x8CA9, [ 7886] = 0x8CAC,
  [ 7887] = 0x8CAB, [ 7888] = 0x8CA8, [ 7889] = 0x8CAA, [ 7890] = 0x8CA7,
  [ 7891] = 0x8D67, [ 7892] = 0x8D66, [ 7893] = 0x8DBE, [ 7894] = 0x8DBA,
  [ 7895] = 0x8EDB, [ 7896] = 0x8EDF, [ 7897] = 0x9019, [ 7898] = 0x900D,
  [ 7899] = 0x901A, [ 7900] = 0x9017, [ 7901] = 0x9023, [ 7902] = 0x901F,
  [ 7903] = 0x901D, [ 7904] = 0x9010, [ 7905] = 0x9015, [ 7906] = 0x901E,
  [ 7907] = 0x9020, [ 7908] = 0x900F, [ 7909] = 0x9022, [ 7910] = 0x9016,
  [ 7911] = 0x901B, [ 7912] = 0x9014, [ 7913] = 0x90E8, [ 7914] = 0x90ED,
  [ 7915] = 0x90FD, [ 7916] = 0x9157, [ 7917] = 0x91CE, [ 7918] = 0x91F5,
  [ 7919] = 0x91E6, [ 7920] = 0x91E3, [ 7921] = 0x91E7, [ 7922] = 0x91ED,
  [ 7923] = 0x91E9, [ 7924] = 0x9589, [ 7925] = 0x966A, [ 7926] = 0x9675,
  [ 7927] = 0x9673, [ 7928] = 0x9678, [ 7929] = 0x9670, [ 7930] = 0x9674,
  [ 7931] = 0x9676, [ 7932] = 0x9677, [ 7933] = 0x966C, [ 7934] = 0x96C0,
  [ 7935] = 0x96EA, [ 7936] = 0x96E9, [ 7937] = 0x7AE0, [ 7938] = 0x7ADF,
  [ 7939] = 0x9802, [ 7940] = 0x9803, [ 7941] = 0x9B5A, [ 7942] = 0x9CE5,
  [ 7943] = 0x9E75, [ 7944] = 0x9E7F, [ 7945] = 0x9EA5, [ 7946] = 0x9EBB,
  [ 7947] = 0x50A2, [ 7948] = 0x508D, [ 7949] = 0x5085, [ 7950] = 0x5099,
  [ 7951] = 0x5091, [ 7952] = 0x5080, [ 7953] = 0x5096, [ 7954] = 0x5098,
  [ 7955] = 0x509A, [ 7956] = 0x6700, [ 7957] = 0x51F1, [ 7958] = 0x5272,
  [ 7959] = 0x5274, [ 7960] = 0x5275, [ 7961] = 0x5269, [ 7962] = 0x52DE,
  [ 7963] = 0x52DD, [ 7964] = 0x52DB, [ 7965] = 0x535A, [ 7966] = 0x53A5,
  [ 7967] = 0x557B, [ 7968] = 0x5580, [ 7969] = 0x55A7, [ 7970] = 0x557C,
  [ 7971] = 0x558A, [ 7972] = 0x559D, [ 7973] = 0x5598, [ 7974] = 0x5582,
  [ 7975] = 0x559C, [ 7976] = 0x55AA, [ 7977] = 0x5594, [ 7978] = 0x5587,
  [ 7979] = 0x558B, [ 7980] = 0x5583, [ 7981] = 0x55B3, [ 7982] = 0x55AE,
  [ 7983] = 0x559F, [ 7984] = 0x553E, [ 7985] = 0x55B2, [ 7986] = 0x559A,
  [ 7987] = 0x55BB, [ 7988] = 0x55AC, [ 7989] = 0x55B1, [ 7990] = 0x557E,
  [ 7991] = 0x5589, [ 7992] = 0x55AB, [ 7993] = 0x5599, [ 7994] = 0x570D,
  [ 7995] = 0x582F, [ 7996] = 0x582A, [ 7997] = 0x5834, [ 7998] = 0x5824,
  [ 7999] = 0x5830, [ 8000] = 0x5831, [ 8001] = 0x5821, [ 8002] = 0x581D,
  [ 8003] = 0x5820, [ 8004] = 0x58F9, [ 8005] = 0x58FA, [ 8006] = 0x5960,
  [ 8007] = 0x5A77, [ 8008] = 0x5A9A, [ 8009] = 0x5A7F, [ 8010] = 0x5A92,
  [ 8011] = 0x5A9B, [ 8012] = 0x5AA7, [ 8013] = 0x5B73, [ 8014] = 0x5B71,
  [ 8015] = 0x5BD2, [ 8016] = 0x5BCC, [ 8017] = 0x5BD3, [ 8018] = 0x5BD0,
  [ 8019] = 0x5C0A, [ 8020] = 0x5C0B, [ 8021] = 0x5C31, [ 8022] = 0x5D4C,
  [ 8023] = 0x5D50, [ 8024] = 0x5D34, [ 8025] = 0x5D47, [ 8026] = 0x5DFD,
  [ 8027] = 0x5E45, [ 8028] = 0x5E3D, [ 8029] = 0x5E40, [ 8030] = 0x5E43,
  [ 8031] = 0x5E7E, [ 8032] = 0x5ECA, [ 8033] = 0x5EC1, [ 8034] = 0x5EC2,
  [ 8035] = 0x5EC4, [ 8036] = 0x5F3C, [ 8037] = 0x5F6D, [ 8038] = 0x5FA9,
  [ 8039] = 0x5FAA, [ 8040] = 0x5FA8, [ 8041] = 0x60D1, [ 8042] = 0x60E1,
  [ 8043] = 0x60B2, [ 8044] = 0x60B6, [ 8045] = 0x60E0, [ 8046] = 0x611C,
  [ 8047] = 0x6123, [ 8048] = 0x60FA, [ 8049] = 0x6115, [ 8050] = 0x60F0,
  [ 8051] = 0x60FB, [ 8052] = 0x60F4, [ 8053] = 0x6168, [ 8054] = 0x60F1,
  [ 8055] = 0x610E, [ 8056] = 0x60F6, [ 8057] = 0x6109, [ 8058] = 0x6100,
  [ 8059] = 0x6112, [ 8060] = 0x621F, [ 8061] = 0x6249, [ 8062] = 0x63A3,
  [ 8063] = 0x638C, [ 8064] = 0x63CF, [ 8065] = 0x63C0, [ 8066] = 0x63E9,
  [ 8067] = 0x63C9, [ 8068] = 0x63C6, [ 8069] = 0x63CD, [ 8070] = 0x63D2,
  [ 8071] = 0x63E3, [ 8072] = 0x63D0, [ 8073] = 0x63E1, [ 8074] = 0x63D6,
  [ 8075] = 0x63ED, [ 8076] = 0x63EE, [ 8077] = 0x6376, [ 8078] = 0x63F4,
  [ 8079] = 0x63EA, [ 8080] = 0x63DB, [ 8081] = 0x6452, [ 8082] = 0x63DA,
  [ 8083] = 0x63F9, [ 8084] = 0x655E, [ 8085] = 0x6566, [ 8086] = 0x6562,
  [ 8087] = 0x6563, [ 8088] = 0x6591, [ 8089] = 0x6590, [ 8090] = 0x65AF,
  [ 8091] = 0x666E, [ 8092] = 0x6670, [ 8093] = 0x6674, [ 8094] = 0x6676,
  [ 8095] = 0x666F, [ 8096] = 0x6691, [ 8097] = 0x667A, [ 8098] = 0x667E,
  [ 8099] = 0x6677, [ 8100] = 0x66FE, [ 8101] = 0x66FF, [ 8102] = 0x671F,
  [ 8103] = 0x671D, [ 8104] = 0x68FA, [ 8105] = 0x68D5, [ 8106] = 0x68E0,
  [ 8107] = 0x68D8, [ 8108] = 0x68D7, [ 8109] = 0x6905, [ 8110] = 0x68DF,
  [ 8111] = 0x68F5, [ 8112] = 0x68EE, [ 8113] = 0x68E7, [ 8114] = 0x68F9,
  [ 8115] = 0x68D2, [ 8116] = 0x68F2, [ 8117] = 0x68E3, [ 8118] = 0x68CB,
  [ 8119] = 0x68CD, [ 8120] = 0x690D, [ 8121] = 0x6912, [ 8122] = 0x690E,
  [ 8123] = 0x68C9, [ 8124] = 0x68DA, [ 8125] = 0x696E, [ 8126] = 0x68FB,
  [ 8127] = 0x6B3E, [ 8128] = 0x6B3A, [ 8129] = 0x6B3D, [ 8130] = 0x6B98,
  [ 8131] = 0x6B96, [ 8132] = 0x6BBC, [ 8133] = 0x6BEF, [ 8134] = 0x6C2E,
  [ 8135] = 0x6C2F, [ 8136] = 0x6C2C, [ 8137] = 0x6E2F, [ 8138] = 0x6E38,
  [ 8139] = 0x6E54, [ 8140] = 0x6E21, [ 8141] = 0x6E32, [ 8142] = 0x6E67,
  [ 8143] = 0x6E4A, [ 8144] = 0x6E20, [ 8145] = 0x6E25, [ 8146] = 0x6E23,
  [ 8147] = 0x6E1B, [ 8148] = 0x6E5B, [ 8149] = 0x6E58, [ 8150] = 0x6E24,
  [ 8151] = 0x6E56, [ 8152] = 0x6E6E, [ 8153] = 0x6E2D, [ 8154] = 0x6E26,
  [ 8155] = 0x6E6F, [ 8156] = 0x6E34, [ 8157] = 0x6E4D, [ 8158] = 0x6E3A,
  [ 8159] = 0x6E2C, [ 8160] = 0x6E43, [ 8161] = 0x6E1D, [ 8162] = 0x6E3E,
  [ 8163] = 0x6ECB, [ 8164] = 0x6E89, [ 8165] = 0x6E19, [ 8166] = 0x6E4E,
  [ 8167] = 0x6E63, [ 8168] = 0x6E44, [ 8169] = 0x6E72, [ 8170] = 0x6E69,
  [ 8171] = 0x6E5F, [ 8172] = 0x7119, [ 8173] = 0x711A, [ 8174] = 0x7126,
  [ 8175] = 0x7130, [ 8176] = 0x7121, [ 8177] = 0x7136, [ 8178] = 0x716E,
  [ 8179] = 0x711C, [ 8180] = 0x724C, [ 8181] = 0x7284, [ 8182] = 0x7280,
  [ 8183] = 0x7336, [ 8184] = 0x7325, [ 8185] = 0x7334, [ 8186] = 0x7329,
  [ 8187] = 0x743A, [ 8188] = 0x742A, [ 8189] = 0x7433, [ 8190] = 0x7422,
  [ 8191] = 0x7425, [ 8192] = 0x7435, [ 8193] = 0x7436, [ 8194] = 0x7434,
  [ 8195] = 0x742F, [ 8196] = 0x741B, [ 8197] = 0x7426, [ 8198] = 0x7428,
  [ 8199] = 0x7525, [ 8200] = 0x7526, [ 8201] = 0x756B, [ 8202] = 0x756A,
  [ 8203] = 0x75E2, [ 8204] = 0x75DB, [ 8205] = 0x75E3, [ 8206] = 0x75D9,
  [ 8207] = 0x75D8, [ 8208] = 0x75DE, [ 8209] = 0x75E0, [ 8210] = 0x767B,
  [ 8211] = 0x767C, [ 8212] = 0x7696, [ 8213] = 0x7693, [ 8214] = 0x76B4,
  [ 8215] = 0x76DC, [ 8216] = 0x774F, [ 8217] = 0x77ED, [ 8218] = 0x785D,
  [ 8219] = 0x786C, [ 8220] = 0x786F, [ 8221] = 0x7A0D, [ 8222] = 0x7A08,
  [ 8223] = 0x7A0B, [ 8224] = 0x7A05, [ 8225] = 0x7A00, [ 8226] = 0x7A98,
  [ 8227] = 0x7A97, [ 8228] = 0x7A96, [ 8229] = 0x7AE5, [ 8230] = 0x7AE3,
  [ 8231] = 0x7B49, [ 8232] = 0x7B56, [ 8233] = 0x7B46, [ 8234] = 0x7B50,
  [ 8235] = 0x7B52, [ 8236] = 0x7B54, [ 8237] = 0x7B4D, [ 8238] = 0x7B4B,
  [ 8239] = 0x7B4F, [ 8240] = 0x7B51, [ 8241] = 0x7C9F, [ 8242] = 0x7CA5,
  [ 8243] = 0x7D5E, [ 8244] = 0x7D50, [ 8245] = 0x7D68, [ 8246] = 0x7D55,
  [ 8247] = 0x7D2B, [ 8248] = 0x7D6E, [ 8249] = 0x7D72, [ 8250] = 0x7D61,
  [ 8251] = 0x7D66, [ 8252] = 0x7D62, [ 8253] = 0x7D70, [ 8254] = 0x7D73,
  [ 8255] = 0x5584, [ 8256] = 0x7FD4, [ 8257] = 0x7FD5, [ 8258] = 0x800B,
  [ 8259] = 0x8052, [ 8260] = 0x8085, [ 8261] = 0x8155, [ 8262] = 0x8154,
  [ 8263] = 0x814B, [ 8264] = 0x8151, [ 8265] = 0x814E, [ 8266] = 0x8139,
  [ 8267] = 0x8146, [ 8268] = 0x813E, [ 8269] = 0x814C, [ 8270] = 0x8153,
  [ 8271] = 0x8174, [ 8272] = 0x8212, [ 8273] = 0x821C, [ 8274] = 0x83E9,
  [ 8275] = 0x8403, [ 8276] = 0x83F8, [ 8277] = 0x840D, [ 8278] = 0x83E0,
  [ 8279] = 0x83C5, [ 8280] = 0x840B, [ 8281] = 0x83C1, [ 8282] = 0x83EF,
  [ 8283] = 0x83F1, [ 8284] = 0x83F4, [ 8285] = 0x8457, [ 8286] = 0x840A,
  [ 8287] = 0x83F0, [ 8288] = 0x840C, [ 8289] = 0x83CC, [ 8290] = 0x83FD,
  [ 8291] = 0x83F2, [ 8292] = 0x83CA, [ 8293] = 0x8438, [ 8294] = 0x840E,
  [ 8295] = 0x8404, [ 8296] = 0x83DC, [ 8297] = 0x8407, [ 8298] = 0x83D4,
  [ 8299] = 0x83DF, [ 8300] = 0x865B, [ 8301] = 0x86DF, [ 8302] = 0x86D9,
  [ 8303] = 0x86ED, [ 8304] = 0x86D4, [ 8305] = 0x86DB, [ 8306] = 0x86E4,
  [ 8307] = 0x86D0, [ 8308] = 0x86DE, [ 8309] = 0x8857, [ 8310] = 0x88C1,
  [ 8311] = 0x88C2, [ 8312] = 0x88B1, [ 8313] = 0x8983, [ 8314] = 0x8996,
  [ 8315] = 0x8A3B, [ 8316] = 0x8A60, [ 8317] = 0x8A55, [ 8318] = 0x8A5E,
  [ 8319] = 0x8A3C, [ 8320] = 0x8A41, [ 8321] = 0x8A54, [ 8322] = 0x8A5B,
  [ 8323] = 0x8A50, [ 8324] = 0x8A46, [ 8325] = 0x8A34, [ 8326] = 0x8A3A,
  [ 8327] = 0x8A36, [ 8328] = 0x8A56, [ 8329] = 0x8C61, [ 8330] = 0x8C82,
  [ 8331] = 0x8CAF, [ 8332] = 0x8CBC, [ 8333] = 0x8CB3, [ 8334] = 0x8CBD,
  [ 8335] = 0x8CC1, [ 8336] = 0x8CBB, [ 8337] = 0x8CC0, [ 8338] = 0x8CB4,
  [ 8339] = 0x8CB7, [ 8340] = 0x8CB6, [ 8341] = 0x8CBF, [ 8342] = 0x8CB8,
  [ 8343] = 0x8D8A, [ 8344] = 0x8D85, [ 8345] = 0x8D81, [ 8346] = 0x8DCE,
  [ 8347] = 0x8DDD, [ 8348] = 0x8DCB, [ 8349] = 0x8DDA, [ 8350] = 0x8DD1,
  [ 8351] = 0x8DCC, [ 8352] = 0x8DDB, [ 8353] = 0x8DC6, [ 8354] = 0x8EFB,
  [ 8355] = 0x8EF8, [ 8356] = 0x8EFC, [ 8357] = 0x8F9C, [ 8358] = 0x902E,
  [ 8359] = 0x9035, [ 8360] = 0x9031, [ 8361] = 0x9038, [ 8362] = 0x9032,
  [ 8363] = 0x9036, [ 8364] = 0x9102, [ 8365] = 0x90F5, [ 8366] = 0x9109,
  [ 8367] = 0x90FE, [ 8368] = 0x9163, [ 8369] = 0x9165, [ 8370] = 0x91CF,
  [ 8371] = 0x9214, [ 8372] = 0x9215, [ 8373] = 0x9223, [ 8374] = 0x9209,
  [ 8375] = 0x921E, [ 8376] = 0x920D, [ 8377] = 0x9210, [ 8378] = 0x9207,
  [ 8379] = 0x9211, [ 8380] = 0x9594, [ 8381] = 0x958F, [ 8382] = 0x958B,
  [ 8383] = 0x9591, [ 8384] = 0x9593, [ 8385] = 0x9592, [ 8386] = 0x958E,
  [ 8387] = 0x968A, [ 8388] = 0x968E, [ 8389] = 0x968B, [ 8390] = 0x967D,
  [ 8391] = 0x9685, [ 8392] = 0x9686, [ 8393] = 0x968D, [ 8394] = 0x9672,
  [ 8395] = 0x9684, [ 8396] = 0x96C1, [ 8397] = 0x96C5, [ 8398] = 0x96C4,
  [ 8399] = 0x96C6, [ 8400] = 0x96C7, [ 8401] = 0x96EF, [ 8402] = 0x96F2,
  [ 8403] = 0x97CC, [ 8404] = 0x9805, [ 8405] = 0x9806, [ 8406] = 0x9808,
  [ 8407] = 0x98E7, [ 8408] = 0x98EA, [ 8409] = 0x98EF, [ 8410] = 0x98E9,
  [ 8411] = 0x98F2, [ 8412] = 0x98ED, [ 8413] = 0x99AE, [ 8414] = 0x99AD,
  [ 8415] = 0x9EC3, [ 8416] = 0x9ECD, [ 8417] = 0x9ED1, [ 8418] = 0x4E82,
  [ 8419] = 0x50AD, [ 8420] = 0x50B5, [ 8421] = 0x50B2, [ 8422] = 0x50B3,
  [ 8423] = 0x50C5, [ 8424] = 0x50BE, [ 8425] = 0x50AC, [ 8426] = 0x50B7,
  [ 8427] = 0x50BB, [ 8428] = 0x50AF, [ 8429] = 0x50C7, [ 8430] = 0x527F,
  [ 8431] = 0x5277, [ 8432] = 0x527D, [ 8433] = 0x52DF, [ 8434] = 0x52E6,
  [ 8435] = 0x52E4, [ 8436] = 0x52E2, [ 8437] = 0x52E3, [ 8438] = 0x532F,
  [ 8439] = 0x55DF, [ 8440] = 0x55E8, [ 8441] = 0x55D3, [ 8442] = 0x55E6,
  [ 8443] = 0x55CE, [ 8444] = 0x55DC, [ 8445] = 0x55C7, [ 8446] = 0x55D1,
  [ 8447] = 0x55E3, [ 8448] = 0x55E4, [ 8449] = 0x55EF, [ 8450] = 0x55DA,
  [ 8451] = 0x55E1, [ 8452] = 0x55C5, [ 8453] = 0x55C6, [ 8454] = 0x55E5,
  [ 8455] = 0x55C9, [ 8456] = 0x5712, [ 8457] = 0x5713, [ 8458] = 0x585E,
  [ 8459] = 0x5851, [ 8460] = 0x5858, [ 8461] = 0x5857, [ 8462] = 0x585A,
  [ 8463] = 0x5854, [ 8464] = 0x586B, [ 8465] = 0x584C, [ 8466] = 0x586D,
  [ 8467] = 0x584A, [ 8468] = 0x5862, [ 8469] = 0x5852, [ 8470] = 0x584B,
  [ 8471] = 0x5967, [ 8472] = 0x5AC1, [ 8473] = 0x5AC9, [ 8474] = 0x5ACC,
  [ 8475] = 0x5ABE, [ 8476] = 0x5ABD, [ 8477] = 0x5ABC, [ 8478] = 0x5AB3,
  [ 8479] = 0x5AC2, [ 8480] = 0x5AB2, [ 8481] = 0x5D69, [ 8482] = 0x5D6F,
  [ 8483] = 0x5E4C, [ 8484] = 0x5E79, [ 8485] = 0x5EC9, [ 8486] = 0x5EC8,
  [ 8487] = 0x5F12, [ 8488] = 0x5F59, [ 8489] = 0x5FAC, [ 8490] = 0x5FAE,
  [ 8491] = 0x611A, [ 8492] = 0x610F, [ 8493] = 0x6148, [ 8494] = 0x611F,
  [ 8495] = 0x60F3, [ 8496] = 0x611B, [ 8497] = 0x60F9, [ 8498] = 0x6101,
  [ 8499] = 0x6108, [ 8500] = 0x614E, [ 8501] = 0x614C, [ 8502] = 0x6144,
  [ 8503] = 0x614D, [ 8504] = 0x613E, [ 8505] = 0x6134, [ 8506] = 0x6127,
  [ 8507] = 0x610D, [ 8508] = 0x6106, [ 8509] = 0x6137, [ 8510] = 0x6221,
  [ 8511] = 0x6222, [ 8512] = 0x6413, [ 8513] = 0x643E, [ 8514] = 0x641E,
  [ 8515] = 0x642A, [ 8516] = 0x642D, [ 8517] = 0x643D, [ 8518] = 0x642C,
  [ 8519] = 0x640F, [ 8520] = 0x641C, [ 8521] = 0x6414, [ 8522] = 0x640D,
  [ 8523] = 0x6436, [ 8524] = 0x6416, [ 8525] = 0x6417, [ 8526] = 0x6406,
  [ 8527] = 0x656C, [ 8528] = 0x659F, [ 8529] = 0x65B0, [ 8530] = 0x6697,
  [ 8531] = 0x6689, [ 8532] = 0x6687, [ 8533] = 0x6688, [ 8534] = 0x6696,
  [ 8535] = 0x6684, [ 8536] = 0x6698, [ 8537] = 0x668D, [ 8538] = 0x6703,
  [ 8539] = 0x6994, [ 8540] = 0x696D, [ 8541] = 0x695A, [ 8542] = 0x6977,
  [ 8543] = 0x6960, [ 8544] = 0x6954, [ 8545] = 0x6975, [ 8546] = 0x6930,
  [ 8547] = 0x6982, [ 8548] = 0x694A, [ 8549] = 0x6968, [ 8550] = 0x696B,
  [ 8551] = 0x695E, [ 8552] = 0x6953, [ 8553] = 0x6979, [ 8554] = 0x6986,
  [ 8555] = 0x695D, [ 8556] = 0x6963, [ 8557] = 0x695B, [ 8558] = 0x6B47,
  [ 8559] = 0x6B72, [ 8560] = 0x6BC0, [ 8561] = 0x6BBF, [ 8562] = 0x6BD3,
  [ 8563] = 0x6BFD, [ 8564] = 0x6EA2, [ 8565] = 0x6EAF, [ 8566] = 0x6ED3,
  [ 8567] = 0x6EB6, [ 8568] = 0x6EC2, [ 8569] = 0x6E90, [ 8570] = 0x6E9D,
  [ 8571] = 0x6EC7, [ 8572] = 0x6EC5, [ 8573] = 0x6EA5, [ 8574] = 0x6E98,
  [ 8575] = 0x6EBC, [ 8576] = 0x6EBA, [ 8577] = 0x6EAB, [ 8578] = 0x6ED1,
  [ 8579] = 0x6E96, [ 8580] = 0x6E9C, [ 8581] = 0x6EC4, [ 8582] = 0x6ED4,
  [ 8583] = 0x6EAA, [ 8584] = 0x6EA7, [ 8585] = 0x6EB4, [ 8586] = 0x714E,
  [ 8587] = 0x7159, [ 8588] = 0x7169, [ 8589] = 0x7164, [ 8590] = 0x7149,
  [ 8591] = 0x7167, [ 8592] = 0x715C, [ 8593] = 0x716C, [ 8594] = 0x7166,
  [ 8595] = 0x714C, [ 8596] = 0x7165, [ 8597] = 0x715E, [ 8598] = 0x7146,
  [ 8599] = 0x7168, [ 8600] = 0x7156, [ 8601] = 0x723A, [ 8602] = 0x7252,
  [ 8603] = 0x7337, [ 8604] = 0x7345, [ 8605] = 0x733F, [ 8606] = 0x733E,
  [ 8607] = 0x746F, [ 8608] = 0x745A, [ 8609] = 0x7455, [ 8610] = 0x745F,
  [ 8611] = 0x745E, [ 8612] = 0x7441, [ 8613] = 0x743F, [ 8614] = 0x7459,
  [ 8615] = 0x745B, [ 8616] = 0x745C, [ 8617] = 0x7576, [ 8618] = 0x7578,
  [ 8619] = 0x7600, [ 8620] = 0x75F0, [ 8621] = 0x7601, [ 8622] = 0x75F2,
  [ 8623] = 0x75F1, [ 8624] = 0x75FA, [ 8625] = 0x75FF, [ 8626] = 0x75F4,
  [ 8627] = 0x75F3, [ 8628] = 0x76DE, [ 8629] = 0x76DF, [ 8630] = 0x775B,
  [ 8631] = 0x776B, [ 8632] = 0x7766, [ 8633] = 0x775E, [ 8634] = 0x7763,
  [ 8635] = 0x7779, [ 8636] = 0x776A, [ 8637] = 0x776C, [ 8638] = 0x775C,
  [ 8639] = 0x7765, [ 8640] = 0x7768, [ 8641] = 0x7762, [ 8642] = 0x77EE,
  [ 8643] = 0x788E, [ 8644] = 0x78B0, [ 8645] = 0x7897, [ 8646] = 0x7898,
  [ 8647] = 0x788C, [ 8648] = 0x7889, [ 8649] = 0x787C, [ 8650] = 0x7891,
  [ 8651] = 0x7893, [ 8652] = 0x787F, [ 8653] = 0x797A, [ 8654] = 0x797F,
  [ 8655] = 0x7981, [ 8656] = 0x842C, [ 8657] = 0x79BD, [ 8658] = 0x7A1C,
  [ 8659] = 0x7A1A, [ 8660] = 0x7A20, [ 8661] = 0x7A14, [ 8662] = 0x7A1F,
  [ 8663] = 0x7A1E, [ 8664] = 0x7A9F, [ 8665] = 0x7AA0, [ 8666] = 0x7B77,
  [ 8667] = 0x7BC0, [ 8668] = 0x7B60, [ 8669] = 0x7B6E, [ 8670] = 0x7B67,
  [ 8671] = 0x7CB1, [ 8672] = 0x7CB3, [ 8673] = 0x7CB5, [ 8674] = 0x7D93,
  [ 8675] = 0x7D79, [ 8676] = 0x7D91, [ 8677] = 0x7D81, [ 8678] = 0x7D8F,
  [ 8679] = 0x7D5B, [ 8680] = 0x7F6E, [ 8681] = 0x7F69, [ 8682] = 0x7F6A,
  [ 8683] = 0x7F72, [ 8684] = 0x7FA9, [ 8685] = 0x7FA8, [ 8686] = 0x7FA4,
  [ 8687] = 0x8056, [ 8688] = 0x8058, [ 8689] = 0x8086, [ 8690] = 0x8084,
  [ 8691] = 0x8171, [ 8692] = 0x8170, [ 8693] = 0x8178, [ 8694] = 0x8165,
  [ 8695] = 0x816E, [ 8696] = 0x8173, [ 8697] = 0x816B, [ 8698] = 0x8179,
  [ 8699] = 0x817A, [ 8700] = 0x8166, [ 8701] = 0x8205, [ 8702] = 0x8247,
  [ 8703] = 0x8482, [ 8704] = 0x8477, [ 8705] = 0x843D, [ 8706] = 0x8431,
  [ 8707] = 0x8475, [ 8708] = 0x8466, [ 8709] = 0x846B, [ 8710] = 0x8449,
  [ 8711] = 0x846C, [ 8712] = 0x845B, [ 8713] = 0x843C, [ 8714] = 0x8435,
  [ 8715] = 0x8461, [ 8716] = 0x8463, [ 8717] = 0x8469, [ 8718] = 0x846D,
  [ 8719] = 0x8446, [ 8720] = 0x865E, [ 8721] = 0x865C, [ 8722] = 0x865F,
  [ 8723] = 0x86F9, [ 8724] = 0x8713, [ 8725] = 0x8708, [ 8726] = 0x8707,
  [ 8727] = 0x8700, [ 8728] = 0x86FE, [ 8729] = 0x86FB, [ 8730] = 0x8702,
  [ 8731] = 0x8703, [ 8732] = 0x8706, [ 8733] = 0x870A, [ 8734] = 0x8859,
  [ 8735] = 0x88DF, [ 8736] = 0x88D4, [ 8737] = 0x88D9, [ 8738] = 0x88DC,
  [ 8739] = 0x88D8, [ 8740] = 0x88DD, [ 8741] = 0x88E1, [ 8742] = 0x88CA,
  [ 8743] = 0x88D5, [ 8744] = 0x88D2, [ 8745] = 0x899C, [ 8746] = 0x89E3,
  [ 8747] = 0x8A6B, [ 8748] = 0x8A72, [ 8749] = 0x8A73, [ 8750] = 0x8A66,
  [ 8751] = 0x8A69, [ 8752] = 0x8A70, [ 8753] = 0x8A87, [ 8754] = 0x8A7C,
  [ 8755] = 0x8A63, [ 8756] = 0x8AA0, [ 8757] = 0x8A71, [ 8758] = 0x8A85,
  [ 8759] = 0x8A6D, [ 8760] = 0x8A62, [ 8761] = 0x8A6E, [ 8762] = 0x8A6C,
  [ 8763] = 0x8A79, [ 8764] = 0x8A7B, [ 8765] = 0x8A3E, [ 8766] = 0x8A68,
  [ 8767] = 0x8C62, [ 8768] = 0x8C8A, [ 8769] = 0x8C89, [ 8770] = 0x8CCA,
  [ 8771] = 0x8CC7, [ 8772] = 0x8CC8, [ 8773] = 0x8CC4, [ 8774] = 0x8CB2,
  [ 8775] = 0x8CC3, [ 8776] = 0x8CC2, [ 8777] = 0x8CC5, [ 8778] = 0x8DE1,
  [ 8779] = 0x8DDF, [ 8780] = 0x8DE8, [ 8781] = 0x8DEF, [ 8782] = 0x8DF3,
  [ 8783] = 0x8DFA, [ 8784] = 0x8DEA, [ 8785] = 0x8DE4, [ 8786] = 0x8DE6,
  [ 8787] = 0x8EB2, [ 8788] = 0x8F03, [ 8789] = 0x8F09, [ 8790] = 0x8EFE,
  [ 8791] = 0x8F0A, [ 8792] = 0x8F9F, [ 8793] = 0x8FB2, [ 8794] = 0x904B,
  [ 8795] = 0x904A, [ 8796] = 0x9053, [ 8797] = 0x9042, [ 8798] = 0x9054,
  [ 8799] = 0x903C, [ 8800] = 0x9055, [ 8801] = 0x9050, [ 8802] = 0x9047,
  [ 8803] = 0x904F, [ 8804] = 0x904E, [ 8805] = 0x904D, [ 8806] = 0x9051,
  [ 8807] = 0x903E, [ 8808] = 0x9041, [ 8809] = 0x9112, [ 8810] = 0x9117,
  [ 8811] = 0x916C, [ 8812] = 0x916A, [ 8813] = 0x9169, [ 8814] = 0x91C9,
  [ 8815] = 0x9237, [ 8816] = 0x9257, [ 8817] = 0x9238, [ 8818] = 0x923D,
  [ 8819] = 0x9240, [ 8820] = 0x923E, [ 8821] = 0x925B, [ 8822] = 0x924B,
  [ 8823] = 0x9264, [ 8824] = 0x9251, [ 8825] = 0x9234, [ 8826] = 0x9249,
  [ 8827] = 0x924D, [ 8828] = 0x9245, [ 8829] = 0x9239, [ 8830] = 0x923F,
  [ 8831] = 0x925A, [ 8832] = 0x9598, [ 8833] = 0x9698, [ 8834] = 0x9694,
  [ 8835] = 0x9695, [ 8836] = 0x96CD, [ 8837] = 0x96CB, [ 8838] = 0x96C9,
  [ 8839] = 0x96CA, [ 8840] = 0x96F7, [ 8841] = 0x96FB, [ 8842] = 0x96F9,
  [ 8843] = 0x96F6, [ 8844] = 0x9756, [ 8845] = 0x9774, [ 8846] = 0x9776,
  [ 8847] = 0x9810, [ 8848] = 0x9811, [ 8849] = 0x9813, [ 8850] = 0x980A,
  [ 8851] = 0x9812, [ 8852] = 0x980C, [ 8853] = 0x98FC, [ 8854] = 0x98F4,
  [ 8855] = 0x98FD, [ 8856] = 0x98FE, [ 8857] = 0x99B3, [ 8858] = 0x99B1,
  [ 8859] = 0x99B4, [ 8860] = 0x9AE1, [ 8861] = 0x9CE9, [ 8862] = 0x9E82,
  [ 8863] = 0x9F0E, [ 8864] = 0x9F13, [ 8865] = 0x9F20, [ 8866] = 0x50E7,
  [ 8867] = 0x50EE, [ 8868] = 0x50E5, [ 8869] = 0x50D6, [ 8870] = 0x50ED,
  [ 8871] = 0x50DA, [ 8872] = 0x50D5, [ 8873] = 0x50CF, [ 8874] = 0x50D1,
  [ 8875] = 0x50F1, [ 8876] = 0x50CE, [ 8877] = 0x50E9, [ 8878] = 0x5162,
  [ 8879] = 0x51F3, [ 8880] = 0x5283, [ 8881] = 0x5282, [ 8882] = 0x5331,
  [ 8883] = 0x53AD, [ 8884] = 0x55FE, [ 8885] = 0x5600, [ 8886] = 0x561B,
  [ 8887] = 0x5617, [ 8888] = 0x55FD, [ 8889] = 0x5614, [ 8890] = 0x5606,
  [ 8891] = 0x5609, [ 8892] = 0x560D, [ 8893] = 0x560E, [ 8894] = 0x55F7,
  [ 8895] = 0x5616, [ 8896] = 0x561F, [ 8897] = 0x5608, [ 8898] = 0x5610,
  [ 8899] = 0x55F6, [ 8900] = 0x5718, [ 8901] = 0x5716, [ 8902] = 0x5875,
  [ 8903] = 0x587E, [ 8904] = 0x5883, [ 8905] = 0x5893, [ 8906] = 0x588A,
  [ 8907] = 0x5879, [ 8908] = 0x5885, [ 8909] = 0x587D, [ 8910] = 0x58FD,
  [ 8911] = 0x5925, [ 8912] = 0x5922, [ 8913] = 0x5924, [ 8914] = 0x596A,
  [ 8915] = 0x5969, [ 8916] = 0x5AE1, [ 8917] = 0x5AE6, [ 8918] = 0x5AE9,
  [ 8919] = 0x5AD7, [ 8920] = 0x5AD6, [ 8921] = 0x5AD8, [ 8922] = 0x5AE3,
  [ 8923] = 0x5B75, [ 8924] = 0x5BDE, [ 8925] = 0x5BE7, [ 8926] = 0x5BE1,
  [ 8927] = 0x5BE5, [ 8928] = 0x5BE6, [ 8929] = 0x5BE8, [ 8930] = 0x5BE2,
  [ 8931] = 0x5BE4, [ 8932] = 0x5BDF, [ 8933] = 0x5C0D, [ 8934] = 0x5C62,
  [ 8935] = 0x5D84, [ 8936] = 0x5D87, [ 8937] = 0x5E5B, [ 8938] = 0x5E63,
  [ 8939] = 0x5E55, [ 8940] = 0x5E57, [ 8941] = 0x5E54, [ 8942] = 0x5ED3,
  [ 8943] = 0x5ED6, [ 8944] = 0x5F0A, [ 8945] = 0x5F46, [ 8946] = 0x5F70,
  [ 8947] = 0x5FB9, [ 8948] = 0x6147, [ 8949] = 0x613F, [ 8950] = 0x614B,
  [ 8951] = 0x6177, [ 8952] = 0x6162, [ 8953] = 0x6163, [ 8954] = 0x615F,
  [ 8955] = 0x615A, [ 8956] = 0x6158, [ 8957] = 0x6175, [ 8958] = 0x622A,
  [ 8959] = 0x6487, [ 8960] = 0x6458, [ 8961] = 0x6454, [ 8962] = 0x64A4,
  [ 8963] = 0x6478, [ 8964] = 0x645F, [ 8965] = 0x647A, [ 8966] = 0x6451,
  [ 8967] = 0x6467, [ 8968] = 0x6434, [ 8969] = 0x646D, [ 8970] = 0x647B,
  [ 8971] = 0x6572, [ 8972] = 0x65A1, [ 8973] = 0x65D7, [ 8974] = 0x65D6,
  [ 8975] = 0x66A2, [ 8976] = 0x66A8, [ 8977] = 0x669D, [ 8978] = 0x699C,
  [ 8979] = 0x69A8, [ 8980] = 0x6995, [ 8981] = 0x69C1, [ 8982] = 0x69AE,
  [ 8983] = 0x69D3, [ 8984] = 0x69CB, [ 8985] = 0x699B, [ 8986] = 0x69B7,
  [ 8987] = 0x69BB, [ 8988] = 0x69AB, [ 8989] = 0x69B4, [ 8990] = 0x69D0,
  [ 8991] = 0x69CD, [ 8992] = 0x69AD, [ 8993] = 0x69CC, [ 8994] = 0x69A6,
  [ 8995] = 0x69C3, [ 8996] = 0x69A3, [ 8997] = 0x6B49, [ 8998] = 0x6B4C,
  [ 8999] = 0x6C33, [ 9000] = 0x6F33, [ 9001] = 0x6F14, [ 9002] = 0x6EFE,
  [ 9003] = 0x6F13, [ 9004] = 0x6EF4, [ 9005] = 0x6F29, [ 9006] = 0x6F3E,
  [ 9007] = 0x6F20, [ 9008] = 0x6F2C, [ 9009] = 0x6F0F, [ 9010] = 0x6F02,
  [ 9011] = 0x6F22, [ 9012] = 0x6EFF, [ 9013] = 0x6EEF, [ 9014] = 0x6F06,
  [ 9015] = 0x6F31, [ 9016] = 0x6F38, [ 9017] = 0x6F32, [ 9018] = 0x6F23,
  [ 9019] = 0x6F15, [ 9020] = 0x6F2B, [ 9021] = 0x6F2F, [ 9022] = 0x6F88,
  [ 9023] = 0x6F2A, [ 9024] = 0x6EEC, [ 9025] = 0x6F01, [ 9026] = 0x6EF2,
  [ 9027] = 0x6ECC, [ 9028] = 0x6EF7, [ 9029] = 0x7194, [ 9030] = 0x7199,
  [ 9031] = 0x717D, [ 9032] = 0x718A, [ 9033] = 0x7184, [ 9034] = 0x7192,
  [ 9035] = 0x723E, [ 9036] = 0x7292, [ 9037] = 0x7296, [ 9038] = 0x7344,
  [ 9039] = 0x7350, [ 9040] = 0x7464, [ 9041] = 0x7463, [ 9042] = 0x746A,
  [ 9043] = 0x7470, [ 9044] = 0x746D, [ 9045] = 0x7504, [ 9046] = 0x7591,
  [ 9047] = 0x7627, [ 9048] = 0x760D, [ 9049] = 0x760B, [ 9050] = 0x7609,
  [ 9051] = 0x7613, [ 9052] = 0x76E1, [ 9053] = 0x76E3, [ 9054] = 0x7784,
  [ 9055] = 0x777D, [ 9056] = 0x777F, [ 9057] = 0x7761, [ 9058] = 0x78C1,
  [ 9059] = 0x789F, [ 9060] = 0x78A7, [ 9061] = 0x78B3, [ 9062] = 0x78A9,
  [ 9063] = 0x78A3, [ 9064] = 0x798E, [ 9065] = 0x798F, [ 9066] = 0x798D,
  [ 9067] = 0x7A2E, [ 9068] = 0x7A31, [ 9069] = 0x7AAA, [ 9070] = 0x7AA9,
  [ 9071] = 0x7AED, [ 9072] = 0x7AEF, [ 9073] = 0x7BA1, [ 9074] = 0x7B95,
  [ 9075] = 0x7B8B, [ 9076] = 0x7B75, [ 9077] = 0x7B97, [ 9078] = 0x7B9D,
  [ 9079] = 0x7B94, [ 9080] = 0x7B8F, [ 9081] = 0x7BB8, [ 9082] = 0x7B87,
  [ 9083] = 0x7B84, [ 9084] = 0x7CB9, [ 9085] = 0x7CBD, [ 9086] = 0x7CBE,
  [ 9087] = 0x7DBB, [ 9088] = 0x7DB0, [ 9089] = 0x7D9C, [ 9090] = 0x7DBD,
  [ 9091] = 0x7DBE, [ 9092] = 0x7DA0, [ 9093] = 0x7DCA, [ 9094] = 0x7DB4,
  [ 9095] = 0x7DB2, [ 9096] = 0x7DB1, [ 9097] = 0x7DBA, [ 9098] = 0x7DA2,
  [ 9099] = 0x7DBF, [ 9100] = 0x7DB5, [ 9101] = 0x7DB8, [ 9102] = 0x7DAD,
  [ 9103] = 0x7DD2, [ 9104] = 0x7DC7, [ 9105] = 0x7DAC, [ 9106] = 0x7F70,
  [ 9107] = 0x7FE0, [ 9108] = 0x7FE1, [ 9109] = 0x7FDF, [ 9110] = 0x805E,
  [ 9111] = 0x805A, [ 9112] = 0x8087, [ 9113] = 0x8150, [ 9114] = 0x8180,
  [ 9115] = 0x818F, [ 9116] = 0x8188, [ 9117] = 0x818A, [ 9118] = 0x817F,
  [ 9119] = 0x8182, [ 9120] = 0x81E7, [ 9121] = 0x81FA, [ 9122] = 0x8207,
  [ 9123] = 0x8214, [ 9124] = 0x821E, [ 9125] = 0x824B, [ 9126] = 0x84C9,
  [ 9127] = 0x84BF, [ 9128] = 0x84C6, [ 9129] = 0x84C4, [ 9130] = 0x8499,
  [ 9131] = 0x849E, [ 9132] = 0x84B2, [ 9133] = 0x849C, [ 9134] = 0x84CB,
  [ 9135] = 0x84B8, [ 9136] = 0x84C0, [ 9137] = 0x84D3, [ 9138] = 0x8490,
  [ 9139] = 0x84BC, [ 9140] = 0x84D1, [ 9141] = 0x84CA, [ 9142] = 0x873F,
  [ 9143] = 0x871C, [ 9144] = 0x873B, [ 9145] = 0x8722, [ 9146] = 0x8725,
  [ 9147] = 0x8734, [ 9148] = 0x8718, [ 9149] = 0x8755, [ 9150] = 0x8737,
  [ 9151] = 0x8729, [ 9152] = 0x88F3, [ 9153] = 0x8902, [ 9154] = 0x88F4,
  [ 9155] = 0x88F9, [ 9156] = 0x88F8, [ 9157] = 0x88FD, [ 9158] = 0x88E8,
  [ 9159] = 0x891A, [ 9160] = 0x88EF, [ 9161] = 0x8AA6, [ 9162] = 0x8A8C,
  [ 9163] = 0x8A9E, [ 9164] = 0x8AA3, [ 9165] = 0x8A8D, [ 9166] = 0x8AA1,
  [ 9167] = 0x8A93, [ 9168] = 0x8AA4, [ 9169] = 0x8AAA, [ 9170] = 0x8AA5,
  [ 9171] = 0x8AA8, [ 9172] = 0x8A98, [ 9173] = 0x8A91, [ 9174] = 0x8A9A,
  [ 9175] = 0x8AA7, [ 9176] = 0x8C6A, [ 9177] = 0x8C8D, [ 9178] = 0x8C8C,
  [ 9179] = 0x8CD3, [ 9180] = 0x8CD1, [ 9181] = 0x8CD2, [ 9182] = 0x8D6B,
  [ 9183] = 0x8D99, [ 9184] = 0x8D95, [ 9185] = 0x8DFC, [ 9186] = 0x8F14,
  [ 9187] = 0x8F12, [ 9188] = 0x8F15, [ 9189] = 0x8F13, [ 9190] = 0x8FA3,
  [ 9191] = 0x9060, [ 9192] = 0x9058, [ 9193] = 0x905C, [ 9194] = 0x9063,
  [ 9195] = 0x9059, [ 9196] = 0x905E, [ 9197] = 0x9062, [ 9198] = 0x905D,
  [ 9199] = 0x905B, [ 9200] = 0x9119, [ 9201] = 0x9118, [ 9202] = 0x911E,
  [ 9203] = 0x9175, [ 9204] = 0x9178, [ 9205] = 0x9177, [ 9206] = 0x9174,
  [ 9207] = 0x9278, [ 9208] = 0x9280, [ 9209] = 0x9285, [ 9210] = 0x9298,
  [ 9211] = 0x9296, [ 9212] = 0x927B, [ 9213] = 0x9293, [ 9214] = 0x929C,
  [ 9215] = 0x92A8, [ 9216] = 0x927C, [ 9217] = 0x9291, [ 9218] = 0x95A1,
  [ 9219] = 0x95A8, [ 9220] = 0x95A9, [ 9221] = 0x95A3, [ 9222] = 0x95A5,
  [ 9223] = 0x95A4, [ 9224] = 0x9699, [ 9225] = 0x969C, [ 9226] = 0x969B,
  [ 9227] = 0x96CC, [ 9228] = 0x96D2, [ 9229] = 0x9700, [ 9230] = 0x977C,
  [ 9231] = 0x9785, [ 9232] = 0x97F6, [ 9233] = 0x9817, [ 9234] = 0x9818,
  [ 9235] = 0x98AF, [ 9236] = 0x98B1, [ 9237] = 0x9903, [ 9238] = 0x9905,
  [ 9239] = 0x990C, [ 9240] = 0x9909, [ 9241] = 0x99C1, [ 9242] = 0x9AAF,
  [ 9243] = 0x9AB0, [ 9244] = 0x9AE6, [ 9245] = 0x9B41, [ 9246] = 0x9B42,
  [ 9247] = 0x9CF4, [ 9248] = 0x9CF6, [ 9249] = 0x9CF3, [ 9250] = 0x9EBC,
  [ 9251] = 0x9F3B, [ 9252] = 0x9F4A, [ 9253] = 0x5104, [ 9254] = 0x5100,
  [ 9255] = 0x50FB, [ 9256] = 0x50F5, [ 9257] = 0x50F9, [ 9258] = 0x5102,
  [ 9259] = 0x5108, [ 9260] = 0x5109, [ 9261] = 0x5105, [ 9262] = 0x51DC,
  [ 9263] = 0x5287, [ 9264] = 0x5288, [ 9265] = 0x5289, [ 9266] = 0x528D,
  [ 9267] = 0x528A, [ 9268] = 0x52F0, [ 9269] = 0x53B2, [ 9270] = 0x562E,
  [ 9271] = 0x563B, [ 9272] = 0x5639, [ 9273] = 0x5632, [ 9274] = 0x563F,
  [ 9275] = 0x5634, [ 9276] = 0x5629, [ 9277] = 0x5653, [ 9278] = 0x564E,
  [ 9279] = 0x5657, [ 9280] = 0x5674, [ 9281] = 0x5636, [ 9282] = 0x562F,
  [ 9283] = 0x5630, [ 9284] = 0x5880, [ 9285] = 0x589F, [ 9286] = 0x589E,
  [ 9287] = 0x58B3, [ 9288] = 0x589C, [ 9289] = 0x58AE, [ 9290] = 0x58A9,
  [ 9291] = 0x58A6, [ 9292] = 0x596D, [ 9293] = 0x5B09, [ 9294] = 0x5AFB,
  [ 9295] = 0x5B0B, [ 9296] = 0x5AF5, [ 9297] = 0x5B0C, [ 9298] = 0x5B08,
  [ 9299] = 0x5BEE, [ 9300] = 0x5BEC, [ 9301] = 0x5BE9, [ 9302] = 0x5BEB,
  [ 9303] = 0x5C64, [ 9304] = 0x5C65, [ 9305] = 0x5D9D, [ 9306] = 0x5D94,
  [ 9307] = 0x5E62, [ 9308] = 0x5E5F, [ 9309] = 0x5E61, [ 9310] = 0x5EE2,
  [ 9311] = 0x5EDA, [ 9312] = 0x5EDF, [ 9313] = 0x5EDD, [ 9314] = 0x5EE3,
  [ 9315] = 0x5EE0, [ 9316] = 0x5F48, [ 9317] = 0x5F71, [ 9318] = 0x5FB7,
  [ 9319] = 0x5FB5, [ 9320] = 0x6176, [ 9321] = 0x6167, [ 9322] = 0x616E,
  [ 9323] = 0x615D, [ 9324] = 0x6155, [ 9325] = 0x6182, [ 9326] = 0x617C,
  [ 9327] = 0x6170, [ 9328] = 0x616B, [ 9329] = 0x617E, [ 9330] = 0x61A7,
  [ 9331] = 0x6190, [ 9332] = 0x61AB, [ 9333] = 0x618E, [ 9334] = 0x61AC,
  [ 9335] = 0x619A, [ 9336] = 0x61A4, [ 9337] = 0x6194, [ 9338] = 0x61AE,
  [ 9339] = 0x622E, [ 9340] = 0x6469, [ 9341] = 0x646F, [ 9342] = 0x6479,
  [ 9343] = 0x649E, [ 9344] = 0x64B2, [ 9345] = 0x6488, [ 9346] = 0x6490,
  [ 9347] = 0x64B0, [ 9348] = 0x64A5, [ 9349] = 0x6493, [ 9350] = 0x6495,
  [ 9351] = 0x64A9, [ 9352] = 0x6492, [ 9353] = 0x64AE, [ 9354] = 0x64AD,
  [ 9355] = 0x64AB, [ 9356] = 0x649A, [ 9357] = 0x64AC, [ 9358] = 0x6499,
  [ 9359] = 0x64A2, [ 9360] = 0x64B3, [ 9361] = 0x6575, [ 9362] = 0x6577,
  [ 9363] = 0x6578, [ 9364] = 0x66AE, [ 9365] = 0x66AB, [ 9366] = 0x66B4,
  [ 9367] = 0x66B1, [ 9368] = 0x6A23, [ 9369] = 0x6A1F, [ 9370] = 0x69E8,
  [ 9371] = 0x6A01, [ 9372] = 0x6A1E, [ 9373] = 0x6A19, [ 9374] = 0x69FD,
  [ 9375] = 0x6A21, [ 9376] = 0x6A13, [ 9377] = 0x6A0A, [ 9378] = 0x69F3,
  [ 9379] = 0x6A02, [ 9380] = 0x6A05, [ 9381] = 0x69ED, [ 9382] = 0x6A11,
  [ 9383] = 0x6B50, [ 9384] = 0x6B4E, [ 9385] = 0x6BA4, [ 9386] = 0x6BC5,
  [ 9387] = 0x6BC6, [ 9388] = 0x6F3F, [ 9389] = 0x6F7C, [ 9390] = 0x6F84,
  [ 9391] = 0x6F51, [ 9392] = 0x6F66, [ 9393] = 0x6F54, [ 9394] = 0x6F86,
  [ 9395] = 0x6F6D, [ 9396] = 0x6F5B, [ 9397] = 0x6F78, [ 9398] = 0x6F6E,
  [ 9399] = 0x6F8E, [ 9400] = 0x6F7A, [ 9401] = 0x6F70, [ 9402] = 0x6F64,
  [ 9403] = 0x6F97, [ 9404] = 0x6F58, [ 9405] = 0x6ED5, [ 9406] = 0x6F6F,
  [ 9407] = 0x6F60, [ 9408] = 0x6F5F, [ 9409] = 0x719F, [ 9410] = 0x71AC,
  [ 9411] = 0x71B1, [ 9412] = 0x71A8, [ 9413] = 0x7256, [ 9414] = 0x729B,
  [ 9415] = 0x734E, [ 9416] = 0x7357, [ 9417] = 0x7469, [ 9418] = 0x748B,
  [ 9419] = 0x7483, [ 9420] = 0x747E, [ 9421] = 0x7480, [ 9422] = 0x757F,
  [ 9423] = 0x7620, [ 9424] = 0x7629, [ 9425] = 0x761F, [ 9426] = 0x7624,
  [ 9427] = 0x7626, [ 9428] = 0x7621, [ 9429] = 0x7622, [ 9430] = 0x769A,
  [ 9431] = 0x76BA, [ 9432] = 0x76E4, [ 9433] = 0x778E, [ 9434] = 0x7787,
  [ 9435] = 0x778C, [ 9436] = 0x7791, [ 9437] = 0x778B, [ 9438] = 0x78CB,
  [ 9439] = 0x78C5, [ 9440] = 0x78BA, [ 9441] = 0x78CA, [ 9442] = 0x78BE,
  [ 9443] = 0x78D5, [ 9444] = 0x78BC, [ 9445] = 0x78D0, [ 9446] = 0x7A3F,
  [ 9447] = 0x7A3C, [ 9448] = 0x7A40, [ 9449] = 0x7A3D, [ 9450] = 0x7A37,
  [ 9451] = 0x7A3B, [ 9452] = 0x7AAF, [ 9453] = 0x7AAE, [ 9454] = 0x7BAD,
  [ 9455] = 0x7BB1, [ 9456] = 0x7BC4, [ 9457] = 0x7BB4, [ 9458] = 0x7BC6,
  [ 9459] = 0x7BC7, [ 9460] = 0x7BC1, [ 9461] = 0x7BA0, [ 9462] = 0x7BCC,
  [ 9463] = 0x7CCA, [ 9464] = 0x7DE0, [ 9465] = 0x7DF4, [ 9466] = 0x7DEF,
  [ 9467] = 0x7DFB, [ 9468] = 0x7DD8, [ 9469] = 0x7DEC, [ 9470] = 0x7DDD,
  [ 9471] = 0x7DE8, [ 9472] = 0x7DE3, [ 9473] = 0x7DDA, [ 9474] = 0x7DDE,
  [ 9475] = 0x7DE9, [ 9476] = 0x7D9E, [ 9477] = 0x7DD9, [ 9478] = 0x7DF2,
  [ 9479] = 0x7DF9, [ 9480] = 0x7F75, [ 9481] = 0x7F77, [ 9482] = 0x7FAF,
  [ 9483] = 0x7FE9, [ 9484] = 0x8026, [ 9485] = 0x819B, [ 9486] = 0x819C,
  [ 9487] = 0x819D, [ 9488] = 0x81A0, [ 9489] = 0x819A, [ 9490] = 0x8198,
  [ 9491] = 0x8517, [ 9492] = 0x853D, [ 9493] = 0x851A, [ 9494] = 0x84EE,
  [ 9495] = 0x852C, [ 9496] = 0x852D, [ 9497] = 0x8513, [ 9498] = 0x8511,
  [ 9499] = 0x8523, [ 9500] = 0x8521, [ 9501] = 0x8514, [ 9502] = 0x84EC,
  [ 9503] = 0x8525, [ 9504] = 0x84FF, [ 9505] = 0x8506, [ 9506] = 0x8782,
  [ 9507] = 0x8774, [ 9508] = 0x8776, [ 9509] = 0x8760, [ 9510] = 0x8766,
  [ 9511] = 0x8778, [ 9512] = 0x8768, [ 9513] = 0x8759, [ 9514] = 0x8757,
  [ 9515] = 0x874C, [ 9516] = 0x8753, [ 9517] = 0x885B, [ 9518] = 0x885D,
  [ 9519] = 0x8910, [ 9520] = 0x8907, [ 9521] = 0x8912, [ 9522] = 0x8913,
  [ 9523] = 0x8915, [ 9524] = 0x890A, [ 9525] = 0x8ABC, [ 9526] = 0x8AD2,
  [ 9527] = 0x8AC7, [ 9528] = 0x8AC4, [ 9529] = 0x8A95, [ 9530] = 0x8ACB,
  [ 9531] = 0x8AF8, [ 9532] = 0x8AB2, [ 9533] = 0x8AC9, [ 9534] = 0x8AC2,
  [ 9535] = 0x8ABF, [ 9536] = 0x8AB0, [ 9537] = 0x8AD6, [ 9538] = 0x8ACD,
  [ 9539] = 0x8AB6, [ 9540] = 0x8AB9, [ 9541] = 0x8ADB, [ 9542] = 0x8C4C,
  [ 9543] = 0x8C4E, [ 9544] = 0x8C6C, [ 9545] = 0x8CE0, [ 9546] = 0x8CDE,
  [ 9547] = 0x8CE6, [ 9548] = 0x8CE4, [ 9549] = 0x8CEC, [ 9550] = 0x8CED,
  [ 9551] = 0x8CE2, [ 9552] = 0x8CE3, [ 9553] = 0x8CDC, [ 9554] = 0x8CEA,
  [ 9555] = 0x8CE1, [ 9556] = 0x8D6D, [ 9557] = 0x8D9F, [ 9558] = 0x8DA3,
  [ 9559] = 0x8E2B, [ 9560] = 0x8E10, [ 9561] = 0x8E1D, [ 9562] = 0x8E22,
  [ 9563] = 0x8E0F, [ 9564] = 0x8E29, [ 9565] = 0x8E1F, [ 9566] = 0x8E21,
  [ 9567] = 0x8E1E, [ 9568] = 0x8EBA, [ 9569] = 0x8F1D, [ 9570] = 0x8F1B,
  [ 9571] = 0x8F1F, [ 9572] = 0x8F29, [ 9573] = 0x8F26, [ 9574] = 0x8F2A,
  [ 9575] = 0x8F1C, [ 9576] = 0x8F1E, [ 9577] = 0x8F25, [ 9578] = 0x9069,
  [ 9579] = 0x906E, [ 9580] = 0x9068, [ 9581] = 0x906D, [ 9582] = 0x9077,
  [ 9583] = 0x9130, [ 9584] = 0x912D, [ 9585] = 0x9127, [ 9586] = 0x9131,
  [ 9587] = 0x9187, [ 9588] = 0x9189, [ 9589] = 0x918B, [ 9590] = 0x9183,
  [ 9591] = 0x92C5, [ 9592] = 0x92BB, [ 9593] = 0x92B7, [ 9594] = 0x92EA,
  [ 9595] = 0x92AC, [ 9596] = 0x92E4, [ 9597] = 0x92C1, [ 9598] = 0x92B3,
  [ 9599] = 0x92BC, [ 9600] = 0x92D2, [ 9601] = 0x92C7, [ 9602] = 0x92F0,
  [ 9603] = 0x92B2, [ 9604] = 0x95AD, [ 9605] = 0x95B1, [ 9606] = 0x9704,
  [ 9607] = 0x9706, [ 9608] = 0x9707, [ 9609] = 0x9709, [ 9610] = 0x9760,
  [ 9611] = 0x978D, [ 9612] = 0x978B, [ 9613] = 0x978F, [ 9614] = 0x9821,
  [ 9615] = 0x982B, [ 9616] = 0x981C, [ 9617] = 0x98B3, [ 9618] = 0x990A,
  [ 9619] = 0x9913, [ 9620] = 0x9912, [ 9621] = 0x9918, [ 9622] = 0x99DD,
  [ 9623] = 0x99D0, [ 9624] = 0x99DF, [ 9625] = 0x99DB, [ 9626] = 0x99D1,
  [ 9627] = 0x99D5, [ 9628] = 0x99D2, [ 9629] = 0x99D9, [ 9630] = 0x9AB7,
  [ 9631] = 0x9AEE, [ 9632] = 0x9AEF, [ 9633] = 0x9B27, [ 9634] = 0x9B45,
  [ 9635] = 0x9B44, [ 9636] = 0x9B77, [ 9637] = 0x9B6F, [ 9638] = 0x9D06,
  [ 9639] = 0x9D09, [ 9640] = 0x9D03, [ 9641] = 0x9EA9, [ 9642] = 0x9EBE,
  [ 9643] = 0x9ECE, [ 9644] = 0x58A8, [ 9645] = 0x9F52, [ 9646] = 0x5112,
  [ 9647] = 0x5118, [ 9648] = 0x5114, [ 9649] = 0x5110, [ 9650] = 0x5115,
  [ 9651] = 0x5180, [ 9652] = 0x51AA, [ 9653] = 0x51DD, [ 9654] = 0x5291,
  [ 9655] = 0x5293, [ 9656] = 0x52F3, [ 9657] = 0x5659, [ 9658] = 0x566B,
  [ 9659] = 0x5679, [ 9660] = 0x5669, [ 9661] = 0x5664, [ 9662] = 0x5678,
  [ 9663] = 0x566A, [ 9664] = 0x5668, [ 9665] = 0x5665, [ 9666] = 0x5671,
  [ 9667] = 0x566F, [ 9668] = 0x566C, [ 9669] = 0x5662, [ 9670] = 0x5676,
  [ 9671] = 0x58C1, [ 9672] = 0x58BE, [ 9673] = 0x58C7, [ 9674] = 0x58C5,
  [ 9675] = 0x596E, [ 9676] = 0x5B1D, [ 9677] = 0x5B34, [ 9678] = 0x5B78,
  [ 9679] = 0x5BF0, [ 9680] = 0x5C0E, [ 9681] = 0x5F4A, [ 9682] = 0x61B2,
  [ 9683] = 0x6191, [ 9684] = 0x61A9, [ 9685] = 0x618A, [ 9686] = 0x61CD,
  [ 9687] = 0x61B6, [ 9688] = 0x61BE, [ 9689] = 0x61CA, [ 9690] = 0x61C8,
  [ 9691] = 0x6230, [ 9692] = 0x64C5, [ 9693] = 0x64C1, [ 9694] = 0x64CB,
  [ 9695] = 0x64BB, [ 9696] = 0x64BC, [ 9697] = 0x64DA, [ 9698] = 0x64C4,
  [ 9699] = 0x64C7, [ 9700] = 0x64C2, [ 9701] = 0x64CD, [ 9702] = 0x64BF,
  [ 9703] = 0x64D2, [ 9704] = 0x64D4, [ 9705] = 0x64BE, [ 9706] = 0x6574,
  [ 9707] = 0x66C6, [ 9708] = 0x66C9, [ 9709] = 0x66B9, [ 9710] = 0x66C4,
  [ 9711] = 0x66C7, [ 9712] = 0x66B8, [ 9713] = 0x6A3D, [ 9714] = 0x6A38,
  [ 9715] = 0x6A3A, [ 9716] = 0x6A59, [ 9717] = 0x6A6B, [ 9718] = 0x6A58,
  [ 9719] = 0x6A39, [ 9720] = 0x6A44, [ 9721] = 0x6A62, [ 9722] = 0x6A61,
  [ 9723] = 0x6A4B, [ 9724] = 0x6A47, [ 9725] = 0x6A35, [ 9726] = 0x6A5F,
  [ 9727] = 0x6A48, [ 9728] = 0x6B59, [ 9729] = 0x6B77, [ 9730] = 0x6C05,
  [ 9731] = 0x6FC2, [ 9732] = 0x6FB1, [ 9733] = 0x6FA1, [ 9734] = 0x6FC3,
  [ 9735] = 0x6FA4, [ 9736] = 0x6FC1, [ 9737] = 0x6FA7, [ 9738] = 0x6FB3,
  [ 9739] = 0x6FC0, [ 9740] = 0x6FB9, [ 9741] = 0x6FB6, [ 9742] = 0x6FA6,
  [ 9743] = 0x6FA0, [ 9744] = 0x6FB4, [ 9745] = 0x71BE, [ 9746] = 0x71C9,
  [ 9747] = 0x71D0, [ 9748] = 0x71D2, [ 9749] = 0x71C8, [ 9750] = 0x71D5,
  [ 9751] = 0x71B9, [ 9752] = 0x71CE, [ 9753] = 0x71D9, [ 9754] = 0x71DC,
  [ 9755] = 0x71C3, [ 9756] = 0x71C4, [ 9757] = 0x7368, [ 9758] = 0x749C,
  [ 9759] = 0x74A3, [ 9760] = 0x7498, [ 9761] = 0x749F, [ 9762] = 0x749E,
  [ 9763] = 0x74E2, [ 9764] = 0x750C, [ 9765] = 0x750D, [ 9766] = 0x7634,
  [ 9767] = 0x7638, [ 9768] = 0x763A, [ 9769] = 0x76E7, [ 9770] = 0x76E5,
  [ 9771] = 0x77A0, [ 9772] = 0x779E, [ 9773] = 0x779F, [ 9774] = 0x77A5,
  [ 9775] = 0x78E8, [ 9776] = 0x78DA, [ 9777] = 0x78EC, [ 9778] = 0x78E7,
  [ 9779] = 0x79A6, [ 9780] = 0x7A4D, [ 9781] = 0x7A4E, [ 9782] = 0x7A46,
  [ 9783] = 0x7A4C, [ 9784] = 0x7A4B, [ 9785] = 0x7ABA, [ 9786] = 0x7BD9,
  [ 9787] = 0x7C11, [ 9788] = 0x7BC9, [ 9789] = 0x7BE4, [ 9790] = 0x7BDB,
  [ 9791] = 0x7BE1, [ 9792] = 0x7BE9, [ 9793] = 0x7BE6, [ 9794] = 0x7CD5,
  [ 9795] = 0x7CD6, [ 9796] = 0x7E0A, [ 9797] = 0x7E11, [ 9798] = 0x7E08,
  [ 9799] = 0x7E1B, [ 9800] = 0x7E23, [ 9801] = 0x7E1E, [ 9802] = 0x7E1D,
  [ 9803] = 0x7E09, [ 9804] = 0x7E10, [ 9805] = 0x7F79, [ 9806] = 0x7FB2,
  [ 9807] = 0x7FF0, [ 9808] = 0x7FF1, [ 9809] = 0x7FEE, [ 9810] = 0x8028,
  [ 9811] = 0x81B3, [ 9812] = 0x81A9, [ 9813] = 0x81A8, [ 9814] = 0x81FB,
  [ 9815] = 0x8208, [ 9816] = 0x8258, [ 9817] = 0x8259, [ 9818] = 0x854A,
  [ 9819] = 0x8559, [ 9820] = 0x8548, [ 9821] = 0x8568, [ 9822] = 0x8569,
  [ 9823] = 0x8543, [ 9824] = 0x8549, [ 9825] = 0x856D, [ 9826] = 0x856A,
  [ 9827] = 0x855E, [ 9828] = 0x8783, [ 9829] = 0x879F, [ 9830] = 0x879E,
  [ 9831] = 0x87A2, [ 9832] = 0x878D, [ 9833] = 0x8861, [ 9834] = 0x892A,
  [ 9835] = 0x8932, [ 9836] = 0x8925, [ 9837] = 0x892B, [ 9838] = 0x8921,
  [ 9839] = 0x89AA, [ 9840] = 0x89A6, [ 9841] = 0x8AE6, [ 9842] = 0x8AFA,
  [ 9843] = 0x8AEB, [ 9844] = 0x8AF1, [ 9845] = 0x8B00, [ 9846] = 0x8ADC,
  [ 9847] = 0x8AE7, [ 9848] = 0x8AEE, [ 9849] = 0x8AFE, [ 9850] = 0x8B01,
  [ 9851] = 0x8B02, [ 9852] = 0x8AF7, [ 9853] = 0x8AED, [ 9854] = 0x8AF3,
  [ 9855] = 0x8AF6, [ 9856] = 0x8AFC, [ 9857] = 0x8C6B, [ 9858] = 0x8C6D,
  [ 9859] = 0x8C93, [ 9860] = 0x8CF4, [ 9861] = 0x8E44, [ 9862] = 0x8E31,
  [ 9863] = 0x8E34, [ 9864] = 0x8E42, [ 9865] = 0x8E39, [ 9866] = 0x8E35,
  [ 9867] = 0x8F3B, [ 9868] = 0x8F2F, [ 9869] = 0x8F38, [ 9870] = 0x8F33,
  [ 9871] = 0x8FA8, [ 9872] = 0x8FA6, [ 9873] = 0x9075, [ 9874] = 0x9074,
  [ 9875] = 0x9078, [ 9876] = 0x9072, [ 9877] = 0x907C, [ 9878] = 0x907A,
  [ 9879] = 0x9134, [ 9880] = 0x9192, [ 9881] = 0x9320, [ 9882] = 0x9336,
  [ 9883] = 0x92F8, [ 9884] = 0x9333, [ 9885] = 0x932F, [ 9886] = 0x9322,
  [ 9887] = 0x92FC, [ 9888] = 0x932B, [ 9889] = 0x9304, [ 9890] = 0x931A,
  [ 9891] = 0x9310, [ 9892] = 0x9326, [ 9893] = 0x9321, [ 9894] = 0x9315,
  [ 9895] = 0x932E, [ 9896] = 0x9319, [ 9897] = 0x95BB, [ 9898] = 0x96A7,
  [ 9899] = 0x96A8, [ 9900] = 0x96AA, [ 9901] = 0x96D5, [ 9902] = 0x970E,
  [ 9903] = 0x9711, [ 9904] = 0x9716, [ 9905] = 0x970D, [ 9906] = 0x9713,
  [ 9907] = 0x970F, [ 9908] = 0x975B, [ 9909] = 0x975C, [ 9910] = 0x9766,
  [ 9911] = 0x9798, [ 9912] = 0x9830, [ 9913] = 0x9838, [ 9914] = 0x983B,
  [ 9915] = 0x9837, [ 9916] = 0x982D, [ 9917] = 0x9839, [ 9918] = 0x9824,
  [ 9919] = 0x9910, [ 9920] = 0x9928, [ 9921] = 0x991E, [ 9922] = 0x991B,
  [ 9923] = 0x9921, [ 9924] = 0x991A, [ 9925] = 0x99ED, [ 9926] = 0x99E2,
  [ 9927] = 0x99F1, [ 9928] = 0x9AB8, [ 9929] = 0x9ABC, [ 9930] = 0x9AFB,
  [ 9931] = 0x9AED, [ 9932] = 0x9B28, [ 9933] = 0x9B91, [ 9934] = 0x9D15,
  [ 9935] = 0x9D23, [ 9936] = 0x9D26, [ 9937] = 0x9D28, [ 9938] = 0x9D12,
  [ 9939] = 0x9D1B, [ 9940] = 0x9ED8, [ 9941] = 0x9ED4, [ 9942] = 0x9F8D,
  [ 9943] = 0x9F9C, [ 9944] = 0x512A, [ 9945] = 0x511F, [ 9946] = 0x5121,
  [ 9947] = 0x5132, [ 9948] = 0x52F5, [ 9949] = 0x568E, [ 9950] = 0x5680,
  [ 9951] = 0x5690, [ 9952] = 0x5685, [ 9953] = 0x5687, [ 9954] = 0x568F,
  [ 9955] = 0x58D5, [ 9956] = 0x58D3, [ 9957] = 0x58D1, [ 9958] = 0x58CE,
  [ 9959] = 0x5B30, [ 9960] = 0x5B2A, [ 9961] = 0x5B24, [ 9962] = 0x5B7A,
  [ 9963] = 0x5C37, [ 9964] = 0x5C68, [ 9965] = 0x5DBC, [ 9966] = 0x5DBA,
  [ 9967] = 0x5DBD, [ 9968] = 0x5DB8, [ 9969] = 0x5E6B, [ 9970] = 0x5F4C,
  [ 9971] = 0x5FBD, [ 9972] = 0x61C9, [ 9973] = 0x61C2, [ 9974] = 0x61C7,
  [ 9975] = 0x61E6, [ 9976] = 0x61CB, [ 9977] = 0x6232, [ 9978] = 0x6234,
  [ 9979] = 0x64CE, [ 9980] = 0x64CA, [ 9981] = 0x64D8, [ 9982] = 0x64E0,
  [ 9983] = 0x64F0, [ 9984] = 0x64E6, [ 9985] = 0x64EC, [ 9986] = 0x64F1,
  [ 9987] = 0x64E2, [ 9988] = 0x64ED, [ 9989] = 0x6582, [ 9990] = 0x6583,
  [ 9991] = 0x66D9, [ 9992] = 0x66D6, [ 9993] = 0x6A80, [ 9994] = 0x6A94,
  [ 9995] = 0x6A84, [ 9996] = 0x6AA2, [ 9997] = 0x6A9C, [ 9998] = 0x6ADB,
  [ 9999] = 0x6AA3, [10000] = 0x6A7E, [10001] = 0x6A97, [10002] = 0x6A90,
  [10003] = 0x6AA0, [10004] = 0x6B5C, [10005] = 0x6BAE, [10006] = 0x6BDA,
  [10007] = 0x6C08, [10008] = 0x6FD8, [10009] = 0x6FF1, [10010] = 0x6FDF,
  [10011] = 0x6FE0, [10012] = 0x6FDB, [10013] = 0x6FE4, [10014] = 0x6FEB,
  [10015] = 0x6FEF, [10016] = 0x6F80, [10017] = 0x6FEC, [10018] = 0x6FE1,
  [10019] = 0x6FE9, [10020] = 0x6FD5, [10021] = 0x6FEE, [10022] = 0x6FF0,
  [10023] = 0x71E7, [10024] = 0x71DF, [10025] = 0x71EE, [10026] = 0x71E6,
  [10027] = 0x71E5, [10028] = 0x71ED, [10029] = 0x71EC, [10030] = 0x71F4,
  [10031] = 0x71E0, [10032] = 0x7235, [10033] = 0x7246, [10034] = 0x7370,
  [10035] = 0x7372, [10036] = 0x74A9, [10037] = 0x74B0, [10038] = 0x74A6,
  [10039] = 0x74A8, [10040] = 0x7646, [10041] = 0x7642, [10042] = 0x764C,
  [10043] = 0x76EA, [10044] = 0x77B3, [10045] = 0x77AA, [10046] = 0x77B0,
  [10047] = 0x77AC, [10048] = 0x77A7, [10049] = 0x77AD, [10050] = 0x77EF,
  [10051] = 0x78F7, [10052] = 0x78FA, [10053] = 0x78F4, [10054] = 0x78EF,
  [10055] = 0x7901, [10056] = 0x79A7, [10057] = 0x79AA, [10058] = 0x7A57,
  [10059] = 0x7ABF, [10060] = 0x7C07, [10061] = 0x7C0D, [10062] = 0x7BFE,
  [10063] = 0x7BF7, [10064] = 0x7C0C, [10065] = 0x7BE0, [10066] = 0x7CE0,
  [10067] = 0x7CDC, [10068] = 0x7CDE, [10069] = 0x7CE2, [10070] = 0x7CDF,
  [10071] = 0x7CD9, [10072] = 0x7CDD, [10073] = 0x7E2E, [10074] = 0x7E3E,
  [10075] = 0x7E46, [10076] = 0x7E37, [10077] = 0x7E32, [10078] = 0x7E43,
  [10079] = 0x7E2B, [10080] = 0x7E3D, [10081] = 0x7E31, [10082] = 0x7E45,
  [10083] = 0x7E41, [10084] = 0x7E34, [10085] = 0x7E39, [10086] = 0x7E48,
  [10087] = 0x7E35, [10088] = 0x7E3F, [10089] = 0x7E2F, [10090] = 0x7F44,
  [10091] = 0x7FF3, [10092] = 0x7FFC, [10093] = 0x8071, [10094] = 0x8072,
  [10095] = 0x8070, [10096] = 0x806F, [10097] = 0x8073, [10098] = 0x81C6,
  [10099] = 0x81C3, [10100] = 0x81BA, [10101] = 0x81C2, [10102] = 0x81C0,
  [10103] = 0x81BF, [10104] = 0x81BD, [10105] = 0x81C9, [10106] = 0x81BE,
  [10107] = 0x81E8, [10108] = 0x8209, [10109] = 0x8271, [10110] = 0x85AA,
  [10111] = 0x8584, [10112] = 0x857E, [10113] = 0x859C, [10114] = 0x8591,
  [10115] = 0x8594, [10116] = 0x85AF, [10117] = 0x859B, [10118] = 0x8587,
  [10119] = 0x85A8, [10120] = 0x858A, [10121] = 0x8667, [10122] = 0x87C0,
  [10123] = 0x87D1, [10124] = 0x87B3, [10125] = 0x87D2, [10126] = 0x87C6,
  [10127] = 0x87AB, [10128] = 0x87BB, [10129] = 0x87BA, [10130] = 0x87C8,
  [10131] = 0x87CB, [10132] = 0x893B, [10133] = 0x8936, [10134] = 0x8944,
  [10135] = 0x8938, [10136] = 0x893D, [10137] = 0x89AC, [10138] = 0x8B0E,
  [10139] = 0x8B17, [10140] = 0x8B19, [10141] = 0x8B1B, [10142] = 0x8B0A,
  [10143] = 0x8B20, [10144] = 0x8B1D, [10145] = 0x8B04, [10146] = 0x8B10,
  [10147] = 0x8C41, [10148] = 0x8C3F, [10149] = 0x8C73, [10150] = 0x8CFA,
  [10151] = 0x8CFD, [10152] = 0x8CFC, [10153] = 0x8CF8, [10154] = 0x8CFB,
  [10155] = 0x8DA8, [10156] = 0x8E49, [10157] = 0x8E4B, [10158] = 0x8E48,
  [10159] = 0x8E4A, [10160] = 0x8F44, [10161] = 0x8F3E, [10162] = 0x8F42,
  [10163] = 0x8F45, [10164] = 0x8F3F, [10165] = 0x907F, [10166] = 0x907D,
  [10167] = 0x9084, [10168] = 0x9081, [10169] = 0x9082, [10170] = 0x9080,
  [10171] = 0x9139, [10172] = 0x91A3, [10173] = 0x919E, [10174] = 0x919C,
  [10175] = 0x934D, [10176] = 0x9382, [10177] = 0x9328, [10178] = 0x9375,
  [10179] = 0x934A, [10180] = 0x9365, [10181] = 0x934B, [10182] = 0x9318,
  [10183] = 0x937E, [10184] = 0x936C, [10185] = 0x935B, [10186] = 0x9370,
  [10187] = 0x935A, [10188] = 0x9354, [10189] = 0x95CA, [10190] = 0x95CB,
  [10191] = 0x95CC, [10192] = 0x95C8, [10193] = 0x95C6, [10194] = 0x96B1,
  [10195] = 0x96B8, [10196] = 0x96D6, [10197] = 0x971C, [10198] = 0x971E,
  [10199] = 0x97A0, [10200] = 0x97D3, [10201] = 0x9846, [10202] = 0x98B6,
  [10203] = 0x9935, [10204] = 0x9A01, [10205] = 0x99FF, [10206] = 0x9BAE,
  [10207] = 0x9BAB, [10208] = 0x9BAA, [10209] = 0x9BAD, [10210] = 0x9D3B,
  [10211] = 0x9D3F, [10212] = 0x9E8B, [10213] = 0x9ECF, [10214] = 0x9EDE,
  [10215] = 0x9EDC, [10216] = 0x9EDD, [10217] = 0x9EDB, [10218] = 0x9F3E,
  [10219] = 0x9F4B, [10220] = 0x53E2, [10221] = 0x5695, [10222] = 0x56AE,
  [10223] = 0x58D9, [10224] = 0x58D8, [10225] = 0x5B38, [10226] = 0x5F5D,
  [10227] = 0x61E3, [10228] = 0x6233, [10229] = 0x64F4, [10230] = 0x64F2,
  [10231] = 0x64FE, [10232] = 0x6506, [10233] = 0x64FA, [10234] = 0x64FB,
  [10235] = 0x64F7, [10236] = 0x65B7, [10237] = 0x66DC, [10238] = 0x6726,
  [10239] = 0x6AB3, [10240] = 0x6AAC, [10241] = 0x6AC3, [10242] = 0x6ABB,
  [10243] = 0x6AB8, [10244] = 0x6AC2, [10245] = 0x6AAE, [10246] = 0x6AAF,
  [10247] = 0x6B5F, [10248] = 0x6B78, [10249] = 0x6BAF, [10250] = 0x7009,
  [10251] = 0x700B, [10252] = 0x6FFE, [10253] = 0x7006, [10254] = 0x6FFA,
  [10255] = 0x7011, [10256] = 0x700F, [10257] = 0x71FB, [10258] = 0x71FC,
  [10259] = 0x71FE, [10260] = 0x71F8, [10261] = 0x7377, [10262] = 0x7375,
  [10263] = 0x74A7, [10264] = 0x74BF, [10265] = 0x7515, [10266] = 0x7656,
  [10267] = 0x7658, [10268] = 0x7652, [10269] = 0x77BD, [10270] = 0x77BF,
  [10271] = 0x77BB, [10272] = 0x77BC, [10273] = 0x790E, [10274] = 0x79AE,
  [10275] = 0x7A61, [10276] = 0x7A62, [10277] = 0x7A60, [10278] = 0x7AC4,
  [10279] = 0x7AC5, [10280] = 0x7C2B, [10281] = 0x7C27, [10282] = 0x7C2A,
  [10283] = 0x7C1E, [10284] = 0x7C23, [10285] = 0x7C21, [10286] = 0x7CE7,
  [10287] = 0x7E54, [10288] = 0x7E55, [10289] = 0x7E5E, [10290] = 0x7E5A,
  [10291] = 0x7E61, [10292] = 0x7E52, [10293] = 0x7E59, [10294] = 0x7F48,
  [10295] = 0x7FF9, [10296] = 0x7FFB, [10297] = 0x8077, [10298] = 0x8076,
  [10299] = 0x81CD, [10300] = 0x81CF, [10301] = 0x820A, [10302] = 0x85CF,
  [10303] = 0x85A9, [10304] = 0x85CD, [10305] = 0x85D0, [10306] = 0x85C9,
  [10307] = 0x85B0, [10308] = 0x85BA, [10309] = 0x85B9, [10310] = 0x85A6,
  [10311] = 0x87EF, [10312] = 0x87EC, [10313] = 0x87F2, [10314] = 0x87E0,
  [10315] = 0x8986, [10316] = 0x89B2, [10317] = 0x89F4, [10318] = 0x8B28,
  [10319] = 0x8B39, [10320] = 0x8B2C, [10321] = 0x8B2B, [10322] = 0x8C50,
  [10323] = 0x8D05, [10324] = 0x8E59, [10325] = 0x8E63, [10326] = 0x8E66,
  [10327] = 0x8E64, [10328] = 0x8E5F, [10329] = 0x8E55, [10330] = 0x8EC0,
  [10331] = 0x8F49, [10332] = 0x8F4D, [10333] = 0x9087, [10334] = 0x9083,
  [10335] = 0x9088, [10336] = 0x91AB, [10337] = 0x91AC, [10338] = 0x91D0,
  [10339] = 0x9394, [10340] = 0x938A, [10341] = 0x9396, [10342] = 0x93A2,
  [10343] = 0x93B3, [10344] = 0x93AE, [10345] = 0x93AC, [10346] = 0x93B0,
  [10347] = 0x9398, [10348] = 0x939A, [10349] = 0x9397, [10350] = 0x95D4,
  [10351] = 0x95D6, [10352] = 0x95D0, [10353] = 0x95D5, [10354] = 0x96E2,
  [10355] = 0x96DC, [10356] = 0x96D9, [10357] = 0x96DB, [10358] = 0x96DE,
  [10359] = 0x9724, [10360] = 0x97A3, [10361] = 0x97A6, [10362] = 0x97AD,
  [10363] = 0x97F9, [10364] = 0x984D, [10365] = 0x984F, [10366] = 0x984C,
  [10367] = 0x984E, [10368] = 0x9853, [10369] = 0x98BA, [10370] = 0x993E,
  [10371] = 0x993F, [10372] = 0x993D, [10373] = 0x992E, [10374] = 0x99A5,
  [10375] = 0x9A0E, [10376] = 0x9AC1, [10377] = 0x9B03, [10378] = 0x9B06,
  [10379] = 0x9B4F, [10380] = 0x9B4E, [10381] = 0x9B4D, [10382] = 0x9BCA,
  [10383] = 0x9BC9, [10384] = 0x9BFD, [10385] = 0x9BC8, [10386] = 0x9BC0,
  [10387] = 0x9D51, [10388] = 0x9D5D, [10389] = 0x9D60, [10390] = 0x9EE0,
  [10391] = 0x9F15, [10392] = 0x9F2C, [10393] = 0x5133, [10394] = 0x56A5,
  [10395] = 0x58DE, [10396] = 0x58DF, [10397] = 0x58E2, [10398] = 0x5BF5,
  [10399] = 0x9F90, [10400] = 0x5EEC, [10401] = 0x61F2, [10402] = 0x61F7,
  [10403] = 0x61F6, [10404] = 0x61F5, [10405] = 0x6500, [10406] = 0x650F,
  [10407] = 0x66E0, [10408] = 0x66DD, [10409] = 0x6AE5, [10410] = 0x6ADD,
  [10411] = 0x6ADA, [10412] = 0x6AD3, [10413] = 0x701B, [10414] = 0x701F,
  [10415] = 0x7028, [10416] = 0x701A, [10417] = 0x701D, [10418] = 0x7015,
  [10419] = 0x7018, [10420] = 0x7206, [10421] = 0x720D, [10422] = 0x7258,
  [10423] = 0x72A2, [10424] = 0x7378, [10425] = 0x737A, [10426] = 0x74BD,
  [10427] = 0x74CA, [10428] = 0x74E3, [10429] = 0x7587, [10430] = 0x7586,
  [10431] = 0x765F, [10432] = 0x7661, [10433] = 0x77C7, [10434] = 0x7919,
  [10435] = 0x79B1, [10436] = 0x7A6B, [10437] = 0x7A69, [10438] = 0x7C3E,
  [10439] = 0x7C3F, [10440] = 0x7C38, [10441] = 0x7C3D, [10442] = 0x7C37,
  [10443] = 0x7C40, [10444] = 0x7E6B, [10445] = 0x7E6D, [10446] = 0x7E79,
  [10447] = 0x7E69, [10448] = 0x7E6A, [10449] = 0x7F85, [10450] = 0x7E73,
  [10451] = 0x7FB6, [10452] = 0x7FB9, [10453] = 0x7FB8, [10454] = 0x81D8,
  [10455] = 0x85E9, [10456] = 0x85DD, [10457] = 0x85EA, [10458] = 0x85D5,
  [10459] = 0x85E4, [10460] = 0x85E5, [10461] = 0x85F7, [10462] = 0x87FB,
  [10463] = 0x8805, [10464] = 0x880D, [10465] = 0x87F9, [10466] = 0x87FE,
  [10467] = 0x8960, [10468] = 0x895F, [10469] = 0x8956, [10470] = 0x895E,
  [10471] = 0x8B41, [10472] = 0x8B5C, [10473] = 0x8B58, [10474] = 0x8B49,
  [10475] = 0x8B5A, [10476] = 0x8B4E, [10477] = 0x8B4F, [10478] = 0x8B46,
  [10479] = 0x8B59, [10480] = 0x8D08, [10481] = 0x8D0A, [10482] = 0x8E7C,
  [10483] = 0x8E72, [10484] = 0x8E87, [10485] = 0x8E76, [10486] = 0x8E6C,
  [10487] = 0x8E7A, [10488] = 0x8E74, [10489] = 0x8F54, [10490] = 0x8F4E,
  [10491] = 0x8FAD, [10492] = 0x908A, [10493] = 0x908B, [10494] = 0x91B1,
  [10495] = 0x91AE, [10496] = 0x93E1, [10497] = 0x93D1, [10498] = 0x93DF,
  [10499] = 0x93C3, [10500] = 0x93C8, [10501] = 0x93DC, [10502] = 0x93DD,
  [10503] = 0x93D6, [10504] = 0x93E2, [10505] = 0x93CD, [10506] = 0x93D8,
  [10507] = 0x93E4, [10508] = 0x93D7, [10509] = 0x93E8, [10510] = 0x95DC,
  [10511] = 0x96B4, [10512] = 0x96E3, [10513] = 0x972A, [10514] = 0x9727,
  [10515] = 0x9761, [10516] = 0x97DC, [10517] = 0x97FB, [10518] = 0x985E,
  [10519] = 0x9858, [10520] = 0x985B, [10521] = 0x98BC, [10522] = 0x9945,
  [10523] = 0x9949, [10524] = 0x9A16, [10525] = 0x9A19, [10526] = 0x9B0D,
  [10527] = 0x9BE8, [10528] = 0x9BE7, [10529] = 0x9BD6, [10530] = 0x9BDB,
  [10531] = 0x9D89, [10532] = 0x9D61, [10533] = 0x9D72, [10534] = 0x9D6A,
  [10535] = 0x9D6C, [10536] = 0x9E92, [10537] = 0x9E97, [10538] = 0x9E93,
  [10539] = 0x9EB4, [10540] = 0x52F8, [10541] = 0x56A8, [10542] = 0x56B7,
  [10543] = 0x56B6, [10544] = 0x56B4, [10545] = 0x56BC, [10546] = 0x58E4,
  [10547] = 0x5B40, [10548] = 0x5B43, [10549] = 0x5B7D, [10550] = 0x5BF6,
  [10551] = 0x5DC9, [10552] = 0x61F8, [10553] = 0x61FA, [10554] = 0x6518,
  [10555] = 0x6514, [10556] = 0x6519, [10557] = 0x66E6, [10558] = 0x6727,
  [10559] = 0x6AEC, [10560] = 0x703E, [10561] = 0x7030, [10562] = 0x7032,
  [10563] = 0x7210, [10564] = 0x737B, [10565] = 0x74CF, [10566] = 0x7662,
  [10567] = 0x7665, [10568] = 0x7926, [10569] = 0x792A, [10570] = 0x792C,
  [10571] = 0x792B, [10572] = 0x7AC7, [10573] = 0x7AF6, [10574] = 0x7C4C,
  [10575] = 0x7C43, [10576] = 0x7C4D, [10577] = 0x7CEF, [10578] = 0x7CF0,
  [10579] = 0x8FAE, [10580] = 0x7E7D, [10581] = 0x7E7C, [10582] = 0x7E82,
  [10583] = 0x7F4C, [10584] = 0x8000, [10585] = 0x81DA, [10586] = 0x8266,
  [10587] = 0x85FB, [10588] = 0x85F9, [10589] = 0x8611, [10590] = 0x85FA,
  [10591] = 0x8606, [10592] = 0x860B, [10593] = 0x8607, [10594] = 0x860A,
  [10595] = 0x8814, [10596] = 0x8815, [10597] = 0x8964, [10598] = 0x89BA,
  [10599] = 0x89F8, [10600] = 0x8B70, [10601] = 0x8B6C, [10602] = 0x8B66,
  [10603] = 0x8B6F, [10604] = 0x8B5F, [10605] = 0x8B6B, [10606] = 0x8D0F,
  [10607] = 0x8D0D, [10608] = 0x8E89, [10609] = 0x8E81, [10610] = 0x8E85,
  [10611] = 0x8E82, [10612] = 0x91B4, [10613] = 0x91CB, [10614] = 0x9418,
  [10615] = 0x9403, [10616] = 0x93FD, [10617] = 0x95E1, [10618] = 0x9730,
  [10619] = 0x98C4, [10620] = 0x9952, [10621] = 0x9951, [10622] = 0x99A8,
  [10623] = 0x9A2B, [10624] = 0x9A30, [10625] = 0x9A37, [10626] = 0x9A35,
  [10627] = 0x9C13, [10628] = 0x9C0D, [10629] = 0x9E79, [10630] = 0x9EB5,
  [10631] = 0x9EE8, [10632] = 0x9F2F, [10633] = 0x9F5F, [10634] = 0x9F63,
  [10635] = 0x9F61, [10636] = 0x5137, [10637] = 0x5138, [10638] = 0x56C1,
  [10639] = 0x56C0, [10640] = 0x56C2, [10641] = 0x5914, [10642] = 0x5C6C,
  [10643] = 0x5DCD, [10644] = 0x61FC, [10645] = 0x61FE, [10646] = 0x651D,
  [10647] = 0x651C, [10648] = 0x6595, [10649] = 0x66E9, [10650] = 0x6AFB,
  [10651] = 0x6B04, [10652] = 0x6AFA, [10653] = 0x6BB2, [10654] = 0x704C,
  [10655] = 0x721B, [10656] = 0x72A7, [10657] = 0x74D6, [10658] = 0x74D4,
  [10659] = 0x7669, [10660] = 0x77D3, [10661] = 0x7C50, [10662] = 0x7E8F,
  [10663] = 0x7E8C, [10664] = 0x7FBC, [10665] = 0x8617, [10666] = 0x862D,
  [10667] = 0x861A, [10668] = 0x8823, [10669] = 0x8822, [10670] = 0x8821,
  [10671] = 0x881F, [10672] = 0x896A, [10673] = 0x896C, [10674] = 0x89BD,
  [10675] = 0x8B74, [10676] = 0x8B77, [10677] = 0x8B7D, [10678] = 0x8D13,
  [10679] = 0x8E8A, [10680] = 0x8E8D, [10681] = 0x8E8B, [10682] = 0x8F5F,
  [10683] = 0x8FAF, [10684] = 0x91BA, [10685] = 0x942E, [10686] = 0x9433,
  [10687] = 0x9435, [10688] = 0x943A, [10689] = 0x9438, [10690] = 0x9432,
  [10691] = 0x942B, [10692] = 0x95E2, [10693] = 0x9738, [10694] = 0x9739,
  [10695] = 0x9732, [10696] = 0x97FF, [10697] = 0x9867, [10698] = 0x9865,
  [10699] = 0x9957, [10700] = 0x9A45, [10701] = 0x9A43, [10702] = 0x9A40,
  [10703] = 0x9A3E, [10704] = 0x9ACF, [10705] = 0x9B54, [10706] = 0x9B51,
  [10707] = 0x9C2D, [10708] = 0x9C25, [10709] = 0x9DAF, [10710] = 0x9DB4,
  [10711] = 0x9DC2, [10712] = 0x9DB8, [10713] = 0x9E9D, [10714] = 0x9EEF,
  [10715] = 0x9F19, [10716] = 0x9F5C, [10717] = 0x9F66, [10718] = 0x9F67,
  [10719] = 0x513C, [10720] = 0x513B, [10721] = 0x56C8, [10722] = 0x56CA,
  [10723] = 0x56C9, [10724] = 0x5B7F, [10725] = 0x5DD4, [10726] = 0x5DD2,
  [10727] = 0x5F4E, [10728] = 0x61FF, [10729] = 0x6524, [10730] = 0x6B0A,
  [10731] = 0x6B61, [10732] = 0x7051, [10733] = 0x7058, [10734] = 0x7380,
  [10735] = 0x74E4, [10736] = 0x758A, [10737] = 0x766E, [10738] = 0x766C,
  [10739] = 0x79B3, [10740] = 0x7C60, [10741] = 0x7C5F, [10742] = 0x807E,
  [10743] = 0x807D, [10744] = 0x81DF, [10745] = 0x8972, [10746] = 0x896F,
  [10747] = 0x89FC, [10748] = 0x8B80, [10749] = 0x8D16, [10750] = 0x8D17,
  [10751] = 0x8E91, [10752] = 0x8E93, [10753] = 0x8F61, [10754] = 0x9148,
  [10755] = 0x9444, [10756] = 0x9451, [10757] = 0x9452, [10758] = 0x973D,
  [10759] = 0x973E, [10760] = 0x97C3, [10761] = 0x97C1, [10762] = 0x986B,
  [10763] = 0x9955, [10764] = 0x9A55, [10765] = 0x9A4D, [10766] = 0x9AD2,
  [10767] = 0x9B1A, [10768] = 0x9C49, [10769] = 0x9C31, [10770] = 0x9C3E,
  [10771] = 0x9C3B, [10772] = 0x9DD3, [10773] = 0x9DD7, [10774] = 0x9F34,
  [10775] = 0x9F6C, [10776] = 0x9F6A, [10777] = 0x9F94, [10778] = 0x56CC,
  [10779] = 0x5DD6, [10780] = 0x6200, [10781] = 0x6523, [10782] = 0x652B,
  [10783] = 0x652A, [10784] = 0x66EC, [10785] = 0x6B10, [10786] = 0x74DA,
  [10787] = 0x7ACA, [10788] = 0x7C64, [10789] = 0x7C63, [10790] = 0x7C65,
  [10791] = 0x7E93, [10792] = 0x7E96, [10793] = 0x7E94, [10794] = 0x81E2,
  [10795] = 0x8638, [10796] = 0x863F, [10797] = 0x8831, [10798] = 0x8B8A,
  [10799] = 0x9090, [10800] = 0x908F, [10801] = 0x9463, [10802] = 0x9460,
  [10803] = 0x9464, [10804] = 0x9768, [10805] = 0x986F, [10806] = 0x995C,
  [10807] = 0x9A5A, [10808] = 0x9A5B, [10809] = 0x9A57, [10810] = 0x9AD3,
  [10811] = 0x9AD4, [10812] = 0x9AD1, [10813] = 0x9C54, [10814] = 0x9C57,
  [10815] = 0x9C56, [10816] = 0x9DE5, [10817] = 0x9E9F, [10818] = 0x9EF4,
  [10819] = 0x56D1, [10820] = 0x58E9, [10821] = 0x652C, [10822] = 0x705E,
  [10823] = 0x7671, [10824] = 0x7672, [10825] = 0x77D7, [10826] = 0x7F50,
  [10827] = 0x7F88, [10828] = 0x8836, [10829] = 0x8839, [10830] = 0x8862,
  [10831] = 0x8B93, [10832] = 0x8B92, [10833] = 0x8B96, [10834] = 0x8277,
  [10835] = 0x8D1B, [10836] = 0x91C0, [10837] = 0x946A, [10838] = 0x9742,
  [10839] = 0x9748, [10840] = 0x9744, [10841] = 0x97C6, [10842] = 0x9870,
  [10843] = 0x9A5F, [10844] = 0x9B22, [10845] = 0x9B58, [10846] = 0x9C5F,
  [10847] = 0x9DF9, [10848] = 0x9DFA, [10849] = 0x9E7C, [10850] = 0x9E7D,
  [10851] = 0x9F07, [10852] = 0x9F77, [10853] = 0x9F72, [10854] = 0x5EF3,
  [10855] = 0x6B16, [10856] = 0x7063, [10857] = 0x7C6C, [10858] = 0x7C6E,
  [10859] = 0x883B, [10860] = 0x89C0, [10861] = 0x8EA1, [10862] = 0x91C1,
  [10863] = 0x9472, [10864] = 0x9470, [10865] = 0x9871, [10866] = 0x995E,
  [10867] = 0x9AD6, [10868] = 0x9B23, [10869] = 0x9ECC, [10870] = 0x7064,
  [10871] = 0x77DA, [10872] = 0x8B9A, [10873] = 0x9477, [10874] = 0x97C9,
  [10875] = 0x9A62, [10876] = 0x9A65, [10877] = 0x7E9C, [10878] = 0x8B9C,
  [10879] = 0x8EAA, [10880] = 0x91C5, [10881] = 0x947D, [10882] = 0x947E,
  [10883] = 0x947C, [10884] = 0x9C77, [10885] = 0x9C78, [10886] = 0x9EF7,
  [10887] = 0x8C54, [10888] = 0x947F, [10889] = 0x9E1A, [10890] = 0x7228,
  [10891] = 0x9A6A, [10892] = 0x9B31, [10893] = 0x9E1B, [10894] = 0x9E1E,
  [10895] = 0x7C72, [10896] = 0x2460, [10897] = 0x2461, [10898] = 0x2462,
  [10899] = 0x2463, [10900] = 0x2464, [10901] = 0x2465, [10902] = 0x2466,
  [10903] = 0x2467, [10904] = 0x2468, [10905] = 0x2469, [10906] = 0x2474,
  [10907] = 0x2475, [10908] = 0x2476, [10909] = 0x2477, [10910] = 0x2478,
  [10911] = 0x2479, [10912] = 0x247A, [10913] = 0x247B, [10914] = 0x247C,
  [10915] = 0x247D, [10916] = 0x2170, [10917] = 0x2171, [10918] = 0x2172,
  [10919] = 0x2173, [10920] = 0x2174, [10921] = 0x2175, [10922] = 0x2176,
  [10923] = 0x2177, [10924] = 0x2178, [10925] = 0x2179, [10926] = 0x4E36,
  [10927] = 0x4E3F, [10928] = 0x4E85, [10929] = 0x4EA0, [10930] = 0x5182,
  [10931] = 0x5196, [10932] = 0x51AB, [10933] = 0x52F9, [10934] = 0x5338,
  [10935] = 0x5369, [10936] = 0x53B6, [10937] = 0x590A, [10938] = 0x5B80,
  [10939] = 0x5DDB, [10941] = 0x5E7F, [10943] = 0x5F50, [10944] = 0x5F61,
  [10945] = 0x6534, [10947] = 0x7592, [10949] = 0x8FB5, [10951] = 0x00A8,
  [10952] = 0x02C6, [10953] = 0x30FD, [10954] = 0x30FE, [10955] = 0x309D,
  [10956] = 0x309E, [10959] = 0x3005, [10960] = 0x3006, [10961] = 0x3007,
  [10962] = 0x30FC, [10963] = 0xFF3B, [10964] = 0xFF3D, [10965] = 0x273D,
  [10966] = 0x3041, [10967] = 0x3042, [10968] = 0x3043, [10969] = 0x3044,
  [10970] = 0x3045, [10971] = 0x3046, [10972] = 0x3047, [10973] = 0x3048,
  [10974] = 0x3049, [10975] = 0x304A, [10976] = 0x304B, [10977] = 0x304C,
  [10978] = 0x304D, [10979] = 0x304E, [10980] = 0x304F, [10981] = 0x3050,
  [10982] = 0x3051, [10983] = 0x3052, [10984] = 0x3053, [10985] = 0x3054,
  [10986] = 0x3055, [10987] = 0x3056, [10988] = 0x3057, [10989] = 0x3058,
  [10990] = 0x3059, [10991] = 0x305A, [10992] = 0x305B, [10993] = 0x305C,
  [10994] = 0x305D, [10995] = 0x305E, [10996] = 0x305F, [10997] = 0x3060,
  [10998] = 0x3061, [10999] = 0x3062, [11000] = 0x3063, [11001] = 0x3064,
  [11002] = 0x3065, [11003] = 0x3066, [11004] = 0x3067, [11005] = 0x3068,
  [11006] = 0x3069, [11007] = 0x306A, [11008] = 0x306B, [11009] = 0x306C,
  [11010] = 0x306D, [11011] = 0x306E, [11012] = 0x306F, [11013] = 0x3070,
  [11014] = 0x3071, [11015] = 0x3072, [11016] = 0x3073, [11017] = 0x3074,
  [11018] = 0x3075, [11019] = 0x3076, [11020] = 0x3077, [11021] = 0x3078,
  [11022] = 0x3079, [11023] = 0x307A, [11024] = 0x307B, [11025] = 0x307C,
  [11026] = 0x307D, [11027] = 0x307E, [11028] = 0x307F, [11029] = 0x3080,
  [11030] = 0x3081, [11031] = 0x3082, [11032] = 0x3083, [11033] = 0x3084,
  [11034] = 0x3085, [11035] = 0x3086, [11036] = 0x3087, [11037] = 0x3088,
  [11038] = 0x3089, [11039] = 0x308A, [11040] = 0x308B, [11041] = 0x308C,
  [11042] = 0x308D, [11043] = 0x308E, [11044] = 0x308F, [11045] = 0x3090,
  [11046] = 0x3091, [11047] = 0x3092, [11048] = 0x3093, [11049] = 0x30A1,
  [11050] = 0x30A2, [11051] = 0x30A3, [11052] = 0x30A4, [11053] = 0x30A5,
  [11054] = 0x30A6, [11055] = 0x30A7, [11056] = 0x30A8, [11057] = 0x30A9,
  [11058] = 0x30AA, [11059] = 0x30AB, [11060] = 0x30AC, [11061] = 0x30AD,
  [11062] = 0x30AE, [11063] = 0x30AF, [11064] = 0x30B0, [11065] = 0x30B1,
  [11066] = 0x30B2, [11067] = 0x30B3, [11068] = 0x30B4, [11069] = 0x30B5,
  [11070] = 0x30B6, [11071] = 0x30B7, [11072] = 0x30B8, [11073] = 0x30B9,
  [11074] = 0x30BA, [11075] = 0x30BB, [11076] = 0x30BC, [11077] = 0x30BD,
  [11078] = 0x30BE, [11079] = 0x30BF, [11080] = 0x30C0, [11081] = 0x30C1,
  [11082] = 0x30C2, [11083] = 0x30C3, [11084] = 0x30C4, [11085] = 0x30C5,
  [11086] = 0x30C6, [11087] = 0x30C7, [11088] = 0x30C8, [11089] = 0x30C9,
  [11090] = 0x30CA, [11091] = 0x30CB, [11092] = 0x30CC, [11093] = 0x30CD,
  [11094] = 0x30CE, [11095] = 0x30CF, [11096] = 0x30D0, [11097] = 0x30D1,
  [11098] = 0x30D2, [11099] = 0x30D3, [11100] = 0x30D4, [11101] = 0x30D5,
  [11102] = 0x30D6, [11103] = 0x30D7, [11104] = 0x30D8, [11105] = 0x30D9,
  [11106] = 0x30DA, [11107] = 0x30DB, [11108] = 0x30DC, [11109] = 0x30DD,
  [11110] = 0x30DE, [11111] = 0x30DF, [11112] = 0x30E0, [11113] = 0x30E1,
  [11114] = 0x30E2, [11115] = 0x30E3, [11116] = 0x30E4, [11117] = 0x30E5,
  [11118] = 0x30E6, [11119] = 0x30E7, [11120] = 0x30E8, [11121] = 0x30E9,
  [11122] = 0x30EA, [11123] = 0x30EB, [11124] = 0x30EC, [11125] = 0x30ED,
  [11126] = 0x30EE, [11127] = 0x30EF, [11128] = 0x30F0, [11129] = 0x30F1,
  [11130] = 0x30F2, [11131] = 0x30F3, [11132] = 0x30F4, [11133] = 0x30F5,
  [11134] = 0x30F6, [11135] = 0x0410, [11136] = 0x0411, [11137] = 0x0412,
  [11138] = 0x0413, [11139] = 0x0414, [11140] = 0x0415, [11141] = 0x0401,
  [11142] = 0x0416, [11143] = 0x0417, [11144] = 0x0418, [11145] = 0x0419,
  [11146] = 0x041A, [11147] = 0x041B, [11148] = 0x041C, [11149] = 0x041D,
  [11150] = 0x041E, [11151] = 0x041F, [11152] = 0x0420, [11153] = 0x0421,
  [11154] = 0x0422, [11155] = 0x0423, [11156] = 0x0424, [11157] = 0x0425,
  [11158] = 0x0426, [11159] = 0x0427, [11160] = 0x0428, [11161] = 0x0429,
  [11162] = 0x042A, [11163] = 0x042B, [11164] = 0x042C, [11165] = 0x042D,
  [11166] = 0x042E, [11167] = 0x042F, [11168] = 0x0430, [11169] = 0x0431,
  [11170] = 0x0432, [11171] = 0x0433, [11172] = 0x0434, [11173] = 0x0435,
  [11174] = 0x0451, [11175] = 0x0436, [11176] = 0x0437, [11177] = 0x0438,
  [11178] = 0x0439, [11179] = 0x043A, [11180] = 0x043B, [11181] = 0x043C,
  [11182] = 0x043D, [11183] = 0x043E, [11184] = 0x043F, [11185] = 0x0440,
  [11186] = 0x0441, [11187] = 0x0442, [11188] = 0x0443, [11189] = 0x0444,
  [11190] = 0x0445, [11191] = 0x0446, [11192] = 0x0447, [11193] = 0x0448,
  [11194] = 0x0449, [11195] = 0x044A, [11196] = 0x044B, [11197] = 0x044C,
  [11198] = 0x044D, [11199] = 0x044E, [11200] = 0x044F, [11201] = 0x21E7,
  [11202] = 0x21B8, [11203] = 0x21B9, [11206] = 0x4E5A, [11208] = 0x5202,
  [11211] = 0x5188, [11254] = 0xFFE2, [11255] = 0xFFE4, [11256] = 0xFF07,
  [11257] = 0xFF02, [11258] = 0x3231, [11259] = 0x2116, [11260] = 0x2121,
  [11261] = 0x309B, [11262] = 0x309C, [11294] = 0x0283, [11295] = 0x0250,
  [11296] = 0x025B, [11297] = 0x0254, [11298] = 0x0275, [11299] = 0x0153,
  [11300] = 0x00F8, [11301] = 0x014B, [11302] = 0x028A, [11303] = 0x026A,
  [11304] = 0x4E42, [11305] = 0x4E5C, [11306] = 0x51F5, [11307] = 0x531A,
  [11308] = 0x5382, [11309] = 0x4E07, [11310] = 0x4E0C, [11311] = 0x4E47,
  [11312] = 0x4E8D, [11313] = 0x56D7, [11314] = 0xFA0C, [11315] = 0x5C6E,
  [11316] = 0x5F73, [11317] = 0x4E0F, [11318] = 0x5187, [11319] = 0x4E0E,
  [11320] = 0x4E2E, [11321] = 0x4E93, [11322] = 0x4EC2, [11323] = 0x4EC9,
  [11324] = 0x4EC8, [11325] = 0x5198, [11326] = 0x52FC, [11327] = 0x536C,
  [11328] = 0x53B9, [11329] = 0x5720, [11330] = 0x5903, [11331] = 0x592C,
  [11332] = 0x5C10, [11333] = 0x5DFF, [11334] = 0x65E1, [11335] = 0x6BB3,
  [11336] = 0x6BCC, [11337] = 0x6C14, [11338] = 0x723F, [11339] = 0x4E31,
  [11340] = 0x4E3C, [11341] = 0x4EE8, [11342] = 0x4EDC, [11343] = 0x4EE9,
  [11344] = 0x4EE1, [11345] = 0x4EDD, [11346] = 0x4EDA, [11347] = 0x520C,
  [11348] = 0x531C, [11349] = 0x534C, [11350] = 0x5722, [11351] = 0x5723,
  [11352] = 0x5917, [11353] = 0x592F, [11354] = 0x5B81, [11355] = 0x5B84,
  [11356] = 0x5C12, [11357] = 0x5C3B, [11358] = 0x5C74, [11359] = 0x5C73,
  [11360] = 0x5E04, [11361] = 0x5E80, [11362] = 0x5E82, [11363] = 0x5FC9,
  [11364] = 0x6209, [11365] = 0x6250, [11366] = 0x6C15, [11367] = 0x6C36,
  [11368] = 0x6C43, [11369] = 0x6C3F, [11370] = 0x6C3B, [11371] = 0x72AE,
  [11372] = 0x72B0, [11373] = 0x738A, [11374] = 0x79B8, [11375] = 0x808A,
  [11376] = 0x961E, [11377] = 0x4F0E, [11378] = 0x4F18, [11379] = 0x4F2C,
  [11380] = 0x4EF5, [11381] = 0x4F14, [11382] = 0x4EF1, [11383] = 0x4F00,
  [11384] = 0x4EF7, [11385] = 0x4F08, [11386] = 0x4F1D, [11387] = 0x4F02,
  [11388] = 0x4F05, [11389] = 0x4F22, [11390] = 0x4F13, [11391] = 0x4F04,
  [11392] = 0x4EF4, [11393] = 0x4F12, [11394] = 0x51B1, [11395] = 0x5213,
  [11396] = 0x5209, [11397] = 0x5210, [11398] = 0x52A6, [11399] = 0x5322,
  [11400] = 0x531F, [11401] = 0x534D, [11402] = 0x538A, [11403] = 0x5407,
  [11404] = 0x56E1, [11405] = 0x56DF, [11406] = 0x572E, [11407] = 0x572A,
  [11408] = 0x5734, [11409] = 0x593C, [11410] = 0x5980, [11411] = 0x597C,
  [11412] = 0x5985, [11413] = 0x597B, [11414] = 0x597E, [11415] = 0x5977,
  [11416] = 0x597F, [11417] = 0x5B56, [11418] = 0x5C15, [11419] = 0x5C25,
  [11420] = 0x5C7C, [11421] = 0x5C7A, [11422] = 0x5C7B, [11423] = 0x5C7E,
  [11424] = 0x5DDF, [11425] = 0x5E75, [11426] = 0x5E84, [11427] = 0x5F02,
  [11428] = 0x5F1A, [11429] = 0x5F74, [11430] = 0x5FD5, [11431] = 0x5FD4,
  [11432] = 0x5FCF, [11433] = 0x625C, [11434] = 0x625E, [11435] = 0x6264,
  [11436] = 0x6261, [11437] = 0x6266, [11438] = 0x6262, [11439] = 0x6259,
  [11440] = 0x6260, [11441] = 0x625A, [11442] = 0x6265, [11443] = 0x65EF,
  [11444] = 0x65EE, [11445] = 0x673E, [11446] = 0x6739, [11447] = 0x6738,
  [11448] = 0x673B, [11449] = 0x673A, [11450] = 0x673F, [11451] = 0x673C,
  [11452] = 0x6733, [11453] = 0x6C18, [11454] = 0x6C46, [11455] = 0x6C52,
  [11456] = 0x6C5C, [11457] = 0x6C4F, [11458] = 0x6C4A, [11459] = 0x6C54,
  [11460] = 0x6C4B, [11461] = 0x6C4C, [11462] = 0x7071, [11463] = 0x725E,
  [11464] = 0x72B4, [11465] = 0x72B5, [11466] = 0x738E, [11467] = 0x752A,
  [11468] = 0x767F, [11469] = 0x7A75, [11470] = 0x7F51, [11471] = 0x8278,
  [11472] = 0x827C, [11473] = 0x8280, [11474] = 0x827D, [11475] = 0x827F,
  [11476] = 0x864D, [11477] = 0x897E, [11478] = 0x9099, [11479] = 0x9097,
  [11480] = 0x9098, [11481] = 0x909B, [11482] = 0x9094, [11483] = 0x9622,
  [11484] = 0x9624, [11485] = 0x9620, [11486] = 0x9623, [11487] = 0x4F56,
  [11488] = 0x4F3B, [11489] = 0x4F62, [11490] = 0x4F49, [11491] = 0x4F53,
  [11492] = 0x4F64, [11493] = 0x4F3E, [11494] = 0x4F67, [11495] = 0x4F52,
  [11496] = 0x4F5F, [11497] = 0x4F41, [11498] = 0x4F58, [11499] = 0x4F2D,
  [11500] = 0x4F33, [11501] = 0x4F3F, [11502] = 0x4F61, [11503] = 0x518F,
  [11504] = 0x51B9, [11505] = 0x521C, [11506] = 0x521E, [11507] = 0x5221,
  [11508] = 0x52AD, [11509] = 0x52AE, [11510] = 0x5309, [11511] = 0x5363,
  [11512] = 0x5372, [11513] = 0x538E, [11514] = 0x538F, [11515] = 0x5430,
  [11516] = 0x5437, [11517] = 0x542A, [11518] = 0x5454, [11519] = 0x5445,
  [11520] = 0x5419, [11521] = 0x541C, [11522] = 0x5425, [11523] = 0x5418,
  [11524] = 0x543D, [11525] = 0x544F, [11526] = 0x5441, [11527] = 0x5428,
  [11528] = 0x5424, [11529] = 0x5447, [11530] = 0x56EE, [11531] = 0x56E7,
  [11532] = 0x56E5, [11533] = 0x5741, [11534] = 0x5745, [11535] = 0x574C,
  [11536] = 0x5749, [11537] = 0x574B, [11538] = 0x5752, [11539] = 0x5906,
  [11540] = 0x5940, [11541] = 0x59A6, [11542] = 0x5998, [11543] = 0x59A0,
  [11544] = 0x5997, [11545] = 0x598E, [11546] = 0x59A2, [11547] = 0x5990,
  [11548] = 0x598F, [11549] = 0x59A7, [11550] = 0x59A1, [11551] = 0x5B8E,
  [11552] = 0x5B92, [11553] = 0x5C28, [11554] = 0x5C2A, [11555] = 0x5C8D,
  [11556] = 0x5C8F, [11557] = 0x5C88, [11558] = 0x5C8B, [11559] = 0x5C89,
  [11560] = 0x5C92, [11561] = 0x5C8A, [11562] = 0x5C86, [11563] = 0x5C93,
  [11564] = 0x5C95, [11565] = 0x5DE0, [11566] = 0x5E0A, [11567] = 0x5E0E,
  [11568] = 0x5E8B, [11569] = 0x5E89, [11570] = 0x5E8C, [11571] = 0x5E88,
  [11572] = 0x5E8D, [11573] = 0x5F05, [11574] = 0x5F1D, [11575] = 0x5F78,
  [11576] = 0x5F76, [11577] = 0x5FD2, [11578] = 0x5FD1, [11579] = 0x5FD0,
  [11580] = 0x5FED, [11581] = 0x5FE8, [11582] = 0x5FEE, [11583] = 0x5FF3,
  [11584] = 0x5FE1, [11585] = 0x5FE4, [11586] = 0x5FE3, [11587] = 0x5FFA,
  [11588] = 0x5FEF, [11589] = 0x5FF7, [11590] = 0x5FFB, [11591] = 0x6000,
  [11592] = 0x5FF4, [11593] = 0x623A, [11594] = 0x6283, [11595] = 0x628C,
  [11596] = 0x628E, [11597] = 0x628F, [11598] = 0x6294, [11599] = 0x6287,
  [11600] = 0x6271, [11601] = 0x627B, [11602] = 0x627A, [11603] = 0x6270,
  [11604] = 0x6281, [11605] = 0x6288, [11606] = 0x6277, [11607] = 0x627D,
  [11608] = 0x6272, [11609] = 0x6274, [11610] = 0x6537, [11611] = 0x65F0,
  [11612] = 0x65F4, [11613] = 0x65F3, [11614] = 0x65F2, [11615] = 0x65F5,
  [11616] = 0x6745, [11617] = 0x6747, [11618] = 0x6759, [11619] = 0x6755,
  [11620] = 0x674C, [11621] = 0x6748, [11622] = 0x675D, [11623] = 0x674D,
  [11624] = 0x675A, [11625] = 0x674B, [11626] = 0x6BD0, [11627] = 0x6C19,
  [11628] = 0x6C1A, [11629] = 0x6C78, [11630] = 0x6C67, [11631] = 0x6C6B,
  [11632] = 0x6C84, [11633] = 0x6C8B, [11634] = 0x6C8F, [11635] = 0x6C71,
  [11636] = 0x6C6F, [11637] = 0x6C69, [11638] = 0x6C9A, [11639] = 0x6C6D,
  [11640] = 0x6C87, [11641] = 0x6C95, [11642] = 0x6C9C, [11643] = 0x6C66,
  [11644] = 0x6C73, [11645] = 0x6C65, [11646] = 0x6C7B, [11647] = 0x6C8E,
  [11648] = 0x7074, [11649] = 0x707A, [11650] = 0x7263, [11651] = 0x72BF,
  [11652] = 0x72BD, [11653] = 0x72C3, [11654] = 0x72C6, [11655] = 0x72C1,
  [11656] = 0x72BA, [11657] = 0x72C5, [11658] = 0x7395, [11659] = 0x7397,
  [11660] = 0x7393, [11661] = 0x7394, [11662] = 0x7392, [11663] = 0x753A,
  [11664] = 0x7539, [11665] = 0x7594, [11666] = 0x7595, [11667] = 0x7681,
  [11668] = 0x793D, [11669] = 0x8034, [11670] = 0x8095, [11671] = 0x8099,
  [11672] = 0x8090, [11673] = 0x8092, [11674] = 0x809C, [11675] = 0x8290,
  [11676] = 0x828F, [11677] = 0x8285, [11678] = 0x828E, [11679] = 0x8291,
  [11680] = 0x8293, [11681] = 0x828A, [11682] = 0x8283, [11683] = 0x8284,
  [11684] = 0x8C78, [11685] = 0x8FC9, [11686] = 0x8FBF, [11687] = 0x909F,
  [11688] = 0x90A1, [11689] = 0x90A5, [11690] = 0x909E, [11691] = 0x90A7,
  [11692] = 0x90A0, [11693] = 0x9630, [11694] = 0x9628, [11695] = 0x962F,
  [11696] = 0x962D, [11697] = 0x4E33, [11698] = 0x4F98, [11699] = 0x4F7C,
  [11700] = 0x4F85, [11701] = 0x4F7D, [11702] = 0x4F80, [11703] = 0x4F87,
  [11704] = 0x4F76, [11705] = 0x4F74, [11706] = 0x4F89, [11707] = 0x4F84,
  [11708] = 0x4F77, [11709] = 0x4F4C, [11710] = 0x4F97, [11711] = 0x4F6A,
  [11712] = 0x4F9A, [11713] = 0x4F79, [11714] = 0x4F81, [11715] = 0x4F78,
  [11716] = 0x4F90, [11717] = 0x4F9C, [11718] = 0x4F94, [11719] = 0x4F9E,
  [11720] = 0x4F92, [11721] = 0x4F82, [11722] = 0x4F95, [11723] = 0x4F6B,
  [11724] = 0x4F6E, [11725] = 0x519E, [11726] = 0x51BC, [11727] = 0x51BE,
  [11728] = 0x5235, [11729] = 0x5232, [11730] = 0x5233, [11731] = 0x5246,
  [11732] = 0x5231, [11733] = 0x52BC, [11734] = 0x530A, [11735] = 0x530B,
  [11736] = 0x533C, [11737] = 0x5392, [11738] = 0x5394, [11739] = 0x5487,
  [11740] = 0x547F, [11741] = 0x5481, [11742] = 0x5491, [11743] = 0x5482,
  [11744] = 0x5488, [11745] = 0x546B, [11746] = 0x547A, [11747] = 0x547E,
  [11748] = 0x5465, [11749] = 0x546C, [11750] = 0x5474, [11751] = 0x5466,
  [11752] = 0x548D, [11753] = 0x546F, [11754] = 0x5461, [11755] = 0x5460,
  [11756] = 0x5498, [11757] = 0x5463, [11758] = 0x5467, [11759] = 0x5464,
  [11760] = 0x56F7, [11761] = 0x56F9, [11762] = 0x576F, [11763] = 0x5772,
  [11764] = 0x576D, [11765] = 0x576B, [11766] = 0x5771, [11767] = 0x5770,
  [11768] = 0x5776, [11769] = 0x5780, [11770] = 0x5775, [11771] = 0x577B,
  [11772] = 0x5773, [11773] = 0x5774, [11774] = 0x5762, [11775] = 0x5768,
  [11776] = 0x577D, [11777] = 0x590C, [11778] = 0x5945, [11779] = 0x59B5,
  [11780] = 0x59BA, [11781] = 0x59CF, [11782] = 0x59CE, [11783] = 0x59B2,
  [11784] = 0x59CC, [11785] = 0x59C1, [11786] = 0x59B6, [11787] = 0x59BC,
  [11788] = 0x59C3, [11789] = 0x59D6, [11790] = 0x59B1, [11791] = 0x59BD,
  [11792] = 0x59C0, [11793] = 0x59C8, [11794] = 0x59B4, [11795] = 0x59C7,
  [11796] = 0x5B62, [11797] = 0x5B65, [11798] = 0x5B93, [11799] = 0x5B95,
  [11800] = 0x5C44, [11801] = 0x5C47, [11802] = 0x5CAE, [11803] = 0x5CA4,
  [11804] = 0x5CA0, [11805] = 0x5CB5, [11806] = 0x5CAF, [11807] = 0x5CA8,
  [11808] = 0x5CAC, [11809] = 0x5C9F, [11810] = 0x5CA3, [11811] = 0x5CAD,
  [11812] = 0x5CA2, [11813] = 0x5CAA, [11814] = 0x5CA7, [11815] = 0x5C9D,
  [11816] = 0x5CA5, [11817] = 0x5CB6, [11818] = 0x5CB0, [11819] = 0x5CA6,
  [11820] = 0x5E17, [11821] = 0x5E14, [11822] = 0x5E19, [11823] = 0x5F28,
  [11824] = 0x5F22, [11825] = 0x5F23, [11826] = 0x5F24, [11827] = 0x5F54,
  [11828] = 0x5F82, [11829] = 0x5F7E, [11830] = 0x5F7D, [11831] = 0x5FDE,
  [11832] = 0x5FE5, [11833] = 0x602D, [11834] = 0x6026, [11835] = 0x6019,
  [11836] = 0x6032, [11837] = 0x600B, [11838] = 0x6034, [11839] = 0x600A,
  [11840] = 0x6017, [11841] = 0x6033, [11842] = 0x601A, [11843] = 0x601E,
  [11844] = 0x602C, [11845] = 0x6022, [11846] = 0x600D, [11847] = 0x6010,
  [11848] = 0x602E, [11849] = 0x6013, [11850] = 0x6011, [11851] = 0x600C,
  [11852] = 0x6009, [11853] = 0x601C, [11854] = 0x6214, [11855] = 0x623D,
  [11856] = 0x62AD, [11857] = 0x62B4, [11858] = 0x62D1, [11859] = 0x62BE,
  [11860] = 0x62AA, [11861] = 0x62B6, [11862] = 0x62CA, [11863] = 0x62AE,
  [11864] = 0x62B3, [11865] = 0x62AF, [11866] = 0x62BB, [11867] = 0x62A9,
  [11868] = 0x62B0, [11869] = 0x62B8, [11870] = 0x653D, [11871] = 0x65A8,
  [11872] = 0x65BB, [11873] = 0x6609, [11874] = 0x65FC, [11875] = 0x6604,
  [11876] = 0x6612, [11877] = 0x6608, [11878] = 0x65FB, [11879] = 0x6603,
  [11880] = 0x660B, [11881] = 0x660D, [11882] = 0x6605, [11883] = 0x65FD,
  [11884] = 0x6611, [11885] = 0x6610, [11886] = 0x66F6, [11887] = 0x670A,
  [11888] = 0x6785, [11889] = 0x676C, [11890] = 0x678E, [11891] = 0x6792,
  [11892] = 0x6776, [11893] = 0x677B, [11894] = 0x6798, [11895] = 0x6786,
  [11896] = 0x6784, [11897] = 0x6774, [11898] = 0x678D, [11899] = 0x678C,
  [11900] = 0x677A, [11901] = 0x679F, [11902] = 0x6791, [11903] = 0x6799,
  [11904] = 0x6783, [11905] = 0x677D, [11906] = 0x6781, [11907] = 0x6778,
  [11908] = 0x6779, [11909] = 0x6794, [11910] = 0x6B25, [11911] = 0x6B80,
  [11912] = 0x6B7E, [11913] = 0x6BDE, [11914] = 0x6C1D, [11915] = 0x6C93,
  [11916] = 0x6CEC, [11917] = 0x6CEB, [11918] = 0x6CEE, [11919] = 0x6CD9,
  [11920] = 0x6CB6, [11921] = 0x6CD4, [11922] = 0x6CAD, [11923] = 0x6CE7,
  [11924] = 0x6CB7, [11925] = 0x6CD0, [11926] = 0x6CC2, [11927] = 0x6CBA,
  [11928] = 0x6CC3, [11929] = 0x6CC6, [11930] = 0x6CED, [11931] = 0x6CF2,
  [11932] = 0x6CD2, [11933] = 0x6CDD, [11934] = 0x6CB4, [11935] = 0x6C8A,
  [11936] = 0x6C9D, [11937] = 0x6C80, [11938] = 0x6CDE, [11939] = 0x6CC0,
  [11940] = 0x6D30, [11941] = 0x6CCD, [11942] = 0x6CC7, [11943] = 0x6CB0,
  [11944] = 0x6CF9, [11945] = 0x6CCF, [11946] = 0x6CE9, [11947] = 0x6CD1,
  [11948] = 0x7094, [11949] = 0x7098, [11950] = 0x7085, [11951] = 0x7093,
  [11952] = 0x7086, [11953] = 0x7084, [11954] = 0x7091, [11955] = 0x7096,
  [11956] = 0x7082, [11957] = 0x709A, [11958] = 0x7083, [11959] = 0x726A,
  [11960] = 0x72D6, [11961] = 0x72CB, [11962] = 0x72D8, [11963] = 0x72C9,
  [11964] = 0x72DC, [11965] = 0x72D2, [11966] = 0x72D4, [11967] = 0x72DA,
  [11968] = 0x72CC, [11969] = 0x72D1, [11970] = 0x73A4, [11971] = 0x73A1,
  [11972] = 0x73AD, [11973] = 0x73A6, [11974] = 0x73A2, [11975] = 0x73A0,
  [11976] = 0x73AC, [11977] = 0x739D, [11978] = 0x74DD, [11979] = 0x74E8,
  [11980] = 0x753F, [11981] = 0x7540, [11982] = 0x753E, [11983] = 0x758C,
  [11984] = 0x7598, [11985] = 0x76AF, [11986] = 0x76F3, [11987] = 0x76F1,
  [11988] = 0x76F0, [11989] = 0x76F5, [11990] = 0x77F8, [11991] = 0x77FC,
  [11992] = 0x77F9, [11993] = 0x77FB, [11994] = 0x77FA, [11995] = 0x77F7,
  [11996] = 0x7942, [11997] = 0x793F, [11998] = 0x79C5, [11999] = 0x7A78,
  [12000] = 0x7A7B, [12001] = 0x7AFB, [12002] = 0x7C75, [12003] = 0x7CFD,
  [12004] = 0x8035, [12005] = 0x808F, [12006] = 0x80AE, [12007] = 0x80A3,
  [12008] = 0x80B8, [12009] = 0x80B5, [12010] = 0x80AD, [12011] = 0x8220,
  [12012] = 0x82A0, [12013] = 0x82C0, [12014] = 0x82AB, [12015] = 0x829A,
  [12016] = 0x8298, [12017] = 0x829B, [12018] = 0x82B5, [12019] = 0x82A7,
  [12020] = 0x82AE, [12021] = 0x82BC, [12022] = 0x829E, [12023] = 0x82BA,
  [12024] = 0x82B4, [12025] = 0x82A8, [12026] = 0x82A1, [12027] = 0x82A9,
  [12028] = 0x82C2, [12029] = 0x82A4, [12030] = 0x82C3, [12031] = 0x82B6,
  [12032] = 0x82A2, [12033] = 0x8670, [12034] = 0x866F, [12035] = 0x866D,
  [12036] = 0x866E, [12037] = 0x8C56, [12038] = 0x8FD2, [12039] = 0x8FCB,
  [12040] = 0x8FD3, [12041] = 0x8FCD, [12042] = 0x8FD6, [12043] = 0x8FD5,
  [12044] = 0x8FD7, [12045] = 0x90B2, [12046] = 0x90B4, [12047] = 0x90AF,
  [12048] = 0x90B3, [12049] = 0x90B0, [12050] = 0x9639, [12051] = 0x963D,
  [12052] = 0x963C, [12053] = 0x963A, [12054] = 0x9643, [12055] = 0x4FCD,
  [12056] = 0x4FC5, [12057] = 0x4FD3, [12058] = 0x4FB2, [12059] = 0x4FC9,
  [12060] = 0x4FCB, [12061] = 0x4FC1, [12062] = 0x4FD4, [12063] = 0x4FDC,
  [12064] = 0x4FD9, [12065] = 0x4FBB, [12066] = 0x4FB3, [12067] = 0x4FDB,
  [12068] = 0x4FC7, [12069] = 0x4FD6, [12070] = 0x4FBA, [12071] = 0x4FC0,
  [12072] = 0x4FB9, [12073] = 0x4FEC, [12074] = 0x5244, [12075] = 0x5249,
  [12076] = 0x52C0, [12077] = 0x52C2, [12078] = 0x533D, [12079] = 0x537C,
  [12080] = 0x5397, [12081] = 0x5396, [12082] = 0x5399, [12083] = 0x5398,
  [12084] = 0x54BA, [12085] = 0x54A1, [12086] = 0x54AD, [12087] = 0x54A5,
  [12088] = 0x54CF, [12089] = 0x54C3, [12090] = 0x830D, [12091] = 0x54B7,
  [12092] = 0x54AE, [12093] = 0x54D6, [12094] = 0x54B6, [12095] = 0x54C5,
  [12096] = 0x54C6, [12097] = 0x54A0, [12098] = 0x5470, [12099] = 0x54BC,
  [12100] = 0x54A2, [12101] = 0x54BE, [12102] = 0x5472, [12103] = 0x54DE,
  [12104] = 0x54B0, [12105] = 0x57B5, [12106] = 0x579E, [12107] = 0x579F,
  [12108] = 0x57A4, [12109] = 0x578C, [12110] = 0x5797, [12111] = 0x579D,
  [12112] = 0x579B, [12113] = 0x5794, [12114] = 0x5798, [12115] = 0x578F,
  [12116] = 0x5799, [12117] = 0x57A5, [12118] = 0x579A, [12119] = 0x5795,
  [12120] = 0x58F4, [12121] = 0x590D, [12122] = 0x5953, [12123] = 0x59E1,
  [12124] = 0x59DE, [12125] = 0x59EE, [12126] = 0x5A00, [12127] = 0x59F1,
  [12128] = 0x59DD, [12129] = 0x59FA, [12130] = 0x59FD, [12131] = 0x59FC,
  [12132] = 0x59F6, [12133] = 0x59E4, [12134] = 0x59F2, [12135] = 0x59F7,
  [12136] = 0x59DB, [12137] = 0x59E9, [12138] = 0x59F3, [12139] = 0x59F5,
  [12140] = 0x59E0, [12141] = 0x59FE, [12142] = 0x59F4, [12143] = 0x59ED,
  [12144] = 0x5BA8, [12145] = 0x5C4C, [12146] = 0x5CD0, [12147] = 0x5CD8,
  [12148] = 0x5CCC, [12149] = 0x5CD7, [12150] = 0x5CCB, [12151] = 0x5CDB,
  [12152] = 0x5CDE, [12153] = 0x5CDA, [12154] = 0x5CC9, [12155] = 0x5CC7,
  [12156] = 0x5CCA, [12157] = 0x5CD6, [12158] = 0x5CD3, [12159] = 0x5CD4,
  [12160] = 0x5CCF, [12161] = 0x5CC8, [12162] = 0x5CC6, [12163] = 0x5CCE,
  [12164] = 0x5CDF, [12165] = 0x5CF8, [12166] = 0x5DF9, [12167] = 0x5E21,
  [12168] = 0x5E22, [12169] = 0x5E23, [12170] = 0x5E20, [12171] = 0x5E24,
  [12172] = 0x5EB0, [12173] = 0x5EA4, [12174] = 0x5EA2, [12175] = 0x5E9B,
  [12176] = 0x5EA3, [12177] = 0x5EA5, [12178] = 0x5F07, [12179] = 0x5F2E,
  [12180] = 0x5F56, [12181] = 0x5F86, [12182] = 0x6037, [12183] = 0x6039,
  [12184] = 0x6054, [12185] = 0x6072, [12186] = 0x605E, [12187] = 0x6045,
  [12188] = 0x6053, [12189] = 0x6047, [12190] = 0x6049, [12191] = 0x605B,
  [12192] = 0x604C, [12193] = 0x6040, [12194] = 0x6042, [12195] = 0x605F,
  [12196] = 0x6024, [12197] = 0x6044, [12198] = 0x6058, [12199] = 0x6066,
  [12200] = 0x606E, [12201] = 0x6242, [12202] = 0x6243, [12203] = 0x62CF,
  [12204] = 0x630D, [12205] = 0x630B, [12206] = 0x62F5, [12207] = 0x630E,
  [12208] = 0x6303, [12209] = 0x62EB, [12210] = 0x62F9, [12211] = 0x630F,
  [12212] = 0x630C, [12213] = 0x62F8, [12214] = 0x62F6, [12215] = 0x6300,
  [12216] = 0x6313, [12217] = 0x6314, [12218] = 0x62FA, [12219] = 0x6315,
  [12220] = 0x62FB, [12221] = 0x62F0, [12222] = 0x6541, [12223] = 0x6543,
  [12224] = 0x65AA, [12225] = 0x65BF, [12226] = 0x6636, [12227] = 0x6621,
  [12228] = 0x6632, [12229] = 0x6635, [12230] = 0x661C, [12231] = 0x6626,
  [12232] = 0x6622, [12233] = 0x6633, [12234] = 0x662B, [12235] = 0x663A,
  [12236] = 0x661D, [12237] = 0x6634, [12238] = 0x6639, [12239] = 0x662E,
  [12240] = 0x670F, [12241] = 0x6710, [12242] = 0x67C1, [12243] = 0x67F2,
  [12244] = 0x67C8, [12245] = 0x67BA, [12246] = 0x67DC, [12247] = 0x67BB,
  [12248] = 0x67F8, [12249] = 0x67D8, [12250] = 0x67C0, [12251] = 0x67B7,
  [12252] = 0x67C5, [12253] = 0x67EB, [12254] = 0x67E4, [12255] = 0x67DF,
  [12256] = 0x67B5, [12257] = 0x67CD, [12258] = 0x67B3, [12259] = 0x67F7,
  [12260] = 0x67F6, [12261] = 0x67EE, [12262] = 0x67E3, [12263] = 0x67C2,
  [12264] = 0x67B9, [12265] = 0x67CE, [12266] = 0x67E7, [12267] = 0x67F0,
  [12268] = 0x67B2, [12269] = 0x67FC, [12270] = 0x67C6, [12271] = 0x67ED,
  [12272] = 0x67CC, [12273] = 0x67AE, [12274] = 0x67E6, [12275] = 0x67DB,
  [12276] = 0x67FA, [12277] = 0x67C9, [12278] = 0x67CA, [12279] = 0x67C3,
  [12280] = 0x67EA, [12281] = 0x67CB, [12282] = 0x6B28, [12283] = 0x6B82,
  [12284] = 0x6B84, [12285] = 0x6BB6, [12286] = 0x6BD6, [12287] = 0x6BD8,
  [12288] = 0x6BE0, [12289] = 0x6C20, [12290] = 0x6C21, [12291] = 0x6D28,
  [12292] = 0x6D34, [12293] = 0x6D2D, [12294] = 0x6D1F, [12295] = 0x6D3C,
  [12296] = 0x6D3F, [12297] = 0x6D12, [12298] = 0x6D0A, [12299] = 0x6CDA,
  [12300] = 0x6D33, [12301] = 0x6D04, [12302] = 0x6D19, [12303] = 0x6D3A,
  [12304] = 0x6D1A, [12305] = 0x6D11, [12306] = 0x6D00, [12307] = 0x6D1D,
  [12308] = 0x6D42, [12309] = 0x6D01, [12310] = 0x6D18, [12311] = 0x6D37,
  [12312] = 0x6D03, [12313] = 0x6D0F, [12314] = 0x6D40, [12315] = 0x6D07,
  [12316] = 0x6D20, [12317] = 0x6D2C, [12318] = 0x6D08, [12319] = 0x6D22,
  [12320] = 0x6D09, [12321] = 0x6D10, [12322] = 0x70B7, [12323] = 0x709F,
  [12324] = 0x70BE, [12325] = 0x70B1, [12326] = 0x70B0, [12327] = 0x70A1,
  [12328] = 0x70B4, [12329] = 0x70B5, [12330] = 0x70A9, [12331] = 0x7241,
  [12332] = 0x7249, [12333] = 0x724A, [12334] = 0x726C, [12335] = 0x7270,
  [12336] = 0x7273, [12337] = 0x726E, [12338] = 0x72CA, [12339] = 0x72E4,
  [12340] = 0x72E8, [12341] = 0x72EB, [12342] = 0x72DF, [12343] = 0x72EA,
  [12344] = 0x72E6, [12345] = 0x72E3, [12346] = 0x7385, [12347] = 0x73CC,
  [12348] = 0x73C2, [12349] = 0x73C8, [12350] = 0x73C5, [12351] = 0x73B9,
  [12352] = 0x73B6, [12353] = 0x73B5, [12354] = 0x73B4, [12355] = 0x73EB,
  [12356] = 0x73BF, [12357] = 0x73C7, [12358] = 0x73BE, [12359] = 0x73C3,
  [12360] = 0x73C6, [12361] = 0x73B8, [12362] = 0x73CB, [12363] = 0x74EC,
  [12364] = 0x74EE, [12365] = 0x752E, [12366] = 0x7547, [12367] = 0x7548,
  [12368] = 0x75A7, [12369] = 0x75AA, [12370] = 0x7679, [12371] = 0x76C4,
  [12372] = 0x7708, [12373] = 0x7703, [12374] = 0x7704, [12375] = 0x7705,
  [12376] = 0x770A, [12377] = 0x76F7, [12378] = 0x76FB, [12379] = 0x76FA,
  [12380] = 0x77E7, [12381] = 0x77E8, [12382] = 0x7806, [12383] = 0x7811,
  [12384] = 0x7812, [12385] = 0x7805, [12386] = 0x7810, [12387] = 0x780F,
  [12388] = 0x780E, [12389] = 0x7809, [12390] = 0x7803, [12391] = 0x7813,
  [12392] = 0x794A, [12393] = 0x794C, [12394] = 0x794B, [12395] = 0x7945,
  [12396] = 0x7944, [12397] = 0x79D5, [12398] = 0x79CD, [12399] = 0x79CF,
  [12400] = 0x79D6, [12401] = 0x79CE, [12402] = 0x7A80, [12403] = 0x7A7E,
  [12404] = 0x7AD1, [12405] = 0x7B00, [12406] = 0x7B01, [12407] = 0x7C7A,
  [12408] = 0x7C78, [12409] = 0x7C79, [12410] = 0x7C7F, [12411] = 0x7C80,
  [12412] = 0x7C81, [12413] = 0x7D03, [12414] = 0x7D08, [12415] = 0x7D01,
  [12416] = 0x7F58, [12417] = 0x7F91, [12418] = 0x7F8D, [12419] = 0x7FBE,
  [12420] = 0x8007, [12421] = 0x800E, [12422] = 0x800F, [12423] = 0x8014,
  [12424] = 0x8037, [12425] = 0x80D8, [12426] = 0x80C7, [12427] = 0x80E0,
  [12428] = 0x80D1, [12429] = 0x80C8, [12430] = 0x80C2, [12431] = 0x80D0,
  [12432] = 0x80C5, [12433] = 0x80E3, [12434] = 0x80D9, [12435] = 0x80DC,
  [12436] = 0x80CA, [12437] = 0x80D5, [12438] = 0x80C9, [12439] = 0x80CF,
  [12440] = 0x80D7, [12441] = 0x80E6, [12442] = 0x80CD, [12443] = 0x81FF,
  [12444] = 0x8221, [12445] = 0x8294, [12446] = 0x82D9, [12447] = 0x82FE,
  [12448] = 0x82F9, [12449] = 0x8307, [12450] = 0x82E8, [12451] = 0x8300,
  [12452] = 0x82D5, [12453] = 0x833A, [12454] = 0x82EB, [12455] = 0x82D6,
  [12456] = 0x82F4, [12457] = 0x82EC, [12458] = 0x82E1, [12459] = 0x82F2,
  [12460] = 0x82F5, [12461] = 0x830C, [12462] = 0x82FB, [12463] = 0x82F6,
  [12464] = 0x82F0, [12465] = 0x82EA, [12466] = 0x82E4, [12467] = 0x82E0,
  [12468] = 0x82FA, [12469] = 0x82F3, [12470] = 0x82ED, [12471] = 0x8677,
  [12472] = 0x8674, [12473] = 0x867C, [12474] = 0x8673, [12475] = 0x8841,
  [12476] = 0x884E, [12477] = 0x8867, [12478] = 0x886A, [12479] = 0x8869,
  [12480] = 0x89D3, [12481] = 0x8A04, [12482] = 0x8A07, [12483] = 0x8D72,
  [12484] = 0x8FE3, [12485] = 0x8FE1, [12486] = 0x8FEE, [12487] = 0x8FE0,
  [12488] = 0x90F1, [12489] = 0x90BD, [12490] = 0x90BF, [12491] = 0x90D5,
  [12492] = 0x90C5, [12493] = 0x90BE, [12494] = 0x90C7, [12495] = 0x90CB,
  [12496] = 0x90C8, [12497] = 0x91D4, [12498] = 0x91D3, [12499] = 0x9654,
  [12500] = 0x964F, [12501] = 0x9651, [12502] = 0x9653, [12503] = 0x964A,
  [12504] = 0x964E, [12505] = 0x501E, [12506] = 0x5005, [12507] = 0x5007,
  [12508] = 0x5013, [12509] = 0x5022, [12510] = 0x5030, [12511] = 0x501B,
  [12512] = 0x4FF5, [12513] = 0x4FF4, [12514] = 0x5033, [12515] = 0x5037,
  [12516] = 0x502C, [12517] = 0x4FF6, [12518] = 0x4FF7, [12519] = 0x5017,
  [12520] = 0x501C, [12521] = 0x5020, [12522] = 0x5027, [12523] = 0x5035,
  [12524] = 0x502F, [12525] = 0x5031, [12526] = 0x500E, [12527] = 0x515A,
  [12528] = 0x5194, [12529] = 0x5193, [12530] = 0x51CA, [12531] = 0x51C4,
  [12532] = 0x51C5, [12533] = 0x51C8, [12534] = 0x51CE, [12535] = 0x5261,
  [12536] = 0x525A, [12537] = 0x5252, [12538] = 0x525E, [12539] = 0x525F,
  [12540] = 0x5255, [12541] = 0x5262, [12542] = 0x52CD, [12543] = 0x530E,
  [12544] = 0x539E, [12545] = 0x5526, [12546] = 0x54E2, [12547] = 0x5517,
  [12548] = 0x5512, [12549] = 0x54E7, [12550] = 0x54F3, [12551] = 0x54E4,
  [12552] = 0x551A, [12553] = 0x54FF, [12554] = 0x5504, [12555] = 0x5508,
  [12556] = 0x54EB, [12557] = 0x5511, [12558] = 0x5505, [12559] = 0x54F1,
  [12560] = 0x550A, [12561] = 0x54FB, [12562] = 0x54F7, [12563] = 0x54F8,
  [12564] = 0x54E0, [12565] = 0x550E, [12566] = 0x5503, [12567] = 0x550B,
  [12568] = 0x5701, [12569] = 0x5702, [12570] = 0x57CC, [12571] = 0x5832,
  [12572] = 0x57D5, [12573] = 0x57D2, [12574] = 0x57BA, [12575] = 0x57C6,
  [12576] = 0x57BD, [12577] = 0x57BC, [12578] = 0x57B8, [12579] = 0x57B6,
  [12580] = 0x57BF, [12581] = 0x57C7, [12582] = 0x57D0, [12583] = 0x57B9,
  [12584] = 0x57C1, [12585] = 0x590E, [12586] = 0x594A, [12587] = 0x5A19,
  [12588] = 0x5A16, [12589] = 0x5A2D, [12590] = 0x5A2E, [12591] = 0x5A15,
  [12592] = 0x5A0F, [12593] = 0x5A17, [12594] = 0x5A0A, [12595] = 0x5A1E,
  [12596] = 0x5A33, [12597] = 0x5B6C, [12598] = 0x5BA7, [12599] = 0x5BAD,
  [12600] = 0x5BAC, [12601] = 0x5C03, [12602] = 0x5C56, [12603] = 0x5C54,
  [12604] = 0x5CEC, [12605] = 0x5CFF, [12606] = 0x5CEE, [12607] = 0x5CF1,
  [12608] = 0x5CF7, [12609] = 0x5D00, [12610] = 0x5CF9, [12611] = 0x5E29,
  [12612] = 0x5E28, [12613] = 0x5EA8, [12614] = 0x5EAE, [12615] = 0x5EAA,
  [12616] = 0x5EAC, [12617] = 0x5F33, [12618] = 0x5F30, [12619] = 0x5F67,
  [12620] = 0x605D, [12621] = 0x605A, [12622] = 0x6067, [12623] = 0x6041,
  [12624] = 0x60A2, [12625] = 0x6088, [12626] = 0x6080, [12627] = 0x6092,
  [12628] = 0x6081, [12629] = 0x609D, [12630] = 0x6083, [12631] = 0x6095,
  [12632] = 0x609B, [12633] = 0x6097, [12634] = 0x6087, [12635] = 0x609C,
  [12636] = 0x608E, [12637] = 0x6219, [12638] = 0x6246, [12639] = 0x62F2,
  [12640] = 0x6310, [12641] = 0x6356, [12642] = 0x632C, [12643] = 0x6344,
  [12644] = 0x6345, [12645] = 0x6336, [12646] = 0x6343, [12647] = 0x63E4,
  [12648] = 0x6339, [12649] = 0x634B, [12650] = 0x634A, [12651] = 0x633C,
  [12652] = 0x6329, [12653] = 0x6341, [12654] = 0x6334, [12655] = 0x6358,
  [12656] = 0x6354, [12657] = 0x6359, [12658] = 0x632D, [12659] = 0x6347,
  [12660] = 0x6333, [12661] = 0x635A, [12662] = 0x6351, [12663] = 0x6338,
  [12664] = 0x6357, [12665] = 0x6340, [12666] = 0x6348, [12667] = 0x654A,
  [12668] = 0x6546, [12669] = 0x65C6, [12670] = 0x65C3, [12671] = 0x65C4,
  [12672] = 0x65C2, [12673] = 0x664A, [12674] = 0x665F, [12675] = 0x6647,
  [12676] = 0x6651, [12677] = 0x6712, [12678] = 0x6713, [12679] = 0x681F,
  [12680] = 0x681A, [12681] = 0x6849, [12682] = 0x6832, [12683] = 0x6833,
  [12684] = 0x683B, [12685] = 0x684B, [12686] = 0x684F, [12687] = 0x6816,
  [12688] = 0x6831, [12689] = 0x681C, [12690] = 0x6835, [12691] = 0x682B,
  [12692] = 0x682D, [12693] = 0x682F, [12694] = 0x684E, [12695] = 0x6844,
  [12696] = 0x6834, [12697] = 0x681D, [12698] = 0x6812, [12699] = 0x6814,
  [12700] = 0x6826, [12701] = 0x6828, [12702] = 0x682E, [12703] = 0x684D,
  [12704] = 0x683A, [12705] = 0x6825, [12706] = 0x6820, [12707] = 0x6B2C,
  [12708] = 0x6B2F, [12709] = 0x6B2D, [12710] = 0x6B31, [12711] = 0x6B34,
  [12712] = 0x6B6D, [12713] = 0x8082, [12714] = 0x6B88, [12715] = 0x6BE6,
  [12716] = 0x6BE4, [12717] = 0x6BE8, [12718] = 0x6BE3, [12719] = 0x6BE2,
  [12720] = 0x6BE7, [12721] = 0x6C25, [12722] = 0x6D7A, [12723] = 0x6D63,
  [12724] = 0x6D64, [12725] = 0x6D76, [12726] = 0x6D0D, [12727] = 0x6D61,
  [12728] = 0x6D92, [12729] = 0x6D58, [12730] = 0x6D62, [12731] = 0x6D6D,
  [12732] = 0x6D6F, [12733] = 0x6D91, [12734] = 0x6D8D, [12735] = 0x6DEF,
  [12736] = 0x6D7F, [12737] = 0x6D86, [12738] = 0x6D5E, [12739] = 0x6D67,
  [12740] = 0x6D60, [12741] = 0x6D97, [12742] = 0x6D70, [12743] = 0x6D7C,
  [12744] = 0x6D5F, [12745] = 0x6D82, [12746] = 0x6D98, [12747] = 0x6D2F,
  [12748] = 0x6D68, [12749] = 0x6D8B, [12750] = 0x6D7E, [12751] = 0x6D80,
  [12752] = 0x6D84, [12753] = 0x6D16, [12754] = 0x6D83, [12755] = 0x6D7B,
  [12756] = 0x6D7D, [12757] = 0x6D75, [12758] = 0x6D90, [12759] = 0x70DC,
  [12760] = 0x70D3, [12761] = 0x70D1, [12762] = 0x70DD, [12763] = 0x70CB,
  [12764] = 0x7F39, [12765] = 0x70E2, [12766] = 0x70D7, [12767] = 0x70D2,
  [12768] = 0x70DE, [12769] = 0x70E0, [12770] = 0x70D4, [12771] = 0x70CD,
  [12772] = 0x70C5, [12773] = 0x70C6, [12774] = 0x70C7, [12775] = 0x70DA,
  [12776] = 0x70CE, [12777] = 0x70E1, [12778] = 0x7242, [12779] = 0x7278,
  [12780] = 0x7277, [12781] = 0x7276, [12782] = 0x7300, [12783] = 0x72FA,
  [12784] = 0x72F4, [12785] = 0x72FE, [12786] = 0x72F6, [12787] = 0x72F3,
  [12788] = 0x72FB, [12789] = 0x7301, [12790] = 0x73D3, [12791] = 0x73D9,
  [12792] = 0x73E5, [12793] = 0x73D6, [12794] = 0x73BC, [12795] = 0x73E7,
  [12796] = 0x73E3, [12797] = 0x73E9, [12798] = 0x73DC, [12799] = 0x73D2,
  [12800] = 0x73DB, [12801] = 0x73D4, [12802] = 0x73DD, [12803] = 0x73DA,
  [12804] = 0x73D7, [12805] = 0x73D8, [12806] = 0x73E8, [12807] = 0x74DE,
  [12808] = 0x74DF, [12809] = 0x74F4, [12810] = 0x74F5, [12811] = 0x7521,
  [12812] = 0x755B, [12813] = 0x755F, [12814] = 0x75B0, [12815] = 0x75C1,
  [12816] = 0x75BB, [12817] = 0x75C4, [12818] = 0x75C0, [12819] = 0x75BF,
  [12820] = 0x75B6, [12821] = 0x75BA, [12822] = 0x768A, [12823] = 0x76C9,
  [12824] = 0x771D, [12825] = 0x771B, [12826] = 0x7710, [12827] = 0x7713,
  [12828] = 0x7712, [12829] = 0x7723, [12830] = 0x7711, [12831] = 0x7715,
  [12832] = 0x7719, [12833] = 0x771A, [12834] = 0x7722, [12835] = 0x7727,
  [12836] = 0x7823, [12837] = 0x782C, [12838] = 0x7822, [12839] = 0x7835,
  [12840] = 0x782F, [12841] = 0x7828, [12842] = 0x782E, [12843] = 0x782B,
  [12844] = 0x7821, [12845] = 0x7829, [12846] = 0x7833, [12847] = 0x782A,
  [12848] = 0x7831, [12849] = 0x7954, [12850] = 0x795B, [12851] = 0x794F,
  [12852] = 0x795C, [12853] = 0x7953, [12854] = 0x7952, [12855] = 0x7951,
  [12856] = 0x79EB, [12857] = 0x79EC, [12858] = 0x79E0, [12859] = 0x79EE,
  [12860] = 0x79ED, [12861] = 0x79EA, [12862] = 0x79DC, [12863] = 0x79DE,
  [12864] = 0x79DD, [12865] = 0x7A86, [12866] = 0x7A89, [12867] = 0x7A85,
  [12868] = 0x7A8B, [12869] = 0x7A8C, [12870] = 0x7A8A, [12871] = 0x7A87,
  [12872] = 0x7AD8, [12873] = 0x7B10, [12874] = 0x7B04, [12875] = 0x7B13,
  [12876] = 0x7B05, [12877] = 0x7B0F, [12878] = 0x7B08, [12879] = 0x7B0A,
  [12880] = 0x7B0E, [12881] = 0x7B09, [12882] = 0x7B12, [12883] = 0x7C84,
  [12884] = 0x7C91, [12885] = 0x7C8A, [12886] = 0x7C8C, [12887] = 0x7C88,
  [12888] = 0x7C8D, [12889] = 0x7C85, [12890] = 0x7D1E, [12891] = 0x7D1D,
  [12892] = 0x7D11, [12893] = 0x7D0E, [12894] = 0x7D18, [12895] = 0x7D16,
  [12896] = 0x7D13, [12897] = 0x7D1F, [12898] = 0x7D12, [12899] = 0x7D0F,
  [12900] = 0x7D0C, [12901] = 0x7F5C, [12902] = 0x7F61, [12903] = 0x7F5E,
  [12904] = 0x7F60, [12905] = 0x7F5D, [12906] = 0x7F5B, [12907] = 0x7F96,
  [12908] = 0x7F92, [12909] = 0x7FC3, [12910] = 0x7FC2, [12911] = 0x7FC0,
  [12912] = 0x8016, [12913] = 0x803E, [12914] = 0x8039, [12915] = 0x80FA,
  [12916] = 0x80F2, [12917] = 0x80F9, [12918] = 0x80F5, [12919] = 0x8101,
  [12920] = 0x80FB, [12921] = 0x8100, [12922] = 0x8201, [12923] = 0x822F,
  [12924] = 0x8225, [12925] = 0x8333, [12926] = 0x832D, [12927] = 0x8344,
  [12928] = 0x8319, [12929] = 0x8351, [12930] = 0x8325, [12931] = 0x8356,
  [12932] = 0x833F, [12933] = 0x8341, [12934] = 0x8326, [12935] = 0x831C,
  [12936] = 0x8322, [12937] = 0x8342, [12938] = 0x834E, [12939] = 0x831B,
  [12940] = 0x832A, [12941] = 0x8308, [12942] = 0x833C, [12943] = 0x834D,
  [12944] = 0x8316, [12945] = 0x8324, [12946] = 0x8320, [12947] = 0x8337,
  [12948] = 0x832F, [12949] = 0x8329, [12950] = 0x8347, [12951] = 0x8345,
  [12952] = 0x834C, [12953] = 0x8353, [12954] = 0x831E, [12955] = 0x832C,
  [12956] = 0x834B, [12957] = 0x8327, [12958] = 0x8348, [12959] = 0x8653,
  [12960] = 0x8652, [12961] = 0x86A2, [12962] = 0x86A8, [12963] = 0x8696,
  [12964] = 0x868D, [12965] = 0x8691, [12966] = 0x869E, [12967] = 0x8687,
  [12968] = 0x8697, [12969] = 0x8686, [12970] = 0x868B, [12971] = 0x869A,
  [12972] = 0x8685, [12973] = 0x86A5, [12974] = 0x8699, [12975] = 0x86A1,
  [12976] = 0x86A7, [12977] = 0x8695, [12978] = 0x8698, [12979] = 0x868E,
  [12980] = 0x869D, [12981] = 0x8690, [12982] = 0x8694, [12983] = 0x8843,
  [12984] = 0x8844, [12985] = 0x886D, [12986] = 0x8875, [12987] = 0x8876,
  [12988] = 0x8872, [12989] = 0x8880, [12990] = 0x8871, [12991] = 0x887F,
  [12992] = 0x886F, [12993] = 0x8883, [12994] = 0x887E, [12995] = 0x8874,
  [12996] = 0x887C, [12997] = 0x8A12, [12998] = 0x8C47, [12999] = 0x8C57,
  [13000] = 0x8C7B, [13001] = 0x8CA4, [13002] = 0x8CA3, [13003] = 0x8D76,
  [13004] = 0x8D78, [13005] = 0x8DB5, [13006] = 0x8DB7, [13007] = 0x8DB6,
  [13008] = 0x8ED1, [13009] = 0x8ED3, [13010] = 0x8FFE, [13011] = 0x8FF5,
  [13012] = 0x9002, [13013] = 0x8FFF, [13014] = 0x8FFB, [13015] = 0x9004,
  [13016] = 0x8FFC, [13017] = 0x8FF6, [13018] = 0x90D6, [13019] = 0x90E0,
  [13020] = 0x90D9, [13021] = 0x90DA, [13022] = 0x90E3, [13023] = 0x90DF,
  [13024] = 0x90E5, [13025] = 0x90D8, [13026] = 0x90DB, [13027] = 0x90D7,
  [13028] = 0x90DC, [13029] = 0x90E4, [13030] = 0x9150, [13031] = 0x914E,
  [13032] = 0x914F, [13033] = 0x91D5, [13034] = 0x91E2, [13035] = 0x91DA,
  [13036] = 0x965C, [13037] = 0x965F, [13038] = 0x96BC, [13039] = 0x98E3,
  [13040] = 0x9ADF, [13041] = 0x9B2F, [13042] = 0x4E7F, [13043] = 0x5070,
  [13044] = 0x506A, [13045] = 0x5061, [13046] = 0x505E, [13047] = 0x5060,
  [13048] = 0x5053, [13049] = 0x504B, [13050] = 0x505D, [13051] = 0x5072,
  [13052] = 0x5048, [13053] = 0x504D, [13054] = 0x5041, [13055] = 0x505B,
  [13056] = 0x504A, [13057] = 0x5062, [13058] = 0x5015, [13059] = 0x5045,
  [13060] = 0x505F, [13061] = 0x5069, [13062] = 0x506B, [13063] = 0x5063,
  [13064] = 0x5064, [13065] = 0x5046, [13066] = 0x5040, [13067] = 0x506E,
  [13068] = 0x5073, [13069] = 0x5057, [13070] = 0x5051, [13071] = 0x51D0,
  [13072] = 0x526B, [13073] = 0x526D, [13074] = 0x526C, [13075] = 0x526E,
  [13076] = 0x52D6, [13077] = 0x52D3, [13078] = 0x532D, [13079] = 0x539C,
  [13080] = 0x5575, [13081] = 0x5576, [13082] = 0x553C, [13083] = 0x554D,
  [13084] = 0x5550, [13085] = 0x5534, [13086] = 0x552A, [13087] = 0x5551,
  [13088] = 0x5562, [13089] = 0x5536, [13090] = 0x5535, [13091] = 0x5530,
  [13092] = 0x5552, [13093] = 0x5545, [13094] = 0x550C, [13095] = 0x5532,
  [13096] = 0x5565, [13097] = 0x554E, [13098] = 0x5539, [13099] = 0x5548,
  [13100] = 0x552D, [13101] = 0x553B, [13102] = 0x5540, [13103] = 0x554B,
  [13104] = 0x570A, [13105] = 0x5707, [13106] = 0x57FB, [13107] = 0x5814,
  [13108] = 0x57E2, [13109] = 0x57F6, [13110] = 0x57DC, [13111] = 0x57F4,
  [13112] = 0x5800, [13113] = 0x57ED, [13114] = 0x57FD, [13115] = 0x5808,
  [13116] = 0x57F8, [13117] = 0x580B, [13118] = 0x57F3, [13119] = 0x57CF,
  [13120] = 0x5807, [13121] = 0x57EE, [13122] = 0x57E3, [13123] = 0x57F2,
  [13124] = 0x57E5, [13125] = 0x57EC, [13126] = 0x57E1, [13127] = 0x580E,
  [13128] = 0x57FC, [13129] = 0x5810, [13130] = 0x57E7, [13131] = 0x5801,
  [13132] = 0x580C, [13133] = 0x57F1, [13134] = 0x57E9, [13135] = 0x57F0,
  [13136] = 0x580D, [13137] = 0x5804, [13138] = 0x595C, [13139] = 0x5A60,
  [13140] = 0x5A58, [13141] = 0x5A55, [13142] = 0x5A67, [13143] = 0x5A5E,
  [13144] = 0x5A38, [13145] = 0x5A35, [13146] = 0x5A6D, [13147] = 0x5A50,
  [13148] = 0x5A5F, [13149] = 0x5A65, [13150] = 0x5A6C, [13151] = 0x5A53,
  [13152] = 0x5A64, [13153] = 0x5A57, [13154] = 0x5A43, [13155] = 0x5A5D,
  [13156] = 0x5A52, [13157] = 0x5A44, [13158] = 0x5A5B, [13159] = 0x5A48,
  [13160] = 0x5A8E, [13161] = 0x5A3E, [13162] = 0x5A4D, [13163] = 0x5A39,
  [13164] = 0x5A4C, [13165] = 0x5A70, [13166] = 0x5A69, [13167] = 0x5A47,
  [13168] = 0x5A51, [13169] = 0x5A56, [13170] = 0x5A42, [13171] = 0x5A5C,
  [13172] = 0x5B72, [13173] = 0x5B6E, [13174] = 0x5BC1, [13175] = 0x5BC0,
  [13176] = 0x5C59, [13177] = 0x5D1E, [13178] = 0x5D0B, [13179] = 0x5D1D,
  [13180] = 0x5D1A, [13181] = 0x5D20, [13182] = 0x5D0C, [13183] = 0x5D28,
  [13184] = 0x5D0D, [13185] = 0x5D26, [13186] = 0x5D25, [13187] = 0x5D0F,
  [13188] = 0x5D30, [13189] = 0x5D12, [13190] = 0x5D23, [13191] = 0x5D1F,
  [13192] = 0x5D2E, [13193] = 0x5E3E, [13194] = 0x5E34, [13195] = 0x5EB1,
  [13196] = 0x5EB4, [13197] = 0x5EB9, [13198] = 0x5EB2, [13199] = 0x5EB3,
  [13200] = 0x5F36, [13201] = 0x5F38, [13202] = 0x5F9B, [13203] = 0x5F96,
  [13204] = 0x5F9F, [13205] = 0x608A, [13206] = 0x6090, [13207] = 0x6086,
  [13208] = 0x60BE, [13209] = 0x60B0, [13210] = 0x60BA, [13211] = 0x60D3,
  [13212] = 0x60D4, [13213] = 0x60CF, [13214] = 0x60E4, [13215] = 0x60D9,
  [13216] = 0x60DD, [13217] = 0x60C8, [13218] = 0x60B1, [13219] = 0x60DB,
  [13220] = 0x60B7, [13221] = 0x60CA, [13222] = 0x60BF, [13223] = 0x60C3,
  [13224] = 0x60CD, [13225] = 0x60C0, [13226] = 0x6332, [13227] = 0x6365,
  [13228] = 0x638A, [13229] = 0x6382, [13230] = 0x637D, [13231] = 0x63BD,
  [13232] = 0x639E, [13233] = 0x63AD, [13234] = 0x639D, [13235] = 0x6397,
  [13236] = 0x63AB, [13237] = 0x638E, [13238] = 0x636F, [13239] = 0x6387,
  [13240] = 0x6390, [13241] = 0x636E, [13242] = 0x63AF, [13243] = 0x6375,
  [13244] = 0x639C, [13245] = 0x636D, [13246] = 0x63AE, [13247] = 0x637C,
  [13248] = 0x63A4, [13249] = 0x633B, [13250] = 0x639F, [13251] = 0x6378,
  [13252] = 0x6385, [13253] = 0x6381, [13254] = 0x6391, [13255] = 0x638D,
  [13256] = 0x6370, [13257] = 0x6553, [13258] = 0x65CD, [13259] = 0x6665,
  [13260] = 0x6661, [13261] = 0x665B, [13262] = 0x6659, [13263] = 0x665C,
  [13264] = 0x6662, [13265] = 0x6718, [13266] = 0x6879, [13267] = 0x6887,
  [13268] = 0x6890, [13269] = 0x689C, [13270] = 0x686D, [13271] = 0x686E,
  [13272] = 0x68AE, [13273] = 0x68AB, [13274] = 0x6956, [13275] = 0x686F,
  [13276] = 0x68A3, [13277] = 0x68AC, [13278] = 0x68A9, [13279] = 0x6875,
  [13280] = 0x6874, [13281] = 0x68B2, [13282] = 0x688F, [13283] = 0x6877,
  [13284] = 0x6892, [13285] = 0x687C, [13286] = 0x686B, [13287] = 0x6872,
  [13288] = 0x68AA, [13289] = 0x6880, [13290] = 0x6871, [13291] = 0x687E,
  [13292] = 0x689B, [13293] = 0x6896, [13294] = 0x688B, [13295] = 0x68A0,
  [13296] = 0x6889, [13297] = 0x68A4, [13298] = 0x6878, [13299] = 0x687B,
  [13300] = 0x6891, [13301] = 0x688C, [13302] = 0x688A, [13303] = 0x687D,
  [13304] = 0x6B36, [13305] = 0x6B33, [13306] = 0x6B37, [13307] = 0x6B38,
  [13308] = 0x6B91, [13309] = 0x6B8F, [13310] = 0x6B8D, [13311] = 0x6B8E,
  [13312] = 0x6B8C, [13313] = 0x6C2A, [13314] = 0x6DC0, [13315] = 0x6DAB,
  [13316] = 0x6DB4, [13317] = 0x6DB3, [13318] = 0x6E74, [13319] = 0x6DAC,
  [13320] = 0x6DE9, [13321] = 0x6DE2, [13322] = 0x6DB7, [13323] = 0x6DF6,
  [13324] = 0x6DD4, [13325] = 0x6E00, [13326] = 0x6DC8, [13327] = 0x6DE0,
  [13328] = 0x6DDF, [13329] = 0x6DD6, [13330] = 0x6DBE, [13331] = 0x6DE5,
  [13332] = 0x6DDC, [13333] = 0x6DDD, [13334] = 0x6DDB, [13335] = 0x6DF4,
  [13336] = 0x6DCA, [13337] = 0x6DBD, [13338] = 0x6DED, [13339] = 0x6DF0,
  [13340] = 0x6DBA, [13341] = 0x6DD5, [13342] = 0x6DC2, [13343] = 0x6DCF,
  [13344] = 0x6DC9, [13345] = 0x6DD0, [13346] = 0x6DF2, [13347] = 0x6DD3,
  [13348] = 0x6DFD, [13349] = 0x6DD7, [13350] = 0x6DCD, [13351] = 0x6DE3,
  [13352] = 0x6DBB, [13353] = 0x70FA, [13354] = 0x710D, [13355] = 0x70F7,
  [13356] = 0x7117, [13357] = 0x70F4, [13358] = 0x710C, [13359] = 0x70F0,
  [13360] = 0x7104, [13361] = 0x70F3, [13362] = 0x7110, [13363] = 0x70FC,
  [13364] = 0x70FF, [13365] = 0x7106, [13366] = 0x7113, [13367] = 0x7100,
  [13368] = 0x70F8, [13369] = 0x70F6, [13370] = 0x710B, [13371] = 0x7102,
  [13372] = 0x710E, [13373] = 0x727E, [13374] = 0x727B, [13375] = 0x727C,
  [13376] = 0x727F, [13377] = 0x731D, [13378] = 0x7317, [13379] = 0x7307,
  [13380] = 0x7311, [13381] = 0x7318, [13382] = 0x730A, [13383] = 0x7308,
  [13384] = 0x72FF, [13385] = 0x730F, [13386] = 0x731E, [13387] = 0x7388,
  [13388] = 0x73F6, [13389] = 0x73F8, [13390] = 0x73F5, [13391] = 0x7404,
  [13392] = 0x7401, [13393] = 0x73FD, [13394] = 0x7407, [13395] = 0x7400,
  [13396] = 0x73FA, [13397] = 0x73FC, [13398] = 0x73FF, [13399] = 0x740C,
  [13400] = 0x740B, [13401] = 0x73F4, [13402] = 0x7408, [13403] = 0x7564,
  [13404] = 0x7563, [13405] = 0x75CE, [13406] = 0x75D2, [13407] = 0x75CF,
  [13408] = 0x75CB, [13409] = 0x75CC, [13410] = 0x75D1, [13411] = 0x75D0,
  [13412] = 0x768F, [13413] = 0x7689, [13414] = 0x76D3, [13415] = 0x7739,
  [13416] = 0x772F, [13417] = 0x772D, [13418] = 0x7731, [13419] = 0x7732,
  [13420] = 0x7734, [13421] = 0x7733, [13422] = 0x773D, [13423] = 0x7725,
  [13424] = 0x773B, [13425] = 0x7735, [13426] = 0x7848, [13427] = 0x7852,
  [13428] = 0x7849, [13429] = 0x784D, [13430] = 0x784A, [13431] = 0x784C,
  [13432] = 0x7826, [13433] = 0x7845, [13434] = 0x7850, [13435] = 0x7964,
  [13436] = 0x7967, [13437] = 0x7969, [13438] = 0x796A, [13439] = 0x7963,
  [13440] = 0x796B, [13441] = 0x7961, [13442] = 0x79BB, [13443] = 0x79FA,
  [13444] = 0x79F8, [13445] = 0x79F6, [13446] = 0x79F7, [13447] = 0x7A8F,
  [13448] = 0x7A94, [13449] = 0x7A90, [13450] = 0x7B35, [13451] = 0x7B47,
  [13452] = 0x7B34, [13453] = 0x7B25, [13454] = 0x7B30, [13455] = 0x7B22,
  [13456] = 0x7B24, [13457] = 0x7B33, [13458] = 0x7B18, [13459] = 0x7B2A,
  [13460] = 0x7B1D, [13461] = 0x7B31, [13462] = 0x7B2B, [13463] = 0x7B2D,
  [13464] = 0x7B2F, [13465] = 0x7B32, [13466] = 0x7B38, [13467] = 0x7B1A,
  [13468] = 0x7B23, [13469] = 0x7C94, [13470] = 0x7C98, [13471] = 0x7C96,
  [13472] = 0x7CA3, [13473] = 0x7D35, [13474] = 0x7D3D, [13475] = 0x7D38,
  [13476] = 0x7D36, [13477] = 0x7D3A, [13478] = 0x7D45, [13479] = 0x7D2C,
  [13480] = 0x7D29, [13481] = 0x7D41, [13482] = 0x7D47, [13483] = 0x7D3E,
  [13484] = 0x7D3F, [13485] = 0x7D4A, [13486] = 0x7D3B, [13487] = 0x7D28,
  [13488] = 0x7F63, [13489] = 0x7F95, [13490] = 0x7F9C, [13491] = 0x7F9D,
  [13492] = 0x7F9B, [13493] = 0x7FCA, [13494] = 0x7FCB, [13495] = 0x7FCD,
  [13496] = 0x7FD0, [13497] = 0x7FD1, [13498] = 0x7FC7, [13499] = 0x7FCF,
  [13500] = 0x7FC9, [13501] = 0x801F, [13502] = 0x801E, [13503] = 0x801B,
  [13504] = 0x8047, [13505] = 0x8043, [13506] = 0x8048, [13507] = 0x8118,
  [13508] = 0x8125, [13509] = 0x8119, [13510] = 0x811B, [13511] = 0x812D,
  [13512] = 0x811F, [13513] = 0x812C, [13514] = 0x811E, [13515] = 0x8121,
  [13516] = 0x8115, [13517] = 0x8127, [13518] = 0x811D, [13519] = 0x8122,
  [13520] = 0x8211, [13521] = 0x8238, [13522] = 0x8233, [13523] = 0x823A,
  [13524] = 0x8234, [13525] = 0x8232, [13526] = 0x8274, [13527] = 0x8390,
  [13528] = 0x83A3, [13529] = 0x83A8, [13530] = 0x838D, [13531] = 0x837A,
  [13532] = 0x8373, [13533] = 0x83A4, [13534] = 0x8374, [13535] = 0x838F,
  [13536] = 0x8381, [13537] = 0x8395, [13538] = 0x8399, [13539] = 0x8375,
  [13540] = 0x8394, [13541] = 0x83A9, [13542] = 0x837D, [13543] = 0x8383,
  [13544] = 0x838C, [13545] = 0x839D, [13546] = 0x839B, [13547] = 0x83AA,
  [13548] = 0x838B, [13549] = 0x837E, [13550] = 0x83A5, [13551] = 0x83AF,
  [13552] = 0x8388, [13553] = 0x8397, [13554] = 0x83B0, [13555] = 0x837F,
  [13556] = 0x83A6, [13557] = 0x8387, [13558] = 0x83AE, [13559] = 0x8376,
  [13560] = 0x839A, [13561] = 0x8659, [13562] = 0x8656, [13563] = 0x86BF,
  [13564] = 0x86B7, [13565] = 0x86C2, [13566] = 0x86C1, [13567] = 0x86C5,
  [13568] = 0x86BA, [13569] = 0x86B0, [13570] = 0x86C8, [13571] = 0x86B9,
  [13572] = 0x86B3, [13573] = 0x86B8, [13574] = 0x86CC, [13575] = 0x86B4,
  [13576] = 0x86BB, [13577] = 0x86BC, [13578] = 0x86C3, [13579] = 0x86BD,
  [13580] = 0x86BE, [13581] = 0x8852, [13582] = 0x8889, [13583] = 0x8895,
  [13584] = 0x88A8, [13585] = 0x88A2, [13586] = 0x88AA, [13587] = 0x889A,
  [13588] = 0x8891, [13589] = 0x88A1, [13590] = 0x889F, [13591] = 0x8898,
  [13592] = 0x88A7, [13593] = 0x8899, [13594] = 0x889B, [13595] = 0x8897,
  [13596] = 0x88A4, [13597] = 0x88AC, [13598] = 0x888C, [13599] = 0x8893,
  [13600] = 0x888E, [13601] = 0x8982, [13602] = 0x89D6, [13603] = 0x89D9,
  [13604] = 0x89D5, [13605] = 0x8A30, [13606] = 0x8A27, [13607] = 0x8A2C,
  [13608] = 0x8A1E, [13609] = 0x8C39, [13610] = 0x8C3B, [13611] = 0x8C5C,
  [13612] = 0x8C5D, [13613] = 0x8C7D, [13614] = 0x8CA5, [13615] = 0x8D7D,
  [13616] = 0x8D7B, [13617] = 0x8D79, [13618] = 0x8DBC, [13619] = 0x8DC2,
  [13620] = 0x8DB9, [13621] = 0x8DBF, [13622] = 0x8DC1, [13623] = 0x8ED8,
  [13624] = 0x8EDE, [13625] = 0x8EDD, [13626] = 0x8EDC, [13627] = 0x8ED7,
  [13628] = 0x8EE0, [13629] = 0x8EE1, [13630] = 0x9024, [13631] = 0x900B,
  [13632] = 0x9011, [13633] = 0x901C, [13634] = 0x900C, [13635] = 0x9021,
  [13636] = 0x90EF, [13637] = 0x90EA, [13638] = 0x90F0, [13639] = 0x90F4,
  [13640] = 0x90F2, [13641] = 0x90F3, [13642] = 0x90D4, [13643] = 0x90EB,
  [13644] = 0x90EC, [13645] = 0x90E9, [13646] = 0x9156, [13647] = 0x9158,
  [13648] = 0x915A, [13649] = 0x9153, [13650] = 0x9155, [13651] = 0x91EC,
  [13652] = 0x91F4, [13653] = 0x91F1, [13654] = 0x91F3, [13655] = 0x91F8,
  [13656] = 0x91E4, [13657] = 0x91F9, [13658] = 0x91EA, [13659] = 0x91EB,
  [13660] = 0x91F7, [13661] = 0x91E8, [13662] = 0x91EE, [13663] = 0x957A,
  [13664] = 0x9586, [13665] = 0x9588, [13666] = 0x967C, [13667] = 0x966D,
  [13668] = 0x966B, [13669] = 0x9671, [13670] = 0x966F, [13671] = 0x96BF,
  [13672] = 0x976A, [13673] = 0x9804, [13674] = 0x98E5, [13675] = 0x9997,
  [13676] = 0x509B, [13677] = 0x5095, [13678] = 0x5094, [13679] = 0x509E,
  [13680] = 0x508B, [13681] = 0x50A3, [13682] = 0x5083, [13683] = 0x508C,
  [13684] = 0x508E, [13685] = 0x509D, [13686] = 0x5068, [13687] = 0x509C,
  [13688] = 0x5092, [13689] = 0x5082, [13690] = 0x5087, [13691] = 0x515F,
  [13692] = 0x51D4, [13693] = 0x5312, [13694] = 0x5311, [13695] = 0x53A4,
  [13696] = 0x53A7, [13697] = 0x5591, [13698] = 0x55A8, [13699] = 0x55A5,
  [13700] = 0x55AD, [13701] = 0x5577, [13702] = 0x5645, [13703] = 0x55A2,
  [13704] = 0x5593, [13705] = 0x5588, [13706] = 0x558F, [13707] = 0x55B5,
  [13708] = 0x5581, [13709] = 0x55A3, [13710] = 0x5592, [13711] = 0x55A4,
  [13712] = 0x557D, [13713] = 0x558C, [13714] = 0x55A6, [13715] = 0x557F,
  [13716] = 0x5595, [13717] = 0x55A1, [13718] = 0x558E, [13719] = 0x570C,
  [13720] = 0x5829, [13721] = 0x5837, [13722] = 0x5819, [13723] = 0x581E,
  [13724] = 0x5827, [13725] = 0x5823, [13726] = 0x5828, [13727] = 0x57F5,
  [13728] = 0x5848, [13729] = 0x5825, [13730] = 0x581C, [13731] = 0x581B,
  [13732] = 0x5833, [13733] = 0x583F, [13734] = 0x5836, [13735] = 0x582E,
  [13736] = 0x5839, [13737] = 0x5838, [13738] = 0x582D, [13739] = 0x582C,
  [13740] = 0x583B, [13741] = 0x5961, [13742] = 0x5AAF, [13743] = 0x5A94,
  [13744] = 0x5A9F, [13745] = 0x5A7A, [13746] = 0x5AA2, [13747] = 0x5A9E,
  [13748] = 0x5A78, [13749] = 0x5AA6, [13750] = 0x5A7C, [13751] = 0x5AA5,
  [13752] = 0x5AAC, [13753] = 0x5A95, [13754] = 0x5AAE, [13755] = 0x5A37,
  [13756] = 0x5A84, [13757] = 0x5A8A, [13758] = 0x5A97, [13759] = 0x5A83,
  [13760] = 0x5A8B, [13761] = 0x5AA9, [13762] = 0x5A7B, [13763] = 0x5A7D,
  [13764] = 0x5A8C, [13765] = 0x5A9C, [13766] = 0x5A8F, [13767] = 0x5A93,
  [13768] = 0x5A9D, [13769] = 0x5BEA, [13770] = 0x5BCD, [13771] = 0x5BCB,
  [13772] = 0x5BD4, [13773] = 0x5BD1, [13774] = 0x5BCA, [13775] = 0x5BCE,
  [13776] = 0x5C0C, [13777] = 0x5C30, [13778] = 0x5D37, [13779] = 0x5D43,
  [13780] = 0x5D6B, [13781] = 0x5D41, [13782] = 0x5D4B, [13783] = 0x5D3F,
  [13784] = 0x5D35, [13785] = 0x5D51, [13786] = 0x5D4E, [13787] = 0x5D55,
  [13788] = 0x5D33, [13789] = 0x5D3A, [13790] = 0x5D52, [13791] = 0x5D3D,
  [13792] = 0x5D31, [13793] = 0x5D59, [13794] = 0x5D42, [13795] = 0x5D39,
  [13796] = 0x5D49, [13797] = 0x5D38, [13798] = 0x5D3C, [13799] = 0x5D32,
  [13800] = 0x5D36, [13801] = 0x5D40, [13802] = 0x5D45, [13803] = 0x5E44,
  [13804] = 0x5E41, [13805] = 0x5F58, [13806] = 0x5FA6, [13807] = 0x5FA5,
  [13808] = 0x5FAB, [13809] = 0x60C9, [13810] = 0x60B9, [13811] = 0x60CC,
  [13812] = 0x60E2, [13813] = 0x60CE, [13814] = 0x60C4, [13815] = 0x6114,
  [13816] = 0x60F2, [13817] = 0x610A, [13818] = 0x6116, [13819] = 0x6105,
  [13820] = 0x60F5, [13821] = 0x6113, [13822] = 0x60F8, [13823] = 0x60FC,
  [13824] = 0x60FE, [13825] = 0x60C1, [13826] = 0x6103, [13827] = 0x6118,
  [13828] = 0x611D, [13829] = 0x6110, [13830] = 0x60FF, [13831] = 0x6104,
  [13832] = 0x610B, [13833] = 0x624A, [13834] = 0x6394, [13835] = 0x63B1,
  [13836] = 0x63B0, [13837] = 0x63CE, [13838] = 0x63E5, [13839] = 0x63E8,
  [13840] = 0x63EF, [13841] = 0x63C3, [13842] = 0x649D, [13843] = 0x63F3,
  [13844] = 0x63CA, [13845] = 0x63E0, [13846] = 0x63F6, [13847] = 0x63D5,
  [13848] = 0x63F2, [13849] = 0x63F5, [13850] = 0x6461, [13851] = 0x63DF,
  [13852] = 0x63BE, [13853] = 0x63DD, [13854] = 0x63DC, [13855] = 0x63C4,
  [13856] = 0x63D8, [13857] = 0x63D3, [13858] = 0x63C2, [13859] = 0x63C7,
  [13860] = 0x63CC, [13861] = 0x63CB, [13862] = 0x63C8, [13863] = 0x63F0,
  [13864] = 0x63D7, [13865] = 0x63D9, [13866] = 0x6532, [13867] = 0x6567,
  [13868] = 0x656A, [13869] = 0x6564, [13870] = 0x655C, [13871] = 0x6568,
  [13872] = 0x6565, [13873] = 0x658C, [13874] = 0x659D, [13875] = 0x659E,
  [13876] = 0x65AE, [13877] = 0x65D0, [13878] = 0x65D2, [13879] = 0x667C,
  [13880] = 0x666C, [13881] = 0x667B, [13882] = 0x6680, [13883] = 0x6671,
  [13884] = 0x6679, [13885] = 0x666A, [13886] = 0x6672, [13887] = 0x6701,
  [13888] = 0x690C, [13889] = 0x68D3, [13890] = 0x6904, [13891] = 0x68DC,
  [13892] = 0x692A, [13893] = 0x68EC, [13894] = 0x68EA, [13895] = 0x68F1,
  [13896] = 0x690F, [13897] = 0x68D6, [13898] = 0x68F7, [13899] = 0x68EB,
  [13900] = 0x68E4, [13901] = 0x68F6, [13902] = 0x6913, [13903] = 0x6910,
  [13904] = 0x68F3, [13905] = 0x68E1, [13906] = 0x6907, [13907] = 0x68CC,
  [13908] = 0x6908, [13909] = 0x6970, [13910] = 0x68B4, [13911] = 0x6911,
  [13912] = 0x68EF, [13913] = 0x68C6, [13914] = 0x6914, [13915] = 0x68F8,
  [13916] = 0x68D0, [13917] = 0x68FD, [13918] = 0x68FC, [13919] = 0x68E8,
  [13920] = 0x690B, [13921] = 0x690A, [13922] = 0x6917, [13923] = 0x68CE,
  [13924] = 0x68C8, [13925] = 0x68DD, [13926] = 0x68DE, [13927] = 0x68E6,
  [13928] = 0x68F4, [13929] = 0x68D1, [13930] = 0x6906, [13931] = 0x68D4,
  [13932] = 0x68E9, [13933] = 0x6915, [13934] = 0x6925, [13935] = 0x68C7,
  [13936] = 0x6B39, [13937] = 0x6B3B, [13938] = 0x6B3F, [13939] = 0x6B3C,
  [13940] = 0x6B94, [13941] = 0x6B97, [13942] = 0x6B99, [13943] = 0x6B95,
  [13944] = 0x6BBD, [13945] = 0x6BF0, [13946] = 0x6BF2, [13947] = 0x6BF3,
  [13948] = 0x6C30, [13949] = 0x6DFC, [13950] = 0x6E46, [13951] = 0x6E47,
  [13952] = 0x6E1F, [13953] = 0x6E49, [13954] = 0x6E88, [13955] = 0x6E3C,
  [13956] = 0x6E3D, [13957] = 0x6E45, [13958] = 0x6E62, [13959] = 0x6E2B,
  [13960] = 0x6E3F, [13961] = 0x6E41, [13962] = 0x6E5D, [13963] = 0x6E73,
  [13964] = 0x6E1C, [13965] = 0x6E33, [13966] = 0x6E4B, [13967] = 0x6E40,
  [13968] = 0x6E51, [13969] = 0x6E3B, [13970] = 0x6E03, [13971] = 0x6E2E,
  [13972] = 0x6E5E, [13973] = 0x6E68, [13974] = 0x6E5C, [13975] = 0x6E61,
  [13976] = 0x6E31, [13977] = 0x6E28, [13978] = 0x6E60, [13979] = 0x6E71,
  [13980] = 0x6E6B, [13981] = 0x6E39, [13982] = 0x6E22, [13983] = 0x6E30,
  [13984] = 0x6E53, [13985] = 0x6E65, [13986] = 0x6E27, [13987] = 0x6E78,
  [13988] = 0x6E64, [13989] = 0x6E77, [13990] = 0x6E55, [13991] = 0x6E79,
  [13992] = 0x6E52, [13993] = 0x6E66, [13994] = 0x6E35, [13995] = 0x6E36,
  [13996] = 0x6E5A, [13997] = 0x7120, [13998] = 0x711E, [13999] = 0x712F,
  [14000] = 0x70FB, [14001] = 0x712E, [14002] = 0x7131, [14003] = 0x7123,
  [14004] = 0x7125, [14005] = 0x7122, [14006] = 0x7132, [14007] = 0x711F,
  [14008] = 0x7128, [14009] = 0x713A, [14010] = 0x711B, [14011] = 0x724B,
  [14012] = 0x725A, [14013] = 0x7288, [14014] = 0x7289, [14015] = 0x7286,
  [14016] = 0x7285, [14017] = 0x728B, [14018] = 0x7312, [14019] = 0x730B,
  [14020] = 0x7330, [14021] = 0x7322, [14022] = 0x7331, [14023] = 0x7333,
  [14024] = 0x7327, [14025] = 0x7332, [14026] = 0x732D, [14027] = 0x7326,
  [14028] = 0x7323, [14029] = 0x7335, [14030] = 0x730C, [14031] = 0x742E,
  [14032] = 0x742C, [14033] = 0x7430, [14034] = 0x742B, [14035] = 0x7416,
  [14036] = 0x741A, [14037] = 0x7421, [14038] = 0x742D, [14039] = 0x7431,
  [14040] = 0x7424, [14041] = 0x7423, [14042] = 0x741D, [14043] = 0x7429,
  [14044] = 0x7420, [14045] = 0x7432, [14046] = 0x74FB, [14047] = 0x752F,
  [14048] = 0x756F, [14049] = 0x756C, [14050] = 0x75E7, [14051] = 0x75DA,
  [14052] = 0x75E1, [14053] = 0x75E6, [14054] = 0x75DD, [14055] = 0x75DF,
  [14056] = 0x75E4, [14057] = 0x75D7, [14058] = 0x7695, [14059] = 0x7692,
  [14060] = 0x76DA, [14061] = 0x7746, [14062] = 0x7747, [14063] = 0x7744,
  [14064] = 0x774D, [14065] = 0x7745, [14066] = 0x774A, [14067] = 0x774E,
  [14068] = 0x774B, [14069] = 0x774C, [14070] = 0x77DE, [14071] = 0x77EC,
  [14072] = 0x7860, [14073] = 0x7864, [14074] = 0x7865, [14075] = 0x785C,
  [14076] = 0x786D, [14077] = 0x7871, [14078] = 0x786A, [14079] = 0x786E,
  [14080] = 0x7870, [14081] = 0x7869, [14082] = 0x7868, [14083] = 0x785E,
  [14084] = 0x7862, [14085] = 0x7974, [14086] = 0x7973, [14087] = 0x7972,
  [14088] = 0x7970, [14089] = 0x7A02, [14090] = 0x7A0A, [14091] = 0x7A03,
  [14092] = 0x7A0C, [14093] = 0x7A04, [14094] = 0x7A99, [14095] = 0x7AE6,
  [14096] = 0x7AE4, [14097] = 0x7B4A, [14098] = 0x7B3B, [14099] = 0x7B44,
  [14100] = 0x7B48, [14101] = 0x7B4C, [14102] = 0x7B4E, [14103] = 0x7B40,
  [14104] = 0x7B58, [14105] = 0x7B45, [14106] = 0x7CA2, [14107] = 0x7C9E,
  [14108] = 0x7CA8, [14109] = 0x7CA1, [14110] = 0x7D58, [14111] = 0x7D6F,
  [14112] = 0x7D63, [14113] = 0x7D53, [14114] = 0x7D56, [14115] = 0x7D67,
  [14116] = 0x7D6A, [14117] = 0x7D4F, [14118] = 0x7D6D, [14119] = 0x7D5C,
  [14120] = 0x7D6B, [14121] = 0x7D52, [14122] = 0x7D54, [14123] = 0x7D69,
  [14124] = 0x7D51, [14125] = 0x7D5F, [14126] = 0x7D4E, [14127] = 0x7F3E,
  [14128] = 0x7F3F, [14129] = 0x7F65, [14130] = 0x7F66, [14131] = 0x7FA2,
  [14132] = 0x7FA0, [14133] = 0x7FA1, [14134] = 0x7FD7, [14135] = 0x8051,
  [14136] = 0x804F, [14137] = 0x8050, [14138] = 0x80FE, [14139] = 0x80D4,
  [14140] = 0x8143, [14141] = 0x814A, [14142] = 0x8152, [14143] = 0x814F,
  [14144] = 0x8147, [14145] = 0x813D, [14146] = 0x814D, [14147] = 0x813A,
  [14148] = 0x81E6, [14149] = 0x81EE, [14150] = 0x81F7, [14151] = 0x81F8,
  [14152] = 0x81F9, [14153] = 0x8204, [14154] = 0x823C, [14155] = 0x823D,
  [14156] = 0x823F, [14157] = 0x8275, [14158] = 0x833B, [14159] = 0x83CF,
  [14160] = 0x83F9, [14161] = 0x8423, [14162] = 0x83C0, [14163] = 0x83E8,
  [14164] = 0x8412, [14165] = 0x83E7, [14166] = 0x83E4, [14167] = 0x83FC,
  [14168] = 0x83F6, [14169] = 0x8410, [14170] = 0x83C6, [14171] = 0x83C8,
  [14172] = 0x83EB, [14173] = 0x83E3, [14174] = 0x83BF, [14175] = 0x8401,
  [14176] = 0x83DD, [14177] = 0x83E5, [14178] = 0x83D8, [14179] = 0x83FF,
  [14180] = 0x83E1, [14181] = 0x83CB, [14182] = 0x83CE, [14183] = 0x83D6,
  [14184] = 0x83F5, [14185] = 0x83C9, [14186] = 0x8409, [14187] = 0x840F,
  [14188] = 0x83DE, [14189] = 0x8411, [14190] = 0x8406, [14191] = 0x83C2,
  [14192] = 0x83F3, [14193] = 0x83D5, [14194] = 0x83FA, [14195] = 0x83C7,
  [14196] = 0x83D1, [14197] = 0x83EA, [14198] = 0x8413, [14199] = 0x83C3,
  [14200] = 0x83EC, [14201] = 0x83EE, [14202] = 0x83C4, [14203] = 0x83FB,
  [14204] = 0x83D7, [14205] = 0x83E2, [14206] = 0x841B, [14207] = 0x83DB,
  [14208] = 0x83FE, [14209] = 0x86D8, [14210] = 0x86E2, [14211] = 0x86E6,
  [14212] = 0x86D3, [14213] = 0x86E3, [14214] = 0x86DA, [14215] = 0x86EA,
  [14216] = 0x86DD, [14217] = 0x86EB, [14218] = 0x86DC, [14219] = 0x86EC,
  [14220] = 0x86E9, [14221] = 0x86D7, [14222] = 0x86E8, [14223] = 0x86D1,
  [14224] = 0x8848, [14225] = 0x8856, [14226] = 0x8855, [14227] = 0x88BA,
  [14228] = 0x88D7, [14229] = 0x88B9, [14230] = 0x88B8, [14231] = 0x88C0,
  [14232] = 0x88BE, [14233] = 0x88B6, [14234] = 0x88BC, [14235] = 0x88B7,
  [14236] = 0x88BD, [14237] = 0x88B2, [14238] = 0x8901, [14239] = 0x88C9,
  [14240] = 0x8995, [14241] = 0x8998, [14242] = 0x8997, [14243] = 0x89DD,
  [14244] = 0x89DA, [14245] = 0x89DB, [14246] = 0x8A4E, [14247] = 0x8A4D,
  [14248] = 0x8A39, [14249] = 0x8A59, [14250] = 0x8A40, [14251] = 0x8A57,
  [14252] = 0x8A58, [14253] = 0x8A44, [14254] = 0x8A45, [14255] = 0x8A52,
  [14256] = 0x8A48, [14257] = 0x8A51, [14258] = 0x8A4A, [14259] = 0x8A4C,
  [14260] = 0x8A4F, [14261] = 0x8C5F, [14262] = 0x8C81, [14263] = 0x8C80,
  [14264] = 0x8CBA, [14265] = 0x8CBE, [14266] = 0x8CB0, [14267] = 0x8CB9,
  [14268] = 0x8CB5, [14269] = 0x8D84, [14270] = 0x8D80, [14271] = 0x8D89,
  [14272] = 0x8DD8, [14273] = 0x8DD3, [14274] = 0x8DCD, [14275] = 0x8DC7,
  [14276] = 0x8DD6, [14277] = 0x8DDC, [14278] = 0x8DCF, [14279] = 0x8DD5,
  [14280] = 0x8DD9, [14281] = 0x8DC8, [14282] = 0x8DD7, [14283] = 0x8DC5,
  [14284] = 0x8EEF, [14285] = 0x8EF7, [14286] = 0x8EFA, [14287] = 0x8EF9,
  [14288] = 0x8EE6, [14289] = 0x8EEE, [14290] = 0x8EE5, [14291] = 0x8EF5,
  [14292] = 0x8EE7, [14293] = 0x8EE8, [14294] = 0x8EF6, [14295] = 0x8EEB,
  [14296] = 0x8EF1, [14297] = 0x8EEC, [14298] = 0x8EF4, [14299] = 0x8EE9,
  [14300] = 0x902D, [14301] = 0x9034, [14302] = 0x902F, [14303] = 0x9106,
  [14304] = 0x912C, [14305] = 0x9104, [14306] = 0x90FF, [14307] = 0x90FC,
  [14308] = 0x9108, [14309] = 0x90F9, [14310] = 0x90FB, [14311] = 0x9101,
  [14312] = 0x9100, [14313] = 0x9107, [14314] = 0x9105, [14315] = 0x9103,
  [14316] = 0x9161, [14317] = 0x9164, [14318] = 0x915F, [14319] = 0x9162,
  [14320] = 0x9160, [14321] = 0x9201, [14322] = 0x920A, [14323] = 0x9225,
  [14324] = 0x9203, [14325] = 0x921A, [14326] = 0x9226, [14327] = 0x920F,
  [14328] = 0x920C, [14329] = 0x9200, [14330] = 0x9212, [14331] = 0x91FF,
  [14332] = 0x91FD, [14333] = 0x9206, [14334] = 0x9204, [14335] = 0x9227,
  [14336] = 0x9202, [14337] = 0x921C, [14338] = 0x9224, [14339] = 0x9219,
  [14340] = 0x9217, [14341] = 0x9205, [14342] = 0x9216, [14343] = 0x957B,
  [14344] = 0x958D, [14345] = 0x958C, [14346] = 0x9590, [14347] = 0x9687,
  [14348] = 0x967E, [14349] = 0x9688, [14350] = 0x9689, [14351] = 0x9683,
  [14352] = 0x9680, [14353] = 0x96C2, [14354] = 0x96C8, [14355] = 0x96C3,
  [14356] = 0x96F1, [14357] = 0x96F0, [14358] = 0x976C, [14359] = 0x9770,
  [14360] = 0x976E, [14361] = 0x9807, [14362] = 0x98A9, [14363] = 0x98EB,
  [14364] = 0x9CE6, [14365] = 0x9EF9, [14366] = 0x4E83, [14367] = 0x4E84,
  [14368] = 0x4EB6, [14369] = 0x50BD, [14370] = 0x50BF, [14371] = 0x50C6,
  [14372] = 0x50AE, [14373] = 0x50C4, [14374] = 0x50CA, [14375] = 0x50B4,
  [14376] = 0x50C8, [14377] = 0x50C2, [14378] = 0x50B0, [14379] = 0x50C1,
  [14380] = 0x50BA, [14381] = 0x50B1, [14382] = 0x50CB, [14383] = 0x50C9,
  [14384] = 0x50B6, [14385] = 0x50B8, [14386] = 0x51D7, [14387] = 0x527A,
  [14388] = 0x5278, [14389] = 0x527B, [14390] = 0x527C, [14391] = 0x55C3,
  [14392] = 0x55DB, [14393] = 0x55CC, [14394] = 0x55D0, [14395] = 0x55CB,
  [14396] = 0x55CA, [14397] = 0x55DD, [14398] = 0x55C0, [14399] = 0x55D4,
  [14400] = 0x55C4, [14401] = 0x55E9, [14402] = 0x55BF, [14403] = 0x55D2,
  [14404] = 0x558D, [14405] = 0x55CF, [14406] = 0x55D5, [14407] = 0x55E2,
  [14408] = 0x55D6, [14409] = 0x55C8, [14410] = 0x55F2, [14411] = 0x55CD,
  [14412] = 0x55D9, [14413] = 0x55C2, [14414] = 0x5714, [14415] = 0x5853,
  [14416] = 0x5868, [14417] = 0x5864, [14418] = 0x584F, [14419] = 0x584D,
  [14420] = 0x5849, [14421] = 0x586F, [14422] = 0x5855, [14423] = 0x584E,
  [14424] = 0x585D, [14425] = 0x5859, [14426] = 0x5865, [14427] = 0x585B,
  [14428] = 0x583D, [14429] = 0x5863, [14430] = 0x5871, [14431] = 0x58FC,
  [14432] = 0x5AC7, [14433] = 0x5AC4, [14434] = 0x5ACB, [14435] = 0x5ABA,
  [14436] = 0x5AB8, [14437] = 0x5AB1, [14438] = 0x5AB5, [14439] = 0x5AB0,
  [14440] = 0x5ABF, [14441] = 0x5AC8, [14442] = 0x5ABB, [14443] = 0x5AC6,
  [14444] = 0x5AB7, [14445] = 0x5AC0, [14446] = 0x5ACA, [14447] = 0x5AB4,
  [14448] = 0x5AB6, [14449] = 0x5ACD, [14450] = 0x5AB9, [14451] = 0x5A90,
  [14452] = 0x5BD6, [14453] = 0x5BD8, [14454] = 0x5BD9, [14455] = 0x5C1F,
  [14456] = 0x5C33, [14457] = 0x5D71, [14458] = 0x5D63, [14459] = 0x5D4A,
  [14460] = 0x5D65, [14461] = 0x5D72, [14462] = 0x5D6C, [14463] = 0x5D5E,
  [14464] = 0x5D68, [14465] = 0x5D67, [14466] = 0x5D62, [14467] = 0x5DF0,
  [14468] = 0x5E4F, [14469] = 0x5E4E, [14470] = 0x5E4A, [14471] = 0x5E4D,
  [14472] = 0x5E4B, [14473] = 0x5EC5, [14474] = 0x5ECC, [14475] = 0x5EC6,
  [14476] = 0x5ECB, [14477] = 0x5EC7, [14478] = 0x5F40, [14479] = 0x5FAF,
  [14480] = 0x5FAD, [14481] = 0x60F7, [14482] = 0x6149, [14483] = 0x614A,
  [14484] = 0x612B, [14485] = 0x6145, [14486] = 0x6136, [14487] = 0x6132,
  [14488] = 0x612E, [14489] = 0x6146, [14490] = 0x612F, [14491] = 0x614F,
  [14492] = 0x6129, [14493] = 0x6140, [14494] = 0x6220, [14495] = 0x9168,
  [14496] = 0x6223, [14497] = 0x6225, [14498] = 0x6224, [14499] = 0x63C5,
  [14500] = 0x63F1, [14501] = 0x63EB, [14502] = 0x6410, [14503] = 0x6412,
  [14504] = 0x6409, [14505] = 0x6420, [14506] = 0x6424, [14507] = 0x6433,
  [14508] = 0x6443, [14509] = 0x641F, [14510] = 0x6415, [14511] = 0x6418,
  [14512] = 0x6439, [14513] = 0x6437, [14514] = 0x6422, [14515] = 0x6423,
  [14516] = 0x640C, [14517] = 0x6426, [14518] = 0x6430, [14519] = 0x6428,
  [14520] = 0x6441, [14521] = 0x6435, [14522] = 0x642F, [14523] = 0x640A,
  [14524] = 0x641A, [14525] = 0x6440, [14526] = 0x6425, [14527] = 0x6427,
  [14528] = 0x640B, [14529] = 0x63E7, [14530] = 0x641B, [14531] = 0x642E,
  [14532] = 0x6421, [14533] = 0x640E, [14534] = 0x656F, [14535] = 0x6592,
  [14536] = 0x65D3, [14537] = 0x6686, [14538] = 0x668C, [14539] = 0x6695,
  [14540] = 0x6690, [14541] = 0x668B, [14542] = 0x668A, [14543] = 0x6699,
  [14544] = 0x6694, [14545] = 0x6678, [14546] = 0x6720, [14547] = 0x6966,
  [14548] = 0x695F, [14549] = 0x6938, [14550] = 0x694E, [14551] = 0x6962,
  [14552] = 0x6971, [14553] = 0x693F, [14554] = 0x6945, [14555] = 0x696A,
  [14556] = 0x6939, [14557] = 0x6942, [14558] = 0x6957, [14559] = 0x6959,
  [14560] = 0x697A, [14561] = 0x6948, [14562] = 0x6949, [14563] = 0x6935,
  [14564] = 0x696C, [14565] = 0x6933, [14566] = 0x693D, [14567] = 0x6965,
  [14568] = 0x68F0, [14569] = 0x6978, [14570] = 0x6934, [14571] = 0x6969,
  [14572] = 0x6940, [14573] = 0x696F, [14574] = 0x6944, [14575] = 0x6976,
  [14576] = 0x6958, [14577] = 0x6941, [14578] = 0x6974, [14579] = 0x694C,
  [14580] = 0x693B, [14581] = 0x694B, [14582] = 0x6937, [14583] = 0x695C,
  [14584] = 0x694F, [14585] = 0x6951, [14586] = 0x6932, [14587] = 0x6952,
  [14588] = 0x692F, [14589] = 0x697B, [14590] = 0x693C, [14591] = 0x6B46,
  [14592] = 0x6B45, [14593] = 0x6B43, [14594] = 0x6B42, [14595] = 0x6B48,
  [14596] = 0x6B41, [14597] = 0x6B9B, [14598] = 0xFA0D, [14599] = 0x6BFB,
  [14600] = 0x6BFC, [14601] = 0x6BF9, [14602] = 0x6BF7, [14603] = 0x6BF8,
  [14604] = 0x6E9B, [14605] = 0x6ED6, [14606] = 0x6EC8, [14607] = 0x6E8F,
  [14608] = 0x6EC0, [14609] = 0x6E9F, [14610] = 0x6E93, [14611] = 0x6E94,
  [14612] = 0x6EA0, [14613] = 0x6EB1, [14614] = 0x6EB9, [14615] = 0x6EC6,
  [14616] = 0x6ED2, [14617] = 0x6EBD, [14618] = 0x6EC1, [14619] = 0x6E9E,
  [14620] = 0x6EC9, [14621] = 0x6EB7, [14622] = 0x6EB0, [14623] = 0x6ECD,
  [14624] = 0x6EA6, [14625] = 0x6ECF, [14626] = 0x6EB2, [14627] = 0x6EBE,
  [14628] = 0x6EC3, [14629] = 0x6EDC, [14630] = 0x6ED8, [14631] = 0x6E99,
  [14632] = 0x6E92, [14633] = 0x6E8E, [14634] = 0x6E8D, [14635] = 0x6EA4,
  [14636] = 0x6EA1, [14637] = 0x6EBF, [14638] = 0x6EB3, [14639] = 0x6ED0,
  [14640] = 0x6ECA, [14641] = 0x6E97, [14642] = 0x6EAE, [14643] = 0x6EA3,
  [14644] = 0x7147, [14645] = 0x7154, [14646] = 0x7152, [14647] = 0x7163,
  [14648] = 0x7160, [14649] = 0x7141, [14650] = 0x715D, [14651] = 0x7162,
  [14652] = 0x7172, [14653] = 0x7178, [14654] = 0x716A, [14655] = 0x7161,
  [14656] = 0x7142, [14657] = 0x7158, [14658] = 0x7143, [14659] = 0x714B,
  [14660] = 0x7170, [14661] = 0x715F, [14662] = 0x7150, [14663] = 0x7153,
  [14664] = 0x7144, [14665] = 0x714D, [14666] = 0x715A, [14667] = 0x724F,
  [14668] = 0x728D, [14669] = 0x728C, [14670] = 0x7291, [14671] = 0x7290,
  [14672] = 0x728E, [14673] = 0x733C, [14674] = 0x7342, [14675] = 0x733B,
  [14676] = 0x733A, [14677] = 0x7340, [14678] = 0x734A, [14679] = 0x7349,
  [14680] = 0x7444, [14681] = 0x744A, [14682] = 0x744B, [14683] = 0x7452,
  [14684] = 0x7451, [14685] = 0x7457, [14686] = 0x7440, [14687] = 0x744F,
  [14688] = 0x7450, [14689] = 0x744E, [14690] = 0x7442, [14691] = 0x7446,
  [14692] = 0x744D, [14693] = 0x7454, [14694] = 0x74E1, [14695] = 0x74FF,
  [14696] = 0x74FE, [14697] = 0x74FD, [14698] = 0x751D, [14699] = 0x7579,
  [14700] = 0x7577, [14701] = 0x6983, [14702] = 0x75EF, [14703] = 0x760F,
  [14704] = 0x7603, [14705] = 0x75F7, [14706] = 0x75FE, [14707] = 0x75FC,
  [14708] = 0x75F9, [14709] = 0x75F8, [14710] = 0x7610, [14711] = 0x75FB,
  [14712] = 0x75F6, [14713] = 0x75ED, [14714] = 0x75F5, [14715] = 0x75FD,
  [14716] = 0x7699, [14717] = 0x76B5, [14718] = 0x76DD, [14719] = 0x7755,
  [14720] = 0x775F, [14721] = 0x7760, [14722] = 0x7752, [14723] = 0x7756,
  [14724] = 0x775A, [14725] = 0x7769, [14726] = 0x7767, [14727] = 0x7754,
  [14728] = 0x7759, [14729] = 0x776D, [14730] = 0x77E0, [14731] = 0x7887,
  [14732] = 0x789A, [14733] = 0x7894, [14734] = 0x788F, [14735] = 0x7884,
  [14736] = 0x7895, [14737] = 0x7885, [14738] = 0x7886, [14739] = 0x78A1,
  [14740] = 0x7883, [14741] = 0x7879, [14742] = 0x7899, [14743] = 0x7880,
  [14744] = 0x7896, [14745] = 0x787B, [14746] = 0x797C, [14747] = 0x7982,
  [14748] = 0x797D, [14749] = 0x7979, [14750] = 0x7A11, [14751] = 0x7A18,
  [14752] = 0x7A19, [14753] = 0x7A12, [14754] = 0x7A17, [14755] = 0x7A15,
  [14756] = 0x7A22, [14757] = 0x7A13, [14758] = 0x7A1B, [14759] = 0x7A10,
  [14760] = 0x7AA3, [14761] = 0x7AA2, [14762] = 0x7A9E, [14763] = 0x7AEB,
  [14764] = 0x7B66, [14765] = 0x7B64, [14766] = 0x7B6D, [14767] = 0x7B74,
  [14768] = 0x7B69, [14769] = 0x7B72, [14770] = 0x7B65, [14771] = 0x7B73,
  [14772] = 0x7B71, [14773] = 0x7B70, [14774] = 0x7B61, [14775] = 0x7B78,
  [14776] = 0x7B76, [14777] = 0x7B63, [14778] = 0x7CB2, [14779] = 0x7CB4,
  [14780] = 0x7CAF, [14781] = 0x7D88, [14782] = 0x7D86, [14783] = 0x7D80,
  [14784] = 0x7D8D, [14785] = 0x7D7F, [14786] = 0x7D85, [14787] = 0x7D7A,
  [14788] = 0x7D8E, [14789] = 0x7D7B, [14790] = 0x7D83, [14791] = 0x7D7C,
  [14792] = 0x7D8C, [14793] = 0x7D94, [14794] = 0x7D84, [14795] = 0x7D7D,
  [14796] = 0x7D92, [14797] = 0x7F6D, [14798] = 0x7F6B, [14799] = 0x7F67,
  [14800] = 0x7F68, [14801] = 0x7F6C, [14802] = 0x7FA6, [14803] = 0x7FA5,
  [14804] = 0x7FA7, [14805] = 0x7FDB, [14806] = 0x7FDC, [14807] = 0x8021,
  [14808] = 0x8164, [14809] = 0x8160, [14810] = 0x8177, [14811] = 0x815C,
  [14812] = 0x8169, [14813] = 0x815B, [14814] = 0x8162, [14815] = 0x8172,
  [14816] = 0x6721, [14817] = 0x815E, [14818] = 0x8176, [14819] = 0x8167,
  [14820] = 0x816F, [14821] = 0x8144, [14822] = 0x8161, [14823] = 0x821D,
  [14824] = 0x8249, [14825] = 0x8244, [14826] = 0x8240, [14827] = 0x8242,
  [14828] = 0x8245, [14829] = 0x84F1, [14830] = 0x843F, [14831] = 0x8456,
  [14832] = 0x8476, [14833] = 0x8479, [14834] = 0x848F, [14835] = 0x848D,
  [14836] = 0x8465, [14837] = 0x8451, [14838] = 0x8440, [14839] = 0x8486,
  [14840] = 0x8467, [14841] = 0x8430, [14842] = 0x844D, [14843] = 0x847D,
  [14844] = 0x845A, [14845] = 0x8459, [14846] = 0x8474, [14847] = 0x8473,
  [14848] = 0x845D, [14849] = 0x8507, [14850] = 0x845E, [14851] = 0x8437,
  [14852] = 0x843A, [14853] = 0x8434, [14854] = 0x847A, [14855] = 0x8443,
  [14856] = 0x8478, [14857] = 0x8432, [14858] = 0x8445, [14859] = 0x8429,
  [14860] = 0x83D9, [14861] = 0x844B, [14862] = 0x842F, [14863] = 0x8442,
  [14864] = 0x842D, [14865] = 0x845F, [14866] = 0x8470, [14867] = 0x8439,
  [14868] = 0x844E, [14869] = 0x844C, [14870] = 0x8452, [14871] = 0x846F,
  [14872] = 0x84C5, [14873] = 0x848E, [14874] = 0x843B, [14875] = 0x8447,
  [14876] = 0x8436, [14877] = 0x8433, [14878] = 0x8468, [14879] = 0x847E,
  [14880] = 0x8444, [14881] = 0x842B, [14882] = 0x8460, [14883] = 0x8454,
  [14884] = 0x846E, [14885] = 0x8450, [14886] = 0x870B, [14887] = 0x8704,
  [14888] = 0x86F7, [14889] = 0x870C, [14890] = 0x86FA, [14891] = 0x86D6,
  [14892] = 0x86F5, [14893] = 0x874D, [14894] = 0x86F8, [14895] = 0x870E,
  [14896] = 0x8709, [14897] = 0x8701, [14898] = 0x86F6, [14899] = 0x870D,
  [14900] = 0x8705, [14901] = 0x88D6, [14902] = 0x88CB, [14903] = 0x88CD,
  [14904] = 0x88CE, [14905] = 0x88DE, [14906] = 0x88DB, [14907] = 0x88DA,
  [14908] = 0x88CC, [14909] = 0x88D0, [14910] = 0x8985, [14911] = 0x899B,
  [14912] = 0x89DF, [14913] = 0x89E5, [14914] = 0x89E4, [14915] = 0x89E1,
  [14916] = 0x89E0, [14917] = 0x89E2, [14918] = 0x89DC, [14919] = 0x89E6,
  [14920] = 0x8A76, [14921] = 0x8A86, [14922] = 0x8A7F, [14923] = 0x8A61,
  [14924] = 0x8A3F, [14925] = 0x8A77, [14926] = 0x8A82, [14927] = 0x8A84,
  [14928] = 0x8A75, [14929] = 0x8A83, [14930] = 0x8A81, [14931] = 0x8A74,
  [14932] = 0x8A7A, [14933] = 0x8C3C, [14934] = 0x8C4B, [14935] = 0x8C4A,
  [14936] = 0x8C65, [14937] = 0x8C64, [14938] = 0x8C66, [14939] = 0x8C86,
  [14940] = 0x8C84, [14941] = 0x8C85, [14942] = 0x8CCC, [14943] = 0x8D68,
  [14944] = 0x8D69, [14945] = 0x8D91, [14946] = 0x8D8C, [14947] = 0x8D8E,
  [14948] = 0x8D8F, [14949] = 0x8D8D, [14950] = 0x8D93, [14951] = 0x8D94,
  [14952] = 0x8D90, [14953] = 0x8D92, [14954] = 0x8DF0, [14955] = 0x8DE0,
  [14956] = 0x8DEC, [14957] = 0x8DF1, [14958] = 0x8DEE, [14959] = 0x8DD0,
  [14960] = 0x8DE9, [14961] = 0x8DE3, [14962] = 0x8DE2, [14963] = 0x8DE7,
  [14964] = 0x8DF2, [14965] = 0x8DEB, [14966] = 0x8DF4, [14967] = 0x8F06,
  [14968] = 0x8EFF, [14969] = 0x8F01, [14970] = 0x8F00, [14971] = 0x8F05,
  [14972] = 0x8F07, [14973] = 0x8F08, [14974] = 0x8F02, [14975] = 0x8F0B,
  [14976] = 0x9052, [14977] = 0x903F, [14978] = 0x9044, [14979] = 0x9049,
  [14980] = 0x903D, [14981] = 0x9110, [14982] = 0x910D, [14983] = 0x910F,
  [14984] = 0x9111, [14985] = 0x9116, [14986] = 0x9114, [14987] = 0x910B,
  [14988] = 0x910E, [14989] = 0x916E, [14990] = 0x916F, [14991] = 0x9248,
  [14992] = 0x9252, [14993] = 0x9230, [14994] = 0x923A, [14995] = 0x9266,
  [14996] = 0x9233, [14997] = 0x9265, [14998] = 0x925E, [14999] = 0x9283,
  [15000] = 0x922E, [15001] = 0x924A, [15002] = 0x9246, [15003] = 0x926D,
  [15004] = 0x926C, [15005] = 0x924F, [15006] = 0x9260, [15007] = 0x9267,
  [15008] = 0x926F, [15009] = 0x9236, [15010] = 0x9261, [15011] = 0x9270,
  [15012] = 0x9231, [15013] = 0x9254, [15014] = 0x9263, [15015] = 0x9250,
  [15016] = 0x9272, [15017] = 0x924E, [15018] = 0x9253, [15019] = 0x924C,
  [15020] = 0x9256, [15021] = 0x9232, [15022] = 0x959F, [15023] = 0x959C,
  [15024] = 0x959E, [15025] = 0x959B, [15026] = 0x9692, [15027] = 0x9693,
  [15028] = 0x9691, [15029] = 0x9697, [15030] = 0x96CE, [15031] = 0x96FA,
  [15032] = 0x96FD, [15033] = 0x96F8, [15034] = 0x96F5, [15035] = 0x9773,
  [15036] = 0x9777, [15037] = 0x9778, [15038] = 0x9772, [15039] = 0x980F,
  [15040] = 0x980D, [15041] = 0x980E, [15042] = 0x98AC, [15043] = 0x98F6,
  [15044] = 0x98F9, [15045] = 0x99AF, [15046] = 0x99B2, [15047] = 0x99B0,
  [15048] = 0x99B5, [15049] = 0x9AAD, [15050] = 0x9AAB, [15051] = 0x9B5B,
  [15052] = 0x9CEA, [15053] = 0x9CED, [15054] = 0x9CE7, [15055] = 0x9E80,
  [15056] = 0x9EFD, [15057] = 0x50E6, [15058] = 0x50D4, [15059] = 0x50D7,
  [15060] = 0x50E8, [15061] = 0x50F3, [15062] = 0x50DB, [15063] = 0x50EA,
  [15064] = 0x50DD, [15065] = 0x50E4, [15066] = 0x50D3, [15067] = 0x50EC,
  [15068] = 0x50F0, [15069] = 0x50EF, [15070] = 0x50E3, [15071] = 0x50E0,
  [15072] = 0x51D8, [15073] = 0x5280, [15074] = 0x5281, [15075] = 0x52E9,
  [15076] = 0x52EB, [15077] = 0x5330, [15078] = 0x53AC, [15079] = 0x5627,
  [15080] = 0x5615, [15081] = 0x560C, [15082] = 0x5612, [15083] = 0x55FC,
  [15084] = 0x560F, [15085] = 0x561C, [15086] = 0x5601, [15087] = 0x5613,
  [15088] = 0x5602, [15089] = 0x55FA, [15090] = 0x561D, [15091] = 0x5604,
  [15092] = 0x55FF, [15093] = 0x55F9, [15094] = 0x5889, [15095] = 0x587C,
  [15096] = 0x5890, [15097] = 0x5898, [15098] = 0x5886, [15099] = 0x5881,
  [15100] = 0x587F, [15101] = 0x5874, [15102] = 0x588B, [15103] = 0x587A,
  [15104] = 0x5887, [15105] = 0x5891, [15106] = 0x588E, [15107] = 0x5876,
  [15108] = 0x5882, [15109] = 0x5888, [15110] = 0x587B, [15111] = 0x5894,
  [15112] = 0x588F, [15113] = 0x58FE, [15114] = 0x596B, [15115] = 0x5ADC,
  [15116] = 0x5AEE, [15117] = 0x5AE5, [15118] = 0x5AD5, [15119] = 0x5AEA,
  [15120] = 0x5ADA, [15121] = 0x5AED, [15122] = 0x5AEB, [15123] = 0x5AF3,
  [15124] = 0x5AE2, [15125] = 0x5AE0, [15126] = 0x5ADB, [15127] = 0x5AEC,
  [15128] = 0x5ADE, [15129] = 0x5ADD, [15130] = 0x5AD9, [15131] = 0x5AE8,
  [15132] = 0x5ADF, [15133] = 0x5B77, [15134] = 0x5BE0, [15135] = 0x5BE3,
  [15136] = 0x5C63, [15137] = 0x5D82, [15138] = 0x5D80, [15139] = 0x5D7D,
  [15140] = 0x5D86, [15141] = 0x5D7A, [15142] = 0x5D81, [15143] = 0x5D77,
  [15144] = 0x5D8A, [15145] = 0x5D89, [15146] = 0x5D88, [15147] = 0x5D7E,
  [15148] = 0x5D7C, [15149] = 0x5D8D, [15150] = 0x5D79, [15151] = 0x5D7F,
  [15152] = 0x5E58, [15153] = 0x5E59, [15154] = 0x5E53, [15155] = 0x5ED8,
  [15156] = 0x5ED1, [15157] = 0x5ED7, [15158] = 0x5ECE, [15159] = 0x5EDC,
  [15160] = 0x5ED5, [15161] = 0x5ED9, [15162] = 0x5ED2, [15163] = 0x5ED4,
  [15164] = 0x5F44, [15165] = 0x5F43, [15166] = 0x5F6F, [15167] = 0x5FB6,
  [15168] = 0x612C, [15169] = 0x6128, [15170] = 0x6141, [15171] = 0x615E,
  [15172] = 0x6171, [15173] = 0x6173, [15174] = 0x6152, [15175] = 0x6153,
  [15176] = 0x6172, [15177] = 0x616C, [15178] = 0x6180, [15179] = 0x6174,
  [15180] = 0x6154, [15181] = 0x617A, [15182] = 0x615B, [15183] = 0x6165,
  [15184] = 0x613B, [15185] = 0x616A, [15186] = 0x6161, [15187] = 0x6156,
  [15188] = 0x6229, [15189] = 0x6227, [15190] = 0x622B, [15191] = 0x642B,
  [15192] = 0x644D, [15193] = 0x645B, [15194] = 0x645D, [15195] = 0x6474,
  [15196] = 0x6476, [15197] = 0x6472, [15198] = 0x6473, [15199] = 0x647D,
  [15200] = 0x6475, [15201] = 0x6466, [15202] = 0x64A6, [15203] = 0x644E,
  [15204] = 0x6482, [15205] = 0x645E, [15206] = 0x645C, [15207] = 0x644B,
  [15208] = 0x6453, [15209] = 0x6460, [15210] = 0x6450, [15211] = 0x647F,
  [15212] = 0x643F, [15213] = 0x646C, [15214] = 0x646B, [15215] = 0x6459,
  [15216] = 0x6465, [15217] = 0x6477, [15218] = 0x6573, [15219] = 0x65A0,
  [15220] = 0x66A1, [15221] = 0x66A0, [15222] = 0x669F, [15223] = 0x6705,
  [15224] = 0x6704, [15225] = 0x6722, [15226] = 0x69B1, [15227] = 0x69B6,
  [15228] = 0x69C9, [15229] = 0x69A0, [15230] = 0x69CE, [15231] = 0x6996,
  [15232] = 0x69B0, [15233] = 0x69AC, [15234] = 0x69BC, [15235] = 0x6991,
  [15236] = 0x6999, [15237] = 0x698E, [15238] = 0x69A7, [15239] = 0x698D,
  [15240] = 0x69A9, [15241] = 0x69BE, [15242] = 0x69AF, [15243] = 0x69BF,
  [15244] = 0x69C4, [15245] = 0x69BD, [15246] = 0x69A4, [15247] = 0x69D4,
  [15248] = 0x69B9, [15249] = 0x69CA, [15250] = 0x699A, [15251] = 0x69CF,
  [15252] = 0x69B3, [15253] = 0x6993, [15254] = 0x69AA, [15255] = 0x69A1,
  [15256] = 0x699E, [15257] = 0x69D9, [15258] = 0x6997, [15259] = 0x6990,
  [15260] = 0x69C2, [15261] = 0x69B5, [15262] = 0x69A5, [15263] = 0x69C6,
  [15264] = 0x6B4A, [15265] = 0x6B4D, [15266] = 0x6B4B, [15267] = 0x6B9E,
  [15268] = 0x6B9F, [15269] = 0x6BA0, [15270] = 0x6BC3, [15271] = 0x6BC4,
  [15272] = 0x6BFE, [15273] = 0x6ECE, [15274] = 0x6EF5, [15275] = 0x6EF1,
  [15276] = 0x6F03, [15277] = 0x6F25, [15278] = 0x6EF8, [15279] = 0x6F37,
  [15280] = 0x6EFB, [15281] = 0x6F2E, [15282] = 0x6F09, [15283] = 0x6F4E,
  [15284] = 0x6F19, [15285] = 0x6F1A, [15286] = 0x6F27, [15287] = 0x6F18,
  [15288] = 0x6F3B, [15289] = 0x6F12, [15290] = 0x6EED, [15291] = 0x6F0A,
  [15292] = 0x6F36, [15293] = 0x6F73, [15294] = 0x6EF9, [15295] = 0x6EEE,
  [15296] = 0x6F2D, [15297] = 0x6F40, [15298] = 0x6F30, [15299] = 0x6F3C,
  [15300] = 0x6F35, [15301] = 0x6EEB, [15302] = 0x6F07, [15303] = 0x6F0E,
  [15304] = 0x6F43, [15305] = 0x6F05, [15306] = 0x6EFD, [15307] = 0x6EF6,
  [15308] = 0x6F39, [15309] = 0x6F1C, [15310] = 0x6EFC, [15311] = 0x6F3A,
  [15312] = 0x6F1F, [15313] = 0x6F0D, [15314] = 0x6F1E, [15315] = 0x6F08,
  [15316] = 0x6F21, [15317] = 0x7187, [15318] = 0x7190, [15319] = 0x7189,
  [15320] = 0x7180, [15321] = 0x7185, [15322] = 0x7182, [15323] = 0x718F,
  [15324] = 0x717B, [15325] = 0x7186, [15326] = 0x7181, [15327] = 0x7197,
  [15328] = 0x7244, [15329] = 0x7253, [15330] = 0x7297, [15331] = 0x7295,
  [15332] = 0x7293, [15333] = 0x7343, [15334] = 0x734D, [15335] = 0x7351,
  [15336] = 0x734C, [15337] = 0x7462, [15338] = 0x7473, [15339] = 0x7471,
  [15340] = 0x7475, [15341] = 0x7472, [15342] = 0x7467, [15343] = 0x746E,
  [15344] = 0x7500, [15345] = 0x7502, [15346] = 0x7503, [15347] = 0x757D,
  [15348] = 0x7590, [15349] = 0x7616, [15350] = 0x7608, [15351] = 0x760C,
  [15352] = 0x7615, [15353] = 0x7611, [15354] = 0x760A, [15355] = 0x7614,
  [15356] = 0x76B8, [15357] = 0x7781, [15358] = 0x777C, [15359] = 0x7785,
  [15360] = 0x7782, [15361] = 0x776E, [15362] = 0x7780, [15363] = 0x776F,
  [15364] = 0x777E, [15365] = 0x7783, [15366] = 0x78B2, [15367] = 0x78AA,
  [15368] = 0x78B4, [15369] = 0x78AD, [15370] = 0x78A8, [15371] = 0x787E,
  [15372] = 0x78AB, [15373] = 0x789E, [15374] = 0x78A5, [15375] = 0x78A0,
  [15376] = 0x78AC, [15377] = 0x78A2, [15378] = 0x78A4, [15379] = 0x7998,
  [15380] = 0x798A, [15381] = 0x798B, [15382] = 0x7996, [15383] = 0x7995,
  [15384] = 0x7994, [15385] = 0x7993, [15386] = 0x7997, [15387] = 0x7988,
  [15388] = 0x7992, [15389] = 0x7990, [15390] = 0x7A2B, [15391] = 0x7A4A,
  [15392] = 0x7A30, [15393] = 0x7A2F, [15394] = 0x7A28, [15395] = 0x7A26,
  [15396] = 0x7AA8, [15397] = 0x7AAB, [15398] = 0x7AAC, [15399] = 0x7AEE,
  [15400] = 0x7B88, [15401] = 0x7B9C, [15402] = 0x7B8A, [15403] = 0x7B91,
  [15404] = 0x7B90, [15405] = 0x7B96, [15406] = 0x7B8D, [15407] = 0x7B8C,
  [15408] = 0x7B9B, [15409] = 0x7B8E, [15410] = 0x7B85, [15411] = 0x7B98,
  [15412] = 0x5284, [15413] = 0x7B99, [15414] = 0x7BA4, [15415] = 0x7B82,
  [15416] = 0x7CBB, [15417] = 0x7CBF, [15418] = 0x7CBC, [15419] = 0x7CBA,
  [15420] = 0x7DA7, [15421] = 0x7DB7, [15422] = 0x7DC2, [15423] = 0x7DA3,
  [15424] = 0x7DAA, [15425] = 0x7DC1, [15426] = 0x7DC0, [15427] = 0x7DC5,
  [15428] = 0x7D9D, [15429] = 0x7DCE, [15430] = 0x7DC4, [15431] = 0x7DC6,
  [15432] = 0x7DCB, [15433] = 0x7DCC, [15434] = 0x7DAF, [15435] = 0x7DB9,
  [15436] = 0x7D96, [15437] = 0x7DBC, [15438] = 0x7D9F, [15439] = 0x7DA6,
  [15440] = 0x7DAE, [15441] = 0x7DA9, [15442] = 0x7DA1, [15443] = 0x7DC9,
  [15444] = 0x7F73, [15445] = 0x7FE2, [15446] = 0x7FE3, [15447] = 0x7FE5,
  [15448] = 0x7FDE, [15449] = 0x8024, [15450] = 0x805D, [15451] = 0x805C,
  [15452] = 0x8189, [15453] = 0x8186, [15454] = 0x8183, [15455] = 0x8187,
  [15456] = 0x818D, [15457] = 0x818C, [15458] = 0x818B, [15459] = 0x8215,
  [15460] = 0x8497, [15461] = 0x84A4, [15462] = 0x84A1, [15463] = 0x849F,
  [15464] = 0x84BA, [15465] = 0x84CE, [15466] = 0x84C2, [15467] = 0x84AC,
  [15468] = 0x84AE, [15469] = 0x84AB, [15470] = 0x84B9, [15471] = 0x84B4,
  [15472] = 0x84C1, [15473] = 0x84CD, [15474] = 0x84AA, [15475] = 0x849A,
  [15476] = 0x84B1, [15477] = 0x84D0, [15478] = 0x849D, [15479] = 0x84A7,
  [15480] = 0x84BB, [15481] = 0x84A2, [15482] = 0x8494, [15483] = 0x84C7,
  [15484] = 0x84CC, [15485] = 0x849B, [15486] = 0x84A9, [15487] = 0x84AF,
  [15488] = 0x84A8, [15489] = 0x84D6, [15490] = 0x8498, [15491] = 0x84B6,
  [15492] = 0x84CF, [15493] = 0x84A0, [15494] = 0x84D7, [15495] = 0x84D4,
  [15496] = 0x84D2, [15497] = 0x84DB, [15498] = 0x84B0, [15499] = 0x8491,
  [15500] = 0x8661, [15501] = 0x8733, [15502] = 0x8723, [15503] = 0x8728,
  [15504] = 0x876B, [15505] = 0x8740, [15506] = 0x872E, [15507] = 0x871E,
  [15508] = 0x8721, [15509] = 0x8719, [15510] = 0x871B, [15511] = 0x8743,
  [15512] = 0x872C, [15513] = 0x8741, [15514] = 0x873E, [15515] = 0x8746,
  [15516] = 0x8720, [15517] = 0x8732, [15518] = 0x872A, [15519] = 0x872D,
  [15520] = 0x873C, [15521] = 0x8712, [15522] = 0x873A, [15523] = 0x8731,
  [15524] = 0x8735, [15525] = 0x8742, [15526] = 0x8726, [15527] = 0x8727,
  [15528] = 0x8738, [15529] = 0x8724, [15530] = 0x871A, [15531] = 0x8730,
  [15532] = 0x8711, [15533] = 0x88F7, [15534] = 0x88E7, [15535] = 0x88F1,
  [15536] = 0x88F2, [15537] = 0x88FA, [15538] = 0x88FE, [15539] = 0x88EE,
  [15540] = 0x88FC, [15541] = 0x88F6, [15542] = 0x88FB, [15543] = 0x88F0,
  [15544] = 0x88EC, [15545] = 0x88EB, [15546] = 0x899D, [15547] = 0x89A1,
  [15548] = 0x899F, [15549] = 0x899E, [15550] = 0x89E9, [15551] = 0x89EB,
  [15552] = 0x89E8, [15553] = 0x8AAB, [15554] = 0x8A99, [15555] = 0x8A8B,
  [15556] = 0x8A92, [15557] = 0x8A8F, [15558] = 0x8A96, [15559] = 0x8C3D,
  [15560] = 0x8C68, [15561] = 0x8C69, [15562] = 0x8CD5, [15563] = 0x8CCF,
  [15564] = 0x8CD7, [15565] = 0x8D96, [15566] = 0x8E09, [15567] = 0x8E02,
  [15568] = 0x8DFF, [15569] = 0x8E0D, [15570] = 0x8DFD, [15571] = 0x8E0A,
  [15572] = 0x8E03, [15573] = 0x8E07, [15574] = 0x8E06, [15575] = 0x8E05,
  [15576] = 0x8DFE, [15577] = 0x8E00, [15578] = 0x8E04, [15579] = 0x8F10,
  [15580] = 0x8F11, [15581] = 0x8F0E, [15582] = 0x8F0D, [15583] = 0x9123,
  [15584] = 0x911C, [15585] = 0x9120, [15586] = 0x9122, [15587] = 0x911F,
  [15588] = 0x911D, [15589] = 0x911A, [15590] = 0x9124, [15591] = 0x9121,
  [15592] = 0x911B, [15593] = 0x917A, [15594] = 0x9172, [15595] = 0x9179,
  [15596] = 0x9173, [15597] = 0x92A5, [15598] = 0x92A4, [15599] = 0x9276,
  [15600] = 0x929B, [15601] = 0x927A, [15602] = 0x92A0, [15603] = 0x9294,
  [15604] = 0x92AA, [15605] = 0x928D, [15606] = 0x92A6, [15607] = 0x929A,
  [15608] = 0x92AB, [15609] = 0x9279, [15610] = 0x9297, [15611] = 0x927F,
  [15612] = 0x92A3, [15613] = 0x92EE, [15614] = 0x928E, [15615] = 0x9282,
  [15616] = 0x9295, [15617] = 0x92A2, [15618] = 0x927D, [15619] = 0x9288,
  [15620] = 0x92A1, [15621] = 0x928A, [15622] = 0x9286, [15623] = 0x928C,
  [15624] = 0x9299, [15625] = 0x92A7, [15626] = 0x927E, [15627] = 0x9287,
  [15628] = 0x92A9, [15629] = 0x929D, [15630] = 0x928B, [15631] = 0x922D,
  [15632] = 0x969E, [15633] = 0x96A1, [15634] = 0x96FF, [15635] = 0x9758,
  [15636] = 0x977D, [15637] = 0x977A, [15638] = 0x977E, [15639] = 0x9783,
  [15640] = 0x9780, [15641] = 0x9782, [15642] = 0x977B, [15643] = 0x9784,
  [15644] = 0x9781, [15645] = 0x977F, [15646] = 0x97CE, [15647] = 0x97CD,
  [15648] = 0x9816, [15649] = 0x98AD, [15650] = 0x98AE, [15651] = 0x9902,
  [15652] = 0x9900, [15653] = 0x9907, [15654] = 0x999D, [15655] = 0x999C,
  [15656] = 0x99C3, [15657] = 0x99B9, [15658] = 0x99BB, [15659] = 0x99BA,
  [15660] = 0x99C2, [15661] = 0x99BD, [15662] = 0x99C7, [15663] = 0x9AB1,
  [15664] = 0x9AE3, [15665] = 0x9AE7, [15666] = 0x9B3E, [15667] = 0x9B3F,
  [15668] = 0x9B60, [15669] = 0x9B61, [15670] = 0x9B5F, [15671] = 0x9CF1,
  [15672] = 0x9CF2, [15673] = 0x9CF5, [15674] = 0x9EA7, [15675] = 0x50FF,
  [15676] = 0x5103, [15677] = 0x5130, [15678] = 0x50F8, [15679] = 0x5106,
  [15680] = 0x5107, [15681] = 0x50F6, [15682] = 0x50FE, [15683] = 0x510B,
  [15684] = 0x510C, [15685] = 0x50FD, [15686] = 0x510A, [15687] = 0x528B,
  [15688] = 0x528C, [15689] = 0x52F1, [15690] = 0x52EF, [15691] = 0x5648,
  [15692] = 0x5642, [15693] = 0x564C, [15694] = 0x5635, [15695] = 0x5641,
  [15696] = 0x564A, [15697] = 0x5649, [15698] = 0x5646, [15699] = 0x5658,
  [15700] = 0x565A, [15701] = 0x5640, [15702] = 0x5633, [15703] = 0x563D,
  [15704] = 0x562C, [15705] = 0x563E, [15706] = 0x5638, [15707] = 0x562A,
  [15708] = 0x563A, [15709] = 0x571A, [15710] = 0x58AB, [15711] = 0x589D,
  [15712] = 0x58B1, [15713] = 0x58A0, [15714] = 0x58A3, [15715] = 0x58AF,
  [15716] = 0x58AC, [15717] = 0x58A5, [15718] = 0x58A1, [15719] = 0x58FF,
  [15720] = 0x5AFF, [15721] = 0x5AF4, [15722] = 0x5AFD, [15723] = 0x5AF7,
  [15724] = 0x5AF6, [15725] = 0x5B03, [15726] = 0x5AF8, [15727] = 0x5B02,
  [15728] = 0x5AF9, [15729] = 0x5B01, [15730] = 0x5B07, [15731] = 0x5B05,
  [15732] = 0x5B0F, [15733] = 0x5C67, [15734] = 0x5D99, [15735] = 0x5D97,
  [15736] = 0x5D9F, [15737] = 0x5D92, [15738] = 0x5DA2, [15739] = 0x5D93,
  [15740] = 0x5D95, [15741] = 0x5DA0, [15742] = 0x5D9C, [15743] = 0x5DA1,
  [15744] = 0x5D9A, [15745] = 0x5D9E, [15746] = 0x5E69, [15747] = 0x5E5D,
  [15748] = 0x5E60, [15749] = 0x5E5C, [15750] = 0x7DF3, [15751] = 0x5EDB,
  [15752] = 0x5EDE, [15753] = 0x5EE1, [15754] = 0x5F49, [15755] = 0x5FB2,
  [15756] = 0x618B, [15757] = 0x6183, [15758] = 0x6179, [15759] = 0x61B1,
  [15760] = 0x61B0, [15761] = 0x61A2, [15762] = 0x6189, [15763] = 0x619B,
  [15764] = 0x6193, [15765] = 0x61AF, [15766] = 0x61AD, [15767] = 0x619F,
  [15768] = 0x6192, [15769] = 0x61AA, [15770] = 0x61A1, [15771] = 0x618D,
  [15772] = 0x6166, [15773] = 0x61B3, [15774] = 0x622D, [15775] = 0x646E,
  [15776] = 0x6470, [15777] = 0x6496, [15778] = 0x64A0, [15779] = 0x6485,
  [15780] = 0x6497, [15781] = 0x649C, [15782] = 0x648F, [15783] = 0x648B,
  [15784] = 0x648A, [15785] = 0x648C, [15786] = 0x64A3, [15787] = 0x649F,
  [15788] = 0x6468, [15789] = 0x64B1, [15790] = 0x6498, [15791] = 0x6576,
  [15792] = 0x657A, [15793] = 0x6579, [15794] = 0x657B, [15795] = 0x65B2,
  [15796] = 0x65B3, [15797] = 0x66B5, [15798] = 0x66B0, [15799] = 0x66A9,
  [15800] = 0x66B2, [15801] = 0x66B7, [15802] = 0x66AA, [15803] = 0x66AF,
  [15804] = 0x6A00, [15805] = 0x6A06, [15806] = 0x6A17, [15807] = 0x69E5,
  [15808] = 0x69F8, [15809] = 0x6A15, [15810] = 0x69F1, [15811] = 0x69E4,
  [15812] = 0x6A20, [15813] = 0x69FF, [15814] = 0x69EC, [15815] = 0x69E2,
  [15816] = 0x6A1B, [15817] = 0x6A1D, [15818] = 0x69FE, [15819] = 0x6A27,
  [15820] = 0x69F2, [15821] = 0x69EE, [15822] = 0x6A14, [15823] = 0x69F7,
  [15824] = 0x69E7, [15825] = 0x6A40, [15826] = 0x6A08, [15827] = 0x69E6,
  [15828] = 0x69FB, [15829] = 0x6A0D, [15830] = 0x69FC, [15831] = 0x69EB,
  [15832] = 0x6A09, [15833] = 0x6A04, [15834] = 0x6A18, [15835] = 0x6A25,
  [15836] = 0x6A0F, [15837] = 0x69F6, [15838] = 0x6A26, [15839] = 0x6A07,
  [15840] = 0x69F4, [15841] = 0x6A16, [15842] = 0x6B51, [15843] = 0x6BA5,
  [15844] = 0x6BA3, [15845] = 0x6BA2, [15846] = 0x6BA6, [15847] = 0x6C01,
  [15848] = 0x6C00, [15849] = 0x6BFF, [15850] = 0x6C02, [15851] = 0x6F41,
  [15852] = 0x6F26, [15853] = 0x6F7E, [15854] = 0x6F87, [15855] = 0x6FC6,
  [15856] = 0x6F92, [15857] = 0x6F8D, [15858] = 0x6F89, [15859] = 0x6F8C,
  [15860] = 0x6F62, [15861] = 0x6F4F, [15862] = 0x6F85, [15863] = 0x6F5A,
  [15864] = 0x6F96, [15865] = 0x6F76, [15866] = 0x6F6C, [15867] = 0x6F82,
  [15868] = 0x6F55, [15869] = 0x6F72, [15870] = 0x6F52, [15871] = 0x6F50,
  [15872] = 0x6F57, [15873] = 0x6F94, [15874] = 0x6F93, [15875] = 0x6F5D,
  [15876] = 0x6F00, [15877] = 0x6F61, [15878] = 0x6F6B, [15879] = 0x6F7D,
  [15880] = 0x6F67, [15881] = 0x6F90, [15882] = 0x6F53, [15883] = 0x6F8B,
  [15884] = 0x6F69, [15885] = 0x6F7F, [15886] = 0x6F95, [15887] = 0x6F63,
  [15888] = 0x6F77, [15889] = 0x6F6A, [15890] = 0x6F7B, [15891] = 0x71B2,
  [15892] = 0x71AF, [15893] = 0x719B, [15894] = 0x71B0, [15895] = 0x71A0,
  [15896] = 0x719A, [15897] = 0x71A9, [15898] = 0x71B5, [15899] = 0x719D,
  [15900] = 0x71A5, [15901] = 0x719E, [15902] = 0x71A4, [15903] = 0x71A1,
  [15904] = 0x71AA, [15905] = 0x719C, [15906] = 0x71A7, [15907] = 0x71B3,
  [15908] = 0x7298, [15909] = 0x729A, [15910] = 0x7358, [15911] = 0x7352,
  [15912] = 0x735E, [15913] = 0x735F, [15914] = 0x7360, [15915] = 0x735D,
  [15916] = 0x735B, [15917] = 0x7361, [15918] = 0x735A, [15919] = 0x7359,
  [15920] = 0x7362, [15921] = 0x7487, [15922] = 0x7489, [15923] = 0x748A,
  [15924] = 0x7486, [15925] = 0x7481, [15926] = 0x747D, [15927] = 0x7485,
  [15928] = 0x7488, [15929] = 0x747C, [15930] = 0x7479, [15931] = 0x7508,
  [15932] = 0x7507, [15933] = 0x757E, [15934] = 0x7625, [15935] = 0x761E,
  [15936] = 0x7619, [15937] = 0x761D, [15938] = 0x761C, [15939] = 0x7623,
  [15940] = 0x761A, [15941] = 0x7628, [15942] = 0x761B, [15943] = 0x769C,
  [15944] = 0x769D, [15945] = 0x769E, [15946] = 0x769B, [15947] = 0x778D,
  [15948] = 0x778F, [15949] = 0x7789, [15950] = 0x7788, [15951] = 0x78CD,
  [15952] = 0x78BB, [15953] = 0x78CF, [15954] = 0x78CC, [15955] = 0x78D1,
  [15956] = 0x78CE, [15957] = 0x78D4, [15958] = 0x78C8, [15959] = 0x78C3,
  [15960] = 0x78C4, [15961] = 0x78C9, [15962] = 0x799A, [15963] = 0x79A1,
  [15964] = 0x79A0, [15965] = 0x799C, [15966] = 0x79A2, [15967] = 0x799B,
  [15968] = 0x6B76, [15969] = 0x7A39, [15970] = 0x7AB2, [15971] = 0x7AB4,
  [15972] = 0x7AB3, [15973] = 0x7BB7, [15974] = 0x7BCB, [15975] = 0x7BBE,
  [15976] = 0x7BAC, [15977] = 0x7BCE, [15978] = 0x7BAF, [15979] = 0x7BB9,
  [15980] = 0x7BCA, [15981] = 0x7BB5, [15982] = 0x7CC5, [15983] = 0x7CC8,
  [15984] = 0x7CCC, [15985] = 0x7CCB, [15986] = 0x7DF7, [15987] = 0x7DDB,
  [15988] = 0x7DEA, [15989] = 0x7DE7, [15990] = 0x7DD7, [15991] = 0x7DE1,
  [15992] = 0x7E03, [15993] = 0x7DFA, [15994] = 0x7DE6, [15995] = 0x7DF6,
  [15996] = 0x7DF1, [15997] = 0x7DF0, [15998] = 0x7DEE, [15999] = 0x7DDF,
  [16000] = 0x7F76, [16001] = 0x7FAC, [16002] = 0x7FB0, [16003] = 0x7FAD,
  [16004] = 0x7FED, [16005] = 0x7FEB, [16006] = 0x7FEA, [16007] = 0x7FEC,
  [16008] = 0x7FE6, [16009] = 0x7FE8, [16010] = 0x8064, [16011] = 0x8067,
  [16012] = 0x81A3, [16013] = 0x819F, [16014] = 0x819E, [16015] = 0x8195,
  [16016] = 0x81A2, [16017] = 0x8199, [16018] = 0x8197, [16019] = 0x8216,
  [16020] = 0x824F, [16021] = 0x8253, [16022] = 0x8252, [16023] = 0x8250,
  [16024] = 0x824E, [16025] = 0x8251, [16026] = 0x8524, [16027] = 0x853B,
  [16028] = 0x850F, [16029] = 0x8500, [16030] = 0x8529, [16031] = 0x850E,
  [16032] = 0x8509, [16033] = 0x850D, [16034] = 0x851F, [16035] = 0x850A,
  [16036] = 0x8527, [16037] = 0x851C, [16038] = 0x84FB, [16039] = 0x852B,
  [16040] = 0x84FA, [16041] = 0x8508, [16042] = 0x850C, [16043] = 0x84F4,
  [16044] = 0x852A, [16045] = 0x84F2, [16046] = 0x8515, [16047] = 0x84F7,
  [16048] = 0x84EB, [16049] = 0x84F3, [16050] = 0x84FC, [16051] = 0x8512,
  [16052] = 0x84EA, [16053] = 0x84E9, [16054] = 0x8516, [16055] = 0x84FE,
  [16056] = 0x8528, [16057] = 0x851D, [16058] = 0x852E, [16059] = 0x8502,
  [16060] = 0x84FD, [16061] = 0x851E, [16062] = 0x84F6, [16063] = 0x8531,
  [16064] = 0x8526, [16065] = 0x84E7, [16066] = 0x84E8, [16067] = 0x84F0,
  [16068] = 0x84EF, [16069] = 0x84F9, [16070] = 0x8518, [16071] = 0x8520,
  [16072] = 0x8530, [16073] = 0x850B, [16074] = 0x8519, [16075] = 0x852F,
  [16076] = 0x8662, [16077] = 0x8756, [16078] = 0x8763, [16079] = 0x8764,
  [16080] = 0x8777, [16081] = 0x87E1, [16082] = 0x8773, [16083] = 0x8758,
  [16084] = 0x8754, [16085] = 0x875B, [16086] = 0x8752, [16087] = 0x8761,
  [16088] = 0x875A, [16089] = 0x8751, [16090] = 0x875E, [16091] = 0x876D,
  [16092] = 0x876A, [16093] = 0x8750, [16094] = 0x874E, [16095] = 0x875F,
  [16096] = 0x875D, [16097] = 0x876F, [16098] = 0x876C, [16099] = 0x877A,
  [16100] = 0x876E, [16101] = 0x875C, [16102] = 0x8765, [16103] = 0x874F,
  [16104] = 0x877B, [16105] = 0x8775, [16106] = 0x8762, [16107] = 0x8767,
  [16108] = 0x8769, [16109] = 0x885A, [16110] = 0x8905, [16111] = 0x890C,
  [16112] = 0x8914, [16113] = 0x890B, [16114] = 0x8917, [16115] = 0x8918,
  [16116] = 0x8919, [16117] = 0x8906, [16118] = 0x8916, [16119] = 0x8911,
  [16120] = 0x890E, [16121] = 0x8909, [16122] = 0x89A2, [16123] = 0x89A4,
  [16124] = 0x89A3, [16125] = 0x89ED, [16126] = 0x89F0, [16127] = 0x89EC,
  [16128] = 0x8ACF, [16129] = 0x8AC6, [16130] = 0x8AB8, [16131] = 0x8AD3,
  [16132] = 0x8AD1, [16133] = 0x8AD4, [16134] = 0x8AD5, [16135] = 0x8ABB,
  [16136] = 0x8AD7, [16137] = 0x8ABE, [16138] = 0x8AC0, [16139] = 0x8AC5,
  [16140] = 0x8AD8, [16141] = 0x8AC3, [16142] = 0x8ABA, [16143] = 0x8ABD,
  [16144] = 0x8AD9, [16145] = 0x8C3E, [16146] = 0x8C4D, [16147] = 0x8C8F,
  [16148] = 0x8CE5, [16149] = 0x8CDF, [16150] = 0x8CD9, [16151] = 0x8CE8,
  [16152] = 0x8CDA, [16153] = 0x8CDD, [16154] = 0x8CE7, [16155] = 0x8DA0,
  [16156] = 0x8D9C, [16157] = 0x8DA1, [16158] = 0x8D9B, [16159] = 0x8E20,
  [16160] = 0x8E23, [16161] = 0x8E25, [16162] = 0x8E24, [16163] = 0x8E2E,
  [16164] = 0x8E15, [16165] = 0x8E1B, [16166] = 0x8E16, [16167] = 0x8E11,
  [16168] = 0x8E19, [16169] = 0x8E26, [16170] = 0x8E27, [16171] = 0x8E14,
  [16172] = 0x8E12, [16173] = 0x8E18, [16174] = 0x8E13, [16175] = 0x8E1C,
  [16176] = 0x8E17, [16177] = 0x8E1A, [16178] = 0x8F2C, [16179] = 0x8F24,
  [16180] = 0x8F18, [16181] = 0x8F1A, [16182] = 0x8F20, [16183] = 0x8F23,
  [16184] = 0x8F16, [16185] = 0x8F17, [16186] = 0x9073, [16187] = 0x9070,
  [16188] = 0x906F, [16189] = 0x9067, [16190] = 0x906B, [16191] = 0x912F,
  [16192] = 0x912B, [16193] = 0x9129, [16194] = 0x912A, [16195] = 0x9132,
  [16196] = 0x9126, [16197] = 0x912E, [16198] = 0x9185, [16199] = 0x9186,
  [16200] = 0x918A, [16201] = 0x9181, [16202] = 0x9182, [16203] = 0x9184,
  [16204] = 0x9180, [16205] = 0x92D0, [16206] = 0x92C3, [16207] = 0x92C4,
  [16208] = 0x92C0, [16209] = 0x92D9, [16210] = 0x92B6, [16211] = 0x92CF,
  [16212] = 0x92F1, [16213] = 0x92DF, [16214] = 0x92D8, [16215] = 0x92E9,
  [16216] = 0x92D7, [16217] = 0x92DD, [16218] = 0x92CC, [16219] = 0x92EF,
  [16220] = 0x92C2, [16221] = 0x92E8, [16222] = 0x92CA, [16223] = 0x92C8,
  [16224] = 0x92CE, [16225] = 0x92E6, [16226] = 0x92CD, [16227] = 0x92D5,
  [16228] = 0x92C9, [16229] = 0x92E0, [16230] = 0x92DE, [16231] = 0x92E7,
  [16232] = 0x92D1, [16233] = 0x92D3, [16234] = 0x92B5, [16235] = 0x92E1,
  [16236] = 0x92C6, [16237] = 0x92B4, [16238] = 0x957C, [16239] = 0x95AC,
  [16240] = 0x95AB, [16241] = 0x95AE, [16242] = 0x95B0, [16243] = 0x96A4,
  [16244] = 0x96A2, [16245] = 0x96D3, [16246] = 0x9705, [16247] = 0x9708,
  [16248] = 0x9702, [16249] = 0x975A, [16250] = 0x978A, [16251] = 0x978E,
  [16252] = 0x9788, [16253] = 0x97D0, [16254] = 0x97CF, [16255] = 0x981E,
  [16256] = 0x981D, [16257] = 0x9826, [16258] = 0x9829, [16259] = 0x9828,
  [16260] = 0x9820, [16261] = 0x981B, [16262] = 0x9827, [16263] = 0x98B2,
  [16264] = 0x9908, [16265] = 0x98FA, [16266] = 0x9911, [16267] = 0x9914,
  [16268] = 0x9916, [16269] = 0x9917, [16270] = 0x9915, [16271] = 0x99DC,
  [16272] = 0x99CD, [16273] = 0x99CF, [16274] = 0x99D3, [16275] = 0x99D4,
  [16276] = 0x99CE, [16277] = 0x99C9, [16278] = 0x99D6, [16279] = 0x99D8,
  [16280] = 0x99CB, [16281] = 0x99D7, [16282] = 0x99CC, [16283] = 0x9AB3,
  [16284] = 0x9AEC, [16285] = 0x9AEB, [16286] = 0x9AF3, [16287] = 0x9AF2,
  [16288] = 0x9AF1, [16289] = 0x9B46, [16290] = 0x9B43, [16291] = 0x9B67,
  [16292] = 0x9B74, [16293] = 0x9B71, [16294] = 0x9B66, [16295] = 0x9B76,
  [16296] = 0x9B75, [16297] = 0x9B70, [16298] = 0x9B68, [16299] = 0x9B64,
  [16300] = 0x9B6C, [16301] = 0x9CFC, [16302] = 0x9CFA, [16303] = 0x9CFD,
  [16304] = 0x9CFF, [16305] = 0x9CF7, [16306] = 0x9D07, [16307] = 0x9D00,
  [16308] = 0x9CF9, [16309] = 0x9CFB, [16310] = 0x9D08, [16311] = 0x9D05,
  [16312] = 0x9D04, [16313] = 0x9E83, [16314] = 0x9ED3, [16315] = 0x9F0F,
  [16316] = 0x9F10, [16317] = 0x511C, [16318] = 0x5113, [16319] = 0x5117,
  [16320] = 0x511A, [16321] = 0x5111, [16322] = 0x51DE, [16323] = 0x5334,
  [16324] = 0x53E1, [16325] = 0x5670, [16326] = 0x5660, [16327] = 0x566E,
  [16328] = 0x5673, [16329] = 0x5666, [16330] = 0x5663, [16331] = 0x566D,
  [16332] = 0x5672, [16333] = 0x565E, [16334] = 0x5677, [16335] = 0x571C,
  [16336] = 0x571B, [16337] = 0x58C8, [16338] = 0x58BD, [16339] = 0x58C9,
  [16340] = 0x58BF, [16341] = 0x58BA, [16342] = 0x58C2, [16343] = 0x58BC,
  [16344] = 0x58C6, [16345] = 0x5B17, [16346] = 0x5B19, [16347] = 0x5B1B,
  [16348] = 0x5B21, [16349] = 0x5B14, [16350] = 0x5B13, [16351] = 0x5B10,
  [16352] = 0x5B16, [16353] = 0x5B28, [16354] = 0x5B1A, [16355] = 0x5B20,
  [16356] = 0x5B1E, [16357] = 0x5BEF, [16358] = 0x5DAC, [16359] = 0x5DB1,
  [16360] = 0x5DA9, [16361] = 0x5DA7, [16362] = 0x5DB5, [16363] = 0x5DB0,
  [16364] = 0x5DAE, [16365] = 0x5DAA, [16366] = 0x5DA8, [16367] = 0x5DB2,
  [16368] = 0x5DAD, [16369] = 0x5DAF, [16370] = 0x5DB4, [16371] = 0x5E67,
  [16372] = 0x5E68, [16373] = 0x5E66, [16374] = 0x5E6F, [16375] = 0x5EE9,
  [16376] = 0x5EE7, [16377] = 0x5EE6, [16378] = 0x5EE8, [16379] = 0x5EE5,
  [16380] = 0x5F4B, [16381] = 0x5FBC, [16382] = 0x619D, [16383] = 0x61A8,
  [16384] = 0x6196, [16385] = 0x61C5, [16386] = 0x61B4, [16387] = 0x61C6,
  [16388] = 0x61C1, [16389] = 0x61CC, [16390] = 0x61BA, [16391] = 0x61BF,
  [16392] = 0x61B8, [16393] = 0x618C, [16394] = 0x64D7, [16395] = 0x64D6,
  [16396] = 0x64D0, [16397] = 0x64CF, [16398] = 0x64C9, [16399] = 0x64BD,
  [16400] = 0x6489, [16401] = 0x64C3, [16402] = 0x64DB, [16403] = 0x64F3,
  [16404] = 0x64D9, [16405] = 0x6533, [16406] = 0x657F, [16407] = 0x657C,
  [16408] = 0x65A2, [16409] = 0x66C8, [16410] = 0x66BE, [16411] = 0x66C0,
  [16412] = 0x66CA, [16413] = 0x66CB, [16414] = 0x66CF, [16415] = 0x66BD,
  [16416] = 0x66BB, [16417] = 0x66BA, [16418] = 0x66CC, [16419] = 0x6723,
  [16420] = 0x6A34, [16421] = 0x6A66, [16422] = 0x6A49, [16423] = 0x6A67,
  [16424] = 0x6A32, [16425] = 0x6A68, [16426] = 0x6A3E, [16427] = 0x6A5D,
  [16428] = 0x6A6D, [16429] = 0x6A76, [16430] = 0x6A5B, [16431] = 0x6A51,
  [16432] = 0x6A28, [16433] = 0x6A5A, [16434] = 0x6A3B, [16435] = 0x6A3F,
  [16436] = 0x6A41, [16437] = 0x6A6A, [16438] = 0x6A64, [16439] = 0x6A50,
  [16440] = 0x6A4F, [16441] = 0x6A54, [16442] = 0x6A6F, [16443] = 0x6A69,
  [16444] = 0x6A60, [16445] = 0x6A3C, [16446] = 0x6A5E, [16447] = 0x6A56,
  [16448] = 0x6A55, [16449] = 0x6A4D, [16450] = 0x6A4E, [16451] = 0x6A46,
  [16452] = 0x6B55, [16453] = 0x6B54, [16454] = 0x6B56, [16455] = 0x6BA7,
  [16456] = 0x6BAA, [16457] = 0x6BAB, [16458] = 0x6BC8, [16459] = 0x6BC7,
  [16460] = 0x6C04, [16461] = 0x6C03, [16462] = 0x6C06, [16463] = 0x6FAD,
  [16464] = 0x6FCB, [16465] = 0x6FA3, [16466] = 0x6FC7, [16467] = 0x6FBC,
  [16468] = 0x6FCE, [16469] = 0x6FC8, [16470] = 0x6F5E, [16471] = 0x6FC4,
  [16472] = 0x6FBD, [16473] = 0x6F9E, [16474] = 0x6FCA, [16475] = 0x6FA8,
  [16476] = 0x7004, [16477] = 0x6FA5, [16478] = 0x6FAE, [16479] = 0x6FBA,
  [16480] = 0x6FAC, [16481] = 0x6FAA, [16482] = 0x6FCF, [16483] = 0x6FBF,
  [16484] = 0x6FB8, [16485] = 0x6FA2, [16486] = 0x6FC9, [16487] = 0x6FAB,
  [16488] = 0x6FCD, [16489] = 0x6FAF, [16490] = 0x6FB2, [16491] = 0x6FB0,
  [16492] = 0x71C5, [16493] = 0x71C2, [16494] = 0x71BF, [16495] = 0x71B8,
  [16496] = 0x71D6, [16497] = 0x71C0, [16498] = 0x71C1, [16499] = 0x71CB,
  [16500] = 0x71D4, [16501] = 0x71CA, [16502] = 0x71C7, [16503] = 0x71CF,
  [16504] = 0x71BD, [16505] = 0x71D8, [16506] = 0x71BC, [16507] = 0x71C6,
  [16508] = 0x71DA, [16509] = 0x71DB, [16510] = 0x729D, [16511] = 0x729E,
  [16512] = 0x7369, [16513] = 0x7366, [16514] = 0x7367, [16515] = 0x736C,
  [16516] = 0x7365, [16517] = 0x736B, [16518] = 0x736A, [16519] = 0x747F,
  [16520] = 0x749A, [16521] = 0x74A0, [16522] = 0x7494, [16523] = 0x7492,
  [16524] = 0x7495, [16525] = 0x74A1, [16526] = 0x750B, [16527] = 0x7580,
  [16528] = 0x762F, [16529] = 0x762D, [16530] = 0x7631, [16531] = 0x763D,
  [16532] = 0x7633, [16533] = 0x763C, [16534] = 0x7635, [16535] = 0x7632,
  [16536] = 0x7630, [16537] = 0x76BB, [16538] = 0x76E6, [16539] = 0x779A,
  [16540] = 0x779D, [16541] = 0x77A1, [16542] = 0x779C, [16543] = 0x779B,
  [16544] = 0x77A2, [16545] = 0x77A3, [16546] = 0x7795, [16547] = 0x7799,
  [16548] = 0x7797, [16549] = 0x78DD, [16550] = 0x78E9, [16551] = 0x78E5,
  [16552] = 0x78EA, [16553] = 0x78DE, [16554] = 0x78E3, [16555] = 0x78DB,
  [16556] = 0x78E1, [16557] = 0x78E2, [16558] = 0x78ED, [16559] = 0x78DF,
  [16560] = 0x78E0, [16561] = 0x79A4, [16562] = 0x7A44, [16563] = 0x7A48,
  [16564] = 0x7A47, [16565] = 0x7AB6, [16566] = 0x7AB8, [16567] = 0x7AB5,
  [16568] = 0x7AB1, [16569] = 0x7AB7, [16570] = 0x7BDE, [16571] = 0x7BE3,
  [16572] = 0x7BE7, [16573] = 0x7BDD, [16574] = 0x7BD5, [16575] = 0x7BE5,
  [16576] = 0x7BDA, [16577] = 0x7BE8, [16578] = 0x7BF9, [16579] = 0x7BD4,
  [16580] = 0x7BEA, [16581] = 0x7BE2, [16582] = 0x7BDC, [16583] = 0x7BEB,
  [16584] = 0x7BD8, [16585] = 0x7BDF, [16586] = 0x7CD2, [16587] = 0x7CD4,
  [16588] = 0x7CD7, [16589] = 0x7CD0, [16590] = 0x7CD1, [16591] = 0x7E12,
  [16592] = 0x7E21, [16593] = 0x7E17, [16594] = 0x7E0C, [16595] = 0x7E1F,
  [16596] = 0x7E20, [16597] = 0x7E13, [16598] = 0x7E0E, [16599] = 0x7E1C,
  [16600] = 0x7E15, [16601] = 0x7E1A, [16602] = 0x7E22, [16603] = 0x7E0B,
  [16604] = 0x7E0F, [16605] = 0x7E16, [16606] = 0x7E0D, [16607] = 0x7E14,
  [16608] = 0x7E25, [16609] = 0x7E24, [16610] = 0x7F43, [16611] = 0x7F7B,
  [16612] = 0x7F7C, [16613] = 0x7F7A, [16614] = 0x7FB1, [16615] = 0x7FEF,
  [16616] = 0x802A, [16617] = 0x8029, [16618] = 0x806C, [16619] = 0x81B1,
  [16620] = 0x81A6, [16621] = 0x81AE, [16622] = 0x81B9, [16623] = 0x81B5,
  [16624] = 0x81AB, [16625] = 0x81B0, [16626] = 0x81AC, [16627] = 0x81B4,
  [16628] = 0x81B2, [16629] = 0x81B7, [16630] = 0x81A7, [16631] = 0x81F2,
  [16632] = 0x8255, [16633] = 0x8256, [16634] = 0x8257, [16635] = 0x8556,
  [16636] = 0x8545, [16637] = 0x856B, [16638] = 0x854D, [16639] = 0x8553,
  [16640] = 0x8561, [16641] = 0x8558, [16642] = 0x8540, [16643] = 0x8546,
  [16644] = 0x8564, [16645] = 0x8541, [16646] = 0x8562, [16647] = 0x8544,
  [16648] = 0x8551, [16649] = 0x8547, [16650] = 0x8563, [16651] = 0x853E,
  [16652] = 0x855B, [16653] = 0x8571, [16654] = 0x854E, [16655] = 0x856E,
  [16656] = 0x8575, [16657] = 0x8555, [16658] = 0x8567, [16659] = 0x8560,
  [16660] = 0x858C, [16661] = 0x8566, [16662] = 0x855D, [16663] = 0x8554,
  [16664] = 0x8565, [16665] = 0x856C, [16666] = 0x8663, [16667] = 0x8665,
  [16668] = 0x8664, [16669] = 0x879B, [16670] = 0x878F, [16671] = 0x8797,
  [16672] = 0x8793, [16673] = 0x8792, [16674] = 0x8788, [16675] = 0x8781,
  [16676] = 0x8796, [16677] = 0x8798, [16678] = 0x8779, [16679] = 0x8787,
  [16680] = 0x87A3, [16681] = 0x8785, [16682] = 0x8790, [16683] = 0x8791,
  [16684] = 0x879D, [16685] = 0x8784, [16686] = 0x8794, [16687] = 0x879C,
  [16688] = 0x879A, [16689] = 0x8789, [16690] = 0x891E, [16691] = 0x8926,
  [16692] = 0x8930, [16693] = 0x892D, [16694] = 0x892E, [16695] = 0x8927,
  [16696] = 0x8931, [16697] = 0x8922, [16698] = 0x8929, [16699] = 0x8923,
  [16700] = 0x892F, [16701] = 0x892C, [16702] = 0x891F, [16703] = 0x89F1,
  [16704] = 0x8AE0, [16705] = 0x8AE2, [16706] = 0x8AF2, [16707] = 0x8AF4,
  [16708] = 0x8AF5, [16709] = 0x8ADD, [16710] = 0x8B14, [16711] = 0x8AE4,
  [16712] = 0x8ADF, [16713] = 0x8AF0, [16714] = 0x8AC8, [16715] = 0x8ADE,
  [16716] = 0x8AE1, [16717] = 0x8AE8, [16718] = 0x8AFF, [16719] = 0x8AEF,
  [16720] = 0x8AFB, [16721] = 0x8C91, [16722] = 0x8C92, [16723] = 0x8C90,
  [16724] = 0x8CF5, [16725] = 0x8CEE, [16726] = 0x8CF1, [16727] = 0x8CF0,
  [16728] = 0x8CF3, [16729] = 0x8D6C, [16730] = 0x8D6E, [16731] = 0x8DA5,
  [16732] = 0x8DA7, [16733] = 0x8E33, [16734] = 0x8E3E, [16735] = 0x8E38,
  [16736] = 0x8E40, [16737] = 0x8E45, [16738] = 0x8E36, [16739] = 0x8E3C,
  [16740] = 0x8E3D, [16741] = 0x8E41, [16742] = 0x8E30, [16743] = 0x8E3F,
  [16744] = 0x8EBD, [16745] = 0x8F36, [16746] = 0x8F2E, [16747] = 0x8F35,
  [16748] = 0x8F32, [16749] = 0x8F39, [16750] = 0x8F37, [16751] = 0x8F34,
  [16752] = 0x9076, [16753] = 0x9079, [16754] = 0x907B, [16755] = 0x9086,
  [16756] = 0x90FA, [16757] = 0x9133, [16758] = 0x9135, [16759] = 0x9136,
  [16760] = 0x9193, [16761] = 0x9190, [16762] = 0x9191, [16763] = 0x918D,
  [16764] = 0x918F, [16765] = 0x9327, [16766] = 0x931E, [16767] = 0x9308,
  [16768] = 0x931F, [16769] = 0x9306, [16770] = 0x930F, [16771] = 0x937A,
  [16772] = 0x9338, [16773] = 0x933C, [16774] = 0x931B, [16775] = 0x9323,
  [16776] = 0x9312, [16777] = 0x9301, [16778] = 0x9346, [16779] = 0x932D,
  [16780] = 0x930E, [16781] = 0x930D, [16782] = 0x92CB, [16783] = 0x931D,
  [16784] = 0x92FA, [16785] = 0x9325, [16786] = 0x9313, [16787] = 0x92F9,
  [16788] = 0x92F7, [16789] = 0x9334, [16790] = 0x9302, [16791] = 0x9324,
  [16792] = 0x92FF, [16793] = 0x9329, [16794] = 0x9339, [16795] = 0x9335,
  [16796] = 0x932A, [16797] = 0x9314, [16798] = 0x930C, [16799] = 0x930B,
  [16800] = 0x92FE, [16801] = 0x9309, [16802] = 0x9300, [16803] = 0x92FB,
  [16804] = 0x9316, [16805] = 0x95BC, [16806] = 0x95CD, [16807] = 0x95BE,
  [16808] = 0x95B9, [16809] = 0x95BA, [16810] = 0x95B6, [16811] = 0x95BF,
  [16812] = 0x95B5, [16813] = 0x95BD, [16814] = 0x96A9, [16815] = 0x96D4,
  [16816] = 0x970B, [16817] = 0x9712, [16818] = 0x9710, [16819] = 0x9799,
  [16820] = 0x9797, [16821] = 0x9794, [16822] = 0x97F0, [16823] = 0x97F8,
  [16824] = 0x9835, [16825] = 0x982F, [16826] = 0x9832, [16827] = 0x9924,
  [16828] = 0x991F, [16829] = 0x9927, [16830] = 0x9929, [16831] = 0x999E,
  [16832] = 0x99EE, [16833] = 0x99EC, [16834] = 0x99E5, [16835] = 0x99E4,
  [16836] = 0x99F0, [16837] = 0x99E3, [16838] = 0x99EA, [16839] = 0x99E9,
  [16840] = 0x99E7, [16841] = 0x9AB9, [16842] = 0x9ABF, [16843] = 0x9AB4,
  [16844] = 0x9ABB, [16845] = 0x9AF6, [16846] = 0x9AFA, [16847] = 0x9AF9,
  [16848] = 0x9AF7, [16849] = 0x9B33, [16850] = 0x9B80, [16851] = 0x9B85,
  [16852] = 0x9B87, [16853] = 0x9B7C, [16854] = 0x9B7E, [16855] = 0x9B7B,
  [16856] = 0x9B82, [16857] = 0x9B93, [16858] = 0x9B92, [16859] = 0x9B90,
  [16860] = 0x9B7A, [16861] = 0x9B95, [16862] = 0x9B7D, [16863] = 0x9B88,
  [16864] = 0x9D25, [16865] = 0x9D17, [16866] = 0x9D20, [16867] = 0x9D1E,
  [16868] = 0x9D14, [16869] = 0x9D29, [16870] = 0x9D1D, [16871] = 0x9D18,
  [16872] = 0x9D22, [16873] = 0x9D10, [16874] = 0x9D19, [16875] = 0x9D1F,
  [16876] = 0x9E88, [16877] = 0x9E86, [16878] = 0x9E87, [16879] = 0x9EAE,
  [16880] = 0x9EAD, [16881] = 0x9ED5, [16882] = 0x9ED6, [16883] = 0x9EFA,
  [16884] = 0x9F12, [16885] = 0x9F3D, [16886] = 0x5126, [16887] = 0x5125,
  [16888] = 0x5122, [16889] = 0x5124, [16890] = 0x5120, [16891] = 0x5129,
  [16892] = 0x52F4, [16893] = 0x5693, [16894] = 0x568C, [16895] = 0x568D,
  [16896] = 0x5686, [16897] = 0x5684, [16898] = 0x5683, [16899] = 0x567E,
  [16900] = 0x5682, [16901] = 0x567F, [16902] = 0x5681, [16903] = 0x58D6,
  [16904] = 0x58D4, [16905] = 0x58CF, [16906] = 0x58D2, [16907] = 0x5B2D,
  [16908] = 0x5B25, [16909] = 0x5B32, [16910] = 0x5B23, [16911] = 0x5B2C,
  [16912] = 0x5B27, [16913] = 0x5B26, [16914] = 0x5B2F, [16915] = 0x5B2E,
  [16916] = 0x5B7B, [16917] = 0x5BF1, [16918] = 0x5BF2, [16919] = 0x5DB7,
  [16920] = 0x5E6C, [16921] = 0x5E6A, [16922] = 0x5FBE, [16923] = 0x5FBB,
  [16924] = 0x61C3, [16925] = 0x61B5, [16926] = 0x61BC, [16927] = 0x61E7,
  [16928] = 0x61E0, [16929] = 0x61E5, [16930] = 0x61E4, [16931] = 0x61E8,
  [16932] = 0x61DE, [16933] = 0x64EF, [16934] = 0x64E9, [16935] = 0x64E3,
  [16936] = 0x64EB, [16937] = 0x64E4, [16938] = 0x64E8, [16939] = 0x6581,
  [16940] = 0x6580, [16941] = 0x65B6, [16942] = 0x65DA, [16943] = 0x66D2,
  [16944] = 0x6A8D, [16945] = 0x6A96, [16946] = 0x6A81, [16947] = 0x6AA5,
  [16948] = 0x6A89, [16949] = 0x6A9F, [16950] = 0x6A9B, [16951] = 0x6AA1,
  [16952] = 0x6A9E, [16953] = 0x6A87, [16954] = 0x6A93, [16955] = 0x6A8E,
  [16956] = 0x6A95, [16957] = 0x6A83, [16958] = 0x6AA8, [16959] = 0x6AA4,
  [16960] = 0x6A91, [16961] = 0x6A7F, [16962] = 0x6AA6, [16963] = 0x6A9A,
  [16964] = 0x6A85, [16965] = 0x6A8C, [16966] = 0x6A92, [16967] = 0x6B5B,
  [16968] = 0x6BAD, [16969] = 0x6C09, [16970] = 0x6FCC, [16971] = 0x6FA9,
  [16972] = 0x6FF4, [16973] = 0x6FD4, [16974] = 0x6FE3, [16975] = 0x6FDC,
  [16976] = 0x6FED, [16977] = 0x6FE7, [16978] = 0x6FE6, [16979] = 0x6FDE,
  [16980] = 0x6FF2, [16981] = 0x6FDD, [16982] = 0x6FE2, [16983] = 0x6FE8,
  [16984] = 0x71E1, [16985] = 0x71F1, [16986] = 0x71E8, [16987] = 0x71F2,
  [16988] = 0x71E4, [16989] = 0x71F0, [16990] = 0x71E2, [16991] = 0x7373,
  [16992] = 0x736E, [16993] = 0x736F, [16994] = 0x7497, [16995] = 0x74B2,
  [16996] = 0x74AB, [16997] = 0x7490, [16998] = 0x74AA, [16999] = 0x74AD,
  [17000] = 0x74B1, [17001] = 0x74A5, [17002] = 0x74AF, [17003] = 0x7510,
  [17004] = 0x7511, [17005] = 0x7512, [17006] = 0x750F, [17007] = 0x7584,
  [17008] = 0x7643, [17009] = 0x7648, [17010] = 0x7649, [17011] = 0x7647,
  [17012] = 0x76A4, [17013] = 0x76E9, [17014] = 0x77B5, [17015] = 0x77AB,
  [17016] = 0x77B2, [17017] = 0x77B7, [17018] = 0x77B6, [17019] = 0x77B4,
  [17020] = 0x77B1, [17021] = 0x77A8, [17022] = 0x77F0, [17023] = 0x78F3,
  [17024] = 0x78FD, [17025] = 0x7902, [17026] = 0x78FB, [17027] = 0x78FC,
  [17028] = 0x78F2, [17029] = 0x7905, [17030] = 0x78F9, [17031] = 0x78FE,
  [17032] = 0x7904, [17033] = 0x79AB, [17034] = 0x79A8, [17035] = 0x7A5C,
  [17036] = 0x7A5B, [17037] = 0x7A56, [17038] = 0x7A58, [17039] = 0x7A54,
  [17040] = 0x7A5A, [17041] = 0x7ABE, [17042] = 0x7AC0, [17043] = 0x7AC1,
  [17044] = 0x7C05, [17045] = 0x7C0F, [17046] = 0x7BF2, [17047] = 0x7C00,
  [17048] = 0x7BFF, [17049] = 0x7BFB, [17050] = 0x7C0E, [17051] = 0x7BF4,
  [17052] = 0x7C0B, [17053] = 0x7BF3, [17054] = 0x7C02, [17055] = 0x7C09,
  [17056] = 0x7C03, [17057] = 0x7C01, [17058] = 0x7BF8, [17059] = 0x7BFD,
  [17060] = 0x7C06, [17061] = 0x7BF0, [17062] = 0x7BF1, [17063] = 0x7C10,
  [17064] = 0x7C0A, [17065] = 0x7CE8, [17066] = 0x7E2D, [17067] = 0x7E3C,
  [17068] = 0x7E42, [17069] = 0x7E33, [17070] = 0x9848, [17071] = 0x7E38,
  [17072] = 0x7E2A, [17073] = 0x7E49, [17074] = 0x7E40, [17075] = 0x7E47,
  [17076] = 0x7E29, [17077] = 0x7E4C, [17078] = 0x7E30, [17079] = 0x7E3B,
  [17080] = 0x7E36, [17081] = 0x7E44, [17082] = 0x7E3A, [17083] = 0x7F45,
  [17084] = 0x7F7F, [17085] = 0x7F7E, [17086] = 0x7F7D, [17087] = 0x7FF4,
  [17088] = 0x7FF2, [17089] = 0x802C, [17090] = 0x81BB, [17091] = 0x81C4,
  [17092] = 0x81CC, [17093] = 0x81CA, [17094] = 0x81C5, [17095] = 0x81C7,
  [17096] = 0x81BC, [17097] = 0x81E9, [17098] = 0x825B, [17099] = 0x825A,
  [17100] = 0x825C, [17101] = 0x8583, [17102] = 0x8580, [17103] = 0x858F,
  [17104] = 0x85A7, [17105] = 0x8595, [17106] = 0x85A0, [17107] = 0x858B,
  [17108] = 0x85A3, [17109] = 0x857B, [17110] = 0x85A4, [17111] = 0x859A,
  [17112] = 0x859E, [17113] = 0x8577, [17114] = 0x857C, [17115] = 0x8589,
  [17116] = 0x85A1, [17117] = 0x857A, [17118] = 0x8578, [17119] = 0x8557,
  [17120] = 0x858E, [17121] = 0x8596, [17122] = 0x8586, [17123] = 0x858D,
  [17124] = 0x8599, [17125] = 0x859D, [17126] = 0x8581, [17127] = 0x85A2,
  [17128] = 0x8582, [17129] = 0x8588, [17130] = 0x8585, [17131] = 0x8579,
  [17132] = 0x8576, [17133] = 0x8598, [17134] = 0x8590, [17135] = 0x859F,
  [17136] = 0x8668, [17137] = 0x87BE, [17138] = 0x87AA, [17139] = 0x87AD,
  [17140] = 0x87C5, [17141] = 0x87B0, [17142] = 0x87AC, [17143] = 0x87B9,
  [17144] = 0x87B5, [17145] = 0x87BC, [17146] = 0x87AE, [17147] = 0x87C9,
  [17148] = 0x87C3, [17149] = 0x87C2, [17150] = 0x87CC, [17151] = 0x87B7,
  [17152] = 0x87AF, [17153] = 0x87C4, [17154] = 0x87CA, [17155] = 0x87B4,
  [17156] = 0x87B6, [17157] = 0x87BF, [17158] = 0x87B8, [17159] = 0x87BD,
  [17160] = 0x87DE, [17161] = 0x87B2, [17162] = 0x8935, [17163] = 0x8933,
  [17164] = 0x893C, [17165] = 0x893E, [17166] = 0x8941, [17167] = 0x8952,
  [17168] = 0x8937, [17169] = 0x8942, [17170] = 0x89AD, [17171] = 0x89AF,
  [17172] = 0x89AE, [17173] = 0x89F2, [17174] = 0x89F3, [17175] = 0x8B1E,
  [17176] = 0x8B18, [17177] = 0x8B16, [17178] = 0x8B11, [17179] = 0x8B05,
  [17180] = 0x8B0B, [17181] = 0x8B22, [17182] = 0x8B0F, [17183] = 0x8B12,
  [17184] = 0x8B15, [17185] = 0x8B07, [17186] = 0x8B0D, [17187] = 0x8B08,
  [17188] = 0x8B06, [17189] = 0x8B1C, [17190] = 0x8B13, [17191] = 0x8B1A,
  [17192] = 0x8C4F, [17193] = 0x8C70, [17194] = 0x8C72, [17195] = 0x8C71,
  [17196] = 0x8C6F, [17197] = 0x8C95, [17198] = 0x8C94, [17199] = 0x8CF9,
  [17200] = 0x8D6F, [17201] = 0x8E4E, [17202] = 0x8E4D, [17203] = 0x8E53,
  [17204] = 0x8E50, [17205] = 0x8E4C, [17206] = 0x8E47, [17207] = 0x8F43,
  [17208] = 0x8F40, [17209] = 0x9085, [17210] = 0x907E, [17211] = 0x9138,
  [17212] = 0x919A, [17213] = 0x91A2, [17214] = 0x919B, [17215] = 0x9199,
  [17216] = 0x919F, [17217] = 0x91A1, [17218] = 0x919D, [17219] = 0x91A0,
  [17220] = 0x93A1, [17221] = 0x9383, [17222] = 0x93AF, [17223] = 0x9364,
  [17224] = 0x9356, [17225] = 0x9347, [17226] = 0x937C, [17227] = 0x9358,
  [17228] = 0x935C, [17229] = 0x9376, [17230] = 0x9349, [17231] = 0x9350,
  [17232] = 0x9351, [17233] = 0x9360, [17234] = 0x936D, [17235] = 0x938F,
  [17236] = 0x934C, [17237] = 0x936A, [17238] = 0x9379, [17239] = 0x9357,
  [17240] = 0x9355, [17241] = 0x9352, [17242] = 0x934F, [17243] = 0x9371,
  [17244] = 0x9377, [17245] = 0x937B, [17246] = 0x9361, [17247] = 0x935E,
  [17248] = 0x9363, [17249] = 0x9367, [17250] = 0x9380, [17251] = 0x934E,
  [17252] = 0x9359, [17253] = 0x95C7, [17254] = 0x95C0, [17255] = 0x95C9,
  [17256] = 0x95C3, [17257] = 0x95C5, [17258] = 0x95B7, [17259] = 0x96AE,
  [17260] = 0x96B0, [17261] = 0x96AC, [17262] = 0x9720, [17263] = 0x971F,
  [17264] = 0x9718, [17265] = 0x971D, [17266] = 0x9719, [17267] = 0x979A,
  [17268] = 0x97A1, [17269] = 0x979C, [17270] = 0x979E, [17271] = 0x979D,
  [17272] = 0x97D5, [17273] = 0x97D4, [17274] = 0x97F1, [17275] = 0x9841,
  [17276] = 0x9844, [17277] = 0x984A, [17278] = 0x9849, [17279] = 0x9845,
  [17280] = 0x9843, [17281] = 0x9925, [17282] = 0x992B, [17283] = 0x992C,
  [17284] = 0x992A, [17285] = 0x9933, [17286] = 0x9932, [17287] = 0x992F,
  [17288] = 0x992D, [17289] = 0x9931, [17290] = 0x9930, [17291] = 0x9998,
  [17292] = 0x99A3, [17293] = 0x99A1, [17294] = 0x9A02, [17295] = 0x99FA,
  [17296] = 0x99F4, [17297] = 0x99F7, [17298] = 0x99F9, [17299] = 0x99F8,
  [17300] = 0x99F6, [17301] = 0x99FB, [17302] = 0x99FD, [17303] = 0x99FE,
  [17304] = 0x99FC, [17305] = 0x9A03, [17306] = 0x9ABE, [17307] = 0x9AFE,
  [17308] = 0x9AFD, [17309] = 0x9B01, [17310] = 0x9AFC, [17311] = 0x9B48,
  [17312] = 0x9B9A, [17313] = 0x9BA8, [17314] = 0x9B9E, [17315] = 0x9B9B,
  [17316] = 0x9BA6, [17317] = 0x9BA1, [17318] = 0x9BA5, [17319] = 0x9BA4,
  [17320] = 0x9B86, [17321] = 0x9BA2, [17322] = 0x9BA0, [17323] = 0x9BAF,
  [17324] = 0x9D33, [17325] = 0x9D41, [17326] = 0x9D67, [17327] = 0x9D36,
  [17328] = 0x9D2E, [17329] = 0x9D2F, [17330] = 0x9D31, [17331] = 0x9D38,
  [17332] = 0x9D30, [17333] = 0x9D45, [17334] = 0x9D42, [17335] = 0x9D43,
  [17336] = 0x9D3E, [17337] = 0x9D37, [17338] = 0x9D40, [17339] = 0x9D3D,
  [17340] = 0x7FF5, [17341] = 0x9D2D, [17342] = 0x9E8A, [17343] = 0x9E89,
  [17344] = 0x9E8D, [17345] = 0x9EB0, [17346] = 0x9EC8, [17347] = 0x9EDA,
  [17348] = 0x9EFB, [17349] = 0x9EFF, [17350] = 0x9F24, [17351] = 0x9F23,
  [17352] = 0x9F22, [17353] = 0x9F54, [17354] = 0x9FA0, [17355] = 0x5131,
  [17356] = 0x512D, [17357] = 0x512E, [17358] = 0x5698, [17359] = 0x569C,
  [17360] = 0x5697, [17361] = 0x569A, [17362] = 0x569D, [17363] = 0x5699,
  [17364] = 0x5970, [17365] = 0x5B3C, [17366] = 0x5C69, [17367] = 0x5C6A,
  [17368] = 0x5DC0, [17369] = 0x5E6D, [17370] = 0x5E6E, [17371] = 0x61D8,
  [17372] = 0x61DF, [17373] = 0x61ED, [17374] = 0x61EE, [17375] = 0x61F1,
  [17376] = 0x61EA, [17377] = 0x61F0, [17378] = 0x61EB, [17379] = 0x61D6,
  [17380] = 0x61E9, [17381] = 0x64FF, [17382] = 0x6504, [17383] = 0x64FD,
  [17384] = 0x64F8, [17385] = 0x6501, [17386] = 0x6503, [17387] = 0x64FC,
  [17388] = 0x6594, [17389] = 0x65DB, [17390] = 0x66DA, [17391] = 0x66DB,
  [17392] = 0x66D8, [17393] = 0x6AC5, [17394] = 0x6AB9, [17395] = 0x6ABD,
  [17396] = 0x6AE1, [17397] = 0x6AC6, [17398] = 0x6ABA, [17399] = 0x6AB6,
  [17400] = 0x6AB7, [17401] = 0x6AC7, [17402] = 0x6AB4, [17403] = 0x6AAD,
  [17404] = 0x6B5E, [17405] = 0x6BC9, [17406] = 0x6C0B, [17407] = 0x7007,
  [17408] = 0x700C, [17409] = 0x700D, [17410] = 0x7001, [17411] = 0x7005,
  [17412] = 0x7014, [17413] = 0x700E, [17414] = 0x6FFF, [17415] = 0x7000,
  [17416] = 0x6FFB, [17417] = 0x7026, [17418] = 0x6FFC, [17419] = 0x6FF7,
  [17420] = 0x700A, [17421] = 0x7201, [17422] = 0x71FF, [17423] = 0x71F9,
  [17424] = 0x7203, [17425] = 0x71FD, [17426] = 0x7376, [17427] = 0x74B8,
  [17428] = 0x74C0, [17429] = 0x74B5, [17430] = 0x74C1, [17431] = 0x74BE,
  [17432] = 0x74B6, [17433] = 0x74BB, [17434] = 0x74C2, [17435] = 0x7514,
  [17436] = 0x7513, [17437] = 0x765C, [17438] = 0x7664, [17439] = 0x7659,
  [17440] = 0x7650, [17441] = 0x7653, [17442] = 0x7657, [17443] = 0x765A,
  [17444] = 0x76A6, [17445] = 0x76BD, [17446] = 0x76EC, [17447] = 0x77C2,
  [17448] = 0x77BA, [17449] = 0x78FF, [17450] = 0x790C, [17451] = 0x7913,
  [17452] = 0x7914, [17453] = 0x7909, [17454] = 0x7910, [17455] = 0x7912,
  [17456] = 0x7911, [17457] = 0x79AD, [17458] = 0x79AC, [17459] = 0x7A5F,
  [17460] = 0x7C1C, [17461] = 0x7C29, [17462] = 0x7C19, [17463] = 0x7C20,
  [17464] = 0x7C1F, [17465] = 0x7C2D, [17466] = 0x7C1D, [17467] = 0x7C26,
  [17468] = 0x7C28, [17469] = 0x7C22, [17470] = 0x7C25, [17471] = 0x7C30,
  [17472] = 0x7E5C, [17473] = 0x7E50, [17474] = 0x7E56, [17475] = 0x7E63,
  [17476] = 0x7E58, [17477] = 0x7E62, [17478] = 0x7E5F, [17479] = 0x7E51,
  [17480] = 0x7E60, [17481] = 0x7E57, [17482] = 0x7E53, [17483] = 0x7FB5,
  [17484] = 0x7FB3, [17485] = 0x7FF7, [17486] = 0x7FF8, [17487] = 0x8075,
  [17488] = 0x81D1, [17489] = 0x81D2, [17490] = 0x81D0, [17491] = 0x825F,
  [17492] = 0x825E, [17493] = 0x85B4, [17494] = 0x85C6, [17495] = 0x85C0,
  [17496] = 0x85C3, [17497] = 0x85C2, [17498] = 0x85B3, [17499] = 0x85B5,
  [17500] = 0x85BD, [17501] = 0x85C7, [17502] = 0x85C4, [17503] = 0x85BF,
  [17504] = 0x85CB, [17505] = 0x85CE, [17506] = 0x85C8, [17507] = 0x85C5,
  [17508] = 0x85B1, [17509] = 0x85B6, [17510] = 0x85D2, [17511] = 0x8624,
  [17512] = 0x85B8, [17513] = 0x85B7, [17514] = 0x85BE, [17515] = 0x8669,
  [17516] = 0x87E7, [17517] = 0x87E6, [17518] = 0x87E2, [17519] = 0x87DB,
  [17520] = 0x87EB, [17521] = 0x87EA, [17522] = 0x87E5, [17523] = 0x87DF,
  [17524] = 0x87F3, [17525] = 0x87E4, [17526] = 0x87D4, [17527] = 0x87DC,
  [17528] = 0x87D3, [17529] = 0x87ED, [17530] = 0x87D8, [17531] = 0x87E3,
  [17532] = 0x87A4, [17533] = 0x87D7, [17534] = 0x87D9, [17535] = 0x8801,
  [17536] = 0x87F4, [17537] = 0x87E8, [17538] = 0x87DD, [17539] = 0x8953,
  [17540] = 0x894B, [17541] = 0x894F, [17542] = 0x894C, [17543] = 0x8946,
  [17544] = 0x8950, [17545] = 0x8951, [17546] = 0x8949, [17547] = 0x8B2A,
  [17548] = 0x8B27, [17549] = 0x8B23, [17550] = 0x8B33, [17551] = 0x8B30,
  [17552] = 0x8B35, [17553] = 0x8B47, [17554] = 0x8B2F, [17555] = 0x8B3C,
  [17556] = 0x8B3E, [17557] = 0x8B31, [17558] = 0x8B25, [17559] = 0x8B37,
  [17560] = 0x8B26, [17561] = 0x8B36, [17562] = 0x8B2E, [17563] = 0x8B24,
  [17564] = 0x8B3B, [17565] = 0x8B3D, [17566] = 0x8B3A, [17567] = 0x8C42,
  [17568] = 0x8C75, [17569] = 0x8C99, [17570] = 0x8C98, [17571] = 0x8C97,
  [17572] = 0x8CFE, [17573] = 0x8D04, [17574] = 0x8D02, [17575] = 0x8D00,
  [17576] = 0x8E5C, [17577] = 0x8E62, [17578] = 0x8E60, [17579] = 0x8E57,
  [17580] = 0x8E56, [17581] = 0x8E5E, [17582] = 0x8E65, [17583] = 0x8E67,
  [17584] = 0x8E5B, [17585] = 0x8E5A, [17586] = 0x8E61, [17587] = 0x8E5D,
  [17588] = 0x8E69, [17589] = 0x8E54, [17590] = 0x8F46, [17591] = 0x8F47,
  [17592] = 0x8F48, [17593] = 0x8F4B, [17594] = 0x9128, [17595] = 0x913A,
  [17596] = 0x913B, [17597] = 0x913E, [17598] = 0x91A8, [17599] = 0x91A5,
  [17600] = 0x91A7, [17601] = 0x91AF, [17602] = 0x91AA, [17603] = 0x93B5,
  [17604] = 0x938C, [17605] = 0x9392, [17606] = 0x93B7, [17607] = 0x939B,
  [17608] = 0x939D, [17609] = 0x9389, [17610] = 0x93A7, [17611] = 0x938E,
  [17612] = 0x93AA, [17613] = 0x939E, [17614] = 0x93A6, [17615] = 0x9395,
  [17616] = 0x9388, [17617] = 0x9399, [17618] = 0x939F, [17619] = 0x938D,
  [17620] = 0x93B1, [17621] = 0x9391, [17622] = 0x93B2, [17623] = 0x93A4,
  [17624] = 0x93A8, [17625] = 0x93B4, [17626] = 0x93A3, [17627] = 0x93A5,
  [17628] = 0x95D2, [17629] = 0x95D3, [17630] = 0x95D1, [17631] = 0x96B3,
  [17632] = 0x96D7, [17633] = 0x96DA, [17634] = 0x5DC2, [17635] = 0x96DF,
  [17636] = 0x96D8, [17637] = 0x96DD, [17638] = 0x9723, [17639] = 0x9722,
  [17640] = 0x9725, [17641] = 0x97AC, [17642] = 0x97AE, [17643] = 0x97A8,
  [17644] = 0x97AB, [17645] = 0x97A4, [17646] = 0x97AA, [17647] = 0x97A2,
  [17648] = 0x97A5, [17649] = 0x97D7, [17650] = 0x97D9, [17651] = 0x97D6,
  [17652] = 0x97D8, [17653] = 0x97FA, [17654] = 0x9850, [17655] = 0x9851,
  [17656] = 0x9852, [17657] = 0x98B8, [17658] = 0x9941, [17659] = 0x993C,
  [17660] = 0x993A, [17661] = 0x9A0F, [17662] = 0x9A0B, [17663] = 0x9A09,
  [17664] = 0x9A0D, [17665] = 0x9A04, [17666] = 0x9A11, [17667] = 0x9A0A,
  [17668] = 0x9A05, [17669] = 0x9A07, [17670] = 0x9A06, [17671] = 0x9AC0,
  [17672] = 0x9ADC, [17673] = 0x9B08, [17674] = 0x9B04, [17675] = 0x9B05,
  [17676] = 0x9B29, [17677] = 0x9B35, [17678] = 0x9B4A, [17679] = 0x9B4C,
  [17680] = 0x9B4B, [17681] = 0x9BC7, [17682] = 0x9BC6, [17683] = 0x9BC3,
  [17684] = 0x9BBF, [17685] = 0x9BC1, [17686] = 0x9BB5, [17687] = 0x9BB8,
  [17688] = 0x9BD3, [17689] = 0x9BB6, [17690] = 0x9BC4, [17691] = 0x9BB9,
  [17692] = 0x9BBD, [17693] = 0x9D5C, [17694] = 0x9D53, [17695] = 0x9D4F,
  [17696] = 0x9D4A, [17697] = 0x9D5B, [17698] = 0x9D4B, [17699] = 0x9D59,
  [17700] = 0x9D56, [17701] = 0x9D4C, [17702] = 0x9D57, [17703] = 0x9D52,
  [17704] = 0x9D54, [17705] = 0x9D5F, [17706] = 0x9D58, [17707] = 0x9D5A,
  [17708] = 0x9E8E, [17709] = 0x9E8C, [17710] = 0x9EDF, [17711] = 0x9F01,
  [17712] = 0x9F00, [17713] = 0x9F16, [17714] = 0x9F25, [17715] = 0x9F2B,
  [17716] = 0x9F2A, [17717] = 0x9F29, [17718] = 0x9F28, [17719] = 0x9F4C,
  [17720] = 0x9F55, [17721] = 0x5134, [17722] = 0x5135, [17723] = 0x5296,
  [17724] = 0x52F7, [17725] = 0x53B4, [17726] = 0x56AB, [17727] = 0x56AD,
  [17728] = 0x56A6, [17729] = 0x56A7, [17730] = 0x56AA, [17731] = 0x56AC,
  [17732] = 0x58DA, [17733] = 0x58DD, [17734] = 0x58DB, [17735] = 0x5912,
  [17736] = 0x5B3D, [17737] = 0x5B3E, [17738] = 0x5B3F, [17739] = 0x5DC3,
  [17740] = 0x5E70, [17741] = 0x5FBF, [17742] = 0x61FB, [17743] = 0x6507,
  [17744] = 0x6510, [17745] = 0x650D, [17746] = 0x6509, [17747] = 0x650C,
  [17748] = 0x650E, [17749] = 0x6584, [17750] = 0x65DE, [17751] = 0x65DD,
  [17752] = 0x66DE, [17753] = 0x6AE7, [17754] = 0x6AE0, [17755] = 0x6ACC,
  [17756] = 0x6AD1, [17757] = 0x6AD9, [17758] = 0x6ACB, [17759] = 0x6ADF,
  [17760] = 0x6ADC, [17761] = 0x6AD0, [17762] = 0x6AEB, [17763] = 0x6ACF,
  [17764] = 0x6ACD, [17765] = 0x6ADE, [17766] = 0x6B60, [17767] = 0x6BB0,
  [17768] = 0x6C0C, [17769] = 0x7019, [17770] = 0x7027, [17771] = 0x7020,
  [17772] = 0x7016, [17773] = 0x702B, [17774] = 0x7021, [17775] = 0x7022,
  [17776] = 0x7023, [17777] = 0x7029, [17778] = 0x7017, [17779] = 0x7024,
  [17780] = 0x701C, [17781] = 0x702A, [17782] = 0x720C, [17783] = 0x720A,
  [17784] = 0x7207, [17785] = 0x7202, [17786] = 0x7205, [17787] = 0x72A5,
  [17788] = 0x72A6, [17789] = 0x72A4, [17790] = 0x72A3, [17791] = 0x72A1,
  [17792] = 0x74CB, [17793] = 0x74C5, [17794] = 0x74B7, [17795] = 0x74C3,
  [17796] = 0x7516, [17797] = 0x7660, [17798] = 0x77C9, [17799] = 0x77CA,
  [17800] = 0x77C4, [17801] = 0x77F1, [17802] = 0x791D, [17803] = 0x791B,
  [17804] = 0x7921, [17805] = 0x791C, [17806] = 0x7917, [17807] = 0x791E,
  [17808] = 0x79B0, [17809] = 0x7A67, [17810] = 0x7A68, [17811] = 0x7C33,
  [17812] = 0x7C3C, [17813] = 0x7C39, [17814] = 0x7C2C, [17815] = 0x7C3B,
  [17816] = 0x7CEC, [17817] = 0x7CEA, [17818] = 0x7E76, [17819] = 0x7E75,
  [17820] = 0x7E78, [17821] = 0x7E70, [17822] = 0x7E77, [17823] = 0x7E6F,
  [17824] = 0x7E7A, [17825] = 0x7E72, [17826] = 0x7E74, [17827] = 0x7E68,
  [17828] = 0x7F4B, [17829] = 0x7F4A, [17830] = 0x7F83, [17831] = 0x7F86,
  [17832] = 0x7FB7, [17833] = 0x7FFD, [17834] = 0x7FFE, [17835] = 0x8078,
  [17836] = 0x81D7, [17837] = 0x81D5, [17838] = 0x8264, [17839] = 0x8261,
  [17840] = 0x8263, [17841] = 0x85EB, [17842] = 0x85F1, [17843] = 0x85ED,
  [17844] = 0x85D9, [17845] = 0x85E1, [17846] = 0x85E8, [17847] = 0x85DA,
  [17848] = 0x85D7, [17849] = 0x85EC, [17850] = 0x85F2, [17851] = 0x85F8,
  [17852] = 0x85D8, [17853] = 0x85DF, [17854] = 0x85E3, [17855] = 0x85DC,
  [17856] = 0x85D1, [17857] = 0x85F0, [17858] = 0x85E6, [17859] = 0x85EF,
  [17860] = 0x85DE, [17861] = 0x85E2, [17862] = 0x8800, [17863] = 0x87FA,
  [17864] = 0x8803, [17865] = 0x87F6, [17866] = 0x87F7, [17867] = 0x8809,
  [17868] = 0x880C, [17869] = 0x880B, [17870] = 0x8806, [17871] = 0x87FC,
  [17872] = 0x8808, [17873] = 0x87FF, [17874] = 0x880A, [17875] = 0x8802,
  [17876] = 0x8962, [17877] = 0x895A, [17878] = 0x895B, [17879] = 0x8957,
  [17880] = 0x8961, [17881] = 0x895C, [17882] = 0x8958, [17883] = 0x895D,
  [17884] = 0x8959, [17885] = 0x8988, [17886] = 0x89B7, [17887] = 0x89B6,
  [17888] = 0x89F6, [17889] = 0x8B50, [17890] = 0x8B48, [17891] = 0x8B4A,
  [17892] = 0x8B40, [17893] = 0x8B53, [17894] = 0x8B56, [17895] = 0x8B54,
  [17896] = 0x8B4B, [17897] = 0x8B55, [17898] = 0x8B51, [17899] = 0x8B42,
  [17900] = 0x8B52, [17901] = 0x8B57, [17902] = 0x8C43, [17903] = 0x8C77,
  [17904] = 0x8C76, [17905] = 0x8C9A, [17906] = 0x8D06, [17907] = 0x8D07,
  [17908] = 0x8D09, [17909] = 0x8DAC, [17910] = 0x8DAA, [17911] = 0x8DAD,
  [17912] = 0x8DAB, [17913] = 0x8E6D, [17914] = 0x8E78, [17915] = 0x8E73,
  [17916] = 0x8E6A, [17917] = 0x8E6F, [17918] = 0x8E7B, [17919] = 0x8EC2,
  [17920] = 0x8F52, [17921] = 0x8F51, [17922] = 0x8F4F, [17923] = 0x8F50,
  [17924] = 0x8F53, [17925] = 0x8FB4, [17926] = 0x9140, [17927] = 0x913F,
  [17928] = 0x91B0, [17929] = 0x91AD, [17930] = 0x93DE, [17931] = 0x93C7,
  [17932] = 0x93CF, [17933] = 0x93C2, [17934] = 0x93DA, [17935] = 0x93D0,
  [17936] = 0x93F9, [17937] = 0x93EC, [17938] = 0x93CC, [17939] = 0x93D9,
  [17940] = 0x93A9, [17941] = 0x93E6, [17942] = 0x93CA, [17943] = 0x93D4,
  [17944] = 0x93EE, [17945] = 0x93E3, [17946] = 0x93D5, [17947] = 0x93C4,
  [17948] = 0x93CE, [17949] = 0x93C0, [17950] = 0x93D2, [17951] = 0x93E7,
  [17952] = 0x957D, [17953] = 0x95DA, [17954] = 0x95DB, [17955] = 0x96E1,
  [17956] = 0x9729, [17957] = 0x972B, [17958] = 0x972C, [17959] = 0x9728,
  [17960] = 0x9726, [17961] = 0x97B3, [17962] = 0x97B7, [17963] = 0x97B6,
  [17964] = 0x97DD, [17965] = 0x97DE, [17966] = 0x97DF, [17967] = 0x985C,
  [17968] = 0x9859, [17969] = 0x985D, [17970] = 0x9857, [17971] = 0x98BF,
  [17972] = 0x98BD, [17973] = 0x98BB, [17974] = 0x98BE, [17975] = 0x9948,
  [17976] = 0x9947, [17977] = 0x9943, [17978] = 0x99A6, [17979] = 0x99A7,
  [17980] = 0x9A1A, [17981] = 0x9A15, [17982] = 0x9A25, [17983] = 0x9A1D,
  [17984] = 0x9A24, [17985] = 0x9A1B, [17986] = 0x9A22, [17987] = 0x9A20,
  [17988] = 0x9A27, [17989] = 0x9A23, [17990] = 0x9A1E, [17991] = 0x9A1C,
  [17992] = 0x9A14, [17993] = 0x9AC2, [17994] = 0x9B0B, [17995] = 0x9B0A,
  [17996] = 0x9B0E, [17997] = 0x9B0C, [17998] = 0x9B37, [17999] = 0x9BEA,
  [18000] = 0x9BEB, [18001] = 0x9BE0, [18002] = 0x9BDE, [18003] = 0x9BE4,
  [18004] = 0x9BE6, [18005] = 0x9BE2, [18006] = 0x9BF0, [18007] = 0x9BD4,
  [18008] = 0x9BD7, [18009] = 0x9BEC, [18010] = 0x9BDC, [18011] = 0x9BD9,
  [18012] = 0x9BE5, [18013] = 0x9BD5, [18014] = 0x9BE1, [18015] = 0x9BDA,
  [18016] = 0x9D77, [18017] = 0x9D81, [18018] = 0x9D8A, [18019] = 0x9D84,
  [18020] = 0x9D88, [18021] = 0x9D71, [18022] = 0x9D80, [18023] = 0x9D78,
  [18024] = 0x9D86, [18025] = 0x9D8B, [18026] = 0x9D8C, [18027] = 0x9D7D,
  [18028] = 0x9D6B, [18029] = 0x9D74, [18030] = 0x9D75, [18031] = 0x9D70,
  [18032] = 0x9D69, [18033] = 0x9D85, [18034] = 0x9D73, [18035] = 0x9D7B,
  [18036] = 0x9D82, [18037] = 0x9D6F, [18038] = 0x9D79, [18039] = 0x9D7F,
  [18040] = 0x9D87, [18041] = 0x9D68, [18042] = 0x9E94, [18043] = 0x9E91,
  [18044] = 0x9EC0, [18045] = 0x9EFC, [18046] = 0x9F2D, [18047] = 0x9F40,
  [18048] = 0x9F41, [18049] = 0x9F4D, [18050] = 0x9F56, [18051] = 0x9F57,
  [18052] = 0x9F58, [18053] = 0x5337, [18054] = 0x56B2, [18055] = 0x56B5,
  [18056] = 0x56B3, [18057] = 0x58E3, [18058] = 0x5B45, [18059] = 0x5DC6,
  [18060] = 0x5DC7, [18061] = 0x5EEE, [18062] = 0x5EEF, [18063] = 0x5FC0,
  [18064] = 0x5FC1, [18065] = 0x61F9, [18066] = 0x6517, [18067] = 0x6516,
  [18068] = 0x6515, [18069] = 0x6513, [18070] = 0x65DF, [18071] = 0x66E8,
  [18072] = 0x66E3, [18073] = 0x66E4, [18074] = 0x6AF3, [18075] = 0x6AF0,
  [18076] = 0x6AEA, [18077] = 0x6AE8, [18078] = 0x6AF9, [18079] = 0x6AF1,
  [18080] = 0x6AEE, [18081] = 0x6AEF, [18082] = 0x703C, [18083] = 0x7035,
  [18084] = 0x702F, [18085] = 0x7037, [18086] = 0x7034, [18087] = 0x7031,
  [18088] = 0x7042, [18089] = 0x7038, [18090] = 0x703F, [18091] = 0x703A,
  [18092] = 0x7039, [18093] = 0x7040, [18094] = 0x703B, [18095] = 0x7033,
  [18096] = 0x7041, [18097] = 0x7213, [18098] = 0x7214, [18099] = 0x72A8,
  [18100] = 0x737D, [18101] = 0x737C, [18102] = 0x74BA, [18103] = 0x76AB,
  [18104] = 0x76AA, [18105] = 0x76BE, [18106] = 0x76ED, [18107] = 0x77CC,
  [18108] = 0x77CE, [18109] = 0x77CF, [18110] = 0x77CD, [18111] = 0x77F2,
  [18112] = 0x7925, [18113] = 0x7923, [18114] = 0x7927, [18115] = 0x7928,
  [18116] = 0x7924, [18117] = 0x7929, [18118] = 0x79B2, [18119] = 0x7A6E,
  [18120] = 0x7A6C, [18121] = 0x7A6D, [18122] = 0x7AF7, [18123] = 0x7C49,
  [18124] = 0x7C48, [18125] = 0x7C4A, [18126] = 0x7C47, [18127] = 0x7C45,
  [18128] = 0x7CEE, [18129] = 0x7E7B, [18130] = 0x7E7E, [18131] = 0x7E81,
  [18132] = 0x7E80, [18133] = 0x7FBA, [18134] = 0x7FFF, [18135] = 0x8079,
  [18136] = 0x81DB, [18137] = 0x81D9, [18138] = 0x820B, [18139] = 0x8268,
  [18140] = 0x8269, [18141] = 0x8622, [18142] = 0x85FF, [18143] = 0x8601,
  [18144] = 0x85FE, [18145] = 0x861B, [18146] = 0x8600, [18147] = 0x85F6,
  [18148] = 0x8604, [18149] = 0x8609, [18150] = 0x8605, [18151] = 0x860C,
  [18152] = 0x85FD, [18153] = 0x8819, [18154] = 0x8810, [18155] = 0x8811,
  [18156] = 0x8817, [18157] = 0x8813, [18158] = 0x8816, [18159] = 0x8963,
  [18160] = 0x8966, [18161] = 0x89B9, [18162] = 0x89F7, [18163] = 0x8B60,
  [18164] = 0x8B6A, [18165] = 0x8B5D, [18166] = 0x8B68, [18167] = 0x8B63,
  [18168] = 0x8B65, [18169] = 0x8B67, [18170] = 0x8B6D, [18171] = 0x8DAE,
  [18172] = 0x8E86, [18173] = 0x8E88, [18174] = 0x8E84, [18175] = 0x8F59,
  [18176] = 0x8F56, [18177] = 0x8F57, [18178] = 0x8F55, [18179] = 0x8F58,
  [18180] = 0x8F5A, [18181] = 0x908D, [18182] = 0x9143, [18183] = 0x9141,
  [18184] = 0x91B7, [18185] = 0x91B5, [18186] = 0x91B2, [18187] = 0x91B3,
  [18188] = 0x940B, [18189] = 0x9413, [18190] = 0x93FB, [18191] = 0x9420,
  [18192] = 0x940F, [18193] = 0x9414, [18194] = 0x93FE, [18195] = 0x9415,
  [18196] = 0x9410, [18197] = 0x9428, [18198] = 0x9419, [18199] = 0x940D,
  [18200] = 0x93F5, [18201] = 0x9400, [18202] = 0x93F7, [18203] = 0x9407,
  [18204] = 0x940E, [18205] = 0x9416, [18206] = 0x9412, [18207] = 0x93FA,
  [18208] = 0x9409, [18209] = 0x93F8, [18210] = 0x940A, [18211] = 0x93FF,
  [18212] = 0x93FC, [18213] = 0x940C, [18214] = 0x93F6, [18215] = 0x9411,
  [18216] = 0x9406, [18217] = 0x95DE, [18218] = 0x95E0, [18219] = 0x95DF,
  [18220] = 0x972E, [18221] = 0x972F, [18222] = 0x97B9, [18223] = 0x97BB,
  [18224] = 0x97FD, [18225] = 0x97FE, [18226] = 0x9860, [18227] = 0x9862,
  [18228] = 0x9863, [18229] = 0x985F, [18230] = 0x98C1, [18231] = 0x98C2,
  [18232] = 0x9950, [18233] = 0x994E, [18234] = 0x9959, [18235] = 0x994C,
  [18236] = 0x994B, [18237] = 0x9953, [18238] = 0x9A32, [18239] = 0x9A34,
  [18240] = 0x9A31, [18241] = 0x9A2C, [18242] = 0x9A2A, [18243] = 0x9A36,
  [18244] = 0x9A29, [18245] = 0x9A2E, [18246] = 0x9A38, [18247] = 0x9A2D,
  [18248] = 0x9AC7, [18249] = 0x9ACA, [18250] = 0x9AC6, [18251] = 0x9B10,
  [18252] = 0x9B12, [18253] = 0x9B11, [18254] = 0x9C0B, [18255] = 0x9C08,
  [18256] = 0x9BF7, [18257] = 0x9C05, [18258] = 0x9C12, [18259] = 0x9BF8,
  [18260] = 0x9C40, [18261] = 0x9C07, [18262] = 0x9C0E, [18263] = 0x9C06,
  [18264] = 0x9C17, [18265] = 0x9C14, [18266] = 0x9C09, [18267] = 0x9D9F,
  [18268] = 0x9D99, [18269] = 0x9DA4, [18270] = 0x9D9D, [18271] = 0x9D92,
  [18272] = 0x9D98, [18273] = 0x9D90, [18274] = 0x9D9B, [18275] = 0x9DA0,
  [18276] = 0x9D94, [18277] = 0x9D9C, [18278] = 0x9DAA, [18279] = 0x9D97,
  [18280] = 0x9DA1, [18281] = 0x9D9A, [18282] = 0x9DA2, [18283] = 0x9DA8,
  [18284] = 0x9D9E, [18285] = 0x9DA3, [18286] = 0x9DBF, [18287] = 0x9DA9,
  [18288] = 0x9D96, [18289] = 0x9DA6, [18290] = 0x9DA7, [18291] = 0x9E99,
  [18292] = 0x9E9B, [18293] = 0x9E9A, [18294] = 0x9EE5, [18295] = 0x9EE4,
  [18296] = 0x9EE7, [18297] = 0x9EE6, [18298] = 0x9F30, [18299] = 0x9F2E,
  [18300] = 0x9F5B, [18301] = 0x9F60, [18302] = 0x9F5E, [18303] = 0x9F5D,
  [18304] = 0x9F59, [18305] = 0x9F91, [18306] = 0x513A, [18307] = 0x5139,
  [18308] = 0x5298, [18309] = 0x5297, [18310] = 0x56C3, [18311] = 0x56BD,
  [18312] = 0x56BE, [18313] = 0x5B48, [18314] = 0x5B47, [18315] = 0x5DCB,
  [18316] = 0x5DCF, [18317] = 0x5EF1, [18318] = 0x61FD, [18319] = 0x651B,
  [18320] = 0x6B02, [18321] = 0x6AFC, [18322] = 0x6B03, [18323] = 0x6AF8,
  [18324] = 0x6B00, [18325] = 0x7043, [18326] = 0x7044, [18327] = 0x704A,
  [18328] = 0x7048, [18329] = 0x7049, [18330] = 0x7045, [18331] = 0x7046,
  [18332] = 0x721D, [18333] = 0x721A, [18334] = 0x7219, [18335] = 0x737E,
  [18336] = 0x7517, [18337] = 0x766A, [18338] = 0x77D0, [18339] = 0x792D,
  [18340] = 0x7931, [18341] = 0x792F, [18342] = 0x7C54, [18343] = 0x7C53,
  [18344] = 0x7CF2, [18345] = 0x7E8A, [18346] = 0x7E87, [18347] = 0x7E88,
  [18348] = 0x7E8B, [18349] = 0x7E86, [18350] = 0x7E8D, [18351] = 0x7F4D,
  [18352] = 0x7FBB, [18353] = 0x8030, [18354] = 0x81DD, [18355] = 0x8618,
  [18356] = 0x862A, [18357] = 0x8626, [18358] = 0x861F, [18359] = 0x8623,
  [18360] = 0x861C, [18361] = 0x8619, [18362] = 0x8627, [18363] = 0x862E,
  [18364] = 0x8621, [18365] = 0x8620, [18366] = 0x8629, [18367] = 0x861E,
  [18368] = 0x8625, [18369] = 0x8829, [18370] = 0x881D, [18371] = 0x881B,
  [18372] = 0x8820, [18373] = 0x8824, [18374] = 0x881C, [18375] = 0x882B,
  [18376] = 0x884A, [18377] = 0x896D, [18378] = 0x8969, [18379] = 0x896E,
  [18380] = 0x896B, [18381] = 0x89FA, [18382] = 0x8B79, [18383] = 0x8B78,
  [18384] = 0x8B45, [18385] = 0x8B7A, [18386] = 0x8B7B, [18387] = 0x8D10,
  [18388] = 0x8D14, [18389] = 0x8DAF, [18390] = 0x8E8E, [18391] = 0x8E8C,
  [18392] = 0x8F5E, [18393] = 0x8F5B, [18394] = 0x8F5D, [18395] = 0x9146,
  [18396] = 0x9144, [18397] = 0x9145, [18398] = 0x91B9, [18399] = 0x943F,
  [18400] = 0x943B, [18401] = 0x9436, [18402] = 0x9429, [18403] = 0x943D,
  [18404] = 0x943C, [18405] = 0x9430, [18406] = 0x9439, [18407] = 0x942A,
  [18408] = 0x9437, [18409] = 0x942C, [18410] = 0x9440, [18411] = 0x9431,
  [18412] = 0x95E5, [18413] = 0x95E4, [18414] = 0x95E3, [18415] = 0x9735,
  [18416] = 0x973A, [18417] = 0x97BF, [18418] = 0x97E1, [18419] = 0x9864,
  [18420] = 0x98C9, [18421] = 0x98C6, [18422] = 0x98C0, [18423] = 0x9958,
  [18424] = 0x9956, [18425] = 0x9A39, [18426] = 0x9A3D, [18427] = 0x9A46,
  [18428] = 0x9A44, [18429] = 0x9A42, [18430] = 0x9A41, [18431] = 0x9A3A,
  [18432] = 0x9A3F, [18433] = 0x9ACD, [18434] = 0x9B15, [18435] = 0x9B17,
  [18436] = 0x9B18, [18437] = 0x9B16, [18438] = 0x9B3A, [18439] = 0x9B52,
  [18440] = 0x9C2B, [18441] = 0x9C1D, [18442] = 0x9C1C, [18443] = 0x9C2C,
  [18444] = 0x9C23, [18445] = 0x9C28, [18446] = 0x9C29, [18447] = 0x9C24,
  [18448] = 0x9C21, [18449] = 0x9DB7, [18450] = 0x9DB6, [18451] = 0x9DBC,
  [18452] = 0x9DC1, [18453] = 0x9DC7, [18454] = 0x9DCA, [18455] = 0x9DCF,
  [18456] = 0x9DBE, [18457] = 0x9DC5, [18458] = 0x9DC3, [18459] = 0x9DBB,
  [18460] = 0x9DB5, [18461] = 0x9DCE, [18462] = 0x9DB9, [18463] = 0x9DBA,
  [18464] = 0x9DAC, [18465] = 0x9DC8, [18466] = 0x9DB1, [18467] = 0x9DAD,
  [18468] = 0x9DCC, [18469] = 0x9DB3, [18470] = 0x9DCD, [18471] = 0x9DB2,
  [18472] = 0x9E7A, [18473] = 0x9E9C, [18474] = 0x9EEB, [18475] = 0x9EEE,
  [18476] = 0x9EED, [18477] = 0x9F1B, [18478] = 0x9F18, [18479] = 0x9F1A,
  [18480] = 0x9F31, [18481] = 0x9F4E, [18482] = 0x9F65, [18483] = 0x9F64,
  [18484] = 0x9F92, [18485] = 0x4EB9, [18486] = 0x56C6, [18487] = 0x56C5,
  [18488] = 0x56CB, [18489] = 0x5971, [18490] = 0x5B4B, [18491] = 0x5B4C,
  [18492] = 0x5DD5, [18493] = 0x5DD1, [18494] = 0x5EF2, [18495] = 0x6521,
  [18496] = 0x6520, [18497] = 0x6526, [18498] = 0x6522, [18499] = 0x6B0B,
  [18500] = 0x6B08, [18501] = 0x6B09, [18502] = 0x6C0D, [18503] = 0x7055,
  [18504] = 0x7056, [18505] = 0x7057, [18506] = 0x7052, [18507] = 0x721E,
  [18508] = 0x721F, [18509] = 0x72A9, [18510] = 0x737F, [18511] = 0x74D8,
  [18512] = 0x74D5, [18513] = 0x74D9, [18514] = 0x74D7, [18515] = 0x766D,
  [18516] = 0x76AD, [18517] = 0x7935, [18518] = 0x79B4, [18519] = 0x7A70,
  [18520] = 0x7A71, [18521] = 0x7C57, [18522] = 0x7C5C, [18523] = 0x7C59,
  [18524] = 0x7C5B, [18525] = 0x7C5A, [18526] = 0x7CF4, [18527] = 0x7CF1,
  [18528] = 0x7E91, [18529] = 0x7F4F, [18530] = 0x7F87, [18531] = 0x81DE,
  [18532] = 0x826B, [18533] = 0x8634, [18534] = 0x8635, [18535] = 0x8633,
  [18536] = 0x862C, [18537] = 0x8632, [18538] = 0x8636, [18539] = 0x882C,
  [18540] = 0x8828, [18541] = 0x8826, [18542] = 0x882A, [18543] = 0x8825,
  [18544] = 0x8971, [18545] = 0x89BF, [18546] = 0x89BE, [18547] = 0x89FB,
  [18548] = 0x8B7E, [18549] = 0x8B84, [18550] = 0x8B82, [18551] = 0x8B86,
  [18552] = 0x8B85, [18553] = 0x8B7F, [18554] = 0x8D15, [18555] = 0x8E95,
  [18556] = 0x8E94, [18557] = 0x8E9A, [18558] = 0x8E92, [18559] = 0x8E90,
  [18560] = 0x8E96, [18561] = 0x8E97, [18562] = 0x8F60, [18563] = 0x8F62,
  [18564] = 0x9147, [18565] = 0x944C, [18566] = 0x9450, [18567] = 0x944A,
  [18568] = 0x944B, [18569] = 0x944F, [18570] = 0x9447, [18571] = 0x9445,
  [18572] = 0x9448, [18573] = 0x9449, [18574] = 0x9446, [18575] = 0x973F,
  [18576] = 0x97E3, [18577] = 0x986A, [18578] = 0x9869, [18579] = 0x98CB,
  [18580] = 0x9954, [18581] = 0x995B, [18582] = 0x9A4E, [18583] = 0x9A53,
  [18584] = 0x9A54, [18585] = 0x9A4C, [18586] = 0x9A4F, [18587] = 0x9A48,
  [18588] = 0x9A4A, [18589] = 0x9A49, [18590] = 0x9A52, [18591] = 0x9A50,
  [18592] = 0x9AD0, [18593] = 0x9B19, [18594] = 0x9B2B, [18595] = 0x9B3B,
  [18596] = 0x9B56, [18597] = 0x9B55, [18598] = 0x9C46, [18599] = 0x9C48,
  [18600] = 0x9C3F, [18601] = 0x9C44, [18602] = 0x9C39, [18603] = 0x9C33,
  [18604] = 0x9C41, [18605] = 0x9C3C, [18606] = 0x9C37, [18607] = 0x9C34,
  [18608] = 0x9C32, [18609] = 0x9C3D, [18610] = 0x9C36, [18611] = 0x9DDB,
  [18612] = 0x9DD2, [18613] = 0x9DDE, [18614] = 0x9DDA, [18615] = 0x9DCB,
  [18616] = 0x9DD0, [18617] = 0x9DDC, [18618] = 0x9DD1, [18619] = 0x9DDF,
  [18620] = 0x9DE9, [18621] = 0x9DD9, [18622] = 0x9DD8, [18623] = 0x9DD6,
  [18624] = 0x9DF5, [18625] = 0x9DD5, [18626] = 0x9DDD, [18627] = 0x9EB6,
  [18628] = 0x9EF0, [18629] = 0x9F35, [18630] = 0x9F33, [18631] = 0x9F32,
  [18632] = 0x9F42, [18633] = 0x9F6B, [18634] = 0x9F95, [18635] = 0x9FA2,
  [18636] = 0x513D, [18637] = 0x5299, [18638] = 0x58E8, [18639] = 0x58E7,
  [18640] = 0x5972, [18641] = 0x5B4D, [18642] = 0x5DD8, [18643] = 0x882F,
  [18644] = 0x5F4F, [18645] = 0x6201, [18646] = 0x6203, [18647] = 0x6204,
  [18648] = 0x6529, [18649] = 0x6525, [18650] = 0x6596, [18651] = 0x66EB,
  [18652] = 0x6B11, [18653] = 0x6B12, [18654] = 0x6B0F, [18655] = 0x6BCA,
  [18656] = 0x705B, [18657] = 0x705A, [18658] = 0x7222, [18659] = 0x7382,
  [18660] = 0x7381, [18661] = 0x7383, [18662] = 0x7670, [18663] = 0x77D4,
  [18664] = 0x7C67, [18665] = 0x7C66, [18666] = 0x7E95, [18667] = 0x826C,
  [18668] = 0x863A, [18669] = 0x8640, [18670] = 0x8639, [18671] = 0x863C,
  [18672] = 0x8631, [18673] = 0x863B, [18674] = 0x863E, [18675] = 0x8830,
  [18676] = 0x8832, [18677] = 0x882E, [18678] = 0x8833, [18679] = 0x8976,
  [18680] = 0x8974, [18681] = 0x8973, [18682] = 0x89FE, [18683] = 0x8B8C,
  [18684] = 0x8B8E, [18685] = 0x8B8B, [18686] = 0x8B88, [18687] = 0x8C45,
  [18688] = 0x8D19, [18689] = 0x8E98, [18690] = 0x8F64, [18691] = 0x8F63,
  [18692] = 0x91BC, [18693] = 0x9462, [18694] = 0x9455, [18695] = 0x945D,
  [18696] = 0x9457, [18697] = 0x945E, [18698] = 0x97C4, [18699] = 0x97C5,
  [18700] = 0x9800, [18701] = 0x9A56, [18702] = 0x9A59, [18703] = 0x9B1E,
  [18704] = 0x9B1F, [18705] = 0x9B20, [18706] = 0x9C52, [18707] = 0x9C58,
  [18708] = 0x9C50, [18709] = 0x9C4A, [18710] = 0x9C4D, [18711] = 0x9C4B,
  [18712] = 0x9C55, [18713] = 0x9C59, [18714] = 0x9C4C, [18715] = 0x9C4E,
  [18716] = 0x9DFB, [18717] = 0x9DF7, [18718] = 0x9DEF, [18719] = 0x9DE3,
  [18720] = 0x9DEB, [18721] = 0x9DF8, [18722] = 0x9DE4, [18723] = 0x9DF6,
  [18724] = 0x9DE1, [18725] = 0x9DEE, [18726] = 0x9DE6, [18727] = 0x9DF2,
  [18728] = 0x9DF0, [18729] = 0x9DE2, [18730] = 0x9DEC, [18731] = 0x9DF4,
  [18732] = 0x9DF3, [18733] = 0x9DE8, [18734] = 0x9DED, [18735] = 0x9EC2,
  [18736] = 0x9ED0, [18737] = 0x9EF2, [18738] = 0x9EF3, [18739] = 0x9F06,
  [18740] = 0x9F1C, [18741] = 0x9F38, [18742] = 0x9F37, [18743] = 0x9F36,
  [18744] = 0x9F43, [18745] = 0x9F4F, [18746] = 0x9F71, [18747] = 0x9F70,
  [18748] = 0x9F6E, [18749] = 0x9F6F, [18750] = 0x56D3, [18751] = 0x56CD,
  [18752] = 0x5B4E, [18753] = 0x5C6D, [18754] = 0x652D, [18755] = 0x66ED,
  [18756] = 0x66EE, [18757] = 0x6B13, [18758] = 0x705F, [18759] = 0x7061,
  [18760] = 0x705D, [18761] = 0x7060, [18762] = 0x7223, [18763] = 0x74DB,
  [18764] = 0x74E5, [18765] = 0x77D5, [18766] = 0x7938, [18767] = 0x79B7,
  [18768] = 0x79B6, [18769] = 0x7C6A, [18770] = 0x7E97, [18771] = 0x7F89,
  [18772] = 0x826D, [18773] = 0x8643, [18774] = 0x8838, [18775] = 0x8837,
  [18776] = 0x8835, [18777] = 0x884B, [18778] = 0x8B94, [18779] = 0x8B95,
  [18780] = 0x8E9E, [18781] = 0x8E9F, [18782] = 0x8EA0, [18783] = 0x8E9D,
  [18784] = 0x91BE, [18785] = 0x91BD, [18786] = 0x91C2, [18787] = 0x946B,
  [18788] = 0x9468, [18789] = 0x9469, [18790] = 0x96E5, [18791] = 0x9746,
  [18792] = 0x9743, [18793] = 0x9747, [18794] = 0x97C7, [18795] = 0x97E5,
  [18796] = 0x9A5E, [18797] = 0x9AD5, [18798] = 0x9B59, [18799] = 0x9C63,
  [18800] = 0x9C67, [18801] = 0x9C66, [18802] = 0x9C62, [18803] = 0x9C5E,
  [18804] = 0x9C60, [18805] = 0x9E02, [18806] = 0x9DFE, [18807] = 0x9E07,
  [18808] = 0x9E03, [18809] = 0x9E06, [18810] = 0x9E05, [18811] = 0x9E00,
  [18812] = 0x9E01, [18813] = 0x9E09, [18814] = 0x9DFF, [18815] = 0x9DFD,
  [18816] = 0x9E04, [18817] = 0x9EA0, [18818] = 0x9F1E, [18819] = 0x9F46,
  [18820] = 0x9F74, [18821] = 0x9F75, [18822] = 0x9F76, [18823] = 0x56D4,
  [18824] = 0x652E, [18825] = 0x65B8, [18826] = 0x6B18, [18827] = 0x6B19,
  [18828] = 0x6B17, [18829] = 0x6B1A, [18830] = 0x7062, [18831] = 0x7226,
  [18832] = 0x72AA, [18833] = 0x77D8, [18834] = 0x77D9, [18835] = 0x7939,
  [18836] = 0x7C69, [18837] = 0x7C6B, [18838] = 0x7CF6, [18839] = 0x7E9A,
  [18840] = 0x7E98, [18841] = 0x7E9B, [18842] = 0x7E99, [18843] = 0x81E0,
  [18844] = 0x81E1, [18845] = 0x8646, [18846] = 0x8647, [18847] = 0x8648,
  [18848] = 0x8979, [18849] = 0x897A, [18850] = 0x897C, [18851] = 0x897B,
  [18852] = 0x89FF, [18853] = 0x8B98, [18854] = 0x8B99, [18855] = 0x8EA5,
  [18856] = 0x8EA4, [18857] = 0x8EA3, [18858] = 0x946E, [18859] = 0x946D,
  [18860] = 0x946F, [18861] = 0x9471, [18862] = 0x9473, [18863] = 0x9749,
  [18864] = 0x9872, [18865] = 0x995F, [18866] = 0x9C68, [18867] = 0x9C6E,
  [18868] = 0x9C6D, [18869] = 0x9E0B, [18870] = 0x9E0D, [18871] = 0x9E10,
  [18872] = 0x9E0F, [18873] = 0x9E12, [18874] = 0x9E11, [18875] = 0x9EA1,
  [18876] = 0x9EF5, [18877] = 0x9F09, [18878] = 0x9F47, [18879] = 0x9F78,
  [18880] = 0x9F7B, [18881] = 0x9F7A, [18882] = 0x9F79, [18883] = 0x571E,
  [18884] = 0x7066, [18885] = 0x7C6F, [18886] = 0x883C, [18887] = 0x8DB2,
  [18888] = 0x8EA6, [18889] = 0x91C3, [18890] = 0x9474, [18891] = 0x9478,
  [18892] = 0x9476, [18893] = 0x9475, [18894] = 0x9A60, [18895] = 0x9C74,
  [18896] = 0x9C73, [18897] = 0x9C71, [18898] = 0x9C75, [18899] = 0x9E14,
  [18900] = 0x9E13, [18901] = 0x9EF6, [18902] = 0x9F0A, [18903] = 0x9FA4,
  [18904] = 0x7068, [18905] = 0x7065, [18906] = 0x7CF7, [18907] = 0x866A,
  [18908] = 0x883E, [18909] = 0x883D, [18910] = 0x883F, [18911] = 0x8B9E,
  [18912] = 0x8C9C, [18913] = 0x8EA9, [18914] = 0x8EC9, [18915] = 0x974B,
  [18916] = 0x9873, [18917] = 0x9874, [18918] = 0x98CC, [18919] = 0x9961,
  [18920] = 0x99AB, [18921] = 0x9A64, [18922] = 0x9A66, [18923] = 0x9A67,
  [18924] = 0x9B24, [18925] = 0x9E15, [18926] = 0x9E17, [18927] = 0x9F48,
  [18928] = 0x6207, [18929] = 0x6B1E, [18930] = 0x7227, [18931] = 0x864C,
  [18932] = 0x8EA8, [18933] = 0x9482, [18934] = 0x9480, [18935] = 0x9481,
  [18936] = 0x9A69, [18937] = 0x9A68, [18938] = 0x9B2E, [18939] = 0x9E19,
  [18940] = 0x7229, [18941] = 0x864B, [18942] = 0x8B9F, [18943] = 0x9483,
  [18944] = 0x9C79, [18945] = 0x9EB7, [18946] = 0x7675, [18947] = 0x9A6B,
  [18948] = 0x9C7A, [18949] = 0x9E1D, [18950] = 0x7069, [18951] = 0x706A,
  [18952] = 0x9EA4, [18953] = 0x9F7E, [18954] = 0x9F49, [18955] = 0x9F98,
  [18956] = 0x7881, [18957] = 0x92B9, [18958] = 0x88CF, [18959] = 0x58BB,
  [18960] = 0x6052, [18961] = 0x7CA7, [18962] = 0x5AFA, [18963] = 0x2554,
  [18964] = 0x2566, [18965] = 0x2557, [18966] = 0x2560, [18967] = 0x256C,
  [18968] = 0x2563, [18969] = 0x255A, [18970] = 0x2569, [18971] = 0x255D,
  [18972] = 0x2552, [18973] = 0x2564, [18974] = 0x2555, [18975] = 0x255E,
  [18976] = 0x256A, [18977] = 0x2561, [18978] = 0x2558, [18979] = 0x2567,
  [18980] = 0x255B, [18981] = 0x2553, [18982] = 0x2565, [18983] = 0x2556,
  [18984] = 0x255F, [18985] = 0x256B, [18986] = 0x2562, [18987] = 0x2559,
  [18988] = 0x2568, [18989] = 0x255C, [18990] = 0x2551, [18991] = 0x2550,
  [18992] = 0x256D, [18993] = 0x256E, [18994] = 0x2570, [18995] = 0x256F,
  [18996] = 0x2593, [18997] = 0xE000, [18998] = 0x92DB, [18999] = 0xE002,
  [19000] = 0xE003, [19001] = 0x854C, [19002] = 0xE005, [19003] = 0x73EF,
  [19004] = 0x51B5, [19005] = 0xE008, [19006] = 0xE009, [19007] = 0xE00A,
  [19008] = 0x9344, [19009] = 0xE00C, [19010] = 0x82EE, [19011] = 0xE00E,
  [19012] = 0x783C, [19013] = 0x6744, [19014] = 0x62DF, [19015] = 0xE012,
  [19016] = 0xE013, [19017] = 0xE014, [19018] = 0xE015, [19019] = 0xE016,
  [19020] = 0x4FAB, [19021] = 0xE018, [19022] = 0x5008, [19023] = 0xE01A,
  [19024] = 0xE01B, [19025] = 0xE01C, [19026] = 0xE01D, [19027] = 0xE01E,
  [19028] = 0xE01F, [19029] = 0xE020, [19030] = 0x5FA4, [19031] = 0xE022,
  [19032] = 0xE023, [19033] = 0x6EDB, [19034] = 0xE025, [19035] = 0xE026,
  [19036] = 0x5101, [19037] = 0xE028, [19038] = 0x510E, [19039] = 0x986C,
  [19040] = 0xE02B, [19041] = 0x8416, [19042] = 0xE02D, [19043] = 0xE02E,
  [19044] = 0x5160, [19045] = 0xE030, [19046] = 0x516A, [19047] = 0xE032,
  [19048] = 0xE033, [19049] = 0xE034, [19050] = 0xE035, [19051] = 0xE036,
  [19052] = 0xE037, [19053] = 0xE038, [19054] = 0x5B82, [19055] = 0x877D,
  [19056] = 0xE03B, [19057] = 0xE03C, [19058] = 0x51B2, [19059] = 0x51B8,
  [19060] = 0x9D34, [19061] = 0x51C9, [19062] = 0x51CF, [19063] = 0x51D1,
  [19064] = 0xE043, [19065] = 0x51D3, [19066] = 0xE045, [19067] = 0x51B3,
  [19068] = 0x51E2, [19069] = 0x5342, [19070] = 0x51ED, [19071] = 0x83CD,
  [19072] = 0x693E, [19073] = 0xE04C, [19074] = 0x5F7B, [19075] = 0x520B,
  [19076] = 0x5226, [19077] = 0x523C, [19078] = 0x52B5, [19079] = 0x5257,
  [19080] = 0x5294, [19081] = 0x52B9, [19082] = 0x52C5, [19083] = 0x7C15,
  [19084] = 0x8542, [19085] = 0x52E0, [19086] = 0x860D, [19087] = 0xE05A,
  [19088] = 0xE05B, [19089] = 0xE05C, [19090] = 0x5549, [19091] = 0x6ED9,
  [19092] = 0xE05F, [19093] = 0xE060, [19094] = 0xE061, [19095] = 0x5333,
  [19096] = 0xE063, [19097] = 0xE064, [19098] = 0x6CCB, [19099] = 0xE066,
  [19100] = 0x681B, [19101] = 0x73D5, [19102] = 0x604A, [19103] = 0xE06A,
  [19104] = 0xE06B, [19105] = 0xE06C, [19106] = 0x71DD, [19107] = 0xE06E,
  [19108] = 0x536D, [19109] = 0x5374, [19110] = 0xE071, [19111] = 0x537E,
  [19112] = 0xE073, [19113] = 0xE074, [19114] = 0xE075, [19115] = 0x77E6,
  [19116] = 0x5393, [19117] = 0xE078, [19118] = 0x53A0, [19119] = 0x53AB,
  [19120] = 0x53AE, [19121] = 0x73A7, [19122] = 0xE07D, [19123] = 0xE07E,
  [19124] = 0x739C, [19125] = 0x53C1, [19126] = 0x53C5, [19127] = 0x6C49,
  [19128] = 0x4E49, [19129] = 0x57FE, [19130] = 0x53D9, [19131] = 0xE086,
  [19132] = 0xE087, [19133] = 0x53E0, [19134] = 0xE089, [19135] = 0xE08A,
  [19136] = 0x53F6, [19137] = 0xE08C, [19138] = 0x5413, [19139] = 0x7079,
  [19140] = 0x552B, [19141] = 0x6657, [19142] = 0x6D5B, [19143] = 0x546D,
  [19144] = 0xE093, [19145] = 0xE094, [19146] = 0x555D, [19147] = 0x548F,
  [19148] = 0x54A4, [19149] = 0xE098, [19150] = 0xE099, [19151] = 0xE09A,
  [19152] = 0xE09B, [19153] = 0xE09C, [19154] = 0xE09D, [19155] = 0xE09E,
  [19156] = 0x5547, [19157] = 0xE0A0, [19158] = 0x542F, [19159] = 0x7417,
  [19160] = 0x5586, [19161] = 0x55A9, [19162] = 0xE0A5, [19163] = 0xE0A6,
  [19164] = 0xE0A7, [19165] = 0xE0A8, [19166] = 0xE0A9, [19167] = 0x66B3,
  [19168] = 0xE0AB, [19169] = 0x5637, [19170] = 0x66CD, [19171] = 0xE0AE,
  [19172] = 0x66A4, [19173] = 0x66AD, [19174] = 0x564D, [19175] = 0x564F,
  [19176] = 0x78F1, [19177] = 0x56F1, [19178] = 0x9787, [19179] = 0x53FE,
  [19180] = 0x5700, [19181] = 0x56EF, [19182] = 0x56ED, [19183] = 0xE0BA,
  [19184] = 0xE0BB, [19185] = 0xE0BC, [19186] = 0x5746, [19187] = 0xE0BE,
  [19188] = 0x6C6E, [19189] = 0x708B, [19190] = 0x5742, [19191] = 0xE0C2,
  [19192] = 0xE0C3, [19193] = 0x57E6, [19194] = 0xE0C5, [19195] = 0x5803,
  [19196] = 0xE0C7, [19197] = 0xE0C8, [19198] = 0x5826, [19199] = 0xE0CA,
  [19200] = 0x585C, [19201] = 0x58AA, [19202] = 0xE0CD, [19203] = 0x58E0,
  [19204] = 0x58DC, [19205] = 0xE0D0, [19206] = 0x58FB, [19207] = 0x5BFF,
  [19208] = 0x5743, [19209] = 0xE0D4, [19210] = 0xE0D5, [19211] = 0x93D3,
  [19212] = 0xE0D7, [19213] = 0x591F, [19214] = 0x68A6, [19215] = 0xE0DA,
  [19216] = 0x6E59, [19217] = 0xE0DC, [19218] = 0x5A24, [19219] = 0x5553,
  [19220] = 0xE0DF, [19221] = 0x8505, [19222] = 0x59C9, [19223] = 0xE0E2,
  [19224] = 0xE0E3, [19225] = 0xE0E4, [19226] = 0xE0E5, [19227] = 0x59D9,
  [19228] = 0xE0E7, [19229] = 0xE0E8, [19230] = 0xE0E9, [19231] = 0x6D71,
  [19232] = 0xE0EB, [19233] = 0xE0EC, [19234] = 0x59F9, [19235] = 0xE0EE,
  [19236] = 0x5AAB, [19237] = 0x5A63, [19238] = 0xE0F1, [19239] = 0xE0F2,
  [19240] = 0xE0F3, [19241] = 0xE0F4, [19242] = 0x5A96, [19243] = 0x7465,
  [19244] = 0x5AD3, [19245] = 0xE0F8, [19246] = 0xE0F9, [19247] = 0xE0FA,
  [19248] = 0xE0FB, [19249] = 0xE0FC, [19250] = 0xE0FD, [19251] = 0x5E83,
  [19252] = 0x52D0, [19253] = 0x5B76, [19254] = 0x6588, [19255] = 0x5B7C,
  [19256] = 0xE103, [19257] = 0xE104, [19258] = 0xE105, [19259] = 0xE106,
  [19260] = 0x5BD5, [19261] = 0x6160, [19262] = 0xE109, [19263] = 0xE10A,
  [19264] = 0xE10B, [19265] = 0x5BF3, [19266] = 0x5B9D, [19267] = 0xE10E,
  [19268] = 0x5C05, [19269] = 0xE110, [19270] = 0x5C13, [19271] = 0x73CE,
  [19272] = 0x5C14, [19273] = 0xE114, [19274] = 0xE115, [19275] = 0x5C49,
  [19276] = 0xE117, [19277] = 0x5C85, [19278] = 0x5CE9, [19279] = 0x5CEF,
  [19280] = 0x5D8B, [19281] = 0xE11C, [19282] = 0xE11D, [19283] = 0x5D10,
  [19284] = 0x5D18, [19285] = 0x5D46, [19286] = 0xE121, [19287] = 0x5CBA,
  [19288] = 0x5DD7, [19289] = 0x82FC, [19290] = 0xE125, [19291] = 0xE126,
  [19292] = 0xE127, [19293] = 0xE128, [19294] = 0x8287, [19295] = 0xE12A,
  [19296] = 0xE12B, [19297] = 0x5E2E, [19298] = 0x6A8A, [19299] = 0xE12E,
  [19300] = 0x5E7A, [19301] = 0xE130, [19302] = 0xE131, [19303] = 0x53A6,
  [19304] = 0x4EB7, [19305] = 0xE134, [19306] = 0x53A8, [19307] = 0xE136,
  [19308] = 0x5E09, [19309] = 0x5EF4, [19310] = 0xE139, [19311] = 0x5EF9,
  [19312] = 0x5EFB, [19313] = 0xE13C, [19314] = 0x5EFC, [19315] = 0x683E,
  [19316] = 0x941B, [19317] = 0x5F0D, [19318] = 0xE141, [19319] = 0xE142,
  [19320] = 0xE143, [19321] = 0xE144, [19322] = 0xE145, [19323] = 0x5F3A,
  [19324] = 0xE147, [19325] = 0xE148, [19326] = 0xE149, [19327] = 0xE14A,
  [19328] = 0x5F63, [19329] = 0x97BD, [19330] = 0xE14D, [19331] = 0x5F72,
  [19332] = 0x9340, [19333] = 0xE150, [19334] = 0x5FA7, [19335] = 0x5DB6,
  [19336] = 0xE153, [19337] = 0xE154, [19338] = 0xE155, [19339] = 0xE156,
  [19340] = 0xE157, [19341] = 0x91D6, [19342] = 0xE159, [19343] = 0xE15A,
  [19344] = 0x6031, [19345] = 0x6685, [19346] = 0xE15D, [19347] = 0xE15E,
  [19348] = 0xE15F, [19349] = 0xE160, [19350] = 0x5790, [19351] = 0xE162,
  [19352] = 0x7971, [19353] = 0xE164, [19354] = 0x609E, [19355] = 0xE166,
  [19356] = 0x60B3, [19357] = 0xE168, [19358] = 0xE169, [19359] = 0xE16A,
  [19360] = 0x74A4, [19361] = 0x50E1, [19362] = 0x5AA0, [19363] = 0x6164,
  [19364] = 0x8424, [19365] = 0x6142, [19366] = 0xE171, [19367] = 0xE172,
  [19368] = 0x6181, [19369] = 0x51F4, [19370] = 0xE175, [19371] = 0x6187,
  [19372] = 0x5BAA, [19373] = 0xE178, [19374] = 0xE179, [19375] = 0x61D3,
  [19376] = 0xE17B, [19377] = 0xE17C, [19378] = 0x61D0, [19379] = 0xE17E,
  [19380] = 0xE17F, [19381] = 0xE180, [19382] = 0x6023, [19383] = 0x615C,
  [19384] = 0x651E, [19385] = 0x638B, [19386] = 0xE185, [19387] = 0x62C5,
  [19388] = 0xE187, [19389] = 0x62D5, [19390] = 0xE189, [19391] = 0x636C,
  [19392] = 0xE18B, [19393] = 0xE18C, [19394] = 0x6438, [19395] = 0x63F8,
  [19396] = 0xE18F, [19397] = 0xE190, [19398] = 0xE191, [19399] = 0x6F8A,
  [19400] = 0xE193, [19401] = 0x9814, [19402] = 0xE195, [19403] = 0xE196,
  [19404] = 0x64E1, [19405] = 0x64E5, [19406] = 0x947B, [19407] = 0xE19A,
  [19408] = 0x643A, [19409] = 0xE19C, [19410] = 0x654D, [19411] = 0x6F16,
  [19412] = 0xE19F, [19413] = 0xE1A0, [19414] = 0x6585, [19415] = 0x656D,
  [19416] = 0x655F, [19417] = 0xE1A4, [19418] = 0x65B5, [19419] = 0xE1A6,
  [19420] = 0xE1A7, [19421] = 0x65D1, [19422] = 0xE1A9, [19423] = 0xE1AA,
  [19424] = 0x65E0, [19425] = 0x65E3, [19426] = 0x5FDF, [19427] = 0xE1AE,
  [19428] = 0x6618, [19429] = 0xE1B0, [19430] = 0xE1B1, [19431] = 0x6644,
  [19432] = 0xE1B3, [19433] = 0xE1B4, [19434] = 0x664B, [19435] = 0xE1B6,
  [19436] = 0x6667, [19437] = 0xE1B8, [19438] = 0x6673, [19439] = 0xE1BA,
  [19440] = 0xE1BB, [19441] = 0xE1BC, [19442] = 0xE1BD, [19443] = 0xE1BE,
  [19444] = 0xE1BF, [19445] = 0x77C5, [19446] = 0xE1C1, [19447] = 0x99A4,
  [19448] = 0x6702, [19449] = 0xE1C4, [19450] = 0xE1C5, [19451] = 0xE1C6,
  [19452] = 0x69FA, [19453] = 0xE1C8, [19454] = 0xE1C9, [19455] = 0x6767,
  [19456] = 0x6762, [19457] = 0xE1CC, [19458] = 0xE1CD, [19459] = 0x67D7,
  [19460] = 0xE1CF, [19461] = 0x6822, [19462] = 0x6E50, [19463] = 0x923C,
  [19464] = 0x6801, [19465] = 0xE1D4, [19466] = 0xE1D5, [19467] = 0x685D,
  [19468] = 0xE1D7, [19469] = 0x69E1, [19470] = 0x6A0B, [19471] = 0xE1DA,
  [19472] = 0x6973, [19473] = 0x68C3, [19474] = 0xE1DD, [19475] = 0x6901,
  [19476] = 0x6900, [19477] = 0xE1E0, [19478] = 0xE1E1, [19479] = 0xE1E2,
  [19480] = 0xE1E3, [19481] = 0x67AC, [19482] = 0x6961, [19483] = 0xE1E6,
  [19484] = 0xE1E7, [19485] = 0x6936, [19486] = 0x6998, [19487] = 0xE1EA,
  [19488] = 0xE1EB, [19489] = 0x8363, [19490] = 0x5090, [19491] = 0x69F9,
  [19492] = 0xE1EF, [19493] = 0xE1F0, [19494] = 0x6A45, [19495] = 0xE1F2,
  [19496] = 0x6A9D, [19497] = 0xE1F4, [19498] = 0x67B1, [19499] = 0x6AC8,
  [19500] = 0xE1F7, [19501] = 0xE1F8, [19502] = 0x6B1D, [19503] = 0xE1FA,
  [19504] = 0x60DE, [19505] = 0x6B35, [19506] = 0x6B74, [19507] = 0xE1FE,
  [19508] = 0x6EB5, [19509] = 0xE200, [19510] = 0xE201, [19511] = 0xE202,
  [19512] = 0xE203, [19513] = 0x5421, [19514] = 0xE205, [19515] = 0x6BE1,
  [19516] = 0xE207, [19517] = 0x6BDC, [19518] = 0x6C37, [19519] = 0xE20A,
  [19520] = 0xE20B, [19521] = 0xE20C, [19522] = 0x6C5A, [19523] = 0x8226,
  [19524] = 0x6C79, [19525] = 0xE210, [19526] = 0xE211, [19527] = 0xE212,
  [19528] = 0xE213, [19529] = 0xE214, [19530] = 0xE215, [19531] = 0xE216,
  [19532] = 0xE217, [19533] = 0xE218, [19534] = 0xE219, [19535] = 0x9B83,
  [19536] = 0xE21B, [19537] = 0xE21C, [19538] = 0x7F8F, [19539] = 0x6837,
  [19540] = 0xE21F, [19541] = 0xE220, [19542] = 0xE221, [19543] = 0x6D96,
  [19544] = 0x6D5C, [19545] = 0x6E7C, [19546] = 0x6F04, [19547] = 0xE226,
  [19548] = 0xE227, [19549] = 0xE228, [19550] = 0x8533, [19551] = 0xE22A,
  [19552] = 0x51C7, [19553] = 0xE22C, [19554] = 0xE22D, [19555] = 0x842E,
  [19556] = 0xE22F, [19557] = 0xE230, [19558] = 0xE231, [19559] = 0x7453,
  [19560] = 0xE233, [19561] = 0x79CC, [19562] = 0x6E4F, [19563] = 0x5A91,
  [19564] = 0xE237, [19565] = 0x6FF8, [19566] = 0xE239, [19567] = 0x6F9D,
  [19568] = 0xE23B, [19569] = 0x6EFA, [19570] = 0xE23D, [19571] = 0xE23E,
  [19572] = 0xE23F, [19573] = 0x93F0, [19574] = 0x6F44, [19575] = 0x6F5C,
  [19576] = 0xE243, [19577] = 0x6F74, [19578] = 0xE245, [19579] = 0xE246,
  [19580] = 0x6F9F, [19581] = 0xE248, [19582] = 0x6FD3, [19583] = 0xE24A,
  [19584] = 0xE24B, [19585] = 0xE24C, [19586] = 0xE24D, [19587] = 0xE24E,
  [19588] = 0xE24F, [19589] = 0x51DF, [19590] = 0xE251, [19591] = 0xE252,
  [19592] = 0xE253, [19593] = 0xE254, [19594] = 0x704B, [19595] = 0x707E,
  [19596] = 0x70A7, [19597] = 0x7081, [19598] = 0x70CC, [19599] = 0x70D5,
  [19600] = 0x70D6, [19601] = 0x70DF, [19602] = 0xE25D, [19603] = 0xE25E,
  [19604] = 0x71B4, [19605] = 0x7196, [19606] = 0xE261, [19607] = 0x712B,
  [19608] = 0x7145, [19609] = 0x5A88, [19610] = 0x714A, [19611] = 0xE266,
  [19612] = 0x5C9C, [19613] = 0xE268, [19614] = 0x714F, [19615] = 0x9362,
  [19616] = 0xE26B, [19617] = 0x712C, [19618] = 0xE26D, [19619] = 0xE26E,
  [19620] = 0xE26F, [19621] = 0x71BA, [19622] = 0xE271, [19623] = 0x70BD,
  [19624] = 0x720E, [19625] = 0x9442, [19626] = 0x7215, [19627] = 0x5911,
  [19628] = 0x9443, [19629] = 0x7224, [19630] = 0x9341, [19631] = 0xE27A,
  [19632] = 0x722E, [19633] = 0x7240, [19634] = 0xE27D, [19635] = 0x68BD,
  [19636] = 0x7255, [19637] = 0x7257, [19638] = 0xE281, [19639] = 0xE282,
  [19640] = 0x680D, [19641] = 0x6F3D, [19642] = 0x7282, [19643] = 0xE286,
  [19644] = 0x732B, [19645] = 0xE288, [19646] = 0xE289, [19647] = 0xE28A,
  [19648] = 0xE28B, [19649] = 0x7328, [19650] = 0x732E, [19651] = 0x73CF,
  [19652] = 0x73AA, [19653] = 0xE290, [19654] = 0xE291, [19655] = 0x73C9,
  [19656] = 0x7449, [19657] = 0xE294, [19658] = 0xE295, [19659] = 0xE296,
  [19660] = 0x6623, [19661] = 0xE298, [19662] = 0xE299, [19663] = 0xE29A,
  [19664] = 0xE29B, [19665] = 0x73F7, [19666] = 0x7415, [19667] = 0x6903,
  [19668] = 0xE29F, [19669] = 0x7439, [19670] = 0xE2A1, [19671] = 0xE2A2,
  [19672] = 0xE2A3, [19673] = 0xE2A4, [19674] = 0x7460, [19675] = 0xE2A6,
  [19676] = 0x7447, [19677] = 0x73E4, [19678] = 0x7476, [19679] = 0x83B9,
  [19680] = 0x746C, [19681] = 0xE2AC, [19682] = 0x7474, [19683] = 0x93F1,
  [19684] = 0x6A2C, [19685] = 0x7482, [19686] = 0xE2B1, [19687] = 0xE2B2,
  [19688] = 0xE2B3, [19689] = 0xE2B4, [19690] = 0xE2B5, [19691] = 0x5B46,
  [19692] = 0xE2B7, [19693] = 0xE2B8, [19694] = 0x74C8, [19695] = 0xE2BA,
  [19696] = 0x750E, [19697] = 0xE2BC, [19698] = 0x751E, [19699] = 0xE2BE,
  [19700] = 0xE2BF, [19701] = 0x5BD7, [19702] = 0xE2C1, [19703] = 0x9385,
  [19704] = 0x754D, [19705] = 0x754A, [19706] = 0x7567, [19707] = 0x756E,
  [19708] = 0xE2C7, [19709] = 0xE2C8, [19710] = 0xE2C9, [19711] = 0x758E,
  [19712] = 0x745D, [19713] = 0x759E, [19714] = 0x75B4, [19715] = 0x7602,
  [19716] = 0x762C, [19717] = 0x7651, [19718] = 0x764F, [19719] = 0x766F,
  [19720] = 0x7676, [19721] = 0xE2D4, [19722] = 0x7690, [19723] = 0x81EF,
  [19724] = 0xE2D7, [19725] = 0xE2D8, [19726] = 0xE2D9, [19727] = 0x76A1,
  [19728] = 0x76A5, [19729] = 0x76B7, [19730] = 0x76CC, [19731] = 0xE2DE,
  [19732] = 0x8462, [19733] = 0xE2E0, [19734] = 0xE2E1, [19735] = 0xE2E2,
  [19736] = 0x771E, [19737] = 0x7726, [19738] = 0x7740, [19739] = 0x64AF,
  [19740] = 0xE2E7, [19741] = 0x7758, [19742] = 0xE2E9, [19743] = 0x77AF,
  [19744] = 0xE2EB, [19745] = 0xE2EC, [19746] = 0xE2ED, [19747] = 0x77F4,
  [19748] = 0xE2EF, [19749] = 0xE2F0, [19750] = 0xE2F1, [19751] = 0x68CA,
  [19752] = 0x78AF, [19753] = 0x78C7, [19754] = 0x78D3, [19755] = 0x96A5,
  [19756] = 0x792E, [19757] = 0xE2F8, [19758] = 0x78D7, [19759] = 0x7934,
  [19760] = 0x78B1, [19761] = 0xE2FC, [19762] = 0x8FB8, [19763] = 0x8884,
  [19764] = 0xE2FF, [19765] = 0xE300, [19766] = 0xE301, [19767] = 0x7986,
  [19768] = 0x8900, [19769] = 0x6902, [19770] = 0x7980, [19771] = 0xE306,
  [19772] = 0x799D, [19773] = 0xE308, [19774] = 0x793C, [19775] = 0x79A9,
  [19776] = 0x6E2A, [19777] = 0xE30C, [19778] = 0xE30D, [19779] = 0x79C6,
  [19780] = 0xE30F, [19781] = 0x79D4,
};


/* The mapping from Unicode to Big5 is not easy since we have several
   separate blocks.  First a block of characters with the second byte
   being zero. */
static const char from_ucs4_tab1[256][2] =
{
  /* 0x00 */ "\x00\x00", "\x01\x00", "\x02\x00", "\x03\x00", "\x04\x00",
  /* 0x05 */ "\x05\x00", "\x06\x00", "\x07\x00", "\x08\x00", "\x09\x00",
  /* 0x0a */ "\x0a\x00", "\x0b\x00", "\x0c\x00", "\x0d\x00", "\x0e\x00",
  /* 0x0f */ "\x0f\x00", "\x10\x00", "\x11\x00", "\x12\x00", "\x13\x00",
  /* 0x14 */ "\x14\x00", "\x15\x00", "\x16\x00", "\x17\x00", "\x18\x00",
  /* 0x19 */ "\x19\x00", "\x1a\x00", "\x1b\x00", "\x1c\x00", "\x1d\x00",
  /* 0x1e */ "\x1e\x00", "\x1f\x00", "\x20\x00", "\x21\x00", "\x22\x00",
  /* 0x23 */ "\x23\x00", "\x24\x00", "\x25\x00", "\x26\x00", "\x27\x00",
  /* 0x28 */ "\x28\x00", "\x29\x00", "\x2a\x00", "\x2b\x00", "\x2c\x00",
  /* 0x2d */ "\x2d\x00", "\x2e\x00", "\x2f\x00", "\x30\x00", "\x31\x00",
  /* 0x32 */ "\x32\x00", "\x33\x00", "\x34\x00", "\x35\x00", "\x36\x00",
  /* 0x37 */ "\x37\x00", "\x38\x00", "\x39\x00", "\x3a\x00", "\x3b\x00",
  /* 0x3c */ "\x3c\x00", "\x3d\x00", "\x3e\x00", "\x3f\x00", "\x40\x00",
  /* 0x41 */ "\x41\x00", "\x42\x00", "\x43\x00", "\x44\x00", "\x45\x00",
  /* 0x46 */ "\x46\x00", "\x47\x00", "\x48\x00", "\x49\x00", "\x4a\x00",
  /* 0x4b */ "\x4b\x00", "\x4c\x00", "\x4d\x00", "\x4e\x00", "\x4f\x00",
  /* 0x50 */ "\x50\x00", "\x51\x00", "\x52\x00", "\x53\x00", "\x54\x00",
  /* 0x55 */ "\x55\x00", "\x56\x00", "\x57\x00", "\x58\x00", "\x59\x00",
  /* 0x5a */ "\x5a\x00", "\x5b\x00", "\x5c\x00", "\x5d\x00", "\x5e\x00",
  /* 0x5f */ "\x5f\x00", "\x60\x00", "\x61\x00", "\x62\x00", "\x63\x00",
  /* 0x64 */ "\x64\x00", "\x65\x00", "\x66\x00", "\x67\x00", "\x68\x00",
  /* 0x69 */ "\x69\x00", "\x6a\x00", "\x6b\x00", "\x6c\x00", "\x6d\x00",
  /* 0x6e */ "\x6e\x00", "\x6f\x00", "\x70\x00", "\x71\x00", "\x72\x00",
  /* 0x73 */ "\x73\x00", "\x74\x00", "\x75\x00", "\x76\x00", "\x77\x00",
  /* 0x78 */ "\x78\x00", "\x79\x00", "\x7a\x00", "\x7b\x00", "\x7c\x00",
  /* 0x7d */ "\x7d\x00", "\x7e\x00", "\x7f\x00", "\x80\x00", "\x81\x00",
  /* 0x82 */ "\x82\x00", "\x83\x00", "\x84\x00", "\x85\x00", "\x86\x00",
  /* 0x87 */ "\x87\x00", "\x88\x00", "\x89\x00", "\x8a\x00", "\x8b\x00",
  /* 0x8c */ "\x8c\x00", "\x8d\x00", "\x8e\x00", "\x8f\x00", "\x90\x00",
  /* 0x91 */ "\x91\x00", "\x92\x00", "\x93\x00", "\x94\x00", "\x95\x00",
  /* 0x96 */ "\x96\x00", "\x97\x00", "\x98\x00", "\x99\x00", "\x9a\x00",
  /* 0x9b */ "\x9b\x00", "\x9c\x00", "\x9d\x00", "\x9e\x00", "\x9f\x00",
  /* 0xa0 */ "\xa0\x00", "\x00\x00", "\xa2\x46", "\xa2\x47", "\x00\x00",
  /* 0xa5 */ "\xa2\x44", "\x00\x00", "\xa1\xb1", "\x00\x00", "\x00\x00",
  /* 0xaa */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xaf */ "\x00\x00", "\xa2\x58", "\xa1\xd3", "\x00\x00", "\x00\x00",
  /* 0xb4 */ "\x00\x00", "\x00\x00", "\x00\x00", "\xa1\x50", "\x00\x00",
  /* 0xb9 */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xbe */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xc3 */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xc8 */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xcd */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xd2 */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xd7 */ "\xa1\xd1", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xdc */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xe1 */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xe6 */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xeb */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xf0 */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0xf5 */ "\x00\x00", "\x00\x00", "\xa1\xd2"
};

/* The next table is for a few phonetic modifiers and diacritics.  They
   start at U+02c7.  */
static const char from_ucs4_tab2[][2] =
{
  /* 0x2c7 */ "\xa3\xbe", "\x00\x00", "\xa3\xbc", "\xa3\xbd", "\xa3\xbf",
  /* 0x2cc */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0x2d1 */ "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00", "\x00\x00",
  /* 0x2d6 */ "\x00\x00", "\x00\x00", "\x00\x00", "\xa3\xbb"
};

/* The next table is for the greek alphabet.  The table can be created
   using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0x0[34]' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0x391,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab3[][2] =
{
  [0x0000] = "\xa3\x44", [0x0001] = "\xa3\x45", [0x0002] = "\xa3\x46",
  [0x0003] = "\xa3\x47", [0x0004] = "\xa3\x48", [0x0005] = "\xa3\x49",
  [0x0006] = "\xa3\x4a", [0x0007] = "\xa3\x4b", [0x0008] = "\xa3\x4c",
  [0x0009] = "\xa3\x4d", [0x000a] = "\xa3\x4e", [0x000b] = "\xa3\x4f",
  [0x000c] = "\xa3\x50", [0x000d] = "\xa3\x51", [0x000e] = "\xa3\x52",
  [0x000f] = "\xa3\x53", [0x0010] = "\xa3\x54", [0x0012] = "\xa3\x55",
  [0x0013] = "\xa3\x56", [0x0014] = "\xa3\x57", [0x0015] = "\xa3\x58",
  [0x0016] = "\xa3\x59", [0x0017] = "\xa3\x5a", [0x0018] = "\xa3\x5b",
  [0x0020] = "\xa3\x5c", [0x0021] = "\xa3\x5d", [0x0022] = "\xa3\x5e",
  [0x0023] = "\xa3\x5f", [0x0024] = "\xa3\x60", [0x0025] = "\xa3\x61",
  [0x0026] = "\xa3\x62", [0x0027] = "\xa3\x63", [0x0028] = "\xa3\x64",
  [0x0029] = "\xa3\x65", [0x002a] = "\xa3\x66", [0x002b] = "\xa3\x67",
  [0x002c] = "\xa3\x68", [0x002d] = "\xa3\x69", [0x002e] = "\xa3\x6a",
  [0x002f] = "\xa3\x6b", [0x0030] = "\xa3\x6c", [0x0032] = "\xa3\x6d",
  [0x0033] = "\xa3\x6e", [0x0034] = "\xa3\x6f", [0x0035] = "\xa3\x70",
  [0x0036] = "\xa3\x71", [0x0037] = "\xa3\x72", [0x0038] = "\xa3\x73",
  [0x0070] = "\xc7\xb3", [0x0083] = "\xc7\xb1", [0x0084] = "\xc7\xb2",
  [0x0085] = "\xc7\xb4", [0x0086] = "\xc7\xb5", [0x0087] = "\xc7\xb6",
  [0x0088] = "\xc7\xb7", [0x0089] = "\xc7\xb8", [0x008a] = "\xc7\xb9",
  [0x008b] = "\xc7\xba", [0x0092] = "\xc7\xbb", [0x0093] = "\xc7\xbc",
  [0x0094] = "\xc7\xbd", [0x0095] = "\xc7\xbe", [0x0096] = "\xc7\xbf",
  [0x0097] = "\xc7\xc0", [0x0098] = "\xc7\xc1", [0x0099] = "\xc7\xc2",
  [0x009a] = "\xc7\xc3", [0x009b] = "\xc7\xc4", [0x009c] = "\xc7\xc5",
  [0x009d] = "\xc7\xc6", [0x009e] = "\xc7\xc7", [0x009f] = "\xc7\xc8",
  [0x00a0] = "\xc7\xc9", [0x00a1] = "\xc7\xca", [0x00a2] = "\xc7\xcb",
  [0x00a3] = "\xc7\xcc", [0x00a4] = "\xc7\xcd", [0x00a5] = "\xc7\xcf",
  [0x00a6] = "\xc7\xd0", [0x00a7] = "\xc7\xd1", [0x00a8] = "\xc7\xd2",
  [0x00a9] = "\xc7\xd3", [0x00aa] = "\xc7\xd4", [0x00ab] = "\xc7\xd5",
  [0x00ac] = "\xc7\xd6", [0x00ad] = "\xc7\xd7", [0x00ae] = "\xc7\xd8",
  [0x00af] = "\xc7\xd9", [0x00b0] = "\xc7\xda", [0x00b1] = "\xc7\xdb",
  [0x00b2] = "\xc7\xdc", [0x00b3] = "\xc7\xdd", [0x00b4] = "\xc7\xde",
  [0x00b5] = "\xc7\xdf", [0x00b6] = "\xc7\xe0", [0x00b7] = "\xc7\xe1",
  [0x00b8] = "\xc7\xe2", [0x00b9] = "\xc7\xe3", [0x00ba] = "\xc7\xe4",
  [0x00bb] = "\xc7\xe5", [0x00bc] = "\xc7\xe6", [0x00bd] = "\xc7\xe7",
  [0x00be] = "\xc7\xe8", [0x00c0] = "\xc7\xce",
};

/* The next table is for the general punctuation characters.
   The table can be created using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0x20' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0x2013,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab4[][2] =
{

  [0x0000] = "\xa1\x56", [0x0001] = "\xa1\x58", [0x0005] = "\xa1\xa5",
  [0x0006] = "\xa1\xa6", [0x0009] = "\xa1\xa7", [0x000a] = "\xa1\xa8",
  [0x000f] = "\xa1\x45", [0x0012] = "\xa1\x4c", [0x0013] = "\xa1\x4b",
  [0x001f] = "\xa1\xac", [0x0022] = "\xa1\xab", [0x0028] = "\xa1\xb0",
  [0x002b] = "\xa1\xc2",
};

/* Arrow characters, starting at U+2190.  */
static const char from_ucs4_tab5[][2] =
{
  [0x00] = "\xa1\xf6", [0x01] = "\xa1\xf4", [0x02] = "\xa1\xf7",
  [0x03] = "\xa1\xf5", [0x06] = "\xa1\xf8", [0x07] = "\xa1\xf9",
  [0x08] = "\xa1\xfb", [0x09] = "\xa1\xfa"
};

/* Mathematical operators, starting at U+221a.
   The table can be created using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0x22' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0x221a,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab6[][2] =
{
  [0x0000] = "\xa1\xd4", [0x0004] = "\xa1\xdb", [0x0005] = "\xa1\xe8",
  [0x0006] = "\xa1\xe7", [0x0009] = "\xa1\xfd", [0x000b] = "\xa1\xfc",
  [0x000f] = "\xa1\xe4", [0x0010] = "\xa1\xe5", [0x0011] = "\xa1\xec",
  [0x0014] = "\xa1\xed", [0x001a] = "\xa1\xef", [0x001b] = "\xa1\xee",
  [0x0022] = "\xa1\xe3", [0x0038] = "\xa1\xdc", [0x0046] = "\xa1\xda",
  [0x0047] = "\xa1\xdd", [0x004c] = "\xa1\xd8", [0x004d] = "\xa1\xd9",
  [0x008b] = "\xa1\xe6", [0x00a5] = "\xa1\xe9",
};

/* Circled and parenthesized numbers.
   The table can be created using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0x24' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0x2460,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab7[][2] =
{
  [0x0000] = "\xc7\xe9", [0x0001] = "\xc7\xea", [0x0002] = "\xc7\xeb",
  [0x0003] = "\xc7\xec", [0x0004] = "\xc7\xed", [0x0005] = "\xc7\xee",
  [0x0006] = "\xc7\xef", [0x0007] = "\xc7\xf0", [0x0008] = "\xc7\xf1",
  [0x0009] = "\xc7\xf2", [0x0014] = "\xc7\xf3", [0x0015] = "\xc7\xf4",
  [0x0016] = "\xc7\xf5", [0x0017] = "\xc7\xf6", [0x0018] = "\xc7\xf7",
  [0x0019] = "\xc7\xf8", [0x001a] = "\xc7\xf9", [0x001b] = "\xc7\xfa",
  [0x001c] = "\xc7\xfb", [0x001d] = "\xc7\xfc",
};

/* Form and chart components.
   The table can be created using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0x2[56]' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0x2500,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab8[][2] =
{
  [0x0000] = "\xa2\x77", [0x0002] = "\xa2\x78", [0x000c] = "\xa2\x7a",
  [0x0010] = "\xa2\x7b", [0x0014] = "\xa2\x7c", [0x0018] = "\xa2\x7d",
  [0x001c] = "\xa2\x75", [0x0024] = "\xa2\x74", [0x002c] = "\xa2\x73",
  [0x0034] = "\xa2\x72", [0x003c] = "\xa2\x71", [0x0050] = "\xa2\xa4",
  [0x005e] = "\xa2\xa5", [0x0061] = "\xa2\xa7", [0x006a] = "\xa2\xa6",
  [0x006d] = "\xa2\x7e", [0x006e] = "\xa2\xa1", [0x006f] = "\xa2\xa3",
  [0x0070] = "\xa2\xa2", [0x0071] = "\xa2\xac", [0x0072] = "\xa2\xad",
  [0x0073] = "\xa2\xae", [0x0081] = "\xa2\x62", [0x0082] = "\xa2\x63",
  [0x0083] = "\xa2\x64", [0x0084] = "\xa2\x65", [0x0085] = "\xa2\x66",
  [0x0086] = "\xa2\x67", [0x0087] = "\xa2\x68", [0x0088] = "\xa2\x69",
  [0x0089] = "\xa2\x70", [0x008a] = "\xa2\x6f", [0x008b] = "\xa2\x6e",
  [0x008c] = "\xa2\x6d", [0x008d] = "\xa2\x6c", [0x008e] = "\xa2\x6b",
  [0x008f] = "\xa2\x6a", [0x0094] = "\xa2\x76", [0x0095] = "\xa2\x79",
  [0x00a0] = "\xa1\xbd", [0x00a1] = "\xa1\xbc", [0x00b2] = "\xa1\xb6",
  [0x00b3] = "\xa1\xb5", [0x00bc] = "\xa1\xbf", [0x00bd] = "\xa1\xbe",
  [0x00c6] = "\xa1\xbb", [0x00c7] = "\xa1\xba", [0x00cb] = "\xa1\xb3",
  [0x00ce] = "\xa1\xb7", [0x00cf] = "\xa1\xb4", [0x00e2] = "\xa2\xa8",
  [0x00e3] = "\xa2\xa9", [0x00e4] = "\xa2\xab", [0x00e5] = "\xa2\xaa",
  [0x0105] = "\xa1\xb9", [0x0106] = "\xa1\xb8", [0x0109] = "\xa1\xf3",
  [0x0140] = "\xa1\xf0", [0x0141] = "\xa1\xf2", [0x0142] = "\xa1\xf1",
};

/* CJK symbols and punctuation.
   The table can be created using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0x3[01]' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0x3000,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab9[][2] =
{
  [0x0000] = "\xa1\x40", [0x0001] = "\xa1\x42", [0x0002] = "\xa1\x43",
  [0x0003] = "\xa1\xb2", [0x0005] = "\xc6\xa4", [0x0008] = "\xa1\x71",
  [0x0009] = "\xa1\x72", [0x000a] = "\xa1\x6d", [0x000b] = "\xa1\x6e",
  [0x000c] = "\xa1\x75", [0x000d] = "\xa1\x76", [0x000e] = "\xa1\x79",
  [0x000f] = "\xa1\x7a", [0x0010] = "\xa1\x69", [0x0011] = "\xa1\x6a",
  [0x0012] = "\xa2\x45", [0x0014] = "\xa1\x65", [0x0015] = "\xa1\x66",
  [0x001d] = "\xa1\xa9", [0x001e] = "\xa1\xaa", [0x0021] = "\xa2\xc3",
  [0x0022] = "\xa2\xc4", [0x0023] = "\xa2\xc5", [0x0024] = "\xa2\xc6",
  [0x0025] = "\xa2\xc7", [0x0026] = "\xa2\xc8", [0x0027] = "\xa2\xc9",
  [0x0028] = "\xa2\xca", [0x0029] = "\xa2\xcb", [0x0041] = "\xc6\xa5",
  [0x0042] = "\xc6\xa6", [0x0043] = "\xc6\xa7", [0x0044] = "\xc6\xa8",
  [0x0045] = "\xc6\xa9", [0x0046] = "\xc6\xaa", [0x0047] = "\xc6\xab",
  [0x0048] = "\xc6\xac", [0x0049] = "\xc6\xad", [0x004a] = "\xc6\xae",
  [0x004b] = "\xc6\xaf", [0x004c] = "\xc6\xb0", [0x004d] = "\xc6\xb1",
  [0x004e] = "\xc6\xb2", [0x004f] = "\xc6\xb3", [0x0050] = "\xc6\xb4",
  [0x0051] = "\xc6\xb5", [0x0052] = "\xc6\xb6", [0x0053] = "\xc6\xb7",
  [0x0054] = "\xc6\xb8", [0x0055] = "\xc6\xb9", [0x0056] = "\xc6\xba",
  [0x0057] = "\xc6\xbb", [0x0058] = "\xc6\xbc", [0x0059] = "\xc6\xbd",
  [0x005a] = "\xc6\xbe", [0x005b] = "\xc6\xbf", [0x005c] = "\xc6\xc0",
  [0x005d] = "\xc6\xc1", [0x005e] = "\xc6\xc2", [0x005f] = "\xc6\xc3",
  [0x0060] = "\xc6\xc4", [0x0061] = "\xc6\xc5", [0x0062] = "\xc6\xc6",
  [0x0063] = "\xc6\xc7", [0x0064] = "\xc6\xc8", [0x0065] = "\xc6\xc9",
  [0x0066] = "\xc6\xca", [0x0067] = "\xc6\xcb", [0x0068] = "\xc6\xcc",
  [0x0069] = "\xc6\xcd", [0x006a] = "\xc6\xce", [0x006b] = "\xc6\xcf",
  [0x006c] = "\xc6\xd0", [0x006d] = "\xc6\xd1", [0x006e] = "\xc6\xd2",
  [0x006f] = "\xc6\xd3", [0x0070] = "\xc6\xd4", [0x0071] = "\xc6\xd5",
  [0x0072] = "\xc6\xd6", [0x0073] = "\xc6\xd7", [0x0074] = "\xc6\xd8",
  [0x0075] = "\xc6\xd9", [0x0076] = "\xc6\xda", [0x0077] = "\xc6\xdb",
  [0x0078] = "\xc6\xdc", [0x0079] = "\xc6\xdd", [0x007a] = "\xc6\xde",
  [0x007b] = "\xc6\xdf", [0x007c] = "\xc6\xe0", [0x007d] = "\xc6\xe1",
  [0x007e] = "\xc6\xe2", [0x007f] = "\xc6\xe3", [0x0080] = "\xc6\xe4",
  [0x0081] = "\xc6\xe5", [0x0082] = "\xc6\xe6", [0x0083] = "\xc6\xe7",
  [0x0084] = "\xc6\xe8", [0x0085] = "\xc6\xe9", [0x0086] = "\xc6\xea",
  [0x0087] = "\xc6\xeb", [0x0088] = "\xc6\xec", [0x0089] = "\xc6\xed",
  [0x008a] = "\xc6\xee", [0x008b] = "\xc6\xef", [0x008c] = "\xc6\xf0",
  [0x008d] = "\xc6\xf1", [0x008e] = "\xc6\xf2", [0x008f] = "\xc6\xf3",
  [0x0090] = "\xc6\xf4", [0x0091] = "\xc6\xf5", [0x0092] = "\xc6\xf6",
  [0x0093] = "\xc6\xf7", [0x009d] = "\xc6\xa2", [0x009e] = "\xc6\xa3",
  [0x00a1] = "\xc6\xf8", [0x00a2] = "\xc6\xf9", [0x00a3] = "\xc6\xfa",
  [0x00a4] = "\xc6\xfb", [0x00a5] = "\xc6\xfc", [0x00a6] = "\xc6\xfd",
  [0x00a7] = "\xc6\xfe", [0x00a8] = "\xc7\x40", [0x00a9] = "\xc7\x41",
  [0x00aa] = "\xc7\x42", [0x00ab] = "\xc7\x43", [0x00ac] = "\xc7\x44",
  [0x00ad] = "\xc7\x45", [0x00ae] = "\xc7\x46", [0x00af] = "\xc7\x47",
  [0x00b0] = "\xc7\x48", [0x00b1] = "\xc7\x49", [0x00b2] = "\xc7\x4a",
  [0x00b3] = "\xc7\x4b", [0x00b4] = "\xc7\x4c", [0x00b5] = "\xc7\x4d",
  [0x00b6] = "\xc7\x4e", [0x00b7] = "\xc7\x4f", [0x00b8] = "\xc7\x50",
  [0x00b9] = "\xc7\x51", [0x00ba] = "\xc7\x52", [0x00bb] = "\xc7\x53",
  [0x00bc] = "\xc7\x54", [0x00bd] = "\xc7\x55", [0x00be] = "\xc7\x56",
  [0x00bf] = "\xc7\x57", [0x00c0] = "\xc7\x58", [0x00c1] = "\xc7\x59",
  [0x00c2] = "\xc7\x5a", [0x00c3] = "\xc7\x5b", [0x00c4] = "\xc7\x5c",
  [0x00c5] = "\xc7\x5d", [0x00c6] = "\xc7\x5e", [0x00c7] = "\xc7\x5f",
  [0x00c8] = "\xc7\x60", [0x00c9] = "\xc7\x61", [0x00ca] = "\xc7\x62",
  [0x00cb] = "\xc7\x63", [0x00cc] = "\xc7\x64", [0x00cd] = "\xc7\x65",
  [0x00ce] = "\xc7\x66", [0x00cf] = "\xc7\x67", [0x00d0] = "\xc7\x68",
  [0x00d1] = "\xc7\x69", [0x00d2] = "\xc7\x6a", [0x00d3] = "\xc7\x6b",
  [0x00d4] = "\xc7\x6c", [0x00d5] = "\xc7\x6d", [0x00d6] = "\xc7\x6e",
  [0x00d7] = "\xc7\x6f", [0x00d8] = "\xc7\x70", [0x00d9] = "\xc7\x71",
  [0x00da] = "\xc7\x72", [0x00db] = "\xc7\x73", [0x00dc] = "\xc7\x74",
  [0x00dd] = "\xc7\x75", [0x00de] = "\xc7\x76", [0x00df] = "\xc7\x77",
  [0x00e0] = "\xc7\x78", [0x00e1] = "\xc7\x79", [0x00e2] = "\xc7\x7a",
  [0x00e3] = "\xc7\x7b", [0x00e4] = "\xc7\x7c", [0x00e5] = "\xc7\x7d",
  [0x00e6] = "\xc7\x7e", [0x00e7] = "\xc7\xa1", [0x00e8] = "\xc7\xa2",
  [0x00e9] = "\xc7\xa3", [0x00ea] = "\xc7\xa4", [0x00eb] = "\xc7\xa5",
  [0x00ec] = "\xc7\xa6", [0x00ed] = "\xc7\xa7", [0x00ee] = "\xc7\xa8",
  [0x00ef] = "\xc7\xa9", [0x00f0] = "\xc7\xaa", [0x00f1] = "\xc7\xab",
  [0x00f2] = "\xc7\xac", [0x00f3] = "\xc7\xad", [0x00f4] = "\xc7\xae",
  [0x00f5] = "\xc7\xaf", [0x00f6] = "\xc7\xb0", [0x00fe] = "\xc6\xa1",
  [0x0105] = "\xa3\x74", [0x0106] = "\xa3\x75", [0x0107] = "\xa3\x76",
  [0x0108] = "\xa3\x77", [0x0109] = "\xa3\x78", [0x010a] = "\xa3\x79",
  [0x010b] = "\xa3\x7a", [0x010c] = "\xa3\x7b", [0x010d] = "\xa3\x7c",
  [0x010e] = "\xa3\x7d", [0x010f] = "\xa3\x7e", [0x0110] = "\xa3\xa1",
  [0x0111] = "\xa3\xa2", [0x0112] = "\xa3\xa3", [0x0113] = "\xa3\xa4",
  [0x0114] = "\xa3\xa5", [0x0115] = "\xa3\xa6", [0x0116] = "\xa3\xa7",
  [0x0117] = "\xa3\xa8", [0x0118] = "\xa3\xa9", [0x0119] = "\xa3\xaa",
  [0x011a] = "\xa3\xab", [0x011b] = "\xa3\xac", [0x011c] = "\xa3\xad",
  [0x011d] = "\xa3\xae", [0x011e] = "\xa3\xaf", [0x011f] = "\xa3\xb0",
  [0x0120] = "\xa3\xb1", [0x0121] = "\xa3\xb2", [0x0122] = "\xa3\xb3",
  [0x0123] = "\xa3\xb4", [0x0124] = "\xa3\xb5", [0x0125] = "\xa3\xb6",
  [0x0126] = "\xa3\xb7", [0x0127] = "\xa3\xb8", [0x0128] = "\xa3\xb9",
  [0x0129] = "\xa3\xba",
};

/* Aquared latin abbreviations.
   The table can be created using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0x33' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0x338e,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab10[][2] =
{
  [0x0000] = "\xa2\x55", [0x0001] = "\xa2\x56", [0x000e] = "\xa2\x50",
  [0x000f] = "\xa2\x51", [0x0010] = "\xa2\x52", [0x0013] = "\xa2\x54",
  [0x0036] = "\xa2\x57", [0x0040] = "\xa2\x53", [0x0043] = "\xa1\xeb",
  [0x0044] = "\xa1\xea", [0x0047] = "\xa2\x4f",
};

/* CJK ideographs.

   (with HKSCS for ISO-10646 v1 mappings added - spacehunt 07/01/2000)

   The table can be created using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0x[4-9]' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0x4e00,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab11[][2] =
{
  [0x0000] = "\xa4\x40", [0x0001] = "\xa4\x42", [0x0003] = "\xa4\x43",
  [0x0004] = "\x9e\xb3", [0x0007] = "\xc9\x45", [0x0008] = "\xa4\x56",
  [0x0009] = "\xa4\x54", [0x000a] = "\xa4\x57", [0x000b] = "\xa4\x55",
  [0x000c] = "\xc9\x46", [0x000d] = "\xa4\xa3", [0x000e] = "\xc9\x4f",
  [0x000f] = "\xc9\x4d", [0x0010] = "\xa4\xa2", [0x0011] = "\xa4\xa1",
  [0x0014] = "\xa5\x42", [0x0015] = "\xa5\x41", [0x0016] = "\xa5\x40",
  [0x0018] = "\xa5\x43", [0x0019] = "\xa4\xfe", [0x001a] = "\x9e\xb2",
  [0x001c] = "\x9d\xd6", [0x001e] = "\xa5\xe0", [0x001f] = "\xa5\xe1",
  [0x0021] = "\x99\x4f", [0x0024] = "\x89\xce", [0x0026] = "\xa8\xc3",
  [0x0028] = "\x8b\xc0", [0x002a] = "\x9f\xc4", [0x002b] = "\xa4\x58",
  [0x002c] = "\x8b\xd4", [0x002d] = "\xa4\xa4", [0x002e] = "\xc9\x50",
  [0x0030] = "\xa4\xa5", [0x0031] = "\xc9\x63", [0x0032] = "\xa6\xea",
  [0x0033] = "\xcb\xb1", [0x0036] = "\xc6\xbf", [0x0037] = "\x8b\xf9",
  [0x0038] = "\xa4\x59", [0x0039] = "\xa4\xa6", [0x003b] = "\xa5\x44",
  [0x003c] = "\xc9\x64", [0x003d] = "\x89\x46", [0x003f] = "\xc6\xc0",
  [0x0042] = "\xc9\x40", [0x0043] = "\xa4\x44", [0x0045] = "\xa4\x5b",
  [0x0047] = "\xc9\x47", [0x0048] = "\xa4\x5c", [0x0049] = "\xfa\xe5",
  [0x004b] = "\xa4\xa7", [0x004d] = "\xa5\x45", [0x004e] = "\xa5\x47",
  [0x004f] = "\xa5\x46", [0x0052] = "\xa5\xe2", [0x0053] = "\xa5\xe3",
  [0x0056] = "\xa8\xc4", [0x0058] = "\xad\xbc", [0x0059] = "\xa4\x41",
  [0x005a] = "\xc8\x7b", [0x005b] = "\x8b\xc6", [0x005c] = "\xc9\x41",
  [0x005d] = "\xa4\x45", [0x005e] = "\xa4\x5e", [0x005f] = "\xa4\x5d",
  [0x0069] = "\xa5\xe4", [0x006a] = "\x9c\x57", [0x0073] = "\xa8\xc5",
  [0x0078] = "\x9a\xfb", [0x007e] = "\xb0\xae", [0x007f] = "\xd4\x4b",
  [0x0080] = "\x89\xd0", [0x0081] = "\x89\xcf", [0x0082] = "\xb6\xc3",
  [0x0083] = "\xdc\xb1", [0x0084] = "\xdc\xb2", [0x0085] = "\xc6\xc1",
  [0x0086] = "\xa4\x46", [0x0087] = "\x89\xd1", [0x0088] = "\xa4\xa9",
  [0x0089] = "\x89\xe2", [0x008b] = "\xa8\xc6", [0x008c] = "\xa4\x47",
  [0x008d] = "\xc9\x48", [0x008e] = "\xa4\x5f", [0x0091] = "\xa4\xaa",
  [0x0092] = "\xa4\xac", [0x0093] = "\xc9\x51", [0x0094] = "\xa4\xad",
  [0x0095] = "\xa4\xab", [0x0098] = "\x92\x7e", [0x0099] = "\xa5\xe5",
  [0x009a] = "\x9d\xba", [0x009b] = "\xa8\xc7", [0x009e] = "\xa8\xc8",
  [0x009f] = "\xab\x45", [0x00a0] = "\xc6\xc2", [0x00a1] = "\xa4\x60",
  [0x00a2] = "\xa4\xae", [0x00a4] = "\xa5\xe6", [0x00a5] = "\xa5\xe8",
  [0x00a6] = "\xa5\xe7", [0x00a8] = "\xa6\xeb", [0x00ab] = "\xa8\xc9",
  [0x00ac] = "\xa8\xca", [0x00ad] = "\xab\x46", [0x00ae] = "\xab\x47",
  [0x00b3] = "\xad\xbd", [0x00b6] = "\xdc\xb3", [0x00b7] = "\xfb\xf8",
  [0x00b9] = "\xf6\xd6", [0x00ba] = "\xa4\x48", [0x00bb] = "\x8b\xc7",
  [0x00bc] = "\x92\x6b", [0x00bf] = "\x89\xd2", [0x00c0] = "\xa4\xb0",
  [0x00c1] = "\xa4\xaf", [0x00c2] = "\xc9\x52", [0x00c3] = "\xa4\xb1",
  [0x00c4] = "\xa4\xb7", [0x00c6] = "\xa4\xb2", [0x00c7] = "\xa4\xb3",
  [0x00c8] = "\xc9\x54", [0x00c9] = "\xc9\x53", [0x00ca] = "\xa4\xb5",
  [0x00cb] = "\xa4\xb6", [0x00cd] = "\xa4\xb4", [0x00ce] = "\x9f\xcf",
  [0x00d4] = "\xa5\x4a", [0x00d5] = "\xa5\x4b", [0x00d6] = "\xa5\x4c",
  [0x00d7] = "\xa5\x4d", [0x00d8] = "\xa5\x49", [0x00d9] = "\xa5\x50",
  [0x00da] = "\xc9\x6a", [0x00dc] = "\xc9\x66", [0x00dd] = "\xc9\x69",
  [0x00de] = "\xa5\x51", [0x00df] = "\xa5\x61", [0x00e1] = "\xc9\x68",
  [0x00e3] = "\xa5\x4e", [0x00e4] = "\xa5\x4f", [0x00e5] = "\xa5\x48",
  [0x00e8] = "\xc9\x65", [0x00e9] = "\xc9\x67", [0x00ea] = "\x9d\xa9",
  [0x00eb] = "\x89\xd3", [0x00ee] = "\x99\xe2", [0x00f0] = "\xa5\xf5",
  [0x00f1] = "\xc9\xb0", [0x00f2] = "\xa5\xf2", [0x00f3] = "\xa5\xf6",
  [0x00f4] = "\xc9\xba", [0x00f5] = "\xc9\xae", [0x00f6] = "\xa5\xf3",
  [0x00f7] = "\xc9\xb2", [0x00f8] = "\x92\x67", [0x00fb] = "\xa5\xf4",
  [0x00fd] = "\xa5\xf7", [0x00ff] = "\xa5\xe9", [0x0100] = "\xc9\xb1",
  [0x0101] = "\xa5\xf8", [0x0102] = "\xc9\xb5", [0x0103] = "\x92\xa4",
  [0x0104] = "\xc9\xb9", [0x0105] = "\xc9\xb6", [0x0108] = "\xc9\xb3",
  [0x0109] = "\xa5\xea", [0x010a] = "\xa5\xec", [0x010b] = "\xa5\xf9",
  [0x010d] = "\xa5\xee", [0x010e] = "\xc9\xab", [0x010f] = "\xa5\xf1",
  [0x0110] = "\xa5\xef", [0x0111] = "\xa5\xf0", [0x0112] = "\xc9\xbb",
  [0x0113] = "\xc9\xb8", [0x0114] = "\xc9\xaf", [0x0115] = "\xa5\xed",
  [0x0118] = "\xc9\xac", [0x0119] = "\xa5\xeb", [0x011a] = "\x89\x4e",
  [0x011d] = "\xc9\xb4", [0x0122] = "\xc9\xb7", [0x0128] = "\x89\x4f",
  [0x0129] = "\x92\x78", [0x012c] = "\xc9\xad", [0x012d] = "\xca\x66",
  [0x012f] = "\xa7\x42", [0x0130] = "\xa6\xf4", [0x0132] = "\x91\xb6",
  [0x0133] = "\xca\x67", [0x0134] = "\xa6\xf1", [0x0136] = "\xa7\x44",
  [0x0137] = "\x89\xd4", [0x0138] = "\xa6\xf9", [0x0139] = "\x9f\xd2",
  [0x013a] = "\xa6\xf8", [0x013b] = "\xca\x5b", [0x013c] = "\xa6\xfc",
  [0x013d] = "\xa6\xf7", [0x013e] = "\xca\x60", [0x013f] = "\xca\x68",
  [0x0141] = "\xca\x64", [0x0142] = "\x92\xa7", [0x0143] = "\xa6\xfa",
  [0x0145] = "\x95\xa2", [0x0146] = "\xa6\xfd", [0x0147] = "\xa6\xee",
  [0x0148] = "\xa7\x47", [0x0149] = "\xca\x5d", [0x014b] = "\x92\x6e",
  [0x014c] = "\xcb\xbd", [0x014d] = "\xa6\xec", [0x014e] = "\xa7\x43",
  [0x014f] = "\xa6\xed", [0x0150] = "\xa6\xf5", [0x0151] = "\xa6\xf6",
  [0x0152] = "\xca\x62", [0x0153] = "\xca\x5e", [0x0154] = "\xa6\xfb",
  [0x0155] = "\xa6\xf3", [0x0156] = "\xca\x5a", [0x0157] = "\xa6\xef",
  [0x0158] = "\xca\x65", [0x0159] = "\xa7\x45", [0x015a] = "\xa7\x48",
  [0x015b] = "\xa6\xf2", [0x015c] = "\xa7\x40", [0x015d] = "\xa7\x46",
  [0x015e] = "\xa6\xf0", [0x015f] = "\xca\x63", [0x0160] = "\xa7\x41",
  [0x0161] = "\xca\x69", [0x0162] = "\xca\x5c", [0x0163] = "\xa6\xfe",
  [0x0164] = "\xca\x5f", [0x0167] = "\xca\x61", [0x0169] = "\xa8\xd8",
  [0x016a] = "\xcb\xbf", [0x016b] = "\xcb\xcb", [0x016c] = "\xa8\xd0",
  [0x016e] = "\xcb\xcc", [0x016f] = "\xa8\xcb", [0x0170] = "\xa8\xd5",
  [0x0172] = "\x96\xea", [0x0173] = "\xa8\xce", [0x0174] = "\xcb\xb9",
  [0x0175] = "\xa8\xd6", [0x0176] = "\xcb\xb8", [0x0177] = "\xcb\xbc",
  [0x0178] = "\xcb\xc3", [0x0179] = "\xcb\xc1", [0x017a] = "\xa8\xde",
  [0x017b] = "\xa8\xd9", [0x017c] = "\xcb\xb3", [0x017d] = "\xcb\xb5",
  [0x017e] = "\xa8\xdb", [0x017f] = "\xa8\xcf", [0x0180] = "\xcb\xb6",
  [0x0181] = "\xcb\xc2", [0x0182] = "\xcb\xc9", [0x0183] = "\xa8\xd4",
  [0x0184] = "\xcb\xbb", [0x0185] = "\xcb\xb4", [0x0186] = "\xa8\xd3",
  [0x0187] = "\xcb\xb7", [0x0188] = "\xa8\xd7", [0x0189] = "\xcb\xba",
  [0x018a] = "\x92\x6f", [0x018b] = "\xa8\xd2", [0x018d] = "\xa8\xcd",
  [0x018f] = "\xa8\xdc", [0x0190] = "\xcb\xc4", [0x0191] = "\xa8\xdd",
  [0x0192] = "\xcb\xc8", [0x0194] = "\xcb\xc6", [0x0195] = "\xcb\xca",
  [0x0196] = "\xa8\xda", [0x0197] = "\xcb\xbe", [0x0198] = "\xcb\xb2",
  [0x019a] = "\xcb\xc0", [0x019b] = "\xa8\xd1", [0x019c] = "\xcb\xc5",
  [0x019d] = "\xa8\xcc", [0x019e] = "\xcb\xc7", [0x01a2] = "\x92\xa3",
  [0x01a8] = "\x89\x50", [0x01ab] = "\xfa\x57", [0x01ae] = "\xab\x56",
  [0x01af] = "\xab\x4a", [0x01b0] = "\x98\x66", [0x01b2] = "\xcd\xe0",
  [0x01b3] = "\xcd\xe8", [0x01b5] = "\xab\x49", [0x01b6] = "\xab\x51",
  [0x01b7] = "\xab\x5d", [0x01b9] = "\xcd\xee", [0x01ba] = "\xcd\xec",
  [0x01bb] = "\xcd\xe7", [0x01bd] = "\x89\xd6", [0x01bf] = "\xab\x4b",
  [0x01c0] = "\xcd\xed", [0x01c1] = "\xcd\xe3", [0x01c2] = "\xab\x59",
  [0x01c3] = "\xab\x50", [0x01c4] = "\xab\x58", [0x01c5] = "\xcd\xde",
  [0x01c7] = "\xcd\xea", [0x01c8] = "\x98\xb2", [0x01c9] = "\xcd\xe1",
  [0x01ca] = "\xab\x54", [0x01cb] = "\xcd\xe2", [0x01cc] = "\x92\xab",
  [0x01cd] = "\xcd\xdd", [0x01ce] = "\xab\x5b", [0x01cf] = "\xab\x4e",
  [0x01d0] = "\xab\x57", [0x01d1] = "\xab\x4d", [0x01d3] = "\xcd\xdf",
  [0x01d4] = "\xcd\xe4", [0x01d6] = "\xcd\xeb", [0x01d7] = "\xab\x55",
  [0x01d8] = "\xab\x52", [0x01d9] = "\xcd\xe6", [0x01da] = "\xab\x5a",
  [0x01db] = "\xcd\xe9", [0x01dc] = "\xcd\xe5", [0x01dd] = "\xab\x4f",
  [0x01de] = "\xab\x5c", [0x01df] = "\xab\x53", [0x01e0] = "\xab\x4c",
  [0x01e1] = "\xab\x48", [0x01e4] = "\x96\xde", [0x01e5] = "\x92\xac",
  [0x01ec] = "\xcd\xef", [0x01ee] = "\xad\xd7", [0x01ef] = "\xad\xc1",
  [0x01f1] = "\xad\xd1", [0x01f2] = "\x9f\x6e", [0x01f3] = "\xad\xd6",
  [0x01f4] = "\xd0\xd0", [0x01f5] = "\xd0\xcf", [0x01f6] = "\xd0\xd4",
  [0x01f7] = "\xd0\xd5", [0x01f8] = "\xad\xc4", [0x01f9] = "\x8e\xf2",
  [0x01fa] = "\xad\xcd", [0x01fd] = "\x9f\x6c", [0x01fe] = "\xad\xda",
  [0x0200] = "\xad\xce", [0x0203] = "\x89\xd8", [0x0205] = "\xd0\xc9",
  [0x0206] = "\xad\xc7", [0x0207] = "\xd0\xca", [0x0208] = "\xfa\x59",
  [0x0209] = "\xad\xdc", [0x020b] = "\xad\xd3", [0x020c] = "\xad\xbe",
  [0x020d] = "\xad\xbf", [0x020e] = "\xd0\xdd", [0x020f] = "\xb0\xbf",
  [0x0211] = "\xad\xcc", [0x0212] = "\xad\xcb", [0x0213] = "\xd0\xcb",
  [0x0214] = "\xad\xcf", [0x0215] = "\xd4\x5b", [0x0216] = "\xad\xc6",
  [0x0217] = "\xd0\xd6", [0x0218] = "\xad\xd5", [0x0219] = "\xad\xd4",
  [0x021a] = "\xad\xca", [0x021b] = "\xd0\xce", [0x021c] = "\xd0\xd7",
  [0x021e] = "\xd0\xc8", [0x021f] = "\xad\xc9", [0x0220] = "\xd0\xd8",
  [0x0221] = "\xad\xd2", [0x0222] = "\xd0\xcc", [0x0223] = "\xad\xc0",
  [0x0225] = "\xad\xc3", [0x0226] = "\xad\xc2", [0x0227] = "\xd0\xd9",
  [0x0228] = "\xad\xd0", [0x0229] = "\xad\xc5", [0x022a] = "\xad\xd9",
  [0x022b] = "\xad\xdb", [0x022c] = "\xd0\xd3", [0x022d] = "\xad\xd8",
  [0x022e] = "\x92\xa8", [0x022f] = "\xd0\xdb", [0x0230] = "\xd0\xcd",
  [0x0231] = "\xd0\xdc", [0x0233] = "\xd0\xd1", [0x0234] = "\x91\x63",
  [0x0235] = "\xd0\xda", [0x0237] = "\xd0\xd2", [0x023c] = "\xad\xc8",
  [0x0240] = "\xd4\x63", [0x0241] = "\xd4\x57", [0x0243] = "\xb0\xb3",
  [0x0245] = "\xd4\x5c", [0x0246] = "\xd4\x62", [0x0247] = "\xb0\xb2",
  [0x0248] = "\xd4\x55", [0x0249] = "\xb0\xb6", [0x024a] = "\xd4\x59",
  [0x024b] = "\xd4\x52", [0x024c] = "\xb0\xb4", [0x024d] = "\xd4\x56",
  [0x024e] = "\xb0\xb9", [0x024f] = "\xb0\xbe", [0x0251] = "\xd4\x67",
  [0x0253] = "\xd4\x51", [0x0255] = "\xb0\xba", [0x0256] = "\x9f\x73",
  [0x0257] = "\xd4\x66", [0x0258] = "\x92\xad", [0x025a] = "\xb0\xb5",
  [0x025b] = "\xd4\x58", [0x025c] = "\xb0\xb1", [0x025d] = "\xd4\x53",
  [0x025e] = "\xd4\x4f", [0x025f] = "\xd4\x5d", [0x0260] = "\xd4\x50",
  [0x0261] = "\xd4\x4e", [0x0262] = "\xd4\x5a", [0x0263] = "\xd4\x60",
  [0x0264] = "\xd4\x61", [0x0265] = "\xb0\xb7", [0x0266] = "\x9b\xe9",
  [0x0268] = "\xd8\x5b", [0x0269] = "\xd4\x5e", [0x026a] = "\xd4\x4d",
  [0x026b] = "\xd4\x5f", [0x026c] = "\x92\xa9", [0x026d] = "\xb0\xc1",
  [0x026e] = "\xd4\x64", [0x026f] = "\xb0\xc0", [0x0270] = "\xd4\x4c",
  [0x0272] = "\xd4\x54", [0x0273] = "\xd4\x65", [0x0274] = "\xb0\xbc",
  [0x0275] = "\xb0\xbb", [0x0276] = "\xb0\xb8", [0x0277] = "\xb0\xbd",
  [0x027a] = "\xb0\xaf", [0x027d] = "\xb0\xb0", [0x0280] = "\xb3\xc8",
  [0x0281] = "\x92\xaa", [0x0282] = "\xd8\x5e", [0x0283] = "\xd8\x57",
  [0x0285] = "\xb3\xc5", [0x0287] = "\xd8\x5f", [0x0288] = "\x89\xd9",
  [0x028b] = "\xd8\x55", [0x028c] = "\xd8\x58", [0x028d] = "\xb3\xc4",
  [0x028e] = "\xd8\x59", [0x0290] = "\xfd\x56", [0x0291] = "\xb3\xc7",
  [0x0292] = "\xd8\x5d", [0x0294] = "\xd8\x53", [0x0295] = "\xd8\x52",
  [0x0296] = "\xb3\xc9", [0x0298] = "\xb3\xca", [0x0299] = "\xb3\xc6",
  [0x029a] = "\xb3\xcb", [0x029b] = "\xd8\x51", [0x029c] = "\xd8\x5c",
  [0x029d] = "\xd8\x5a", [0x029e] = "\xd8\x54", [0x02a2] = "\xb3\xc3",
  [0x02a3] = "\xd8\x56", [0x02a6] = "\x9f\xa8", [0x02ac] = "\xb6\xca",
  [0x02ad] = "\xb6\xc4", [0x02ae] = "\xdc\xb7", [0x02af] = "\xb6\xcd",
  [0x02b0] = "\xdc\xbd", [0x02b1] = "\xdc\xc0", [0x02b2] = "\xb6\xc6",
  [0x02b3] = "\xb6\xc7", [0x02b4] = "\xdc\xba", [0x02b5] = "\xb6\xc5",
  [0x02b6] = "\xdc\xc3", [0x02b7] = "\xb6\xcb", [0x02b8] = "\xdc\xc4",
  [0x02ba] = "\xdc\xbf", [0x02bb] = "\xb6\xcc", [0x02bd] = "\xdc\xb4",
  [0x02be] = "\xb6\xc9", [0x02bf] = "\xdc\xb5", [0x02c1] = "\xdc\xbe",
  [0x02c2] = "\xdc\xbc", [0x02c4] = "\xdc\xb8", [0x02c5] = "\xb6\xc8",
  [0x02c6] = "\xdc\xb6", [0x02c7] = "\xb6\xce", [0x02c8] = "\xdc\xbb",
  [0x02c9] = "\xdc\xc2", [0x02ca] = "\xdc\xb9", [0x02cb] = "\xdc\xc1",
  [0x02cd] = "\x92\xa1", [0x02ce] = "\xb9\xb6", [0x02cf] = "\xb9\xb3",
  [0x02d0] = "\x90\xe3", [0x02d1] = "\xb9\xb4", [0x02d3] = "\xe0\xf9",
  [0x02d4] = "\xe0\xf1", [0x02d5] = "\xb9\xb2", [0x02d6] = "\xb9\xaf",
  [0x02d7] = "\xe0\xf2", [0x02d9] = "\xa0\xa6", [0x02da] = "\xb9\xb1",
  [0x02db] = "\xe0\xf5", [0x02dd] = "\xe0\xf7", [0x02df] = "\x94\xab",
  [0x02e0] = "\xe0\xfe", [0x02e1] = "\xfc\x72", [0x02e3] = "\xe0\xfd",
  [0x02e4] = "\xe0\xf8", [0x02e5] = "\xb9\xae", [0x02e6] = "\xe0\xf0",
  [0x02e7] = "\xb9\xac", [0x02e8] = "\xe0\xf3", [0x02e9] = "\xb9\xb7",
  [0x02ea] = "\xe0\xf6", [0x02ec] = "\xe0\xfa", [0x02ed] = "\xb9\xb0",
  [0x02ee] = "\xb9\xad", [0x02ef] = "\xe0\xfc", [0x02f0] = "\xe0\xfb",
  [0x02f1] = "\xb9\xb5", [0x02f3] = "\xe0\xf4", [0x02f4] = "\x97\xc4",
  [0x02f5] = "\xbb\xf8", [0x02f6] = "\xe4\xec", [0x02f8] = "\xe4\xe9",
  [0x02f9] = "\xbb\xf9", [0x02fb] = "\xbb\xf7", [0x02fc] = "\x92\xae",
  [0x02fd] = "\xe4\xf0", [0x02fe] = "\xe4\xed", [0x02ff] = "\xe4\xe6",
  [0x0300] = "\xbb\xf6", [0x0301] = "\xfa\x67", [0x0302] = "\xbb\xfa",
  [0x0303] = "\xe4\xe7", [0x0304] = "\xbb\xf5", [0x0305] = "\xbb\xfd",
  [0x0306] = "\xe4\xea", [0x0307] = "\xe4\xeb", [0x0308] = "\xbb\xfb",
  [0x0309] = "\xbb\xfc", [0x030a] = "\xe4\xf1", [0x030b] = "\xe4\xee",
  [0x030c] = "\xe4\xef", [0x030d] = "\x92\xa2", [0x030e] = "\xfa\x69",
  [0x0310] = "\xbe\xaa", [0x0311] = "\xe8\xf8", [0x0312] = "\xbe\xa7",
  [0x0313] = "\xe8\xf5", [0x0314] = "\xbe\xa9", [0x0315] = "\xbe\xab",
  [0x0317] = "\xe8\xf6", [0x0318] = "\xbe\xa8", [0x031a] = "\xe8\xf7",
  [0x031c] = "\xe8\xf4", [0x031f] = "\xc0\x76", [0x0320] = "\xec\xbd",
  [0x0321] = "\xc0\x77", [0x0322] = "\xec\xbb", [0x0324] = "\xec\xbc",
  [0x0325] = "\xec\xba", [0x0326] = "\xec\xb9", [0x0329] = "\xec\xbe",
  [0x032a] = "\xc0\x75", [0x032b] = "\x92\x68", [0x032d] = "\xef\xb8",
  [0x032e] = "\xef\xb9", [0x0330] = "\xe4\xe8", [0x0331] = "\xef\xb7",
  [0x0332] = "\xc0\x78", [0x0333] = "\xc3\x5f", [0x0334] = "\xf1\xeb",
  [0x0335] = "\xf1\xec", [0x0337] = "\xc4\xd7", [0x0338] = "\xc4\xd8",
  [0x0339] = "\xf5\xc1", [0x033a] = "\xf5\xc0", [0x033b] = "\xc5\x6c",
  [0x033c] = "\xc5\x6b", [0x033d] = "\xf7\xd0", [0x033f] = "\xa4\x49",
  [0x0340] = "\xa4\x61", [0x0341] = "\xa4\xb9", [0x0343] = "\xa4\xb8",
  [0x0344] = "\xa5\x53", [0x0345] = "\xa5\x52", [0x0346] = "\xa5\xfc",
  [0x0347] = "\xa5\xfb", [0x0348] = "\xa5\xfd", [0x0349] = "\xa5\xfa",
  [0x034b] = "\xa7\x4a", [0x034c] = "\xa7\x49", [0x034d] = "\xa7\x4b",
  [0x0352] = "\xa8\xe0", [0x0354] = "\xa8\xdf", [0x0355] = "\xa8\xe1",
  [0x0356] = "\x89\x51", [0x0357] = "\xab\x5e", [0x0359] = "\xa2\x59",
  [0x035a] = "\xd0\xde", [0x035b] = "\xa2\x5a", [0x035c] = "\xb0\xc2",
  [0x035d] = "\xa2\x5c", [0x035e] = "\xa2\x5b", [0x035f] = "\xd8\x60",
  [0x0360] = "\xfa\x6f", [0x0361] = "\xa2\x5d", [0x0362] = "\xb9\xb8",
  [0x0363] = "\xa2\x5e", [0x0365] = "\xa4\x4a", [0x0367] = "\xa4\xba",
  [0x0368] = "\xa5\xfe", [0x0369] = "\xa8\xe2", [0x036a] = "\xfa\x71",
  [0x036b] = "\xa4\x4b", [0x036c] = "\xa4\xbd", [0x036d] = "\xa4\xbb",
  [0x036e] = "\xa4\xbc", [0x0371] = "\xa6\x40", [0x0374] = "\x89\x52",
  [0x0375] = "\xa7\x4c", [0x0376] = "\xa8\xe4", [0x0377] = "\xa8\xe3",
  [0x0378] = "\xa8\xe5", [0x0379] = "\x94\x5a", [0x037c] = "\xad\xdd",
  [0x0380] = "\xbe\xac", [0x0382] = "\xc6\xc3", [0x0386] = "\x89\xdd",
  [0x0387] = "\xc9\x4e", [0x0388] = "\xc8\xa2", [0x0389] = "\xa5\x54",
  [0x038a] = "\xa5\x55", [0x038d] = "\xa6\x41", [0x038f] = "\xca\x6a",
  [0x0391] = "\xab\x60", [0x0392] = "\xab\x5f", [0x0393] = "\xd0\xe0",
  [0x0394] = "\xd0\xdf", [0x0395] = "\xb0\xc3", [0x0396] = "\xc6\xc4",
  [0x0397] = "\xa4\xbe", [0x0398] = "\xc9\x55", [0x039a] = "\x9e\x52",
  [0x039c] = "\x89\x53", [0x039e] = "\xcb\xcd", [0x03a0] = "\xab\x61",
  [0x03a2] = "\xad\xe0", [0x03a4] = "\xad\xde", [0x03a5] = "\xad\xdf",
  [0x03a7] = "\x9e\x55", [0x03a8] = "\x92\xba", [0x03aa] = "\xbe\xad",
  [0x03ab] = "\xc6\xc5", [0x03ac] = "\xa5\x56", [0x03b0] = "\xa6\x42",
  [0x03b1] = "\xc9\xbc", [0x03b2] = "\xfa\x7d", [0x03b3] = "\xfa\xa8",
  [0x03b4] = "\x9a\x68", [0x03b5] = "\xfa\x47", [0x03b6] = "\xa7\x4d",
  [0x03b7] = "\xa7\x4e", [0x03b8] = "\xfa\x7e", [0x03b9] = "\xca\x6b",
  [0x03bc] = "\xcb\xce", [0x03bd] = "\xa8\xe6", [0x03be] = "\xcb\xcf",
  [0x03c3] = "\x92\xbb", [0x03c4] = "\xd0\xe2", [0x03c5] = "\xd0\xe3",
  [0x03c6] = "\xad\xe3", [0x03c7] = "\xfd\xb6", [0x03c8] = "\xd0\xe4",
  [0x03c9] = "\xfa\xa2", [0x03ca] = "\xd0\xe1", [0x03cb] = "\xad\xe4",
  [0x03cc] = "\xad\xe2", [0x03cd] = "\xad\xe1", [0x03ce] = "\xd0\xe5",
  [0x03cf] = "\xfa\xa3", [0x03d0] = "\xd4\x68", [0x03d1] = "\xfa\xa4",
  [0x03d2] = "\x9b\xb4", [0x03d3] = "\xfa\xa6", [0x03d4] = "\xd8\x61",
  [0x03d7] = "\xdc\xc5", [0x03d8] = "\xe1\x40", [0x03db] = "\x89\xdf",
  [0x03dc] = "\xbb\xfe", [0x03dd] = "\xbe\xae", [0x03de] = "\xe8\xf9",
  [0x03df] = "\xfd\xdb", [0x03e0] = "\xa4\x4c", [0x03e1] = "\xa4\x5a",
  [0x03e2] = "\xfa\xa9", [0x03e4] = "\x89\x54", [0x03ed] = "\xfa\xab",
  [0x03f0] = "\xb0\xc4", [0x03f1] = "\xb3\xcd", [0x03f3] = "\xb9\xb9",
  [0x03f4] = "\xfc\x7a", [0x03f5] = "\xc9\x42", [0x03f6] = "\xa4\xbf",
  [0x03f8] = "\xa5\x59", [0x03f9] = "\xa5\x57", [0x03fa] = "\xa5\x58",
  [0x03fc] = "\x89\xe0", [0x03fd] = "\xa8\xe7", [0x03fe] = "\x9f\x4f",
  [0x0400] = "\xa4\x4d", [0x0401] = "\xa4\x4e", [0x0402] = "\xc8\x7d",
  [0x0403] = "\xa4\x62", [0x0405] = "\x89\xe1", [0x0406] = "\xa4\xc0",
  [0x0407] = "\xa4\xc1", [0x0408] = "\xa4\xc2", [0x0409] = "\xc9\xbe",
  [0x040a] = "\xa5\x5a", [0x040b] = "\xfa\xb0", [0x040c] = "\xc9\x6b",
  [0x040e] = "\xa6\x46", [0x0410] = "\xc9\xbf", [0x0411] = "\xa6\x44",
  [0x0412] = "\xa6\x45", [0x0413] = "\xc9\xbd", [0x0416] = "\xa6\x47",
  [0x0417] = "\xa6\x43", [0x041c] = "\xca\x6c", [0x041d] = "\xaa\xec",
  [0x041e] = "\xca\x6d", [0x041f] = "\x9f\xcd", [0x0420] = "\xa0\xe7",
  [0x0421] = "\xca\x6e", [0x0424] = "\xa7\x50", [0x0425] = "\xa7\x4f",
  [0x0426] = "\xfa\xb1", [0x0427] = "\x89\xa6", [0x0428] = "\xa7\x53",
  [0x0429] = "\xa7\x51", [0x042a] = "\xa7\x52", [0x042e] = "\xa8\xed",
  [0x0430] = "\xa8\xec", [0x0431] = "\xcb\xd4", [0x0432] = "\xcb\xd1",
  [0x0433] = "\xcb\xd2", [0x0434] = "\x9e\xfa", [0x0435] = "\xcb\xd0",
  [0x0436] = "\xa8\xee", [0x0437] = "\xa8\xea", [0x0438] = "\xa8\xe9",
  [0x043a] = "\xa8\xeb", [0x043b] = "\xa8\xe8", [0x043c] = "\xfa\xb2",
  [0x0441] = "\xa8\xef", [0x0443] = "\xab\x63", [0x0444] = "\xcd\xf0",
  [0x0446] = "\xcb\xd3", [0x0447] = "\xab\x68", [0x0449] = "\xcd\xf1",
  [0x044a] = "\xab\x64", [0x044b] = "\xab\x67", [0x044c] = "\xab\x66",
  [0x044d] = "\xab\x65", [0x044e] = "\xab\x62", [0x0452] = "\xd0\xe8",
  [0x0454] = "\xad\xe7", [0x0455] = "\xd0\xeb", [0x0456] = "\xad\xe5",
  [0x0457] = "\xfa\xb4", [0x0459] = "\x92\xc4", [0x045a] = "\xd0\xe7",
  [0x045b] = "\xad\xe8", [0x045c] = "\xad\xe6", [0x045d] = "\xad\xe9",
  [0x045e] = "\xd0\xe9", [0x045f] = "\xd0\xea", [0x0460] = "\x9f\x6f",
  [0x0461] = "\xd0\xe6", [0x0462] = "\xd0\xec", [0x0468] = "\x8b\xb0",
  [0x0469] = "\xb3\xd1", [0x046a] = "\xb0\xc5", [0x046b] = "\xd4\x69",
  [0x046c] = "\xd4\x6b", [0x046d] = "\xd4\x6a", [0x046e] = "\xd4\x6c",
  [0x046f] = "\xb0\xc6", [0x0472] = "\xb3\xce", [0x0473] = "\x9f\xac",
  [0x0474] = "\xb3\xcf", [0x0475] = "\xb3\xd0", [0x0477] = "\xb6\xd0",
  [0x0478] = "\xdc\xc7", [0x0479] = "\x89\xe3", [0x047a] = "\xdc\xc6",
  [0x047b] = "\xdc\xc8", [0x047c] = "\xdc\xc9", [0x047d] = "\xb6\xd1",
  [0x047f] = "\xb6\xcf", [0x0480] = "\xe1\x41", [0x0481] = "\xe1\x42",
  [0x0482] = "\xb9\xbb", [0x0483] = "\xb9\xba", [0x0484] = "\xe3\x5a",
  [0x0487] = "\xbc\x40", [0x0488] = "\xbc\x41", [0x0489] = "\xbc\x42",
  [0x048a] = "\xbc\x44", [0x048b] = "\xe4\xf2", [0x048c] = "\xe4\xf3",
  [0x048d] = "\xbc\x43", [0x048f] = "\x9b\xd3", [0x0490] = "\x89\xe4",
  [0x0491] = "\xbe\xaf", [0x0493] = "\xbe\xb0", [0x0494] = "\xfa\xb5",
  [0x0496] = "\xf1\xed", [0x0497] = "\xf5\xc3", [0x0498] = "\xf5\xc2",
  [0x0499] = "\xf7\xd1", [0x049a] = "\x9f\xd5", [0x049b] = "\xa4\x4f",
  [0x049f] = "\xa5\x5c", [0x04a0] = "\xa5\x5b", [0x04a1] = "\x89\x55",
  [0x04a3] = "\xa6\x48", [0x04a4] = "\x92\xc5", [0x04a6] = "\xc9\xc0",
  [0x04a8] = "\x89\x56", [0x04a9] = "\xa7\x55", [0x04aa] = "\xa7\x56",
  [0x04ab] = "\xa7\x54", [0x04ac] = "\xa7\x57", [0x04ad] = "\xca\x6f",
  [0x04ae] = "\xca\x70", [0x04b5] = "\xfa\xb3", [0x04b9] = "\xfa\xb6",
  [0x04bb] = "\xa8\xf1", [0x04bc] = "\xcb\xd5", [0x04be] = "\xa8\xf0",
  [0x04c0] = "\xcd\xf2", [0x04c1] = "\xab\x6c", [0x04c2] = "\xcd\xf3",
  [0x04c3] = "\xab\x6b", [0x04c5] = "\xfa\xb7", [0x04c7] = "\xab\x69",
  [0x04c9] = "\xab\x6a", [0x04cc] = "\x9e\xdc", [0x04cd] = "\xd0\xed",
  [0x04d0] = "\xfb\xc4", [0x04d1] = "\x9f\x71", [0x04d2] = "\xb0\xc7",
  [0x04d3] = "\xd4\x6e", [0x04d5] = "\xb0\xca", [0x04d6] = "\xd4\x6d",
  [0x04d7] = "\xb1\xe5", [0x04d8] = "\xb0\xc9", [0x04d9] = "\xb0\xc8",
  [0x04db] = "\xb3\xd4", [0x04dd] = "\xb3\xd3", [0x04de] = "\xb3\xd2",
  [0x04df] = "\xb6\xd2", [0x04e0] = "\xfa\xba", [0x04e1] = "\x92\xc7",
  [0x04e2] = "\xb6\xd5", [0x04e3] = "\xb6\xd6", [0x04e4] = "\xb6\xd4",
  [0x04e6] = "\xb6\xd3", [0x04e9] = "\xe1\x43", [0x04eb] = "\xe1\x44",
  [0x04ef] = "\xe4\xf5", [0x04f0] = "\xbc\x45", [0x04f1] = "\xe4\xf4",
  [0x04f3] = "\xbe\xb1", [0x04f4] = "\xec\xbf", [0x04f5] = "\xc0\x79",
  [0x04f7] = "\xf1\xee", [0x04f8] = "\xc4\x55", [0x04f9] = "\xc6\xc6",
  [0x04fa] = "\xa4\x63", [0x04fb] = "\xa4\xc3", [0x04fc] = "\xc9\x56",
  [0x04fe] = "\xa4\xc4", [0x04ff] = "\xa4\xc5", [0x0501] = "\x9a\x4c",
  [0x0505] = "\xa5\x5d", [0x0506] = "\xa5\x5e", [0x0508] = "\xa6\x49",
  [0x0509] = "\xca\x71", [0x050a] = "\xcb\xd6", [0x050b] = "\xcb\xd7",
  [0x050d] = "\xab\x6d", [0x050e] = "\xd0\xee", [0x050f] = "\xb0\xcc",
  [0x0510] = "\xb0\xcb", [0x0511] = "\xd8\x63", [0x0512] = "\xd8\x62",
  [0x0515] = "\xa4\x50", [0x0516] = "\xa4\xc6", [0x0517] = "\xa5\x5f",
  [0x0519] = "\xb0\xcd", [0x051a] = "\xc9\x43", [0x051c] = "\xc9\x6c",
  [0x051d] = "\xa5\x60", [0x051f] = "\xc9\xc2", [0x0520] = "\xa6\x4b",
  [0x0521] = "\xa6\x4a", [0x0522] = "\xc9\xc1", [0x0523] = "\xa7\x58",
  [0x0527] = "\x89\xe5", [0x052a] = "\xad\xea", [0x052c] = "\x9f\x7d",
  [0x052d] = "\xd4\x6f", [0x052f] = "\xb6\xd7", [0x0530] = "\xe1\x45",
  [0x0531] = "\xb9\xbc", [0x0532] = "\xa0\xa9", [0x0533] = "\xfa\xc4",
  [0x0534] = "\xe8\xfa", [0x0537] = "\xf3\xfd", [0x0538] = "\xc6\xc7",
  [0x0539] = "\xa4\xc7", [0x053b] = "\x89\x57", [0x053c] = "\xcb\xd8",
  [0x053d] = "\xcd\xf4", [0x053e] = "\xb0\xd0", [0x053f] = "\xb0\xce",
  [0x0540] = "\xb0\xcf", [0x0541] = "\xa2\xcc", [0x0542] = "\xfa\xaa",
  [0x0543] = "\xa4\x64", [0x0544] = "\xa2\xcd", [0x0545] = "\xa4\xca",
  [0x0547] = "\xa4\xc9", [0x0548] = "\xa4\xc8", [0x0549] = "\xa5\x63",
  [0x054a] = "\xa5\x62", [0x054c] = "\xc9\x6d", [0x054d] = "\xc9\xc3",
  [0x054e] = "\x89\x58", [0x0551] = "\xa8\xf5", [0x0552] = "\xa8\xf2",
  [0x0553] = "\xa8\xf4", [0x0554] = "\xa8\xf3", [0x0557] = "\xab\x6e",
  [0x055a] = "\xb3\xd5", [0x055c] = "\xa4\x52", [0x055d] = "\x8b\xe3",
  [0x055e] = "\xa4\xcb", [0x055f] = "\x8b\x61", [0x0560] = "\xa5\x65",
  [0x0561] = "\xa5\x64", [0x0563] = "\xca\x72", [0x0564] = "\x9a\xf1",
  [0x0566] = "\xa8\xf6", [0x0567] = "\x9e\xb7", [0x0569] = "\xc6\xc8",
  [0x056c] = "\xc9\x57", [0x056d] = "\xfa\xd1", [0x056e] = "\xa5\x67",
  [0x056f] = "\xa5\x66", [0x0570] = "\xa6\x4c", [0x0571] = "\xa6\x4d",
  [0x0572] = "\xca\x73", [0x0573] = "\xa7\x59", [0x0574] = "\xfa\xd2",
  [0x0575] = "\xa7\x5a", [0x0577] = "\xa8\xf7", [0x0578] = "\xa8\xf8",
  [0x0579] = "\xa8\xf9", [0x057b] = "\xab\x6f", [0x057c] = "\xcd\xf5",
  [0x057d] = "\x9e\xba", [0x057e] = "\xfa\xd4", [0x057f] = "\xad\xeb",
  [0x0582] = "\xc9\x44", [0x0584] = "\xa4\xcc", [0x058a] = "\xc9\xc4",
  [0x058e] = "\xca\x74", [0x058f] = "\xca\x75", [0x0592] = "\xcb\xd9",
  [0x0593] = "\xfa\xd9", [0x0594] = "\xcb\xda", [0x0596] = "\xcd\xf7",
  [0x0597] = "\xcd\xf6", [0x0598] = "\xcd\xf9", [0x0599] = "\xcd\xf8",
  [0x059a] = "\xab\x70", [0x059c] = "\xd4\x70", [0x059d] = "\xad\xed",
  [0x059e] = "\xd0\xef", [0x059f] = "\xad\xec", [0x05a0] = "\xfa\xdb",
  [0x05a2] = "\x9c\xe0", [0x05a4] = "\xd8\x64", [0x05a5] = "\xb3\xd6",
  [0x05a6] = "\xfb\xf7", [0x05a7] = "\xd8\x65", [0x05a8] = "\xfb\xfa",
  [0x05a9] = "\x89\xe7", [0x05aa] = "\xa0\x7a", [0x05ab] = "\xfa\xdc",
  [0x05ac] = "\xe1\x46", [0x05ad] = "\xb9\xbd", [0x05ae] = "\xfa\xdd",
  [0x05b0] = "\x89\xe9", [0x05b2] = "\xbc\x46", [0x05b4] = "\xf1\xef",
  [0x05b6] = "\xc6\xc9", [0x05b9] = "\xc9\x58", [0x05bb] = "\xa5\x68",
  [0x05c1] = "\xfa\xe2", [0x05c2] = "\x89\xeb", [0x05c3] = "\xb0\xd1",
  [0x05c5] = "\xfa\xe3", [0x05c8] = "\xa4\x53", [0x05c9] = "\xa4\x65",
  [0x05ca] = "\xa4\xce", [0x05cb] = "\xa4\xcd", [0x05cc] = "\x90\xc8",
  [0x05cd] = "\xa4\xcf", [0x05d0] = "\x92\xda", [0x05d1] = "\x89\x59",
  [0x05d2] = "\x9c\xf5", [0x05d4] = "\xa8\xfb", [0x05d6] = "\xa8\xfa",
  [0x05d7] = "\xa8\xfc", [0x05d8] = "\x89\x5a", [0x05d9] = "\xfa\xe7",
  [0x05da] = "\x9f\xa2", [0x05db] = "\xab\x71", [0x05df] = "\xad\xee",
  [0x05e0] = "\xfa\xea", [0x05e1] = "\xe8\xfb", [0x05e2] = "\xc2\x4f",
  [0x05e3] = "\xa4\x66", [0x05e4] = "\xa5\x6a", [0x05e5] = "\xa5\x79",
  [0x05e6] = "\xa5\x74", [0x05e8] = "\xa5\x6f", [0x05e9] = "\xa5\x6e",
  [0x05ea] = "\xa5\x75", [0x05eb] = "\xa5\x73", [0x05ec] = "\xa5\x6c",
  [0x05ed] = "\xa5\x7a", [0x05ee] = "\xa5\x6d", [0x05ef] = "\xa5\x69",
  [0x05f0] = "\xa5\x78", [0x05f1] = "\xa5\x77", [0x05f2] = "\xa5\x76",
  [0x05f3] = "\xa5\x6b", [0x05f5] = "\xa5\x72", [0x05f6] = "\xfa\xed",
  [0x05f7] = "\x8f\xad", [0x05f8] = "\xa5\x71", [0x05fb] = "\xa5\x7b",
  [0x05fc] = "\xa5\x70", [0x05fe] = "\xfb\x59", [0x0601] = "\xa6\x53",
  [0x0603] = "\xa6\x59", [0x0604] = "\xa6\x55", [0x0606] = "\xa6\x5b",
  [0x0607] = "\xc9\xc5", [0x0608] = "\xa6\x58", [0x0609] = "\xa6\x4e",
  [0x060a] = "\xa6\x51", [0x060b] = "\xa6\x54", [0x060c] = "\xa6\x50",
  [0x060d] = "\xa6\x57", [0x060e] = "\xa6\x5a", [0x060f] = "\xa6\x4f",
  [0x0610] = "\xa6\x52", [0x0611] = "\xa6\x56", [0x0612] = "\xa6\x5c",
  [0x0613] = "\xfa\xef", [0x0614] = "\x96\xef", [0x0616] = "\x9d\xec",
  [0x0618] = "\xca\x7e", [0x0619] = "\xca\x7b", [0x061a] = "\x9d\xca",
  [0x061b] = "\xa7\x67", [0x061c] = "\xca\x7c", [0x061d] = "\xa7\x5b",
  [0x061e] = "\xa7\x5d", [0x061f] = "\xa7\x75", [0x0620] = "\xa7\x70",
  [0x0621] = "\xfd\x6d", [0x0623] = "\x89\xec", [0x0624] = "\xca\xa5",
  [0x0625] = "\xca\x7d", [0x0626] = "\xa7\x5f", [0x0627] = "\xa7\x61",
  [0x0628] = "\xca\xa4", [0x0629] = "\xa7\x68", [0x062a] = "\xca\x78",
  [0x062b] = "\xa7\x74", [0x062c] = "\xa7\x76", [0x062d] = "\xa7\x5c",
  [0x062e] = "\xa7\x6d", [0x062f] = "\xfb\x44", [0x0630] = "\xca\x76",
  [0x0631] = "\xa7\x73", [0x0632] = "\x9d\xe2", [0x0633] = "\xa7\x64",
  [0x0635] = "\xa7\x6e", [0x0636] = "\xa7\x6f", [0x0637] = "\xca\x77",
  [0x0638] = "\xa7\x6c", [0x0639] = "\xa7\x6a", [0x063b] = "\xa7\x6b",
  [0x063c] = "\xa7\x71", [0x063d] = "\xca\xa1", [0x063e] = "\xa7\x5e",
  [0x0640] = "\xa7\x72", [0x0641] = "\xca\xa3", [0x0642] = "\xa7\x66",
  [0x0643] = "\xa7\x63", [0x0645] = "\xca\x7a", [0x0646] = "\xa7\x62",
  [0x0647] = "\xca\xa6", [0x0648] = "\xa7\x65", [0x064a] = "\xa7\x69",
  [0x064b] = "\x9e\xc0", [0x064d] = "\x9e\x56", [0x064e] = "\xa7\x60",
  [0x064f] = "\xca\xa2", [0x0654] = "\xca\x79", [0x0660] = "\xcb\xeb",
  [0x0661] = "\xcb\xea", [0x0662] = "\xa9\x4f", [0x0663] = "\xcb\xed",
  [0x0664] = "\xcb\xef", [0x0665] = "\xcb\xe4", [0x0666] = "\xcb\xe7",
  [0x0667] = "\xcb\xee", [0x0668] = "\xa9\x50", [0x0669] = "\x9f\x79",
  [0x066a] = "\x9a\xc7", [0x066b] = "\xcb\xe1", [0x066c] = "\xcb\xe5",
  [0x066d] = "\xfa\xf4", [0x066f] = "\xcb\xe9", [0x0670] = "\xce\x49",
  [0x0671] = "\xa9\x4b", [0x0672] = "\xce\x4d", [0x0673] = "\xa8\xfd",
  [0x0674] = "\xcb\xe6", [0x0675] = "\xa8\xfe", [0x0676] = "\xa9\x4c",
  [0x0677] = "\xa9\x45", [0x0678] = "\xa9\x41", [0x067a] = "\xcb\xe2",
  [0x067b] = "\xa9\x44", [0x067c] = "\xa9\x49", [0x067d] = "\xa9\x52",
  [0x067e] = "\xcb\xe3", [0x067f] = "\xcb\xdc", [0x0680] = "\xa9\x43",
  [0x0681] = "\xcb\xdd", [0x0682] = "\xcb\xdf", [0x0684] = "\xa9\x46",
  [0x0685] = "\x98\xa1", [0x0686] = "\xa9\x48", [0x0687] = "\xcb\xdb",
  [0x0688] = "\xcb\xe0", [0x068b] = "\xa9\x51", [0x068c] = "\xa9\x4d",
  [0x068d] = "\xcb\xe8", [0x068e] = "\xa9\x53", [0x068f] = "\xfa\xf8",
  [0x0690] = "\xa9\x4a", [0x0691] = "\xcb\xde", [0x0692] = "\xa9\x47",
  [0x0693] = "\x89\xf0", [0x0694] = "\x9e\x47", [0x0695] = "\xa9\x42",
  [0x0696] = "\xa9\x40", [0x0697] = "\x9d\xf7", [0x0698] = "\xcb\xec",
  [0x069a] = "\xa9\x4e", [0x069c] = "\x9f\xd3", [0x069e] = "\x9a\xca",
  [0x06a0] = "\xce\x48", [0x06a1] = "\xcd\xfb", [0x06a2] = "\xce\x4b",
  [0x06a3] = "\x89\xf1", [0x06a4] = "\xfa\xf9", [0x06a5] = "\xcd\xfd",
  [0x06a6] = "\xab\x78", [0x06a7] = "\xab\xa8", [0x06a8] = "\xab\x74",
  [0x06a9] = "\xab\xa7", [0x06aa] = "\xab\x7d", [0x06ab] = "\xab\xa4",
  [0x06ac] = "\xab\x72", [0x06ad] = "\xcd\xfc", [0x06ae] = "\xce\x43",
  [0x06af] = "\xab\xa3", [0x06b0] = "\xce\x4f", [0x06b1] = "\xab\xa5",
  [0x06b2] = "\x8e\x5a", [0x06b3] = "\xab\x79", [0x06b4] = "\x89\xf2",
  [0x06b6] = "\xce\x45", [0x06b7] = "\xce\x42", [0x06b8] = "\xab\x77",
  [0x06b9] = "\x89\xf3", [0x06ba] = "\xcd\xfa", [0x06bb] = "\xab\xa6",
  [0x06bc] = "\xce\x4a", [0x06bd] = "\xab\x7c", [0x06be] = "\xce\x4c",
  [0x06bf] = "\xab\xa9", [0x06c0] = "\xab\x73", [0x06c1] = "\xab\x7e",
  [0x06c2] = "\xab\x7b", [0x06c3] = "\xce\x40", [0x06c4] = "\xab\xa1",
  [0x06c5] = "\xce\x46", [0x06c6] = "\xce\x47", [0x06c7] = "\xab\x7a",
  [0x06c8] = "\xab\xa2", [0x06c9] = "\xab\x76", [0x06cb] = "\x92\x5d",
  [0x06cc] = "\x8b\x51", [0x06cd] = "\x92\xe0", [0x06ce] = "\xab\x75",
  [0x06cf] = "\xcd\xfe", [0x06d0] = "\x89\xf4", [0x06d6] = "\xce\x44",
  [0x06da] = "\x9f\xd4", [0x06de] = "\xce\x4e", [0x06e0] = "\xd1\x44",
  [0x06e1] = "\xad\xfb", [0x06e2] = "\xd0\xf1", [0x06e3] = "\x8a\x79",
  [0x06e4] = "\xd0\xf6", [0x06e5] = "\xad\xf4", [0x06e6] = "\xae\x40",
  [0x06e7] = "\xd0\xf4", [0x06e8] = "\xad\xef", [0x06e9] = "\xad\xf9",
  [0x06ea] = "\xad\xfe", [0x06eb] = "\xd0\xfb", [0x06ed] = "\xad\xfa",
  [0x06ee] = "\xad\xfd", [0x06ef] = "\x89\xf5", [0x06f1] = "\xd0\xfe",
  [0x06f2] = "\xad\xf5", [0x06f3] = "\xd0\xf5", [0x06f7] = "\xd1\x42",
  [0x06f8] = "\xd1\x43", [0x06fa] = "\xad\xf7", [0x06fb] = "\xd1\x41",
  [0x06fc] = "\xad\xf3", [0x06fd] = "\xae\x43", [0x06ff] = "\xd0\xf8",
  [0x0701] = "\xad\xf1", [0x0702] = "\x97\xa7", [0x0703] = "\xd1\x46",
  [0x0704] = "\xd0\xf9", [0x0705] = "\xd0\xfd", [0x0706] = "\xad\xf6",
  [0x0707] = "\xae\x42", [0x0708] = "\xd0\xfa", [0x0709] = "\xad\xfc",
  [0x070a] = "\xd1\x40", [0x070b] = "\xd1\x47", [0x070c] = "\xd4\xa1",
  [0x070d] = "\x93\xba", [0x070e] = "\xd1\x45", [0x070f] = "\xae\x44",
  [0x0710] = "\xad\xf0", [0x0711] = "\xd0\xfc", [0x0712] = "\xd0\xf3",
  [0x0713] = "\x9e\x58", [0x0714] = "\xad\xf8", [0x0717] = "\xd0\xf2",
  [0x0718] = "\x89\xf6", [0x071a] = "\xd0\xf7", [0x071e] = "\x9e\x57",
  [0x0723] = "\x89\xf7", [0x0725] = "\x8a\x41", [0x0726] = "\xd0\xf0",
  [0x0727] = "\xae\x41", [0x0728] = "\x89\xf8", [0x072a] = "\xd4\x77",
  [0x072b] = "\xfa\xf1", [0x072c] = "\xb0\xe4", [0x072d] = "\xd4\xa7",
  [0x072e] = "\xb0\xe2", [0x072f] = "\xb0\xdf", [0x0730] = "\xd4\x7c",
  [0x0731] = "\xb0\xdb", [0x0732] = "\xd4\xa2", [0x0733] = "\xb0\xe6",
  [0x0734] = "\xd4\x76", [0x0735] = "\xd4\x7b", [0x0736] = "\xd4\x7a",
  [0x0737] = "\xad\xf2", [0x0738] = "\xb0\xe1", [0x0739] = "\xd4\xa5",
  [0x073b] = "\xd4\xa8", [0x073c] = "\xd4\x73", [0x073e] = "\xb3\xe8",
  [0x073f] = "\x89\xfa", [0x0740] = "\xd4\xa9", [0x0741] = "\xb0\xe7",
  [0x0743] = "\xb0\xd9", [0x0744] = "\xb0\xd6", [0x0745] = "\xd4\x7e",
  [0x0746] = "\xb0\xd3", [0x0747] = "\xfb\x42", [0x0748] = "\xd4\xa6",
  [0x0749] = "\xfa\xbf", [0x074a] = "\xb0\xda", [0x074b] = "\xd4\xaa",
  [0x074d] = "\xd4\x74", [0x074e] = "\xd4\xa4", [0x074f] = "\xb0\xdd",
  [0x0750] = "\xd4\x75", [0x0751] = "\xd4\x78", [0x0752] = "\xd4\x7d",
  [0x0753] = "\xfb\xa3", [0x0755] = "\xb0\xde", [0x0756] = "\xb0\xdc",
  [0x0757] = "\xb0\xe8", [0x075c] = "\xb0\xe3", [0x075d] = "\xfa\xf7",
  [0x075e] = "\xb0\xd7", [0x075f] = "\xb1\xd2", [0x0761] = "\xb0\xd8",
  [0x0762] = "\xd4\x79", [0x0763] = "\xb0\xe5", [0x0764] = "\xb0\xe0",
  [0x0765] = "\xd4\xa3", [0x0766] = "\xb0\xd5", [0x0769] = "\x9e\x4e",
  [0x076a] = "\xb0\xd4", [0x076b] = "\x94\xdc", [0x0771] = "\x95\xda",
  [0x0772] = "\x9d\xf8", [0x0773] = "\x9f\x6a", [0x0775] = "\xd4\x71",
  [0x0776] = "\xd4\x72", [0x0777] = "\xd8\x6a", [0x0779] = "\x8a\xb7",
  [0x077b] = "\xb3\xd7", [0x077c] = "\xb3\xda", [0x077d] = "\xd8\x75",
  [0x077e] = "\xb3\xee", [0x077f] = "\xd8\x78", [0x0780] = "\xb3\xd8",
  [0x0781] = "\xd8\x71", [0x0782] = "\xb3\xde", [0x0783] = "\xb3\xe4",
  [0x0784] = "\xb5\xbd", [0x0786] = "\xfb\x46", [0x0787] = "\xb3\xe2",
  [0x0788] = "\xd8\x6e", [0x0789] = "\xb3\xef", [0x078a] = "\xb3\xdb",
  [0x078b] = "\xb3\xe3", [0x078c] = "\xd8\x76", [0x078d] = "\xdc\xd7",
  [0x078e] = "\xd8\x7b", [0x078f] = "\xd8\x6f", [0x0790] = "\x8a\x46",
  [0x0791] = "\xd8\x66", [0x0792] = "\xd8\x73", [0x0793] = "\xd8\x6d",
  [0x0794] = "\xb3\xe1", [0x0795] = "\xd8\x79", [0x0798] = "\xb3\xdd",
  [0x0799] = "\xb3\xf1", [0x079a] = "\xb3\xea", [0x079c] = "\xb3\xdf",
  [0x079d] = "\xb3\xdc", [0x079f] = "\xb3\xe7", [0x07a1] = "\xd8\x7a",
  [0x07a2] = "\xd8\x6c", [0x07a3] = "\xd8\x72", [0x07a4] = "\xd8\x74",
  [0x07a5] = "\xd8\x68", [0x07a6] = "\xd8\x77", [0x07a7] = "\xb3\xd9",
  [0x07a8] = "\xd8\x67", [0x07a9] = "\xfb\x47", [0x07aa] = "\xb3\xe0",
  [0x07ab] = "\xb3\xf0", [0x07ac] = "\xb3\xec", [0x07ad] = "\xd8\x69",
  [0x07ae] = "\xb3\xe6", [0x07b0] = "\x91\x48", [0x07b1] = "\xb3\xed",
  [0x07b2] = "\xb3\xe9", [0x07b3] = "\xb3\xe5", [0x07b4] = "\x92\xde",
  [0x07b5] = "\xd8\x70", [0x07b9] = "\x8b\x53", [0x07ba] = "\x9d\xf6",
  [0x07bb] = "\xb3\xeb", [0x07bc] = "\x9b\xda", [0x07bf] = "\xdc\xd5",
  [0x07c0] = "\xdc\xd1", [0x07c1] = "\x9d\x7e", [0x07c2] = "\xdc\xe0",
  [0x07c3] = "\xdc\xca", [0x07c4] = "\xdc\xd3", [0x07c5] = "\xb6\xe5",
  [0x07c6] = "\xb6\xe6", [0x07c7] = "\xb6\xde", [0x07c8] = "\xdc\xdc",
  [0x07c9] = "\xb6\xe8", [0x07ca] = "\xdc\xcf", [0x07cb] = "\xdc\xce",
  [0x07cc] = "\xdc\xcc", [0x07cd] = "\xdc\xde", [0x07ce] = "\xb6\xdc",
  [0x07cf] = "\xdc\xd8", [0x07d0] = "\xdc\xcd", [0x07d1] = "\xb6\xdf",
  [0x07d2] = "\xdc\xd6", [0x07d3] = "\xb6\xda", [0x07d4] = "\xdc\xd2",
  [0x07d5] = "\xdc\xd9", [0x07d6] = "\xdc\xdb", [0x07d7] = "\x89\xfd",
  [0x07d8] = "\x99\xe4", [0x07d9] = "\xdc\xdf", [0x07da] = "\xb6\xe3",
  [0x07db] = "\xdc\xcb", [0x07dc] = "\xb6\xdd", [0x07dd] = "\xdc\xd0",
  [0x07de] = "\x9e\x43", [0x07df] = "\xb6\xd8", [0x07e1] = "\xb6\xe4",
  [0x07e2] = "\xdc\xda", [0x07e3] = "\xb6\xe0", [0x07e4] = "\xb6\xe1",
  [0x07e5] = "\xb6\xe7", [0x07e6] = "\xb6\xdb", [0x07e7] = "\xa2\x5f",
  [0x07e8] = "\xb6\xd9", [0x07e9] = "\xdc\xd4", [0x07ea] = "\x9d\xe9",
  [0x07ec] = "\x8f\x52", [0x07ef] = "\xb6\xe2", [0x07f0] = "\x9d\xf5",
  [0x07f1] = "\x9d\xf0", [0x07f2] = "\xdc\xdd", [0x07f5] = "\x99\xe7",
  [0x07f6] = "\xb9\xcd", [0x07f7] = "\xb9\xc8", [0x07f9] = "\xe1\x55",
  [0x07fa] = "\xe1\x51", [0x07fb] = "\x8b\xbd", [0x07fc] = "\xe1\x4b",
  [0x07fd] = "\xb9\xc2", [0x07fe] = "\xb9\xbe", [0x07ff] = "\xe1\x54",
  [0x0800] = "\xb9\xbf", [0x0801] = "\xe1\x4e", [0x0802] = "\xe1\x50",
  [0x0804] = "\xe1\x53", [0x0805] = "\x9d\xef", [0x0806] = "\xb9\xc4",
  [0x0808] = "\xb9\xcb", [0x0809] = "\xb9\xc5", [0x080c] = "\xe1\x49",
  [0x080d] = "\xb9\xc6", [0x080e] = "\xb9\xc7", [0x080f] = "\xe1\x4c",
  [0x0810] = "\xb9\xcc", [0x0811] = "\x9f\xb7", [0x0812] = "\xe1\x4a",
  [0x0813] = "\xe1\x4f", [0x0814] = "\xb9\xc3", [0x0815] = "\xe1\x48",
  [0x0816] = "\xb9\xc9", [0x0817] = "\xb9\xc1", [0x081b] = "\xb9\xc0",
  [0x081c] = "\xe1\x4d", [0x081d] = "\xe1\x52", [0x081e] = "\x9d\xd0",
  [0x081f] = "\xb9\xca", [0x0820] = "\x9f\xeb", [0x0821] = "\x8d\xa9",
  [0x0822] = "\x9d\xcf", [0x0823] = "\x98\xe1", [0x0825] = "\x9d\xe5",
  [0x0827] = "\xe1\x47", [0x0829] = "\xbc\x4d", [0x082a] = "\xe5\x47",
  [0x082c] = "\xe5\x44", [0x082d] = "\x9d\xc8", [0x082e] = "\xbc\x47",
  [0x082f] = "\xbc\x53", [0x0830] = "\xbc\x54", [0x0832] = "\xbc\x4a",
  [0x0833] = "\xe5\x42", [0x0834] = "\xbc\x4c", [0x0835] = "\xe4\xf9",
  [0x0836] = "\xbc\x52", [0x0837] = "\xfb\x4f", [0x0838] = "\xe5\x46",
  [0x0839] = "\xbc\x49", [0x083a] = "\xe5\x48", [0x083b] = "\xbc\x48",
  [0x083d] = "\xe5\x43", [0x083e] = "\xe5\x45", [0x083f] = "\xbc\x4b",
  [0x0840] = "\xe5\x41", [0x0841] = "\xe4\xfa", [0x0842] = "\xe4\xf7",
  [0x0843] = "\x9d\xeb", [0x0845] = "\xd8\x6b", [0x0846] = "\xe4\xfd",
  [0x0848] = "\xe4\xf6", [0x0849] = "\xe4\xfc", [0x084a] = "\xe4\xfb",
  [0x084c] = "\xe4\xf8", [0x084d] = "\xfb\x54", [0x084e] = "\xbc\x4f",
  [0x084f] = "\xfb\x55", [0x0850] = "\x9a\xa2", [0x0852] = "\x8a\xd6",
  [0x0853] = "\xbc\x4e", [0x0854] = "\x9a\x5f", [0x0857] = "\xbc\x50",
  [0x0858] = "\xe4\xfe", [0x0859] = "\xbe\xb2", [0x085a] = "\xe5\x40",
  [0x085d] = "\x9e\xf5", [0x085e] = "\xe9\x45", [0x0860] = "\xe8\xfd",
  [0x0861] = "\x8f\xb7", [0x0862] = "\xbe\xbe", [0x0863] = "\xe9\x42",
  [0x0864] = "\xbe\xb6", [0x0865] = "\xbe\xba", [0x0866] = "\xe9\x41",
  [0x0868] = "\xbe\xb9", [0x0869] = "\xbe\xb5", [0x086a] = "\xbe\xb8",
  [0x086b] = "\xbe\xb3", [0x086c] = "\xbe\xbd", [0x086d] = "\xe9\x43",
  [0x086e] = "\xe8\xfe", [0x086f] = "\xbe\xbc", [0x0870] = "\xe8\xfc",
  [0x0871] = "\xbe\xbb", [0x0872] = "\xe9\x44", [0x0873] = "\xe9\x40",
  [0x0874] = "\xbc\x51", [0x0876] = "\xbe\xbf", [0x0877] = "\xe9\x46",
  [0x0878] = "\xbe\xb7", [0x0879] = "\xbe\xb4", [0x087a] = "\x9a\xd2",
  [0x087b] = "\x9e\x6a", [0x087c] = "\x9e\xe8", [0x087e] = "\xec\xc6",
  [0x087f] = "\xec\xc8", [0x0880] = "\xc0\x7b", [0x0881] = "\xec\xc9",
  [0x0882] = "\xec\xc7", [0x0883] = "\xec\xc5", [0x0884] = "\xec\xc4",
  [0x0885] = "\xc0\x7d", [0x0886] = "\xec\xc3", [0x0887] = "\xc0\x7e",
  [0x0889] = "\x8b\xbf", [0x088a] = "\x91\xc2", [0x088b] = "\x9d\x62",
  [0x088c] = "\xec\xc1", [0x088d] = "\xec\xc2", [0x088e] = "\xc0\x7a",
  [0x088f] = "\xc0\xa1", [0x0890] = "\xc0\x7c", [0x0892] = "\x92\x60",
  [0x0893] = "\xec\xc0", [0x0895] = "\xc2\x50", [0x0897] = "\xef\xbc",
  [0x0898] = "\xef\xba", [0x0899] = "\xef\xbf", [0x089a] = "\xef\xbd",
  [0x089c] = "\xef\xbb", [0x089d] = "\xef\xbe", [0x089e] = "\x92\x5e",
  [0x089f] = "\x91\xc1", [0x08a1] = "\x8a\xc5", [0x08a4] = "\x97\xa3",
  [0x08a5] = "\xc3\x60", [0x08a6] = "\xf1\xf2", [0x08a7] = "\xf1\xf3",
  [0x08a8] = "\xc4\x56", [0x08aa] = "\xf1\xf4", [0x08ab] = "\xf1\xf0",
  [0x08ac] = "\xf1\xf5", [0x08ad] = "\xf1\xf1", [0x08ae] = "\xc2\x51",
  [0x08af] = "\x8b\x6c", [0x08b1] = "\x8d\x7e", [0x08b2] = "\xf3\xfe",
  [0x08b3] = "\xf4\x41", [0x08b4] = "\xc4\x59", [0x08b5] = "\xf4\x40",
  [0x08b6] = "\xc4\x58", [0x08b7] = "\xc4\x57", [0x08b9] = "\x9c\x54",
  [0x08bc] = "\xc4\x5a", [0x08bd] = "\xf5\xc5", [0x08be] = "\xf5\xc6",
  [0x08bf] = "\x9d\xbd", [0x08c0] = "\xc4\xda", [0x08c1] = "\xc4\xd9",
  [0x08c2] = "\xc4\xdb", [0x08c3] = "\xf5\xc4", [0x08c5] = "\xf6\xd8",
  [0x08c6] = "\xf6\xd7", [0x08c8] = "\xc5\x6d", [0x08c9] = "\xc5\x6f",
  [0x08ca] = "\xc5\x6e", [0x08cb] = "\xf6\xd9", [0x08cc] = "\xc5\xc8",
  [0x08cd] = "\xf8\xa6", [0x08d1] = "\xc5\xf1", [0x08d3] = "\xf8\xa5",
  [0x08d4] = "\xf8\xee", [0x08d6] = "\x9c\xc5", [0x08d7] = "\xc9\x49",
  [0x08da] = "\xa5\x7d", [0x08db] = "\xa5\x7c", [0x08dd] = "\xa6\x5f",
  [0x08de] = "\xa6\x5e", [0x08df] = "\xc9\xc7", [0x08e0] = "\xa6\x5d",
  [0x08e1] = "\xc9\xc6", [0x08e2] = "\x89\x5b", [0x08e4] = "\xa7\x79",
  [0x08e5] = "\xca\xa9", [0x08e7] = "\xca\xa8", [0x08ea] = "\xa7\x77",
  [0x08eb] = "\xa7\x7a", [0x08ed] = "\xfb\x5c", [0x08ee] = "\xca\xa7",
  [0x08ef] = "\xfb\x5b", [0x08f0] = "\xa7\x78", [0x08f1] = "\xfb\x57",
  [0x08f7] = "\xcb\xf0", [0x08f9] = "\xcb\xf1", [0x08fa] = "\xa9\x54",
  [0x08fd] = "\x98\xc7", [0x08ff] = "\xab\xaa", [0x0900] = "\xfb\x5a",
  [0x0901] = "\xd1\x48", [0x0902] = "\xd1\x49", [0x0903] = "\xae\x45",
  [0x0904] = "\xae\x46", [0x0907] = "\xd4\xac", [0x0908] = "\xb0\xe9",
  [0x0909] = "\xb0\xeb", [0x090a] = "\xd4\xab", [0x090b] = "\xb0\xea",
  [0x090c] = "\xd8\x7c", [0x090d] = "\xb3\xf2", [0x0912] = "\xb6\xe9",
  [0x0913] = "\xb6\xea", [0x0914] = "\xdc\xe1", [0x0915] = "\x9c\xee",
  [0x0916] = "\xb9\xcf", [0x0918] = "\xb9\xce", [0x091a] = "\xe5\x49",
  [0x091b] = "\xe9\x48", [0x091c] = "\xe9\x47", [0x091d] = "\x92\xe2",
  [0x091e] = "\xf9\x6b", [0x091f] = "\xa4\x67", [0x0920] = "\xc9\x59",
  [0x0922] = "\xc9\x6e", [0x0923] = "\xc9\x6f", [0x0928] = "\xa6\x62",
  [0x0929] = "\xa6\x66", [0x092a] = "\xc9\xc9", [0x092c] = "\xa6\x64",
  [0x092d] = "\xa6\x63", [0x092e] = "\xc9\xc8", [0x092f] = "\xa6\x65",
  [0x0930] = "\xa6\x61", [0x0932] = "\x94\xa7", [0x0933] = "\xa6\x60",
  [0x0934] = "\xc9\xca", [0x093b] = "\xa7\xa6", [0x093e] = "\xa7\xa3",
  [0x093f] = "\x9b\xd4", [0x0940] = "\xa7\x7d", [0x0941] = "\xca\xaa",
  [0x0942] = "\xfb\x64", [0x0943] = "\xfb\x76", [0x0945] = "\xca\xab",
  [0x0946] = "\xfb\x60", [0x0947] = "\xa7\xa1", [0x0949] = "\xca\xad",
  [0x094a] = "\xa7\x7b", [0x094b] = "\xca\xae", [0x094c] = "\xca\xac",
  [0x094d] = "\xa7\x7e", [0x094e] = "\xa7\xa2", [0x094f] = "\xa7\xa5",
  [0x0950] = "\xa7\xa4", [0x0951] = "\xa7\x7c", [0x0952] = "\xca\xaf",
  [0x0954] = "\x99\xe5", [0x0957] = "\x9a\xc2", [0x095b] = "\x91\xfb",
  [0x095f] = "\xa0\x73", [0x0961] = "\xa9\x59", [0x0962] = "\xcb\xfe",
  [0x0964] = "\xa9\x5b", [0x0966] = "\xa9\x5a", [0x0967] = "\x9f\x72",
  [0x0968] = "\xcc\x40", [0x0969] = "\xa9\x58", [0x096a] = "\xa9\x57",
  [0x096b] = "\xcb\xf5", [0x096d] = "\xcb\xf4", [0x096f] = "\xcb\xf2",
  [0x0970] = "\xcb\xf7", [0x0971] = "\xcb\xf6", [0x0972] = "\xcb\xf3",
  [0x0973] = "\xcb\xfc", [0x0974] = "\xcb\xfd", [0x0975] = "\xcb\xfa",
  [0x0976] = "\xcb\xf8", [0x0977] = "\xa9\x56", [0x097a] = "\x9f\xcc",
  [0x097b] = "\xcb\xfb", [0x097c] = "\xa9\x5c", [0x097d] = "\xcc\x41",
  [0x097e] = "\x98\xa5", [0x097f] = "\x92\xe8", [0x0980] = "\xcb\xf9",
  [0x0982] = "\xab\xab", [0x0983] = "\xa9\x55", [0x0988] = "\x9b\xbc",
  [0x098a] = "\x96\xf3", [0x098b] = "\xab\xac", [0x098c] = "\xce\x54",
  [0x098d] = "\x92\xe7", [0x098f] = "\xce\x5a", [0x0990] = "\xfc\x67",
  [0x0993] = "\xab\xb2", [0x0994] = "\xce\x58", [0x0995] = "\xce\x5e",
  [0x0997] = "\xce\x55", [0x0998] = "\xce\x59", [0x0999] = "\xce\x5b",
  [0x099a] = "\xce\x5d", [0x099b] = "\xce\x57", [0x099c] = "\x8b\x7d",
  [0x099d] = "\xce\x56", [0x099e] = "\xce\x51", [0x099f] = "\xce\x52",
  [0x09a0] = "\xab\xad", [0x09a1] = "\x9b\xf4", [0x09a2] = "\xab\xaf",
  [0x09a3] = "\xab\xae", [0x09a4] = "\xce\x53", [0x09a5] = "\xce\x5c",
  [0x09a7] = "\x9e\xf7", [0x09aa] = "\x9e\xc1", [0x09ae] = "\xab\xb1",
  [0x09b4] = "\x99\x6f", [0x09b5] = "\xce\x50", [0x09b6] = "\xd1\x53",
  [0x09b8] = "\xd1\x52", [0x09b9] = "\xd1\x57", [0x09ba] = "\xd1\x4e",
  [0x09bb] = "\x96\xf1", [0x09bc] = "\xd1\x51", [0x09bd] = "\xd1\x50",
  [0x09be] = "\x8e\x41", [0x09bf] = "\xd1\x54", [0x09c1] = "\xd1\x58",
  [0x09c2] = "\xae\x47", [0x09c3] = "\xae\x4a", [0x09c4] = "\x95\x4a",
  [0x09c6] = "\xd1\x4f", [0x09c7] = "\xd1\x55", [0x09c8] = "\x97\xe6",
  [0x09cb] = "\xae\x49", [0x09cc] = "\xd1\x4a", [0x09ce] = "\xab\xb0",
  [0x09cf] = "\xd4\xba", [0x09d0] = "\xd1\x56", [0x09d2] = "\xd1\x4d",
  [0x09d4] = "\xae\x48", [0x09d5] = "\xd1\x4c", [0x09d7] = "\x96\xf5",
  [0x09dc] = "\xd4\xb1", [0x09dd] = "\x92\xe6", [0x09de] = "\x9f\x42",
  [0x09df] = "\xb0\xec", [0x09e0] = "\xb0\xf0", [0x09e1] = "\xd4\xc1",
  [0x09e2] = "\xd4\xaf", [0x09e3] = "\xd4\xbd", [0x09e4] = "\xb0\xf1",
  [0x09e5] = "\xd4\xbf", [0x09e6] = "\xfb\x67", [0x09e7] = "\xd4\xc5",
  [0x09e9] = "\xd4\xc9", [0x09ec] = "\xd4\xc0", [0x09ed] = "\xd4\xb4",
  [0x09ee] = "\xd4\xbc", [0x09ef] = "\x99\xa9", [0x09f0] = "\xd4\xca",
  [0x09f1] = "\xd4\xc8", [0x09f2] = "\xd4\xbe", [0x09f3] = "\xd4\xb9",
  [0x09f4] = "\xd4\xb2", [0x09f5] = "\xd8\xa6", [0x09f6] = "\xd4\xb0",
  [0x09f7] = "\xb0\xf5", [0x09f8] = "\xd4\xb7", [0x09f9] = "\xb0\xf6",
  [0x09fa] = "\xb0\xf2", [0x09fb] = "\xd4\xad", [0x09fc] = "\xd4\xc3",
  [0x09fd] = "\xd4\xb5", [0x09fe] = "\xfa\xe6", [0x0a00] = "\xd4\xb3",
  [0x0a01] = "\xd4\xc6", [0x0a02] = "\xb0\xf3", [0x0a03] = "\xfb\x69",
  [0x0a04] = "\xd4\xcc", [0x0a05] = "\xb0\xed", [0x0a06] = "\xb0\xef",
  [0x0a07] = "\xd4\xbb", [0x0a08] = "\xd4\xb6", [0x0a09] = "\xae\x4b",
  [0x0a0a] = "\xb0\xee", [0x0a0b] = "\xd4\xb8", [0x0a0c] = "\xd4\xc7",
  [0x0a0d] = "\xd4\xcb", [0x0a0e] = "\xd4\xc2", [0x0a10] = "\xd4\xc4",
  [0x0a12] = "\x97\xe5", [0x0a14] = "\xd4\xae", [0x0a19] = "\xd8\xa1",
  [0x0a1b] = "\xd8\xaa", [0x0a1c] = "\xd8\xa9", [0x0a1d] = "\xb3\xfa",
  [0x0a1e] = "\xd8\xa2", [0x0a20] = "\xb3\xfb", [0x0a21] = "\xb3\xf9",
  [0x0a22] = "\x96\x7d", [0x0a23] = "\xd8\xa4", [0x0a24] = "\xb3\xf6",
  [0x0a25] = "\xd8\xa8", [0x0a26] = "\xfb\x6c", [0x0a27] = "\xd8\xa3",
  [0x0a28] = "\xd8\xa5", [0x0a29] = "\xd8\x7d", [0x0a2a] = "\xb3\xf4",
  [0x0a2c] = "\xd8\xb2", [0x0a2d] = "\xd8\xb1", [0x0a2e] = "\xd8\xae",
  [0x0a2f] = "\xb3\xf3", [0x0a30] = "\xb3\xf7", [0x0a31] = "\xb3\xf8",
  [0x0a32] = "\xd1\x4b", [0x0a33] = "\xd8\xab", [0x0a34] = "\xb3\xf5",
  [0x0a35] = "\xb0\xf4", [0x0a36] = "\xd8\xad", [0x0a37] = "\xd8\x7e",
  [0x0a38] = "\xd8\xb0", [0x0a39] = "\xd8\xaf", [0x0a3a] = "\x99\xa2",
  [0x0a3b] = "\xd8\xb3", [0x0a3d] = "\xdc\xef", [0x0a3f] = "\xd8\xac",
  [0x0a40] = "\x9a\xbb", [0x0a44] = "\x9a\x65", [0x0a47] = "\x94\x4e",
  [0x0a48] = "\xd8\xa7", [0x0a49] = "\xdc\xe7", [0x0a4a] = "\xb6\xf4",
  [0x0a4b] = "\xb6\xf7", [0x0a4c] = "\xb6\xf2", [0x0a4d] = "\xdc\xe6",
  [0x0a4e] = "\xdc\xea", [0x0a4f] = "\xdc\xe5", [0x0a51] = "\xb6\xec",
  [0x0a52] = "\xb6\xf6", [0x0a53] = "\xdc\xe2", [0x0a54] = "\xb6\xf0",
  [0x0a55] = "\xdc\xe9", [0x0a57] = "\xb6\xee", [0x0a58] = "\xb6\xed",
  [0x0a59] = "\xdc\xec", [0x0a5a] = "\xb6\xef", [0x0a5b] = "\xdc\xee",
  [0x0a5c] = "\xfb\x6e", [0x0a5d] = "\xdc\xeb", [0x0a5e] = "\xb6\xeb",
  [0x0a5f] = "\x99\xdf", [0x0a62] = "\xb6\xf5", [0x0a63] = "\xdc\xf0",
  [0x0a64] = "\xdc\xe4", [0x0a65] = "\xdc\xed", [0x0a68] = "\xdc\xe3",
  [0x0a69] = "\x98\xe3", [0x0a6b] = "\xb6\xf1", [0x0a6c] = "\x92\x54",
  [0x0a6d] = "\xb6\xf3", [0x0a6f] = "\xdc\xe8", [0x0a71] = "\xdc\xf1",
  [0x0a72] = "\x96\x7b", [0x0a73] = "\x8a\xaf", [0x0a74] = "\xe1\x5d",
  [0x0a75] = "\xb9\xd0", [0x0a76] = "\xe1\x63", [0x0a79] = "\xb9\xd5",
  [0x0a7a] = "\xe1\x5f", [0x0a7b] = "\xe1\x66", [0x0a7c] = "\xe1\x57",
  [0x0a7d] = "\xb9\xd7", [0x0a7e] = "\xb9\xd1", [0x0a7f] = "\xe1\x5c",
  [0x0a80] = "\xbc\x55", [0x0a81] = "\xe1\x5b", [0x0a82] = "\xe1\x64",
  [0x0a83] = "\xb9\xd2", [0x0a85] = "\xb9\xd6", [0x0a86] = "\xe1\x5a",
  [0x0a87] = "\xe1\x60", [0x0a88] = "\xe1\x65", [0x0a89] = "\xe1\x56",
  [0x0a8a] = "\xb9\xd4", [0x0a8b] = "\xe1\x5e", [0x0a8e] = "\xe1\x62",
  [0x0a8f] = "\xe1\x68", [0x0a90] = "\xe1\x58", [0x0a91] = "\xe1\x61",
  [0x0a93] = "\xb9\xd3", [0x0a94] = "\xe1\x67", [0x0a98] = "\xe1\x59",
  [0x0a99] = "\x8b\xaf", [0x0a9a] = "\x9e\xbd", [0x0a9c] = "\xbc\x59",
  [0x0a9d] = "\xe5\x4b", [0x0a9e] = "\xbc\x57", [0x0a9f] = "\xbc\x56",
  [0x0aa0] = "\xe5\x4d", [0x0aa1] = "\xe5\x52", [0x0aa3] = "\xe5\x4e",
  [0x0aa5] = "\xe5\x51", [0x0aa6] = "\xbc\x5c", [0x0aa7] = "\x9e\xe6",
  [0x0aa8] = "\xbe\xa5", [0x0aa9] = "\xbc\x5b", [0x0aaa] = "\xfb\x6f",
  [0x0aab] = "\xe5\x4a", [0x0aac] = "\xe5\x50", [0x0aae] = "\xbc\x5a",
  [0x0aaf] = "\xe5\x4f", [0x0ab0] = "\x8e\xe1", [0x0ab1] = "\xe5\x4c",
  [0x0ab3] = "\xbc\x58", [0x0ab5] = "\x9b\x7d", [0x0ab6] = "\x9c\x7e",
  [0x0aba] = "\xe9\x4d", [0x0abb] = "\xf9\xd9", [0x0abc] = "\xe9\x4f",
  [0x0abd] = "\xe9\x4a", [0x0abe] = "\xbe\xc1", [0x0abf] = "\xe9\x4c",
  [0x0ac1] = "\xbe\xc0", [0x0ac2] = "\xe9\x4e", [0x0ac5] = "\xbe\xc3",
  [0x0ac6] = "\xe9\x50", [0x0ac7] = "\xbe\xc2", [0x0ac8] = "\xe9\x49",
  [0x0ac9] = "\xe9\x4b", [0x0acb] = "\x92\xea", [0x0ace] = "\xc0\xa5",
  [0x0acf] = "\xec\xcc", [0x0ad1] = "\xc0\xa4", [0x0ad2] = "\xec\xcd",
  [0x0ad3] = "\xc0\xa3", [0x0ad4] = "\xec\xcb", [0x0ad5] = "\xc0\xa2",
  [0x0ad6] = "\xec\xca", [0x0ad8] = "\xc2\x53", [0x0ad9] = "\xc2\x52",
  [0x0ada] = "\xf1\xf6", [0x0adb] = "\xf1\xf8", [0x0adc] = "\xfb\x72",
  [0x0add] = "\xf1\xf7", [0x0ade] = "\xc3\x61", [0x0adf] = "\xc3\x62",
  [0x0ae0] = "\xfb\x71", [0x0ae2] = "\xc3\x63", [0x0ae3] = "\xf4\x42",
  [0x0ae4] = "\xc4\x5b", [0x0ae7] = "\xf7\xd3", [0x0ae8] = "\xf7\xd2",
  [0x0ae9] = "\xc5\xf2", [0x0aeb] = "\xa4\x68", [0x0aec] = "\xa4\xd0",
  [0x0aef] = "\xa7\xa7", [0x0af0] = "\x89\x5c", [0x0af2] = "\x98\xf0",
  [0x0af3] = "\x96\xf2", [0x0af4] = "\xce\x5f", [0x0af9] = "\xb3\xfc",
  [0x0afa] = "\xb3\xfd", [0x0afb] = "\xfb\x74", [0x0afc] = "\xdc\xf2",
  [0x0afd] = "\xb9\xd8", [0x0afe] = "\xe1\x69", [0x0aff] = "\xe5\x53",
  [0x0b02] = "\x8b\xc1", [0x0b03] = "\xc9\x5a", [0x0b04] = "\x89\x5d",
  [0x0b05] = "\x89\xde", [0x0b06] = "\xca\xb0", [0x0b07] = "\x89\x5e",
  [0x0b0a] = "\xc6\xca", [0x0b0c] = "\xcc\x42", [0x0b0d] = "\xce\x60",
  [0x0b0e] = "\xd1\x59", [0x0b0f] = "\xae\x4c", [0x0b11] = "\xfe\x42",
  [0x0b12] = "\xf1\xf9", [0x0b14] = "\xc4\xdc", [0x0b15] = "\xa4\x69",
  [0x0b16] = "\xa5\x7e", [0x0b17] = "\xc9\x70", [0x0b19] = "\xa6\x67",
  [0x0b1a] = "\xa6\x68", [0x0b1c] = "\xa9\x5d", [0x0b1f] = "\xfb\x7b",
  [0x0b20] = "\xb0\xf7", [0x0b22] = "\xb9\xda", [0x0b24] = "\xb9\xdb",
  [0x0b25] = "\xb9\xd9", [0x0b27] = "\xa4\x6a", [0x0b29] = "\xa4\xd1",
  [0x0b2a] = "\xa4\xd3", [0x0b2b] = "\xa4\xd2", [0x0b2c] = "\xc9\x5b",
  [0x0b2d] = "\xa4\xd4", [0x0b2e] = "\xa5\xa1", [0x0b2f] = "\xc9\x71",
  [0x0b31] = "\xa5\xa2", [0x0b32] = "\x89\x5f", [0x0b34] = "\x89\x60",
  [0x0b37] = "\xa6\x69", [0x0b38] = "\xa6\x6a", [0x0b3c] = "\xc9\xcb",
  [0x0b3e] = "\xa7\xa8", [0x0b40] = "\xca\xb1", [0x0b44] = "\xa9\x61",
  [0x0b45] = "\xcc\x43", [0x0b47] = "\xa9\x5f", [0x0b48] = "\xa9\x60",
  [0x0b49] = "\xa9\x5e", [0x0b4a] = "\xd1\x5a", [0x0b4e] = "\xab\xb6",
  [0x0b4f] = "\xab\xb5", [0x0b50] = "\xab\xb7", [0x0b51] = "\xab\xb4",
  [0x0b53] = "\xce\x61", [0x0b54] = "\xa9\x62", [0x0b55] = "\xab\xb3",
  [0x0b57] = "\xae\x4d", [0x0b58] = "\xae\x4e", [0x0b5a] = "\xae\x4f",
  [0x0b5c] = "\xd4\xcd", [0x0b60] = "\xb3\xfe", [0x0b61] = "\xd8\xb4",
  [0x0b62] = "\xb0\xf8", [0x0b65] = "\x9b\xcd", [0x0b67] = "\xb6\xf8",
  [0x0b69] = "\xb9\xdd", [0x0b6a] = "\xb9\xdc", [0x0b6b] = "\xe1\x6a",
  [0x0b6d] = "\xbc\x5d", [0x0b6e] = "\xbe\xc4", [0x0b70] = "\xef\xc0",
  [0x0b71] = "\xf6\xda", [0x0b72] = "\xf7\xd4", [0x0b73] = "\xa4\x6b",
  [0x0b74] = "\xa5\xa3", [0x0b75] = "\x9d\xd3", [0x0b76] = "\xa5\xa4",
  [0x0b77] = "\xc9\xd1", [0x0b78] = "\xa6\x6c", [0x0b79] = "\xa6\x6f",
  [0x0b7b] = "\xc9\xcf", [0x0b7c] = "\xc9\xcd", [0x0b7d] = "\xa6\x6e",
  [0x0b7e] = "\xc9\xd0", [0x0b7f] = "\xc9\xd2", [0x0b80] = "\xc9\xcc",
  [0x0b81] = "\xa6\x71", [0x0b82] = "\xa6\x70", [0x0b83] = "\xa6\x6d",
  [0x0b84] = "\xa6\x6b", [0x0b85] = "\xc9\xce", [0x0b89] = "\x98\x4c",
  [0x0b8a] = "\xa7\xb3", [0x0b8d] = "\xa7\xb0", [0x0b8e] = "\xca\xb6",
  [0x0b8f] = "\xca\xb9", [0x0b90] = "\xca\xb8", [0x0b92] = "\xa7\xaa",
  [0x0b93] = "\xa7\xb2", [0x0b94] = "\x97\x52", [0x0b96] = "\xa7\xaf",
  [0x0b97] = "\xca\xb5", [0x0b98] = "\xca\xb3", [0x0b99] = "\xa7\xae",
  [0x0b9a] = "\x95\xc3", [0x0b9d] = "\xa7\xa9", [0x0b9e] = "\xa7\xac",
  [0x0b9f] = "\x9b\xb6", [0x0ba0] = "\xca\xb4", [0x0ba1] = "\xca\xbb",
  [0x0ba2] = "\xca\xb7", [0x0ba3] = "\xa7\xad", [0x0ba4] = "\xa7\xb1",
  [0x0ba5] = "\xa7\xb4", [0x0ba6] = "\xca\xb2", [0x0ba7] = "\xca\xba",
  [0x0ba8] = "\xa7\xab", [0x0bac] = "\x9a\xb9", [0x0bae] = "\xa9\x67",
  [0x0baf] = "\xa9\x6f", [0x0bb0] = "\x97\xb3", [0x0bb1] = "\xcc\x4f",
  [0x0bb2] = "\xcc\x48", [0x0bb3] = "\xa9\x70", [0x0bb4] = "\xcc\x53",
  [0x0bb5] = "\xcc\x44", [0x0bb6] = "\xcc\x4b", [0x0bb7] = "\x9f\x74",
  [0x0bb8] = "\x92\xf1", [0x0bb9] = "\xa9\x66", [0x0bba] = "\xcc\x45",
  [0x0bbb] = "\xa9\x64", [0x0bbc] = "\xcc\x4c", [0x0bbd] = "\xcc\x50",
  [0x0bbe] = "\xa9\x63", [0x0bc0] = "\xcc\x51", [0x0bc1] = "\xcc\x4a",
  [0x0bc3] = "\xcc\x4d", [0x0bc4] = "\x97\xdf", [0x0bc5] = "\xa9\x72",
  [0x0bc6] = "\xa9\x69", [0x0bc7] = "\xcc\x54", [0x0bc8] = "\xcc\x52",
  [0x0bc9] = "\xfb\xa6", [0x0bca] = "\xa9\x6e", [0x0bcb] = "\xa9\x6c",
  [0x0bcc] = "\xcc\x49", [0x0bcd] = "\xa9\x6b", [0x0bce] = "\xcc\x47",
  [0x0bcf] = "\xcc\x46", [0x0bd0] = "\xa9\x6a", [0x0bd1] = "\xa9\x68",
  [0x0bd2] = "\xa9\x71", [0x0bd3] = "\xa9\x6d", [0x0bd4] = "\xa9\x65",
  [0x0bd6] = "\xcc\x4e", [0x0bd8] = "\xab\xb9", [0x0bd9] = "\xfb\xab",
  [0x0bda] = "\xab\xc0", [0x0bdb] = "\xce\x6f", [0x0bdc] = "\xab\xb8",
  [0x0bdd] = "\xce\x67", [0x0bde] = "\xce\x63", [0x0be0] = "\xce\x73",
  [0x0be1] = "\xce\x62", [0x0be3] = "\xab\xbb", [0x0be4] = "\xce\x6c",
  [0x0be5] = "\xab\xbe", [0x0be6] = "\xab\xc1", [0x0be8] = "\xab\xbc",
  [0x0be9] = "\xce\x70", [0x0bea] = "\xab\xbf", [0x0beb] = "\x98\x77",
  [0x0bec] = "\xae\x56", [0x0bed] = "\xce\x76", [0x0bee] = "\xce\x64",
  [0x0bef] = "\x98\x54", [0x0bf0] = "\x95\xc5", [0x0bf1] = "\xce\x66",
  [0x0bf2] = "\xce\x6d", [0x0bf3] = "\xce\x71", [0x0bf4] = "\xce\x75",
  [0x0bf5] = "\xce\x72", [0x0bf6] = "\xce\x6b", [0x0bf7] = "\xce\x6e",
  [0x0bf8] = "\x9d\x55", [0x0bf9] = "\xfb\xb2", [0x0bfa] = "\xce\x68",
  [0x0bfb] = "\xab\xc3", [0x0bfc] = "\xce\x6a", [0x0bfd] = "\xce\x69",
  [0x0bfe] = "\xce\x74", [0x0bff] = "\xab\xba", [0x0c00] = "\xce\x65",
  [0x0c01] = "\xab\xc2", [0x0c02] = "\x95\x7e", [0x0c03] = "\xab\xbd",
  [0x0c09] = "\xae\x5c", [0x0c0a] = "\xd1\x62", [0x0c0b] = "\x97\x42",
  [0x0c0c] = "\xae\x5b", [0x0c0d] = "\x94\xe6", [0x0c0f] = "\xd1\x60",
  [0x0c11] = "\xae\x50", [0x0c12] = "\x92\xf5", [0x0c13] = "\xae\x55",
  [0x0c15] = "\xd1\x5f", [0x0c16] = "\xd1\x5c", [0x0c17] = "\xd1\x61",
  [0x0c18] = "\xae\x51", [0x0c19] = "\xd1\x5b", [0x0c1b] = "\xae\x54",
  [0x0c1c] = "\xae\x52", [0x0c1e] = "\xd1\x63", [0x0c1f] = "\xae\x53",
  [0x0c20] = "\xae\x57", [0x0c21] = "\x92\xfd", [0x0c23] = "\xae\x58",
  [0x0c24] = "\xfb\xa2", [0x0c25] = "\xae\x5a", [0x0c27] = "\x9c\x51",
  [0x0c29] = "\xae\x59", [0x0c2a] = "\x94\xe9", [0x0c2b] = "\x98\x5c",
  [0x0c2c] = "\x92\xf0", [0x0c2d] = "\xd1\x5d", [0x0c2e] = "\xd1\x5e",
  [0x0c33] = "\xd1\x64", [0x0c35] = "\xd4\xd4", [0x0c36] = "\xb0\xf9",
  [0x0c37] = "\xd8\xc2", [0x0c38] = "\xd4\xd3", [0x0c39] = "\xd4\xe6",
  [0x0c3c] = "\xb1\x40", [0x0c3d] = "\x94\x4c", [0x0c3e] = "\xd4\xe4",
  [0x0c40] = "\xb0\xfe", [0x0c41] = "\xb0\xfa", [0x0c42] = "\xd4\xed",
  [0x0c43] = "\xd4\xdd", [0x0c44] = "\xd4\xe0", [0x0c45] = "\x91\x6b",
  [0x0c46] = "\xb1\x43", [0x0c47] = "\xd4\xea", [0x0c48] = "\xd4\xe2",
  [0x0c49] = "\xb0\xfb", [0x0c4a] = "\xb1\x44", [0x0c4c] = "\xd4\xe7",
  [0x0c4d] = "\xd4\xe5", [0x0c50] = "\xd4\xd6", [0x0c51] = "\xd4\xeb",
  [0x0c52] = "\xd4\xdf", [0x0c53] = "\xd4\xda", [0x0c54] = "\x8b\x78",
  [0x0c55] = "\xd4\xd0", [0x0c56] = "\xd4\xec", [0x0c57] = "\xd4\xdc",
  [0x0c58] = "\xd4\xcf", [0x0c59] = "\x94\xe2", [0x0c5a] = "\xb1\x42",
  [0x0c5b] = "\xd4\xe1", [0x0c5c] = "\xd4\xee", [0x0c5d] = "\xd4\xde",
  [0x0c5e] = "\xd4\xd2", [0x0c5f] = "\xd4\xd7", [0x0c60] = "\xd4\xce",
  [0x0c61] = "\x98\x4f", [0x0c62] = "\xb1\x41", [0x0c63] = "\xfb\xb5",
  [0x0c64] = "\xd4\xdb", [0x0c65] = "\xd4\xd8", [0x0c66] = "\xb0\xfc",
  [0x0c67] = "\xd4\xd1", [0x0c68] = "\x92\x71", [0x0c69] = "\xd4\xe9",
  [0x0c6a] = "\xb0\xfd", [0x0c6b] = "\x93\x65", [0x0c6c] = "\xd4\xd9",
  [0x0c6d] = "\xd4\xd5", [0x0c6e] = "\x98\x5b", [0x0c70] = "\xd4\xe8",
  [0x0c71] = "\x98\x50", [0x0c77] = "\xb4\x40", [0x0c78] = "\xd8\xbb",
  [0x0c79] = "\x97\xbc", [0x0c7a] = "\xd8\xb8", [0x0c7b] = "\xd8\xc9",
  [0x0c7c] = "\xd8\xbd", [0x0c7d] = "\xd8\xca", [0x0c7e] = "\x92\xf3",
  [0x0c7f] = "\xb4\x42", [0x0c81] = "\x93\x40", [0x0c82] = "\x98\x4d",
  [0x0c83] = "\xd8\xc6", [0x0c84] = "\xd8\xc3", [0x0c86] = "\x95\x72",
  [0x0c88] = "\xfd\xef", [0x0c8a] = "\xd8\xc4", [0x0c8b] = "\xd8\xc7",
  [0x0c8c] = "\xd8\xcb", [0x0c8e] = "\xd4\xe3", [0x0c8f] = "\xd8\xcd",
  [0x0c90] = "\xdd\x47", [0x0c91] = "\xfd\xc1", [0x0c92] = "\xb4\x43",
  [0x0c93] = "\xd8\xce", [0x0c94] = "\xd8\xb6", [0x0c95] = "\xd8\xc0",
  [0x0c96] = "\xfb\xba", [0x0c97] = "\xd8\xc5", [0x0c99] = "\x92\xeb",
  [0x0c9a] = "\xb4\x41", [0x0c9b] = "\xb4\x44", [0x0c9c] = "\xd8\xcc",
  [0x0c9d] = "\xd8\xcf", [0x0c9e] = "\xd8\xba", [0x0c9f] = "\xd8\xb7",
  [0x0ca0] = "\xfc\x73", [0x0ca1] = "\x97\xb7", [0x0ca2] = "\xd8\xb9",
  [0x0ca5] = "\xd8\xbe", [0x0ca6] = "\xd8\xbc", [0x0ca7] = "\xb4\x45",
  [0x0ca9] = "\xd8\xc8", [0x0cab] = "\xfb\xb4", [0x0cac] = "\xd8\xbf",
  [0x0cae] = "\xd8\xc1", [0x0caf] = "\xd8\xb5", [0x0cb0] = "\xdc\xfa",
  [0x0cb1] = "\xdc\xf8", [0x0cb2] = "\xb7\x42", [0x0cb3] = "\xb7\x40",
  [0x0cb4] = "\xdd\x43", [0x0cb5] = "\xdc\xf9", [0x0cb6] = "\xdd\x44",
  [0x0cb7] = "\xdd\x40", [0x0cb8] = "\xdc\xf7", [0x0cb9] = "\xdd\x46",
  [0x0cba] = "\xdc\xf6", [0x0cbb] = "\xdc\xfd", [0x0cbc] = "\xb6\xfe",
  [0x0cbd] = "\xb6\xfd", [0x0cbe] = "\xb6\xfc", [0x0cbf] = "\xdc\xfb",
  [0x0cc0] = "\xdd\x41", [0x0cc1] = "\xb6\xf9", [0x0cc2] = "\xb7\x41",
  [0x0cc3] = "\x90\xa7", [0x0cc4] = "\xdc\xf4", [0x0cc6] = "\xdc\xfe",
  [0x0cc7] = "\xdc\xf3", [0x0cc8] = "\xdc\xfc", [0x0cc9] = "\xb6\xfa",
  [0x0cca] = "\xdd\x42", [0x0ccb] = "\xdc\xf5", [0x0ccc] = "\xb6\xfb",
  [0x0ccd] = "\xdd\x45", [0x0cce] = "\x97\x41", [0x0ccf] = "\x92\xf4",
  [0x0cd3] = "\xfb\xbc", [0x0cd5] = "\xe1\x6e", [0x0cd6] = "\xb9\xe2",
  [0x0cd7] = "\xb9\xe1", [0x0cd8] = "\xb9\xe3", [0x0cd9] = "\xe1\x7a",
  [0x0cda] = "\xe1\x70", [0x0cdb] = "\xe1\x76", [0x0cdc] = "\xe1\x6b",
  [0x0cdd] = "\xe1\x79", [0x0cde] = "\xe1\x78", [0x0cdf] = "\xe1\x7c",
  [0x0ce0] = "\xe1\x75", [0x0ce1] = "\xb9\xde", [0x0ce2] = "\xe1\x74",
  [0x0ce3] = "\xb9\xe4", [0x0ce4] = "\x95\x77", [0x0ce5] = "\xe1\x6d",
  [0x0ce6] = "\xb9\xdf", [0x0ce8] = "\xe1\x7b", [0x0ce9] = "\xb9\xe0",
  [0x0cea] = "\xe1\x6f", [0x0ceb] = "\xe1\x72", [0x0cec] = "\xe1\x77",
  [0x0ced] = "\xe1\x71", [0x0cee] = "\xe1\x6c", [0x0cf0] = "\x9e\xe2",
  [0x0cf2] = "\x8f\x78", [0x0cf3] = "\xe1\x73", [0x0cf4] = "\xe5\x55",
  [0x0cf5] = "\xbc\x61", [0x0cf6] = "\xe5\x58", [0x0cf7] = "\xe5\x57",
  [0x0cf8] = "\xe5\x5a", [0x0cf9] = "\xe5\x5c", [0x0cfa] = "\xf9\xdc",
  [0x0cfb] = "\xbc\x5f", [0x0cfd] = "\xe5\x56", [0x0cfe] = "\x96\x72",
  [0x0cff] = "\xe5\x54", [0x0d01] = "\xe5\x5d", [0x0d02] = "\xe5\x5b",
  [0x0d03] = "\xe5\x59", [0x0d05] = "\xe5\x5f", [0x0d07] = "\xe5\x5e",
  [0x0d08] = "\xbc\x63", [0x0d09] = "\xbc\x5e", [0x0d0b] = "\xbc\x60",
  [0x0d0c] = "\xbc\x62", [0x0d0d] = "\x9e\xb5", [0x0d0f] = "\xe5\x60",
  [0x0d10] = "\xe9\x57", [0x0d11] = "\x96\x4b", [0x0d13] = "\xe9\x56",
  [0x0d14] = "\xe9\x55", [0x0d16] = "\xe9\x58", [0x0d17] = "\xe9\x51",
  [0x0d19] = "\xe9\x52", [0x0d1a] = "\xe9\x5a", [0x0d1b] = "\xe9\x53",
  [0x0d1d] = "\xbe\xc5", [0x0d1e] = "\xe9\x5c", [0x0d1f] = "\xa0\xfa",
  [0x0d20] = "\xe9\x5b", [0x0d21] = "\xe9\x54", [0x0d23] = "\xec\xd1",
  [0x0d24] = "\xc0\xa8", [0x0d25] = "\xec\xcf", [0x0d26] = "\xec\xd4",
  [0x0d27] = "\xec\xd3", [0x0d28] = "\xe9\x59", [0x0d2a] = "\xc0\xa7",
  [0x0d2b] = "\x95\x75", [0x0d2c] = "\xec\xd2", [0x0d2d] = "\xec\xce",
  [0x0d2e] = "\xec\xd6", [0x0d2f] = "\xec\xd5", [0x0d30] = "\xc0\xa6",
  [0x0d32] = "\xec\xd0", [0x0d34] = "\xbe\xc6", [0x0d38] = "\xc2\x54",
  [0x0d3c] = "\xef\xc1", [0x0d3d] = "\xf1\xfa", [0x0d3e] = "\xf1\xfb",
  [0x0d3f] = "\xf1\xfc", [0x0d40] = "\xc4\x5c", [0x0d41] = "\x90\xda",
  [0x0d43] = "\xc4\x5d", [0x0d44] = "\x93\x67", [0x0d45] = "\xf4\x43",
  [0x0d46] = "\xfe\xa4", [0x0d47] = "\xf5\xc8", [0x0d48] = "\xf5\xc7",
  [0x0d4a] = "\x90\xdf", [0x0d4b] = "\xf6\xdb", [0x0d4c] = "\xf6\xdc",
  [0x0d4d] = "\xf7\xd5", [0x0d4e] = "\xf8\xa7", [0x0d4f] = "\x93\x54",
  [0x0d50] = "\xa4\x6c", [0x0d51] = "\xa4\x6d", [0x0d53] = "\xa4\x6e",
  [0x0d54] = "\xa4\xd5", [0x0d55] = "\xa5\xa5", [0x0d56] = "\xc9\xd3",
  [0x0d57] = "\xa6\x72", [0x0d58] = "\xa6\x73", [0x0d5a] = "\xa7\xb7",
  [0x0d5b] = "\xa7\xb8", [0x0d5c] = "\xa7\xb6", [0x0d5d] = "\xa7\xb5",
  [0x0d5f] = "\xa9\x73", [0x0d62] = "\xcc\x55", [0x0d63] = "\xa9\x75",
  [0x0d64] = "\xa9\x74", [0x0d65] = "\xcc\x56", [0x0d66] = "\x89\x61",
  [0x0d68] = "\x8b\xb4", [0x0d69] = "\xab\xc4", [0x0d6b] = "\xae\x5d",
  [0x0d6c] = "\xd1\x65", [0x0d6d] = "\x9d\xc0", [0x0d6e] = "\xd4\xf0",
  [0x0d70] = "\xb1\x45", [0x0d71] = "\xb4\x47", [0x0d72] = "\xd4\xef",
  [0x0d73] = "\xb4\x46", [0x0d74] = "\x8e\x48", [0x0d75] = "\xb9\xe5",
  [0x0d76] = "\xfb\xc5", [0x0d77] = "\xe1\x7d", [0x0d78] = "\xbe\xc7",
  [0x0d7a] = "\xc0\xa9", [0x0d7b] = "\xec\xd7", [0x0d7c] = "\xfb\xc7",
  [0x0d7d] = "\xc4\x5e", [0x0d7f] = "\xc5\x70", [0x0d80] = "\xc6\xcb",
  [0x0d81] = "\xc9\x72", [0x0d82] = "\xfa\x79", [0x0d83] = "\xa5\xa6",
  [0x0d84] = "\xc9\x73", [0x0d85] = "\xa6\x76", [0x0d87] = "\xa6\x74",
  [0x0d88] = "\xa6\x75", [0x0d89] = "\xa6\x77", [0x0d8b] = "\xa7\xba",
  [0x0d8c] = "\xa7\xb9", [0x0d8e] = "\xca\xbc", [0x0d8f] = "\xa7\xbb",
  [0x0d90] = "\x9e\x67", [0x0d92] = "\xca\xbd", [0x0d93] = "\xcc\x57",
  [0x0d95] = "\xcc\x58", [0x0d97] = "\xa9\x76", [0x0d98] = "\xa9\x78",
  [0x0d99] = "\xa9\x7a", [0x0d9a] = "\xa9\x77", [0x0d9b] = "\xa9\x7b",
  [0x0d9c] = "\xa9\x79", [0x0d9d] = "\xfb\xd2", [0x0d9e] = "\x89\x62",
  [0x0d9f] = "\x89\x63", [0x0da2] = "\xab\xc8", [0x0da3] = "\xab\xc5",
  [0x0da4] = "\xab\xc7", [0x0da5] = "\xab\xc9", [0x0da6] = "\xab\xc6",
  [0x0da7] = "\xd1\x66", [0x0da8] = "\xce\x77", [0x0daa] = "\xfc\x7d",
  [0x0dac] = "\xd1\x68", [0x0dad] = "\xd1\x67", [0x0dae] = "\xae\x63",
  [0x0db0] = "\xae\x5f", [0x0db3] = "\xae\x60", [0x0db4] = "\xae\x62",
  [0x0db5] = "\xae\x64", [0x0db6] = "\xae\x61", [0x0db8] = "\xae\x66",
  [0x0db9] = "\xae\x65", [0x0dbf] = "\xb1\x4a", [0x0dc0] = "\xd4\xf2",
  [0x0dc1] = "\xd4\xf1", [0x0dc2] = "\xb1\x49", [0x0dc3] = "\x9f\x6b",
  [0x0dc4] = "\xb1\x48", [0x0dc5] = "\xb1\x47", [0x0dc6] = "\xb1\x4b",
  [0x0dc7] = "\xb1\x46", [0x0dca] = "\xd8\xd5", [0x0dcb] = "\xd8\xd2",
  [0x0dcc] = "\xb4\x49", [0x0dcd] = "\xd8\xd1", [0x0dce] = "\xd8\xd6",
  [0x0dd0] = "\xb4\x4b", [0x0dd1] = "\xd8\xd4", [0x0dd2] = "\xb4\x48",
  [0x0dd3] = "\xb4\x4a", [0x0dd4] = "\xd8\xd3", [0x0dd5] = "\xfb\xcc",
  [0x0dd6] = "\xdd\x48", [0x0dd7] = "\xfe\xae", [0x0dd8] = "\xdd\x49",
  [0x0dd9] = "\xdd\x4a", [0x0dde] = "\xb9\xe6", [0x0ddf] = "\xb9\xee",
  [0x0de0] = "\xe1\x7e", [0x0de1] = "\xb9\xe8", [0x0de2] = "\xb9\xec",
  [0x0de3] = "\xe1\xa1", [0x0de4] = "\xb9\xed", [0x0de5] = "\xb9\xe9",
  [0x0de6] = "\xb9\xea", [0x0de7] = "\xb9\xe7", [0x0de8] = "\xb9\xeb",
  [0x0de9] = "\xbc\x66", [0x0dea] = "\xd8\xd0", [0x0deb] = "\xbc\x67",
  [0x0dec] = "\xbc\x65", [0x0dee] = "\xbc\x64", [0x0def] = "\xe9\x5d",
  [0x0df0] = "\xbe\xc8", [0x0df1] = "\xec\xd8", [0x0df2] = "\xec\xd9",
  [0x0df3] = "\xfb\xd1", [0x0df5] = "\xc3\x64", [0x0df6] = "\xc4\x5f",
  [0x0df8] = "\xa4\x6f", [0x0dfa] = "\xa6\x78", [0x0dff] = "\xfb\x75",
  [0x0e01] = "\xab\xca", [0x0e03] = "\xd1\x69", [0x0e04] = "\xae\x67",
  [0x0e05] = "\xfb\xd4", [0x0e07] = "\xb1\x4e", [0x0e08] = "\xb1\x4d",
  [0x0e09] = "\xb1\x4c", [0x0e0a] = "\xb4\x4c", [0x0e0b] = "\xb4\x4d",
  [0x0e0c] = "\xd8\xd7", [0x0e0d] = "\xb9\xef", [0x0e0e] = "\xbe\xc9",
  [0x0e0f] = "\xa4\x70", [0x0e10] = "\xc9\x5c", [0x0e11] = "\xa4\xd6",
  [0x0e12] = "\xc9\x74", [0x0e13] = "\xfb\xd6", [0x0e14] = "\xfb\xd8",
  [0x0e15] = "\xc9\xd4", [0x0e16] = "\xa6\x79", [0x0e1a] = "\xa9\x7c",
  [0x0e1c] = "\x8b\x5d", [0x0e1e] = "\x93\x4c", [0x0e1f] = "\xdd\x4b",
  [0x0e20] = "\x9a\xe2", [0x0e22] = "\xa4\x71", [0x0e23] = "\x8b\xc9",
  [0x0e24] = "\xa4\xd7", [0x0e25] = "\xc9\xd5", [0x0e28] = "\xca\xbe",
  [0x0e2a] = "\xca\xbf", [0x0e2c] = "\xa7\xbc", [0x0e30] = "\xd8\xd8",
  [0x0e31] = "\xb4\x4e", [0x0e33] = "\xdd\x4c", [0x0e37] = "\xc0\xaa",
  [0x0e38] = "\xa4\x72", [0x0e39] = "\xa4\xa8", [0x0e3a] = "\xa4\xd8",
  [0x0e3b] = "\xc9\x75", [0x0e3c] = "\xa5\xa7", [0x0e3e] = "\xa7\xc0",
  [0x0e3f] = "\xa7\xbf", [0x0e40] = "\xa7\xbd", [0x0e41] = "\xa7\xbe",
  [0x0e44] = "\xcc\x59", [0x0e45] = "\xa9\x7e", [0x0e46] = "\xa9\xa1",
  [0x0e47] = "\xcc\x5a", [0x0e48] = "\xa9\x7d", [0x0e49] = "\xfb\xdb",
  [0x0e4a] = "\x9f\xc9", [0x0e4b] = "\xab\xce", [0x0e4c] = "\xce\x78",
  [0x0e4d] = "\xab\xcd", [0x0e4e] = "\xab\xcb", [0x0e4f] = "\xab\xcc",
  [0x0e50] = "\xae\x6a", [0x0e51] = "\xae\x68", [0x0e53] = "\x9f\x44",
  [0x0e54] = "\xd1\x6b", [0x0e55] = "\xae\x69", [0x0e56] = "\xd1\x6a",
  [0x0e58] = "\xae\x5e", [0x0e59] = "\xd4\xf3", [0x0e5c] = "\xb1\x50",
  [0x0e5d] = "\xb1\x51", [0x0e5e] = "\x98\xed", [0x0e60] = "\xb1\x4f",
  [0x0e62] = "\xb9\xf0", [0x0e63] = "\xe1\xa2", [0x0e64] = "\xbc\x68",
  [0x0e65] = "\xbc\x69", [0x0e67] = "\xe5\x61", [0x0e68] = "\xc0\xab",
  [0x0e69] = "\xef\xc2", [0x0e6a] = "\xef\xc3", [0x0e6c] = "\xc4\xdd",
  [0x0e6d] = "\xf8\xa8", [0x0e6e] = "\xc9\x4b", [0x0e6f] = "\xa4\xd9",
  [0x0e71] = "\xa4\x73", [0x0e73] = "\xc9\x77", [0x0e74] = "\xc9\x76",
  [0x0e79] = "\xa6\x7a", [0x0e7a] = "\xc9\xd7", [0x0e7b] = "\xc9\xd8",
  [0x0e7c] = "\xc9\xd6", [0x0e7e] = "\xc9\xd9", [0x0e85] = "\xfb\xdd",
  [0x0e86] = "\xca\xc7", [0x0e88] = "\xca\xc2", [0x0e89] = "\xca\xc4",
  [0x0e8a] = "\xca\xc6", [0x0e8b] = "\xca\xc3", [0x0e8c] = "\xa7\xc4",
  [0x0e8d] = "\xca\xc0", [0x0e8f] = "\xca\xc1", [0x0e90] = "\xa7\xc1",
  [0x0e91] = "\xa7\xc2", [0x0e92] = "\xca\xc5", [0x0e93] = "\xca\xc8",
  [0x0e94] = "\xa7\xc3", [0x0e95] = "\xca\xc9", [0x0e99] = "\x8d\xf2",
  [0x0e9a] = "\x89\x64", [0x0e9c] = "\xfd\xf2", [0x0e9d] = "\xcc\x68",
  [0x0e9e] = "\x93\x4d", [0x0e9f] = "\xcc\x62", [0x0ea0] = "\xcc\x5d",
  [0x0ea1] = "\xa9\xa3", [0x0ea2] = "\xcc\x65", [0x0ea3] = "\xcc\x63",
  [0x0ea4] = "\xcc\x5c", [0x0ea5] = "\xcc\x69", [0x0ea6] = "\xcc\x6c",
  [0x0ea7] = "\xcc\x67", [0x0ea8] = "\xcc\x60", [0x0ea9] = "\xa9\xa5",
  [0x0eaa] = "\xcc\x66", [0x0eab] = "\xa9\xa6", [0x0eac] = "\xcc\x61",
  [0x0ead] = "\xcc\x64", [0x0eae] = "\xcc\x5b", [0x0eaf] = "\xcc\x5f",
  [0x0eb0] = "\xcc\x6b", [0x0eb1] = "\xa9\xa7", [0x0eb3] = "\xa9\xa8",
  [0x0eb5] = "\xcc\x5e", [0x0eb6] = "\xcc\x6a", [0x0eb7] = "\xa9\xa2",
  [0x0eb8] = "\xa9\xa4", [0x0eba] = "\xfb\xe7", [0x0ec1] = "\xa0\xf2",
  [0x0ec2] = "\x98\x68", [0x0ec6] = "\xce\xab", [0x0ec7] = "\xce\xa4",
  [0x0ec8] = "\xce\xaa", [0x0ec9] = "\xce\xa3", [0x0eca] = "\xce\xa5",
  [0x0ecb] = "\xce\x7d", [0x0ecc] = "\xce\x7b", [0x0ece] = "\xce\xac",
  [0x0ecf] = "\xce\xa9", [0x0ed0] = "\xce\x79", [0x0ed1] = "\x9f\x58",
  [0x0ed2] = "\xab\xd0", [0x0ed3] = "\xce\xa7", [0x0ed4] = "\xce\xa8",
  [0x0ed6] = "\xce\xa6", [0x0ed7] = "\xce\x7c", [0x0ed8] = "\xce\x7a",
  [0x0ed9] = "\xab\xcf", [0x0eda] = "\xce\xa2", [0x0edb] = "\xce\x7e",
  [0x0ede] = "\xce\xa1", [0x0edf] = "\xce\xad", [0x0ee5] = "\x8d\x73",
  [0x0ee8] = "\xae\x6f", [0x0ee9] = "\xfb\xde", [0x0eea] = "\xae\x6e",
  [0x0eec] = "\xd1\x6c", [0x0eed] = "\xae\x6b", [0x0eee] = "\xd1\x6e",
  [0x0eef] = "\xfb\xdf", [0x0ef0] = "\xae\x70", [0x0ef1] = "\xd1\x6f",
  [0x0ef4] = "\xae\x73", [0x0ef6] = "\xae\x71", [0x0ef7] = "\xd1\x70",
  [0x0ef8] = "\xce\xae", [0x0ef9] = "\xd1\x72", [0x0efb] = "\xae\x6d",
  [0x0efd] = "\xae\x6c", [0x0eff] = "\xd1\x6d", [0x0f00] = "\xd1\x71",
  [0x0f01] = "\xae\x72", [0x0f06] = "\xb1\x53", [0x0f07] = "\xb1\x52",
  [0x0f0b] = "\xd4\xf5", [0x0f0c] = "\xd4\xf9", [0x0f0d] = "\xd4\xfb",
  [0x0f0e] = "\xb1\x54", [0x0f0f] = "\xd4\xfe", [0x0f10] = "\xfb\xe3",
  [0x0f11] = "\xb1\x58", [0x0f12] = "\xd5\x41", [0x0f14] = "\xb1\x5a",
  [0x0f15] = "\x8d\xa8", [0x0f16] = "\xb1\x56", [0x0f17] = "\xb1\x5e",
  [0x0f18] = "\xfb\xe4", [0x0f19] = "\xb1\x5b", [0x0f1a] = "\xd4\xf7",
  [0x0f1b] = "\xb1\x55", [0x0f1d] = "\xd4\xf6", [0x0f1e] = "\xd4\xf4",
  [0x0f1f] = "\xd5\x43", [0x0f20] = "\xd4\xf8", [0x0f22] = "\xb1\x57",
  [0x0f23] = "\xd5\x42", [0x0f24] = "\xb1\x5c", [0x0f25] = "\xd4\xfd",
  [0x0f26] = "\xd4\xfc", [0x0f27] = "\xb1\x5d", [0x0f28] = "\xd4\xfa",
  [0x0f29] = "\xb1\x59", [0x0f2c] = "\x9c\x75", [0x0f2e] = "\xd5\x44",
  [0x0f2f] = "\x98\x78", [0x0f30] = "\xd5\x40", [0x0f31] = "\xd8\xe7",
  [0x0f32] = "\xd8\xee", [0x0f33] = "\xd8\xe3", [0x0f34] = "\xb4\x51",
  [0x0f35] = "\xd8\xdf", [0x0f36] = "\xd8\xef", [0x0f37] = "\xd8\xd9",
  [0x0f38] = "\xd8\xec", [0x0f39] = "\xd8\xea", [0x0f3a] = "\xd8\xe4",
  [0x0f3c] = "\xd8\xed", [0x0f3d] = "\xd8\xe6", [0x0f3e] = "\x8d\x60",
  [0x0f3f] = "\xd8\xde", [0x0f40] = "\xd8\xf0", [0x0f41] = "\xd8\xdc",
  [0x0f42] = "\xd8\xe9", [0x0f43] = "\xd8\xda", [0x0f45] = "\xd8\xf1",
  [0x0f46] = "\xfb\xe5", [0x0f47] = "\xb4\x52", [0x0f48] = "\x8d\x61",
  [0x0f49] = "\xd8\xeb", [0x0f4a] = "\xdd\x4f", [0x0f4b] = "\xd8\xdd",
  [0x0f4c] = "\xb4\x4f", [0x0f4e] = "\xd8\xe1", [0x0f50] = "\xb4\x50",
  [0x0f51] = "\xd8\xe0", [0x0f52] = "\xd8\xe5", [0x0f55] = "\xd8\xe2",
  [0x0f56] = "\x8d\x62", [0x0f57] = "\xa0\xa1", [0x0f59] = "\xd8\xe8",
  [0x0f5b] = "\x9c\x40", [0x0f5e] = "\xdd\x53", [0x0f62] = "\xdd\x56",
  [0x0f63] = "\xdd\x4e", [0x0f65] = "\xdd\x50", [0x0f67] = "\xdd\x55",
  [0x0f68] = "\xdd\x54", [0x0f69] = "\xb7\x43", [0x0f6b] = "\xd8\xdb",
  [0x0f6c] = "\xdd\x52", [0x0f6f] = "\xb7\x44", [0x0f70] = "\x98\xad",
  [0x0f71] = "\xdd\x4d", [0x0f72] = "\xdd\x51", [0x0f74] = "\x9e\xea",
  [0x0f77] = "\xe1\xa9", [0x0f79] = "\xe1\xb0", [0x0f7a] = "\xe1\xa7",
  [0x0f7c] = "\xe1\xae", [0x0f7d] = "\xe1\xa5", [0x0f7e] = "\xe1\xad",
  [0x0f7f] = "\xe1\xb1", [0x0f80] = "\xe1\xa4", [0x0f81] = "\xe1\xa8",
  [0x0f82] = "\xe1\xa3", [0x0f84] = "\xb9\xf1", [0x0f85] = "\x9c\xeb",
  [0x0f86] = "\xe1\xa6", [0x0f87] = "\xb9\xf2", [0x0f88] = "\xe1\xac",
  [0x0f89] = "\xe1\xab", [0x0f8a] = "\xe1\xaa", [0x0f8b] = "\xfb\xe0",
  [0x0f8d] = "\xe1\xaf", [0x0f8e] = "\x9f\x51", [0x0f92] = "\xe5\x65",
  [0x0f93] = "\xe5\x67", [0x0f94] = "\xbc\x6b", [0x0f95] = "\xe5\x68",
  [0x0f97] = "\xe5\x63", [0x0f99] = "\xe5\x62", [0x0f9a] = "\xe5\x6c",
  [0x0f9c] = "\xe5\x6a", [0x0f9d] = "\xbc\x6a", [0x0f9e] = "\xe5\x6d",
  [0x0f9f] = "\xe5\x64", [0x0fa0] = "\xe5\x69", [0x0fa1] = "\xe5\x6b",
  [0x0fa2] = "\xe5\x66", [0x0fa4] = "\x8d\x65", [0x0fa7] = "\xe9\x61",
  [0x0fa8] = "\xe9\x66", [0x0fa9] = "\xe9\x60", [0x0faa] = "\xe9\x65",
  [0x0fab] = "\x9c\xf1", [0x0fac] = "\xe9\x5e", [0x0fad] = "\xe9\x68",
  [0x0fae] = "\xe9\x64", [0x0faf] = "\xe9\x69", [0x0fb0] = "\xe9\x63",
  [0x0fb1] = "\xe9\x5f", [0x0fb2] = "\xe9\x67", [0x0fb4] = "\xe9\x6a",
  [0x0fb5] = "\xe9\x62", [0x0fb6] = "\xfc\x58", [0x0fb7] = "\xec\xda",
  [0x0fb8] = "\xc0\xaf", [0x0fb9] = "\x8d\x66", [0x0fba] = "\xc0\xad",
  [0x0fbc] = "\xc0\xac", [0x0fbd] = "\xc0\xae", [0x0fc0] = "\xef\xc4",
  [0x0fc1] = "\x96\x54", [0x0fc2] = "\xf1\x72", [0x0fc3] = "\xf1\xfd",
  [0x0fc6] = "\xf4\x44", [0x0fc7] = "\xf4\x45", [0x0fc9] = "\xc4\x60",
  [0x0fcb] = "\xf5\xc9", [0x0fcd] = "\xc4\xde", [0x0fcf] = "\xf5\xca",
  [0x0fd1] = "\xf6\xde", [0x0fd2] = "\xc5\x72", [0x0fd4] = "\xc5\x71",
  [0x0fd5] = "\xf6\xdd", [0x0fd6] = "\xc5\xc9", [0x0fd7] = "\xfb\xe8",
  [0x0fd8] = "\xf7\xd6", [0x0fdb] = "\xc6\xcc", [0x0fdd] = "\xa4\x74",
  [0x0fde] = "\xa6\x7b", [0x0fdf] = "\xc9\xda", [0x0fe0] = "\xca\xca",
  [0x0fe1] = "\xa8\xb5", [0x0fe2] = "\xb1\x5f", [0x0fe5] = "\xa4\x75",
  [0x0fe6] = "\xa5\xaa", [0x0fe7] = "\xa5\xa9", [0x0fe8] = "\xa5\xa8",
  [0x0feb] = "\xa7\xc5", [0x0fee] = "\xae\x74", [0x0ff0] = "\xdd\x57",
  [0x0ff1] = "\xa4\x76", [0x0ff2] = "\xa4\x77", [0x0ff3] = "\xa4\x78",
  [0x0ff4] = "\xa4\xda", [0x0ff5] = "\x9f\xce", [0x0ff7] = "\xab\xd1",
  [0x0ff9] = "\xce\xaf", [0x0ffd] = "\xb4\x53", [0x0ffe] = "\xa4\x79",
  [0x0fff] = "\xc9\x5d", [0x1002] = "\xa5\xab", [0x1003] = "\xa5\xac",
  [0x1004] = "\xc9\x78", [0x1006] = "\xa6\x7c", [0x1009] = "\xfb\xfc",
  [0x100a] = "\xca\xcb", [0x100b] = "\x9a\xe4", [0x100c] = "\xa7\xc6",
  [0x100e] = "\xca\xcc", [0x1011] = "\xa9\xae", [0x1012] = "\x9f\x75",
  [0x1014] = "\xcc\x6e", [0x1015] = "\xa9\xac", [0x1016] = "\xa9\xab",
  [0x1017] = "\xcc\x6d", [0x1018] = "\xa9\xa9", [0x1019] = "\xcc\x6f",
  [0x101a] = "\xa9\xaa", [0x101b] = "\xa9\xad", [0x101d] = "\xab\xd2",
  [0x101f] = "\xab\xd4", [0x1020] = "\xce\xb3", [0x1021] = "\xce\xb0",
  [0x1022] = "\xce\xb1", [0x1023] = "\xce\xb2", [0x1024] = "\xce\xb4",
  [0x1025] = "\xab\xd3", [0x1028] = "\xd1\x74", [0x1029] = "\xd1\x73",
  [0x102b] = "\xae\x76", [0x102d] = "\xae\x75", [0x102e] = "\xfb\xf1",
  [0x1033] = "\xb1\x62", [0x1034] = "\xd5\x46", [0x1036] = "\xb1\x61",
  [0x1037] = "\xb1\x63", [0x1038] = "\xb1\x60", [0x103d] = "\xb4\x55",
  [0x103e] = "\xd5\x45", [0x1040] = "\xb4\x56", [0x1041] = "\xd8\xf3",
  [0x1042] = "\x8d\x69", [0x1043] = "\xb4\x57", [0x1044] = "\xd8\xf2",
  [0x1045] = "\xb4\x54", [0x1048] = "\x93\x4f", [0x104a] = "\xdd\x5a",
  [0x104b] = "\xdd\x5c", [0x104c] = "\xb7\x45", [0x104d] = "\xdd\x5b",
  [0x104e] = "\xdd\x59", [0x104f] = "\xdd\x58", [0x1053] = "\xe1\xb4",
  [0x1054] = "\xb9\xf7", [0x1055] = "\xb9\xf5", [0x1057] = "\xb9\xf6",
  [0x1058] = "\xe1\xb2", [0x1059] = "\xe1\xb3", [0x105b] = "\xb9\xf3",
  [0x105c] = "\xe5\x71", [0x105d] = "\xe5\x6f", [0x105e] = "\x93\x4e",
  [0x105f] = "\xbc\x6d", [0x1060] = "\xe5\x70", [0x1061] = "\xbc\x6e",
  [0x1062] = "\xbc\x6c", [0x1063] = "\xb9\xf4", [0x1066] = "\xe9\x6d",
  [0x1067] = "\xe9\x6b", [0x1068] = "\xe9\x6c", [0x1069] = "\xe5\x6e",
  [0x106a] = "\xec\xdc", [0x106b] = "\xc0\xb0", [0x106c] = "\xec\xdb",
  [0x106d] = "\xef\xc5", [0x106e] = "\xef\xc6", [0x106f] = "\xe9\x6e",
  [0x1070] = "\xf1\xfe", [0x1072] = "\xa4\x7a", [0x1073] = "\xa5\xad",
  [0x1074] = "\xa6\x7e", [0x1075] = "\xc9\xdb", [0x1076] = "\xa6\x7d",
  [0x1078] = "\xa9\xaf", [0x1079] = "\xb7\x46", [0x107a] = "\xfb\xf4",
  [0x107b] = "\xa4\xdb", [0x107c] = "\xa5\xae", [0x107d] = "\xab\xd5",
  [0x107e] = "\xb4\x58", [0x107f] = "\xc6\xce", [0x1080] = "\xc9\x79",
  [0x1082] = "\xc9\x7a", [0x1083] = "\xfb\xc3", [0x1084] = "\xc9\xdc",
  [0x1086] = "\x89\x65", [0x1087] = "\xa7\xc8", [0x1088] = "\xca\xd0",
  [0x1089] = "\xca\xce", [0x108a] = "\xa7\xc9", [0x108b] = "\xca\xcd",
  [0x108c] = "\xca\xcf", [0x108d] = "\xca\xd1", [0x108f] = "\xa7\xc7",
  [0x1095] = "\xa9\xb3", [0x1096] = "\xa9\xb4", [0x1097] = "\xa9\xb1",
  [0x109a] = "\xa9\xb0", [0x109b] = "\xce\xb8", [0x109c] = "\xa9\xb2",
  [0x10a0] = "\xab\xd6", [0x10a2] = "\xce\xb7", [0x10a3] = "\xce\xb9",
  [0x10a4] = "\xce\xb6", [0x10a5] = "\xce\xba", [0x10a6] = "\xab\xd7",
  [0x10a7] = "\xae\x79", [0x10a8] = "\xd1\x75", [0x10aa] = "\xd1\x77",
  [0x10ab] = "\xae\x77", [0x10ac] = "\xd1\x78", [0x10ad] = "\xae\x78",
  [0x10ae] = "\xd1\x76", [0x10b0] = "\xce\xb5", [0x10b1] = "\xd5\x47",
  [0x10b2] = "\xd5\x4a", [0x10b3] = "\xd5\x4b", [0x10b4] = "\xd5\x48",
  [0x10b5] = "\xb1\x67", [0x10b6] = "\xb1\x66", [0x10b7] = "\xb1\x64",
  [0x10b8] = "\xb1\x65", [0x10b9] = "\xd5\x49", [0x10bd] = "\x8d\x6a",
  [0x10be] = "\xb1\x68", [0x10c1] = "\xb4\x5a", [0x10c2] = "\xb4\x5b",
  [0x10c4] = "\xb4\x5c", [0x10c5] = "\xdd\x5d", [0x10c6] = "\xdd\x5f",
  [0x10c7] = "\xdd\x61", [0x10c8] = "\xb7\x48", [0x10c9] = "\xb7\x47",
  [0x10ca] = "\xb4\x59", [0x10cb] = "\xdd\x60", [0x10cc] = "\xdd\x5e",
  [0x10cd] = "\x93\x53", [0x10ce] = "\xe1\xb8", [0x10d0] = "\x9d\xfb",
  [0x10d1] = "\xe1\xb6", [0x10d2] = "\xe1\xbc", [0x10d3] = "\xb9\xf8",
  [0x10d4] = "\xe1\xbd", [0x10d5] = "\xe1\xba", [0x10d6] = "\xb9\xf9",
  [0x10d7] = "\xe1\xb7", [0x10d8] = "\xe1\xb5", [0x10d9] = "\xe1\xbb",
  [0x10da] = "\xbc\x70", [0x10db] = "\xe5\x73", [0x10dc] = "\xe1\xb9",
  [0x10dd] = "\xbc\x72", [0x10de] = "\xe5\x74", [0x10df] = "\xbc\x71",
  [0x10e0] = "\xbc\x74", [0x10e1] = "\xe5\x75", [0x10e2] = "\xbc\x6f",
  [0x10e3] = "\xbc\x73", [0x10e5] = "\xe9\x73", [0x10e6] = "\xe9\x71",
  [0x10e7] = "\xe9\x70", [0x10e8] = "\xe9\x72", [0x10e9] = "\xe9\x6f",
  [0x10ec] = "\xc3\x66", [0x10ee] = "\xf4\x46", [0x10ef] = "\xf4\x47",
  [0x10f1] = "\xf5\xcb", [0x10f2] = "\xf6\xdf", [0x10f3] = "\xc6\x55",
  [0x10f4] = "\xfb\xfd", [0x10f6] = "\xa9\xb5", [0x10f7] = "\xa7\xca",
  [0x10f8] = "\x90\x59", [0x10f9] = "\xfc\x40", [0x10fa] = "\xab\xd8",
  [0x10fb] = "\xfc\x41", [0x10fc] = "\xfc\x43", [0x10fe] = "\xa4\x7b",
  [0x10ff] = "\xa4\xdc", [0x1101] = "\xa5\xaf", [0x1102] = "\xc9\xdd",
  [0x1104] = "\xa7\xcb", [0x1105] = "\xca\xd2", [0x1107] = "\xce\xbb",
  [0x1108] = "\xab\xd9", [0x110a] = "\xb9\xfa", [0x110b] = "\xa4\x7c",
  [0x110c] = "\x93\x61", [0x110d] = "\xfc\x46", [0x110e] = "\x93\x62",
  [0x110f] = "\xa6\xa1", [0x1112] = "\xb7\x49", [0x1113] = "\xa4\x7d",
  [0x1114] = "\xa4\xdd", [0x1115] = "\xa4\xde", [0x1117] = "\xa5\xb1",
  [0x1118] = "\xa5\xb0", [0x111a] = "\xc9\xde", [0x111b] = "\xa6\xa2",
  [0x111d] = "\xca\xd3", [0x111f] = "\xa7\xcc", [0x1122] = "\xcc\x71",
  [0x1123] = "\xcc\x72", [0x1124] = "\xcc\x73", [0x1125] = "\x8d\x6b",
  [0x1126] = "\xa9\xb6", [0x1127] = "\xa9\xb7", [0x1128] = "\xcc\x70",
  [0x1129] = "\xa9\xb8", [0x112d] = "\xab\xda", [0x112e] = "\xce\xbc",
  [0x1130] = "\xd1\x7a", [0x1131] = "\xae\x7a", [0x1133] = "\xd1\x79",
  [0x1135] = "\xb1\x69", [0x1136] = "\xd5\x4c", [0x1137] = "\xb1\x6a",
  [0x1138] = "\xd5\x4d", [0x113a] = "\xfc\x4c", [0x113c] = "\xb4\x5d",
  [0x1140] = "\xdd\x62", [0x1143] = "\xe1\xbf", [0x1144] = "\xe1\xbe",
  [0x1146] = "\xb9\xfb", [0x1148] = "\xbc\x75", [0x1149] = "\xe5\x76",
  [0x114a] = "\xbe\xca", [0x114b] = "\xe9\x74", [0x114c] = "\xc0\xb1",
  [0x114d] = "\x95\xb8", [0x114e] = "\xc5\x73", [0x114f] = "\xf7\xd8",
  [0x1150] = "\xc6\xd0", [0x1151] = "\x8b\xca", [0x1154] = "\xcc\x74",
  [0x1156] = "\xce\xbd", [0x1157] = "\xb1\x6b", [0x1158] = "\xd8\xf4",
  [0x1159] = "\xb7\x4a", [0x115c] = "\x98\x7a", [0x115d] = "\xc2\x55",
  [0x1161] = "\xc6\xd1", [0x1162] = "\xa7\xce", [0x1163] = "\xfc\x51",
  [0x1164] = "\xa7\xcd", [0x1165] = "\xab\xdb", [0x1167] = "\xd1\x7b",
  [0x1169] = "\xb1\x6d", [0x116a] = "\xb3\x43", [0x116b] = "\xb1\x6e",
  [0x116c] = "\xb1\x6c", [0x116d] = "\xb4\x5e", [0x116f] = "\xe1\xc0",
  [0x1170] = "\xb9\xfc", [0x1171] = "\xbc\x76", [0x1172] = "\xfc\x54",
  [0x1173] = "\xc9\x4c", [0x1174] = "\xc9\xdf", [0x1176] = "\xca\xd5",
  [0x1177] = "\xa7\xcf", [0x1178] = "\xca\xd4", [0x1179] = "\xa7\xd0",
  [0x117b] = "\xfa\xaf", [0x117c] = "\xa9\xbc", [0x117d] = "\xcc\x77",
  [0x117e] = "\xcc\x76", [0x117f] = "\xa9\xbb", [0x1180] = "\xa9\xb9",
  [0x1181] = "\xa9\xba", [0x1182] = "\xcc\x75", [0x1183] = "\x8d\x6c",
  [0x1185] = "\xab\xdd", [0x1186] = "\xce\xbe", [0x1187] = "\xab\xe0",
  [0x1188] = "\xab\xdc", [0x1189] = "\xab\xe2", [0x118a] = "\xab\xde",
  [0x118b] = "\xab\xdf", [0x118c] = "\xab\xe1", [0x1190] = "\xae\x7d",
  [0x1191] = "\xae\x7c", [0x1192] = "\xae\x7b", [0x1196] = "\xd5\x4f",
  [0x1197] = "\xb1\x6f", [0x1198] = "\xb1\x72", [0x1199] = "\xb1\x70",
  [0x119b] = "\xd5\x4e", [0x119c] = "\xb1\x75", [0x119e] = "\xb1\x71",
  [0x119f] = "\xd5\x50", [0x11a0] = "\xb1\x74", [0x11a1] = "\xb1\x73",
  [0x11a4] = "\xfa\x61", [0x11a5] = "\xd8\xf6", [0x11a6] = "\xd8\xf5",
  [0x11a7] = "\xfc\x57", [0x11a8] = "\xb4\x61", [0x11a9] = "\xb4\x5f",
  [0x11aa] = "\xb4\x60", [0x11ab] = "\xd8\xf7", [0x11ac] = "\xb7\x4b",
  [0x11ad] = "\xdd\x64", [0x11ae] = "\xb7\x4c", [0x11af] = "\xdd\x63",
  [0x11b1] = "\x9b\x70", [0x11b2] = "\xe5\x77", [0x11b5] = "\xbc\x78",
  [0x11b6] = "\xe1\xc1", [0x11b7] = "\xbc\x77", [0x11b9] = "\xb9\xfd",
  [0x11ba] = "\xa0\x51", [0x11bb] = "\xec\xde", [0x11bc] = "\xe9\x75",
  [0x11bd] = "\xc0\xb2", [0x11be] = "\xec\xdd", [0x11bf] = "\xf2\x40",
  [0x11c0] = "\xf4\x48", [0x11c1] = "\xf4\x49", [0x11c3] = "\xa4\xdf",
  [0x11c4] = "\x8b\xcb", [0x11c5] = "\xa5\xb2", [0x11c9] = "\xc9\x7b",
  [0x11cc] = "\xa7\xd2", [0x11cd] = "\xa7\xd4", [0x11cf] = "\xc9\xe2",
  [0x11d0] = "\xca\xd8", [0x11d1] = "\xca\xd7", [0x11d2] = "\xca\xd6",
  [0x11d4] = "\xc9\xe1", [0x11d5] = "\xc9\xe0", [0x11d6] = "\xa6\xa4",
  [0x11d7] = "\xa7\xd3", [0x11d8] = "\xa7\xd1", [0x11d9] = "\xa6\xa3",
  [0x11db] = "\x93\x6e", [0x11dd] = "\xa9\xbd", [0x11de] = "\xcc\x78",
  [0x11df] = "\xfc\xd5", [0x11e0] = "\xa9\xbe", [0x11e1] = "\xca\xdd",
  [0x11e3] = "\xca\xdf", [0x11e4] = "\xca\xde", [0x11e5] = "\xcc\x79",
  [0x11e8] = "\xca\xda", [0x11ea] = "\xa7\xd8", [0x11eb] = "\xa7\xd6",
  [0x11ed] = "\xca\xd9", [0x11ee] = "\xca\xdb", [0x11ef] = "\xca\xe1",
  [0x11f1] = "\xa7\xd5", [0x11f3] = "\xca\xdc", [0x11f4] = "\xca\xe5",
  [0x11f5] = "\xa9\xc0", [0x11f7] = "\xca\xe2", [0x11f8] = "\xa7\xd7",
  [0x11fa] = "\xca\xe0", [0x11fb] = "\xca\xe3", [0x11fd] = "\xa9\xbf",
  [0x11ff] = "\xa9\xc1", [0x1200] = "\xca\xe4", [0x1209] = "\xcc\xaf",
  [0x120a] = "\xcc\xa2", [0x120b] = "\xcc\x7e", [0x120c] = "\xcc\xae",
  [0x120d] = "\xcc\xa9", [0x120e] = "\xab\xe7", [0x120f] = "\xa9\xc2",
  [0x1210] = "\xcc\xaa", [0x1211] = "\xcc\xad", [0x1212] = "\xab\xe3",
  [0x1213] = "\xcc\xac", [0x1214] = "\xa9\xc3", [0x1215] = "\xa9\xc8",
  [0x1216] = "\xa9\xc6", [0x1217] = "\xcc\xa3", [0x1219] = "\xcc\x7c",
  [0x121a] = "\xcc\xa5", [0x121b] = "\xa9\xcd", [0x121c] = "\xcc\xb0",
  [0x121d] = "\xab\xe4", [0x121e] = "\xcc\xa6", [0x1220] = "\xab\xe5",
  [0x1221] = "\xa9\xc9", [0x1222] = "\xcc\xa8", [0x1223] = "\xfc\xa9",
  [0x1224] = "\xce\xcd", [0x1225] = "\xab\xe6", [0x1226] = "\xcc\x7b",
  [0x1227] = "\xa9\xca", [0x1228] = "\xab\xe8", [0x1229] = "\xa9\xcb",
  [0x122a] = "\xa9\xc7", [0x122b] = "\xa9\xcc", [0x122c] = "\xcc\xa7",
  [0x122d] = "\xcc\x7a", [0x122e] = "\xcc\xab", [0x122f] = "\xa9\xc4",
  [0x1231] = "\xfc\x61", [0x1232] = "\xcc\x7d", [0x1233] = "\xcc\xa4",
  [0x1234] = "\xcc\xa1", [0x1235] = "\xa9\xc5", [0x1237] = "\xce\xbf",
  [0x1239] = "\xce\xc0", [0x123b] = "\x89\x66", [0x1240] = "\xce\xca",
  [0x1241] = "\xd1\xa1", [0x1242] = "\xce\xcb", [0x1243] = "\xab\xee",
  [0x1244] = "\xce\xce", [0x1245] = "\xce\xc4", [0x1246] = "\xab\xed",
  [0x1247] = "\xce\xc6", [0x1249] = "\xce\xc7", [0x124a] = "\xfa\xcb",
  [0x124c] = "\xce\xc9", [0x124d] = "\xab\xe9", [0x1250] = "\xae\xa3",
  [0x1252] = "\xf9\xda", [0x1253] = "\xce\xc5", [0x1254] = "\xce\xc1",
  [0x1255] = "\xae\xa4", [0x1258] = "\xce\xcf", [0x1259] = "\xae\x7e",
  [0x125a] = "\xd1\x7d", [0x125b] = "\xce\xc8", [0x125d] = "\xd1\x7c",
  [0x125e] = "\xce\xc3", [0x125f] = "\xce\xcc", [0x1262] = "\xab\xec",
  [0x1263] = "\xae\xa1", [0x1264] = "\xab\xf2", [0x1265] = "\xae\xa2",
  [0x1266] = "\xce\xd0", [0x1267] = "\xd1\x7e", [0x1268] = "\xab\xeb",
  [0x1269] = "\xae\xa6", [0x126a] = "\xab\xf1", [0x126b] = "\xab\xf0",
  [0x126c] = "\xab\xef", [0x126d] = "\xae\xa5", [0x126e] = "\xce\xd1",
  [0x126f] = "\xae\xa7", [0x1270] = "\xab\xea", [0x1272] = "\xce\xc2",
  [0x1275] = "\x93\x7a", [0x1277] = "\xa0\xe0", [0x127e] = "\x93\x6b",
  [0x127f] = "\xb1\x76", [0x1280] = "\xd1\xa4", [0x1281] = "\xd1\xa6",
  [0x1283] = "\xd1\xa8", [0x1284] = "\xae\xa8", [0x1285] = "\xae\xae",
  [0x1286] = "\xd5\x53", [0x1287] = "\xd1\xac", [0x1288] = "\xd1\xa3",
  [0x1289] = "\xb1\x78", [0x128a] = "\xd5\x51", [0x128c] = "\xae\xad",
  [0x128d] = "\xae\xab", [0x128e] = "\xd1\xae", [0x1290] = "\xd5\x52",
  [0x1292] = "\xd1\xa5", [0x1294] = "\xae\xac", [0x1295] = "\xd1\xa9",
  [0x1296] = "\xae\xaf", [0x1297] = "\xd1\xab", [0x129a] = "\xae\xaa",
  [0x129b] = "\xd1\xaa", [0x129c] = "\xd1\xad", [0x129d] = "\xd1\xa7",
  [0x129e] = "\xfc\x6b", [0x129f] = "\xae\xa9", [0x12a0] = "\xb1\x79",
  [0x12a2] = "\xd1\xa2", [0x12a3] = "\xb1\x77", [0x12a4] = "\xa0\xdc",
  [0x12a7] = "\x94\x68", [0x12a8] = "\xb1\x7a", [0x12b0] = "\xd5\x55",
  [0x12b1] = "\xd5\x5e", [0x12b2] = "\xb4\x64", [0x12b3] = "\xfc\x6d",
  [0x12b4] = "\xb1\x7c", [0x12b5] = "\xb1\xa3", [0x12b6] = "\xb4\x65",
  [0x12b7] = "\xd5\x60", [0x12b8] = "\xb1\xaa", [0x12b9] = "\xd8\xf9",
  [0x12ba] = "\xd5\x56", [0x12bb] = "\xb1\xa2", [0x12bc] = "\xb1\xa5",
  [0x12bd] = "\xb1\x7e", [0x12be] = "\xd5\x54", [0x12bf] = "\xd5\x62",
  [0x12c0] = "\xd5\x65", [0x12c1] = "\xd9\x49", [0x12c3] = "\xd5\x63",
  [0x12c4] = "\xd8\xfd", [0x12c5] = "\xb1\xa1", [0x12c6] = "\xb1\xa8",
  [0x12c7] = "\xb1\xac", [0x12c8] = "\xd5\x5d", [0x12c9] = "\xd8\xf8",
  [0x12ca] = "\xd5\x61", [0x12cb] = "\xb1\x7b", [0x12cc] = "\xd8\xfa",
  [0x12cd] = "\xd5\x64", [0x12ce] = "\xd8\xfc", [0x12cf] = "\xd5\x59",
  [0x12d1] = "\xb4\x62", [0x12d3] = "\xd5\x57", [0x12d4] = "\xd5\x58",
  [0x12d5] = "\xb1\xa7", [0x12d7] = "\x8d\x71", [0x12d8] = "\xb1\xa6",
  [0x12d9] = "\xd5\x5b", [0x12da] = "\xb1\xab", [0x12db] = "\xd5\x5f",
  [0x12dc] = "\xb1\xa4", [0x12dd] = "\xd5\x5c", [0x12de] = "\xfd\x64",
  [0x12df] = "\xb1\xa9", [0x12e0] = "\xb4\x66", [0x12e1] = "\xb4\x63",
  [0x12e2] = "\xd8\xfb", [0x12e3] = "\x99\xba", [0x12e4] = "\xd5\x5a",
  [0x12e6] = "\xb1\x7d", [0x12e7] = "\x9a\xd0", [0x12e8] = "\x9a\x61",
  [0x12e9] = "\xa0\xe5", [0x12f0] = "\xb4\x6b", [0x12f1] = "\xb4\x6f",
  [0x12f2] = "\xd9\x40", [0x12f3] = "\xb7\x51", [0x12f4] = "\xb4\x6d",
  [0x12f5] = "\xd9\x44", [0x12f6] = "\xb4\x71", [0x12f7] = "\xdd\x65",
  [0x12f8] = "\xd9\x46", [0x12f9] = "\xb7\x53", [0x12fa] = "\xb4\x69",
  [0x12fb] = "\xb4\x6c", [0x12fc] = "\xd9\x47", [0x12fd] = "\xa0\x5b",
  [0x12fe] = "\xd9\x48", [0x12ff] = "\xd9\x4e", [0x1300] = "\xb4\x73",
  [0x1301] = "\xb7\x54", [0x1303] = "\xd9\x4a", [0x1304] = "\xd9\x4f",
  [0x1305] = "\xd9\x43", [0x1306] = "\xb7\x5e", [0x1307] = "\x96\xac",
  [0x1308] = "\xb7\x55", [0x1309] = "\xb4\x72", [0x130a] = "\xd9\x41",
  [0x130b] = "\xd9\x50", [0x130c] = "\x97\x40", [0x130d] = "\xb7\x5d",
  [0x130e] = "\xb4\x70", [0x130f] = "\xb7\x4e", [0x1310] = "\xd9\x4d",
  [0x1312] = "\xb4\x74", [0x1313] = "\xd9\x45", [0x1314] = "\xd8\xfe",
  [0x1315] = "\xb4\x6a", [0x1316] = "\xd9\x42", [0x1318] = "\xd9\x4b",
  [0x1319] = "\x9e\xf1", [0x131a] = "\xb7\x4d", [0x131b] = "\xb7\x52",
  [0x131c] = "\xb4\x67", [0x131d] = "\xd9\x4c", [0x131f] = "\xb7\x50",
  [0x1323] = "\xb4\x68", [0x1327] = "\xb7\x5c", [0x1328] = "\xe1\xc3",
  [0x1329] = "\xdd\x70", [0x132b] = "\xdd\x68", [0x132c] = "\xe1\xc2",
  [0x132e] = "\xdd\x6c", [0x132f] = "\xdd\x6e", [0x1330] = "\x9f\x7e",
  [0x1332] = "\xdd\x6b", [0x1334] = "\xb7\x5b", [0x1336] = "\xdd\x6a",
  [0x1337] = "\xb7\x5f", [0x133b] = "\xe1\xd2", [0x133d] = "\x8d\x72",
  [0x133e] = "\xb7\x5a", [0x133f] = "\xba\x40", [0x1340] = "\xdd\x71",
  [0x1341] = "\xe1\xc4", [0x1342] = "\xfc\x76", [0x1344] = "\xb7\x58",
  [0x1345] = "\xdd\x69", [0x1346] = "\xdd\x6d", [0x1347] = "\xb9\xfe",
  [0x1348] = "\xb7\x4f", [0x1349] = "\xdd\x66", [0x134a] = "\xdd\x67",
  [0x134b] = "\xba\x41", [0x134c] = "\xb7\x57", [0x134d] = "\xb7\x59",
  [0x134e] = "\xb7\x56", [0x134f] = "\xdd\x6f", [0x1350] = "\x96\xa9",
  [0x1352] = "\xe1\xc8", [0x1353] = "\xe1\xc9", [0x1354] = "\xe1\xce",
  [0x1355] = "\xbc\x7d", [0x1356] = "\xe1\xd5", [0x1358] = "\xba\x47",
  [0x1359] = "\xa0\x6e", [0x135a] = "\xba\x46", [0x135b] = "\xe1\xd0",
  [0x135c] = "\xfc\xaa", [0x135d] = "\xbc\x7c", [0x135e] = "\xe1\xc5",
  [0x135f] = "\xba\x45", [0x1360] = "\xfb\xcd", [0x1361] = "\xe1\xd4",
  [0x1362] = "\xba\x43", [0x1363] = "\xba\x44", [0x1364] = "\xfc\x74",
  [0x1365] = "\xe1\xd1", [0x1366] = "\xe5\xaa", [0x1367] = "\xbc\x7a",
  [0x1368] = "\xb4\x6e", [0x136a] = "\xe1\xd3", [0x136b] = "\xbc\xa3",
  [0x136c] = "\xe1\xcb", [0x136e] = "\xbc\x7b", [0x136f] = "\xa0\x74",
  [0x1370] = "\xbc\xa2", [0x1371] = "\xe1\xc6", [0x1372] = "\xe1\xca",
  [0x1373] = "\xe1\xc7", [0x1374] = "\xe1\xcd", [0x1375] = "\xba\x48",
  [0x1376] = "\xbc\x79", [0x1377] = "\xba\x42", [0x1379] = "\xe5\x7a",
  [0x137a] = "\xe1\xcf", [0x137c] = "\xbc\xa1", [0x137d] = "\xa0\x71",
  [0x137e] = "\xbc\xa4", [0x1380] = "\xe1\xcc", [0x1381] = "\xfc\x79",
  [0x1382] = "\xbc\x7e", [0x1383] = "\xe5\x79", [0x1387] = "\xfc\x7c",
  [0x1389] = "\xe5\x7e", [0x138a] = "\xbe\xce", [0x138b] = "\xe5\x78",
  [0x138c] = "\xe9\xa3", [0x138d] = "\xe5\xa9", [0x138e] = "\xbc\xa8",
  [0x1390] = "\xbc\xa6", [0x1391] = "\xbe\xcc", [0x1392] = "\xe5\xa6",
  [0x1393] = "\xe5\xa2", [0x1394] = "\xbc\xac", [0x1395] = "\x9c\x50",
  [0x1396] = "\xe9\x78", [0x1398] = "\x93\x79", [0x1399] = "\x93\x78",
  [0x139a] = "\xbc\xaa", [0x139b] = "\xe5\xa1", [0x139c] = "\xa0\xdd",
  [0x139d] = "\xe9\x76", [0x139f] = "\xe5\xa5", [0x13a1] = "\xe5\xa8",
  [0x13a2] = "\xe5\x7d", [0x13a4] = "\xbc\xab", [0x13a7] = "\xbc\xa5",
  [0x13a8] = "\xe9\x77", [0x13a9] = "\xbe\xcd", [0x13aa] = "\xe5\xa7",
  [0x13ab] = "\xbc\xa7", [0x13ac] = "\xbc\xa9", [0x13ad] = "\xe5\xa4",
  [0x13ae] = "\xbc\xad", [0x13af] = "\xe5\xa3", [0x13b0] = "\xe5\x7c",
  [0x13b1] = "\xe5\x7b", [0x13b2] = "\xbe\xcb", [0x13b3] = "\xe5\xab",
  [0x13b4] = "\xe9\x7a", [0x13b5] = "\xec\xe0", [0x13b6] = "\xbe\xd0",
  [0x13b7] = "\x8d\x75", [0x13b8] = "\xe9\xa2", [0x13b9] = "\x8d\x76",
  [0x13ba] = "\xe9\x7e", [0x13bc] = "\xec\xe1", [0x13be] = "\xbe\xd1",
  [0x13bf] = "\xe9\xa1", [0x13c0] = "\x93\x74", [0x13c1] = "\xe9\x7c",
  [0x13c2] = "\xc0\xb4", [0x13c3] = "\xec\xdf", [0x13c5] = "\xe9\x79",
  [0x13c6] = "\xe9\x7b", [0x13c7] = "\xc0\xb5", [0x13c8] = "\xbe\xd3",
  [0x13c9] = "\xc0\xb3", [0x13ca] = "\xbe\xd2", [0x13cb] = "\xc0\xb7",
  [0x13cc] = "\xe9\x7d", [0x13cd] = "\xbe\xcf", [0x13cf] = "\x8d\x77",
  [0x13d0] = "\xfc\xa5", [0x13d3] = "\xfc\xa2", [0x13d6] = "\xef\xcf",
  [0x13d8] = "\xef\xc7", [0x13da] = "\x90\xc3", [0x13de] = "\xec\xe7",
  [0x13df] = "\xef\xc8", [0x13e0] = "\xec\xe3", [0x13e2] = "\xa0\x79",
  [0x13e3] = "\xc2\x56", [0x13e4] = "\xec\xe5", [0x13e5] = "\xec\xe4",
  [0x13e6] = "\xc0\xb6", [0x13e7] = "\xec\xe2", [0x13e8] = "\xec\xe6",
  [0x13e9] = "\xef\xd0", [0x13ea] = "\xef\xcc", [0x13eb] = "\xef\xce",
  [0x13ed] = "\xef\xc9", [0x13ee] = "\xef\xca", [0x13f0] = "\xef\xcd",
  [0x13f1] = "\xef\xcb", [0x13f2] = "\xc3\x67", [0x13f5] = "\xc3\x6a",
  [0x13f6] = "\xc3\x69", [0x13f7] = "\xc3\x68", [0x13f8] = "\xc4\x61",
  [0x13f9] = "\xf4\x4a", [0x13fa] = "\xc4\x62", [0x13fb] = "\xf2\x41",
  [0x13fc] = "\xc4\xdf", [0x13fd] = "\xf5\xcc", [0x13fe] = "\xc4\xe0",
  [0x13ff] = "\xc5\x74", [0x1400] = "\xc5\xca", [0x1401] = "\xf7\xd9",
  [0x1403] = "\xf7\xda", [0x1404] = "\xf7\xdb", [0x1407] = "\xf9\xba",
  [0x1408] = "\xa4\xe0", [0x1409] = "\xc9\x7c", [0x140a] = "\xa5\xb3",
  [0x140c] = "\xa6\xa6", [0x140d] = "\xa6\xa7", [0x140e] = "\xa6\xa5",
  [0x1410] = "\xa6\xa8", [0x1411] = "\xa7\xda", [0x1412] = "\xa7\xd9",
  [0x1414] = "\xcc\xb1", [0x1415] = "\xa9\xcf", [0x1416] = "\xa9\xce",
  [0x1419] = "\xd1\xaf", [0x141a] = "\xb1\xad", [0x141b] = "\xb1\xae",
  [0x141f] = "\xb4\x75", [0x1420] = "\xdd\x72", [0x1421] = "\xb7\x60",
  [0x1422] = "\xb7\x61", [0x1423] = "\xdd\x74", [0x1424] = "\xdd\x76",
  [0x1425] = "\xdd\x75", [0x1427] = "\xe1\xd7", [0x1429] = "\xe1\xd6",
  [0x142a] = "\xba\x49", [0x142b] = "\xe1\xd8", [0x142c] = "\x8d\x79",
  [0x142d] = "\xe5\xac", [0x142e] = "\xbc\xae", [0x1430] = "\xbe\xd4",
  [0x1432] = "\xc0\xb8", [0x1433] = "\xc2\x57", [0x1434] = "\xc0\xb9",
  [0x1436] = "\xa4\xe1", [0x1437] = "\x8b\xfc", [0x1439] = "\xa0\x76",
  [0x143a] = "\xca\xe6", [0x143d] = "\xcc\xb2", [0x143e] = "\xa9\xd1",
  [0x143f] = "\xa9\xd0", [0x1440] = "\xa9\xd2", [0x1441] = "\xab\xf3",
  [0x1442] = "\xce\xd2", [0x1443] = "\xce\xd3", [0x1446] = "\xd1\xb0",
  [0x1447] = "\xae\xb0", [0x1448] = "\xb1\xaf", [0x1449] = "\xb4\x76",
  [0x144a] = "\xd9\x51", [0x144b] = "\xa4\xe2", [0x144c] = "\x8b\xcd",
  [0x144d] = "\xa4\x7e", [0x144e] = "\xa4\xe3", [0x1450] = "\xc9\x7d",
  [0x1451] = "\xa5\xb7", [0x1452] = "\xa5\xb6", [0x1453] = "\xa5\xb4",
  [0x1454] = "\xa5\xb5", [0x1458] = "\xa6\xab", [0x1459] = "\xc9\xe9",
  [0x145a] = "\xc9\xeb", [0x145b] = "\xa6\xaa", [0x145c] = "\xc9\xe3",
  [0x145e] = "\xc9\xe4", [0x1460] = "\xc9\xea", [0x1461] = "\xc9\xe6",
  [0x1462] = "\xc9\xe8", [0x1463] = "\xa6\xa9", [0x1464] = "\xc9\xe5",
  [0x1465] = "\xc9\xec", [0x1466] = "\xc9\xe7", [0x1468] = "\x9f\x5a",
  [0x146d] = "\xa7\xe1", [0x146e] = "\xa7\xea", [0x146f] = "\xa7\xe8",
  [0x1470] = "\xca\xf0", [0x1471] = "\xca\xed", [0x1472] = "\xca\xf5",
  [0x1473] = "\xa7\xe6", [0x1474] = "\xca\xf6", [0x1476] = "\xa7\xdf",
  [0x1477] = "\xca\xf3", [0x1479] = "\xa7\xe5", [0x147a] = "\xca\xef",
  [0x147b] = "\xca\xee", [0x147c] = "\xa7\xe3", [0x147d] = "\xca\xf4",
  [0x147e] = "\xa7\xe4", [0x147f] = "\xa9\xd3", [0x1480] = "\xa7\xde",
  [0x1481] = "\xca\xf1", [0x1482] = "\x9f\xf4", [0x1483] = "\xca\xe7",
  [0x1484] = "\xa7\xdb", [0x1485] = "\x9f\xba", [0x1486] = "\xa7\xee",
  [0x1487] = "\xca\xec", [0x1488] = "\xca\xf2", [0x1489] = "\xa7\xe0",
  [0x148a] = "\xa7\xe2", [0x148c] = "\xca\xe8", [0x148e] = "\xca\xe9",
  [0x148f] = "\xca\xea", [0x1490] = "\x8d\x7a", [0x1491] = "\xa7\xed",
  [0x1492] = "\xa7\xe7", [0x1493] = "\xa7\xec", [0x1494] = "\xca\xeb",
  [0x1495] = "\xa7\xeb", [0x1496] = "\xa7\xdd", [0x1497] = "\xa7\xdc",
  [0x1498] = "\xa7\xe9", [0x149d] = "\x9e\x45", [0x14a4] = "\x93\xb0",
  [0x14a6] = "\xa0\x75", [0x14a8] = "\xa9\xe1", [0x14a9] = "\xcc\xbe",
  [0x14aa] = "\xcc\xb7", [0x14ab] = "\xa9\xdc", [0x14ac] = "\xa9\xef",
  [0x14ad] = "\xcc\xb3", [0x14ae] = "\xcc\xba", [0x14af] = "\xcc\xbc",
  [0x14b0] = "\xcc\xbf", [0x14b1] = "\xa9\xea", [0x14b3] = "\xcc\xbb",
  [0x14b4] = "\xcc\xb4", [0x14b5] = "\xa9\xe8", [0x14b6] = "\xcc\xb8",
  [0x14b8] = "\xcc\xc0", [0x14b9] = "\xa9\xd9", [0x14bb] = "\xcc\xbd",
  [0x14bc] = "\xa9\xe3", [0x14bd] = "\xa9\xe2", [0x14be] = "\xcc\xb6",
  [0x14bf] = "\xa9\xd7", [0x14c2] = "\xa9\xd8", [0x14c3] = "\x9b\x46",
  [0x14c4] = "\xa9\xd6", [0x14c5] = "\xfc\xae", [0x14c6] = "\xa9\xee",
  [0x14c7] = "\xa9\xe6", [0x14c8] = "\xa9\xe0", [0x14c9] = "\xa9\xd4",
  [0x14ca] = "\xcc\xb9", [0x14cb] = "\xa9\xdf", [0x14cc] = "\xa9\xd5",
  [0x14cd] = "\xa9\xe7", [0x14ce] = "\xa9\xf0", [0x14cf] = "\xce\xd4",
  [0x14d0] = "\xa9\xe4", [0x14d1] = "\xcc\xb5", [0x14d2] = "\xa9\xda",
  [0x14d3] = "\xa9\xdd", [0x14d4] = "\xa9\xde", [0x14d5] = "\xfc\xb0",
  [0x14d6] = "\xa9\xec", [0x14d7] = "\xa9\xed", [0x14d8] = "\xa9\xeb",
  [0x14d9] = "\xa9\xe5", [0x14da] = "\xa9\xe9", [0x14db] = "\xa9\xdb",
  [0x14dc] = "\xab\xf4", [0x14df] = "\xfa\x51", [0x14e5] = "\x8d\x7b",
  [0x14eb] = "\xce\xda", [0x14ec] = "\xac\x41", [0x14ed] = "\xab\xf8",
  [0x14ee] = "\xab\xfa", [0x14ef] = "\xac\x40", [0x14f0] = "\xce\xe6",
  [0x14f1] = "\xab\xfd", [0x14f2] = "\xd1\xb1", [0x14f3] = "\xae\xb1",
  [0x14f4] = "\xac\x43", [0x14f5] = "\xce\xd7", [0x14f6] = "\xce\xdf",
  [0x14f7] = "\xab\xfe", [0x14f8] = "\xce\xde", [0x14f9] = "\xce\xdb",
  [0x14fa] = "\xce\xe3", [0x14fb] = "\xce\xe5", [0x14fc] = "\xab\xf7",
  [0x14fd] = "\xab\xfb", [0x14fe] = "\xac\x42", [0x14ff] = "\xae\xb3",
  [0x1500] = "\xce\xe0", [0x1501] = "\xab\xf9", [0x1502] = "\xac\x45",
  [0x1503] = "\xce\xd9", [0x1507] = "\xab\xfc", [0x1508] = "\xae\xb2",
  [0x1509] = "\xab\xf6", [0x150b] = "\xce\xd6", [0x150c] = "\xce\xdd",
  [0x150d] = "\xce\xd5", [0x150e] = "\xce\xd8", [0x150f] = "\xce\xdc",
  [0x1510] = "\xd1\xb2", [0x1511] = "\xac\x44", [0x1513] = "\xce\xe1",
  [0x1514] = "\xce\xe2", [0x1515] = "\xce\xe4", [0x1516] = "\xab\xf5",
  [0x1518] = "\x8d\x7c", [0x1528] = "\xae\xc1", [0x1529] = "\xd1\xbe",
  [0x152a] = "\xae\xbf", [0x152b] = "\xae\xc0", [0x152c] = "\xd1\xb4",
  [0x152d] = "\xd1\xc4", [0x152e] = "\x9e\xd6", [0x152f] = "\xae\xb6",
  [0x1531] = "\x93\xac", [0x1532] = "\xd5\x66", [0x1533] = "\xd1\xc6",
  [0x1534] = "\xd1\xc0", [0x1535] = "\x9f\x5b", [0x1536] = "\xd1\xb7",
  [0x1537] = "\x93\xa9", [0x1538] = "\xd1\xc9", [0x1539] = "\xd1\xba",
  [0x153a] = "\xae\xbc", [0x153b] = "\xd5\x7d", [0x153c] = "\xd1\xbd",
  [0x153d] = "\xae\xbe", [0x153e] = "\xae\xb5", [0x1540] = "\xd1\xcb",
  [0x1541] = "\xd1\xbf", [0x1542] = "\xae\xb8", [0x1543] = "\xd1\xb8",
  [0x1544] = "\xd1\xb5", [0x1545] = "\xd1\xb6", [0x1546] = "\xae\xb9",
  [0x1547] = "\xd1\xc5", [0x1548] = "\xd1\xcc", [0x1549] = "\xae\xbb",
  [0x154a] = "\xd1\xbc", [0x154b] = "\xd1\xbb", [0x154c] = "\xae\xc3",
  [0x154d] = "\xae\xc2", [0x154e] = "\xae\xb4", [0x154f] = "\xae\xba",
  [0x1550] = "\xae\xbd", [0x1551] = "\xd1\xc8", [0x1554] = "\xd1\xc2",
  [0x1555] = "\xae\xb7", [0x1556] = "\xd1\xb3", [0x1557] = "\xd1\xca",
  [0x1558] = "\xd1\xc1", [0x1559] = "\xd1\xc3", [0x155a] = "\xd1\xc7",
  [0x1564] = "\xa0\x7c", [0x1565] = "\xd5\x67", [0x1567] = "\xb1\xb7",
  [0x1568] = "\xb1\xcb", [0x1569] = "\xb1\xca", [0x156b] = "\xb1\xbf",
  [0x156c] = "\xfc\xb2", [0x156d] = "\xd5\x79", [0x156e] = "\xd5\x75",
  [0x156f] = "\xd5\x72", [0x1570] = "\xd5\xa6", [0x1571] = "\xb1\xba",
  [0x1572] = "\xb1\xb2", [0x1575] = "\xd5\x77", [0x1576] = "\xb4\xa8",
  [0x1577] = "\xb1\xb6", [0x1578] = "\xd5\xa1", [0x1579] = "\x8a\xc1",
  [0x157a] = "\xb1\xcc", [0x157b] = "\xb1\xc9", [0x157c] = "\xd5\x7b",
  [0x157d] = "\xd5\x6a", [0x157f] = "\x9f\xb4", [0x1580] = "\xb1\xc8",
  [0x1581] = "\xd5\xa3", [0x1582] = "\xd5\x69", [0x1583] = "\xb1\xbd",
  [0x1584] = "\xb1\xc1", [0x1585] = "\xd5\xa2", [0x1587] = "\xd5\x73",
  [0x1588] = "\xb1\xc2", [0x1589] = "\xb1\xbc", [0x158a] = "\xd5\x68",
  [0x158b] = "\xfc\xac", [0x158c] = "\xb4\x78", [0x158d] = "\xd5\xa5",
  [0x158e] = "\xd5\x71", [0x158f] = "\xb1\xc7", [0x1590] = "\xd5\x74",
  [0x1591] = "\xd5\xa4", [0x1592] = "\xb1\xc6", [0x1594] = "\xd9\x52",
  [0x1596] = "\xb1\xb3", [0x1597] = "\xd5\x6f", [0x1598] = "\xb1\xb8",
  [0x1599] = "\xb1\xc3", [0x159b] = "\xb1\xbe", [0x159c] = "\xd5\x78",
  [0x159d] = "\xd5\x6e", [0x159e] = "\xd5\x6c", [0x159f] = "\xd5\x7e",
  [0x15a0] = "\xb1\xb0", [0x15a1] = "\xb1\xc4", [0x15a2] = "\xb1\xb4",
  [0x15a3] = "\xb4\x77", [0x15a4] = "\xd5\x7c", [0x15a5] = "\xb1\xb5",
  [0x15a7] = "\xb1\xb1", [0x15a8] = "\xb1\xc0", [0x15a9] = "\xb1\xbb",
  [0x15aa] = "\xb1\xb9", [0x15ab] = "\xd5\x70", [0x15ac] = "\xb1\xc5",
  [0x15ad] = "\xd5\x6d", [0x15ae] = "\xd5\x7a", [0x15af] = "\xd5\x76",
  [0x15b0] = "\xd9\x54", [0x15b1] = "\xd9\x53", [0x15b9] = "\x9e\x4c",
  [0x15bd] = "\xd5\x6b", [0x15be] = "\xd9\x64", [0x15c0] = "\xb4\x7a",
  [0x15c1] = "\x8f\xc5", [0x15c2] = "\xd9\x6a", [0x15c3] = "\xd9\x59",
  [0x15c4] = "\xd9\x67", [0x15c5] = "\xdd\x77", [0x15c6] = "\xb4\x7d",
  [0x15c7] = "\xd9\x6b", [0x15c8] = "\xd9\x6e", [0x15c9] = "\xb4\x7c",
  [0x15ca] = "\xd9\x5c", [0x15cb] = "\xd9\x6d", [0x15cc] = "\xd9\x6c",
  [0x15cd] = "\xb4\x7e", [0x15ce] = "\xd9\x55", [0x15cf] = "\xb4\x79",
  [0x15d0] = "\xb4\xa3", [0x15d1] = "\x93\xad", [0x15d2] = "\xb4\xa1",
  [0x15d3] = "\xd9\x69", [0x15d5] = "\xd9\x5f", [0x15d6] = "\xb4\xa5",
  [0x15d7] = "\xd9\x70", [0x15d8] = "\xd9\x68", [0x15d9] = "\xd9\x71",
  [0x15da] = "\xb4\xad", [0x15db] = "\xb4\xab", [0x15dc] = "\xd9\x66",
  [0x15dd] = "\xd9\x65", [0x15de] = "\x9d\xc3", [0x15df] = "\xd9\x63",
  [0x15e0] = "\xd9\x5d", [0x15e1] = "\xb4\xa4", [0x15e2] = "\x8d\xa2",
  [0x15e3] = "\xb4\xa2", [0x15e4] = "\xd1\xb9", [0x15e5] = "\xd9\x56",
  [0x15e6] = "\x9d\x4a", [0x15e7] = "\xdd\xb7", [0x15e8] = "\xd9\x57",
  [0x15e9] = "\xb4\x7b", [0x15ea] = "\xb4\xaa", [0x15eb] = "\xdd\x79",
  [0x15ed] = "\xb4\xa6", [0x15ee] = "\xb4\xa7", [0x15ef] = "\xd9\x58",
  [0x15f0] = "\xd9\x6f", [0x15f1] = "\xdd\x78", [0x15f2] = "\xd9\x60",
  [0x15f3] = "\xd9\x5b", [0x15f4] = "\xb4\xa9", [0x15f5] = "\xd9\x61",
  [0x15f6] = "\xd9\x5e", [0x15f8] = "\xfc\xb6", [0x15f9] = "\xb4\xae",
  [0x15fb] = "\x8d\xa3", [0x15fc] = "\x9e\x4b", [0x15fe] = "\x9e\x4d",
  [0x1606] = "\xb7\x70", [0x1607] = "\x8d\xa4", [0x1609] = "\xdd\x7c",
  [0x160a] = "\xdd\xb1", [0x160b] = "\xdd\xb6", [0x160c] = "\xdd\xaa",
  [0x160d] = "\xb7\x6c", [0x160e] = "\xdd\xbb", [0x160f] = "\xb7\x69",
  [0x1610] = "\xdd\x7a", [0x1612] = "\xdd\x7b", [0x1613] = "\xb7\x62",
  [0x1614] = "\xb7\x6b", [0x1615] = "\xdd\xa4", [0x1616] = "\xb7\x6e",
  [0x1617] = "\xb7\x6f", [0x1618] = "\xdd\xa5", [0x161a] = "\xdd\xb2",
  [0x161b] = "\xdd\xb8", [0x161c] = "\xb7\x6a", [0x161e] = "\xb7\x64",
  [0x161f] = "\xdd\xa3", [0x1620] = "\xdd\x7d", [0x1621] = "\xdd\xba",
  [0x1622] = "\xdd\xa8", [0x1623] = "\xdd\xa9", [0x1624] = "\xdd\x7e",
  [0x1625] = "\xdd\xb4", [0x1626] = "\xdd\xab", [0x1627] = "\xdd\xb5",
  [0x1628] = "\xdd\xad", [0x162a] = "\xb7\x65", [0x162b] = "\xe1\xd9",
  [0x162c] = "\xb7\x68", [0x162d] = "\xb7\x66", [0x162e] = "\xdd\xb9",
  [0x162f] = "\xdd\xb0", [0x1630] = "\xdd\xac", [0x1632] = "\x8a\xfd",
  [0x1633] = "\xdd\xa1", [0x1634] = "\xba\x53", [0x1635] = "\xdd\xaf",
  [0x1636] = "\xb7\x6d", [0x1637] = "\xdd\xa7", [0x1638] = "\xfc\xb5",
  [0x1639] = "\xdd\xa6", [0x163a] = "\xfc\xc3", [0x163b] = "\x93\xb2",
  [0x163d] = "\xb7\x67", [0x163e] = "\xb7\x63", [0x163f] = "\xe1\xee",
  [0x1640] = "\xdd\xb3", [0x1641] = "\xdd\xae", [0x1643] = "\xdd\xa2",
  [0x164b] = "\xe1\xe9", [0x164d] = "\xe1\xda", [0x164e] = "\xe1\xe5",
  [0x1650] = "\xe1\xec", [0x1651] = "\xba\x51", [0x1652] = "\xb4\xac",
  [0x1653] = "\xe1\xea", [0x1654] = "\xba\x4c", [0x1658] = "\xba\x4b",
  [0x1659] = "\xe1\xf1", [0x165a] = "\x8d\xa5", [0x165b] = "\xe1\xdb",
  [0x165c] = "\xe1\xe8", [0x165d] = "\xe1\xdc", [0x165e] = "\xe1\xe7",
  [0x165f] = "\xba\x4f", [0x1660] = "\xe1\xeb", [0x1661] = "\xd9\x62",
  [0x1665] = "\xe1\xf2", [0x1666] = "\xe1\xe3", [0x1667] = "\xba\x52",
  [0x1668] = "\xe5\xba", [0x1669] = "\xbc\xaf", [0x166b] = "\xe1\xf0",
  [0x166c] = "\xe1\xef", [0x166d] = "\xba\x54", [0x166e] = "\xe5\xad",
  [0x166f] = "\xbc\xb0", [0x1670] = "\xe5\xae", [0x1671] = "\x93\xa1",
  [0x1672] = "\xe1\xdf", [0x1673] = "\xe1\xe0", [0x1674] = "\xe1\xdd",
  [0x1675] = "\xe1\xe2", [0x1676] = "\xe1\xde", [0x1677] = "\xe1\xf3",
  [0x1678] = "\xba\x4e", [0x1679] = "\xbc\xb1", [0x167a] = "\xba\x50",
  [0x167b] = "\xba\x55", [0x167c] = "\x8a\xc6", [0x167d] = "\xe1\xe1",
  [0x167f] = "\xe1\xed", [0x1682] = "\xe1\xe6", [0x1685] = "\xe5\xb1",
  [0x1687] = "\xba\x4a", [0x1688] = "\xbc\xb4", [0x1689] = "\xe9\xaa",
  [0x168a] = "\xe5\xb6", [0x168b] = "\xe5\xb5", [0x168c] = "\xe5\xb7",
  [0x168d] = "\x8a\x5b", [0x168f] = "\xe5\xb4", [0x1690] = "\xbc\xb5",
  [0x1691] = "\x89\x4d", [0x1692] = "\xbc\xbb", [0x1693] = "\xbc\xb8",
  [0x1695] = "\xbc\xb9", [0x1696] = "\xe5\xaf", [0x1697] = "\xe5\xb2",
  [0x1698] = "\xe5\xbc", [0x1699] = "\xbc\xc1", [0x169a] = "\xbc\xbf",
  [0x169c] = "\xe5\xb3", [0x169d] = "\xd9\x5a", [0x169e] = "\xbc\xb2",
  [0x169f] = "\xe5\xb9", [0x16a0] = "\xe5\xb0", [0x16a2] = "\xbc\xc2",
  [0x16a3] = "\xe5\xb8", [0x16a4] = "\xba\x4d", [0x16a5] = "\xbc\xb7",
  [0x16a6] = "\xe1\xe4", [0x16a9] = "\xbc\xba", [0x16ab] = "\xbc\xbe",
  [0x16ac] = "\xbc\xc0", [0x16ad] = "\xbc\xbd", [0x16ae] = "\xbc\xbc",
  [0x16af] = "\xfe\xd4", [0x16b0] = "\xbc\xb6", [0x16b1] = "\xe5\xbb",
  [0x16b2] = "\xbc\xb3", [0x16b3] = "\xbc\xc3", [0x16b4] = "\x8a\x78",
  [0x16b6] = "\x93\xab", [0x16bb] = "\xbe\xd8", [0x16bc] = "\xbe\xd9",
  [0x16bd] = "\xe9\xa9", [0x16be] = "\xbe\xe2", [0x16bf] = "\xbe\xdf",
  [0x16c0] = "\x8d\xa7", [0x16c1] = "\xbe\xd6", [0x16c2] = "\xbe\xdd",
  [0x16c3] = "\xe9\xab", [0x16c4] = "\xbe\xdb", [0x16c5] = "\xbe\xd5",
  [0x16c7] = "\xbe\xdc", [0x16c9] = "\xe9\xa8", [0x16ca] = "\xc0\xbb",
  [0x16cb] = "\xbe\xd7", [0x16cd] = "\xbe\xde", [0x16ce] = "\xc0\xba",
  [0x16cf] = "\xe9\xa7", [0x16d0] = "\xe9\xa6", [0x16d2] = "\xbe\xe0",
  [0x16d3] = "\x9f\x45", [0x16d4] = "\xbe\xe1", [0x16d6] = "\xe9\xa5",
  [0x16d7] = "\xe9\xa4", [0x16d8] = "\xc0\xbc", [0x16d9] = "\xe9\xae",
  [0x16da] = "\xbe\xda", [0x16db] = "\xe9\xac", [0x16dd] = "\x8a\x56",
  [0x16e0] = "\xc0\xbd", [0x16e1] = "\xfc\xbf", [0x16e2] = "\xc0\xc2",
  [0x16e3] = "\xec\xea", [0x16e4] = "\xec\xec", [0x16e5] = "\xfc\xc0",
  [0x16e6] = "\xc0\xbf", [0x16e7] = "\x8e\xe6", [0x16e8] = "\xec\xed",
  [0x16e9] = "\xec\xe9", [0x16ea] = "\x8a\xa4", [0x16eb] = "\xec\xeb",
  [0x16ec] = "\xc0\xc0", [0x16ed] = "\xc0\xc3", [0x16ef] = "\xec\xe8",
  [0x16f0] = "\xc0\xbe", [0x16f1] = "\xc0\xc1", [0x16f2] = "\xc2\x59",
  [0x16f3] = "\xe9\xad", [0x16f4] = "\xc2\x58", [0x16f7] = "\xc2\x5e",
  [0x16f8] = "\xef\xd4", [0x16fa] = "\xc2\x5c", [0x16fb] = "\xc2\x5d",
  [0x16fc] = "\xef\xd7", [0x16fd] = "\xef\xd3", [0x16fe] = "\xc2\x5a",
  [0x16ff] = "\xef\xd1", [0x1700] = "\xc3\x6b", [0x1701] = "\xef\xd5",
  [0x1703] = "\xef\xd6", [0x1704] = "\xef\xd2", [0x1706] = "\xc2\x5b",
  [0x1707] = "\xf2\x42", [0x1709] = "\xf2\x45", [0x170a] = "\x89\x43",
  [0x170c] = "\xf2\x46", [0x170d] = "\xf2\x44", [0x170e] = "\xf2\x47",
  [0x170f] = "\xc3\x6c", [0x1710] = "\xf2\x43", [0x1711] = "\x93\xf3",
  [0x1713] = "\xf4\x4e", [0x1714] = "\xc4\x64", [0x1715] = "\xf4\x4d",
  [0x1716] = "\xf4\x4c", [0x1717] = "\xf4\x4b", [0x1718] = "\xc4\x63",
  [0x1719] = "\xc4\x65", [0x171b] = "\xf5\xcd", [0x171c] = "\xc4\xe2",
  [0x171d] = "\xc4\xe1", [0x171e] = "\xfc\xab", [0x171f] = "\x9e\xa2",
  [0x1720] = "\xf6\xe1", [0x1721] = "\xf6\xe0", [0x1722] = "\xf6\xe3",
  [0x1723] = "\xc5\xcb", [0x1724] = "\xc5\x75", [0x1725] = "\xf7\xdd",
  [0x1726] = "\xf6\xe2", [0x1729] = "\xf7\xdc", [0x172a] = "\xc5\xcd",
  [0x172b] = "\xc5\xcc", [0x172c] = "\xc5\xf3", [0x172d] = "\xf8\xa9",
  [0x172e] = "\xf8\xef", [0x172f] = "\xa4\xe4", [0x1730] = "\x9d\xc7",
  [0x1732] = "\xd9\x72", [0x1733] = "\xe9\xaf", [0x1734] = "\xc6\xd2",
  [0x1735] = "\x8b\xce", [0x1736] = "\xa6\xac", [0x1737] = "\xca\xf7",
  [0x1738] = "\xa7\xf1", [0x1739] = "\xa7\xef", [0x173b] = "\xa7\xf0",
  [0x173d] = "\xcc\xc1", [0x173e] = "\xa9\xf1", [0x173f] = "\xac\x46",
  [0x1741] = "\xce\xe7", [0x1743] = "\xce\xe8", [0x1745] = "\xac\x47",
  [0x1746] = "\xd1\xce", [0x1748] = "\xae\xc4", [0x1749] = "\xae\xc5",
  [0x174a] = "\xd1\xcd", [0x174d] = "\xfc\xc5", [0x174f] = "\xb1\xd3",
  [0x1751] = "\xb1\xcf", [0x1753] = "\xd5\xa7", [0x1754] = "\xb1\xd6",
  [0x1755] = "\xb1\xd5", [0x1756] = "\xb1\xce", [0x1757] = "\xb1\xd1",
  [0x1758] = "\xb1\xd4", [0x1759] = "\xb1\xd0", [0x175c] = "\xd9\x76",
  [0x175d] = "\xb1\xcd", [0x175e] = "\xb4\xaf", [0x175f] = "\xfc\xcb",
  [0x1762] = "\xb4\xb1", [0x1763] = "\xb4\xb2", [0x1764] = "\xd9\x75",
  [0x1765] = "\xd9\x78", [0x1766] = "\xb4\xb0", [0x1767] = "\xd9\x73",
  [0x1768] = "\xd9\x77", [0x176a] = "\xd9\x74", [0x176b] = "\x93\xb3",
  [0x176c] = "\xb7\x71", [0x176d] = "\xfc\xca", [0x176f] = "\xdd\xbc",
  [0x1772] = "\xba\x56", [0x1773] = "\xe1\xf4", [0x1774] = "\xbe\xe3",
  [0x1775] = "\xbc\xc4", [0x1776] = "\xe5\xbd", [0x1777] = "\xbc\xc5",
  [0x1778] = "\xbc\xc6", [0x1779] = "\xe5\xbf", [0x177a] = "\xe5\xbe",
  [0x177b] = "\xe5\xc0", [0x177c] = "\xe9\xb1", [0x177f] = "\xe9\xb0",
  [0x1780] = "\xec\xef", [0x1781] = "\xec\xee", [0x1782] = "\xc0\xc4",
  [0x1783] = "\xc0\xc5", [0x1784] = "\xf2\x48", [0x1785] = "\xfc\xc9",
  [0x1786] = "\x8d\xac", [0x1787] = "\xa4\xe5", [0x1788] = "\xfb\xc6",
  [0x1789] = "\x89\x67", [0x178c] = "\xd9\x79", [0x1790] = "\xb4\xb4",
  [0x1791] = "\xb4\xb3", [0x1792] = "\xdd\xbd", [0x1794] = "\xef\xd8",
  [0x1795] = "\xc4\xe3", [0x1796] = "\xf7\xde", [0x1797] = "\xa4\xe6",
  [0x1799] = "\xae\xc6", [0x179b] = "\xb1\xd8", [0x179c] = "\xb1\xd7",
  [0x179d] = "\xd9\x7a", [0x179e] = "\xd9\x7b", [0x179f] = "\xb7\x72",
  [0x17a0] = "\xe1\xf5", [0x17a1] = "\xba\x57", [0x17a2] = "\xe9\xb2",
  [0x17a4] = "\xa4\xe7", [0x17a5] = "\xa5\xb8", [0x17a7] = "\xa9\xf2",
  [0x17a8] = "\xcc\xc2", [0x17aa] = "\xce\xe9", [0x17ab] = "\xac\x48",
  [0x17ac] = "\xb1\xd9", [0x17ae] = "\xd9\x7c", [0x17af] = "\xb4\xb5",
  [0x17b0] = "\xb7\x73", [0x17b2] = "\xe5\xc1", [0x17b3] = "\xe5\xc2",
  [0x17b5] = "\xfc\xcd", [0x17b6] = "\xec\xf0", [0x17b7] = "\xc2\x5f",
  [0x17b8] = "\xf8\xf0", [0x17b9] = "\xa4\xe8", [0x17bb] = "\xcc\xc3",
  [0x17bc] = "\xa9\xf3", [0x17bd] = "\xac\x49", [0x17be] = "\x9c\xf3",
  [0x17bf] = "\xce\xea", [0x17c1] = "\xae\xc7", [0x17c2] = "\xd1\xd2",
  [0x17c3] = "\xd1\xd0", [0x17c4] = "\xd1\xd1", [0x17c5] = "\xae\xc8",
  [0x17c6] = "\xd1\xcf", [0x17cb] = "\xb1\xdb", [0x17cc] = "\xb1\xdc",
  [0x17cd] = "\xd5\xa8", [0x17ce] = "\xb1\xdd", [0x17cf] = "\xb1\xda",
  [0x17d0] = "\xd9\x7d", [0x17d1] = "\xfc\xd0", [0x17d2] = "\xd9\x7e",
  [0x17d3] = "\xdd\xbe", [0x17d4] = "\x95\xbb", [0x17d6] = "\xba\x59",
  [0x17d7] = "\xba\x58", [0x17da] = "\xec\xf1", [0x17db] = "\xef\xd9",
  [0x17dd] = "\xf2\x4a", [0x17de] = "\xf2\x49", [0x17df] = "\xf4\x4f",
  [0x17e0] = "\xfc\xd3", [0x17e1] = "\xc9\x5e", [0x17e2] = "\xac\x4a",
  [0x17e3] = "\xfc\xd4", [0x17e5] = "\xa4\xe9", [0x17e6] = "\xa5\xb9",
  [0x17e8] = "\xa6\xae", [0x17e9] = "\xa6\xad", [0x17ec] = "\xa6\xaf",
  [0x17ed] = "\xa6\xb0", [0x17ee] = "\xc9\xee", [0x17ef] = "\xc9\xed",
  [0x17f0] = "\xca\xf8", [0x17f1] = "\xa7\xf2", [0x17f2] = "\xca\xfb",
  [0x17f3] = "\xca\xfa", [0x17f4] = "\xca\xf9", [0x17f5] = "\xca\xfc",
  [0x17fa] = "\xa9\xf4", [0x17fb] = "\xcc\xc9", [0x17fc] = "\xcc\xc5",
  [0x17fd] = "\xcc\xce", [0x17ff] = "\x8d\xae", [0x1800] = "\xa9\xfb",
  [0x1802] = "\xa9\xf9", [0x1803] = "\xcc\xca", [0x1804] = "\xcc\xc6",
  [0x1805] = "\xcc\xcd", [0x1806] = "\xa9\xf8", [0x1807] = "\xaa\x40",
  [0x1808] = "\xcc\xc8", [0x1809] = "\xcc\xc4", [0x180a] = "\xa9\xfe",
  [0x180b] = "\xcc\xcb", [0x180c] = "\xa9\xf7", [0x180d] = "\xcc\xcc",
  [0x180e] = "\xa9\xfa", [0x180f] = "\xa9\xfc", [0x1810] = "\xcc\xd0",
  [0x1811] = "\xcc\xcf", [0x1812] = "\xcc\xc7", [0x1813] = "\xa9\xf6",
  [0x1814] = "\xa9\xf5", [0x1815] = "\xa9\xfd", [0x1818] = "\xfc\xd7",
  [0x181c] = "\xce\xef", [0x181d] = "\xce\xf5", [0x181e] = "\x93\xdb",
  [0x181f] = "\xac\x50", [0x1820] = "\xac\x4d", [0x1821] = "\xce\xec",
  [0x1822] = "\xce\xf1", [0x1823] = "\xfe\x63", [0x1824] = "\xac\x53",
  [0x1825] = "\xac\x4b", [0x1826] = "\xce\xf0", [0x1827] = "\xac\x4e",
  [0x1828] = "\xac\x51", [0x182b] = "\xce\xf3", [0x182d] = "\xac\x4c",
  [0x182e] = "\xce\xf8", [0x182f] = "\xac\x4f", [0x1830] = "\x93\xd5",
  [0x1831] = "\xac\x52", [0x1832] = "\xce\xed", [0x1833] = "\xce\xf2",
  [0x1834] = "\xce\xf6", [0x1835] = "\xce\xee", [0x1836] = "\xce\xeb",
  [0x1839] = "\xce\xf7", [0x183a] = "\xce\xf4", [0x1841] = "\xae\xd0",
  [0x1842] = "\xae\xc9", [0x1843] = "\xae\xcc", [0x1844] = "\xfc\xda",
  [0x1845] = "\xae\xcf", [0x1847] = "\xd1\xd5", [0x1848] = "\x9b\x71",
  [0x1849] = "\xae\xca", [0x184a] = "\xd1\xd3", [0x184b] = "\xfc\xdd",
  [0x184c] = "\xae\xce", [0x184f] = "\xae\xcb", [0x1851] = "\xd1\xd6",
  [0x1852] = "\xae\xcd", [0x1853] = "\x8d\xaf", [0x1857] = "\xfa\xf2",
  [0x1859] = "\xd5\xac", [0x185a] = "\xb1\xdf", [0x185b] = "\xd5\xab",
  [0x185c] = "\xd5\xad", [0x185d] = "\xb1\xde", [0x185e] = "\xb1\xe3",
  [0x185f] = "\xd1\xd4", [0x1861] = "\xd5\xaa", [0x1862] = "\xd5\xae",
  [0x1863] = "\x93\xd8", [0x1864] = "\xb1\xe0", [0x1865] = "\xd5\xa9",
  [0x1866] = "\xb1\xe2", [0x1867] = "\xfc\xdf", [0x1868] = "\xb1\xe1",
  [0x186a] = "\xd9\xa7", [0x186b] = "\x93\xd3", [0x186c] = "\xd9\xa2",
  [0x186e] = "\xb4\xb6", [0x186f] = "\xb4\xba", [0x1870] = "\xb4\xb7",
  [0x1871] = "\xd9\xa5", [0x1872] = "\xd9\xa8", [0x1873] = "\xfc\xe1",
  [0x1874] = "\xb4\xb8", [0x1876] = "\xb4\xb9", [0x1877] = "\xb4\xbe",
  [0x1878] = "\xdd\xc7", [0x1879] = "\xd9\xa6", [0x187a] = "\xb4\xbc",
  [0x187b] = "\xd9\xa3", [0x187c] = "\xd9\xa1", [0x187d] = "\x8e\x76",
  [0x187e] = "\xb4\xbd", [0x1880] = "\xd9\xa4", [0x1884] = "\xb7\x79",
  [0x1885] = "\xfc\x62", [0x1886] = "\xdd\xbf", [0x1887] = "\xb7\x76",
  [0x1888] = "\xb7\x77", [0x1889] = "\xb7\x75", [0x188a] = "\xdd\xc4",
  [0x188b] = "\xdd\xc3", [0x188c] = "\xdd\xc0", [0x188d] = "\xb7\x7b",
  [0x188e] = "\x93\xd1", [0x1890] = "\xdd\xc2", [0x1891] = "\xb4\xbb",
  [0x1892] = "\x8d\xb1", [0x1894] = "\xdd\xc6", [0x1895] = "\xdd\xc1",
  [0x1896] = "\xb7\x78", [0x1897] = "\xb7\x74", [0x1898] = "\xb7\x7a",
  [0x1899] = "\xdd\xc5", [0x189a] = "\x98\x59", [0x189d] = "\xba\x5c",
  [0x189f] = "\xe1\xf8", [0x18a0] = "\xe1\xf7", [0x18a1] = "\xe1\xf6",
  [0x18a2] = "\xba\x5a", [0x18a4] = "\xfb\x52", [0x18a8] = "\xba\x5b",
  [0x18a9] = "\xe5\xc5", [0x18aa] = "\xe5\xc8", [0x18ab] = "\xbc\xc8",
  [0x18ad] = "\xfb\x53", [0x18ae] = "\xbc\xc7", [0x18af] = "\xe5\xc9",
  [0x18b0] = "\xe5\xc4", [0x18b1] = "\xbc\xca", [0x18b2] = "\xe5\xc6",
  [0x18b3] = "\xfb\x4d", [0x18b4] = "\xbc\xc9", [0x18b5] = "\xe5\xc3",
  [0x18b6] = "\x9c\xbf", [0x18b7] = "\xe5\xc7", [0x18b8] = "\xbe\xe9",
  [0x18b9] = "\xbe\xe6", [0x18ba] = "\xe9\xbb", [0x18bb] = "\xe9\xba",
  [0x18bd] = "\xe9\xb9", [0x18be] = "\xe9\xb4", [0x18bf] = "\x9b\x72",
  [0x18c0] = "\xe9\xb5", [0x18c4] = "\xbe\xe7", [0x18c6] = "\xbe\xe4",
  [0x18c7] = "\xbe\xe8", [0x18c8] = "\xe9\xb3", [0x18c9] = "\xbe\xe5",
  [0x18ca] = "\xe9\xb6", [0x18cb] = "\xe9\xb7", [0x18cc] = "\xe9\xbc",
  [0x18cd] = "\xfb\x50", [0x18ce] = "\x93\xbe", [0x18cf] = "\xe9\xb8",
  [0x18d2] = "\xec\xf2", [0x18d6] = "\xc0\xc7", [0x18d8] = "\xef\xdc",
  [0x18d9] = "\xc0\xc6", [0x18da] = "\xef\xda", [0x18db] = "\xef\xdb",
  [0x18dc] = "\xc2\x60", [0x18dd] = "\xc3\x6e", [0x18de] = "\xf2\x4b",
  [0x18e0] = "\xc3\x6d", [0x18e3] = "\xf4\x51", [0x18e4] = "\xf4\x52",
  [0x18e6] = "\xc4\x66", [0x18e8] = "\xf4\x50", [0x18e9] = "\xc4\xe4",
  [0x18eb] = "\xf7\xdf", [0x18ec] = "\xc5\xce", [0x18ed] = "\xf8\xaa",
  [0x18ee] = "\xf8\xab", [0x18f0] = "\xa4\xea", [0x18f1] = "\x9d\xf1",
  [0x18f2] = "\xa6\xb1", [0x18f3] = "\xa6\xb2", [0x18f4] = "\xa7\xf3",
  [0x18f6] = "\xcc\xd1", [0x18f7] = "\xac\x54", [0x18f8] = "\xae\xd1",
  [0x18f9] = "\xb1\xe4", [0x18fc] = "\xb0\xd2", [0x18fe] = "\xb4\xbf",
  [0x18ff] = "\xb4\xc0", [0x1900] = "\xb3\xcc", [0x1901] = "\xd9\xa9",
  [0x1902] = "\xfc\xeb", [0x1903] = "\xb7\x7c", [0x1904] = "\xe1\xfa",
  [0x1905] = "\xe1\xf9", [0x1908] = "\xa4\xeb", [0x1909] = "\xa6\xb3",
  [0x190a] = "\xcc\xd2", [0x190b] = "\xaa\x42", [0x190c] = "\xa0\xbb",
  [0x190d] = "\xaa\x41", [0x190e] = "\x9b\x7e", [0x190f] = "\xce\xf9",
  [0x1910] = "\xce\xfa", [0x1912] = "\xd1\xd7", [0x1913] = "\xd1\xd8",
  [0x1914] = "\xae\xd2", [0x1915] = "\xae\xd3", [0x1916] = "\x8d\xb3",
  [0x1917] = "\xae\xd4", [0x1918] = "\xd5\xaf", [0x191b] = "\xb1\xe6",
  [0x191d] = "\xb4\xc2", [0x191e] = "\x9a\xe8", [0x191f] = "\xb4\xc1",
  [0x1920] = "\xdd\xc8", [0x1921] = "\xdf\x7a", [0x1922] = "\xe1\xfb",
  [0x1923] = "\xe9\xbd", [0x1925] = "\x8e\xdc", [0x1926] = "\xc2\x61",
  [0x1927] = "\xc4\x67", [0x1928] = "\xa4\xec", [0x192a] = "\xa5\xbc",
  [0x192b] = "\xa5\xbd", [0x192c] = "\xa5\xbb", [0x192d] = "\xa5\xbe",
  [0x192e] = "\xa5\xba", [0x1931] = "\xa6\xb6", [0x1933] = "\xc9\xf6",
  [0x1934] = "\xa6\xb5", [0x1935] = "\xa6\xb7", [0x1936] = "\x9c\xf9",
  [0x1938] = "\xc9\xf1", [0x1939] = "\xc9\xf0", [0x193a] = "\xc9\xf3",
  [0x193b] = "\xc9\xf2", [0x193c] = "\xc9\xf5", [0x193d] = "\xa6\xb4",
  [0x193e] = "\xc9\xef", [0x193f] = "\xc9\xf4", [0x1944] = "\xfa\x50",
  [0x1945] = "\xca\xfd", [0x1946] = "\xa7\xfd", [0x1947] = "\xca\xfe",
  [0x1948] = "\xcb\x43", [0x1949] = "\xa7\xfc", [0x194b] = "\xcb\x47",
  [0x194c] = "\xcb\x42", [0x194d] = "\xcb\x45", [0x194e] = "\xa7\xf5",
  [0x194f] = "\xa7\xf6", [0x1950] = "\xa7\xf7", [0x1951] = "\xa7\xf8",
  [0x1953] = "\xa8\x40", [0x1955] = "\xcb\x41", [0x1956] = "\xa7\xfa",
  [0x1957] = "\xa8\x41", [0x1959] = "\xcb\x40", [0x195a] = "\xcb\x46",
  [0x195c] = "\xa7\xf9", [0x195d] = "\xcb\x44", [0x195e] = "\xa7\xfb",
  [0x195f] = "\xa7\xf4", [0x1960] = "\xa7\xfe", [0x1961] = "\x98\xe7",
  [0x1962] = "\xfc\xf3", [0x1967] = "\xfc\xf2", [0x196a] = "\xaa\x57",
  [0x196c] = "\xcc\xd4", [0x196d] = "\xaa\x43", [0x196f] = "\xaa\x4d",
  [0x1970] = "\xaa\x4e", [0x1971] = "\xaa\x46", [0x1972] = "\xaa\x58",
  [0x1973] = "\xaa\x48", [0x1974] = "\xcc\xdc", [0x1975] = "\xaa\x53",
  [0x1976] = "\xcc\xd7", [0x1977] = "\xaa\x49", [0x1978] = "\xcc\xe6",
  [0x1979] = "\xcc\xe7", [0x197a] = "\xcc\xdf", [0x197b] = "\xcc\xd8",
  [0x197c] = "\xaa\x56", [0x197d] = "\xcc\xe4", [0x197e] = "\xaa\x51",
  [0x197f] = "\xaa\x4f", [0x1981] = "\xcc\xe5", [0x1983] = "\xcc\xe3",
  [0x1984] = "\xcc\xdb", [0x1985] = "\xcc\xd3", [0x1986] = "\xcc\xda",
  [0x1987] = "\xaa\x4a", [0x1989] = "\xaa\x50", [0x198b] = "\xaa\x44",
  [0x198c] = "\xcc\xde", [0x198d] = "\xcc\xdd", [0x198e] = "\xcc\xd5",
  [0x198f] = "\x93\xe5", [0x1990] = "\xaa\x52", [0x1991] = "\xcc\xe1",
  [0x1992] = "\xcc\xd6", [0x1993] = "\xaa\x55", [0x1994] = "\xcc\xe8",
  [0x1995] = "\xaa\x45", [0x1997] = "\xaa\x4c", [0x1998] = "\xcc\xd9",
  [0x1999] = "\xcc\xe2", [0x199a] = "\xaa\x54", [0x199c] = "\xaa\x47",
  [0x199d] = "\xaa\x4b", [0x199f] = "\xcc\xe0", [0x19a0] = "\x9a\x59",
  [0x19a4] = "\x8d\xb5", [0x19ac] = "\xfd\x4d", [0x19ae] = "\xcf\x5b",
  [0x19af] = "\xac\x5c", [0x19b0] = "\xac\x69", [0x19b1] = "\xfd\x5e",
  [0x19b2] = "\xcf\x56", [0x19b3] = "\xcf\x4c", [0x19b4] = "\xac\x62",
  [0x19b5] = "\xcf\x4a", [0x19b6] = "\xac\x5b", [0x19b7] = "\xcf\x45",
  [0x19b8] = "\xac\x65", [0x19b9] = "\xcf\x52", [0x19ba] = "\xce\xfe",
  [0x19bb] = "\xcf\x41", [0x19bf] = "\x8f\x7d", [0x19c0] = "\xcf\x44",
  [0x19c1] = "\xce\xfb", [0x19c2] = "\xcf\x51", [0x19c3] = "\xcf\x61",
  [0x19c4] = "\xac\x60", [0x19c5] = "\xcf\x46", [0x19c6] = "\xcf\x58",
  [0x19c8] = "\xce\xfd", [0x19c9] = "\xcf\x5f", [0x19ca] = "\xcf\x60",
  [0x19cb] = "\xcf\x63", [0x19cc] = "\xcf\x5a", [0x19cd] = "\xcf\x4b",
  [0x19ce] = "\xcf\x53", [0x19cf] = "\xac\x66", [0x19d0] = "\xac\x59",
  [0x19d1] = "\xac\x61", [0x19d2] = "\xac\x6d", [0x19d3] = "\xac\x56",
  [0x19d4] = "\xac\x58", [0x19d6] = "\x95\x47", [0x19d7] = "\xfc\xf6",
  [0x19d8] = "\xcf\x43", [0x19d9] = "\xac\x6a", [0x19da] = "\xac\x63",
  [0x19db] = "\xcf\x5d", [0x19dc] = "\xcf\x40", [0x19dd] = "\xac\x6c",
  [0x19de] = "\xac\x67", [0x19df] = "\xcf\x49", [0x19e2] = "\xac\x6b",
  [0x19e3] = "\xcf\x50", [0x19e4] = "\xcf\x48", [0x19e5] = "\xac\x64",
  [0x19e6] = "\xcf\x5c", [0x19e7] = "\xcf\x54", [0x19e9] = "\xac\x5e",
  [0x19ea] = "\xcf\x62", [0x19eb] = "\xcf\x47", [0x19ec] = "\xac\x5a",
  [0x19ed] = "\xcf\x59", [0x19ee] = "\xcf\x4f", [0x19ef] = "\xac\x5f",
  [0x19f0] = "\xcf\x55", [0x19f1] = "\xac\x57", [0x19f2] = "\xce\xfc",
  [0x19f3] = "\xac\x68", [0x19f4] = "\xae\xe3", [0x19f5] = "\xac\x5d",
  [0x19f6] = "\xcf\x4e", [0x19f7] = "\xcf\x4d", [0x19f8] = "\xcf\x42",
  [0x19f9] = "\x92\x50", [0x19fa] = "\xcf\x5e", [0x19fc] = "\xcf\x57",
  [0x19fe] = "\x89\x68", [0x19ff] = "\xac\x55", [0x1a00] = "\x8d\xb6",
  [0x1a01] = "\xfc\xfb", [0x1a02] = "\xa0\x7d", [0x1a03] = "\x98\xfc",
  [0x1a04] = "\x89\x69", [0x1a0d] = "\xfe\x4f", [0x1a10] = "\x92\x56",
  [0x1a12] = "\xd1\xec", [0x1a13] = "\xae\xea", [0x1a14] = "\xd1\xed",
  [0x1a16] = "\xd1\xe1", [0x1a17] = "\xae\xdf", [0x1a18] = "\xae\xeb",
  [0x1a1a] = "\xd1\xda", [0x1a1b] = "\xfa\xc9", [0x1a1c] = "\xd1\xe3",
  [0x1a1d] = "\xd1\xeb", [0x1a1e] = "\x93\xe8", [0x1a1f] = "\xd1\xd9",
  [0x1a20] = "\xd1\xf4", [0x1a21] = "\xae\xd5", [0x1a22] = "\xfc\xf8",
  [0x1a25] = "\xd1\xf3", [0x1a26] = "\xd1\xee", [0x1a28] = "\xd1\xef",
  [0x1a29] = "\xae\xdd", [0x1a2a] = "\xae\xe8", [0x1a2b] = "\xd1\xe5",
  [0x1a2d] = "\xd1\xe6", [0x1a2e] = "\xd1\xf0", [0x1a2f] = "\xd1\xe7",
  [0x1a31] = "\xd1\xe2", [0x1a32] = "\xd1\xdc", [0x1a33] = "\xd1\xdd",
  [0x1a34] = "\xd1\xea", [0x1a35] = "\xd1\xe4", [0x1a36] = "\x9c\xe3",
  [0x1a37] = "\xfd\xa9", [0x1a38] = "\xae\xd6", [0x1a39] = "\xae\xda",
  [0x1a3a] = "\xd1\xf2", [0x1a3b] = "\xd1\xde", [0x1a3c] = "\xae\xe6",
  [0x1a3d] = "\xae\xe2", [0x1a3e] = "\xfc\x44", [0x1a40] = "\xae\xe5",
  [0x1a41] = "\xae\xec", [0x1a42] = "\xae\xdb", [0x1a43] = "\xae\xe7",
  [0x1a44] = "\xd1\xe9", [0x1a45] = "\xae\xe9", [0x1a46] = "\xae\xd8",
  [0x1a47] = "\x96\x40", [0x1a48] = "\xae\xd7", [0x1a49] = "\xd1\xdb",
  [0x1a4a] = "\x8d\xb8", [0x1a4b] = "\xd1\xdf", [0x1a4c] = "\xae\xe0",
  [0x1a4d] = "\xd1\xf1", [0x1a4e] = "\xd1\xe8", [0x1a4f] = "\xd1\xe0",
  [0x1a50] = "\xae\xe4", [0x1a51] = "\xae\xe1", [0x1a53] = "\xae\xd9",
  [0x1a54] = "\xae\xdc", [0x1a55] = "\x9b\x4a", [0x1a56] = "\x8f\xb9",
  [0x1a5d] = "\xfc\xfe", [0x1a65] = "\x89\x6a", [0x1a6b] = "\xd5\xc4",
  [0x1a6d] = "\xd5\xb4", [0x1a6e] = "\xd5\xb5", [0x1a6f] = "\xd5\xb9",
  [0x1a71] = "\xd5\xc8", [0x1a72] = "\xd5\xc5", [0x1a74] = "\xd5\xbe",
  [0x1a75] = "\xd5\xbd", [0x1a76] = "\xb1\xed", [0x1a77] = "\xd5\xc1",
  [0x1a78] = "\xd5\xd0", [0x1a79] = "\xd5\xb0", [0x1a7b] = "\xd5\xd1",
  [0x1a7c] = "\xd5\xc3", [0x1a7d] = "\xd5\xd5", [0x1a7e] = "\xd5\xc9",
  [0x1a7f] = "\xb1\xec", [0x1a80] = "\xd5\xc7", [0x1a81] = "\xb1\xe7",
  [0x1a82] = "\xb1\xfc", [0x1a83] = "\xb1\xf2", [0x1a84] = "\x8d\xb9",
  [0x1a85] = "\xb1\xf6", [0x1a86] = "\xb1\xf5", [0x1a87] = "\xd5\xb1",
  [0x1a88] = "\x91\x7e", [0x1a89] = "\xd5\xce", [0x1a8a] = "\xd5\xd4",
  [0x1a8b] = "\xd5\xcc", [0x1a8c] = "\xd5\xd3", [0x1a8f] = "\xd5\xc0",
  [0x1a90] = "\xd5\xb2", [0x1a91] = "\xd5\xd2", [0x1a92] = "\xd5\xc2",
  [0x1a93] = "\xb1\xea", [0x1a94] = "\xb1\xf7", [0x1a96] = "\xd5\xcb",
  [0x1a97] = "\xb1\xf0", [0x1a98] = "\x93\xf4", [0x1a9b] = "\xd5\xca",
  [0x1a9c] = "\xd5\xb3", [0x1a9d] = "\xb1\xf8", [0x1a9f] = "\xb1\xfa",
  [0x1aa0] = "\xd5\xcd", [0x1aa1] = "\xb1\xfb", [0x1aa2] = "\xb1\xe9",
  [0x1aa3] = "\xd5\xba", [0x1aa4] = "\xd5\xcf", [0x1aa6] = "\xfb\x7c",
  [0x1aa7] = "\xb1\xef", [0x1aa8] = "\xb1\xf9", [0x1aa9] = "\xd5\xbc",
  [0x1aaa] = "\xd5\xc6", [0x1aab] = "\xd5\xb7", [0x1aac] = "\xd5\xbb",
  [0x1aad] = "\xb1\xf4", [0x1aae] = "\xd5\xb6", [0x1aaf] = "\xb1\xe8",
  [0x1ab0] = "\xb1\xf1", [0x1ab1] = "\xb1\xee", [0x1ab2] = "\xd5\xbf",
  [0x1ab3] = "\xae\xde", [0x1ab4] = "\xd9\xc0", [0x1ab5] = "\xb1\xeb",
  [0x1ab6] = "\x93\xe7", [0x1ab9] = "\x97\xef", [0x1abd] = "\xfe\x4a",
  [0x1ac3] = "\xfd\x45", [0x1ac4] = "\xb1\xf3", [0x1ac5] = "\x96\xa5",
  [0x1ac6] = "\xd9\xc3", [0x1ac7] = "\xd9\xd9", [0x1ac8] = "\xd9\xce",
  [0x1ac9] = "\xb4\xd6", [0x1aca] = "\xfe\xe0", [0x1acb] = "\xb4\xd1",
  [0x1acc] = "\xd9\xbd", [0x1acd] = "\xb4\xd2", [0x1ace] = "\xd9\xcd",
  [0x1ad0] = "\xd9\xc6", [0x1ad1] = "\xd9\xd3", [0x1ad2] = "\xb4\xce",
  [0x1ad3] = "\xd9\xab", [0x1ad4] = "\xd9\xd5", [0x1ad5] = "\xb4\xc4",
  [0x1ad6] = "\xd9\xb3", [0x1ad7] = "\xb4\xc7", [0x1ad8] = "\xb4\xc6",
  [0x1ada] = "\xb4\xd7", [0x1adc] = "\xd9\xad", [0x1add] = "\xd9\xcf",
  [0x1ade] = "\xd9\xd0", [0x1adf] = "\xb4\xc9", [0x1ae0] = "\xb4\xc5",
  [0x1ae1] = "\xd9\xbb", [0x1ae3] = "\xb4\xd0", [0x1ae4] = "\xd9\xb6",
  [0x1ae6] = "\xd9\xd1", [0x1ae7] = "\xb4\xcc", [0x1ae8] = "\xd9\xc9",
  [0x1ae9] = "\xd9\xd6", [0x1aea] = "\xd9\xb0", [0x1aeb] = "\xd9\xb5",
  [0x1aec] = "\xd9\xaf", [0x1aee] = "\xb4\xcb", [0x1aef] = "\xd9\xc2",
  [0x1af0] = "\xdd\xde", [0x1af1] = "\xd9\xb1", [0x1af2] = "\xb4\xcf",
  [0x1af3] = "\xd9\xba", [0x1af4] = "\xd9\xd2", [0x1af5] = "\xb4\xca",
  [0x1af6] = "\xd9\xb7", [0x1af7] = "\xd9\xb4", [0x1af8] = "\xd9\xc5",
  [0x1af9] = "\xb4\xcd", [0x1afa] = "\xb4\xc3", [0x1afb] = "\xb4\xd9",
  [0x1afc] = "\xd9\xc8", [0x1afd] = "\xd9\xc7", [0x1b00] = "\xfd\x48",
  [0x1b01] = "\xfd\x47", [0x1b02] = "\xfe\xf2", [0x1b03] = "\xfe\x6a",
  [0x1b04] = "\xd9\xac", [0x1b05] = "\xb4\xc8", [0x1b06] = "\xd9\xd4",
  [0x1b07] = "\xd9\xbc", [0x1b08] = "\xd9\xbe", [0x1b09] = "\x8d\xbd",
  [0x1b0a] = "\xd9\xcb", [0x1b0b] = "\xd9\xca", [0x1b0c] = "\xd9\xaa",
  [0x1b0d] = "\xb4\xd3", [0x1b0e] = "\xb4\xd5", [0x1b0f] = "\xd9\xb2",
  [0x1b10] = "\xd9\xb9", [0x1b11] = "\xd9\xc1", [0x1b12] = "\xb4\xd4",
  [0x1b13] = "\xd9\xb8", [0x1b14] = "\xd9\xc4", [0x1b15] = "\xd9\xd7",
  [0x1b17] = "\xd9\xcc", [0x1b18] = "\x9b\xa1", [0x1b1a] = "\x9a\xb7",
  [0x1b1b] = "\x8e\xfc", [0x1b25] = "\xd9\xd8", [0x1b2a] = "\xd9\xae",
  [0x1b2c] = "\x9f\xa1", [0x1b2f] = "\xdd\xf2", [0x1b30] = "\xb7\xa6",
  [0x1b32] = "\xdd\xf0", [0x1b33] = "\xdd\xdb", [0x1b34] = "\xdd\xe0",
  [0x1b35] = "\xdd\xd9", [0x1b36] = "\xfd\x51", [0x1b37] = "\xdd\xec",
  [0x1b38] = "\xdd\xcb", [0x1b39] = "\xdd\xd2", [0x1b3b] = "\xdd\xea",
  [0x1b3c] = "\xdd\xf4", [0x1b3d] = "\xdd\xdc", [0x1b3e] = "\xfa\xad",
  [0x1b3f] = "\xdd\xcf", [0x1b40] = "\xdd\xe2", [0x1b41] = "\xdd\xe7",
  [0x1b42] = "\xdd\xd3", [0x1b43] = "\x8d\xbe", [0x1b44] = "\xdd\xe4",
  [0x1b45] = "\xdd\xd0", [0x1b46] = "\x89\xa4", [0x1b48] = "\xdd\xd7",
  [0x1b49] = "\xdd\xd8", [0x1b4a] = "\xb7\xa8", [0x1b4b] = "\xdd\xeb",
  [0x1b4c] = "\xdd\xe9", [0x1b4e] = "\xdd\xcc", [0x1b4f] = "\xdd\xee",
  [0x1b51] = "\xdd\xef", [0x1b52] = "\xdd\xf1", [0x1b53] = "\xb7\xac",
  [0x1b54] = "\xb7\xa4", [0x1b55] = "\x9a\xd9", [0x1b56] = "\xd5\xb8",
  [0x1b57] = "\xdd\xd4", [0x1b58] = "\xdd\xe6", [0x1b59] = "\xdd\xd5",
  [0x1b5a] = "\xb7\xa1", [0x1b5b] = "\xb7\xb1", [0x1b5c] = "\xdd\xed",
  [0x1b5d] = "\xb7\xaf", [0x1b5e] = "\xb7\xab", [0x1b5f] = "\xdd\xca",
  [0x1b60] = "\xb7\xa3", [0x1b61] = "\xfd\x4e", [0x1b62] = "\xdd\xcd",
  [0x1b63] = "\xb7\xb0", [0x1b64] = "\x8d\xc0", [0x1b65] = "\xdd\xdd",
  [0x1b66] = "\xdd\xc9", [0x1b67] = "\x97\xf0", [0x1b68] = "\xb7\xa9",
  [0x1b69] = "\xdd\xe1", [0x1b6a] = "\xdd\xd1", [0x1b6b] = "\xb7\xaa",
  [0x1b6c] = "\xdd\xda", [0x1b6d] = "\xb7\x7e", [0x1b6e] = "\xb4\xd8",
  [0x1b6f] = "\xdd\xe3", [0x1b70] = "\xd9\xbf", [0x1b71] = "\xdd\xce",
  [0x1b72] = "\x93\xb4", [0x1b73] = "\xfd\x44", [0x1b74] = "\xdd\xe8",
  [0x1b75] = "\xb7\xa5", [0x1b76] = "\xdd\xe5", [0x1b77] = "\xb7\xa2",
  [0x1b78] = "\xdd\xdf", [0x1b79] = "\xb7\xad", [0x1b7a] = "\xdd\xd6",
  [0x1b7b] = "\xdd\xf3", [0x1b80] = "\x9f\xa7", [0x1b82] = "\xb7\xa7",
  [0x1b83] = "\xde\xc6", [0x1b85] = "\x8d\xc2", [0x1b86] = "\xb7\xae",
  [0x1b8a] = "\x99\xb6", [0x1b8d] = "\xe2\x4a", [0x1b8e] = "\xe2\x48",
  [0x1b90] = "\xe2\x5e", [0x1b91] = "\xe2\x46", [0x1b93] = "\xe2\x58",
  [0x1b94] = "\xb7\x7d", [0x1b95] = "\xba\x5f", [0x1b96] = "\xe2\x42",
  [0x1b97] = "\xe2\x5d", [0x1b98] = "\xfd\x52", [0x1b99] = "\xe2\x47",
  [0x1b9a] = "\xe2\x55", [0x1b9b] = "\xba\x64", [0x1b9c] = "\xba\x5d",
  [0x1b9e] = "\xe2\x5b", [0x1b9f] = "\x8d\xc1", [0x1ba0] = "\xe2\x40",
  [0x1ba1] = "\xe2\x5a", [0x1ba2] = "\x8e\x46", [0x1ba3] = "\xba\x6f",
  [0x1ba4] = "\xe2\x51", [0x1ba5] = "\xe2\x61", [0x1ba6] = "\xba\x6d",
  [0x1ba7] = "\xe2\x49", [0x1ba8] = "\xba\x5e", [0x1ba9] = "\xe2\x4b",
  [0x1baa] = "\xe2\x59", [0x1bab] = "\xba\x67", [0x1bac] = "\xe2\x44",
  [0x1bad] = "\xba\x6b", [0x1bae] = "\xba\x61", [0x1baf] = "\xe2\x4d",
  [0x1bb0] = "\xe2\x43", [0x1bb1] = "\xe1\xfc", [0x1bb2] = "\xa0\xd1",
  [0x1bb3] = "\xe2\x57", [0x1bb4] = "\xba\x68", [0x1bb5] = "\xe2\x60",
  [0x1bb6] = "\xe1\xfd", [0x1bb7] = "\xba\x65", [0x1bb9] = "\xe2\x53",
  [0x1bbb] = "\xba\x66", [0x1bbc] = "\xe2\x45", [0x1bbd] = "\xe2\x50",
  [0x1bbe] = "\xe2\x4c", [0x1bbf] = "\xe2\x4e", [0x1bc0] = "\x9f\xca",
  [0x1bc1] = "\xba\x60", [0x1bc2] = "\xe2\x5f", [0x1bc3] = "\xba\x6e",
  [0x1bc4] = "\xe2\x4f", [0x1bc6] = "\xe2\x62", [0x1bc9] = "\xe1\xfe",
  [0x1bca] = "\xe2\x54", [0x1bcb] = "\xba\x63", [0x1bcc] = "\xba\x6c",
  [0x1bcd] = "\xba\x6a", [0x1bce] = "\xe2\x41", [0x1bcf] = "\xe2\x56",
  [0x1bd0] = "\xba\x69", [0x1bd1] = "\x92\xcf", [0x1bd3] = "\xba\x62",
  [0x1bd4] = "\xe2\x52", [0x1bd5] = "\x9c\xf4", [0x1bd6] = "\x8d\xc4",
  [0x1bd9] = "\xe2\x5c", [0x1be1] = "\xfd\x41", [0x1be2] = "\xe5\xd5",
  [0x1be4] = "\xe5\xd1", [0x1be5] = "\xe5\xcd", [0x1be6] = "\xe5\xe1",
  [0x1be7] = "\xe5\xde", [0x1be8] = "\xbc\xcd", [0x1be9] = "\x9b\x4c",
  [0x1beb] = "\xe5\xe5", [0x1bec] = "\xe5\xd4", [0x1bed] = "\xbc\xd8",
  [0x1bee] = "\xe5\xdb", [0x1bf1] = "\xe5\xd0", [0x1bf2] = "\xe5\xda",
  [0x1bf3] = "\xbc\xd5", [0x1bf4] = "\xe5\xee", [0x1bf6] = "\xe5\xeb",
  [0x1bf7] = "\xe5\xdd", [0x1bf8] = "\xe5\xce", [0x1bf9] = "\xfd\x57",
  [0x1bfa] = "\xfc\xef", [0x1bfb] = "\xe5\xe2", [0x1bfc] = "\xe5\xe4",
  [0x1bfd] = "\xbc\xd1", [0x1bfe] = "\xe5\xd8", [0x1bff] = "\xe5\xd3",
  [0x1c00] = "\xe5\xca", [0x1c01] = "\xbc\xce", [0x1c02] = "\xbc\xd6",
  [0x1c03] = "\x9c\xde", [0x1c04] = "\xe5\xe7", [0x1c05] = "\xbc\xd7",
  [0x1c06] = "\xe5\xcb", [0x1c07] = "\xe5\xed", [0x1c08] = "\xe5\xe0",
  [0x1c09] = "\xe5\xe6", [0x1c0a] = "\xbc\xd4", [0x1c0b] = "\xfd\x42",
  [0x1c0c] = "\x98\x6c", [0x1c0d] = "\xe5\xe3", [0x1c0f] = "\xe5\xea",
  [0x1c11] = "\xbc\xd9", [0x1c13] = "\xbc\xd3", [0x1c14] = "\xe5\xdc",
  [0x1c15] = "\xe5\xcf", [0x1c16] = "\xe5\xef", [0x1c17] = "\xe5\xcc",
  [0x1c18] = "\xe5\xe8", [0x1c19] = "\xbc\xd0", [0x1c1a] = "\x97\xf9",
  [0x1c1b] = "\xe5\xd6", [0x1c1c] = "\x95\x58", [0x1c1d] = "\xe5\xd7",
  [0x1c1e] = "\xbc\xcf", [0x1c1f] = "\xbc\xcc", [0x1c20] = "\xe5\xd2",
  [0x1c21] = "\xbc\xd2", [0x1c23] = "\xbc\xcb", [0x1c25] = "\xe5\xe9",
  [0x1c26] = "\xe5\xec", [0x1c27] = "\xe5\xd9", [0x1c28] = "\xe9\xca",
  [0x1c2b] = "\x98\x5e", [0x1c2c] = "\xfe\x7b", [0x1c2d] = "\x94\xcd",
  [0x1c32] = "\xe9\xc2", [0x1c33] = "\x93\xee", [0x1c34] = "\xe9\xbe",
  [0x1c35] = "\xbe\xf6", [0x1c38] = "\xbe\xeb", [0x1c39] = "\xbe\xf0",
  [0x1c3a] = "\xbe\xec", [0x1c3b] = "\xe9\xcc", [0x1c3c] = "\xe9\xd7",
  [0x1c3d] = "\xbe\xea", [0x1c3e] = "\xe9\xc4", [0x1c3f] = "\xe9\xcd",
  [0x1c40] = "\xe5\xdf", [0x1c41] = "\xe9\xce", [0x1c44] = "\xbe\xf1",
  [0x1c45] = "\xfd\x5a", [0x1c46] = "\xe9\xdd", [0x1c47] = "\xbe\xf5",
  [0x1c48] = "\xbe\xf8", [0x1c49] = "\xe9\xc0", [0x1c4b] = "\xbe\xf4",
  [0x1c4c] = "\x93\xf5", [0x1c4d] = "\xe9\xdb", [0x1c4e] = "\xe9\xdc",
  [0x1c4f] = "\xe9\xd2", [0x1c50] = "\xe9\xd1", [0x1c51] = "\xe9\xc9",
  [0x1c52] = "\x93\xef", [0x1c53] = "\x8e\xea", [0x1c54] = "\xe9\xd3",
  [0x1c55] = "\xe9\xda", [0x1c56] = "\xe9\xd9", [0x1c57] = "\x8f\x5b",
  [0x1c58] = "\xbe\xef", [0x1c59] = "\xbe\xed", [0x1c5a] = "\xe9\xcb",
  [0x1c5b] = "\xe9\xc8", [0x1c5d] = "\xe9\xc5", [0x1c5e] = "\xe9\xd8",
  [0x1c5f] = "\xbe\xf7", [0x1c60] = "\xe9\xd6", [0x1c61] = "\xbe\xf3",
  [0x1c62] = "\xbe\xf2", [0x1c64] = "\xe9\xd0", [0x1c65] = "\x8d\xc6",
  [0x1c66] = "\xe9\xbf", [0x1c67] = "\xe9\xc1", [0x1c68] = "\xe9\xc3",
  [0x1c69] = "\xe9\xd5", [0x1c6a] = "\xe9\xcf", [0x1c6b] = "\xbe\xee",
  [0x1c6d] = "\xe9\xc6", [0x1c6f] = "\xe9\xd4", [0x1c71] = "\x8d\xc8",
  [0x1c74] = "\x8d\xc7", [0x1c76] = "\xe9\xc7", [0x1c7a] = "\x93\xf7",
  [0x1c7e] = "\xc0\xcf", [0x1c7f] = "\xed\x45", [0x1c80] = "\xc0\xc8",
  [0x1c81] = "\xec\xf5", [0x1c82] = "\x8d\xc9", [0x1c83] = "\xed\x41",
  [0x1c84] = "\xc0\xca", [0x1c85] = "\xed\x48", [0x1c87] = "\xec\xfc",
  [0x1c89] = "\xec\xf7", [0x1c8a] = "\xfb\xf2", [0x1c8c] = "\xed\x49",
  [0x1c8d] = "\xec\xf3", [0x1c8e] = "\xec\xfe", [0x1c8f] = "\x96\x70",
  [0x1c90] = "\xc0\xd1", [0x1c91] = "\xed\x44", [0x1c92] = "\xed\x4a",
  [0x1c93] = "\xec\xfd", [0x1c94] = "\xc0\xc9", [0x1c95] = "\xed\x40",
  [0x1c96] = "\xec\xf4", [0x1c97] = "\xc0\xd0", [0x1c99] = "\x8d\xcb",
  [0x1c9a] = "\xed\x47", [0x1c9b] = "\xec\xf9", [0x1c9c] = "\xc0\xcc",
  [0x1c9d] = "\xfd\x5c", [0x1c9e] = "\xec\xfb", [0x1c9f] = "\xec\xf8",
  [0x1ca0] = "\xc0\xd2", [0x1ca1] = "\xec\xfa", [0x1ca2] = "\xc0\xcb",
  [0x1ca3] = "\xc0\xce", [0x1ca4] = "\xed\x43", [0x1ca5] = "\xec\xf6",
  [0x1ca6] = "\xed\x46", [0x1ca7] = "\x8f\x65", [0x1ca8] = "\xed\x42",
  [0x1cab] = "\x8d\xcd", [0x1cac] = "\xc2\x63", [0x1cad] = "\xef\xe7",
  [0x1cae] = "\xc2\x68", [0x1caf] = "\xc2\x69", [0x1cb1] = "\x9d\xa8",
  [0x1cb2] = "\x94\xf9", [0x1cb3] = "\xc2\x62", [0x1cb4] = "\xef\xe6",
  [0x1cb5] = "\x8d\xce", [0x1cb6] = "\xef\xe3", [0x1cb7] = "\xef\xe4",
  [0x1cb8] = "\xc2\x66", [0x1cb9] = "\xef\xde", [0x1cba] = "\xef\xe2",
  [0x1cbb] = "\xc2\x65", [0x1cbd] = "\xef\xdf", [0x1cbe] = "\x93\xea",
  [0x1cc2] = "\xc2\x67", [0x1cc3] = "\xc2\x64", [0x1cc5] = "\xef\xdd",
  [0x1cc6] = "\xef\xe1", [0x1cc7] = "\xef\xe5", [0x1cc8] = "\xfd\x5f",
  [0x1cc9] = "\x93\xf0", [0x1cca] = "\x9f\xb6", [0x1ccb] = "\xf2\x51",
  [0x1ccc] = "\xf2\x4e", [0x1ccd] = "\xf2\x57", [0x1ccf] = "\xf2\x56",
  [0x1cd0] = "\xf2\x54", [0x1cd1] = "\xf2\x4f", [0x1cd3] = "\xc3\x72",
  [0x1cd4] = "\x8d\xcf", [0x1cd8] = "\x97\x63", [0x1cd9] = "\xf2\x50",
  [0x1cda] = "\xc3\x71", [0x1cdb] = "\xc0\xcd", [0x1cdc] = "\xf2\x53",
  [0x1cdd] = "\xc3\x70", [0x1cde] = "\xf2\x58", [0x1cdf] = "\xf2\x52",
  [0x1ce0] = "\xf2\x4d", [0x1ce1] = "\xef\xe0", [0x1ce5] = "\xc3\x6f",
  [0x1ce7] = "\xf2\x4c", [0x1ce8] = "\xf4\x56", [0x1cea] = "\xf4\x55",
  [0x1ceb] = "\xf2\x55", [0x1cec] = "\xc4\x68", [0x1cee] = "\xf4\x59",
  [0x1cef] = "\xf4\x5a", [0x1cf0] = "\xf4\x54", [0x1cf1] = "\xf4\x58",
  [0x1cf3] = "\xf4\x53", [0x1cf6] = "\x8d\xd0", [0x1cf8] = "\xf5\xd1",
  [0x1cf9] = "\xf4\x57", [0x1cfa] = "\xc4\xe7", [0x1cfb] = "\xc4\xe5",
  [0x1cfc] = "\xf5\xcf", [0x1d00] = "\xf5\xd2", [0x1d02] = "\xf5\xce",
  [0x1d03] = "\xf5\xd0", [0x1d04] = "\xc4\xe6", [0x1d05] = "\x93\xf1",
  [0x1d08] = "\xf6\xe5", [0x1d09] = "\xf6\xe6", [0x1d0a] = "\xc5\x76",
  [0x1d0b] = "\xf6\xe4", [0x1d0f] = "\xf7\xe2", [0x1d10] = "\xc5\xcf",
  [0x1d11] = "\xf7\xe0", [0x1d12] = "\xf7\xe1", [0x1d13] = "\xf8\xac",
  [0x1d16] = "\xc6\x56", [0x1d17] = "\xf8\xf3", [0x1d18] = "\xf8\xf1",
  [0x1d19] = "\xf8\xf2", [0x1d1a] = "\xf8\xf4", [0x1d1d] = "\xfd\x62",
  [0x1d1e] = "\xf9\xbb", [0x1d20] = "\xa4\xed", [0x1d21] = "\xa6\xb8",
  [0x1d23] = "\xaa\x59", [0x1d25] = "\xcc\xe9", [0x1d28] = "\xcf\x64",
  [0x1d2c] = "\xd1\xf5", [0x1d2d] = "\xd1\xf7", [0x1d2f] = "\xd1\xf6",
  [0x1d31] = "\xd1\xf8", [0x1d32] = "\xb1\xfd", [0x1d33] = "\xd5\xd7",
  [0x1d34] = "\xd1\xf9", [0x1d35] = "\xfd\x65", [0x1d36] = "\xd5\xd6",
  [0x1d37] = "\xd5\xd8", [0x1d38] = "\xd5\xd9", [0x1d39] = "\xd9\xda",
  [0x1d3a] = "\xb4\xdb", [0x1d3b] = "\xd9\xdb", [0x1d3c] = "\xd9\xdd",
  [0x1d3d] = "\xb4\xdc", [0x1d3e] = "\xb4\xda", [0x1d3f] = "\xd9\xdc",
  [0x1d41] = "\xdd\xfa", [0x1d42] = "\xdd\xf8", [0x1d43] = "\xdd\xf7",
  [0x1d45] = "\xdd\xf6", [0x1d46] = "\xdd\xf5", [0x1d47] = "\xb7\xb2",
  [0x1d48] = "\xdd\xf9", [0x1d49] = "\xba\x70", [0x1d4a] = "\xe2\x63",
  [0x1d4b] = "\xe2\x65", [0x1d4c] = "\xba\x71", [0x1d4d] = "\xe2\x64",
  [0x1d4e] = "\xbc\xdb", [0x1d50] = "\xbc\xda", [0x1d51] = "\xe5\xf0",
  [0x1d52] = "\x9f\xdb", [0x1d54] = "\xe9\xdf", [0x1d55] = "\xe9\xde",
  [0x1d56] = "\xe9\xe0", [0x1d57] = "\x93\xf8", [0x1d59] = "\xbe\xf9",
  [0x1d5b] = "\xed\x4b", [0x1d5c] = "\xc0\xd3", [0x1d5e] = "\xef\xe8",
  [0x1d5f] = "\xc2\x6a", [0x1d60] = "\xf2\x59", [0x1d61] = "\xc5\x77",
  [0x1d62] = "\xa4\xee", [0x1d63] = "\xa5\xbf", [0x1d64] = "\xa6\xb9",
  [0x1d65] = "\xa8\x42", [0x1d66] = "\xaa\x5a", [0x1d67] = "\xaa\x5b",
  [0x1d6a] = "\xac\x6e", [0x1d6d] = "\xd1\xfa", [0x1d6f] = "\x8b\xf7",
  [0x1d72] = "\xb7\xb3", [0x1d74] = "\xfd\x66", [0x1d76] = "\xe6\xd1",
  [0x1d77] = "\xbe\xfa", [0x1d78] = "\xc2\x6b", [0x1d79] = "\xa4\xef",
  [0x1d7a] = "\x8b\xcf", [0x1d7b] = "\xa6\xba", [0x1d7e] = "\xcc\xeb",
  [0x1d7f] = "\xaa\x5c", [0x1d80] = "\xcc\xea", [0x1d81] = "\x8d\xd1",
  [0x1d82] = "\xcf\x65", [0x1d83] = "\xac\x6f", [0x1d84] = "\xcf\x66",
  [0x1d86] = "\xac\x70", [0x1d88] = "\xd1\xfc", [0x1d89] = "\xae\xee",
  [0x1d8a] = "\xae\xed", [0x1d8c] = "\xd5\xde", [0x1d8d] = "\xd5\xdc",
  [0x1d8e] = "\xd5\xdd", [0x1d8f] = "\xd5\xdb", [0x1d91] = "\xd5\xda",
  [0x1d94] = "\xd9\xde", [0x1d95] = "\xd9\xe1", [0x1d96] = "\xb4\xde",
  [0x1d97] = "\xd9\xdf", [0x1d98] = "\xb4\xdd", [0x1d99] = "\xd9\xe0",
  [0x1d9b] = "\xdd\xfb", [0x1d9e] = "\xe2\x66", [0x1d9f] = "\xe2\x67",
  [0x1da0] = "\xe2\x68", [0x1da2] = "\xe5\xf3", [0x1da3] = "\xe5\xf2",
  [0x1da4] = "\xbc\xdc", [0x1da5] = "\xe5\xf1", [0x1da6] = "\xe5\xf4",
  [0x1da7] = "\xe9\xe1", [0x1daa] = "\xe9\xe2", [0x1dab] = "\xe9\xe3",
  [0x1dad] = "\xed\x4c", [0x1dae] = "\xc0\xd4", [0x1daf] = "\xc2\x6c",
  [0x1db0] = "\xf2\x5a", [0x1db2] = "\xc4\xe8", [0x1db3] = "\xc9\x5f",
  [0x1db5] = "\xac\x71", [0x1db6] = "\xcf\x67", [0x1db7] = "\xae\xef",
  [0x1dba] = "\xb1\xfe", [0x1dbc] = "\xb4\xdf", [0x1dbd] = "\xd9\xe2",
  [0x1dbf] = "\xb7\xb5", [0x1dc0] = "\xb7\xb4", [0x1dc1] = "\x8d\xd2",
  [0x1dc3] = "\xe2\x69", [0x1dc4] = "\xe2\x6a", [0x1dc5] = "\xbc\xdd",
  [0x1dc6] = "\xbc\xde", [0x1dc7] = "\xe9\xe5", [0x1dc8] = "\xe9\xe4",
  [0x1dc9] = "\xef\xe9", [0x1dca] = "\xf7\xe3", [0x1dcb] = "\xa4\xf0",
  [0x1dcc] = "\xc9\x60", [0x1dcd] = "\xa5\xc0", [0x1dcf] = "\xa8\x43",
  [0x1dd0] = "\xcb\x48", [0x1dd2] = "\xac\x72", [0x1dd3] = "\xb7\xb6",
  [0x1dd4] = "\xa4\xf1", [0x1dd6] = "\xcf\x68", [0x1dd7] = "\xac\x73",
  [0x1dd8] = "\xcf\x69", [0x1dda] = "\xc0\xd5", [0x1ddb] = "\xa4\xf2",
  [0x1ddc] = "\xfd\x71", [0x1dde] = "\xcc\xec", [0x1de0] = "\xcf\x6a",
  [0x1de1] = "\xfd\x6f", [0x1de2] = "\xd2\x42", [0x1de3] = "\xd2\x41",
  [0x1de4] = "\xd1\xfe", [0x1de6] = "\xd1\xfd", [0x1de7] = "\xd2\x43",
  [0x1de8] = "\xd2\x40", [0x1dea] = "\x8d\xd3", [0x1deb] = "\xb2\x40",
  [0x1dec] = "\xb2\x41", [0x1def] = "\xb4\xe0", [0x1df0] = "\xd9\xe3",
  [0x1df2] = "\xd9\xe4", [0x1df3] = "\xd9\xe5", [0x1df7] = "\xde\x41",
  [0x1df8] = "\xde\x42", [0x1df9] = "\xde\x40", [0x1dfa] = "\x9f\xe7",
  [0x1dfb] = "\xdd\xfd", [0x1dfc] = "\xdd\xfe", [0x1dfd] = "\xb7\xb7",
  [0x1dfe] = "\xe2\x6b", [0x1dff] = "\xe5\xf7", [0x1e00] = "\xe5\xf6",
  [0x1e01] = "\xe5\xf5", [0x1e02] = "\xe5\xf8", [0x1e03] = "\xe9\xe7",
  [0x1e04] = "\xe9\xe6", [0x1e05] = "\xbe\xfb", [0x1e06] = "\xe9\xe8",
  [0x1e08] = "\xc0\xd6", [0x1e09] = "\xed\x4d", [0x1e0b] = "\xef\xea",
  [0x1e0c] = "\xf2\x5b", [0x1e0d] = "\xf6\xe7", [0x1e0f] = "\xa4\xf3",
  [0x1e10] = "\xa5\xc2", [0x1e11] = "\xa5\xc1", [0x1e13] = "\xaa\x5d",
  [0x1e14] = "\xc9\x61", [0x1e15] = "\xc9\x7e", [0x1e16] = "\xa6\xbb",
  [0x1e18] = "\xc9\xf7", [0x1e19] = "\xcb\x49", [0x1e1a] = "\xcb\x4a",
  [0x1e1b] = "\xaa\x5e", [0x1e1c] = "\x90\xbd", [0x1e1d] = "\xcc\xed",
  [0x1e1f] = "\xac\x74", [0x1e20] = "\xcf\x6b", [0x1e21] = "\xcf\x6c",
  [0x1e23] = "\xae\xf0", [0x1e24] = "\xae\xf4", [0x1e25] = "\xd2\x44",
  [0x1e26] = "\xae\xf3", [0x1e27] = "\xae\xf1", [0x1e28] = "\xae\xf2",
  [0x1e2a] = "\xd5\xdf", [0x1e2b] = "\xb2\x42", [0x1e2c] = "\xb4\xe3",
  [0x1e2e] = "\xb4\xe1", [0x1e2f] = "\xb4\xe2", [0x1e30] = "\xd9\xe6",
  [0x1e31] = "\x9f\xd0", [0x1e33] = "\xba\x72", [0x1e34] = "\xa4\xf4",
  [0x1e35] = "\x8b\xd0", [0x1e36] = "\xc9\xa1", [0x1e37] = "\xfd\x72",
  [0x1e38] = "\xa5\xc3", [0x1e39] = "\x9c\xae", [0x1e3a] = "\x8b\xd1",
  [0x1e3b] = "\xc9\xa4", [0x1e3d] = "\x8a\xdb", [0x1e3e] = "\xa5\xc6",
  [0x1e3f] = "\xc9\xa3", [0x1e40] = "\xa5\xc5", [0x1e41] = "\xa5\xc4",
  [0x1e42] = "\xa8\x44", [0x1e43] = "\xc9\xa2", [0x1e46] = "\xc9\xf8",
  [0x1e49] = "\xfa\xe4", [0x1e4a] = "\xc9\xfc", [0x1e4b] = "\xc9\xfe",
  [0x1e4c] = "\xca\x40", [0x1e4d] = "\xa6\xc5", [0x1e4e] = "\xa6\xc6",
  [0x1e4f] = "\xc9\xfb", [0x1e50] = "\xa6\xc1", [0x1e52] = "\xc9\xf9",
  [0x1e54] = "\xc9\xfd", [0x1e55] = "\xa6\xc2", [0x1e57] = "\xa6\xbd",
  [0x1e58] = "\x95\xce", [0x1e59] = "\xa6\xbe", [0x1e5a] = "\xfd\x76",
  [0x1e5b] = "\xa6\xc4", [0x1e5c] = "\xc9\xfa", [0x1e5d] = "\xa6\xbc",
  [0x1e5e] = "\xa8\x45", [0x1e5f] = "\xa6\xbf", [0x1e60] = "\xa6\xc0",
  [0x1e61] = "\xa6\xc3", [0x1e65] = "\xcb\x5b", [0x1e66] = "\xcb\x59",
  [0x1e67] = "\xcb\x4c", [0x1e68] = "\xa8\x51", [0x1e69] = "\xcb\x53",
  [0x1e6a] = "\xa8\x4c", [0x1e6b] = "\xcb\x4d", [0x1e6d] = "\xcb\x55",
  [0x1e6e] = "\xfb\x62", [0x1e6f] = "\xcb\x52", [0x1e70] = "\xa8\x4f",
  [0x1e71] = "\xcb\x51", [0x1e72] = "\xa8\x56", [0x1e73] = "\xcb\x5a",
  [0x1e74] = "\xa8\x58", [0x1e75] = "\x8d\xd4", [0x1e76] = "\xa8\x5a",
  [0x1e78] = "\xcb\x4b", [0x1e79] = "\xfd\x78", [0x1e7a] = "\xa8\x4d",
  [0x1e7b] = "\xcb\x5c", [0x1e7d] = "\xa8\x54", [0x1e7e] = "\xa8\x57",
  [0x1e7f] = "\x8e\xe3", [0x1e80] = "\xcd\x45", [0x1e81] = "\xa8\x47",
  [0x1e82] = "\xa8\x5e", [0x1e83] = "\xa8\x55", [0x1e84] = "\xcb\x4e",
  [0x1e85] = "\xa8\x4a", [0x1e86] = "\xa8\x59", [0x1e87] = "\xcb\x56",
  [0x1e88] = "\xa8\x48", [0x1e89] = "\xa8\x49", [0x1e8a] = "\xcd\x43",
  [0x1e8b] = "\xcb\x4f", [0x1e8c] = "\xa8\x50", [0x1e8d] = "\xa8\x5b",
  [0x1e8e] = "\xcb\x5d", [0x1e8f] = "\xcb\x50", [0x1e90] = "\xa8\x4e",
  [0x1e92] = "\xa8\x53", [0x1e93] = "\xcc\xee", [0x1e94] = "\xa8\x5c",
  [0x1e95] = "\xcb\x57", [0x1e96] = "\xa8\x52", [0x1e98] = "\xa8\x5d",
  [0x1e99] = "\xa8\x46", [0x1e9a] = "\xcb\x54", [0x1e9b] = "\xa8\x4b",
  [0x1e9c] = "\xcb\x58", [0x1e9d] = "\xcd\x44", [0x1e9f] = "\x90\x76",
  [0x1ea2] = "\x98\xc6", [0x1eaa] = "\x8d\xd5", [0x1eab] = "\xaa\x6a",
  [0x1eac] = "\xaa\x7a", [0x1ead] = "\xcc\xf5", [0x1eae] = "\xaa\x71",
  [0x1eaf] = "\x97\xd1", [0x1eb0] = "\xcd\x4b", [0x1eb1] = "\xaa\x62",
  [0x1eb2] = "\x9e\xb6", [0x1eb3] = "\xaa\x65", [0x1eb4] = "\xcd\x42",
  [0x1eb6] = "\xcc\xf3", [0x1eb7] = "\xcc\xf7", [0x1eb8] = "\xaa\x6d",
  [0x1eb9] = "\xaa\x6f", [0x1eba] = "\xcc\xfa", [0x1ebb] = "\xaa\x76",
  [0x1ebc] = "\xaa\x68", [0x1ebd] = "\xaa\x66", [0x1ebe] = "\xaa\x67",
  [0x1ebf] = "\xaa\x75", [0x1ec0] = "\xcd\x47", [0x1ec1] = "\xaa\x70",
  [0x1ec2] = "\xcc\xf9", [0x1ec3] = "\xcc\xfb", [0x1ec4] = "\xaa\x6e",
  [0x1ec5] = "\xaa\x73", [0x1ec6] = "\xcc\xfc", [0x1ec7] = "\xcd\x4a",
  [0x1ec9] = "\xac\x75", [0x1eca] = "\xaa\x79", [0x1ecb] = "\xfa\xc7",
  [0x1ecc] = "\xaa\x63", [0x1ecd] = "\xcd\x49", [0x1ece] = "\xa0\x42",
  [0x1ecf] = "\xcd\x4d", [0x1ed0] = "\xcc\xf8", [0x1ed1] = "\xcd\x4f",
  [0x1ed2] = "\xcd\x40", [0x1ed3] = "\xaa\x6c", [0x1ed4] = "\xcc\xf4",
  [0x1ed5] = "\xaa\x6b", [0x1ed6] = "\xaa\x7d", [0x1ed7] = "\xaa\x72",
  [0x1ed9] = "\xcc\xf2", [0x1eda] = "\xcf\x75", [0x1edb] = "\xaa\x78",
  [0x1edc] = "\xaa\x7c", [0x1edd] = "\xcd\x41", [0x1ede] = "\xcd\x46",
  [0x1edf] = "\x98\x73", [0x1ee0] = "\xaa\x7e", [0x1ee1] = "\xaa\x77",
  [0x1ee2] = "\xaa\x69", [0x1ee3] = "\xaa\x5f", [0x1ee5] = "\xaa\x64",
  [0x1ee7] = "\xcc\xf6", [0x1ee8] = "\xaa\x60", [0x1ee9] = "\xcd\x4e",
  [0x1eea] = "\x9f\xfc", [0x1eeb] = "\xcc\xf0", [0x1eec] = "\xcc\xef",
  [0x1eed] = "\xcc\xfd", [0x1eee] = "\xcc\xf1", [0x1eef] = "\xaa\x7b",
  [0x1ef0] = "\xae\xf5", [0x1ef1] = "\xaa\x74", [0x1ef2] = "\xcc\xfe",
  [0x1ef3] = "\xaa\x61", [0x1ef5] = "\xac\xa6", [0x1ef9] = "\xcd\x4c",
  [0x1f00] = "\xcf\x7c", [0x1f01] = "\xcf\xa1", [0x1f02] = "\x8d\xd7",
  [0x1f03] = "\xcf\xa4", [0x1f04] = "\xcf\x77", [0x1f05] = "\x92\xfb",
  [0x1f06] = "\x8d\xd8", [0x1f07] = "\xcf\xa7", [0x1f08] = "\xcf\xaa",
  [0x1f09] = "\xcf\xac", [0x1f0a] = "\xcf\x74", [0x1f0b] = "\xac\x76",
  [0x1f0c] = "\xac\x7b", [0x1f0d] = "\xd2\x49", [0x1f0e] = "\xac\xad",
  [0x1f0f] = "\xcf\xa5", [0x1f10] = "\xcf\xad", [0x1f11] = "\xcf\x7b",
  [0x1f12] = "\xcf\x73", [0x1f16] = "\xd2\x64", [0x1f17] = "\xac\x7e",
  [0x1f18] = "\xcf\xa2", [0x1f19] = "\xcf\x78", [0x1f1a] = "\xcf\x7a",
  [0x1f1b] = "\xac\xa5", [0x1f1d] = "\xcf\x7d", [0x1f1e] = "\xac\x7d",
  [0x1f1f] = "\xcf\x70", [0x1f20] = "\xcf\xa8", [0x1f22] = "\xcf\xab",
  [0x1f24] = "\x94\x4f", [0x1f25] = "\xac\x7a", [0x1f26] = "\x8d\xd9",
  [0x1f27] = "\xac\xa8", [0x1f28] = "\xcf\x6d", [0x1f29] = "\xac\xaa",
  [0x1f2a] = "\xac\x78", [0x1f2b] = "\xac\xae", [0x1f2c] = "\xcf\xa9",
  [0x1f2d] = "\xcf\x6f", [0x1f2e] = "\xac\xab", [0x1f2f] = "\xd2\x5e",
  [0x1f30] = "\xcd\x48", [0x1f31] = "\xac\x7c", [0x1f32] = "\xac\x77",
  [0x1f33] = "\xcf\x76", [0x1f34] = "\xcf\x6e", [0x1f35] = "\xac\xac",
  [0x1f36] = "\xac\xa4", [0x1f37] = "\xcf\xa3", [0x1f38] = "\xac\xa9",
  [0x1f39] = "\xac\xa7", [0x1f3a] = "\xcf\x79", [0x1f3b] = "\xac\xa1",
  [0x1f3c] = "\xcf\x71", [0x1f3d] = "\xac\xa2", [0x1f3e] = "\xac\xa3",
  [0x1f3f] = "\xcf\x72", [0x1f40] = "\xcf\xa6", [0x1f41] = "\xac\x79",
  [0x1f42] = "\xcf\x7e", [0x1f4e] = "\x89\x6b", [0x1f57] = "\x97\xce",
  [0x1f58] = "\xd2\x4c", [0x1f59] = "\xae\xfd", [0x1f5a] = "\xaf\x43",
  [0x1f5b] = "\xfa\xf3", [0x1f5c] = "\xfd\xae", [0x1f5e] = "\xd2\x55",
  [0x1f5f] = "\xd2\x5b", [0x1f60] = "\xd2\x57", [0x1f61] = "\xd2\x4a",
  [0x1f62] = "\xd2\x4d", [0x1f63] = "\xd2\x46", [0x1f64] = "\xd2\x47",
  [0x1f65] = "\xaf\x4a", [0x1f66] = "\xae\xfa", [0x1f67] = "\xd2\x56",
  [0x1f68] = "\xd2\x5f", [0x1f69] = "\xaf\x45", [0x1f6a] = "\xae\xf6",
  [0x1f6c] = "\xaf\x40", [0x1f6d] = "\xd2\x4e", [0x1f6e] = "\xaf\x42",
  [0x1f6f] = "\xd2\x4f", [0x1f70] = "\xd2\x59", [0x1f71] = "\xfb\xaf",
  [0x1f72] = "\x92\xb7", [0x1f74] = "\xaf\x44", [0x1f75] = "\xd2\x68",
  [0x1f76] = "\xd2\x48", [0x1f77] = "\xae\xfc", [0x1f78] = "\xae\xfb",
  [0x1f79] = "\xaf\x48", [0x1f7a] = "\xd2\x45", [0x1f7b] = "\xd2\x66",
  [0x1f7c] = "\xd2\x5a", [0x1f7d] = "\xd2\x67", [0x1f7e] = "\xd2\x61",
  [0x1f7f] = "\xd2\x53", [0x1f80] = "\xd2\x62", [0x1f81] = "\x8d\xda",
  [0x1f82] = "\xd2\x5c", [0x1f83] = "\xd2\x65", [0x1f84] = "\xd2\x63",
  [0x1f85] = "\xaf\x49", [0x1f86] = "\xd2\x54", [0x1f87] = "\xae\xf9",
  [0x1f88] = "\xae\xf8", [0x1f89] = "\xaf\x41", [0x1f8a] = "\xaf\x47",
  [0x1f8b] = "\xd2\x60", [0x1f8c] = "\xaf\x46", [0x1f8d] = "\xd2\x51",
  [0x1f8e] = "\xb2\x43", [0x1f8f] = "\x9c\x5a", [0x1f90] = "\xd2\x69",
  [0x1f91] = "\xd2\x50", [0x1f92] = "\xd2\x4b", [0x1f93] = "\xae\xfe",
  [0x1f94] = "\xaf\x4b", [0x1f95] = "\xae\xf7", [0x1f96] = "\xfd\xad",
  [0x1f97] = "\xd2\x58", [0x1f98] = "\xd2\x5d", [0x1fa4] = "\x8d\xdc",
  [0x1fa5] = "\x94\x44", [0x1faa] = "\xb2\x65", [0x1fab] = "\xd5\xe1",
  [0x1fac] = "\xd5\xe5", [0x1fae] = "\xb2\x52", [0x1faf] = "\xb2\x50",
  [0x1fb1] = "\x8d\xdd", [0x1fb2] = "\xb2\x47", [0x1fb3] = "\xd5\xe3",
  [0x1fb4] = "\xd5\xe2", [0x1fb5] = "\xb2\x5b", [0x1fb7] = "\xd5\xe8",
  [0x1fb8] = "\xb2\x55", [0x1fb9] = "\xa0\xd6", [0x1fba] = "\xd5\xfa",
  [0x1fbb] = "\xd6\x47", [0x1fbc] = "\xb2\x44", [0x1fbd] = "\xd5\xf7",
  [0x1fbe] = "\xd5\xf0", [0x1fbf] = "\xb2\x67", [0x1fc0] = "\xd5\xe0",
  [0x1fc2] = "\xd5\xfc", [0x1fc4] = "\xb2\x64", [0x1fc5] = "\xb2\x58",
  [0x1fc6] = "\xb2\x63", [0x1fc7] = "\xb2\x4e", [0x1fc8] = "\xd5\xec",
  [0x1fc9] = "\xd5\xfe", [0x1fca] = "\xd5\xf6", [0x1fcb] = "\xb2\x4f",
  [0x1fcc] = "\xb2\x49", [0x1fcd] = "\xd6\x45", [0x1fcf] = "\xd5\xfd",
  [0x1fd0] = "\xd6\x40", [0x1fd1] = "\xb2\x51", [0x1fd2] = "\xb2\x59",
  [0x1fd3] = "\xd6\x42", [0x1fd4] = "\xd5\xea", [0x1fd5] = "\xd5\xfb",
  [0x1fd6] = "\xd5\xef", [0x1fd7] = "\xd6\x44", [0x1fd8] = "\xb2\x5e",
  [0x1fd9] = "\xb2\x46", [0x1fda] = "\xb2\x5c", [0x1fdb] = "\xd5\xf4",
  [0x1fdc] = "\xd5\xf2", [0x1fdd] = "\xd5\xf3", [0x1fde] = "\xb2\x53",
  [0x1fdf] = "\xd5\xee", [0x1fe0] = "\xd5\xed", [0x1fe1] = "\xb2\x48",
  [0x1fe2] = "\xd5\xe7", [0x1fe3] = "\xd6\x46", [0x1fe4] = "\xb2\x4a",
  [0x1fe5] = "\xd5\xf1", [0x1fe6] = "\xb2\x68", [0x1fe8] = "\xb2\x62",
  [0x1fe9] = "\xd5\xe6", [0x1fea] = "\xb2\x5f", [0x1feb] = "\xb2\x5d",
  [0x1fec] = "\xb2\x66", [0x1fed] = "\xd5\xf8", [0x1fee] = "\xb2\x61",
  [0x1fef] = "\xd2\x52", [0x1ff0] = "\xd5\xf9", [0x1ff1] = "\xb2\x60",
  [0x1ff2] = "\xd6\x41", [0x1ff3] = "\xb2\x45", [0x1ff4] = "\xd5\xf5",
  [0x1ff5] = "\xb2\x57", [0x1ff6] = "\xd5\xe9", [0x1ff7] = "\xb2\x56",
  [0x1ff9] = "\xb2\x54", [0x1ffa] = "\xb2\x4c", [0x1ffb] = "\xb2\x4b",
  [0x1ffc] = "\xd9\xe7", [0x1ffd] = "\xd6\x43", [0x2000] = "\xd5\xeb",
  [0x2002] = "\x97\xd5", [0x2003] = "\xd9\xfc", [0x2004] = "\x94\x4a",
  [0x2005] = "\xb2\x4d", [0x200a] = "\x94\x4d", [0x200f] = "\x97\xcb",
  [0x2015] = "\x8d\xde", [0x2018] = "\x8d\xdf", [0x2019] = "\xb5\x41",
  [0x201a] = "\xb2\x5a", [0x201b] = "\xb4\xee", [0x201c] = "\xd9\xf6",
  [0x201d] = "\xb4\xfc", [0x201f] = "\xd9\xea", [0x2020] = "\xb4\xeb",
  [0x2021] = "\xb4\xe7", [0x2022] = "\xda\x49", [0x2023] = "\xb4\xed",
  [0x2024] = "\xb4\xf1", [0x2025] = "\xb4\xec", [0x2026] = "\xb4\xf5",
  [0x2027] = "\xda\x4d", [0x2028] = "\xda\x44", [0x2029] = "\x8d\xe0",
  [0x202a] = "\xfe\xf9", [0x202b] = "\xd9\xf1", [0x202c] = "\xb4\xfa",
  [0x202d] = "\xb4\xf4", [0x202e] = "\xd9\xfd", [0x202f] = "\xb4\xe4",
  [0x2030] = "\xda\x4a", [0x2031] = "\xda\x43", [0x2032] = "\xb4\xe8",
  [0x2033] = "\xd9\xf7", [0x2034] = "\xb4\xf7", [0x2035] = "\xda\x55",
  [0x2036] = "\xda\x56", [0x2038] = "\xb4\xe5", [0x2039] = "\xda\x48",
  [0x203a] = "\xb4\xf9", [0x203b] = "\xd9\xfb", [0x203c] = "\xd9\xed",
  [0x203d] = "\xd9\xee", [0x203e] = "\xb4\xfd", [0x203f] = "\xd9\xf2",
  [0x2040] = "\xd9\xf9", [0x2041] = "\xd9\xf3", [0x2043] = "\xb4\xfb",
  [0x2044] = "\xb5\x44", [0x2045] = "\xd9\xef", [0x2046] = "\xd9\xe8",
  [0x2047] = "\xd9\xe9", [0x2049] = "\xd9\xeb", [0x204a] = "\xb4\xea",
  [0x204b] = "\xd9\xf8", [0x204d] = "\xb4\xf8", [0x204e] = "\xb5\x42",
  [0x204f] = "\xfd\xc0", [0x2050] = "\xfc\xf9", [0x2051] = "\xd9\xfa",
  [0x2052] = "\xda\x53", [0x2053] = "\xda\x4b", [0x2054] = "\xb4\xe6",
  [0x2055] = "\xda\x51", [0x2056] = "\xb4\xf2", [0x2058] = "\xb4\xf0",
  [0x2059] = "\xfb\x7e", [0x205a] = "\xda\x57", [0x205b] = "\xb4\xef",
  [0x205c] = "\xda\x41", [0x205d] = "\xd9\xf4", [0x205e] = "\xd9\xfe",
  [0x205f] = "\xb5\x47", [0x2060] = "\xda\x45", [0x2061] = "\xda\x42",
  [0x2062] = "\xd9\xf0", [0x2063] = "\xb5\x43", [0x2064] = "\xda\x4f",
  [0x2065] = "\xda\x4c", [0x2066] = "\xda\x54", [0x2067] = "\xb4\xe9",
  [0x2068] = "\xda\x40", [0x2069] = "\xb5\x46", [0x206b] = "\xda\x47",
  [0x206e] = "\xb4\xf3", [0x206f] = "\xb4\xf6", [0x2071] = "\xda\x46",
  [0x2072] = "\xb5\x45", [0x2073] = "\xd9\xf5", [0x2074] = "\xd5\xe4",
  [0x2076] = "\x92\xb3", [0x2077] = "\xda\x50", [0x2078] = "\xda\x4e",
  [0x2079] = "\xda\x52", [0x207c] = "\xfd\xaf", [0x2086] = "\x8d\xe1",
  [0x2088] = "\xd9\xec", [0x2089] = "\xb5\x40", [0x208b] = "\x95\xd3",
  [0x208d] = "\xde\x61", [0x208e] = "\xde\x60", [0x208f] = "\xde\x46",
  [0x2090] = "\xb7\xbd", [0x2092] = "\xde\x5f", [0x2093] = "\xde\x49",
  [0x2094] = "\xde\x4a", [0x2096] = "\xb7\xc7", [0x2097] = "\xde\x68",
  [0x2098] = "\xb7\xc2", [0x2099] = "\xde\x5e", [0x209a] = "\x89\xc1",
  [0x209b] = "\xde\x43", [0x209c] = "\xb7\xc8", [0x209d] = "\xb7\xbe",
  [0x209e] = "\xde\x52", [0x209f] = "\xde\x48", [0x20a0] = "\xde\x4b",
  [0x20a1] = "\xde\x63", [0x20a2] = "\xb7\xb8", [0x20a3] = "\xde\x6a",
  [0x20a4] = "\xde\x62", [0x20a5] = "\xb7\xc1", [0x20a6] = "\xde\x57",
  [0x20a7] = "\xb7\xcc", [0x20aa] = "\xb7\xcb", [0x20ab] = "\xb7\xc5",
  [0x20ae] = "\xde\x69", [0x20af] = "\xb7\xb9", [0x20b0] = "\xde\x55",
  [0x20b1] = "\xde\x4c", [0x20b2] = "\xde\x59", [0x20b3] = "\xde\x65",
  [0x20b4] = "\xb7\xcd", [0x20b5] = "\xfd\x68", [0x20b6] = "\xb7\xbb",
  [0x20b7] = "\xde\x54", [0x20b8] = "\x9c\xb7", [0x20b9] = "\xde\x4d",
  [0x20ba] = "\xb7\xc4", [0x20bb] = "\x8d\xe3", [0x20bc] = "\xb7\xc3",
  [0x20bd] = "\xde\x50", [0x20be] = "\xde\x5a", [0x20bf] = "\xde\x64",
  [0x20c0] = "\xde\x47", [0x20c1] = "\xde\x51", [0x20c2] = "\xb7\xbc",
  [0x20c3] = "\xde\x5b", [0x20c4] = "\xb7\xc9", [0x20c5] = "\xb7\xc0",
  [0x20c6] = "\xde\x4e", [0x20c7] = "\xb7\xbf", [0x20c8] = "\xde\x45",
  [0x20c9] = "\xde\x53", [0x20ca] = "\xde\x67", [0x20cb] = "\xb4\xfe",
  [0x20cc] = "\xba\xb0", [0x20cd] = "\xde\x56", [0x20ce] = "\xe2\x6c",
  [0x20cf] = "\xde\x58", [0x20d0] = "\xde\x66", [0x20d1] = "\xb7\xc6",
  [0x20d2] = "\xde\x4f", [0x20d3] = "\xb7\xba", [0x20d4] = "\xb7\xca",
  [0x20d5] = "\xbc\xf0", [0x20d6] = "\xde\x44", [0x20d8] = "\xde\x5d",
  [0x20d9] = "\xfa\xc0", [0x20da] = "\x8d\xe5", [0x20db] = "\xfa\x64",
  [0x20dc] = "\xde\x5c", [0x20dd] = "\x89\x47", [0x20e2] = "\x8d\xe4",
  [0x20e8] = "\x8d\xe7", [0x20e9] = "\x8d\xe8", [0x20eb] = "\xe2\xaa",
  [0x20ec] = "\xba\xad", [0x20ed] = "\xe2\x7d", [0x20ee] = "\xe2\xa4",
  [0x20ef] = "\xba\xa2", [0x20f1] = "\xe2\x6e", [0x20f2] = "\xba\xaf",
  [0x20f4] = "\xba\x77", [0x20f5] = "\xe2\x6d", [0x20f6] = "\xe2\xb0",
  [0x20f7] = "\xba\xb1", [0x20f8] = "\xe2\x71", [0x20f9] = "\xe2\xa3",
  [0x20fa] = "\xfd\xc7", [0x20fb] = "\xe2\x73", [0x20fc] = "\xe2\xb3",
  [0x20fd] = "\xe2\xaf", [0x20fe] = "\xba\x75", [0x20ff] = "\xba\xa1",
  [0x2100] = "\xe6\x53", [0x2101] = "\xba\xae", [0x2102] = "\xba\x7d",
  [0x2103] = "\xe2\x6f", [0x2104] = "\xfd\xb0", [0x2105] = "\xe2\xae",
  [0x2106] = "\xba\xa3", [0x2107] = "\xe2\xab", [0x2108] = "\xe2\xb8",
  [0x2109] = "\xe2\x75", [0x210a] = "\xe2\x7e", [0x210b] = "\x94\x45",
  [0x210c] = "\x97\xd6", [0x210d] = "\xe2\xb6", [0x210e] = "\xe2\xac",
  [0x210f] = "\xba\x7c", [0x2112] = "\xe2\x7c", [0x2113] = "\xba\x76",
  [0x2114] = "\xba\x74", [0x2115] = "\xba\xa8", [0x2116] = "\xfc\xc6",
  [0x2117] = "\x98\x44", [0x2118] = "\xe2\x7a", [0x2119] = "\xe2\x77",
  [0x211a] = "\xe2\x78", [0x211c] = "\xe2\xb2", [0x211e] = "\xe2\xb7",
  [0x211f] = "\xe2\xb5", [0x2120] = "\xba\x7a", [0x2121] = "\xe2\xb9",
  [0x2122] = "\xba\x7e", [0x2123] = "\xba\xa7", [0x2124] = "\x8d\xe9",
  [0x2125] = "\xe2\x70", [0x2126] = "\xe5\xfa", [0x2127] = "\xe2\x79",
  [0x2129] = "\xba\x78", [0x212a] = "\xba\xac", [0x212b] = "\xba\xa9",
  [0x212c] = "\xba\x7b", [0x212d] = "\xe2\xa5", [0x212e] = "\xe2\x74",
  [0x212f] = "\xba\xaa", [0x2130] = "\xe2\xa7", [0x2131] = "\xba\xa4",
  [0x2132] = "\xba\xa6", [0x2133] = "\xba\x73", [0x2134] = "\x8d\xea",
  [0x2135] = "\xe2\xa9", [0x2136] = "\xe2\xa1", [0x2137] = "\xe2\x72",
  [0x2138] = "\xba\xa5", [0x2139] = "\xe2\xb1", [0x213a] = "\xe2\xb4",
  [0x213b] = "\xe2\x7b", [0x213c] = "\xe2\xa8", [0x213d] = "\xfe\x50",
  [0x213e] = "\xba\x79", [0x213f] = "\xbc\xdf", [0x2140] = "\xe2\xa6",
  [0x2141] = "\xe5\xf9", [0x2143] = "\xe2\xad", [0x2144] = "\xfd\xcc",
  [0x214e] = "\xe2\x76", [0x214f] = "\xe6\x44", [0x2150] = "\xe6\x4e",
  [0x2151] = "\xbc\xe2", [0x2152] = "\xe6\x4d", [0x2153] = "\xe6\x59",
  [0x2154] = "\xbc\xe4", [0x2155] = "\xe6\x4b", [0x2156] = "\x9d\xa7",
  [0x2157] = "\xe6\x4f", [0x2158] = "\xbc\xef", [0x215a] = "\xe6\x46",
  [0x215b] = "\xbc\xe7", [0x215c] = "\xfd\xcd", [0x215d] = "\xe6\x52",
  [0x215e] = "\xe9\xf0", [0x215f] = "\xbc\xf3", [0x2160] = "\xbc\xf2",
  [0x2161] = "\xe6\x54", [0x2162] = "\xe6\x43", [0x2163] = "\xe6\x5e",
  [0x2164] = "\xbc\xed", [0x2166] = "\xbc\xe3", [0x2167] = "\xe6\x57",
  [0x2169] = "\xe6\x5b", [0x216a] = "\xe6\x60", [0x216b] = "\xe6\x55",
  [0x216c] = "\xe6\x49", [0x216d] = "\xbc\xe6", [0x216e] = "\xbc\xe9",
  [0x216f] = "\xbc\xf1", [0x2170] = "\xbc\xec", [0x2172] = "\xe6\x4c",
  [0x2173] = "\xe2\xa2", [0x2174] = "\xfd\xcf", [0x2176] = "\xe6\x48",
  [0x2177] = "\xe6\x5f", [0x2178] = "\xbc\xe8", [0x2179] = "\x95\xd2",
  [0x217a] = "\xbc\xeb", [0x217b] = "\xe6\x61", [0x217c] = "\xbc\xe0",
  [0x217d] = "\xe6\x56", [0x217e] = "\xe5\xfb", [0x217f] = "\xe6\x5c",
  [0x2180] = "\xc0\xdf", [0x2181] = "\x8d\xed", [0x2182] = "\xe6\x4a",
  [0x2184] = "\xbc\xe1", [0x2185] = "\xe6\x45", [0x2186] = "\xbc\xe5",
  [0x2187] = "\xe5\xfc", [0x2188] = "\xba\xab", [0x2189] = "\xe6\x41",
  [0x218a] = "\xfc\xba", [0x218b] = "\xe6\x5a", [0x218c] = "\xe6\x42",
  [0x218d] = "\xe6\x40", [0x218e] = "\xbc\xea", [0x2190] = "\xe6\x58",
  [0x2192] = "\xe5\xfe", [0x2193] = "\xe6\x51", [0x2194] = "\xe6\x50",
  [0x2195] = "\xe6\x5d", [0x2196] = "\xe6\x47", [0x2197] = "\xbc\xee",
  [0x219d] = "\xfd\xc5", [0x219e] = "\xe9\xf3", [0x219f] = "\xfd\xd2",
  [0x21a0] = "\xbf\x49", [0x21a1] = "\xbe\xfe", [0x21a2] = "\xea\x40",
  [0x21a3] = "\xe9\xeb", [0x21a4] = "\xbf\x41", [0x21a5] = "\xe9\xf7",
  [0x21a6] = "\xbf\x48", [0x21a7] = "\xbf\x43", [0x21a8] = "\xe9\xf5",
  [0x21a9] = "\xed\x4f", [0x21aa] = "\xe9\xfb", [0x21ab] = "\xea\x42",
  [0x21ac] = "\xe9\xfa", [0x21ad] = "\xe9\xe9", [0x21ae] = "\xe9\xf8",
  [0x21af] = "\xea\x44", [0x21b0] = "\xea\x46", [0x21b1] = "\xbe\xfd",
  [0x21b2] = "\xea\x45", [0x21b3] = "\xbf\x44", [0x21b4] = "\xbf\x4a",
  [0x21b5] = "\x9c\xdc", [0x21b6] = "\xbf\x47", [0x21b8] = "\xe9\xfe",
  [0x21b9] = "\xbf\x46", [0x21ba] = "\xe9\xf9", [0x21bb] = "\x95\xcf",
  [0x21bc] = "\xe9\xed", [0x21bd] = "\xe9\xf2", [0x21be] = "\x8d\xee",
  [0x21bf] = "\xe9\xfd", [0x21c0] = "\xbf\x45", [0x21c1] = "\xbf\x42",
  [0x21c2] = "\xbe\xfc", [0x21c3] = "\xbf\x40", [0x21c4] = "\xe9\xf1",
  [0x21c6] = "\xe5\xfd", [0x21c7] = "\xe9\xec", [0x21c8] = "\xe9\xef",
  [0x21c9] = "\xea\x41", [0x21ca] = "\xe9\xf4", [0x21cb] = "\xe9\xea",
  [0x21cc] = "\xed\x4e", [0x21cd] = "\xea\x43", [0x21ce] = "\xe9\xee",
  [0x21cf] = "\xe9\xfc", [0x21d3] = "\xfd\xd4", [0x21d4] = "\xed\x51",
  [0x21d5] = "\xc0\xe3", [0x21d8] = "\xc0\xd7", [0x21d9] = "\x96\xec",
  [0x21da] = "\x96\xeb", [0x21db] = "\xc0\xdb", [0x21dc] = "\xed\x53",
  [0x21dd] = "\xed\x59", [0x21de] = "\xed\x57", [0x21df] = "\xc0\xd9",
  [0x21e0] = "\xc0\xda", [0x21e1] = "\xc0\xe1", [0x21e2] = "\xed\x5a",
  [0x21e3] = "\xed\x52", [0x21e4] = "\xc0\xdc", [0x21e6] = "\xed\x56",
  [0x21e7] = "\xed\x55", [0x21e8] = "\xed\x5b", [0x21e9] = "\xc0\xe2",
  [0x21eb] = "\xc0\xdd", [0x21ec] = "\xc0\xe0", [0x21ed] = "\xed\x54",
  [0x21ee] = "\xc0\xe4", [0x21ef] = "\xc0\xde", [0x21f0] = "\xc0\xe5",
  [0x21f1] = "\xc0\xd8", [0x21f2] = "\xed\x58", [0x21f4] = "\xed\x50",
  [0x21f6] = "\x90\xb6", [0x21f7] = "\xef\xf7", [0x21f8] = "\xfd\xc3",
  [0x21fa] = "\xc2\x71", [0x21fb] = "\xef\xf4", [0x21fc] = "\xef\xf6",
  [0x21fe] = "\xc2\x6f", [0x21ff] = "\xef\xf2", [0x2200] = "\xef\xf3",
  [0x2201] = "\xef\xee", [0x2203] = "\x98\xab", [0x2204] = "\xe9\xf6",
  [0x2205] = "\xef\xef", [0x2206] = "\xc2\x70", [0x2207] = "\xef\xeb",
  [0x2209] = "\xc2\x6d", [0x220a] = "\xef\xf8", [0x220b] = "\xc2\x6e",
  [0x220c] = "\xef\xec", [0x220d] = "\xef\xed", [0x220e] = "\xef\xf1",
  [0x220f] = "\xc2\x73", [0x2211] = "\xc2\x72", [0x2214] = "\xef\xf0",
  [0x2215] = "\xc3\x78", [0x2216] = "\xf2\x5f", [0x2217] = "\xf2\x65",
  [0x2218] = "\xc3\x79", [0x2219] = "\xf2\x5c", [0x221a] = "\xc3\x76",
  [0x221b] = "\xc3\x73", [0x221c] = "\xf2\x67", [0x221d] = "\xc3\x77",
  [0x221e] = "\x96\xee", [0x221f] = "\xc3\x74", [0x2220] = "\xf2\x5e",
  [0x2221] = "\xf2\x61", [0x2222] = "\xf2\x62", [0x2223] = "\xf2\x63",
  [0x2224] = "\xf2\x66", [0x2226] = "\xef\xf5", [0x2227] = "\xf2\x5d",
  [0x2228] = "\xc3\x75", [0x2229] = "\xf2\x64", [0x222a] = "\xf2\x68",
  [0x222b] = "\xf2\x60", [0x222c] = "\x8d\xf4", [0x222f] = "\xf4\x5d",
  [0x2230] = "\xc4\x6a", [0x2231] = "\xf4\x60", [0x2232] = "\xc4\x6b",
  [0x2233] = "\xf4\x68", [0x2234] = "\xf4\x5f", [0x2235] = "\xf4\x5c",
  [0x2237] = "\xf4\x5e", [0x2238] = "\xf4\x62", [0x2239] = "\xf4\x65",
  [0x223a] = "\xf4\x64", [0x223b] = "\xf4\x67", [0x223c] = "\xf4\x5b",
  [0x223e] = "\xc4\x69", [0x223f] = "\xf4\x63", [0x2240] = "\xf4\x66",
  [0x2241] = "\xf4\x69", [0x2242] = "\xf4\x61", [0x2243] = "\xf5\xd3",
  [0x2244] = "\xf5\xd4", [0x2245] = "\xf5\xd8", [0x2246] = "\xf5\xd9",
  [0x2248] = "\xf5\xd6", [0x2249] = "\xf5\xd7", [0x224a] = "\xf5\xd5",
  [0x224b] = "\xfd\xe0", [0x224c] = "\xc4\xe9", [0x2250] = "\x8d\xf6",
  [0x2251] = "\xc5\x78", [0x2252] = "\xf6\xeb", [0x2254] = "\x8d\xf7",
  [0x2255] = "\xf6\xe8", [0x2256] = "\xf6\xe9", [0x2257] = "\xf6\xea",
  [0x2258] = "\xc5\x79", [0x225a] = "\xf7\xe5", [0x225b] = "\xf7\xe4",
  [0x225c] = "\x8f\xfa", [0x225d] = "\xf8\xaf", [0x225e] = "\xc5\xf4",
  [0x225f] = "\xf8\xad", [0x2260] = "\xf8\xb0", [0x2261] = "\xf8\xae",
  [0x2262] = "\xf8\xf5", [0x2263] = "\xc6\x57", [0x2264] = "\xc6\x65",
  [0x2265] = "\xf9\xa3", [0x2266] = "\xf9\x6c", [0x2267] = "\x97\xd0",
  [0x2268] = "\xf9\xa2", [0x2269] = "\xf9\xd0", [0x226a] = "\xf9\xd1",
  [0x226b] = "\xa4\xf5", [0x226c] = "\x8b\xd2", [0x226f] = "\x8d\xf8",
  [0x2270] = "\xa6\xc7", [0x2271] = "\xca\x41", [0x2274] = "\xcb\x5e",
  [0x2275] = "\x90\xd9", [0x2276] = "\xa8\x5f", [0x2278] = "\xa8\x62",
  [0x2279] = "\xfa\xf0", [0x227a] = "\xcb\x5f", [0x227c] = "\xa8\x60",
  [0x227d] = "\xa8\x61", [0x227e] = "\xfd\xe1", [0x227f] = "\x8d\xf9",
  [0x2281] = "\xfd\xe3", [0x2282] = "\xcd\x58", [0x2283] = "\xcd\x5a",
  [0x2284] = "\xcd\x55", [0x2285] = "\xcd\x52", [0x2286] = "\xcd\x54",
  [0x2289] = "\x8d\xfa", [0x228a] = "\xaa\xa4", [0x228b] = "\xfb\x63",
  [0x228e] = "\xaa\xa2", [0x228f] = "\x90\xa6", [0x2291] = "\xcd\x56",
  [0x2292] = "\xaa\xa3", [0x2293] = "\xcd\x53", [0x2294] = "\xcd\x50",
  [0x2295] = "\xaa\xa1", [0x2296] = "\xcd\x57", [0x2298] = "\xcd\x51",
  [0x2299] = "\xaa\xa5", [0x229a] = "\xcd\x59", [0x229f] = "\xcf\xaf",
  [0x22a0] = "\x99\x70", [0x22a1] = "\xcf\xb3", [0x22a3] = "\x91\xeb",
  [0x22a4] = "\xac\xb7", [0x22a5] = "\x97\x70", [0x22a6] = "\x98\x6f",
  [0x22a7] = "\xfd\xe2", [0x22a9] = "\xcf\xb6", [0x22ab] = "\xac\xaf",
  [0x22ac] = "\xac\xb2", [0x22ad] = "\xac\xb4", [0x22ae] = "\xac\xb6",
  [0x22af] = "\xac\xb3", [0x22b0] = "\xcf\xb2", [0x22b1] = "\xcf\xb1",
  [0x22b3] = "\xac\xb1", [0x22b4] = "\xcf\xb4", [0x22b5] = "\xcf\xb5",
  [0x22b7] = "\xcf\xae", [0x22b8] = "\xac\xb5", [0x22b9] = "\x98\xf2",
  [0x22ba] = "\xac\xb0", [0x22bb] = "\x9a\xfc", [0x22bc] = "\x89\x6c",
  [0x22bd] = "\xfd\xfd", [0x22be] = "\xcf\xb0", [0x22c0] = "\x99\x5e",
  [0x22c4] = "\x95\xbd", [0x22c5] = "\xd2\x77", [0x22c6] = "\xd2\x78",
  [0x22c7] = "\xd2\x79", [0x22c8] = "\xaf\x50", [0x22ca] = "\xaf\x4c",
  [0x22cb] = "\xd2\x6e", [0x22cc] = "\xfd\xe4", [0x22cd] = "\xd2\x76",
  [0x22ce] = "\xd2\x7b", [0x22cf] = "\xaf\x51", [0x22d0] = "\x91\xe6",
  [0x22d1] = "\xd2\x6c", [0x22d2] = "\xd2\x72", [0x22d3] = "\xd2\x6b",
  [0x22d4] = "\xd2\x75", [0x22d5] = "\xfd\xe5", [0x22d6] = "\xfd\xe6",
  [0x22d7] = "\xd2\x71", [0x22d8] = "\xaf\x4d", [0x22d9] = "\xaf\x4f",
  [0x22da] = "\xd2\x7a", [0x22dc] = "\xd2\x6a", [0x22dd] = "\xd2\x6d",
  [0x22de] = "\xd2\x73", [0x22df] = "\xfd\xe7", [0x22e0] = "\xd2\x74",
  [0x22e1] = "\xd2\x7c", [0x22e2] = "\xd2\x70", [0x22e4] = "\xaf\x4e",
  [0x22ef] = "\xb2\x6d", [0x22f0] = "\xd6\x4e", [0x22f1] = "\x94\x54",
  [0x22f3] = "\xd6\x50", [0x22f4] = "\xd6\x4c", [0x22f5] = "\x99\xb8",
  [0x22f6] = "\xd6\x58", [0x22f7] = "\xd6\x4a", [0x22f8] = "\xd6\x57",
  [0x22f9] = "\xb2\x69", [0x22fa] = "\xd6\x48", [0x22fb] = "\xda\x5b",
  [0x22fc] = "\xd6\x52", [0x22fd] = "\xb2\x6c", [0x22fe] = "\x97\xe9",
  [0x22ff] = "\xd6\x53", [0x2300] = "\xd6\x56", [0x2302] = "\xd6\x5a",
  [0x2304] = "\xd6\x4f", [0x2305] = "\x93\x46", [0x2306] = "\xd6\x54",
  [0x2309] = "\xb2\x6a", [0x230a] = "\xb2\x6b", [0x230b] = "\xd6\x59",
  [0x230c] = "\xd6\x4d", [0x230d] = "\xd6\x49", [0x230e] = "\xd6\x5b",
  [0x2310] = "\xd6\x51", [0x2313] = "\xd6\x55", [0x2317] = "\xd6\x4b",
  [0x2319] = "\xb5\x48", [0x231a] = "\xb5\x49", [0x231b] = "\xda\x65",
  [0x231c] = "\xb5\x4f", [0x231d] = "\x98\x63", [0x231e] = "\xda\x59",
  [0x231f] = "\xda\x62", [0x2320] = "\xda\x58", [0x2321] = "\xb5\x4c",
  [0x2322] = "\xda\x60", [0x2323] = "\xda\x5e", [0x2325] = "\xda\x5f",
  [0x2326] = "\xb5\x4a", [0x2328] = "\xda\x63", [0x2329] = "\x95\xbc",
  [0x232b] = "\xfd\xed", [0x232c] = "\xfd\xf7", [0x232e] = "\xda\x5c",
  [0x232f] = "\xda\x5a", [0x2330] = "\xb5\x4b", [0x2331] = "\xda\x5d",
  [0x2332] = "\xda\x61", [0x2333] = "\x98\x70", [0x2334] = "\x96\xf6",
  [0x2335] = "\x8e\xa9", [0x2336] = "\xb5\x4d", [0x233a] = "\xda\x64",
  [0x233b] = "\x94\x51", [0x233e] = "\x8e\x43", [0x2340] = "\x8b\x5a",
  [0x2341] = "\xde\x70", [0x2342] = "\xde\x77", [0x2343] = "\xde\x79",
  [0x2344] = "\xde\xa1", [0x2345] = "\xfd\xee", [0x2346] = "\xb7\xda",
  [0x2347] = "\xde\x6b", [0x2349] = "\xb7\xd2", [0x234a] = "\xfd\xf0",
  [0x234b] = "\xde\x7a", [0x234c] = "\xb7\xd7", [0x234d] = "\xde\xa2",
  [0x234e] = "\xb7\xce", [0x234f] = "\xfd\xf4", [0x2350] = "\xde\x7d",
  [0x2351] = "\x9b\xf5", [0x2352] = "\xde\x6d", [0x2353] = "\xde\x7e",
  [0x2354] = "\xde\x6c", [0x2356] = "\xb7\xdc", [0x2358] = "\xde\x78",
  [0x2359] = "\xb7\xcf", [0x235a] = "\xde\xa3", [0x235c] = "\xb7\xd4",
  [0x235d] = "\xde\x71", [0x235e] = "\xb7\xd9", [0x235f] = "\xde\x7c",
  [0x2360] = "\xde\x6f", [0x2361] = "\xde\x76", [0x2362] = "\xde\x72",
  [0x2363] = "\xde\x6e", [0x2364] = "\xb7\xd1", [0x2365] = "\xb7\xd8",
  [0x2366] = "\xb7\xd6", [0x2367] = "\xb7\xd3", [0x2368] = "\xb7\xdb",
  [0x2369] = "\xb7\xd0", [0x236a] = "\xde\x75", [0x236b] = "\x97\x7e",
  [0x236c] = "\xb7\xd5", [0x236e] = "\xb5\x4e", [0x2370] = "\xde\x7b",
  [0x2371] = "\x9b\xd5", [0x2372] = "\xde\x73", [0x2373] = "\x9a\xc3",
  [0x2375] = "\x97\xc8", [0x2376] = "\xa0\xdb", [0x2377] = "\x91\xd0",
  [0x2378] = "\xde\x74", [0x237a] = "\x9f\xe4", [0x237b] = "\xe2\xc1",
  [0x237c] = "\x8f\xdd", [0x237d] = "\xba\xb4", [0x237e] = "\x91\xe9",
  [0x2380] = "\xe2\xbd", [0x2381] = "\xe2\xc3", [0x2382] = "\xe2\xbf",
  [0x2384] = "\xba\xb6", [0x2385] = "\xe2\xbe", [0x2386] = "\xe2\xc2",
  [0x2387] = "\xe2\xba", [0x2388] = "\x98\xe0", [0x2389] = "\xe2\xbc",
  [0x238a] = "\xba\xb5", [0x238c] = "\x92\xca", [0x238e] = "\x98\x57",
  [0x238f] = "\xe2\xc0", [0x2390] = "\xe2\xbb", [0x2392] = "\xba\xb7",
  [0x2394] = "\xba\xb2", [0x2396] = "\xfd\xeb", [0x2397] = "\xe2\xc4",
  [0x2398] = "\x9b\x49", [0x2399] = "\xba\xb3", [0x239a] = "\xe6\x67",
  [0x239b] = "\xe6\x64", [0x239c] = "\xe6\x70", [0x239d] = "\xe6\x6a",
  [0x239e] = "\xe6\x6c", [0x239f] = "\xbc\xf4", [0x23a0] = "\xe6\x66",
  [0x23a1] = "\xe6\x6e", [0x23a2] = "\x9d\x76", [0x23a3] = "\x9e\xaf",
  [0x23a4] = "\xe6\x6d", [0x23a5] = "\xe6\x6b", [0x23a7] = "\xe6\x71",
  [0x23a8] = "\xbc\xf7", [0x23a9] = "\xe6\x68", [0x23aa] = "\xe6\x6f",
  [0x23ac] = "\xbc\xf5", [0x23ad] = "\x9c\xcc", [0x23af] = "\xe6\x63",
  [0x23b0] = "\xe6\x65", [0x23b1] = "\xbc\xf6", [0x23b2] = "\xe6\x62",
  [0x23b3] = "\xe6\x72", [0x23b4] = "\xfd\xea", [0x23b5] = "\xe6\x69",
  [0x23b7] = "\x8d\xf1", [0x23b8] = "\xea\x4a", [0x23b9] = "\xbf\x51",
  [0x23ba] = "\xfd\xfb", [0x23bc] = "\xea\x55", [0x23bd] = "\xea\x53",
  [0x23be] = "\xbf\x4b", [0x23bf] = "\xea\x49", [0x23c0] = "\xea\x4c",
  [0x23c1] = "\xea\x4d", [0x23c2] = "\xea\x48", [0x23c3] = "\xbf\x55",
  [0x23c4] = "\xbf\x56", [0x23c5] = "\xea\x47", [0x23c6] = "\xea\x56",
  [0x23c7] = "\xea\x51", [0x23c8] = "\xbf\x4f", [0x23c9] = "\xbf\x4c",
  [0x23ca] = "\xea\x50", [0x23cb] = "\xea\x4e", [0x23ce] = "\xbf\x52",
  [0x23cf] = "\xea\x52", [0x23d0] = "\xbf\x4d", [0x23d1] = "\x8e\x53",
  [0x23d2] = "\xbf\x4e", [0x23d4] = "\xea\x4f", [0x23d5] = "\xbf\x50",
  [0x23d6] = "\xea\x4b", [0x23d8] = "\xea\x54", [0x23d9] = "\xbf\x53",
  [0x23da] = "\xea\x57", [0x23db] = "\xea\x58", [0x23dc] = "\xbf\x54",
  [0x23dd] = "\xfa\xcf", [0x23df] = "\xc0\xe7", [0x23e0] = "\xc0\xee",
  [0x23e1] = "\xed\x5c", [0x23e2] = "\xed\x62", [0x23e4] = "\xed\x60",
  [0x23e5] = "\xc0\xea", [0x23e6] = "\xc0\xe9", [0x23e7] = "\xc0\xe6",
  [0x23e8] = "\xed\x5e", [0x23eb] = "\x96\xf9", [0x23ec] = "\xc0\xec",
  [0x23ed] = "\xc0\xeb", [0x23ee] = "\xc0\xe8", [0x23f0] = "\xed\x61",
  [0x23f1] = "\xed\x5d", [0x23f2] = "\xed\x5f", [0x23f4] = "\xc0\xed",
  [0x23f5] = "\x98\xbf", [0x23f6] = "\x9e\x49", [0x23f8] = "\xc2\x77",
  [0x23f9] = "\xef\xfb", [0x23fb] = "\xc2\x74", [0x23fc] = "\xc2\x75",
  [0x23fd] = "\xef\xfd", [0x23fe] = "\xc2\x76", [0x23ff] = "\xef\xfa",
  [0x2401] = "\xef\xf9", [0x2402] = "\xf2\x6c", [0x2403] = "\xef\xfc",
  [0x2405] = "\xf2\x6d", [0x2406] = "\xc3\x7a", [0x2407] = "\xf2\x6b",
  [0x2409] = "\x9b\xca", [0x240a] = "\xf2\x6a", [0x240c] = "\xf2\x69",
  [0x240d] = "\xc3\x7b", [0x240e] = "\xfd\xfe", [0x240f] = "\x92\xdc",
  [0x2410] = "\xc4\x6c", [0x2413] = "\xf4\x6a", [0x2414] = "\xf4\x6b",
  [0x2415] = "\xfe\x41", [0x2416] = "\x91\xcc", [0x2417] = "\x91\xe2",
  [0x2419] = "\xf5\xdc", [0x241a] = "\xf5\xdb", [0x241b] = "\xc4\xea",
  [0x241d] = "\xf5\xda", [0x241e] = "\xf6\xec", [0x241f] = "\xf6\xed",
  [0x2422] = "\xf7\xe6", [0x2423] = "\xf8\xb1", [0x2424] = "\xfe\x44",
  [0x2426] = "\xf8\xf6", [0x2427] = "\xf9\xbc", [0x2428] = "\xc6\x79",
  [0x2429] = "\xf9\xc6", [0x242a] = "\xa4\xf6", [0x242b] = "\x8b\xd3",
  [0x242c] = "\xaa\xa6", [0x242d] = "\xaa\xa7", [0x242e] = "\xfe\x47",
  [0x2430] = "\xac\xb8", [0x2435] = "\xc0\xef", [0x2436] = "\xa4\xf7",
  [0x2438] = "\xaa\xa8", [0x2439] = "\xaf\x52", [0x243a] = "\xb7\xdd",
  [0x243b] = "\xa4\xf8", [0x243d] = "\xb2\x6e", [0x243e] = "\xba\xb8",
  [0x243f] = "\xc9\x62", [0x2440] = "\xfe\x48", [0x2441] = "\xcf\xb7",
  [0x2442] = "\xd2\x7d", [0x2444] = "\xe2\xc5", [0x2446] = "\xc0\xf0",
  [0x2447] = "\xa4\xf9", [0x2448] = "\xaa\xa9", [0x2449] = "\xcf\xb8",
  [0x244a] = "\xcf\xb9", [0x244b] = "\xda\x66", [0x244c] = "\xb5\x50",
  [0x244f] = "\xde\xa4", [0x2450] = "\x94\x55", [0x2452] = "\xb7\xde",
  [0x2453] = "\xe2\xc6", [0x2455] = "\xfe\x4b", [0x2456] = "\xbc\xf8",
  [0x2457] = "\xfe\x4c", [0x2458] = "\xc3\x7c", [0x2459] = "\xa4\xfa",
  [0x245a] = "\xda\x67", [0x245b] = "\xa4\xfb", [0x245c] = "\x8d\xbf",
  [0x245d] = "\xa6\xc9", [0x245e] = "\xca\x42", [0x245f] = "\xa6\xc8",
  [0x2460] = "\xa8\x65", [0x2461] = "\xa8\x64", [0x2462] = "\xa8\x63",
  [0x2463] = "\xcb\x60", [0x2466] = "\x9e\x78", [0x2467] = "\xaa\xaa",
  [0x2469] = "\xaa\xab", [0x246a] = "\xcd\x5b", [0x246c] = "\xcf\xba",
  [0x246e] = "\xcf\xbd", [0x246f] = "\xac\xba", [0x2470] = "\xcf\xbb",
  [0x2472] = "\xac\xb9", [0x2473] = "\xcf\xbc", [0x2474] = "\xac\xbb",
  [0x2476] = "\xd2\xa2", [0x2477] = "\xd2\xa1", [0x2478] = "\xd2\x7e",
  [0x2479] = "\xaf\x53", [0x247b] = "\xd6\x5d", [0x247c] = "\xd6\x5e",
  [0x247d] = "\xb2\x6f", [0x247e] = "\xd6\x5c", [0x247f] = "\xd6\x5f",
  [0x2480] = "\xb5\x52", [0x2481] = "\xb2\x70", [0x2482] = "\xfe\x51",
  [0x2484] = "\xb5\x51", [0x2485] = "\xda\x6b", [0x2486] = "\xda\x6a",
  [0x2487] = "\x94\x56", [0x2488] = "\xda\x68", [0x2489] = "\xda\x69",
  [0x248b] = "\xda\x6c", [0x248c] = "\xde\xa6", [0x248d] = "\xde\xa5",
  [0x248e] = "\xde\xa9", [0x248f] = "\x9d\x61", [0x2490] = "\xde\xa8",
  [0x2491] = "\xde\xa7", [0x2492] = "\xba\xb9", [0x2493] = "\xe2\xc9",
  [0x2494] = "\x94\x57", [0x2495] = "\xe2\xc8", [0x2496] = "\xba\xba",
  [0x2497] = "\xe2\xc7", [0x2498] = "\xe6\x73", [0x249a] = "\xe6\x74",
  [0x249b] = "\xbc\xf9", [0x249d] = "\xea\x59", [0x249e] = "\xea\x5a",
  [0x249f] = "\x99\x66", [0x24a1] = "\xf2\x72", [0x24a2] = "\xc3\x7d",
  [0x24a3] = "\xf2\x71", [0x24a4] = "\xf2\x70", [0x24a5] = "\xf2\x6e",
  [0x24a6] = "\xf2\x6f", [0x24a7] = "\xc4\xeb", [0x24a8] = "\xf4\x6c",
  [0x24a9] = "\xf6\xee", [0x24aa] = "\xf8\xf7", [0x24ac] = "\xa4\xfc",
  [0x24ad] = "\x8b\xd5", [0x24ae] = "\xc9\xa5", [0x24af] = "\xa5\xc7",
  [0x24b0] = "\xc9\xa6", [0x24b2] = "\xa0\x69", [0x24b4] = "\xca\x43",
  [0x24b5] = "\xca\x44", [0x24ba] = "\xcb\x66", [0x24bd] = "\xcb\x62",
  [0x24bf] = "\xcb\x61", [0x24c0] = "\xaa\xac", [0x24c1] = "\xcb\x65",
  [0x24c2] = "\xa8\x67", [0x24c3] = "\xcb\x63", [0x24c4] = "\xa8\x66",
  [0x24c5] = "\xcb\x67", [0x24c6] = "\xcb\x64", [0x24c9] = "\xcd\x5f",
  [0x24ca] = "\xcf\xbe", [0x24cb] = "\xcd\x5d", [0x24cc] = "\xcd\x64",
  [0x24cd] = "\x98\xb4", [0x24ce] = "\xaa\xad", [0x24d0] = "\xaa\xb0",
  [0x24d1] = "\xcd\x65", [0x24d2] = "\xcd\x61", [0x24d4] = "\xcd\x62",
  [0x24d6] = "\xcd\x5c", [0x24d7] = "\xaa\xaf", [0x24d8] = "\xcd\x5e",
  [0x24d9] = "\xaa\xae", [0x24da] = "\xcd\x63", [0x24dc] = "\xcd\x60",
  [0x24df] = "\xcf\xc2", [0x24e0] = "\xac\xbd", [0x24e1] = "\xac\xbe",
  [0x24e2] = "\xa0\x49", [0x24e3] = "\xcf\xc5", [0x24e4] = "\xcf\xbf",
  [0x24e6] = "\xcf\xc4", [0x24e8] = "\xcf\xc0", [0x24e9] = "\xac\xbc",
  [0x24ea] = "\xcf\xc3", [0x24eb] = "\xcf\xc1", [0x24f3] = "\xd2\xa8",
  [0x24f4] = "\xd2\xa5", [0x24f6] = "\xd2\xa7", [0x24f7] = "\xaf\x58",
  [0x24f8] = "\xaf\x57", [0x24f9] = "\xaf\x55", [0x24fa] = "\xd2\xa4",
  [0x24fb] = "\xd2\xa9", [0x24fc] = "\xaf\x54", [0x24fd] = "\xaf\x56",
  [0x24fe] = "\xd2\xa6", [0x24ff] = "\xd6\x67", [0x2500] = "\xd2\xa3",
  [0x2501] = "\xd2\xaa", [0x2502] = "\xa0\x4c", [0x2504] = "\x9e\x65",
  [0x2507] = "\xd6\x62", [0x2508] = "\xd6\x66", [0x250a] = "\xd6\x65",
  [0x250b] = "\xda\x6e", [0x250c] = "\xda\x79", [0x250f] = "\xd6\x68",
  [0x2510] = "\x98\xb5", [0x2511] = "\xd6\x63", [0x2512] = "\xda\x6d",
  [0x2513] = "\xb2\x74", [0x2516] = "\xb2\x73", [0x2517] = "\xd6\x61",
  [0x2518] = "\xd6\x64", [0x2519] = "\xb2\x75", [0x251b] = "\xb2\x72",
  [0x251c] = "\xb2\x71", [0x251d] = "\xd6\x60", [0x251e] = "\xd6\x69",
  [0x2522] = "\xda\x70", [0x2523] = "\xda\x77", [0x2525] = "\xb5\x54",
  [0x2526] = "\xda\x76", [0x2527] = "\xda\x73", [0x2528] = "\xfe\x58",
  [0x2529] = "\xb5\x56", [0x252a] = "\x99\x75", [0x252b] = "\xfe\x53",
  [0x252c] = "\xa0\x65", [0x252d] = "\xda\x75", [0x252e] = "\xfe\x59",
  [0x2530] = "\xda\x6f", [0x2531] = "\xda\x71", [0x2532] = "\xda\x74",
  [0x2533] = "\xda\x72", [0x2534] = "\xb5\x55", [0x2535] = "\xda\x78",
  [0x2536] = "\xb5\x53", [0x2537] = "\xb7\xdf", [0x2538] = "\x98\xb7",
  [0x2539] = "\x98\xb8", [0x253a] = "\xde\xad", [0x253b] = "\xde\xac",
  [0x253c] = "\xde\xaa", [0x253e] = "\xb7\xe2", [0x253f] = "\xb7\xe1",
  [0x2540] = "\xde\xae", [0x2541] = "\x98\xba", [0x2542] = "\xde\xab",
  [0x2543] = "\xe2\xca", [0x2544] = "\xba\xbb", [0x2545] = "\xb7\xe0",
  [0x2548] = "\x98\xbb", [0x2549] = "\xde\xb0", [0x254a] = "\xde\xaf",
  [0x254c] = "\xe2\xcd", [0x254d] = "\xe2\xcb", [0x254e] = "\xbc\xfa",
  [0x254f] = "\x9f\xbc", [0x2550] = "\xba\xbc", [0x2551] = "\xe2\xcc",
  [0x2552] = "\xe6\x76", [0x2557] = "\xbc\xfb", [0x2558] = "\xe6\x75",
  [0x2559] = "\xe6\x7e", [0x255a] = "\xe6\x7d", [0x255b] = "\xe6\x7b",
  [0x255d] = "\xe6\x7a", [0x255e] = "\xe6\x77", [0x255f] = "\xe6\x78",
  [0x2560] = "\xe6\x79", [0x2561] = "\xe6\x7c", [0x2562] = "\xe6\xa1",
  [0x2565] = "\xea\x5f", [0x2566] = "\xea\x5c", [0x2567] = "\xea\x5d",
  [0x2568] = "\xbf\x57", [0x2569] = "\xea\x5b", [0x256a] = "\xea\x61",
  [0x256b] = "\xea\x60", [0x256c] = "\xea\x5e", [0x256e] = "\xed\x64",
  [0x256f] = "\xed\x65", [0x2570] = "\xc0\xf1", [0x2571] = "\xa0\x4a",
  [0x2572] = "\xc0\xf2", [0x2573] = "\xed\x63", [0x2574] = "\x9e\xc7",
  [0x2575] = "\xc2\x79", [0x2576] = "\xef\xfe", [0x2577] = "\xc2\x78",
  [0x2578] = "\xc3\x7e", [0x257a] = "\xc3\xa1", [0x257b] = "\xc4\x6d",
  [0x257c] = "\xf4\x6e", [0x257d] = "\xf4\x6d", [0x257e] = "\xf5\xdd",
  [0x257f] = "\xf6\xef", [0x2580] = "\xc5\x7a", [0x2581] = "\xf7\xe8",
  [0x2582] = "\xf7\xe7", [0x2583] = "\xf7\xe9", [0x2584] = "\xa5\xc8",
  [0x2585] = "\xcf\xc6", [0x2586] = "\xaf\x59", [0x2587] = "\xb2\x76",
  [0x2588] = "\xd6\x6a", [0x2589] = "\xa5\xc9", [0x258a] = "\xc9\xa7",
  [0x258b] = "\xa4\xfd", [0x258e] = "\xca\x45", [0x258f] = "\x98\xae",
  [0x2592] = "\xcb\x6c", [0x2593] = "\xcb\x6a", [0x2594] = "\xcb\x6b",
  [0x2595] = "\xcb\x68", [0x2596] = "\xa8\x68", [0x2597] = "\xcb\x69",
  [0x2598] = "\x92\xd6", [0x259c] = "\xfa\xe1", [0x259d] = "\xcd\x6d",
  [0x259e] = "\x91\xd4", [0x259f] = "\xaa\xb3", [0x25a0] = "\xcd\x6b",
  [0x25a1] = "\xcd\x67", [0x25a2] = "\xcd\x6a", [0x25a4] = "\xcd\x66",
  [0x25a5] = "\xaa\xb5", [0x25a6] = "\xcd\x69", [0x25a7] = "\xfa\xde",
  [0x25a8] = "\xaa\xb2", [0x25a9] = "\xaa\xb1", [0x25aa] = "\xfe\x5b",
  [0x25ab] = "\xaa\xb4", [0x25ac] = "\xcd\x6c", [0x25ad] = "\xcd\x68",
  [0x25b2] = "\xac\xc2", [0x25b3] = "\xac\xc5", [0x25b4] = "\xcf\xce",
  [0x25b5] = "\xcf\xcd", [0x25b6] = "\xcf\xcc", [0x25b7] = "\xac\xbf",
  [0x25b8] = "\xcf\xd5", [0x25b9] = "\xcf\xcb", [0x25bb] = "\xac\xc1",
  [0x25bc] = "\xd2\xaf", [0x25be] = "\xcf\xd2", [0x25bf] = "\xcf\xd0",
  [0x25c0] = "\xac\xc4", [0x25c2] = "\xcf\xc8", [0x25c3] = "\xcf\xd3",
  [0x25c5] = "\xcf\xca", [0x25c6] = "\xcf\xd4", [0x25c7] = "\xcf\xd1",
  [0x25c8] = "\xcf\xc9", [0x25c9] = "\xfe\x5e", [0x25ca] = "\xac\xc0",
  [0x25cb] = "\xcf\xd6", [0x25cc] = "\xcf\xc7", [0x25cd] = "\xac\xc3",
  [0x25ce] = "\xfb\xd7", [0x25cf] = "\xfe\x5a", [0x25d0] = "\x94\xc5",
  [0x25d2] = "\xd2\xb4", [0x25d3] = "\xd2\xab", [0x25d4] = "\xd2\xb6",
  [0x25d5] = "\xfa\xca", [0x25d6] = "\xd2\xae", [0x25d7] = "\xd2\xb9",
  [0x25d8] = "\xd2\xba", [0x25d9] = "\xd2\xac", [0x25da] = "\xd2\xb8",
  [0x25db] = "\xd2\xb5", [0x25dc] = "\xd2\xb3", [0x25dd] = "\xd2\xb7",
  [0x25de] = "\xaf\x5f", [0x25e0] = "\xaf\x5d", [0x25e1] = "\x98\xc1",
  [0x25e2] = "\x97\x5c", [0x25e3] = "\xd2\xb1", [0x25e4] = "\xfe\x74",
  [0x25e5] = "\xd2\xad", [0x25e6] = "\x97\x73", [0x25e7] = "\xd2\xb0",
  [0x25e8] = "\xd2\xbb", [0x25e9] = "\xd2\xb2", [0x25ea] = "\xaf\x5e",
  [0x25eb] = "\xcf\xcf", [0x25ed] = "\xaf\x5a", [0x25ee] = "\xaf\x5c",
  [0x25ef] = "\xfa\x46", [0x25f3] = "\x97\x64", [0x25f4] = "\xd6\x78",
  [0x25f5] = "\xd6\x6d", [0x25f6] = "\xd6\x6b", [0x25f7] = "\xfe\x68",
  [0x25f8] = "\xd6\x6c", [0x25f9] = "\x96\x4e", [0x25fa] = "\xd6\x73",
  [0x25fb] = "\x97\x65", [0x25fc] = "\xd6\x74", [0x25fd] = "\xd6\x70",
  [0x25fe] = "\xb2\x7b", [0x25ff] = "\xd6\x75", [0x2600] = "\xd6\x72",
  [0x2601] = "\xd6\x6f", [0x2603] = "\xb2\x79", [0x2604] = "\xd6\x6e",
  [0x2605] = "\xb2\x77", [0x2606] = "\xb2\x7a", [0x2607] = "\xd6\x71",
  [0x2608] = "\xd6\x79", [0x2609] = "\xaf\x5b", [0x260a] = "\xb2\x78",
  [0x260b] = "\xd6\x77", [0x260c] = "\xd6\x76", [0x260d] = "\xb2\x7c",
  [0x2611] = "\x89\xa1", [0x2612] = "\x95\xfa", [0x2614] = "\x92\xd4",
  [0x2615] = "\xfe\x69", [0x2616] = "\xda\x7e", [0x2617] = "\xfb\x45",
  [0x2619] = "\x98\xc8", [0x261a] = "\xda\xa1", [0x261b] = "\xb5\x60",
  [0x261c] = "\x90\xef", [0x261d] = "\xda\xa7", [0x261e] = "\x98\xc9",
  [0x261f] = "\x98\xca", [0x2620] = "\xda\xa9", [0x2621] = "\xda\xa2",
  [0x2622] = "\xb5\x5a", [0x2623] = "\xda\xa6", [0x2624] = "\xda\xa5",
  [0x2625] = "\xb5\x5b", [0x2626] = "\xb5\x61", [0x2628] = "\xb5\x62",
  [0x2629] = "\xda\xa8", [0x262a] = "\xb5\x58", [0x262b] = "\xda\x7d",
  [0x262c] = "\xda\x7b", [0x262d] = "\xda\xa3", [0x262e] = "\xda\x7a",
  [0x262f] = "\xb5\x5f", [0x2630] = "\xda\x7c", [0x2631] = "\xda\xa4",
  [0x2632] = "\xda\xaa", [0x2633] = "\xb5\x59", [0x2634] = "\xb5\x5e",
  [0x2635] = "\xb5\x5c", [0x2636] = "\xb5\x5d", [0x2637] = "\x94\x6d",
  [0x2638] = "\x94\xb7", [0x2639] = "\xfe\x6c", [0x263a] = "\xb5\x57",
  [0x263c] = "\x94\x6b", [0x263f] = "\xb7\xe9", [0x2640] = "\xde\xb7",
  [0x2641] = "\xb7\xe8", [0x2642] = "\xde\xbb", [0x2643] = "\x92\xfc",
  [0x2644] = "\xde\xb1", [0x2645] = "\x95\xeb", [0x2646] = "\xde\xbc",
  [0x2647] = "\xfe\x73", [0x2648] = "\x97\x6e", [0x2649] = "\xfe\x5f",
  [0x264a] = "\xde\xb2", [0x264b] = "\xde\xb3", [0x264d] = "\xde\xbd",
  [0x264e] = "\xde\xba", [0x264f] = "\xde\xb8", [0x2650] = "\xde\xb9",
  [0x2651] = "\xde\xb5", [0x2652] = "\xde\xb4", [0x2653] = "\xfd\xbd",
  [0x2654] = "\xde\xbe", [0x2655] = "\xb7\xe5", [0x2656] = "\x92\xd5",
  [0x2657] = "\xde\xb6", [0x2659] = "\xb7\xea", [0x265a] = "\xb7\xe4",
  [0x265b] = "\xb7\xeb", [0x265c] = "\xb7\xec", [0x265d] = "\xfe\xb9",
  [0x265e] = "\xb7\xe7", [0x265f] = "\xb7\xe6", [0x2660] = "\xfe\x71",
  [0x2662] = "\xe2\xce", [0x2663] = "\xba\xbe", [0x2664] = "\xba\xbd",
  [0x2665] = "\xfb\xbb", [0x2667] = "\xe2\xd3", [0x2668] = "\x94\x7a",
  [0x2669] = "\xbc\xfc", [0x266a] = "\xba\xbf", [0x266b] = "\x95\xfb",
  [0x266c] = "\xfe\x77", [0x266d] = "\xba\xc1", [0x266e] = "\xe2\xd4",
  [0x266f] = "\xb7\xe3", [0x2670] = "\xba\xc0", [0x2671] = "\xe2\xd0",
  [0x2672] = "\xe2\xd2", [0x2673] = "\xe2\xcf", [0x2674] = "\xfe\x79",
  [0x2675] = "\xe2\xd1", [0x2676] = "\xfe\x75", [0x2679] = "\xe6\xab",
  [0x267a] = "\x94\x5d", [0x267c] = "\xe6\xaa", [0x267d] = "\xe6\xa7",
  [0x267e] = "\xbd\x40", [0x267f] = "\xea\x62", [0x2680] = "\xbd\x41",
  [0x2681] = "\xe6\xa6", [0x2682] = "\xfe\x7c", [0x2683] = "\xbc\xfe",
  [0x2685] = "\xe6\xa8", [0x2686] = "\xe6\xa5", [0x2687] = "\xe6\xa2",
  [0x2688] = "\xe6\xa9", [0x2689] = "\xe6\xa3", [0x268a] = "\xe6\xa4",
  [0x268b] = "\xbc\xfd", [0x268c] = "\x93\x44", [0x268d] = "\x8e\xa6",
  [0x2690] = "\xed\x69", [0x2692] = "\xea\x66", [0x2694] = "\xea\x65",
  [0x2695] = "\xea\x67", [0x2697] = "\xed\x66", [0x2698] = "\xbf\x5a",
  [0x2699] = "\x92\xd3", [0x269a] = "\xea\x63", [0x269b] = "\x94\xb8",
  [0x269c] = "\xbf\x58", [0x269e] = "\xbf\x5c", [0x269f] = "\xbf\x5b",
  [0x26a0] = "\xea\x64", [0x26a1] = "\xea\x68", [0x26a3] = "\xbf\x59",
  [0x26a4] = "\xfc\x71", [0x26a5] = "\xed\x6d", [0x26a6] = "\xc0\xf5",
  [0x26a7] = "\xc2\x7a", [0x26a8] = "\xc0\xf6", [0x26a9] = "\xc0\xf3",
  [0x26aa] = "\xed\x6a", [0x26ab] = "\xed\x68", [0x26ad] = "\xed\x6b",
  [0x26af] = "\xed\x6e", [0x26b0] = "\xc0\xf4", [0x26b1] = "\xed\x6c",
  [0x26b2] = "\xed\x67", [0x26b4] = "\x97\x5e", [0x26b5] = "\xf0\x42",
  [0x26b6] = "\xf0\x45", [0x26b7] = "\xf2\x75", [0x26b8] = "\xf0\x40",
  [0x26ba] = "\xf4\x6f", [0x26bb] = "\xf0\x46", [0x26bd] = "\xc3\xa2",
  [0x26be] = "\xf0\x44", [0x26bf] = "\xc2\x7b", [0x26c0] = "\xf0\x41",
  [0x26c1] = "\xf0\x43", [0x26c2] = "\xf0\x47", [0x26c3] = "\xf2\x76",
  [0x26c5] = "\xf2\x74", [0x26c8] = "\xfe\xa7", [0x26ca] = "\xc3\xa3",
  [0x26cb] = "\xf2\x73", [0x26cc] = "\x94\x6a", [0x26cf] = "\xc4\x6e",
  [0x26d0] = "\x93\xe3", [0x26d3] = "\x98\xcf", [0x26d4] = "\xc4\xed",
  [0x26d5] = "\xf6\xf1", [0x26d6] = "\xc4\xec", [0x26d7] = "\xf6\xf3",
  [0x26d8] = "\xf6\xf0", [0x26d9] = "\xf6\xf2", [0x26da] = "\xc5\xd0",
  [0x26db] = "\xf8\xb2", [0x26dc] = "\xa5\xca", [0x26dd] = "\xcd\x6e",
  [0x26de] = "\xd2\xbc", [0x26df] = "\xd2\xbd", [0x26e0] = "\xb2\x7d",
  [0x26e1] = "\xde\xbf", [0x26e2] = "\xbf\x5d", [0x26e3] = "\xc3\xa4",
  [0x26e4] = "\xc5\x7b", [0x26e5] = "\xf8\xb3", [0x26e6] = "\xa5\xcb",
  [0x26e7] = "\xa0\xd9", [0x26e8] = "\xcd\x6f", [0x26e9] = "\xa2\x60",
  [0x26ec] = "\xcf\xd7", [0x26ee] = "\xcf\xd8", [0x26f0] = "\xa0\xbf",
  [0x26f1] = "\xa0\x4d", [0x26f2] = "\xa0\xb8", [0x26f4] = "\xd2\xbe",
  [0x26f5] = "\xd2\xbf", [0x26f6] = "\xb2\x7e", [0x26f7] = "\xb2\xa1",
  [0x26f8] = "\xa0\xce", [0x26fb] = "\xda\xab", [0x26fd] = "\xde\xc2",
  [0x26fe] = "\xde\xc1", [0x26ff] = "\xde\xc0", [0x2700] = "\xe2\xd5",
  [0x2702] = "\xe2\xd6", [0x2703] = "\xe2\xd7", [0x2704] = "\xba\xc2",
  [0x2705] = "\xa0\xb7", [0x2707] = "\xe6\xad", [0x2708] = "\xe6\xac",
  [0x270b] = "\xea\x69", [0x270c] = "\xbf\x5e", [0x270d] = "\xbf\x5f",
  [0x270e] = "\xfe\xa9", [0x270f] = "\xed\x72", [0x2710] = "\xed\x6f",
  [0x2711] = "\xed\x70", [0x2712] = "\xed\x71", [0x2713] = "\xf0\x49",
  [0x2714] = "\xf0\x48", [0x2715] = "\xc2\x7c", [0x2716] = "\xf2\x77",
  [0x2717] = "\xf5\xde", [0x2718] = "\xa5\xcc", [0x2719] = "\x89\xc3",
  [0x271a] = "\xac\xc6", [0x271c] = "\xb2\xa2", [0x271d] = "\xde\xc3",
  [0x271e] = "\xfe\xab", [0x271f] = "\xa5\xcd", [0x2721] = "\xd2\xc0",
  [0x2722] = "\xb2\xa3", [0x2725] = "\xb5\x63", [0x2726] = "\xb5\x64",
  [0x2728] = "\xa5\xce", [0x2729] = "\xa5\xcf", [0x272a] = "\xca\x46",
  [0x272b] = "\xa8\x6a", [0x272c] = "\xa8\x69", [0x272d] = "\xac\xc7",
  [0x272e] = "\xcf\xd9", [0x272f] = "\xda\xac", [0x2730] = "\xa5\xd0",
  [0x2731] = "\xa5\xd1", [0x2732] = "\xa5\xd2", [0x2733] = "\xa5\xd3",
  [0x2734] = "\x9d\xf4", [0x2735] = "\x89\x6d", [0x2737] = "\xa8\x6b",
  [0x2738] = "\xa8\x6c", [0x2739] = "\xcb\x6e", [0x273a] = "\xcb\x6d",
  [0x273b] = "\x9c\x7b", [0x273d] = "\xaa\xb6", [0x273e] = "\xcd\x72",
  [0x273f] = "\xcd\x70", [0x2740] = "\xcd\x71", [0x2742] = "\x98\xd2",
  [0x2746] = "\x9f\xa9", [0x2747] = "\xcf\xda", [0x2748] = "\xcf\xdb",
  [0x274a] = "\xfe\xb2", [0x274b] = "\xac\xcb", [0x274c] = "\xac\xc9",
  [0x274d] = "\xfe\xb1", [0x274e] = "\xac\xca", [0x274f] = "\xac\xc8",
  [0x2751] = "\x97\xd9", [0x2753] = "\xa0\xc4", [0x2754] = "\xaf\x60",
  [0x2755] = "\x94\x76", [0x2759] = "\xaf\x64", [0x275a] = "\xaf\x63",
  [0x275b] = "\xd2\xc1", [0x275c] = "\xaf\x62", [0x275d] = "\xaf\x61",
  [0x275f] = "\xd2\xc2", [0x2760] = "\x99\x78", [0x2762] = "\xb2\xa6",
  [0x2763] = "\xd6\x7b", [0x2764] = "\xd6\x7a", [0x2765] = "\xb2\xa4",
  [0x2766] = "\xb2\xa5", [0x2767] = "\xfe\xb3", [0x276a] = "\xb5\x66",
  [0x276b] = "\xb5\x65", [0x276c] = "\xda\xae", [0x276d] = "\x98\xd3",
  [0x276e] = "\xfe\xb4", [0x276f] = "\xda\xad", [0x2770] = "\xb2\xa7",
  [0x2772] = "\x98\xd4", [0x2776] = "\xb7\xed", [0x2777] = "\xde\xc5",
  [0x2778] = "\xb7\xee", [0x2779] = "\xde\xc4", [0x277a] = "\x9f\xb9",
  [0x277d] = "\xe2\xd8", [0x277e] = "\xe6\xae", [0x277f] = "\xbd\x42",
  [0x2780] = "\xea\x6a", [0x2783] = "\x94\x71", [0x2784] = "\xed\x73",
  [0x2786] = "\xc3\xa6", [0x2787] = "\xc3\xa5", [0x278a] = "\xc5\x7c",
  [0x278b] = "\xa5\xd4", [0x278c] = "\xcd\x73", [0x278d] = "\x98\xd5",
  [0x278e] = "\xfe\xb8", [0x278f] = "\xb2\xa8", [0x2790] = "\xe2\xd9",
  [0x2791] = "\xba\xc3", [0x2792] = "\xc6\xd4", [0x2794] = "\xcb\x6f",
  [0x2795] = "\xcb\x70", [0x2798] = "\xcd\x74", [0x2799] = "\xaa\xb8",
  [0x279a] = "\xaa\xb9", [0x279d] = "\xaa\xb7", [0x279e] = "\xfe\xba",
  [0x27a2] = "\xac\xcf", [0x27a3] = "\xac\xd0", [0x27a4] = "\xac\xcd",
  [0x27a5] = "\xac\xce", [0x27a7] = "\xcf\xdc", [0x27aa] = "\xcf\xdd",
  [0x27ab] = "\xac\xcc", [0x27b0] = "\xd2\xc3", [0x27b1] = "\x9e\x5c",
  [0x27b2] = "\xaf\x68", [0x27b3] = "\xaf\x69", [0x27b4] = "\xfe\xbb",
  [0x27b5] = "\xb2\xab", [0x27b6] = "\xd2\xc9", [0x27b8] = "\xaf\x6e",
  [0x27b9] = "\xaf\x6c", [0x27ba] = "\xd2\xca", [0x27bb] = "\xd2\xc5",
  [0x27bc] = "\xaf\x6b", [0x27bd] = "\xaf\x6a", [0x27be] = "\xaf\x65",
  [0x27bf] = "\xd2\xc8", [0x27c0] = "\xd2\xc7", [0x27c1] = "\xd2\xc4",
  [0x27c2] = "\xaf\x6d", [0x27c3] = "\xa0\x44", [0x27c4] = "\xd2\xc6",
  [0x27c5] = "\xaf\x66", [0x27c7] = "\xaf\x67", [0x27c8] = "\x98\xd7",
  [0x27ca] = "\xb2\xac", [0x27cb] = "\xd6\xa1", [0x27cc] = "\xd6\xa2",
  [0x27cd] = "\xb2\xad", [0x27ce] = "\xd6\x7c", [0x27cf] = "\xd6\x7e",
  [0x27d0] = "\xd6\xa4", [0x27d1] = "\xd6\xa3", [0x27d2] = "\xd6\x7d",
  [0x27d4] = "\xb2\xa9", [0x27d5] = "\xb2\xaa", [0x27d7] = "\xda\xb6",
  [0x27d8] = "\xb5\x6b", [0x27d9] = "\xb5\x6a", [0x27da] = "\xda\xb0",
  [0x27db] = "\xb5\x68", [0x27dc] = "\x98\xd8", [0x27dd] = "\xda\xb3",
  [0x27de] = "\xb5\x6c", [0x27df] = "\xda\xb4", [0x27e0] = "\xb5\x6d",
  [0x27e1] = "\xda\xb1", [0x27e2] = "\xb5\x67", [0x27e3] = "\xb5\x69",
  [0x27e4] = "\xda\xb5", [0x27e6] = "\xda\xb2", [0x27e7] = "\xda\xaf",
  [0x27ed] = "\xde\xd2", [0x27ef] = "\xde\xc7", [0x27f0] = "\xb7\xf0",
  [0x27f1] = "\xb7\xf3", [0x27f2] = "\xb7\xf2", [0x27f3] = "\xb7\xf7",
  [0x27f4] = "\xb7\xf6", [0x27f5] = "\xde\xd3", [0x27f6] = "\xde\xd1",
  [0x27f7] = "\xde\xca", [0x27f8] = "\xde\xce", [0x27f9] = "\xde\xcd",
  [0x27fa] = "\xb7\xf4", [0x27fb] = "\xde\xd0", [0x27fc] = "\xde\xcc",
  [0x27fd] = "\xde\xd4", [0x27fe] = "\xde\xcb", [0x27ff] = "\xb7\xf5",
  [0x2800] = "\xb7\xef", [0x2801] = "\xb7\xf1", [0x2802] = "\xfe\xbc",
  [0x2803] = "\xde\xc9", [0x2807] = "\x9f\xfe", [0x2808] = "\xe2\xdb",
  [0x2809] = "\xba\xc7", [0x280a] = "\xe2\xdf", [0x280b] = "\xba\xc6",
  [0x280c] = "\xe2\xdc", [0x280d] = "\xba\xc5", [0x280f] = "\xde\xc8",
  [0x2810] = "\xde\xcf", [0x2811] = "\xe2\xde", [0x2813] = "\xba\xc8",
  [0x2814] = "\xe2\xe0", [0x2815] = "\xe2\xdd", [0x2816] = "\xe2\xda",
  [0x2819] = "\xe6\xb1", [0x281a] = "\xe6\xb5", [0x281b] = "\xe6\xb7",
  [0x281c] = "\xe6\xb3", [0x281d] = "\xe6\xb2", [0x281e] = "\xe6\xb0",
  [0x281f] = "\xbd\x45", [0x2820] = "\xbd\x43", [0x2821] = "\xbd\x48",
  [0x2822] = "\xbd\x49", [0x2823] = "\xe6\xb4", [0x2824] = "\xbd\x46",
  [0x2825] = "\xe6\xaf", [0x2826] = "\xbd\x47", [0x2827] = "\xba\xc4",
  [0x2828] = "\xe6\xb6", [0x2829] = "\xbd\x44", [0x282c] = "\xfe\xbd",
  [0x282d] = "\xea\x6c", [0x282f] = "\xea\x6b", [0x2830] = "\xea\x73",
  [0x2831] = "\xea\x6d", [0x2832] = "\xea\x72", [0x2833] = "\xea\x6f",
  [0x2834] = "\xbf\x60", [0x2835] = "\xea\x71", [0x2838] = "\xbf\x61",
  [0x283a] = "\xbf\x62", [0x283b] = "\x9d\xdd", [0x283c] = "\xea\x70",
  [0x283d] = "\xea\x6e", [0x2840] = "\x9e\xe1", [0x2842] = "\xc0\xf8",
  [0x2843] = "\xed\x74", [0x2846] = "\xc0\xf7", [0x2847] = "\xed\x77",
  [0x2848] = "\xed\x75", [0x2849] = "\xed\x76", [0x284c] = "\xc0\xf9",
  [0x284d] = "\x98\xda", [0x284e] = "\x9d\xdf", [0x284f] = "\xfe\xbf",
  [0x2850] = "\xf0\x4d", [0x2851] = "\xfe\xbe", [0x2852] = "\xc2\xa1",
  [0x2853] = "\xf0\x4e", [0x2854] = "\x9e\xeb", [0x2856] = "\xc2\x7d",
  [0x2857] = "\xf0\x4f", [0x2858] = "\xc2\x7e", [0x2859] = "\xf0\x4c",
  [0x285a] = "\xf0\x50", [0x285c] = "\xf0\x4a", [0x285f] = "\xc3\xa7",
  [0x2860] = "\xf2\x78", [0x2861] = "\xc3\xa8", [0x2862] = "\xc4\x6f",
  [0x2864] = "\xf0\x4b", [0x2865] = "\xc4\x70", [0x2866] = "\x9e\x59",
  [0x2867] = "\xa0\x5c", [0x2869] = "\xc4\xee", [0x286a] = "\xf5\xdf",
  [0x286c] = "\xc5\x7e", [0x286d] = "\xf6\xf4", [0x286e] = "\xc5\x7d",
  [0x286f] = "\xfe\xc0", [0x2870] = "\xf7\xea", [0x2871] = "\xc5\xf5",
  [0x2872] = "\xc5\xf6", [0x2873] = "\x94\x77", [0x2874] = "\x98\xdc",
  [0x2875] = "\xf9\xcc", [0x2876] = "\xfe\xc1", [0x2878] = "\xac\xd1",
  [0x2879] = "\xcf\xde", [0x287a] = "\x98\xde", [0x287b] = "\xb5\x6e",
  [0x287c] = "\xb5\x6f", [0x287d] = "\xa5\xd5", [0x287e] = "\xa6\xca",
  [0x287f] = "\xca\x47", [0x2881] = "\xcb\x71", [0x2882] = "\xa8\x6d",
  [0x2884] = "\xaa\xba", [0x2886] = "\xac\xd2", [0x2887] = "\xac\xd3",
  [0x2888] = "\xac\xd4", [0x2889] = "\xd6\xa6", [0x288a] = "\xd2\xcb",
  [0x288b] = "\xaf\x6f", [0x288e] = "\xb2\xae", [0x288f] = "\xd6\xa5",
  [0x2890] = "\xfe\xc3", [0x2892] = "\xda\xb8", [0x2893] = "\xb5\x71",
  [0x2895] = "\xda\xb7", [0x2896] = "\xb5\x70", [0x2899] = "\xde\xd5",
  [0x289a] = "\xbd\x4a", [0x289b] = "\xe6\xbb", [0x289c] = "\xe6\xb8",
  [0x289d] = "\xe6\xb9", [0x289e] = "\xe6\xba", [0x28a1] = "\xfe\xc8",
  [0x28a4] = "\xed\x78", [0x28a5] = "\xfe\xc9", [0x28a6] = "\xf0\x51",
  [0x28aa] = "\xf4\x71", [0x28ab] = "\xf4\x70", [0x28ad] = "\xf6\xf5",
  [0x28ae] = "\xa5\xd6", [0x28af] = "\xcd\x75", [0x28b0] = "\xaf\x70",
  [0x28b4] = "\xb5\x72", [0x28b5] = "\xde\xd6", [0x28b7] = "\xfe\xca",
  [0x28b8] = "\xe2\xe1", [0x28ba] = "\xbd\x4b", [0x28bb] = "\xea\x74",
  [0x28bd] = "\xf0\x52", [0x28be] = "\xf4\x72", [0x28bf] = "\xa5\xd7",
  [0x28c2] = "\xaa\xbb", [0x28c3] = "\xac\xd7", [0x28c4] = "\xcf\xdf",
  [0x28c5] = "\xac\xd8", [0x28c6] = "\xac\xd6", [0x28c8] = "\xac\xd5",
  [0x28c9] = "\xd2\xcc", [0x28ca] = "\xaf\x71", [0x28cc] = "\xfe\xcb",
  [0x28cd] = "\xaf\x72", [0x28ce] = "\xaf\x73", [0x28d2] = "\xb2\xb0",
  [0x28d3] = "\xd6\xa7", [0x28d4] = "\xb2\xaf", [0x28d6] = "\x9f\xc2",
  [0x28da] = "\xda\xb9", [0x28db] = "\xb2\xb1", [0x28dc] = "\xb5\x73",
  [0x28dd] = "\xde\xd7", [0x28de] = "\xb7\xf8", [0x28df] = "\xb7\xf9",
  [0x28e1] = "\xba\xc9", [0x28e3] = "\xba\xca", [0x28e4] = "\xbd\x4c",
  [0x28e5] = "\xbf\x64", [0x28e6] = "\xea\x75", [0x28e7] = "\xbf\x63",
  [0x28e9] = "\xed\x79", [0x28ea] = "\xc0\xfa", [0x28ec] = "\xf0\x53",
  [0x28ed] = "\xf4\x73", [0x28ee] = "\xa5\xd8", [0x28ef] = "\xa8\x6e",
  [0x28f0] = "\xcd\x78", [0x28f1] = "\xcd\x77", [0x28f2] = "\xaa\xbc",
  [0x28f3] = "\xcd\x76", [0x28f4] = "\xaa\xbd", [0x28f5] = "\xcd\x79",
  [0x28f7] = "\xcf\xe5", [0x28f8] = "\xac\xdb", [0x28f9] = "\xac\xda",
  [0x28fa] = "\xcf\xe7", [0x28fb] = "\xcf\xe6", [0x28fc] = "\xac\xdf",
  [0x28fe] = "\xac\xde", [0x2901] = "\xac\xd9", [0x2903] = "\xcf\xe1",
  [0x2904] = "\xcf\xe2", [0x2905] = "\xcf\xe3", [0x2907] = "\xac\xe0",
  [0x2908] = "\xcf\xe0", [0x2909] = "\xac\xdc", [0x290a] = "\xcf\xe4",
  [0x290b] = "\xac\xdd", [0x290c] = "\x98\xc4", [0x290e] = "\x94\xb0",
  [0x290f] = "\x94\xb1", [0x2910] = "\xd2\xcf", [0x2911] = "\xd2\xd3",
  [0x2912] = "\xd2\xd1", [0x2913] = "\xd2\xd0", [0x2915] = "\xd2\xd4",
  [0x2919] = "\xd2\xd5", [0x291a] = "\xd2\xd6", [0x291b] = "\xd2\xce",
  [0x291d] = "\xd2\xcd", [0x291e] = "\xfe\xd1", [0x291f] = "\xaf\x75",
  [0x2920] = "\xaf\x76", [0x2922] = "\xd2\xd7", [0x2923] = "\xd2\xd2",
  [0x2924] = "\xa0\xc1", [0x2925] = "\xd6\xb0", [0x2926] = "\xfe\xd2",
  [0x2927] = "\xd2\xd8", [0x2928] = "\xaf\x77", [0x2929] = "\xaf\x74",
  [0x292b] = "\xa0\xcd", [0x292d] = "\xd6\xaa", [0x292f] = "\xd6\xa9",
  [0x2931] = "\xd6\xab", [0x2932] = "\xd6\xac", [0x2933] = "\xd6\xae",
  [0x2934] = "\xd6\xad", [0x2935] = "\xd6\xb2", [0x2936] = "\xb2\xb5",
  [0x2937] = "\xb2\xb2", [0x2938] = "\xb2\xb6", [0x2939] = "\xd6\xa8",
  [0x293a] = "\xb2\xb7", [0x293b] = "\xd6\xb1", [0x293c] = "\xb2\xb4",
  [0x293d] = "\xd6\xaf", [0x293e] = "\xb2\xb3", [0x2940] = "\xfe\xd3",
  [0x2943] = "\x98\xe5", [0x2944] = "\xda\xbc", [0x2945] = "\xda\xbe",
  [0x2946] = "\xda\xba", [0x2947] = "\xda\xbb", [0x294a] = "\xda\xbf",
  [0x294b] = "\xda\xc1", [0x294c] = "\xda\xc2", [0x294d] = "\xda\xbd",
  [0x294e] = "\xda\xc0", [0x294f] = "\xb5\x74", [0x2952] = "\xde\xdb",
  [0x2954] = "\xde\xe0", [0x2955] = "\xde\xd8", [0x2956] = "\xde\xdc",
  [0x2958] = "\xfe\xd6", [0x2959] = "\xde\xe1", [0x295a] = "\xde\xdd",
  [0x295b] = "\xb7\xfa", [0x295c] = "\xb8\x43", [0x295e] = "\xb7\xfd",
  [0x295f] = "\xde\xd9", [0x2960] = "\xde\xda", [0x2961] = "\xba\xce",
  [0x2962] = "\xb8\x46", [0x2963] = "\xb7\xfe", [0x2965] = "\xb8\x44",
  [0x2966] = "\xb7\xfc", [0x2967] = "\xde\xdf", [0x2968] = "\xb8\x45",
  [0x2969] = "\xde\xde", [0x296a] = "\xb8\x41", [0x296b] = "\xb7\xfb",
  [0x296c] = "\xb8\x42", [0x296d] = "\xde\xe2", [0x296e] = "\xe2\xe6",
  [0x296f] = "\xe2\xe8", [0x2972] = "\x91\xe4", [0x2977] = "\x8f\xc7",
  [0x2978] = "\x94\xae", [0x2979] = "\xb8\x40", [0x297a] = "\x8a\x4f",
  [0x297b] = "\x94\xb2", [0x297c] = "\xe2\xe3", [0x297d] = "\xba\xcc",
  [0x297e] = "\xe2\xe9", [0x297f] = "\xba\xcd", [0x2980] = "\xe2\xe7",
  [0x2981] = "\xe2\xe2", [0x2982] = "\xe2\xe5", [0x2983] = "\xe2\xea",
  [0x2984] = "\xba\xcb", [0x2985] = "\xe2\xe4", [0x2987] = "\xbd\x4e",
  [0x2988] = "\xe6\xbf", [0x2989] = "\xe6\xbe", [0x298b] = "\xbd\x51",
  [0x298c] = "\xbd\x4f", [0x298d] = "\xe6\xbc", [0x298e] = "\xbd\x4d",
  [0x298f] = "\xe6\xbd", [0x2991] = "\xbd\x50", [0x2993] = "\x8f\xd4",
  [0x2995] = "\xea\x7d", [0x2997] = "\xea\xa1", [0x2998] = "\x98\xea",
  [0x2999] = "\xea\x7e", [0x299a] = "\xea\x76", [0x299b] = "\xea\x7a",
  [0x299c] = "\xea\x79", [0x299d] = "\xea\x77", [0x299e] = "\xbf\x66",
  [0x299f] = "\xbf\x67", [0x29a0] = "\xbf\x65", [0x29a1] = "\xea\x78",
  [0x29a2] = "\xea\x7b", [0x29a3] = "\xea\x7c", [0x29a5] = "\xbf\x68",
  [0x29a7] = "\xc1\x40", [0x29a8] = "\xed\xa3", [0x29aa] = "\xc0\xfc",
  [0x29ab] = "\xed\x7b", [0x29ac] = "\xc0\xfe", [0x29ad] = "\xc1\x41",
  [0x29af] = "\xfe\xd8", [0x29b0] = "\xc0\xfd", [0x29b1] = "\xed\xa2",
  [0x29b2] = "\xed\x7c", [0x29b3] = "\xc0\xfb", [0x29b4] = "\xed\xa1",
  [0x29b5] = "\xed\x7a", [0x29b6] = "\xed\x7e", [0x29b7] = "\xed\x7d",
  [0x29b9] = "\x9d\xe0", [0x29ba] = "\xf0\x55", [0x29bb] = "\xc2\xa4",
  [0x29bc] = "\xc2\xa5", [0x29bd] = "\xc2\xa2", [0x29be] = "\x98\xee",
  [0x29bf] = "\xc2\xa3", [0x29c2] = "\xf0\x54", [0x29c3] = "\x95\xc4",
  [0x29c4] = "\xf2\x7b", [0x29c5] = "\xfc\xe8", [0x29c7] = "\xc3\xa9",
  [0x29c9] = "\xf2\x79", [0x29ca] = "\xf2\x7a", [0x29cb] = "\x98\xef",
  [0x29cc] = "\xf4\x74", [0x29cd] = "\xf4\x77", [0x29ce] = "\xf4\x75",
  [0x29cf] = "\xf4\x76", [0x29d0] = "\xf5\xe0", [0x29d3] = "\xc4\xef",
  [0x29d4] = "\xf7\xeb", [0x29d5] = "\xf8\xb4", [0x29d7] = "\xc5\xf7",
  [0x29d8] = "\xf8\xf8", [0x29d9] = "\xf8\xf9", [0x29da] = "\xc6\x66",
  [0x29db] = "\xa5\xd9", [0x29dc] = "\xac\xe1", [0x29de] = "\xda\xc3",
  [0x29e0] = "\xde\xe3", [0x29e2] = "\xa5\xda", [0x29e3] = "\xa8\x6f",
  [0x29e5] = "\xaa\xbe", [0x29e6] = "\xfa\xd8", [0x29e7] = "\xcf\xe8",
  [0x29e8] = "\xcf\xe9", [0x29e9] = "\xaf\x78", [0x29ec] = "\xda\xc4",
  [0x29ed] = "\xb5\x75", [0x29ee] = "\xb8\x47", [0x29ef] = "\xc1\x42",
  [0x29f0] = "\xed\xa4", [0x29f1] = "\xf2\x7c", [0x29f2] = "\xf4\x78",
  [0x29f3] = "\xa5\xdb", [0x29f4] = "\xfe\xdc", [0x29f7] = "\xcd\xa1",
  [0x29f8] = "\xcd\x7a", [0x29f9] = "\xcd\x7c", [0x29fa] = "\xcd\x7e",
  [0x29fb] = "\xcd\x7d", [0x29fc] = "\xcd\x7b", [0x29fd] = "\xaa\xbf",
  [0x29fe] = "\xa0\xae", [0x2a02] = "\xac\xe2", [0x2a03] = "\xcf\xf2",
  [0x2a05] = "\xcf\xed", [0x2a06] = "\xcf\xea", [0x2a08] = "\x9d\x4c",
  [0x2a09] = "\xcf\xf1", [0x2a0c] = "\xac\xe4", [0x2a0d] = "\xac\xe5",
  [0x2a0e] = "\xcf\xf0", [0x2a0f] = "\xcf\xef", [0x2a10] = "\xcf\xee",
  [0x2a11] = "\xcf\xeb", [0x2a12] = "\xcf\xec", [0x2a13] = "\xcf\xf3",
  [0x2a14] = "\xac\xe3", [0x2a18] = "\x98\xf1", [0x2a1c] = "\x98\xf3",
  [0x2a1d] = "\xaf\x7c", [0x2a1e] = "\x94\xc1", [0x2a1f] = "\xaf\xa4",
  [0x2a20] = "\xaf\xa3", [0x2a21] = "\xd2\xe1", [0x2a22] = "\xd2\xdb",
  [0x2a23] = "\xd2\xd9", [0x2a25] = "\xaf\xa1", [0x2a26] = "\xd6\xb9",
  [0x2a27] = "\xaf\x7a", [0x2a28] = "\xd2\xde", [0x2a29] = "\xd2\xe2",
  [0x2a2a] = "\xd2\xe4", [0x2a2b] = "\xd2\xe0", [0x2a2c] = "\xd2\xda",
  [0x2a2d] = "\xaf\xa2", [0x2a2e] = "\xd2\xdf", [0x2a2f] = "\xd2\xdd",
  [0x2a30] = "\xaf\x79", [0x2a31] = "\xd2\xe5", [0x2a32] = "\xaf\xa5",
  [0x2a33] = "\xd2\xe3", [0x2a34] = "\xaf\x7d", [0x2a35] = "\xd2\xdc",
  [0x2a37] = "\xaf\x7e", [0x2a38] = "\xaf\x7b", [0x2a39] = "\x98\xf5",
  [0x2a3c] = "\xfa\x4f", [0x2a3d] = "\x96\xe2", [0x2a42] = "\x94\x50",
  [0x2a43] = "\xb2\xb9", [0x2a44] = "\x96\xa2", [0x2a45] = "\xd6\xba",
  [0x2a47] = "\x98\xf6", [0x2a48] = "\xd6\xb3", [0x2a49] = "\xd6\xb5",
  [0x2a4a] = "\xd6\xb7", [0x2a4b] = "\x96\xe5", [0x2a4c] = "\xd6\xb8",
  [0x2a4d] = "\xd6\xb6", [0x2a4e] = "\xb2\xba", [0x2a50] = "\xd6\xbb",
  [0x2a51] = "\x98\xf7", [0x2a52] = "\xd6\xb4", [0x2a53] = "\xa0\x46",
  [0x2a54] = "\x96\xe3", [0x2a5c] = "\xda\xc8", [0x2a5d] = "\xb5\x76",
  [0x2a5e] = "\xda\xd0", [0x2a60] = "\xda\xc5", [0x2a62] = "\xda\xd1",
  [0x2a64] = "\xda\xc6", [0x2a65] = "\xda\xc7", [0x2a66] = "\x98\xf8",
  [0x2a68] = "\xda\xcf", [0x2a69] = "\xda\xce", [0x2a6a] = "\xda\xcb",
  [0x2a6b] = "\xb2\xb8", [0x2a6c] = "\xb5\x77", [0x2a6d] = "\xda\xc9",
  [0x2a6e] = "\xda\xcc", [0x2a6f] = "\xb5\x78", [0x2a70] = "\xda\xcd",
  [0x2a71] = "\xda\xca", [0x2a79] = "\xde\xee", [0x2a7a] = "\x9e\xe4",
  [0x2a7b] = "\xde\xf2", [0x2a7c] = "\xb8\x4e", [0x2a7e] = "\xe2\xf0",
  [0x2a7f] = "\xb8\x51", [0x2a80] = "\xde\xf0", [0x2a81] = "\xf9\xd6",
  [0x2a83] = "\xde\xed", [0x2a84] = "\xde\xe8", [0x2a85] = "\xde\xea",
  [0x2a86] = "\xde\xeb", [0x2a87] = "\xde\xe4", [0x2a88] = "\x94\xc3",
  [0x2a89] = "\xb8\x4d", [0x2a8c] = "\xb8\x4c", [0x2a8d] = "\x94\xc2",
  [0x2a8e] = "\xb8\x48", [0x2a8f] = "\xde\xe7", [0x2a91] = "\xb8\x4f",
  [0x2a93] = "\xb8\x50", [0x2a94] = "\xde\xe6", [0x2a95] = "\xde\xe9",
  [0x2a96] = "\xde\xf1", [0x2a97] = "\xb8\x4a", [0x2a98] = "\xb8\x4b",
  [0x2a99] = "\xde\xef", [0x2a9a] = "\xde\xe5", [0x2a9e] = "\xe2\xf2",
  [0x2a9f] = "\xba\xd0", [0x2aa0] = "\xe2\xf4", [0x2aa1] = "\xde\xec",
  [0x2aa2] = "\xe2\xf6", [0x2aa3] = "\xba\xd4", [0x2aa4] = "\xe2\xf7",
  [0x2aa5] = "\xe2\xf3", [0x2aa7] = "\xba\xd1", [0x2aa8] = "\xe2\xef",
  [0x2aa9] = "\xba\xd3", [0x2aaa] = "\xe2\xec", [0x2aab] = "\xe2\xf1",
  [0x2aac] = "\xe2\xf5", [0x2aad] = "\xe2\xee", [0x2aaf] = "\xfe\xe1",
  [0x2ab0] = "\xb8\x49", [0x2ab1] = "\xfe\xe9", [0x2ab2] = "\xe2\xeb",
  [0x2ab3] = "\xba\xd2", [0x2ab4] = "\xe2\xed", [0x2ab6] = "\x96\xe4",
  [0x2ab8] = "\x89\xac", [0x2ab9] = "\x96\xdb", [0x2aba] = "\xbd\x54",
  [0x2abb] = "\xe6\xc1", [0x2abc] = "\xbd\x58", [0x2abe] = "\xbd\x56",
  [0x2ac1] = "\xba\xcf", [0x2ac3] = "\xe6\xc8", [0x2ac4] = "\xe6\xc9",
  [0x2ac5] = "\xbd\x53", [0x2ac7] = "\xfe\xe2", [0x2ac8] = "\xe6\xc7",
  [0x2ac9] = "\xe6\xca", [0x2aca] = "\xbd\x55", [0x2acb] = "\xbd\x52",
  [0x2acc] = "\xe6\xc3", [0x2acd] = "\xe6\xc0", [0x2ace] = "\xe6\xc5",
  [0x2acf] = "\xe6\xc2", [0x2ad0] = "\xbd\x59", [0x2ad1] = "\xe6\xc4",
  [0x2ad2] = "\x94\xc4", [0x2ad3] = "\xfe\xe3", [0x2ad4] = "\xe6\xc6",
  [0x2ad5] = "\xbd\x57", [0x2ad7] = "\xfe\xe7", [0x2ad8] = "\x9f\xfb",
  [0x2ada] = "\xbf\x6a", [0x2adb] = "\xea\xa8", [0x2add] = "\xea\xa2",
  [0x2ade] = "\xea\xa6", [0x2adf] = "\xea\xac", [0x2ae0] = "\xea\xad",
  [0x2ae1] = "\xea\xa9", [0x2ae2] = "\xea\xaa", [0x2ae3] = "\xea\xa7",
  [0x2ae5] = "\xea\xa4", [0x2ae7] = "\xbf\x6c", [0x2ae8] = "\xbf\x69",
  [0x2ae9] = "\xea\xa3", [0x2aea] = "\xea\xa5", [0x2aec] = "\xbf\x6b",
  [0x2aed] = "\xea\xab", [0x2aee] = "\x93\xc9", [0x2aef] = "\xc1\x46",
  [0x2af0] = "\x94\xe8", [0x2af1] = "\xfb\x56", [0x2af2] = "\xed\xaa",
  [0x2af3] = "\xed\xa5", [0x2af4] = "\xc1\x45", [0x2af5] = "\x90\xc5",
  [0x2af7] = "\xc1\x43", [0x2af9] = "\xed\xac", [0x2afa] = "\xc1\x44",
  [0x2afb] = "\xed\xa8", [0x2afc] = "\xed\xa9", [0x2afd] = "\xed\xa6",
  [0x2afe] = "\xed\xad", [0x2aff] = "\xf0\x56", [0x2b01] = "\xc1\x47",
  [0x2b02] = "\xed\xa7", [0x2b04] = "\xed\xae", [0x2b05] = "\xed\xab",
  [0x2b06] = "\xa0\xa8", [0x2b09] = "\xf0\x5a", [0x2b0c] = "\xf0\x57",
  [0x2b0e] = "\xc2\xa6", [0x2b10] = "\xf0\x5b", [0x2b11] = "\xf0\x5d",
  [0x2b12] = "\xf0\x5c", [0x2b13] = "\xf0\x58", [0x2b14] = "\xf0\x59",
  [0x2b17] = "\xf2\xa3", [0x2b19] = "\xc3\xaa", [0x2b1b] = "\xf2\x7e",
  [0x2b1c] = "\xf2\xa2", [0x2b1d] = "\xf2\x7d", [0x2b1e] = "\xf2\xa4",
  [0x2b21] = "\xf2\xa1", [0x2b23] = "\xf4\x7a", [0x2b24] = "\xf4\x7d",
  [0x2b25] = "\xf4\x79", [0x2b26] = "\xc4\x71", [0x2b27] = "\xf4\x7b",
  [0x2b28] = "\xf4\x7c", [0x2b29] = "\xf4\x7e", [0x2b2a] = "\xc4\x72",
  [0x2b2b] = "\xc4\x74", [0x2b2c] = "\xc4\x73", [0x2b2d] = "\xf5\xe1",
  [0x2b2e] = "\xfe\xe5", [0x2b2f] = "\xf5\xe3", [0x2b31] = "\xf5\xe2",
  [0x2b32] = "\x98\xfd", [0x2b33] = "\x98\xfb", [0x2b34] = "\xfe\xe8",
  [0x2b35] = "\xf6\xf6", [0x2b36] = "\x8e\xbf", [0x2b38] = "\xf8\xb5",
  [0x2b39] = "\xf8\xfa", [0x2b3a] = "\xa5\xdc", [0x2b3b] = "\x8b\xd8",
  [0x2b3c] = "\xfe\xf7", [0x2b3d] = "\xcb\x72", [0x2b3e] = "\xaa\xc0",
  [0x2b3f] = "\xcd\xa3", [0x2b40] = "\xaa\xc1", [0x2b41] = "\xaa\xc2",
  [0x2b42] = "\xcd\xa2", [0x2b44] = "\xcf\xf8", [0x2b45] = "\xcf\xf7",
  [0x2b46] = "\xac\xe6", [0x2b47] = "\xac\xe9", [0x2b48] = "\xac\xe8",
  [0x2b49] = "\xac\xe7", [0x2b4a] = "\xcf\xf4", [0x2b4b] = "\xcf\xf6",
  [0x2b4c] = "\xcf\xf5", [0x2b4f] = "\xd2\xe8", [0x2b50] = "\xaf\xa7",
  [0x2b51] = "\xd2\xec", [0x2b52] = "\xd2\xeb", [0x2b53] = "\xd2\xea",
  [0x2b54] = "\xd2\xe6", [0x2b55] = "\xaf\xa6", [0x2b56] = "\xaf\xaa",
  [0x2b57] = "\xaf\xad", [0x2b58] = "\x8f\x68", [0x2b59] = "\x94\xc6",
  [0x2b5a] = "\xaf\xae", [0x2b5b] = "\xd2\xe7", [0x2b5c] = "\xd2\xe9",
  [0x2b5d] = "\xaf\xac", [0x2b5e] = "\xaf\xab", [0x2b5f] = "\xaf\xa9",
  [0x2b60] = "\xaf\xa8", [0x2b61] = "\xd6\xc2", [0x2b62] = "\x9d\xea",
  [0x2b63] = "\xd6\xc0", [0x2b64] = "\xd6\xbc", [0x2b65] = "\xb2\xbb",
  [0x2b67] = "\xd6\xbd", [0x2b68] = "\xb2\xbc", [0x2b69] = "\xd6\xbe",
  [0x2b6a] = "\xd6\xbf", [0x2b6b] = "\xd6\xc1", [0x2b6d] = "\xb2\xbd",
  [0x2b70] = "\xda\xd5", [0x2b71] = "\xfc\x69", [0x2b72] = "\xda\xd4",
  [0x2b73] = "\xda\xd3", [0x2b74] = "\xda\xd2", [0x2b79] = "\xde\xf6",
  [0x2b7a] = "\xb8\x52", [0x2b7c] = "\xde\xf3", [0x2b7d] = "\xde\xf5",
  [0x2b7e] = "\x9c\xda", [0x2b7f] = "\xb8\x53", [0x2b80] = "\xfe\xf3",
  [0x2b81] = "\xb8\x54", [0x2b82] = "\xde\xf4", [0x2b83] = "\x9c\x72",
  [0x2b86] = "\xfe\xf0", [0x2b87] = "\x89\xc9", [0x2b88] = "\xe3\x41",
  [0x2b8a] = "\xe2\xf9", [0x2b8b] = "\xe2\xfa", [0x2b8d] = "\xba\xd7",
  [0x2b8e] = "\xba\xd5", [0x2b8f] = "\xba\xd6", [0x2b90] = "\xe3\x43",
  [0x2b91] = "\x99\x41", [0x2b92] = "\xe3\x42", [0x2b93] = "\xe2\xfe",
  [0x2b94] = "\xe2\xfd", [0x2b95] = "\xe2\xfc", [0x2b96] = "\xe2\xfb",
  [0x2b97] = "\xe3\x40", [0x2b98] = "\xe2\xf8", [0x2b99] = "\x99\x42",
  [0x2b9a] = "\xe6\xcb", [0x2b9b] = "\xe6\xd0", [0x2b9c] = "\xe6\xce",
  [0x2b9d] = "\xfe\xf5", [0x2b9f] = "\x91\xd7", [0x2ba0] = "\xe6\xcd",
  [0x2ba1] = "\xe6\xcc", [0x2ba2] = "\xe6\xcf", [0x2ba4] = "\xea\xae",
  [0x2ba5] = "\x94\xcc", [0x2ba6] = "\xbf\x6d", [0x2ba7] = "\xc1\x48",
  [0x2ba8] = "\xed\xb0", [0x2ba9] = "\xfe\xf8", [0x2baa] = "\xc1\x49",
  [0x2bab] = "\xed\xaf", [0x2bac] = "\xf0\x5f", [0x2bad] = "\xf0\x5e",
  [0x2bae] = "\xc2\xa7", [0x2bb0] = "\xf2\xa5", [0x2bb1] = "\xc3\xab",
  [0x2bb2] = "\xf4\xa1", [0x2bb3] = "\xc5\xa1", [0x2bb4] = "\xf6\xf7",
  [0x2bb6] = "\xf8\xb7", [0x2bb7] = "\xf8\xb6", [0x2bb8] = "\xc9\xa8",
  [0x2bb9] = "\xac\xea", [0x2bba] = "\xac\xeb", [0x2bbb] = "\xd6\xc3",
  [0x2bbd] = "\xb8\x56", [0x2bbe] = "\xa5\xdd", [0x2bbf] = "\xa8\x72",
  [0x2bc0] = "\xa8\x71", [0x2bc1] = "\xa8\x70", [0x2bc4] = "\x97\xa8",
  [0x2bc5] = "\xcd\xa4", [0x2bc6] = "\xfe\xfc", [0x2bc8] = "\xaa\xc4",
  [0x2bc9] = "\xaa\xc3", [0x2bcb] = "\xac\xee", [0x2bcc] = "\xfd\xbf",
  [0x2bcd] = "\xcf\xfa", [0x2bce] = "\xcf\xfd", [0x2bcf] = "\xcf\xfb",
  [0x2bd1] = "\xac\xec", [0x2bd2] = "\xac\xed", [0x2bd4] = "\xfe\xfe",
  [0x2bd5] = "\xcf\xf9", [0x2bd6] = "\xcf\xfc", [0x2bd8] = "\xaf\xb5",
  [0x2bdc] = "\xd2\xf3", [0x2bdd] = "\xd2\xf5", [0x2bde] = "\xd2\xf4",
  [0x2bdf] = "\xaf\xb2", [0x2be0] = "\xd2\xef", [0x2be2] = "\x96\xd1",
  [0x2be3] = "\xaf\xb0", [0x2be4] = "\xaf\xaf", [0x2be6] = "\xaf\xb3",
  [0x2be7] = "\xaf\xb1", [0x2be9] = "\xaf\xb4", [0x2bea] = "\xd2\xf2",
  [0x2beb] = "\xd2\xed", [0x2bec] = "\xd2\xee", [0x2bed] = "\xd2\xf1",
  [0x2bee] = "\xd2\xf0", [0x2bf1] = "\x94\xd5", [0x2bf4] = "\x94\xd0",
  [0x2bf6] = "\xd6\xc6", [0x2bf7] = "\xd6\xc7", [0x2bf8] = "\xd6\xc5",
  [0x2bfa] = "\xd6\xc4", [0x2bfb] = "\xb2\xbe", [0x2c00] = "\xb5\x7d",
  [0x2c02] = "\xda\xd6", [0x2c03] = "\xda\xd8", [0x2c04] = "\xda\xda",
  [0x2c05] = "\xb5\x7c", [0x2c06] = "\x99\x44", [0x2c08] = "\xb5\x7a",
  [0x2c0a] = "\xda\xd7", [0x2c0b] = "\xb5\x7b", [0x2c0c] = "\xda\xd9",
  [0x2c0d] = "\xb5\x79", [0x2c10] = "\xdf\x41", [0x2c11] = "\xde\xf7",
  [0x2c12] = "\xde\xfa", [0x2c13] = "\xde\xfe", [0x2c14] = "\xb8\x5a",
  [0x2c15] = "\xde\xfc", [0x2c17] = "\xde\xfb", [0x2c18] = "\xde\xf8",
  [0x2c19] = "\xde\xf9", [0x2c1a] = "\xb8\x58", [0x2c1b] = "\xdf\x40",
  [0x2c1c] = "\xb8\x57", [0x2c1e] = "\xb8\x5c", [0x2c1f] = "\xb8\x5b",
  [0x2c20] = "\xb8\x59", [0x2c22] = "\xde\xfd", [0x2c26] = "\xe3\x49",
  [0x2c28] = "\xe3\x48", [0x2c2b] = "\xe3\x44", [0x2c2d] = "\xa0\xb3",
  [0x2c2e] = "\xba\xd8", [0x2c2f] = "\xe3\x47", [0x2c30] = "\xe3\x46",
  [0x2c31] = "\xba\xd9", [0x2c37] = "\xbd\x5e", [0x2c39] = "\xe6\xd2",
  [0x2c3a] = "\x94\xcf", [0x2c3b] = "\xbd\x5f", [0x2c3c] = "\xbd\x5b",
  [0x2c3d] = "\xbd\x5d", [0x2c3e] = "\x9f\xfa", [0x2c3f] = "\xbd\x5a",
  [0x2c40] = "\xbd\x5c", [0x2c43] = "\x91\xe5", [0x2c44] = "\xea\xaf",
  [0x2c45] = "\x9c\x6a", [0x2c46] = "\xbf\x70", [0x2c47] = "\xea\xb1",
  [0x2c48] = "\xea\xb0", [0x2c49] = "\x8e\x49", [0x2c4a] = "\xe3\x45",
  [0x2c4b] = "\xbf\x72", [0x2c4c] = "\xbf\x71", [0x2c4d] = "\xbf\x6e",
  [0x2c4e] = "\xbf\x6f", [0x2c54] = "\xed\xb5", [0x2c56] = "\xed\xb3",
  [0x2c57] = "\xc1\x4a", [0x2c58] = "\xed\xb4", [0x2c5a] = "\xed\xb6",
  [0x2c5b] = "\xed\xb2", [0x2c5c] = "\xed\xb1", [0x2c5f] = "\xf0\x60",
  [0x2c60] = "\xc2\xaa", [0x2c61] = "\xc2\xa8", [0x2c62] = "\xc2\xa9",
  [0x2c65] = "\x8e\x4c", [0x2c67] = "\xf2\xa6", [0x2c68] = "\xf2\xa7",
  [0x2c69] = "\xc3\xad", [0x2c6b] = "\xc3\xac", [0x2c6c] = "\xf4\xa3",
  [0x2c6d] = "\xf4\xa4", [0x2c6e] = "\xf4\xa2", [0x2c70] = "\xf6\xf8",
  [0x2c71] = "\xf6\xf9", [0x2c74] = "\xa5\xde", [0x2c75] = "\xca\x48",
  [0x2c76] = "\xa8\x73", [0x2c78] = "\xcd\xa5", [0x2c79] = "\xaa\xc6",
  [0x2c7a] = "\xaa\xc5", [0x2c7b] = "\xcd\xa6", [0x2c7d] = "\x8e\x4d",
  [0x2c7e] = "\xd0\x40", [0x2c7f] = "\xac\xef", [0x2c80] = "\xcf\xfe",
  [0x2c81] = "\xac\xf0", [0x2c83] = "\x9a\x73", [0x2c84] = "\xaf\xb6",
  [0x2c85] = "\xd2\xf8", [0x2c86] = "\xd2\xf6", [0x2c87] = "\xd2\xfc",
  [0x2c88] = "\xaf\xb7", [0x2c89] = "\xd2\xf7", [0x2c8a] = "\xd2\xfb",
  [0x2c8b] = "\xd2\xf9", [0x2c8c] = "\xd2\xfa", [0x2c8f] = "\xd6\xc8",
  [0x2c90] = "\xd6\xca", [0x2c91] = "\x99\x47", [0x2c92] = "\xb2\xbf",
  [0x2c94] = "\xd6\xc9", [0x2c95] = "\xb2\xc0", [0x2c96] = "\xb5\xa2",
  [0x2c97] = "\xb5\xa1", [0x2c98] = "\xb5\x7e", [0x2c99] = "\xda\xdb",
  [0x2c9e] = "\xdf\x44", [0x2c9f] = "\xb8\x5d", [0x2ca0] = "\xb8\x5e",
  [0x2ca2] = "\xdf\x43", [0x2ca3] = "\xdf\x42", [0x2ca8] = "\xe3\x4a",
  [0x2ca9] = "\xba\xdb", [0x2caa] = "\xba\xda", [0x2cab] = "\xe3\x4b",
  [0x2cac] = "\xe3\x4c", [0x2cae] = "\xbd\x61", [0x2caf] = "\xbd\x60",
  [0x2cb0] = "\x8e\x50", [0x2cb1] = "\xea\xb5", [0x2cb2] = "\xe6\xd3",
  [0x2cb3] = "\xe6\xd5", [0x2cb4] = "\xe6\xd4", [0x2cb5] = "\xea\xb4",
  [0x2cb6] = "\xea\xb2", [0x2cb7] = "\xea\xb6", [0x2cb8] = "\xea\xb3",
  [0x2cba] = "\xbf\x73", [0x2cbb] = "\x8e\x4f", [0x2cbc] = "\x99\x49",
  [0x2cbe] = "\xed\xb7", [0x2cbf] = "\xc1\x4b", [0x2cc0] = "\xed\xb8",
  [0x2cc1] = "\xed\xb9", [0x2cc2] = "\x8e\x51", [0x2cc3] = "\x8e\x52",
  [0x2cc4] = "\xc2\xab", [0x2cc5] = "\xc2\xac", [0x2cc7] = "\xc4\x75",
  [0x2cc8] = "\x9a\xb2", [0x2cc9] = "\x89\xa5", [0x2cca] = "\xc5\xd1",
  [0x2ccb] = "\xa5\xdf", [0x2ccf] = "\x99\x4c", [0x2cd1] = "\xd0\x41",
  [0x2cd3] = "\x9f\xf8", [0x2cd8] = "\xd2\xfd", [0x2cd9] = "\xaf\xb8",
  [0x2cda] = "\x8e\x56", [0x2cdb] = "\x99\x4d", [0x2cdc] = "\x91\xca",
  [0x2cdd] = "\x8e\x57", [0x2cdf] = "\xb3\xba", [0x2ce0] = "\xb3\xb9",
  [0x2ce2] = "\x94\xe1", [0x2ce3] = "\xb5\xa4", [0x2ce4] = "\xda\xdd",
  [0x2ce5] = "\xb5\xa3", [0x2ce6] = "\xda\xdc", [0x2ce7] = "\x90\x47",
  [0x2ce9] = "\x8f\xd8", [0x2cea] = "\x8e\x58", [0x2ceb] = "\xdf\x45",
  [0x2ced] = "\xba\xdc", [0x2cee] = "\xe3\x4d", [0x2cef] = "\xba\xdd",
  [0x2cf6] = "\xc4\x76", [0x2cf7] = "\xf4\xa5", [0x2cf9] = "\xa6\xcb",
  [0x2cfa] = "\xaa\xc7", [0x2cfb] = "\xcd\xa7", [0x2cfd] = "\xac\xf2",
  [0x2cfe] = "\x94\xeb", [0x2cff] = "\xac\xf1", [0x2d00] = "\xd0\x42",
  [0x2d01] = "\xd0\x43", [0x2d04] = "\xd3\x40", [0x2d05] = "\xd3\x42",
  [0x2d06] = "\xaf\xb9", [0x2d08] = "\xd3\x44", [0x2d09] = "\xd3\x47",
  [0x2d0a] = "\xd3\x45", [0x2d0b] = "\x8e\x5c", [0x2d0c] = "\x95\x53",
  [0x2d0e] = "\xd3\x46", [0x2d0f] = "\xd3\x43", [0x2d10] = "\xd2\xfe",
  [0x2d11] = "\xaf\xba", [0x2d12] = "\xd3\x48", [0x2d13] = "\xd3\x41",
  [0x2d14] = "\x9f\xe5", [0x2d18] = "\xd6\xd3", [0x2d19] = "\xb2\xc6",
  [0x2d1a] = "\xd6\xdc", [0x2d1b] = "\xb2\xc3", [0x2d1d] = "\xd6\xd5",
  [0x2d1e] = "\xb2\xc7", [0x2d1f] = "\x9f\x56", [0x2d20] = "\xb2\xc1",
  [0x2d22] = "\xd6\xd0", [0x2d23] = "\xd6\xdd", [0x2d24] = "\xd6\xd1",
  [0x2d25] = "\xd6\xce", [0x2d26] = "\xb2\xc5", [0x2d27] = "\x95\x4f",
  [0x2d28] = "\xb2\xc2", [0x2d29] = "\x8e\x5e", [0x2d2a] = "\xd6\xd4",
  [0x2d2b] = "\xd6\xd7", [0x2d2c] = "\xb2\xc4", [0x2d2d] = "\xd6\xd8",
  [0x2d2e] = "\xb2\xc8", [0x2d2f] = "\xd6\xd9", [0x2d30] = "\xd6\xcf",
  [0x2d31] = "\xd6\xd6", [0x2d32] = "\xd6\xda", [0x2d33] = "\xd6\xd2",
  [0x2d34] = "\xd6\xcd", [0x2d35] = "\xd6\xcb", [0x2d38] = "\xd6\xdb",
  [0x2d39] = "\x99\x6a", [0x2d3b] = "\xda\xdf", [0x2d40] = "\xda\xe4",
  [0x2d42] = "\x9c\x64", [0x2d43] = "\x9c\xd9", [0x2d44] = "\xda\xe0",
  [0x2d45] = "\xda\xe6", [0x2d46] = "\xb5\xa7", [0x2d47] = "\xd6\xcc",
  [0x2d48] = "\xda\xe1", [0x2d49] = "\xb5\xa5", [0x2d4a] = "\xda\xde",
  [0x2d4b] = "\xb5\xac", [0x2d4c] = "\xda\xe2", [0x2d4d] = "\xb5\xab",
  [0x2d4e] = "\xda\xe3", [0x2d4f] = "\xb5\xad", [0x2d50] = "\xb5\xa8",
  [0x2d51] = "\xb5\xae", [0x2d52] = "\xb5\xa9", [0x2d54] = "\xb5\xaa",
  [0x2d55] = "\x8e\x5d", [0x2d56] = "\xb5\xa6", [0x2d58] = "\xda\xe5",
  [0x2d60] = "\xb8\x61", [0x2d61] = "\xdf\x50", [0x2d62] = "\x99\x50",
  [0x2d63] = "\xdf\x53", [0x2d64] = "\xdf\x47", [0x2d65] = "\xdf\x4c",
  [0x2d66] = "\xdf\x46", [0x2d67] = "\xb8\x63", [0x2d69] = "\xdf\x4a",
  [0x2d6c] = "\x99\x51", [0x2d6d] = "\xdf\x48", [0x2d6e] = "\xb8\x62",
  [0x2d6f] = "\x8e\x62", [0x2d70] = "\xdf\x4f", [0x2d71] = "\xdf\x4e",
  [0x2d72] = "\xdf\x4b", [0x2d73] = "\xdf\x4d", [0x2d74] = "\xdf\x49",
  [0x2d75] = "\xba\xe1", [0x2d76] = "\xdf\x52", [0x2d77] = "\xb8\x5f",
  [0x2d78] = "\xdf\x51", [0x2d7b] = "\x99\x52", [0x2d82] = "\xe3\x5d",
  [0x2d84] = "\xba\xe8", [0x2d85] = "\xe3\x58", [0x2d87] = "\xba\xe7",
  [0x2d88] = "\xe3\x4e", [0x2d8a] = "\xe3\x50", [0x2d8b] = "\xba\xe0",
  [0x2d8c] = "\xe3\x55", [0x2d8d] = "\xe3\x54", [0x2d8e] = "\xe3\x57",
  [0x2d8f] = "\xba\xe5", [0x2d90] = "\xe3\x52", [0x2d91] = "\xe3\x51",
  [0x2d92] = "\x8e\x68", [0x2d94] = "\xba\xe4", [0x2d95] = "\xba\xdf",
  [0x2d96] = "\xe3\x53", [0x2d97] = "\xba\xe2", [0x2d98] = "\xe3\x59",
  [0x2d99] = "\xe3\x5b", [0x2d9b] = "\xe3\x56", [0x2d9c] = "\xe3\x4f",
  [0x2d9d] = "\xba\xe3", [0x2da0] = "\xbd\x69", [0x2da1] = "\xba\xde",
  [0x2da2] = "\x8e\x61", [0x2da3] = "\x9f\x59", [0x2da4] = "\xe3\x5c",
  [0x2dac] = "\xe6\xd9", [0x2dad] = "\xbd\x62", [0x2daf] = "\xe6\xdb",
  [0x2db1] = "\xbd\x63", [0x2db2] = "\x8b\xb3", [0x2db4] = "\xbd\x65",
  [0x2db5] = "\xe6\xde", [0x2db7] = "\xe6\xd6", [0x2db8] = "\xba\xe6",
  [0x2db9] = "\xe6\xdc", [0x2dbe] = "\xe6\xd8", [0x2dc0] = "\xb8\x60",
  [0x2dc1] = "\xbd\x68", [0x2dc4] = "\xbd\x64", [0x2dc6] = "\xbd\x66",
  [0x2dc7] = "\xbd\x67", [0x2dc9] = "\xbf\x76", [0x2dca] = "\xe6\xdd",
  [0x2dcb] = "\xe6\xd7", [0x2dcc] = "\xbd\x6a", [0x2dce] = "\xe6\xda",
  [0x2dcf] = "\x9f\x5d", [0x2dd0] = "\x8e\x66", [0x2dd4] = "\xea\xc0",
  [0x2dd5] = "\xea\xbb", [0x2dd8] = "\xea\xc5", [0x2dd9] = "\xbf\x74",
  [0x2dda] = "\xea\xbd", [0x2ddb] = "\xbf\x78", [0x2ddc] = "\xea\xc3",
  [0x2ddd] = "\xea\xba", [0x2dde] = "\xea\xb7", [0x2ddf] = "\xea\xc6",
  [0x2de0] = "\xc1\x51", [0x2de1] = "\xbf\x79", [0x2de2] = "\xea\xc2",
  [0x2de3] = "\xea\xb8", [0x2de4] = "\xbf\x77", [0x2de5] = "\xea\xbc",
  [0x2de6] = "\xbf\x7b", [0x2de7] = "\xea\xb9", [0x2de8] = "\xea\xbe",
  [0x2de9] = "\xbf\x7a", [0x2dea] = "\xea\xc1", [0x2deb] = "\xea\xc4",
  [0x2df0] = "\xed\xcb", [0x2df1] = "\xed\xcc", [0x2df2] = "\xed\xbc",
  [0x2df3] = "\xed\xc3", [0x2df4] = "\xed\xc1", [0x2df7] = "\xc1\x4f",
  [0x2df8] = "\xed\xc8", [0x2df9] = "\xea\xbf", [0x2dfa] = "\x8e\x6e",
  [0x2dfb] = "\xed\xbf", [0x2dfc] = "\x9f\x64", [0x2dfd] = "\xed\xc9",
  [0x2dfe] = "\xc1\x4e", [0x2dff] = "\xed\xbe", [0x2e00] = "\xed\xbd",
  [0x2e01] = "\xed\xc7", [0x2e02] = "\xed\xc4", [0x2e03] = "\xed\xc6",
  [0x2e05] = "\xed\xba", [0x2e06] = "\xed\xca", [0x2e07] = "\xc1\x4c",
  [0x2e09] = "\xed\xc5", [0x2e0a] = "\xed\xce", [0x2e0b] = "\xed\xc2",
  [0x2e0c] = "\xc1\x50", [0x2e0d] = "\xc1\x4d", [0x2e0e] = "\xed\xc0",
  [0x2e0f] = "\xed\xbb", [0x2e10] = "\xed\xcd", [0x2e11] = "\xbf\x75",
  [0x2e12] = "\x99\x53", [0x2e15] = "\xfa\xb8", [0x2e19] = "\xf0\x63",
  [0x2e1b] = "\x99\x54", [0x2e1c] = "\xf0\x61", [0x2e1d] = "\xf0\x67",
  [0x2e1e] = "\xc2\xb0", [0x2e1f] = "\xf0\x65", [0x2e20] = "\xf0\x64",
  [0x2e21] = "\xc2\xb2", [0x2e22] = "\xf0\x6a", [0x2e23] = "\xc2\xb1",
  [0x2e25] = "\xf0\x6b", [0x2e26] = "\xf0\x68", [0x2e27] = "\xc2\xae",
  [0x2e28] = "\xf0\x69", [0x2e29] = "\xf0\x62", [0x2e2a] = "\xc2\xaf",
  [0x2e2b] = "\xc2\xad", [0x2e2c] = "\xf2\xab", [0x2e2d] = "\xf0\x66",
  [0x2e30] = "\xf0\x6c", [0x2e33] = "\xf2\xa8", [0x2e35] = "\x8e\x70",
  [0x2e37] = "\xc3\xb2", [0x2e38] = "\xc3\xb0", [0x2e39] = "\xf2\xaa",
  [0x2e3b] = "\xf2\xac", [0x2e3c] = "\xf2\xa9", [0x2e3d] = "\xc3\xb1",
  [0x2e3e] = "\xc3\xae", [0x2e3f] = "\xc3\xaf", [0x2e40] = "\xc3\xb3",
  [0x2e42] = "\x9f\x61", [0x2e43] = "\xc4\x78", [0x2e44] = "\x8e\x72",
  [0x2e45] = "\xf4\xaa", [0x2e47] = "\xf4\xa9", [0x2e48] = "\xf4\xa7",
  [0x2e49] = "\xf4\xa6", [0x2e4a] = "\xf4\xa8", [0x2e4c] = "\xc4\x77",
  [0x2e4d] = "\xc4\x79", [0x2e50] = "\xc4\xf0", [0x2e51] = "\xa0\x6b",
  [0x2e53] = "\xf5\xe5", [0x2e54] = "\xf5\xe4", [0x2e56] = "\x9f\x40",
  [0x2e57] = "\xf6\xfa", [0x2e59] = "\xf6\xfc", [0x2e5a] = "\xf6\xfe",
  [0x2e5b] = "\xf6\xfd", [0x2e5c] = "\xf6\xfb", [0x2e5d] = "\x94\xed",
  [0x2e5f] = "\xc5\xa3", [0x2e60] = "\xc5\xa2", [0x2e63] = "\xc5\xd3",
  [0x2e64] = "\xc5\xd2", [0x2e65] = "\xc5\xd4", [0x2e66] = "\xf7\xed",
  [0x2e67] = "\xf7\xec", [0x2e69] = "\xf8\xfb", [0x2e6a] = "\xf8\xb8",
  [0x2e6b] = "\xf8\xfc", [0x2e6c] = "\xc6\x58", [0x2e6d] = "\x94\xee",
  [0x2e6e] = "\xc6\x59", [0x2e6f] = "\xf9\x6d", [0x2e70] = "\x9f\xbd",
  [0x2e72] = "\xc6\x7e", [0x2e73] = "\xa6\xcc", [0x2e74] = "\x8e\x7b",
  [0x2e75] = "\xcd\xa8", [0x2e78] = "\xd0\x45", [0x2e79] = "\xd0\x46",
  [0x2e7a] = "\xd0\x44", [0x2e7b] = "\x99\x57", [0x2e7c] = "\x94\xf7",
  [0x2e7d] = "\xac\xf3", [0x2e7e] = "\x9f\x5f", [0x2e7f] = "\xd0\x47",
  [0x2e80] = "\xd0\x48", [0x2e81] = "\xd0\x49", [0x2e83] = "\x8e\x73",
  [0x2e84] = "\xd3\x49", [0x2e85] = "\xd3\x4f", [0x2e86] = "\x9f\x62",
  [0x2e88] = "\xd3\x4d", [0x2e89] = "\xaf\xbb", [0x2e8a] = "\xd3\x4b",
  [0x2e8c] = "\xd3\x4c", [0x2e8d] = "\xd3\x4e", [0x2e8e] = "\x94\xf6",
  [0x2e91] = "\xd3\x4a", [0x2e92] = "\xb2\xc9", [0x2e94] = "\xd6\xde",
  [0x2e95] = "\xb2\xcb", [0x2e96] = "\xd6\xe0", [0x2e97] = "\xb2\xca",
  [0x2e98] = "\xd6\xdf", [0x2e9c] = "\x99\x58", [0x2e9e] = "\xda\xe8",
  [0x2e9f] = "\xb5\xaf", [0x2ea1] = "\xda\xea", [0x2ea2] = "\xda\xe7",
  [0x2ea3] = "\xd6\xe1", [0x2ea5] = "\xb5\xb0", [0x2ea6] = "\x8e\x75",
  [0x2ea7] = "\xf9\xdb", [0x2ea8] = "\xda\xe9", [0x2eac] = "\x90\x72",
  [0x2eae] = "\x94\xf8", [0x2eaf] = "\xdf\x56", [0x2eb1] = "\xb8\x64",
  [0x2eb2] = "\xdf\x54", [0x2eb3] = "\xb8\x65", [0x2eb4] = "\xdf\x55",
  [0x2eb5] = "\xb8\x66", [0x2eb8] = "\x99\x5a", [0x2eb9] = "\xba\xe9",
  [0x2eba] = "\xe3\x61", [0x2ebb] = "\xe3\x5e", [0x2ebc] = "\xe3\x60",
  [0x2ebd] = "\xba\xea", [0x2ebe] = "\xba\xeb", [0x2ebf] = "\xe3\x5f",
  [0x2ec2] = "\xa0\xb0", [0x2ec5] = "\xe6\xdf", [0x2ec7] = "\x8e\x79",
  [0x2ec8] = "\xe6\xe0", [0x2ec9] = "\x8e\x78", [0x2eca] = "\xbd\x6b",
  [0x2ecb] = "\xe6\xe2", [0x2ecc] = "\xe6\xe1", [0x2ecd] = "\x94\xf3",
  [0x2ece] = "\xa2\x61", [0x2ed0] = "\xea\xca", [0x2ed1] = "\xea\xcb",
  [0x2ed2] = "\xea\xc7", [0x2ed3] = "\x98\xaf", [0x2ed4] = "\xea\xc8",
  [0x2ed5] = "\xbf\x7c", [0x2ed6] = "\xbf\x7d", [0x2ed7] = "\xea\xc9",
  [0x2ed9] = "\xc1\x57", [0x2eda] = "\xa0\xb2", [0x2edc] = "\xc1\x53",
  [0x2edd] = "\xc1\x58", [0x2ede] = "\xc1\x54", [0x2edf] = "\xc1\x56",
  [0x2ee0] = "\xc1\x52", [0x2ee2] = "\xc1\x55", [0x2ee6] = "\x8e\x7a",
  [0x2ee7] = "\xc2\xb3", [0x2ee8] = "\xed\xcf", [0x2eea] = "\xf2\xae",
  [0x2eec] = "\xf2\xad", [0x2eed] = "\x99\x5c", [0x2eee] = "\xf4\xab",
  [0x2eef] = "\xc4\x7a", [0x2ef0] = "\xc4\x7b", [0x2ef1] = "\xf7\x41",
  [0x2ef2] = "\xf5\xe6", [0x2ef3] = "\x8e\x7c", [0x2ef4] = "\xf7\x40",
  [0x2ef5] = "\x8e\x7d", [0x2ef6] = "\xf8\xfd", [0x2ef7] = "\xf9\xa4",
  [0x2ef8] = "\xa6\xcd", [0x2ef9] = "\x8b\xd9", [0x2efb] = "\xa8\x74",
  [0x2efc] = "\x89\xa2", [0x2efd] = "\xcd\xa9", [0x2efe] = "\xaa\xc8",
  [0x2f00] = "\xac\xf6", [0x2f01] = "\xd0\x4c", [0x2f02] = "\xac\xf4",
  [0x2f03] = "\xd0\x4a", [0x2f04] = "\xac\xf9", [0x2f05] = "\xac\xf5",
  [0x2f06] = "\xac\xfa", [0x2f07] = "\xac\xf8", [0x2f08] = "\xd0\x4b",
  [0x2f09] = "\xac\xf7", [0x2f0a] = "\xaf\xbf", [0x2f0b] = "\xaf\xbe",
  [0x2f0c] = "\xd3\x5a", [0x2f0d] = "\xaf\xc7", [0x2f0e] = "\xd3\x53",
  [0x2f0f] = "\xd3\x59", [0x2f10] = "\xaf\xc3", [0x2f11] = "\xd3\x52",
  [0x2f12] = "\xd3\x58", [0x2f13] = "\xd3\x56", [0x2f14] = "\xaf\xc2",
  [0x2f15] = "\xaf\xc4", [0x2f16] = "\xd3\x55", [0x2f17] = "\xaf\xbd",
  [0x2f18] = "\xd3\x54", [0x2f19] = "\xaf\xc8", [0x2f1a] = "\xaf\xc5",
  [0x2f1b] = "\xaf\xc9", [0x2f1c] = "\xaf\xc6", [0x2f1d] = "\xd3\x51",
  [0x2f1e] = "\xd3\x50", [0x2f1f] = "\xd3\x57", [0x2f20] = "\xaf\xc0",
  [0x2f21] = "\xaf\xbc", [0x2f22] = "\xaf\xc1", [0x2f25] = "\x9e\xd7",
  [0x2f28] = "\xd6\xf0", [0x2f29] = "\xd6\xe9", [0x2f2b] = "\xb5\xb5",
  [0x2f2c] = "\xd6\xe8", [0x2f2e] = "\xb2\xcf", [0x2f2f] = "\xb2\xd6",
  [0x2f30] = "\xb2\xd3", [0x2f31] = "\xb2\xd9", [0x2f32] = "\xb2\xd8",
  [0x2f33] = "\xb2\xd4", [0x2f35] = "\xd6\xe2", [0x2f36] = "\xd6\xe5",
  [0x2f38] = "\xd6\xe4", [0x2f39] = "\xb2\xd0", [0x2f3a] = "\xd6\xe6",
  [0x2f3b] = "\xd6\xef", [0x2f3c] = "\xb2\xd1", [0x2f3d] = "\xd6\xe3",
  [0x2f3e] = "\xd6\xec", [0x2f3f] = "\xd6\xed", [0x2f40] = "\xb2\xd2",
  [0x2f41] = "\xd6\xea", [0x2f42] = "\xb2\xd7", [0x2f43] = "\xb2\xcd",
  [0x2f44] = "\xb2\xd5", [0x2f45] = "\xd6\xe7", [0x2f46] = "\xb2\xcc",
  [0x2f47] = "\xd6\xeb", [0x2f4a] = "\xd6\xee", [0x2f4d] = "\xa0\xb6",
  [0x2f4e] = "\xda\xfb", [0x2f4f] = "\xda\xf2", [0x2f50] = "\xb5\xb2",
  [0x2f51] = "\xda\xf9", [0x2f52] = "\xda\xf6", [0x2f53] = "\xda\xee",
  [0x2f54] = "\xda\xf7", [0x2f55] = "\xb5\xb4", [0x2f56] = "\xda\xef",
  [0x2f58] = "\xda\xeb", [0x2f5a] = "\x9e\x42", [0x2f5b] = "\xb8\x6c",
  [0x2f5c] = "\xda\xf4", [0x2f5d] = "\x8e\xa4", [0x2f5e] = "\xb5\xb1",
  [0x2f5f] = "\xda\xfa", [0x2f61] = "\xb5\xb8", [0x2f62] = "\xb5\xba",
  [0x2f63] = "\xda\xed", [0x2f66] = "\xb5\xb9", [0x2f67] = "\xda\xf0",
  [0x2f68] = "\xb5\xb3", [0x2f69] = "\xda\xf8", [0x2f6a] = "\xda\xf1",
  [0x2f6b] = "\xda\xf5", [0x2f6d] = "\xda\xf3", [0x2f6e] = "\xb5\xb6",
  [0x2f6f] = "\xda\xec", [0x2f70] = "\xb5\xbb", [0x2f71] = "\xb2\xce",
  [0x2f72] = "\xb5\xb7", [0x2f73] = "\xb5\xbc", [0x2f79] = "\xb8\x68",
  [0x2f7a] = "\xdf\x5d", [0x2f7b] = "\xdf\x5f", [0x2f7c] = "\xdf\x61",
  [0x2f7d] = "\xdf\x65", [0x2f7f] = "\xdf\x5b", [0x2f80] = "\xdf\x59",
  [0x2f81] = "\xb8\x6a", [0x2f83] = "\xdf\x60", [0x2f84] = "\xdf\x64",
  [0x2f85] = "\xdf\x5c", [0x2f86] = "\xdf\x58", [0x2f88] = "\xdf\x57",
  [0x2f89] = "\x8e\xa7", [0x2f8c] = "\xdf\x62", [0x2f8d] = "\xdf\x5a",
  [0x2f8e] = "\xdf\x5e", [0x2f8f] = "\xb8\x6b", [0x2f91] = "\xb8\x69",
  [0x2f92] = "\xdf\x66", [0x2f93] = "\xb8\x67", [0x2f94] = "\xdf\x63",
  [0x2f96] = "\xe3\x72", [0x2f97] = "\x95\x42", [0x2f9c] = "\xba\xee",
  [0x2f9d] = "\xe3\x6a", [0x2f9e] = "\xbd\x78", [0x2f9f] = "\xe3\x74",
  [0x2fa0] = "\xba\xf1", [0x2fa1] = "\xe3\x78", [0x2fa2] = "\xba\xf7",
  [0x2fa3] = "\xe3\x65", [0x2fa4] = "\x98\x7d", [0x2fa6] = "\xe3\x75",
  [0x2fa7] = "\xe3\x62", [0x2fa8] = "\x97\x55", [0x2fa9] = "\xe3\x77",
  [0x2faa] = "\xe3\x66", [0x2fab] = "\x8e\xa8", [0x2fac] = "\xba\xfe",
  [0x2fad] = "\xba\xfb", [0x2fae] = "\xe3\x76", [0x2faf] = "\xe3\x70",
  [0x2fb0] = "\xba\xed", [0x2fb1] = "\xba\xf5", [0x2fb2] = "\xba\xf4",
  [0x2fb3] = "\x8e\xaa", [0x2fb4] = "\xba\xf3", [0x2fb5] = "\xba\xf9",
  [0x2fb7] = "\xe3\x63", [0x2fb8] = "\xba\xfa", [0x2fb9] = "\xe3\x71",
  [0x2fba] = "\xba\xf6", [0x2fbb] = "\xba\xec", [0x2fbc] = "\xe3\x73",
  [0x2fbd] = "\xba\xef", [0x2fbe] = "\xba\xf0", [0x2fbf] = "\xba\xf8",
  [0x2fc0] = "\xe3\x68", [0x2fc1] = "\xe3\x67", [0x2fc2] = "\xe3\x64",
  [0x2fc4] = "\xe3\x6c", [0x2fc5] = "\xe3\x69", [0x2fc6] = "\xe3\x6d",
  [0x2fc7] = "\xba\xfd", [0x2fc9] = "\xe3\x79", [0x2fca] = "\xba\xf2",
  [0x2fcb] = "\xe3\x6e", [0x2fcc] = "\xe3\x6f", [0x2fcd] = "\x89\xa3",
  [0x2fce] = "\xe3\x6b", [0x2fcf] = "\x99\x60", [0x2fd0] = "\x99\x62",
  [0x2fd2] = "\xba\xfc", [0x2fd3] = "\x94\xfc", [0x2fd4] = "\x99\x61",
  [0x2fd7] = "\xe6\xe7", [0x2fd8] = "\xbd\x70", [0x2fd9] = "\xbd\x79",
  [0x2fda] = "\xbd\x75", [0x2fdb] = "\xe6\xe4", [0x2fdc] = "\x94\xfa",
  [0x2fdd] = "\xbd\x72", [0x2fde] = "\xbd\x76", [0x2fdf] = "\xe6\xf0",
  [0x2fe0] = "\xbd\x6c", [0x2fe1] = "\xe6\xe8", [0x2fe3] = "\xbd\x74",
  [0x2fe4] = "\x8e\xae", [0x2fe5] = "\x8e\xb2", [0x2fe6] = "\xe6\xeb",
  [0x2fe7] = "\xe6\xe6", [0x2fe8] = "\xbd\x73", [0x2fe9] = "\xbd\x77",
  [0x2fea] = "\xe6\xe5", [0x2fec] = "\xbd\x71", [0x2fee] = "\xe6\xef",
  [0x2fef] = "\xbd\x6e", [0x2ff0] = "\xe6\xee", [0x2ff1] = "\xe6\xed",
  [0x2ff2] = "\xbd\x7a", [0x2ff3] = "\xe5\x72", [0x2ff4] = "\xbd\x6d",
  [0x2ff5] = "\x8e\xb0", [0x2ff6] = "\xe6\xec", [0x2ff7] = "\xe6\xe3",
  [0x2ff9] = "\xbd\x7b", [0x2ffa] = "\xe6\xea", [0x2ffb] = "\xbd\x6f",
  [0x2ffd] = "\x99\x63", [0x2ffe] = "\x97\xaa", [0x3003] = "\xe6\xe9",
  [0x3007] = "\x94\xfb", [0x3008] = "\xbf\xa2", [0x3009] = "\xbf\xa7",
  [0x300a] = "\xbf\x7e", [0x300b] = "\xea\xd8", [0x300c] = "\xea\xcf",
  [0x300d] = "\xea\xdb", [0x300e] = "\xea\xd3", [0x300f] = "\xea\xd9",
  [0x3010] = "\xbf\xa8", [0x3011] = "\xbf\xa1", [0x3012] = "\xea\xcc",
  [0x3013] = "\xea\xd2", [0x3014] = "\xea\xdc", [0x3015] = "\xea\xd5",
  [0x3016] = "\xea\xda", [0x3017] = "\xea\xce", [0x301a] = "\xea\xd6",
  [0x301b] = "\xbf\xa3", [0x301c] = "\xea\xd4", [0x301d] = "\xbf\xa6",
  [0x301e] = "\xbf\xa5", [0x301f] = "\xea\xd0", [0x3020] = "\xea\xd1",
  [0x3021] = "\xea\xcd", [0x3022] = "\xea\xd7", [0x3023] = "\xbf\xa4",
  [0x3024] = "\xea\xde", [0x3025] = "\xea\xdd", [0x3027] = "\x8e\xbb",
  [0x3029] = "\xed\xda", [0x302a] = "\xed\xd6", [0x302b] = "\xc1\x5f",
  [0x302d] = "\xed\xd0", [0x302e] = "\xc1\x59", [0x302f] = "\xc1\x69",
  [0x3030] = "\xed\xdc", [0x3031] = "\xc1\x61", [0x3032] = "\xc1\x5d",
  [0x3033] = "\xed\xd3", [0x3034] = "\xc1\x64", [0x3035] = "\xc1\x67",
  [0x3036] = "\xed\xde", [0x3037] = "\xc1\x5c", [0x3038] = "\xed\xd5",
  [0x3039] = "\xc1\x65", [0x303a] = "\xed\xe0", [0x303b] = "\xed\xdd",
  [0x303c] = "\xed\xd1", [0x303d] = "\xc1\x60", [0x303e] = "\xc1\x5a",
  [0x303f] = "\xc1\x68", [0x3040] = "\xed\xd8", [0x3041] = "\xc1\x63",
  [0x3042] = "\xed\xd2", [0x3043] = "\xc1\x5e", [0x3044] = "\xed\xdf",
  [0x3045] = "\xc1\x62", [0x3046] = "\xc1\x5b", [0x3047] = "\xed\xd9",
  [0x3048] = "\xc1\x66", [0x3049] = "\xed\xd7", [0x304c] = "\xed\xdb",
  [0x3050] = "\xf0\x6e", [0x3051] = "\xf0\x74", [0x3052] = "\xc2\xb9",
  [0x3053] = "\xf0\x77", [0x3054] = "\xc2\xb4", [0x3055] = "\xc2\xb5",
  [0x3056] = "\xf0\x6f", [0x3057] = "\xf0\x76", [0x3058] = "\xf0\x71",
  [0x3059] = "\xc2\xba", [0x305a] = "\xc2\xb7", [0x305c] = "\xf0\x6d",
  [0x305e] = "\xc2\xb6", [0x305f] = "\xf0\x73", [0x3060] = "\xf0\x75",
  [0x3061] = "\xc2\xb8", [0x3062] = "\xf0\x72", [0x3063] = "\xf0\x70",
  [0x3065] = "\x98\x76", [0x3067] = "\x8e\xa1", [0x3068] = "\xf2\xb8",
  [0x3069] = "\xc3\xb7", [0x306a] = "\xc3\xb8", [0x306b] = "\xc3\xb4",
  [0x306d] = "\xc3\xb5", [0x306e] = "\x8e\xb7", [0x306f] = "\xf2\xb4",
  [0x3070] = "\xf2\xb2", [0x3072] = "\xf2\xb6", [0x3073] = "\xc3\xba",
  [0x3074] = "\xf2\xb7", [0x3075] = "\xf2\xb0", [0x3076] = "\xf2\xaf",
  [0x3077] = "\xf2\xb3", [0x3078] = "\xf2\xb1", [0x3079] = "\xc3\xb6",
  [0x307a] = "\xf2\xb5", [0x307b] = "\xf4\xac", [0x307c] = "\xc4\x7e",
  [0x307d] = "\xc4\x7d", [0x307e] = "\xf4\xad", [0x307f] = "\x9d\xa6",
  [0x3080] = "\xf4\xaf", [0x3081] = "\xf4\xae", [0x3082] = "\xc4\xa1",
  [0x3086] = "\xf5\xeb", [0x3087] = "\xf5\xe8", [0x3088] = "\xf5\xe9",
  [0x308a] = "\xf5\xe7", [0x308b] = "\xf5\xea", [0x308c] = "\xc4\xf2",
  [0x308d] = "\xf5\xec", [0x308e] = "\x9e\xb0", [0x308f] = "\xc4\xf1",
  [0x3091] = "\xf7\x42", [0x3092] = "\x8e\xb8", [0x3093] = "\xc5\xd5",
  [0x3094] = "\xc5\xd7", [0x3095] = "\xf7\xee", [0x3096] = "\xc5\xd6",
  [0x3097] = "\xf8\xb9", [0x3098] = "\xf9\x40", [0x3099] = "\xf9\x42",
  [0x309a] = "\xf8\xfe", [0x309b] = "\xf9\x41", [0x309c] = "\xc6\x6c",
  [0x309f] = "\x9d\x70", [0x30a4] = "\x89\x6e", [0x30ac] = "\x89\x6f",
  [0x30ba] = "\x89\x70", [0x30c7] = "\x89\x71", [0x30cf] = "\x89\x72",
  [0x30df] = "\x89\x73", [0x3106] = "\x89\x74", [0x3136] = "\xa6\xce",
  [0x3137] = "\x89\x75", [0x3138] = "\xac\xfb", [0x3139] = "\xd2\x6f",
  [0x313a] = "\xaf\xca", [0x313d] = "\xb2\xda", [0x313e] = "\xda\xfc",
  [0x313f] = "\xda\xfd", [0x3140] = "\x8e\xbc", [0x3141] = "\x8e\xbd",
  [0x3143] = "\xea\xdf", [0x3144] = "\xc1\x6a", [0x3145] = "\xed\xe1",
  [0x3147] = "\x8e\xbe", [0x3148] = "\xc2\xbb", [0x3149] = "\x9d\xd1",
  [0x314a] = "\xf2\xba", [0x314b] = "\xf2\xb9", [0x314c] = "\xc4\xa2",
  [0x314d] = "\xf5\xed", [0x314e] = "\x94\xfd", [0x314f] = "\xf7\x43",
  [0x3150] = "\xc5\xf8", [0x3151] = "\xca\x49", [0x3152] = "\x8b\xd7",
  [0x3153] = "\x8b\xda", [0x3154] = "\xaa\xc9", [0x3155] = "\xa8\x75",
  [0x3158] = "\xd0\x4d", [0x315b] = "\xd3\x60", [0x315c] = "\xd3\x5b",
  [0x315d] = "\xd3\x5f", [0x315e] = "\xd3\x5d", [0x315f] = "\xaf\xcb",
  [0x3160] = "\xd3\x5e", [0x3161] = "\xd3\x5c", [0x3163] = "\xd6\xf1",
  [0x3165] = "\xda\xfe", [0x3166] = "\xdb\x40", [0x3167] = "\xdf\x69",
  [0x3168] = "\xdf\x6a", [0x3169] = "\xb8\x6e", [0x316a] = "\xb8\x6f",
  [0x316b] = "\xdf\x68", [0x316c] = "\xdf\x6b", [0x316d] = "\xdf\x67",
  [0x316e] = "\xb8\x6d", [0x3170] = "\xbb\x40", [0x3171] = "\xa0\xe2",
  [0x3172] = "\xb8\x70", [0x3173] = "\xe3\x7a", [0x3175] = "\xbd\x7c",
  [0x3176] = "\xe6\xf1", [0x3177] = "\xbd\x7d", [0x3178] = "\x9f\xe9",
  [0x3179] = "\xbf\xa9", [0x317a] = "\xea\xe2", [0x317b] = "\xea\xe0",
  [0x317c] = "\xea\xe1", [0x317d] = "\xed\xe4", [0x317e] = "\xed\xe3",
  [0x317f] = "\xed\xe2", [0x3183] = "\xf2\xbb", [0x3185] = "\xc3\xb9",
  [0x3186] = "\xf2\xbc", [0x3187] = "\xf7\x44", [0x3188] = "\xc5\xf9",
  [0x3189] = "\xf8\xba", [0x318a] = "\xa6\xcf", [0x318b] = "\xaa\xcb",
  [0x318c] = "\xaa\xca", [0x318d] = "\xd0\x4f", [0x318e] = "\xac\xfc",
  [0x318f] = "\xfd\xa8", [0x3191] = "\xd0\x4e", [0x3192] = "\xd3\x62",
  [0x3193] = "\x8a\xe7", [0x3194] = "\xaf\xcc", [0x3195] = "\xd6\xf2",
  [0x3196] = "\xd3\x61", [0x3197] = "\x8e\xc2", [0x319a] = "\xb2\xdc",
  [0x319b] = "\xd6\xf5", [0x319c] = "\xd6\xf3", [0x319d] = "\xd6\xf4",
  [0x319e] = "\xb2\xdb", [0x31a0] = "\xdb\x42", [0x31a1] = "\xdb\x43",
  [0x31a2] = "\xdb\x41", [0x31a3] = "\x8e\xc4", [0x31a4] = "\xb8\x73",
  [0x31a5] = "\xdf\x6d", [0x31a6] = "\xdf\x6c", [0x31a7] = "\xdf\x6e",
  [0x31a8] = "\xb8\x72", [0x31a9] = "\xb8\x71", [0x31ac] = "\xe6\xf2",
  [0x31ad] = "\xe6\xf4", [0x31ae] = "\x99\x64", [0x31af] = "\xbd\x7e",
  [0x31b0] = "\xe6\xf3", [0x31b1] = "\xea\xe3", [0x31b2] = "\xbf\xaa",
  [0x31b3] = "\xf0\x79", [0x31b4] = "\x99\x65", [0x31b5] = "\xf0\x78",
  [0x31b6] = "\xc3\xbb", [0x31b7] = "\xf2\xbd", [0x31b8] = "\xc3\xbd",
  [0x31b9] = "\xc3\xbc", [0x31ba] = "\xf4\xb0", [0x31bb] = "\xf5\xee",
  [0x31bc] = "\xc4\xf3", [0x31bd] = "\xa6\xd0", [0x31be] = "\xd0\x50",
  [0x31bf] = "\xac\xfd", [0x31c0] = "\xd3\x65", [0x31c1] = "\xaf\xce",
  [0x31c2] = "\xd3\x64", [0x31c3] = "\xd3\x63", [0x31c5] = "\xaf\xcd",
  [0x31c7] = "\xd6\xfb", [0x31c9] = "\xd6\xfd", [0x31ca] = "\xd6\xf6",
  [0x31cb] = "\xd6\xf7", [0x31cc] = "\xb2\xdd", [0x31cd] = "\xd6\xf8",
  [0x31ce] = "\xb2\xde", [0x31cf] = "\xd6\xfc", [0x31d0] = "\xd6\xf9",
  [0x31d1] = "\xd6\xfa", [0x31d2] = "\xb2\xdf", [0x31d4] = "\xb5\xbe",
  [0x31d5] = "\xb5\xbf", [0x31d7] = "\xdb\x44", [0x31db] = "\xdf\x6f",
  [0x31dc] = "\xdf\x70", [0x31dd] = "\x95\x4e", [0x31de] = "\xe3\x7e",
  [0x31df] = "\xbb\x43", [0x31e0] = "\xbb\x41", [0x31e1] = "\xbb\x42",
  [0x31e2] = "\xe3\x7b", [0x31e3] = "\xe3\x7c", [0x31e5] = "\xe3\x7d",
  [0x31e6] = "\xe6\xf9", [0x31e7] = "\x98\xb3", [0x31e8] = "\xe6\xfa",
  [0x31e9] = "\xbd\xa1", [0x31ea] = "\xe6\xf7", [0x31eb] = "\xe6\xf6",
  [0x31ec] = "\xe6\xf8", [0x31ed] = "\xe6\xf5", [0x31ee] = "\xbf\xad",
  [0x31ef] = "\xea\xe4", [0x31f0] = "\xbf\xab", [0x31f1] = "\xbf\xac",
  [0x31f2] = "\xed\xe6", [0x31f3] = "\xc1\x6b", [0x31f4] = "\xed\xe5",
  [0x31f5] = "\xef\xa8", [0x31f7] = "\xf0\x7a", [0x31f8] = "\xf0\x7b",
  [0x31f9] = "\xc2\xbc", [0x31fa] = "\x8e\xcb", [0x31fb] = "\xc2\xbd",
  [0x31fc] = "\xc1\x6c", [0x31fd] = "\xf2\xbe", [0x31fe] = "\xf2\xbf",
  [0x31ff] = "\xf4\xb1", [0x3200] = "\xc4\xa3", [0x3201] = "\xa6\xd1",
  [0x3202] = "\x8b\xdf", [0x3203] = "\xa6\xd2", [0x3204] = "\xac\xfe",
  [0x3205] = "\xaa\xcc", [0x3206] = "\xaf\xcf", [0x3207] = "\xd0\x51",
  [0x3208] = "\x8e\xce", [0x320b] = "\xb5\xc0", [0x320c] = "\xa6\xd3",
  [0x320d] = "\xad\x41", [0x320e] = "\xd0\x52", [0x320f] = "\xd0\x53",
  [0x3210] = "\xad\x40", [0x3211] = "\xad\x42", [0x3212] = "\xa6\xd4",
  [0x3214] = "\xd0\x54", [0x3215] = "\xaf\xd1", [0x3216] = "\xd3\x66",
  [0x3217] = "\xaf\xd3", [0x3218] = "\xaf\xd0", [0x3219] = "\xaf\xd2",
  [0x321b] = "\xd7\x41", [0x321c] = "\xb2\xe0", [0x321d] = "\x8e\xcf",
  [0x321e] = "\xd7\x40", [0x321f] = "\xd6\xfe", [0x3220] = "\x99\x68",
  [0x3221] = "\xdf\x71", [0x3224] = "\xe3\xa1", [0x3225] = "\x99\x69",
  [0x3226] = "\xbd\xa2", [0x3228] = "\xbf\xae", [0x3229] = "\xea\xe6",
  [0x322a] = "\xea\xe5", [0x322c] = "\xed\xe7", [0x322e] = "\x99\x6b",
  [0x322f] = "\x8e\xd1", [0x3230] = "\xf5\xef", [0x3231] = "\x99\x6c",
  [0x3233] = "\xa6\xd5", [0x3234] = "\xcb\x73", [0x3235] = "\xcd\xaa",
  [0x3236] = "\xad\x43", [0x3237] = "\xd0\x55", [0x3239] = "\xd3\x68",
  [0x323b] = "\x8e\xd4", [0x323c] = "\x8e\xd5", [0x323d] = "\xaf\xd4",
  [0x323e] = "\xd3\x67", [0x323f] = "\xaf\xd5", [0x3243] = "\xd7\x43",
  [0x3246] = "\xb2\xe2", [0x3247] = "\xd7\x42", [0x3248] = "\xd7\x44",
  [0x324a] = "\xb2\xe1", [0x324f] = "\xdb\x46", [0x3250] = "\xdb\x47",
  [0x3251] = "\xdb\x45", [0x3252] = "\xb5\xc1", [0x3254] = "\x99\x6d",
  [0x3256] = "\xb8\x74", [0x3258] = "\xb8\x75", [0x325a] = "\xbb\x45",
  [0x325b] = "\xa0\xbe", [0x325c] = "\xe3\xa3", [0x325d] = "\xe3\xa2",
  [0x325e] = "\xbb\x44", [0x3261] = "\x8e\xd6", [0x3262] = "\xa0\xbc",
  [0x3263] = "\xa0\xb5", [0x3264] = "\xe6\xfb", [0x3266] = "\xa0\xb4",
  [0x3267] = "\xe6\xfc", [0x326c] = "\xea\xe7", [0x326f] = "\xc1\x70",
  [0x3270] = "\xc1\x6f", [0x3271] = "\xc1\x6d", [0x3272] = "\xc1\x6e",
  [0x3273] = "\xc1\x71", [0x3275] = "\xf0\x7c", [0x3276] = "\xc2\xbf",
  [0x3277] = "\xc2\xbe", [0x3278] = "\xf2\xc0", [0x3279] = "\xf4\xb2",
  [0x327d] = "\xc5\xa5", [0x327e] = "\xc5\xa4", [0x327f] = "\xa6\xd6",
  [0x3280] = "\x8b\xe0", [0x3282] = "\xd1\xfb", [0x3284] = "\xb8\x77",
  [0x3285] = "\xb5\xc2", [0x3286] = "\xb8\x76", [0x3287] = "\xbb\x46",
  [0x3289] = "\xa6\xd7", [0x328a] = "\xc9\xa9", [0x328b] = "\xa6\xd8",
  [0x328c] = "\xa6\xd9", [0x328f] = "\xcd\xab", [0x3290] = "\xcb\x76",
  [0x3292] = "\xcb\x77", [0x3293] = "\xa8\x77", [0x3295] = "\xcb\x74",
  [0x3296] = "\xa8\x76", [0x3298] = "\xa8\x79", [0x3299] = "\xcb\x75",
  [0x329a] = "\xa8\x7b", [0x329b] = "\xa8\x7a", [0x329c] = "\xcb\x78",
  [0x329d] = "\xa8\x78", [0x329f] = "\x89\xb5", [0x32a1] = "\xaa\xd1",
  [0x32a2] = "\xaa\xcf", [0x32a3] = "\xcd\xad", [0x32a5] = "\xaa\xce",
  [0x32a7] = "\x8e\xdd", [0x32a9] = "\xaa\xd3", [0x32aa] = "\xaa\xd5",
  [0x32ab] = "\xaa\xd2", [0x32ad] = "\xcd\xb0", [0x32ae] = "\xcd\xac",
  [0x32af] = "\xaa\xd6", [0x32b1] = "\xaa\xd0", [0x32b2] = "\xa8\x7c",
  [0x32b4] = "\xaa\xd4", [0x32b5] = "\xcd\xaf", [0x32b6] = "\x9e\x5d",
  [0x32b7] = "\x99\x71", [0x32b8] = "\xcd\xae", [0x32ba] = "\xaa\xcd",
  [0x32bc] = "\x89\xae", [0x32bd] = "\x9d\xe8", [0x32c2] = "\xd0\x5b",
  [0x32c3] = "\xad\x47", [0x32c4] = "\xad\x48", [0x32c5] = "\xd0\x5d",
  [0x32c6] = "\x95\x65", [0x32c7] = "\xd0\x57", [0x32c8] = "\xd0\x5a",
  [0x32c9] = "\xd0\x63", [0x32ca] = "\xd0\x61", [0x32cc] = "\xad\x49",
  [0x32cd] = "\xd0\x67", [0x32ce] = "\xad\x4c", [0x32cf] = "\xd0\x64",
  [0x32d0] = "\xd0\x5c", [0x32d1] = "\xd0\x59", [0x32d4] = "\xdb\x49",
  [0x32d5] = "\xd0\x62", [0x32d6] = "\xad\x44", [0x32d7] = "\xd0\x65",
  [0x32d8] = "\xd0\x56", [0x32d9] = "\xd0\x5f", [0x32da] = "\xad\x46",
  [0x32db] = "\xad\x4b", [0x32dc] = "\xd0\x60", [0x32dd] = "\xad\x4f",
  [0x32de] = "\xad\x4d", [0x32e0] = "\xd0\x58", [0x32e1] = "\xad\x4a",
  [0x32e3] = "\xd0\x5e", [0x32e4] = "\xad\x4e", [0x32e5] = "\xad\x45",
  [0x32e6] = "\xd0\x66", [0x32e9] = "\x99\x72", [0x32ec] = "\x8b\x5c",
  [0x32ed] = "\xaf\xda", [0x32ef] = "\xaf\xe3", [0x32f0] = "\xaf\xd8",
  [0x32f1] = "\xaf\xd6", [0x32f2] = "\xd3\x6a", [0x32f3] = "\xaf\xde",
  [0x32f4] = "\xaf\xdb", [0x32f5] = "\xd3\x6c", [0x32f6] = "\x89\xb1",
  [0x32f8] = "\xaf\xdd", [0x32f9] = "\xd3\x6b", [0x32fa] = "\xd3\x69",
  [0x32fb] = "\xd3\x6e", [0x32fc] = "\xaf\xe2", [0x32fd] = "\xaf\xe0",
  [0x32fe] = "\xdb\x48", [0x3300] = "\xd3\x6f", [0x3301] = "\xd3\x6d",
  [0x3302] = "\xaf\xd7", [0x3303] = "\xa0\xc0", [0x3305] = "\xaf\xd9",
  [0x3306] = "\xaf\xdc", [0x3307] = "\x8e\xdf", [0x3308] = "\xaf\xdf",
  [0x3309] = "\x95\x66", [0x330a] = "\xaf\xe1", [0x330c] = "\x99\x74",
  [0x330e] = "\x99\x76", [0x3312] = "\x99\x77", [0x3314] = "\x99\x79",
  [0x3315] = "\xd7\x4e", [0x3316] = "\xb2\xe4", [0x3317] = "\x9d\xda",
  [0x3318] = "\xd7\x45", [0x3319] = "\xd7\x47", [0x331a] = "\x8e\xe0",
  [0x331b] = "\xd7\x48", [0x331d] = "\xd7\x50", [0x331e] = "\xd7\x4c",
  [0x331f] = "\xd7\x4a", [0x3321] = "\xd7\x4d", [0x3322] = "\xd7\x51",
  [0x3323] = "\xb2\xe5", [0x3324] = "\xb2\xe9", [0x3325] = "\xd7\x46",
  [0x3327] = "\xd7\x4f", [0x3329] = "\xb2\xe7", [0x332a] = "\x93\x5c",
  [0x332b] = "\xb2\xe6", [0x332c] = "\xd7\x4b", [0x332d] = "\xd7\x49",
  [0x332f] = "\xb2\xe3", [0x3330] = "\xb2\xe8", [0x3332] = "\x9d\xe6",
  [0x3334] = "\x8b\x5f", [0x3337] = "\x95\x63", [0x3339] = "\xb5\xc8",
  [0x333a] = "\xdb\x51", [0x333d] = "\xdb\x4f", [0x333e] = "\xb5\xca",
  [0x3342] = "\x95\x67", [0x3343] = "\xdb\x4a", [0x3344] = "\xdf\xa1",
  [0x3346] = "\xb5\xc9", [0x3347] = "\xdb\x4e", [0x3348] = "\x9d\xe3",
  [0x334a] = "\xdb\x4b", [0x334b] = "\xb5\xc5", [0x334c] = "\xb5\xcb",
  [0x334d] = "\xdb\x50", [0x334e] = "\xb5\xc7", [0x334f] = "\xdb\x4d",
  [0x3350] = "\xbb\x47", [0x3351] = "\xb5\xc6", [0x3352] = "\xdb\x4c",
  [0x3353] = "\xb5\xcc", [0x3354] = "\xb5\xc4", [0x3355] = "\xb5\xc3",
  [0x3356] = "\x99\x7c", [0x3359] = "\x99\x7d", [0x335a] = "\x99\x7e",
  [0x335b] = "\xdf\x77", [0x335c] = "\xdf\x75", [0x335e] = "\xdf\x7b",
  [0x3360] = "\xdf\x73", [0x3361] = "\xdf\xa2", [0x3362] = "\xdf\x78",
  [0x3364] = "\xdf\x72", [0x3365] = "\xb8\x7b", [0x3366] = "\xb8\xa3",
  [0x3367] = "\xdf\x7d", [0x3369] = "\xdf\x76", [0x336b] = "\xb8\x7e",
  [0x336d] = "\x8b\x5b", [0x336e] = "\xb8\x7c", [0x336f] = "\xdf\x7e",
  [0x3370] = "\xb8\x79", [0x3371] = "\xb8\x78", [0x3372] = "\xdf\x79",
  [0x3373] = "\xb8\x7d", [0x3374] = "\xb5\xcd", [0x3376] = "\xdf\x7c",
  [0x3377] = "\xdf\x74", [0x3378] = "\xb8\x7a", [0x3379] = "\xb8\xa1",
  [0x337a] = "\xb8\xa2", [0x337c] = "\x99\xa3", [0x337f] = "\xbb\x4c",
  [0x3380] = "\xbb\x48", [0x3382] = "\xbb\x4d", [0x3383] = "\xe3\xa6",
  [0x3384] = "\x99\xa4", [0x3386] = "\xe3\xa5", [0x3387] = "\xe3\xa7",
  [0x3388] = "\xbb\x4a", [0x3389] = "\xe3\xa4", [0x338a] = "\xbb\x4b",
  [0x338b] = "\xe3\xaa", [0x338c] = "\xe3\xa9", [0x338d] = "\xe3\xa8",
  [0x338f] = "\xbb\x49", [0x3393] = "\x99\xa6", [0x3395] = "\xe7\x41",
  [0x3397] = "\xe7\x44", [0x3398] = "\xbd\xa8", [0x3399] = "\xe7\x43",
  [0x339a] = "\xbd\xa7", [0x339b] = "\xbd\xa3", [0x339c] = "\xbd\xa4",
  [0x339d] = "\xbd\xa5", [0x339e] = "\xe7\x40", [0x339f] = "\xe6\xfe",
  [0x33a0] = "\xbd\xa6", [0x33a2] = "\xe7\x42", [0x33a3] = "\xe6\xfd",
  [0x33a5] = "\x99\xa8", [0x33a6] = "\xea\xe9", [0x33a7] = "\xea\xf3",
  [0x33a8] = "\xbf\xb1", [0x33a9] = "\xbf\xb0", [0x33aa] = "\x8a\xbe",
  [0x33ab] = "\xea\xed", [0x33ac] = "\xea\xef", [0x33ae] = "\xea\xea",
  [0x33b0] = "\xea\xee", [0x33b1] = "\xea\xe8", [0x33b2] = "\xea\xf1",
  [0x33b3] = "\xbf\xaf", [0x33b4] = "\xea\xf0", [0x33b5] = "\xea\xec",
  [0x33b6] = "\x9e\x61", [0x33b7] = "\xea\xf2", [0x33b9] = "\xea\xeb",
  [0x33ba] = "\xc1\x74", [0x33bb] = "\xed\xe8", [0x33bc] = "\xed\xee",
  [0x33bd] = "\xc1\x78", [0x33be] = "\xc1\x7a", [0x33bf] = "\xc1\x77",
  [0x33c0] = "\xc1\x76", [0x33c1] = "\x99\xaa", [0x33c2] = "\xc1\x75",
  [0x33c3] = "\xc1\x73", [0x33c4] = "\xed\xe9", [0x33c5] = "\xed\xec",
  [0x33c6] = "\xc1\x72", [0x33c7] = "\xed\xed", [0x33c8] = "\xa0\xc8",
  [0x33c9] = "\xc1\x79", [0x33ca] = "\xed\xeb", [0x33cc] = "\xed\xea",
  [0x33cd] = "\xc2\xc0", [0x33cf] = "\xc2\xc1", [0x33d0] = "\xf0\xa1",
  [0x33d1] = "\xf0\x7d", [0x33d2] = "\xf0\x7e", [0x33d5] = "\xf2\xc2",
  [0x33d7] = "\xf2\xc1", [0x33d8] = "\xc3\xbe", [0x33d9] = "\xf4\xb4",
  [0x33da] = "\xc4\xa4", [0x33db] = "\xf4\xb3", [0x33dd] = "\xf5\xf0",
  [0x33de] = "\xf7\x45", [0x33df] = "\xc5\xa6", [0x33e0] = "\xf9\x43",
  [0x33e1] = "\xf9\x44", [0x33e2] = "\xc5\xd8", [0x33e3] = "\xa6\xda",
  [0x33e4] = "\x99\xab", [0x33e5] = "\xaa\xd7", [0x33e6] = "\xdb\x52",
  [0x33e7] = "\xbb\x4e", [0x33e8] = "\xc1\x7b", [0x33e9] = "\xed\xef",
  [0x33ea] = "\xa6\xdb", [0x33ec] = "\xaf\xe5", [0x33ed] = "\xaf\xe4",
  [0x33ee] = "\xdb\x53", [0x33ef] = "\xfe\xc4", [0x33f2] = "\xea\xf4",
  [0x33f3] = "\xa6\xdc", [0x33f4] = "\xad\x50", [0x33f6] = "\x98\xc2",
  [0x33f7] = "\xdb\x54", [0x33f8] = "\xdb\x55", [0x33f9] = "\xdb\x56",
  [0x33fa] = "\xbb\x4f", [0x33fb] = "\xbf\xb2", [0x33fc] = "\xa6\xdd",
  [0x33fe] = "\xaa\xd8", [0x33ff] = "\xd0\x68", [0x3400] = "\xaf\xe6",
  [0x3401] = "\xd3\x70", [0x3402] = "\xb2\xea", [0x3404] = "\xdb\x57",
  [0x3405] = "\xb8\xa4", [0x3407] = "\xbb\x50", [0x3408] = "\xbf\xb3",
  [0x3409] = "\xc1\x7c", [0x340a] = "\xc2\xc2", [0x340b] = "\xf4\xb5",
  [0x340c] = "\xa6\xde", [0x340d] = "\xaa\xd9", [0x3410] = "\xaf\xe7",
  [0x3411] = "\xd7\x52", [0x3412] = "\xb5\xce", [0x3414] = "\xbb\x51",
  [0x3415] = "\xe3\xab", [0x3416] = "\xe7\x45", [0x3418] = "\x8e\xe8",
  [0x341a] = "\xa0\xba", [0x341b] = "\xa6\xdf", [0x341c] = "\xb5\xcf",
  [0x341d] = "\xdf\xa3", [0x341e] = "\xbb\x52", [0x341f] = "\xa6\xe0",
  [0x3420] = "\xcd\xb1", [0x3421] = "\xd0\x69", [0x3422] = "\xad\x51",
  [0x3425] = "\xd3\x72", [0x3426] = "\xfd\x77", [0x3428] = "\xaf\xea",
  [0x3429] = "\x8e\xee", [0x342a] = "\xaf\xe8", [0x342b] = "\xaf\xe9",
  [0x342c] = "\xaf\xeb", [0x342d] = "\x9e\xbf", [0x342f] = "\xd3\x71",
  [0x3432] = "\xd7\x57", [0x3433] = "\xd7\x54", [0x3434] = "\xd7\x56",
  [0x3435] = "\xb2\xeb", [0x3436] = "\xb2\xed", [0x3437] = "\xb2\xec",
  [0x3438] = "\xd7\x53", [0x3439] = "\xb2\xee", [0x343a] = "\xd7\x55",
  [0x343c] = "\xdb\x58", [0x343d] = "\xdb\x59", [0x343e] = "\x89\xc2",
  [0x343f] = "\xdb\x5a", [0x3440] = "\xdf\xa6", [0x3442] = "\xdf\xa7",
  [0x3444] = "\xdf\xa5", [0x3445] = "\xdf\xa8", [0x3447] = "\xb8\xa5",
  [0x3449] = "\xdf\xa4", [0x344b] = "\xbb\x53", [0x344e] = "\xe7\x4a",
  [0x344f] = "\xe7\x46", [0x3450] = "\xe7\x49", [0x3451] = "\xe7\x4b",
  [0x3452] = "\xe7\x48", [0x3453] = "\xe7\x47", [0x3454] = "\x99\xac",
  [0x3455] = "\xea\xf5", [0x3456] = "\xea\xf6", [0x3457] = "\xea\xf7",
  [0x3458] = "\xbf\xb4", [0x3459] = "\xbf\xb5", [0x345a] = "\xed\xf1",
  [0x345b] = "\xed\xf0", [0x345c] = "\xed\xf2", [0x345e] = "\xf0\xa3",
  [0x345f] = "\xf0\xa2", [0x3461] = "\xf2\xc4", [0x3462] = "\x95\x6b",
  [0x3463] = "\xf2\xc5", [0x3464] = "\xf2\xc3", [0x3465] = "\x95\x6c",
  [0x3466] = "\xc4\xa5", [0x3468] = "\xf4\xb6", [0x3469] = "\xf4\xb7",
  [0x346b] = "\xf7\x46", [0x346c] = "\xf7\xef", [0x346d] = "\xf8\xbb",
  [0x346e] = "\xa6\xe1", [0x346f] = "\xa8\x7d", [0x3471] = "\xc1\x7d",
  [0x3472] = "\xa6\xe2", [0x3474] = "\xd7\x58", [0x3475] = "\xdb\x5b",
  [0x3476] = "\x99\xaf", [0x3477] = "\xc6\x41", [0x3478] = "\xca\x4a",
  [0x3479] = "\x99\x4a", [0x347a] = "\x89\x76", [0x347b] = "\x8f\x48",
  [0x347c] = "\xca\x4b", [0x347d] = "\xca\x4d", [0x347e] = "\xa6\xe3",
  [0x347f] = "\xca\x4e", [0x3480] = "\xca\x4c", [0x3483] = "\xcb\xa2",
  [0x3484] = "\xcb\xa3", [0x3485] = "\xcb\x7b", [0x3487] = "\xfb\xee",
  [0x348a] = "\xcb\xa1", [0x348b] = "\xa8\xa1", [0x348d] = "\xa8\xa2",
  [0x348e] = "\xcb\x7c", [0x348f] = "\xcb\x7a", [0x3490] = "\xcb\x79",
  [0x3491] = "\xcb\x7d", [0x3492] = "\xa8\x7e", [0x3493] = "\xcb\x7e",
  [0x3494] = "\xd0\x6a", [0x3498] = "\xcd\xb6", [0x3499] = "\xaa\xdc",
  [0x349a] = "\xcd\xb5", [0x349b] = "\xcd\xb7", [0x349d] = "\xaa\xdb",
  [0x349e] = "\xcd\xbc", [0x349f] = "\xaa\xdf", [0x34a0] = "\xcd\xb2",
  [0x34a1] = "\xcd\xc0", [0x34a2] = "\xcd\xc6", [0x34a3] = "\xaa\xe6",
  [0x34a4] = "\xcd\xc3", [0x34a5] = "\xaa\xe3", [0x34a6] = "\x99\xae",
  [0x34a7] = "\xcd\xb9", [0x34a8] = "\xcd\xbf", [0x34a9] = "\xcd\xc1",
  [0x34aa] = "\x8e\xfb", [0x34ab] = "\xcd\xb4", [0x34ac] = "\xaa\xe2",
  [0x34ad] = "\xaa\xdd", [0x34ae] = "\xcd\xba", [0x34af] = "\xaa\xe4",
  [0x34b0] = "\xaa\xe7", [0x34b1] = "\xaa\xe1", [0x34b3] = "\xaa\xda",
  [0x34b4] = "\xcd\xbe", [0x34b5] = "\xcd\xb8", [0x34b6] = "\xcd\xc5",
  [0x34b7] = "\xaa\xe9", [0x34b8] = "\xaa\xe5", [0x34b9] = "\xaa\xe0",
  [0x34ba] = "\xcd\xbd", [0x34bb] = "\xaf\xec", [0x34bc] = "\xcd\xbb",
  [0x34bd] = "\xaa\xde", [0x34be] = "\xaa\xe8", [0x34c0] = "\xcd\xb3",
  [0x34c2] = "\xcd\xc2", [0x34c3] = "\xcd\xc4", [0x34c4] = "\x8b\x52",
  [0x34ca] = "\x99\xb0", [0x34cf] = "\x89\x77", [0x34d0] = "\x8f\x41",
  [0x34d1] = "\xad\x62", [0x34d2] = "\xad\x5c", [0x34d3] = "\xad\x64",
  [0x34d4] = "\xad\x61", [0x34d5] = "\xd0\x71", [0x34d6] = "\xd0\x74",
  [0x34d7] = "\xad\x5d", [0x34d8] = "\x99\xb1", [0x34d9] = "\xd0\x6b",
  [0x34db] = "\xad\x56", [0x34dc] = "\xad\x60", [0x34de] = "\xad\x63",
  [0x34df] = "\xad\x65", [0x34e0] = "\xd0\xa2", [0x34e1] = "\xd0\x77",
  [0x34e2] = "\x8f\x49", [0x34e3] = "\xad\x55", [0x34e4] = "\xd0\xa1",
  [0x34e5] = "\xad\x59", [0x34e6] = "\xad\x57", [0x34e7] = "\xad\x52",
  [0x34e8] = "\xd0\x6f", [0x34ea] = "\xd0\x7e", [0x34eb] = "\xd0\x73",
  [0x34ec] = "\xd0\x76", [0x34ed] = "\xd0\xa5", [0x34ee] = "\xfa\x4d",
  [0x34ef] = "\xad\x66", [0x34f0] = "\xd0\x7d", [0x34f1] = "\xad\x5e",
  [0x34f2] = "\xd0\x78", [0x34f3] = "\xd0\xa4", [0x34f4] = "\xd0\x75",
  [0x34f5] = "\xd0\x79", [0x34f6] = "\xd0\x7c", [0x34f7] = "\x9d\xe4",
  [0x34f9] = "\xd0\x6d", [0x34fa] = "\xd0\xa3", [0x34fb] = "\xd0\x7b",
  [0x34fc] = "\xfb\xe9", [0x34fd] = "\x9b\x54", [0x34fe] = "\xd0\x6c",
  [0x34ff] = "\x99\xb2", [0x3500] = "\xd0\x70", [0x3501] = "\xad\x5f",
  [0x3502] = "\xad\x5a", [0x3503] = "\xad\x53", [0x3504] = "\xad\x58",
  [0x3505] = "\xad\x54", [0x3506] = "\xad\x67", [0x3507] = "\xd0\x6e",
  [0x3508] = "\xd3\xa5", [0x3509] = "\xad\x5b", [0x350b] = "\x9e\x68",
  [0x350c] = "\xd0\x7a", [0x350d] = "\xce\x41", [0x3516] = "\xd3\xa8",
  [0x3517] = "\xaf\xfa", [0x3518] = "\x8f\x4a", [0x3519] = "\xd3\x76",
  [0x351a] = "\x8f\x42", [0x351b] = "\xd3\xa3", [0x351c] = "\xd3\x7d",
  [0x351d] = "\x8f\x51", [0x351e] = "\xd3\xb2", [0x3520] = "\xd3\xaa",
  [0x3522] = "\xd3\x7e", [0x3524] = "\xd3\xa9", [0x3525] = "\xd3\x78",
  [0x3526] = "\xd3\x7c", [0x3527] = "\xd3\xb5", [0x3528] = "\xaf\xfd",
  [0x3529] = "\xd3\xad", [0x352a] = "\xd3\xa4", [0x352b] = "\xaf\xed",
  [0x352c] = "\xd3\xb3", [0x352d] = "\xd3\x74", [0x352f] = "\xd3\xac",
  [0x3531] = "\xaf\xfc", [0x3532] = "\xaf\xf7", [0x3533] = "\xd3\x73",
  [0x3534] = "\xaf\xf5", [0x3535] = "\xaf\xf4", [0x3536] = "\xaf\xf9",
  [0x3537] = "\xd3\xab", [0x3538] = "\xaf\xf1", [0x3539] = "\xaf\xf8",
  [0x353a] = "\xd0\x72", [0x353b] = "\xdb\x5c", [0x353c] = "\xd3\xa6",
  [0x353d] = "\x98\x46", [0x353f] = "\xd3\x7a", [0x3540] = "\xaf\xfb",
  [0x3541] = "\xd3\x7b", [0x3542] = "\xd3\xa1", [0x3543] = "\xaf\xfe",
  [0x3544] = "\xd3\x75", [0x3545] = "\xd3\xaf", [0x3547] = "\xd3\xae",
  [0x3548] = "\xd3\xb6", [0x3549] = "\xaf\xf3", [0x354a] = "\xaf\xf0",
  [0x354b] = "\xd3\xb4", [0x354c] = "\xd3\xb0", [0x354d] = "\xd3\xa7",
  [0x354e] = "\xd3\xa2", [0x354f] = "\xaf\xf6", [0x3550] = "\xaf\xf2",
  [0x3551] = "\xd3\x77", [0x3552] = "\xaf\xee", [0x3553] = "\xd3\xb1",
  [0x3554] = "\xaf\xef", [0x3556] = "\xd3\x79", [0x3557] = "\x99\xb4",
  [0x3562] = "\x8e\xf5", [0x3563] = "\xfd\x55", [0x3566] = "\x9c\xcd",
  [0x356f] = "\x89\x78", [0x3573] = "\xd7\x5e", [0x3574] = "\xd7\x60",
  [0x3575] = "\xd7\x65", [0x3576] = "\xd7\x79", [0x3577] = "\xb2\xfc",
  [0x3578] = "\xb2\xf2", [0x357a] = "\xd7\x5d", [0x357b] = "\xb2\xfd",
  [0x357c] = "\xb2\xfe", [0x357d] = "\xd7\x68", [0x357e] = "\xd7\x6f",
  [0x357f] = "\xd7\x75", [0x3581] = "\xd7\x62", [0x3583] = "\xd7\x69",
  [0x3585] = "\x8f\x53", [0x3586] = "\xb3\x40", [0x3587] = "\xd7\x77",
  [0x3588] = "\xd7\x72", [0x3589] = "\xb2\xfa", [0x358a] = "\xb2\xf8",
  [0x358b] = "\xd7\x6e", [0x358c] = "\xd7\x6a", [0x358d] = "\xd7\x5c",
  [0x358e] = "\xb2\xef", [0x358f] = "\xd7\x61", [0x3590] = "\xd7\x59",
  [0x3591] = "\x8f\x6f", [0x3592] = "\xb2\xf7", [0x3593] = "\xb2\xf9",
  [0x3594] = "\xd7\x66", [0x3595] = "\xd7\x63", [0x3596] = "\xb2\xf4",
  [0x3597] = "\xd7\x73", [0x3598] = "\xb2\xf1", [0x3599] = "\xd7\x64",
  [0x359a] = "\xd7\x7a", [0x359b] = "\xd7\x6c", [0x359c] = "\x8e\x63",
  [0x359d] = "\xd7\x6b", [0x359e] = "\xb2\xf0", [0x35a0] = "\xb2\xfb",
  [0x35a2] = "\xb2\xf3", [0x35a3] = "\xd7\x5a", [0x35a4] = "\xd7\x5f",
  [0x35a5] = "\xd7\x70", [0x35a6] = "\xd7\x76", [0x35a7] = "\xb3\x41",
  [0x35a8] = "\xd7\x5b", [0x35a9] = "\xd7\x67", [0x35aa] = "\xd7\x6d",
  [0x35ab] = "\xb2\xf6", [0x35ac] = "\x8f\x56", [0x35ae] = "\xd7\x78",
  [0x35af] = "\xd7\x71", [0x35b0] = "\xd7\x74", [0x35b9] = "\xfe\x76",
  [0x35bd] = "\xb2\xf5", [0x35be] = "\x9f\xc6", [0x35bf] = "\xdb\x6c",
  [0x35c0] = "\xdb\x60", [0x35c1] = "\xb5\xd7", [0x35c2] = "\xdb\x7d",
  [0x35c3] = "\xdb\xa7", [0x35c4] = "\xdb\xaa", [0x35c5] = "\xb5\xd5",
  [0x35c6] = "\xdb\x68", [0x35c7] = "\xdb\xa3", [0x35c8] = "\xdb\x69",
  [0x35c9] = "\xdb\x77", [0x35ca] = "\xb5\xe2", [0x35cb] = "\xdb\x73",
  [0x35cc] = "\xb5\xdf", [0x35cd] = "\xfa\xac", [0x35ce] = "\xdb\x74",
  [0x35cf] = "\xdb\x5d", [0x35d1] = "\xdb\xa4", [0x35d3] = "\x8f\x58",
  [0x35d4] = "\xb5\xe8", [0x35d5] = "\xdb\xa1", [0x35d6] = "\xdb\x75",
  [0x35d7] = "\xdb\xac", [0x35d8] = "\xdb\x70", [0x35d9] = "\xdf\xc8",
  [0x35db] = "\xdb\xaf", [0x35dc] = "\xb5\xe6", [0x35dd] = "\xdb\x6e",
  [0x35de] = "\xdb\x7a", [0x35df] = "\xb5\xe9", [0x35e0] = "\xb5\xd4",
  [0x35e1] = "\xdb\x72", [0x35e2] = "\xdb\xad", [0x35e3] = "\xdb\x6b",
  [0x35e4] = "\xdb\x64", [0x35e5] = "\xdb\x6f", [0x35e7] = "\xdb\x63",
  [0x35e8] = "\xdb\x61", [0x35e9] = "\xb5\xd0", [0x35ea] = "\xdb\xa5",
  [0x35eb] = "\xdb\x6a", [0x35ec] = "\xdb\xa8", [0x35ed] = "\x98\x48",
  [0x35ee] = "\xdb\xa9", [0x35ef] = "\xb5\xd8", [0x35f0] = "\xb5\xdd",
  [0x35f1] = "\xb5\xd9", [0x35f2] = "\xb5\xe1", [0x35f3] = "\xdb\x7e",
  [0x35f4] = "\xb5\xda", [0x35f5] = "\xdb\x76", [0x35f6] = "\xdb\x66",
  [0x35f8] = "\xb5\xd2", [0x35f9] = "\xdb\x5e", [0x35fa] = "\xdb\xa2",
  [0x35fb] = "\xdb\xab", [0x35fc] = "\xdb\x65", [0x35fd] = "\xb5\xe0",
  [0x35fe] = "\xdb\xb0", [0x35ff] = "\xdb\x71", [0x3601] = "\xdb\x6d",
  [0x3603] = "\xb5\xd1", [0x3604] = "\xb5\xe5", [0x3605] = "\x99\xb7",
  [0x3606] = "\xdb\x7c", [0x3607] = "\xb5\xe7", [0x3609] = "\xdb\x78",
  [0x360a] = "\xb5\xdc", [0x360b] = "\xb5\xd6", [0x360c] = "\xb5\xde",
  [0x360d] = "\xb5\xd3", [0x360e] = "\xb5\xe4", [0x360f] = "\xdb\x79",
  [0x3610] = "\xdb\x67", [0x3611] = "\xdb\x7b", [0x3612] = "\xdb\x62",
  [0x3613] = "\xdb\xa6", [0x3614] = "\x96\x65", [0x3616] = "\xfa\x6c",
  [0x3618] = "\x9d\xe7", [0x361b] = "\xdb\xae", [0x361c] = "\x9e\x62",
  [0x3620] = "\x96\xcc", [0x3621] = "\x8e\x67", [0x3623] = "\xdb\x5f",
  [0x3624] = "\xfc\x75", [0x3626] = "\x98\x7e", [0x3629] = "\xdf\xc7",
  [0x362b] = "\xdf\xdd", [0x362c] = "\xb8\x55", [0x362d] = "\xdf\xcc",
  [0x362e] = "\xfd\xb9", [0x362f] = "\xdf\xca", [0x3630] = "\xdf\xb5",
  [0x3631] = "\xb8\xa9", [0x3632] = "\xdf\xc5", [0x3633] = "\xdf\xd9",
  [0x3634] = "\xdf\xc1", [0x3635] = "\xb8\xb1", [0x3636] = "\xdf\xd8",
  [0x3637] = "\xdf\xbf", [0x3638] = "\xb5\xe3", [0x3639] = "\xdf\xcf",
  [0x363a] = "\xdf\xc0", [0x363b] = "\xdf\xd6", [0x363c] = "\xb8\xb0",
  [0x363d] = "\xb8\xa8", [0x363e] = "\x97\xfc", [0x363f] = "\xdf\xaa",
  [0x3640] = "\xdf\xb2", [0x3642] = "\xdf\xcb", [0x3643] = "\xdf\xc3",
  [0x3644] = "\xdf\xdc", [0x3645] = "\xdf\xc6", [0x3646] = "\xb8\xb6",
  [0x3647] = "\xdf\xd7", [0x3648] = "\x98\xf9", [0x3649] = "\xb8\xad",
  [0x364a] = "\x8f\x66", [0x364b] = "\xdf\xc9", [0x364c] = "\xdf\xd1",
  [0x364d] = "\xdf\xb6", [0x364e] = "\xdf\xd0", [0x3650] = "\xdf\xe1",
  [0x3651] = "\xdf\xb1", [0x3652] = "\xdf\xd2", [0x3653] = "\x95\x6e",
  [0x3654] = "\xdf\xdf", [0x3655] = "\x92\x45", [0x3656] = "\xdf\xab",
  [0x3657] = "\xb5\xdb", [0x3658] = "\x8f\x60", [0x3659] = "\xdf\xb9",
  [0x365a] = "\xdf\xb8", [0x365b] = "\xb8\xaf", [0x365c] = "\x9e\xd1",
  [0x365d] = "\xdf\xbc", [0x365e] = "\xdf\xbe", [0x365f] = "\xdf\xcd",
  [0x3660] = "\xdf\xde", [0x3661] = "\xb8\xb2", [0x3662] = "\xfe\xcd",
  [0x3663] = "\xb8\xb3", [0x3664] = "\x99\xb9", [0x3665] = "\xdf\xb0",
  [0x3666] = "\xb8\xab", [0x3667] = "\xdf\xb4", [0x3668] = "\xdf\xda",
  [0x3669] = "\xb8\xb4", [0x366b] = "\xb8\xac", [0x366c] = "\xb8\xae",
  [0x366d] = "\xb8\xb5", [0x366e] = "\xdf\xe0", [0x366f] = "\xdf\xd3",
  [0x3670] = "\xdf\xce", [0x3671] = "\x8f\x62", [0x3672] = "\x97\x4c",
  [0x3673] = "\xdf\xbb", [0x3674] = "\xdf\xba", [0x3675] = "\xb8\xaa",
  [0x3676] = "\xdf\xac", [0x3677] = "\xb8\xa7", [0x3678] = "\xdf\xc4",
  [0x3679] = "\xdf\xad", [0x367a] = "\xdf\xc2", [0x367d] = "\xdf\xb7",
  [0x367e] = "\xdf\xdb", [0x367f] = "\x91\xc7", [0x3680] = "\x95\x5f",
  [0x3682] = "\xb8\xa6", [0x3686] = "\xdf\xb3", [0x3688] = "\x99\xbb",
  [0x368d] = "\xdf\xaf", [0x368e] = "\xdf\xd5", [0x368f] = "\xdf\xae",
  [0x3690] = "\xbb\x60", [0x3691] = "\xe3\xd3", [0x3692] = "\x8e\x6d",
  [0x3693] = "\x8f\x71", [0x3694] = "\xe3\xc2", [0x3696] = "\x94\xcb",
  [0x3697] = "\xe3\xac", [0x3698] = "\xe3\xca", [0x3699] = "\xbb\x58",
  [0x369a] = "\xe3\xbb", [0x369b] = "\xe3\xc5", [0x369c] = "\xbb\x5b",
  [0x369d] = "\xe3\xbe", [0x369e] = "\xbb\x59", [0x369f] = "\xe3\xaf",
  [0x36a0] = "\xe3\xcd", [0x36a1] = "\xe3\xae", [0x36a2] = "\xe3\xc1",
  [0x36a3] = "\x95\xb1", [0x36a4] = "\xe3\xad", [0x36a7] = "\xe3\xbf",
  [0x36a8] = "\xe3\xc8", [0x36a9] = "\xe3\xc6", [0x36aa] = "\xe3\xba",
  [0x36ab] = "\xe3\xb5", [0x36ac] = "\xe3\xb3", [0x36ad] = "\x9a\xf2",
  [0x36ae] = "\xe3\xb4", [0x36af] = "\xe3\xc7", [0x36b0] = "\xe3\xd2",
  [0x36b1] = "\xe3\xbc", [0x36b2] = "\xbb\x5a", [0x36b4] = "\xe3\xb7",
  [0x36b6] = "\xe3\xcb", [0x36b8] = "\xbb\x5d", [0x36b9] = "\xe3\xb6",
  [0x36ba] = "\xe3\xb0", [0x36bb] = "\xe3\xc0", [0x36bc] = "\xbb\x61",
  [0x36bd] = "\x96\xc3", [0x36be] = "\x99\xbd", [0x36bf] = "\xbb\x55",
  [0x36c0] = "\xbb\x5e", [0x36c1] = "\xe3\xb8", [0x36c2] = "\xe3\xb2",
  [0x36c4] = "\xbb\x57", [0x36c5] = "\xdf\xd4", [0x36c6] = "\xbb\x56",
  [0x36c7] = "\xe3\xc3", [0x36c9] = "\xbb\x54", [0x36ca] = "\xbb\x63",
  [0x36cb] = "\xbb\x5c", [0x36cc] = "\xe3\xc4", [0x36cd] = "\xe3\xb9",
  [0x36ce] = "\xe3\xb1", [0x36cf] = "\xe3\xcc", [0x36d0] = "\xe3\xbd",
  [0x36d1] = "\xbb\x62", [0x36d2] = "\xe3\xd0", [0x36d3] = "\xbb\x5f",
  [0x36d4] = "\xe3\xcf", [0x36d6] = "\xe3\xc9", [0x36d7] = "\xe3\xce",
  [0x36da] = "\xa0\xcf", [0x36db] = "\xe3\xd1", [0x36de] = "\x8f\x6d",
  [0x36e1] = "\x99\xbe", [0x36e2] = "\x8e\xf4", [0x36e4] = "\x8f\x72",
  [0x36e5] = "\x95\xe4", [0x36e7] = "\xe7\x73", [0x36e8] = "\xe7\x74",
  [0x36e9] = "\xe7\x67", [0x36ea] = "\xe7\x66", [0x36eb] = "\xe7\x62",
  [0x36ec] = "\xbd\xb4", [0x36ee] = "\xbd\xac", [0x36ef] = "\xe7\x76",
  [0x36f0] = "\xe7\x75", [0x36f1] = "\xdf\xa9", [0x36f2] = "\xe7\x5f",
  [0x36f3] = "\xe7\x63", [0x36f4] = "\xe7\x5d", [0x36f6] = "\xe7\x70",
  [0x36f7] = "\xe7\x61", [0x36f8] = "\x99\xbf", [0x36f9] = "\xe7\x77",
  [0x36fa] = "\xe7\x5a", [0x36fb] = "\xe7\x58", [0x36fc] = "\xe7\x64",
  [0x36fd] = "\xe7\x6e", [0x36fe] = "\xe7\x69", [0x36ff] = "\xbd\xb6",
  [0x3700] = "\xe7\x4f", [0x3702] = "\xe7\x6d", [0x3703] = "\x92\x42",
  [0x3705] = "\xfb\xa5", [0x3706] = "\xbd\xb7", [0x3707] = "\xdf\xbd",
  [0x3708] = "\xe7\x5b", [0x3709] = "\xe7\x52", [0x370a] = "\xe7\x55",
  [0x370b] = "\xe7\x7b", [0x370c] = "\xe7\x5c", [0x370d] = "\xe7\x53",
  [0x370e] = "\xe7\x51", [0x370f] = "\xe7\x4e", [0x3710] = "\x99\xc0",
  [0x3711] = "\xbd\xb0", [0x3712] = "\xe7\x65", [0x3713] = "\xbd\xaf",
  [0x3714] = "\xbd\xb3", [0x3715] = "\xe7\x60", [0x3716] = "\xe7\x68",
  [0x3717] = "\xbd\xa9", [0x3718] = "\xe7\x78", [0x3719] = "\xe7\x7c",
  [0x371a] = "\xbd\xab", [0x371c] = "\xe7\x57", [0x371d] = "\xe7\x6b",
  [0x371e] = "\xe7\x6f", [0x371f] = "\xe7\x54", [0x3720] = "\xe7\x79",
  [0x3721] = "\xbd\xb2", [0x3723] = "\xbd\xb1", [0x3724] = "\xe7\x4c",
  [0x3725] = "\xbd\xb5", [0x3726] = "\xe7\x72", [0x3727] = "\xe7\x56",
  [0x3728] = "\xe7\x6a", [0x3729] = "\xe7\x50", [0x372a] = "\xe7\x5e",
  [0x372b] = "\xe7\x59", [0x372c] = "\xbd\xad", [0x372d] = "\xbd\xae",
  [0x372e] = "\xe7\x6c", [0x372f] = "\xe7\x7d", [0x3730] = "\xe7\x7a",
  [0x3731] = "\xe7\x71", [0x3733] = "\xfd\xb4", [0x3734] = "\x8f\x77",
  [0x3738] = "\x99\xc1", [0x373b] = "\xe7\x4d", [0x373d] = "\xbd\xaa",
  [0x373e] = "\xeb\x49", [0x3740] = "\xeb\x40", [0x3741] = "\xeb\x43",
  [0x3742] = "\xfa\xb9", [0x3743] = "\xbf\xbb", [0x3744] = "\xeb\x45",
  [0x3745] = "\xea\xf9", [0x3746] = "\xeb\x41", [0x3747] = "\xeb\x47",
  [0x3748] = "\xbf\xb8", [0x3749] = "\xbf\xbc", [0x374a] = "\xbf\xb6",
  [0x374b] = "\x8f\x40", [0x374c] = "\xfa\x44", [0x374d] = "\xea\xfb",
  [0x374e] = "\xeb\x4c", [0x3751] = "\xeb\x46", [0x3752] = "\x99\xc2",
  [0x3753] = "\xea\xfc", [0x3754] = "\xeb\x55", [0x3755] = "\xeb\x4f",
  [0x3756] = "\xea\xf8", [0x3757] = "\xee\x46", [0x3758] = "\xea\xfe",
  [0x3759] = "\xbf\xb7", [0x375a] = "\x8f\x5c", [0x375b] = "\xeb\x4a",
  [0x375d] = "\xeb\x54", [0x375e] = "\xbf\xbf", [0x3760] = "\xeb\x51",
  [0x3761] = "\xea\xfd", [0x3762] = "\xeb\x44", [0x3763] = "\xeb\x48",
  [0x3764] = "\xeb\x42", [0x3765] = "\xeb\x56", [0x3766] = "\xeb\x53",
  [0x3767] = "\xeb\x50", [0x3768] = "\xbf\xb9", [0x3769] = "\xbf\xba",
  [0x376a] = "\xbf\xbe", [0x376b] = "\xea\xfa", [0x376c] = "\xeb\x57",
  [0x376d] = "\xbf\xbd", [0x376e] = "\xeb\x4d", [0x376f] = "\x99\xc4",
  [0x3770] = "\x99\xc5", [0x3771] = "\xeb\x4b", [0x3773] = "\x8f\x7b",
  [0x3775] = "\xeb\x4e", [0x3776] = "\xee\x53", [0x3777] = "\xee\x40",
  [0x3778] = "\xee\x45", [0x3779] = "\xee\x52", [0x377a] = "\xee\x44",
  [0x377b] = "\xed\xfb", [0x377c] = "\xee\x41", [0x377e] = "\xc1\xa2",
  [0x3780] = "\xed\xf4", [0x3781] = "\xee\x4d", [0x3782] = "\xee\x4f",
  [0x3783] = "\xed\xf3", [0x3784] = "\xc1\xa1", [0x3785] = "\xee\x51",
  [0x3786] = "\xee\x49", [0x3787] = "\xc1\xa8", [0x3788] = "\xee\x50",
  [0x3789] = "\xee\x42", [0x378a] = "\xc1\xaa", [0x378b] = "\xed\xf9",
  [0x378c] = "\xeb\x52", [0x378d] = "\xee\x4a", [0x378e] = "\xee\x47",
  [0x378f] = "\xed\xf5", [0x3790] = "\xee\x55", [0x3791] = "\xc1\xa4",
  [0x3794] = "\xc1\xa5", [0x3795] = "\xed\xf7", [0x3796] = "\xee\x48",
  [0x3798] = "\xee\x54", [0x3799] = "\xee\x4b", [0x379a] = "\xed\xfd",
  [0x379b] = "\xc1\xa7", [0x379c] = "\xc1\xa3", [0x379d] = "\xee\x4c",
  [0x379e] = "\xed\xfe", [0x379f] = "\xee\x56", [0x37a0] = "\xed\xf8",
  [0x37a1] = "\xee\x43", [0x37a2] = "\xee\x4e", [0x37a3] = "\xed\xfa",
  [0x37a4] = "\xed\xfc", [0x37a6] = "\xc2\xcb", [0x37a7] = "\xed\xf6",
  [0x37a8] = "\xc1\xa9", [0x37a9] = "\xc2\xc4", [0x37aa] = "\xc1\x7e",
  [0x37af] = "\xc1\xa6", [0x37b0] = "\xc2\xc8", [0x37b1] = "\xf0\xb3",
  [0x37b3] = "\xf0\xa9", [0x37b4] = "\xf0\xa4", [0x37b5] = "\xf0\xaa",
  [0x37b6] = "\xf0\xb4", [0x37b7] = "\xf0\xb8", [0x37b8] = "\xf0\xb7",
  [0x37b9] = "\xc2\xca", [0x37ba] = "\xc2\xc9", [0x37bd] = "\xf0\xab",
  [0x37be] = "\xf0\xb9", [0x37bf] = "\xf0\xae", [0x37c0] = "\xf0\xa6",
  [0x37c1] = "\x8f\xa3", [0x37c2] = "\xf0\xa8", [0x37c3] = "\xf0\xa7",
  [0x37c4] = "\xf0\xad", [0x37c5] = "\xf0\xb2", [0x37c6] = "\xf0\xa5",
  [0x37c7] = "\xf0\xac", [0x37c8] = "\xf0\xb1", [0x37c9] = "\xc2\xc7",
  [0x37cb] = "\xf0\xaf", [0x37cd] = "\xc2\xc5", [0x37ce] = "\xf0\xb0",
  [0x37cf] = "\xc2\xc3", [0x37d0] = "\xc2\xc6", [0x37d1] = "\xf2\xd5",
  [0x37d2] = "\xf0\xb5", [0x37d5] = "\xc3\xc2", [0x37d7] = "\xf2\xcd",
  [0x37d8] = "\xf2\xd1", [0x37d9] = "\xf2\xc9", [0x37da] = "\xf2\xcc",
  [0x37dc] = "\xf2\xd4", [0x37dd] = "\xc3\xc0", [0x37de] = "\xf2\xd9",
  [0x37df] = "\xf2\xd2", [0x37e0] = "\x99\xc6", [0x37e1] = "\xf2\xca",
  [0x37e2] = "\xf2\xda", [0x37e3] = "\xf2\xd3", [0x37e4] = "\xc3\xc3",
  [0x37e5] = "\xc3\xc4", [0x37e6] = "\xf2\xd7", [0x37e8] = "\xf2\xcb",
  [0x37e9] = "\xc3\xbf", [0x37ea] = "\xc3\xc1", [0x37eb] = "\xf2\xc6",
  [0x37ec] = "\xf2\xce", [0x37ed] = "\xf2\xc8", [0x37ee] = "\x96\xcd",
  [0x37ef] = "\xf2\xd8", [0x37f0] = "\xf2\xd6", [0x37f1] = "\xf2\xc7",
  [0x37f2] = "\xf2\xcf", [0x37f6] = "\xf4\xbe", [0x37f7] = "\xc3\xc5",
  [0x37f8] = "\xf2\xd0", [0x37f9] = "\xc4\xa7", [0x37fa] = "\xc4\xa9",
  [0x37fb] = "\xc4\xa6", [0x37fc] = "\x96\xc7", [0x37fd] = "\xf4\xc3",
  [0x37fe] = "\xf4\xbb", [0x37ff] = "\xf4\xb9", [0x3800] = "\xf4\xbd",
  [0x3801] = "\xf4\xba", [0x3802] = "\x8f\xa5", [0x3804] = "\xf4\xbf",
  [0x3805] = "\xf4\xc1", [0x3806] = "\xc4\xaa", [0x3807] = "\xc4\xac",
  [0x3809] = "\xf4\xc0", [0x380a] = "\xc4\xad", [0x380b] = "\xc4\xab",
  [0x380c] = "\xf4\xc2", [0x380d] = "\xfa\xbb", [0x3810] = "\x95\x70",
  [0x3811] = "\xc4\xa8", [0x3814] = "\x93\x68", [0x3816] = "\x8f\x7e",
  [0x3817] = "\xc4\xf4", [0x3818] = "\xf5\xf1", [0x3819] = "\xf5\xf7",
  [0x381a] = "\xc4\xf6", [0x381b] = "\xf4\xbc", [0x381c] = "\xf5\xf6",
  [0x381e] = "\xf5\xfd", [0x381f] = "\xf5\xf4", [0x3820] = "\xf5\xfb",
  [0x3821] = "\xf5\xfa", [0x3822] = "\xf4\xb8", [0x3823] = "\xf5\xf5",
  [0x3824] = "\xf0\xb6", [0x3825] = "\xf5\xfe", [0x3826] = "\xf5\xf3",
  [0x3827] = "\xf5\xf8", [0x3828] = "\x8f\xaa", [0x3829] = "\xf5\xfc",
  [0x382a] = "\xf5\xf2", [0x382c] = "\xf7\x4a", [0x382d] = "\xc4\xf5",
  [0x382e] = "\xf5\xf9", [0x382f] = "\xa0\x50", [0x3831] = "\xf7\xf4",
  [0x3832] = "\xf7\x4b", [0x3833] = "\xf7\x49", [0x3834] = "\xf7\x47",
  [0x3835] = "\xf7\x48", [0x3836] = "\xf7\x4c", [0x3838] = "\xc5\xd9",
  [0x3839] = "\xf7\xf2", [0x383a] = "\xf7\xf0", [0x383b] = "\xf7\xf5",
  [0x383c] = "\xf7\xf3", [0x383e] = "\xf7\xf6", [0x383f] = "\xc5\xda",
  [0x3840] = "\xf7\xf1", [0x3842] = "\x90\xd3", [0x3843] = "\xf8\xbc",
  [0x3845] = "\x95\x56", [0x3846] = "\xf9\x45", [0x3847] = "\xf9\x46",
  [0x3848] = "\xf9\x47", [0x384b] = "\xf9\xc7", [0x384c] = "\xf9\xbd",
  [0x384d] = "\xca\x4f", [0x384e] = "\xaa\xea", [0x3850] = "\xad\x68",
  [0x3852] = "\xd3\xb8", [0x3853] = "\xd3\xb7", [0x3854] = "\xb0\x40",
  [0x3855] = "\xb3\x42", [0x3856] = "\xd7\x7c", [0x3859] = "\xd7\x7b",
  [0x385b] = "\xb5\xea", [0x385c] = "\xb8\xb8", [0x385e] = "\xb8\xb7",
  [0x385f] = "\xb8\xb9", [0x3861] = "\xe3\xd4", [0x3862] = "\xe7\x7e",
  [0x3863] = "\xeb\x58", [0x3864] = "\xeb\x5a", [0x3865] = "\xeb\x59",
  [0x3867] = "\xc1\xab", [0x3868] = "\xee\x57", [0x3869] = "\xf0\xba",
  [0x386a] = "\xf9\xa5", [0x386b] = "\xa6\xe4", [0x386c] = "\x8f\xb8",
  [0x386d] = "\xcd\xc9", [0x386e] = "\xcd\xca", [0x386f] = "\xcd\xc8",
  [0x3870] = "\xcd\xc7", [0x3871] = "\xaa\xeb", [0x3872] = "\x99\xc8",
  [0x3873] = "\xd0\xa9", [0x3874] = "\xd0\xa7", [0x3877] = "\xd0\xa6",
  [0x3879] = "\xad\x69", [0x387a] = "\xad\x6b", [0x387b] = "\xad\x6a",
  [0x387c] = "\xd0\xa8", [0x387e] = "\x8f\xaf", [0x3885] = "\xd3\xc4",
  [0x3886] = "\xd3\xc1", [0x3887] = "\xd3\xbf", [0x388a] = "\xb0\x41",
  [0x388b] = "\xd3\xc2", [0x388c] = "\xb0\x46", [0x388d] = "\xd3\xbc",
  [0x388e] = "\xd3\xcb", [0x3890] = "\xd3\xcd", [0x3891] = "\xd3\xbd",
  [0x3892] = "\x99\xc9", [0x3893] = "\xb0\x43", [0x3894] = "\xd3\xce",
  [0x3895] = "\xd3\xc9", [0x3896] = "\xd3\xbb", [0x3897] = "\xd3\xc0",
  [0x3898] = "\xd3\xca", [0x3899] = "\xd3\xc6", [0x389a] = "\xd3\xc3",
  [0x389c] = "\xb0\x48", [0x389d] = "\xd3\xcc", [0x389e] = "\xd3\xbe",
  [0x38a0] = "\x95\x79", [0x38a1] = "\xd3\xc7", [0x38a2] = "\xd3\xb9",
  [0x38a3] = "\xb0\x47", [0x38a4] = "\xb0\x44", [0x38a5] = "\xd3\xc5",
  [0x38a7] = "\xd3\xc8", [0x38a8] = "\xd3\xba", [0x38a9] = "\xb0\x45",
  [0x38aa] = "\xb0\x42", [0x38ad] = "\x9f\x49", [0x38af] = "\xb3\x4c",
  [0x38b0] = "\xd7\xa5", [0x38b1] = "\xb3\x4b", [0x38b2] = "\x99\xca",
  [0x38b3] = "\xd7\xa8", [0x38b4] = "\xd7\xab", [0x38b5] = "\xb3\x48",
  [0x38b6] = "\xb3\x46", [0x38b7] = "\xd7\x7e", [0x38b8] = "\xd7\xa9",
  [0x38b9] = "\xd7\xa7", [0x38ba] = "\xd7\xa4", [0x38bb] = "\xd7\xac",
  [0x38bc] = "\xd7\xad", [0x38bd] = "\xd7\xaf", [0x38be] = "\xd7\xb0",
  [0x38bf] = "\xd7\x7d", [0x38c0] = "\xb3\x45", [0x38c1] = "\xd7\xa2",
  [0x38c2] = "\xd7\xa1", [0x38c3] = "\xd7\xae", [0x38c4] = "\xb3\x47",
  [0x38c5] = "\xd7\xa3", [0x38c6] = "\xb3\x49", [0x38c7] = "\xb3\x44",
  [0x38c8] = "\xd7\xa6", [0x38c9] = "\xb3\x4d", [0x38cb] = "\xb3\x4a",
  [0x38cc] = "\xd7\xaa", [0x38d0] = "\xb5\xf1", [0x38d1] = "\xdb\xbf",
  [0x38d3] = "\xdb\xb4", [0x38d4] = "\xb5\xee", [0x38d6] = "\xdf\xe7",
  [0x38d7] = "\xdb\xbd", [0x38d8] = "\xdb\xb1", [0x38d9] = "\xb5\xec",
  [0x38da] = "\xdb\xb6", [0x38db] = "\xb5\xef", [0x38dc] = "\xdb\xba",
  [0x38dd] = "\xdb\xb8", [0x38de] = "\xb5\xf2", [0x38df] = "\xb5\xeb",
  [0x38e2] = "\xdb\xb2", [0x38e3] = "\xdb\xb5", [0x38e4] = "\xb5\xf0",
  [0x38e6] = "\xdb\xb3", [0x38e8] = "\xdb\xbe", [0x38e9] = "\xdb\xbc",
  [0x38ea] = "\xdb\xb7", [0x38eb] = "\xdb\xb9", [0x38ec] = "\xdb\xbb",
  [0x38ed] = "\xb5\xed", [0x38ef] = "\x99\xcb", [0x38f5] = "\xdf\xe8",
  [0x38f6] = "\xdf\xee", [0x38f7] = "\xdf\xe4", [0x38f8] = "\xdf\xea",
  [0x38f9] = "\xb8\xba", [0x38fa] = "\xdf\xe6", [0x38fb] = "\xb8\xc0",
  [0x38fe] = "\xb8\xbf", [0x3900] = "\xb8\xbe", [0x3901] = "\xdf\xed",
  [0x3902] = "\xb8\xc1", [0x3903] = "\xb8\xc2", [0x3904] = "\xdf\xe3",
  [0x3905] = "\xdf\xf0", [0x3906] = "\xb8\xc3", [0x3907] = "\xb8\xbd",
  [0x3908] = "\xb8\xbc", [0x3909] = "\xdf\xec", [0x390a] = "\xb8\xc4",
  [0x390b] = "\xdf\xe2", [0x390c] = "\xdf\xe5", [0x390d] = "\xdf\xef",
  [0x390e] = "\xdf\xeb", [0x3911] = "\xe3\xf4", [0x3912] = "\xe3\xe9",
  [0x3913] = "\xb8\xbb", [0x3918] = "\xbb\x6a", [0x3919] = "\xe3\xdd",
  [0x391a] = "\xe3\xf2", [0x391b] = "\xe3\xde", [0x391c] = "\xbb\x65",
  [0x391e] = "\xe3\xdb", [0x3920] = "\xe3\xe4", [0x3921] = "\xe3\xdc",
  [0x3922] = "\xbb\x67", [0x3923] = "\xe3\xd6", [0x3924] = "\xe3\xf1",
  [0x3925] = "\xbb\x68", [0x3926] = "\xe3\xee", [0x3927] = "\xe3\xef",
  [0x3928] = "\xe3\xd7", [0x3929] = "\xbb\x6d", [0x392a] = "\xe3\xe6",
  [0x392c] = "\xe3\xe0", [0x392d] = "\xe3\xe7", [0x392e] = "\xe3\xda",
  [0x3930] = "\xe3\xf3", [0x3931] = "\xe3\xeb", [0x3932] = "\xe3\xe5",
  [0x3933] = "\xe3\xd5", [0x3934] = "\xbb\x69", [0x3935] = "\xe3\xec",
  [0x3937] = "\xbb\x6c", [0x3938] = "\xe3\xf0", [0x393a] = "\xe3\xea",
  [0x393b] = "\xbb\x66", [0x393c] = "\xe3\xe8", [0x393e] = "\xe3\xe2",
  [0x393f] = "\xbb\x64", [0x3940] = "\xe3\xd9", [0x3941] = "\xe3\xe1",
  [0x3942] = "\xe3\xed", [0x3943] = "\xe3\xdf", [0x3946] = "\xe3\xe3",
  [0x394c] = "\xbd\xc1", [0x394d] = "\xdf\xe9", [0x394e] = "\xe7\xb2",
  [0x394f] = "\xe7\xbb", [0x3950] = "\xe7\xb1", [0x3951] = "\xe7\xad",
  [0x3952] = "\xe7\xaa", [0x3953] = "\xbd\xc2", [0x3954] = "\xe7\xa8",
  [0x3955] = "\xbb\x6b", [0x3956] = "\xe7\xa1", [0x3957] = "\xbd\xc0",
  [0x3958] = "\xe7\xa7", [0x3959] = "\xbd\xbf", [0x395a] = "\xe7\xac",
  [0x395b] = "\xe7\xa9", [0x395c] = "\xe7\xb9", [0x395d] = "\xe7\xb4",
  [0x395e] = "\xe7\xae", [0x395f] = "\xe7\xb3", [0x3960] = "\xbd\xbb",
  [0x3961] = "\xe7\xab", [0x3962] = "\xe7\xbe", [0x3963] = "\xe7\xa2",
  [0x3964] = "\xe7\xa3", [0x3965] = "\xe7\xba", [0x3966] = "\xbd\xbc",
  [0x3967] = "\xe7\xbf", [0x3968] = "\xbd\xbe", [0x3969] = "\xe7\xc0",
  [0x396a] = "\xe7\xb0", [0x396b] = "\xe3\xd8", [0x396c] = "\xe7\xb6",
  [0x396d] = "\xe7\xaf", [0x396e] = "\xe7\xb8", [0x396f] = "\xe7\xb5",
  [0x3970] = "\x9d\xd5", [0x3971] = "\x8f\xb0", [0x3973] = "\xe7\xa6",
  [0x3974] = "\xbd\xb9", [0x3975] = "\xe7\xbd", [0x3976] = "\xbd\xba",
  [0x3977] = "\xe7\xa4", [0x3978] = "\xbd\xbd", [0x3979] = "\xeb\x64",
  [0x397a] = "\xe7\xb7", [0x397b] = "\xe7\xbc", [0x397d] = "\xfa\x7a",
  [0x3981] = "\xeb\x61", [0x3982] = "\xbd\xb8", [0x3983] = "\xbf\xc0",
  [0x3984] = "\xeb\x6b", [0x3985] = "\xeb\x67", [0x3986] = "\x9e\x5f",
  [0x3987] = "\xeb\x65", [0x3988] = "\xeb\x60", [0x3989] = "\xeb\x6f",
  [0x398b] = "\x99\xcd", [0x398c] = "\xa0\xc9", [0x398d] = "\xbf\xc4",
  [0x398f] = "\xeb\x5c", [0x3990] = "\xeb\x68", [0x3991] = "\xeb\x69",
  [0x3992] = "\xeb\x5f", [0x3993] = "\xeb\x5e", [0x3994] = "\xeb\x6c",
  [0x3996] = "\xeb\x62", [0x3997] = "\xeb\x5d", [0x3998] = "\xeb\x63",
  [0x399a] = "\xeb\x6e", [0x399b] = "\xeb\x5b", [0x399c] = "\xeb\x6d",
  [0x399d] = "\xeb\x6a", [0x399e] = "\xbf\xc2", [0x399f] = "\xbf\xc1",
  [0x39a2] = "\xbf\xc3", [0x39a3] = "\xeb\x66", [0x39a4] = "\xf0\xcb",
  [0x39a5] = "\x9a\xdb", [0x39a9] = "\xa0\xc6", [0x39aa] = "\xee\x59",
  [0x39ab] = "\xc1\xb1", [0x39ac] = "\xee\x5d", [0x39ad] = "\xee\x5a",
  [0x39ae] = "\xee\x61", [0x39af] = "\xee\x67", [0x39b0] = "\xee\x5c",
  [0x39b1] = "\x8f\xb4", [0x39b2] = "\xee\x70", [0x39b3] = "\xc1\xae",
  [0x39b4] = "\xee\x6a", [0x39b5] = "\xee\x5f", [0x39b6] = "\xee\x6b",
  [0x39b7] = "\xee\x66", [0x39b8] = "\xee\x6d", [0x39b9] = "\xee\x5e",
  [0x39ba] = "\xc1\xb3", [0x39bb] = "\xc1\xb2", [0x39bc] = "\xee\x60",
  [0x39bd] = "\xee\x6e", [0x39be] = "\xee\x58", [0x39bf] = "\xee\x6c",
  [0x39c0] = "\xc1\xac", [0x39c1] = "\xa0\xd7", [0x39c2] = "\xee\x64",
  [0x39c3] = "\xee\x63", [0x39c4] = "\xee\x68", [0x39c5] = "\xee\x5b",
  [0x39c6] = "\xc1\xb0", [0x39c8] = "\xc1\xb4", [0x39c9] = "\xee\x62",
  [0x39ca] = "\xee\x69", [0x39cb] = "\xc1\xb5", [0x39cc] = "\xee\x65",
  [0x39ce] = "\xa0\xc7", [0x39d1] = "\xc1\xad", [0x39d2] = "\xc1\xaf",
  [0x39d3] = "\xf0\xc7", [0x39d4] = "\xf0\xc5", [0x39d6] = "\xa0\x43",
  [0x39d7] = "\xf0\xcc", [0x39d8] = "\xf0\xc9", [0x39d9] = "\xf0\xcd",
  [0x39da] = "\x8f\xb5", [0x39db] = "\xf0\xbe", [0x39dc] = "\xf0\xc6",
  [0x39dd] = "\xf0\xd1", [0x39de] = "\xee\x6f", [0x39df] = "\xf0\xc2",
  [0x39e0] = "\xc2\xcf", [0x39e1] = "\xe7\xa5", [0x39e2] = "\xf0\xbd",
  [0x39e3] = "\xf0\xca", [0x39e4] = "\xf0\xc4", [0x39e5] = "\xf0\xc1",
  [0x39e6] = "\xf0\xbc", [0x39e7] = "\xf0\xbb", [0x39e8] = "\xf0\xd0",
  [0x39ea] = "\xf0\xc0", [0x39eb] = "\xf0\xbf", [0x39ec] = "\xc2\xcd",
  [0x39ed] = "\xf0\xc8", [0x39ee] = "\x8f\xb2", [0x39ef] = "\xc2\xcc",
  [0x39f2] = "\xc2\xce", [0x39f3] = "\xf0\xc3", [0x39f4] = "\xf0\xcf",
  [0x39f5] = "\xa0\x61", [0x39f6] = "\xf2\xde", [0x39f7] = "\xf2\xdf",
  [0x39f9] = "\xc3\xc9", [0x39fa] = "\xf2\xdc", [0x39fb] = "\xc3\xc6",
  [0x39fc] = "\xf2\xe4", [0x39fe] = "\xc3\xca", [0x39ff] = "\xf2\xe6",
  [0x3a00] = "\xf2\xdb", [0x3a01] = "\xf0\xce", [0x3a02] = "\xf2\xe8",
  [0x3a03] = "\xf2\xdd", [0x3a04] = "\x9e\x5e", [0x3a05] = "\xc3\xc7",
  [0x3a06] = "\xf2\xe3", [0x3a08] = "\xf2\xe5", [0x3a09] = "\xf2\xe0",
  [0x3a0a] = "\xf2\xe7", [0x3a0b] = "\xf2\xe2", [0x3a0c] = "\xf2\xe1",
  [0x3a0d] = "\xc3\xc8", [0x3a0f] = "\x8f\xb6", [0x3a10] = "\xf4\xc5",
  [0x3a11] = "\xf4\xc6", [0x3a13] = "\xf4\xc8", [0x3a14] = "\xc4\xae",
  [0x3a15] = "\xc4\xaf", [0x3a16] = "\xf4\xc9", [0x3a17] = "\xf4\xc7",
  [0x3a18] = "\x9f\xe8", [0x3a19] = "\xf4\xc4", [0x3a1b] = "\xf6\x42",
  [0x3a1c] = "\xf6\x45", [0x3a1d] = "\xf6\x41", [0x3a1f] = "\xc4\xfa",
  [0x3a20] = "\xf6\x43", [0x3a21] = "\xc4\xf9", [0x3a22] = "\xc4\xf8",
  [0x3a23] = "\xc4\xf7", [0x3a24] = "\xf6\x44", [0x3a25] = "\xf7\x51",
  [0x3a26] = "\xf7\x4f", [0x3a27] = "\x9c\xb2", [0x3a28] = "\xf7\x4e",
  [0x3a29] = "\xf6\x40", [0x3a2a] = "\xf7\x50", [0x3a2b] = "\xf6\x46",
  [0x3a2c] = "\xf7\x4d", [0x3a2d] = "\x95\x7c", [0x3a2e] = "\xf7\xf9",
  [0x3a2f] = "\xf7\xd7", [0x3a30] = "\xf7\xf7", [0x3a31] = "\xc5\xdb",
  [0x3a32] = "\xf7\xf8", [0x3a33] = "\xf7\xfa", [0x3a35] = "\xf8\xbf",
  [0x3a36] = "\xc5\xfa", [0x3a37] = "\xf8\xbe", [0x3a38] = "\xf8\xbd",
  [0x3a39] = "\xc5\xfb", [0x3a3b] = "\xc6\x5a", [0x3a3c] = "\xf9\x6e",
  [0x3a3d] = "\xf9\xa7", [0x3a3e] = "\xf9\xa6", [0x3a3f] = "\xf9\xa8",
  [0x3a40] = "\xa6\xe5", [0x3a41] = "\xd0\xaa", [0x3a42] = "\x9f\xc7",
  [0x3a43] = "\xd3\xcf", [0x3a44] = "\xd3\xd0", [0x3a45] = "\x8f\xbb",
  [0x3a46] = "\x8f\xbc", [0x3a48] = "\xdb\xc0", [0x3a4a] = "\xf6\x47",
  [0x3a4b] = "\xf8\xc0", [0x3a4c] = "\xa6\xe6", [0x3a4d] = "\xad\x6c",
  [0x3a4e] = "\xd0\xab", [0x3a4f] = "\x8f\xec", [0x3a52] = "\xd7\xb1",
  [0x3a53] = "\xb3\x4e", [0x3a55] = "\xdb\xc2", [0x3a56] = "\xdb\xc1",
  [0x3a57] = "\xb5\xf3", [0x3a59] = "\xb8\xc5", [0x3a5a] = "\xe7\xc1",
  [0x3a5b] = "\xbd\xc3", [0x3a5d] = "\xbd\xc4", [0x3a5e] = "\x8f\xc0",
  [0x3a60] = "\x93\x6a", [0x3a61] = "\xbf\xc5", [0x3a62] = "\xc5\xfc",
  [0x3a63] = "\xa6\xe7", [0x3a64] = "\x8b\xe4", [0x3a65] = "\x9c\x7c",
  [0x3a67] = "\xd0\xac", [0x3a68] = "\xaa\xed", [0x3a69] = "\xd0\xae",
  [0x3a6a] = "\xd0\xad", [0x3a6b] = "\xad\x6d", [0x3a6d] = "\xd3\xd1",
  [0x3a6e] = "\x95\xa1", [0x3a6f] = "\xd3\xd8", [0x3a70] = "\xb0\x49",
  [0x3a71] = "\xd3\xd6", [0x3a72] = "\xd3\xd4", [0x3a74] = "\xd3\xdb",
  [0x3a75] = "\xd3\xd2", [0x3a76] = "\xd3\xd3", [0x3a77] = "\xb0\x4a",
  [0x3a79] = "\xb0\x4e", [0x3a7c] = "\xd3\xdc", [0x3a7d] = "\xb0\x4d",
  [0x3a7e] = "\xd3\xda", [0x3a7f] = "\xd3\xd7", [0x3a80] = "\xd3\xd5",
  [0x3a81] = "\xb0\x4b", [0x3a82] = "\xb0\x4c", [0x3a83] = "\xd3\xd9",
  [0x3a84] = "\xfe\xec", [0x3a87] = "\x95\xa3", [0x3a88] = "\xb3\x50",
  [0x3a89] = "\xd7\xb2", [0x3a8b] = "\xb3\x55", [0x3a8c] = "\xd7\xc2",
  [0x3a8d] = "\xb3\x54", [0x3a8e] = "\xd7\xc4", [0x3a91] = "\xd7\xb8",
  [0x3a92] = "\xb3\x52", [0x3a93] = "\xd7\xc3", [0x3a95] = "\xd7\xb3",
  [0x3a96] = "\xb3\x53", [0x3a97] = "\xd7\xbf", [0x3a98] = "\xd7\xbb",
  [0x3a99] = "\xd7\xbd", [0x3a9a] = "\xd7\xb7", [0x3a9b] = "\xd7\xbe",
  [0x3a9c] = "\x8f\xc1", [0x3a9e] = "\xb3\x4f", [0x3a9f] = "\xd7\xba",
  [0x3aa0] = "\xa0\x52", [0x3aa1] = "\xd7\xb9", [0x3aa2] = "\xd7\xb5",
  [0x3aa4] = "\xd7\xc0", [0x3aa7] = "\xd7\xbc", [0x3aa8] = "\xd7\xb4",
  [0x3aaa] = "\xd7\xb6", [0x3aab] = "\xb3\x51", [0x3aac] = "\xd7\xc1",
  [0x3aae] = "\x99\xd0", [0x3ab1] = "\xb5\xf6", [0x3ab2] = "\xdb\xcd",
  [0x3ab4] = "\x8f\xc3", [0x3ab5] = "\x8f\xc4", [0x3ab6] = "\xdb\xc9",
  [0x3ab7] = "\xdb\xcb", [0x3ab8] = "\xdb\xc6", [0x3ab9] = "\xdb\xc5",
  [0x3aba] = "\xdb\xc3", [0x3abc] = "\xdb\xca", [0x3abd] = "\xdb\xcc",
  [0x3abe] = "\xdb\xc8", [0x3abf] = "\x95\xa4", [0x3ac0] = "\xdb\xc7",
  [0x3ac1] = "\xb5\xf4", [0x3ac2] = "\xb5\xf5", [0x3ac5] = "\x8f\xc6",
  [0x3ac7] = "\x9e\x60", [0x3ac9] = "\xdb\xcf", [0x3aca] = "\xb8\xcd",
  [0x3acb] = "\xdf\xf2", [0x3acc] = "\xdf\xf8", [0x3acd] = "\xdf\xf3",
  [0x3ace] = "\xdf\xf4", [0x3acf] = "\xf9\xd8", [0x3ad0] = "\xdf\xf9",
  [0x3ad2] = "\xb8\xcf", [0x3ad4] = "\xb8\xc7", [0x3ad5] = "\xb8\xce",
  [0x3ad6] = "\xdf\xf1", [0x3ad7] = "\xdb\xc4", [0x3ad8] = "\xb8\xca",
  [0x3ad9] = "\xb8\xc8", [0x3ada] = "\xdf\xf7", [0x3adb] = "\xdf\xf6",
  [0x3adc] = "\xb8\xc9", [0x3add] = "\xb8\xcb", [0x3ade] = "\xdf\xf5",
  [0x3adf] = "\xb8\xc6", [0x3ae1] = "\xb8\xcc", [0x3ae6] = "\x95\xa5",
  [0x3ae7] = "\xe3\xf6", [0x3ae8] = "\xbb\x74", [0x3aeb] = "\xe4\x42",
  [0x3aec] = "\xe4\x41", [0x3aee] = "\xe3\xfb", [0x3aef] = "\xbb\x76",
  [0x3af0] = "\xe4\x40", [0x3af1] = "\xe3\xf7", [0x3af2] = "\xe3\xf8",
  [0x3af3] = "\xbb\x6e", [0x3af4] = "\xbb\x70", [0x3af5] = "\x9c\xb3",
  [0x3af6] = "\xe3\xfd", [0x3af7] = "\xe3\xf5", [0x3af8] = "\xbb\x72",
  [0x3af9] = "\xbb\x71", [0x3afa] = "\xe3\xf9", [0x3afb] = "\xe3\xfe",
  [0x3afc] = "\xe3\xfc", [0x3afd] = "\xbb\x73", [0x3afe] = "\xe3\xfa",
  [0x3aff] = "\x99\xd1", [0x3b00] = "\xfe\xf1", [0x3b01] = "\xdb\xce",
  [0x3b02] = "\xbb\x6f", [0x3b05] = "\xe7\xc2", [0x3b06] = "\xe7\xc9",
  [0x3b07] = "\xbd\xc6", [0x3b09] = "\xe7\xcd", [0x3b0a] = "\xbd\xca",
  [0x3b0b] = "\xe7\xc5", [0x3b0c] = "\xe7\xc3", [0x3b0e] = "\xe7\xcc",
  [0x3b10] = "\xbd\xc5", [0x3b11] = "\xe7\xcb", [0x3b12] = "\xbd\xc7",
  [0x3b13] = "\xbd\xc8", [0x3b14] = "\xe7\xc4", [0x3b15] = "\xbd\xc9",
  [0x3b16] = "\xe7\xca", [0x3b17] = "\xe7\xc6", [0x3b18] = "\xe7\xc7",
  [0x3b19] = "\xe7\xc8", [0x3b1a] = "\xbb\x75", [0x3b1e] = "\xeb\x70",
  [0x3b1f] = "\xeb\x7c", [0x3b21] = "\xbf\xca", [0x3b22] = "\xeb\x77",
  [0x3b23] = "\xeb\x79", [0x3b24] = "\x99\xd2", [0x3b25] = "\xbf\xc8",
  [0x3b26] = "\xeb\x71", [0x3b27] = "\xeb\x75", [0x3b29] = "\xeb\x78",
  [0x3b2a] = "\xbf\xc6", [0x3b2b] = "\xbf\xc9", [0x3b2c] = "\xeb\x7b",
  [0x3b2d] = "\xeb\x73", [0x3b2e] = "\xeb\x74", [0x3b2f] = "\xeb\x7a",
  [0x3b30] = "\xeb\x72", [0x3b31] = "\xeb\x76", [0x3b32] = "\xbf\xc7",
  [0x3b33] = "\xee\x72", [0x3b35] = "\xee\x71", [0x3b36] = "\xc1\xb7",
  [0x3b37] = "\xee\x77", [0x3b38] = "\xc1\xb9", [0x3b3b] = "\xc1\xb6",
  [0x3b3c] = "\xee\x73", [0x3b3d] = "\xc1\xba", [0x3b3e] = "\xee\x74",
  [0x3b41] = "\xee\x75", [0x3b42] = "\xee\x78", [0x3b43] = "\x9c\xc2",
  [0x3b44] = "\xc1\xb8", [0x3b46] = "\xf0\xd6", [0x3b47] = "\x99\xd3",
  [0x3b49] = "\xf0\xd9", [0x3b4b] = "\xf0\xd3", [0x3b4c] = "\xf0\xd5",
  [0x3b4d] = "\x95\xa7", [0x3b4f] = "\xf0\xd4", [0x3b50] = "\xf0\xd7",
  [0x3b51] = "\xf0\xd8", [0x3b52] = "\xee\x76", [0x3b53] = "\xf0\xd2",
  [0x3b54] = "\x95\xa9", [0x3b56] = "\xc3\xcd", [0x3b57] = "\xf2\xec",
  [0x3b58] = "\xf2\xef", [0x3b59] = "\xf2\xf1", [0x3b5a] = "\xf2\xea",
  [0x3b5b] = "\xf2\xeb", [0x3b5c] = "\xf2\xee", [0x3b5d] = "\xf2\xf0",
  [0x3b5e] = "\xc3\xce", [0x3b5f] = "\xc3\xcc", [0x3b60] = "\xc3\xcb",
  [0x3b61] = "\xf2\xed", [0x3b62] = "\xf2\xe9", [0x3b63] = "\xf4\xca",
  [0x3b64] = "\xc4\xb0", [0x3b65] = "\x95\xa6", [0x3b66] = "\xf4\xcb",
  [0x3b69] = "\xf6\x49", [0x3b6a] = "\xc4\xfb", [0x3b6b] = "\xf6\x4b",
  [0x3b6c] = "\xc4\xfc", [0x3b6d] = "\xf6\x48", [0x3b6e] = "\xf6\x4a",
  [0x3b6f] = "\xc5\xa8", [0x3b71] = "\xf7\x52", [0x3b72] = "\xc5\xa7",
  [0x3b73] = "\xf7\xfd", [0x3b74] = "\xf7\xfc", [0x3b76] = "\xf7\xfb",
  [0x3b77] = "\x9c\x5d", [0x3b79] = "\xf9\x48", [0x3b7a] = "\xf9\x49",
  [0x3b7b] = "\xf9\x4b", [0x3b7c] = "\xf9\x4a", [0x3b7e] = "\xca\x50",
  [0x3b7f] = "\xa6\xe8", [0x3b80] = "\x98\xe2", [0x3b81] = "\xad\x6e",
  [0x3b82] = "\xd7\xc5", [0x3b83] = "\xb5\xf7", [0x3b85] = "\xdf\xfa",
  [0x3b86] = "\xc2\xd0", [0x3b87] = "\x8f\xc9", [0x3b88] = "\xf2\xf2",
  [0x3b89] = "\xa0\xc2", [0x3b8a] = "\x8f\xca", [0x3b8b] = "\xa8\xa3",
  [0x3b8f] = "\xb3\x57", [0x3b91] = "\x99\xd4", [0x3b93] = "\xb3\x56",
  [0x3b94] = "\xa0\xb9", [0x3b95] = "\xdb\xd0", [0x3b96] = "\xb5\xf8",
  [0x3b97] = "\xdb\xd2", [0x3b98] = "\xdb\xd1", [0x3b9b] = "\xdf\xfb",
  [0x3b9c] = "\xb8\xd0", [0x3b9d] = "\xe4\x43", [0x3b9e] = "\xe4\x46",
  [0x3b9f] = "\xe4\x45", [0x3ba1] = "\xe4\x44", [0x3ba2] = "\xe7\xce",
  [0x3ba3] = "\xe7\xd0", [0x3ba4] = "\xe7\xcf", [0x3ba5] = "\x9b\x58",
  [0x3ba6] = "\xbf\xcc", [0x3ba7] = "\x8f\xcd", [0x3ba9] = "\xa0\xd4",
  [0x3baa] = "\xbf\xcb", [0x3bac] = "\xc1\xbb", [0x3bad] = "\xee\x79",
  [0x3bae] = "\xee\x7b", [0x3baf] = "\xee\x7a", [0x3bb2] = "\xc2\xd1",
  [0x3bb6] = "\xf2\xf4", [0x3bb7] = "\xf2\xf3", [0x3bb9] = "\xf4\xcc",
  [0x3bba] = "\xc4\xb1", [0x3bbc] = "\x8f\xce", [0x3bbd] = "\xc4\xfd",
  [0x3bbe] = "\xf7\x54", [0x3bbf] = "\xf7\x53", [0x3bc0] = "\xc6\x5b",
  [0x3bc1] = "\x8b\xe5", [0x3bc6] = "\x89\x79", [0x3bd2] = "\xa8\xa4",
  [0x3bd3] = "\xd0\xaf", [0x3bd4] = "\xad\x6f", [0x3bd5] = "\xd7\xc8",
  [0x3bd6] = "\xd7\xc6", [0x3bd9] = "\xd7\xc7", [0x3bda] = "\xdb\xd4",
  [0x3bdb] = "\xdb\xd5", [0x3bdc] = "\xe0\x43", [0x3bdd] = "\xdb\xd3",
  [0x3bdf] = "\xdf\xfc", [0x3be0] = "\xe0\x41", [0x3be1] = "\xe0\x40",
  [0x3be2] = "\xe0\x42", [0x3be3] = "\xb8\xd1", [0x3be4] = "\xdf\xfe",
  [0x3be5] = "\xdf\xfd", [0x3be6] = "\xe0\x44", [0x3be7] = "\x8f\xd0",
  [0x3be8] = "\xe4\x49", [0x3be9] = "\xe4\x47", [0x3beb] = "\xe4\x48",
  [0x3bec] = "\xe7\xd3", [0x3bed] = "\xe7\xd1", [0x3bf0] = "\xe7\xd2",
  [0x3bf1] = "\xeb\x7d", [0x3bf2] = "\xee\x7c", [0x3bf3] = "\xee\x7d",
  [0x3bf4] = "\xc2\xd2", [0x3bf6] = "\xf2\xf5", [0x3bf7] = "\xf4\xcd",
  [0x3bf8] = "\xc4\xb2", [0x3bfa] = "\xf6\x4c", [0x3bfb] = "\xf7\x55",
  [0x3bfc] = "\xc5\xa9", [0x3bfe] = "\xf7\xfe", [0x3bff] = "\xf9\x4c",
  [0x3c00] = "\xa8\xa5", [0x3c02] = "\xad\x71", [0x3c03] = "\xad\x72",
  [0x3c04] = "\xd0\xb0", [0x3c07] = "\xd0\xb1", [0x3c08] = "\xad\x70",
  [0x3c0a] = "\xb0\x54", [0x3c0c] = "\xb0\x52", [0x3c0e] = "\xb0\x51",
  [0x3c0f] = "\xb0\x58", [0x3c10] = "\xb0\x50", [0x3c11] = "\xb0\x59",
  [0x3c12] = "\xd3\xdd", [0x3c13] = "\xb0\x56", [0x3c15] = "\xb0\x53",
  [0x3c16] = "\xb0\x57", [0x3c17] = "\xb0\x55", [0x3c18] = "\xb0\x4f",
  [0x3c1b] = "\xb3\x5f", [0x3c1c] = "\x95\xb6", [0x3c1d] = "\xb3\x59",
  [0x3c1e] = "\xd7\xcc", [0x3c1f] = "\xb3\x5e", [0x3c22] = "\xb3\x60",
  [0x3c23] = "\xb3\x5a", [0x3c25] = "\xb3\x5b", [0x3c27] = "\xd7\xca",
  [0x3c29] = "\x99\xd6", [0x3c2a] = "\xb3\x58", [0x3c2b] = "\x95\xe5",
  [0x3c2c] = "\xd7\xcb", [0x3c2d] = "\xb3\x5d", [0x3c30] = "\xd7\xc9",
  [0x3c31] = "\xb3\x5c", [0x3c34] = "\xb6\x44", [0x3c36] = "\xb6\x46",
  [0x3c38] = "\x99\xd7", [0x3c39] = "\xdb\xd8", [0x3c3a] = "\xb6\x45",
  [0x3c3b] = "\xb5\xf9", [0x3c3c] = "\xb5\xfd", [0x3c3d] = "\x95\xb5",
  [0x3c3e] = "\xb8\xe4", [0x3c3f] = "\xe0\x49", [0x3c40] = "\xdb\xda",
  [0x3c41] = "\xb5\xfe", [0x3c44] = "\xdb\xdd", [0x3c45] = "\xdb\xde",
  [0x3c46] = "\xb6\x43", [0x3c48] = "\xdb\xe0", [0x3c49] = "\xa0\xca",
  [0x3c4a] = "\xdb\xe2", [0x3c4c] = "\xdb\xe3", [0x3c4d] = "\xdb\xd7",
  [0x3c4e] = "\xdb\xd6", [0x3c4f] = "\xdb\xe4", [0x3c50] = "\xb6\x42",
  [0x3c51] = "\xdb\xe1", [0x3c52] = "\xdb\xdf", [0x3c54] = "\xb6\x40",
  [0x3c55] = "\xb5\xfb", [0x3c56] = "\xb6\x47", [0x3c57] = "\xdb\xdb",
  [0x3c58] = "\xdb\xdc", [0x3c59] = "\xdb\xd9", [0x3c5b] = "\xb6\x41",
  [0x3c5e] = "\xb5\xfc", [0x3c60] = "\xb5\xfa", [0x3c61] = "\xe0\x48",
  [0x3c62] = "\xb8\xdf", [0x3c63] = "\xb8\xda", [0x3c66] = "\xb8\xd5",
  [0x3c67] = "\x9f\xfd", [0x3c68] = "\xb8\xe5", [0x3c69] = "\xb8\xd6",
  [0x3c6b] = "\xb8\xd2", [0x3c6c] = "\xb8\xe1", [0x3c6d] = "\xb8\xde",
  [0x3c6e] = "\xb8\xe0", [0x3c70] = "\xb8\xd7", [0x3c71] = "\xb8\xdc",
  [0x3c72] = "\xb8\xd3", [0x3c73] = "\xb8\xd4", [0x3c74] = "\xe0\x50",
  [0x3c75] = "\xe0\x4d", [0x3c76] = "\xe0\x45", [0x3c77] = "\xe0\x4a",
  [0x3c79] = "\xb8\xe2", [0x3c7a] = "\xe0\x51", [0x3c7b] = "\xb8\xe3",
  [0x3c7c] = "\xb8\xd9", [0x3c7e] = "\xa0\x58", [0x3c7f] = "\xe0\x47",
  [0x3c81] = "\xe0\x4f", [0x3c82] = "\xe0\x4b", [0x3c83] = "\xe0\x4e",
  [0x3c84] = "\xe0\x4c", [0x3c85] = "\xb8\xdd", [0x3c86] = "\xe0\x46",
  [0x3c87] = "\xb8\xd8", [0x3c8b] = "\xe4\x4c", [0x3c8c] = "\xbb\x78",
  [0x3c8d] = "\xbb\x7b", [0x3c8f] = "\xe4\x4e", [0x3c90] = "\x8f\xd6",
  [0x3c91] = "\xbb\xa5", [0x3c92] = "\xe4\x4d", [0x3c93] = "\xbb\x7d",
  [0x3c94] = "\x99\xd8", [0x3c95] = "\xbd\xcf", [0x3c96] = "\xe4\x4f",
  [0x3c98] = "\xbb\xa4", [0x3c99] = "\xe4\x4b", [0x3c9a] = "\xbb\xa6",
  [0x3c9c] = "\x8f\xd3", [0x3c9e] = "\xbb\x79", [0x3ca0] = "\xb8\xdb",
  [0x3ca1] = "\xbb\x7c", [0x3ca3] = "\xbb\x7a", [0x3ca4] = "\xbb\x7e",
  [0x3ca5] = "\xbb\xa2", [0x3ca6] = "\xbb\x77", [0x3ca7] = "\xbb\xa7",
  [0x3ca8] = "\xbb\xa3", [0x3ca9] = "\x8f\xe5", [0x3caa] = "\xbb\xa1",
  [0x3cab] = "\xe4\x4a", [0x3caf] = "\x8f\xe9", [0x3cb0] = "\xbd\xd6",
  [0x3cb2] = "\xbd\xd2", [0x3cb4] = "\x99\xd9", [0x3cb6] = "\xbd\xd9",
  [0x3cb8] = "\xe7\xd6", [0x3cb9] = "\xbd\xda", [0x3cba] = "\xe7\xe2",
  [0x3cbb] = "\xe7\xdb", [0x3cbc] = "\xbd\xcb", [0x3cbd] = "\xe7\xe3",
  [0x3cbe] = "\xe7\xdd", [0x3cbf] = "\xbd\xd5", [0x3cc0] = "\xe7\xde",
  [0x3cc2] = "\xbd\xd4", [0x3cc3] = "\xe7\xe1", [0x3cc4] = "\xbd\xce",
  [0x3cc5] = "\xe7\xdf", [0x3cc6] = "\xe7\xd5", [0x3cc7] = "\xbd\xcd",
  [0x3cc8] = "\xeb\xaa", [0x3cc9] = "\xbd\xd3", [0x3ccb] = "\xbd\xd0",
  [0x3ccd] = "\xbd\xd8", [0x3ccf] = "\xe7\xd4", [0x3cd1] = "\xe7\xd8",
  [0x3cd2] = "\xbd\xcc", [0x3cd3] = "\xe7\xd7", [0x3cd4] = "\xe7\xd9",
  [0x3cd5] = "\xe7\xda", [0x3cd6] = "\xbd\xd7", [0x3cd7] = "\xe7\xdc",
  [0x3cd8] = "\xe7\xe0", [0x3cd9] = "\xe7\xe4", [0x3cda] = "\x92\x7c",
  [0x3cdb] = "\xbd\xdb", [0x3cdc] = "\xbf\xd2", [0x3cdd] = "\xeb\xa5",
  [0x3cde] = "\xeb\xab", [0x3cdf] = "\xeb\xa8", [0x3ce0] = "\xeb\x7e",
  [0x3ce1] = "\xeb\xac", [0x3ce2] = "\xeb\xa1", [0x3ce4] = "\xeb\xa7",
  [0x3ce6] = "\xbf\xcd", [0x3ce7] = "\xbf\xd3", [0x3ce8] = "\xeb\xad",
  [0x3cea] = "\x9c\x45", [0x3ceb] = "\xbf\xcf", [0x3ced] = "\xbf\xd9",
  [0x3cee] = "\xbf\xd4", [0x3cef] = "\xeb\xaf", [0x3cf0] = "\xeb\xa9",
  [0x3cf1] = "\xbf\xd0", [0x3cf2] = "\xeb\xa2", [0x3cf3] = "\xbf\xda",
  [0x3cf4] = "\xeb\xa3", [0x3cf5] = "\xeb\xa4", [0x3cf6] = "\xbf\xdb",
  [0x3cf7] = "\xbf\xd8", [0x3cf8] = "\xbd\xd1", [0x3cfa] = "\xbf\xce",
  [0x3cfb] = "\xeb\xb0", [0x3cfc] = "\xbf\xdc", [0x3cfe] = "\xbf\xd5",
  [0x3cff] = "\xeb\xae", [0x3d00] = "\xbf\xd1", [0x3d01] = "\xbf\xd6",
  [0x3d02] = "\xbf\xd7", [0x3d04] = "\xc1\xc3", [0x3d05] = "\xee\xa4",
  [0x3d06] = "\xee\xad", [0x3d07] = "\xee\xaa", [0x3d08] = "\xee\xac",
  [0x3d0a] = "\xc1\xc0", [0x3d0b] = "\xee\xa5", [0x3d0c] = "\x8f\xde",
  [0x3d0d] = "\xee\xab", [0x3d0e] = "\xc1\xbc", [0x3d0f] = "\xee\xa7",
  [0x3d10] = "\xc1\xc4", [0x3d11] = "\xee\xa3", [0x3d12] = "\xee\xa8",
  [0x3d13] = "\xee\xaf", [0x3d14] = "\xeb\xa6", [0x3d15] = "\xee\xa9",
  [0x3d16] = "\xee\xa2", [0x3d17] = "\xc1\xbd", [0x3d18] = "\xee\xa1",
  [0x3d19] = "\xc1\xbe", [0x3d1a] = "\xee\xb0", [0x3d1b] = "\xc1\xbf",
  [0x3d1c] = "\xee\xae", [0x3d1d] = "\xc1\xc2", [0x3d1e] = "\xee\x7e",
  [0x3d1f] = "\x8f\xdf", [0x3d20] = "\xc1\xc1", [0x3d22] = "\xee\xa6",
  [0x3d23] = "\xf0\xdc", [0x3d24] = "\xf0\xea", [0x3d25] = "\xf0\xe5",
  [0x3d26] = "\xf0\xe7", [0x3d27] = "\xf0\xdb", [0x3d28] = "\xc2\xd3",
  [0x3d2a] = "\xf0\xda", [0x3d2b] = "\xc2\xd6", [0x3d2c] = "\xc2\xd5",
  [0x3d2d] = "\xa0\x4b", [0x3d2e] = "\xf0\xe9", [0x3d2f] = "\xf0\xe1",
  [0x3d30] = "\xf0\xde", [0x3d31] = "\xf0\xe4", [0x3d33] = "\xf0\xdd",
  [0x3d35] = "\xf0\xdf", [0x3d36] = "\xf0\xe8", [0x3d37] = "\xf0\xe6",
  [0x3d39] = "\xc2\xd4", [0x3d3a] = "\xf0\xed", [0x3d3b] = "\xf0\xeb",
  [0x3d3c] = "\xf0\xe2", [0x3d3d] = "\xf0\xec", [0x3d3e] = "\xf0\xe3",
  [0x3d3f] = "\x8f\xe2", [0x3d40] = "\xf2\xf9", [0x3d41] = "\xc3\xcf",
  [0x3d42] = "\xf3\x41", [0x3d43] = "\xa0\xcc", [0x3d45] = "\xf6\x4f",
  [0x3d46] = "\xc3\xd6", [0x3d47] = "\xf0\xe0", [0x3d48] = "\xf2\xf7",
  [0x3d49] = "\xc3\xd2", [0x3d4a] = "\xf2\xf8", [0x3d4b] = "\xf2\xfd",
  [0x3d4c] = "\x8f\xe3", [0x3d4d] = "\x8f\xe4", [0x3d4e] = "\xc3\xd4",
  [0x3d4f] = "\xc3\xd5", [0x3d50] = "\xf2\xf6", [0x3d51] = "\xf3\x40",
  [0x3d52] = "\xf3\x42", [0x3d53] = "\xf2\xfa", [0x3d54] = "\xf2\xfc",
  [0x3d55] = "\xf2\xfe", [0x3d56] = "\xf2\xfb", [0x3d57] = "\xf3\x43",
  [0x3d58] = "\xc3\xd1", [0x3d59] = "\xc3\xd7", [0x3d5a] = "\xc3\xd3",
  [0x3d5c] = "\xc3\xd0", [0x3d5d] = "\xf4\xd0", [0x3d5e] = "\x9b\xc4",
  [0x3d5f] = "\xc4\xb7", [0x3d60] = "\xf4\xce", [0x3d62] = "\x9b\xfc",
  [0x3d63] = "\xf4\xd2", [0x3d65] = "\xf4\xd3", [0x3d66] = "\xc4\xb5",
  [0x3d67] = "\xf4\xd4", [0x3d68] = "\xf4\xd1", [0x3d69] = "\x96\x4c",
  [0x3d6a] = "\xf4\xcf", [0x3d6b] = "\xc4\xb8", [0x3d6c] = "\xc4\xb4",
  [0x3d6d] = "\xf4\xd5", [0x3d6f] = "\xc4\xb6", [0x3d70] = "\xc4\xb3",
  [0x3d74] = "\xc4\xfe", [0x3d77] = "\xc5\x40", [0x3d78] = "\xf6\x4e",
  [0x3d79] = "\xf6\x4d", [0x3d7a] = "\xf6\x50", [0x3d7b] = "\xf6\x51",
  [0x3d7d] = "\xc5\x41", [0x3d7e] = "\xf7\x56", [0x3d7f] = "\xf7\x5b",
  [0x3d80] = "\xc5\xaa", [0x3d81] = "\x9a\xf6", [0x3d82] = "\xf7\x58",
  [0x3d84] = "\xf7\x57", [0x3d85] = "\xf7\x5a", [0x3d86] = "\xf7\x59",
  [0x3d88] = "\xf8\x43", [0x3d8a] = "\xc5\xdc", [0x3d8b] = "\xf8\x42",
  [0x3d8c] = "\xf8\x40", [0x3d8e] = "\xf8\x41", [0x3d90] = "\x8f\xe7",
  [0x3d92] = "\xc5\xfe", [0x3d93] = "\xc5\xfd", [0x3d94] = "\xf8\xc1",
  [0x3d95] = "\xf8\xc2", [0x3d96] = "\xc6\x40", [0x3d98] = "\xf9\x4d",
  [0x3d99] = "\xf9\x4e", [0x3d9a] = "\xc6\x67", [0x3d9b] = "\x8f\xe8",
  [0x3d9c] = "\xc6\x6d", [0x3d9e] = "\xf9\xa9", [0x3d9f] = "\xf9\xc8",
  [0x3da0] = "\x8b\xe7", [0x3dbe] = "\x89\x7a", [0x3de2] = "\x89\x7b",
  [0x3e37] = "\xa8\xa6", [0x3e39] = "\xd7\xcd", [0x3e3b] = "\xd7\xce",
  [0x3e3c] = "\xe0\x52", [0x3e3d] = "\xe4\x50", [0x3e3e] = "\xe7\xe5",
  [0x3e3f] = "\xc1\xc6", [0x3e41] = "\xc1\xc5", [0x3e42] = "\xf0\xee",
  [0x3e43] = "\xf3\x44", [0x3e45] = "\xf8\x44", [0x3e46] = "\xa8\xa7",
  [0x3e47] = "\xd3\xde", [0x3e48] = "\xb0\x5a", [0x3e49] = "\xb3\x61",
  [0x3e4a] = "\xe0\x54", [0x3e4b] = "\xe0\x53", [0x3e4c] = "\xbd\xdc",
  [0x3e4d] = "\xe7\xe6", [0x3e4e] = "\xbd\xdd", [0x3e4f] = "\xee\xb1",
  [0x3e50] = "\xc2\xd7", [0x3e51] = "\x99\xda", [0x3e54] = "\xc6\x76",
  [0x3e55] = "\xa8\xa8", [0x3e56] = "\xcd\xcb", [0x3e57] = "\xd3\xdf",
  [0x3e5a] = "\xb3\x62", [0x3e5c] = "\xd7\xcf", [0x3e5d] = "\xd7\xd0",
  [0x3e5f] = "\xdb\xe5", [0x3e61] = "\xb6\x48", [0x3e62] = "\xb8\xe6",
  [0x3e64] = "\xe0\x56", [0x3e65] = "\xe0\x55", [0x3e66] = "\xe0\x57",
  [0x3e68] = "\xe4\x51", [0x3e69] = "\xe4\x52", [0x3e6a] = "\xbb\xa8",
  [0x3e6b] = "\xbf\xdd", [0x3e6c] = "\xbd\xde", [0x3e6d] = "\xbf\xde",
  [0x3e6f] = "\xee\xb5", [0x3e70] = "\xee\xb2", [0x3e71] = "\xee\xb4",
  [0x3e72] = "\xee\xb3", [0x3e73] = "\xc1\xc7", [0x3e75] = "\xf0\xef",
  [0x3e76] = "\xf3\x46", [0x3e77] = "\xf3\x45", [0x3e78] = "\xcb\xa4",
  [0x3e79] = "\xb0\x5c", [0x3e7a] = "\xb0\x5b", [0x3e7b] = "\xd3\xe0",
  [0x3e7d] = "\xd7\xd1", [0x3e80] = "\xdb\xe7", [0x3e81] = "\xdb\xe6",
  [0x3e82] = "\xb6\x49", [0x3e84] = "\xe0\x59", [0x3e85] = "\xe0\x5a",
  [0x3e86] = "\xe0\x58", [0x3e89] = "\xb8\xe8", [0x3e8a] = "\xb8\xe7",
  [0x3e8c] = "\xbb\xaa", [0x3e8d] = "\xbb\xa9", [0x3e8f] = "\xe7\xe7",
  [0x3e90] = "\xeb\xb3", [0x3e91] = "\xeb\xb1", [0x3e92] = "\xeb\xb2",
  [0x3e93] = "\xbf\xdf", [0x3e94] = "\xee\xb7", [0x3e95] = "\xee\xb6",
  [0x3e97] = "\xf0\xf2", [0x3e98] = "\xf0\xf1", [0x3e99] = "\xf0\xf0",
  [0x3e9a] = "\xf3\x47", [0x3e9b] = "\x8f\xed", [0x3e9c] = "\xf9\xaa",
  [0x3e9d] = "\xa8\xa9", [0x3e9e] = "\xad\x73", [0x3e9f] = "\x95\xc0",
  [0x3ea0] = "\xad\x74", [0x3ea1] = "\xb0\x5d", [0x3ea2] = "\xb0\x5e",
  [0x3ea3] = "\xd3\xe2", [0x3ea4] = "\xd3\xe1", [0x3ea5] = "\xd7\xd2",
  [0x3ea7] = "\xb3\x68", [0x3ea8] = "\xb3\x66", [0x3ea9] = "\xb3\x63",
  [0x3eaa] = "\xb3\x67", [0x3eab] = "\xb3\x65", [0x3eac] = "\xb3\x64",
  [0x3ead] = "\xa0\xcb", [0x3eaf] = "\xb6\x4a", [0x3eb0] = "\xdb\xea",
  [0x3eb2] = "\xb8\xed", [0x3eb3] = "\xb6\x4c", [0x3eb4] = "\xb6\x51",
  [0x3eb5] = "\xdb\xec", [0x3eb6] = "\xb6\x53", [0x3eb7] = "\xb6\x52",
  [0x3eb8] = "\xb6\x55", [0x3eb9] = "\xdb\xeb", [0x3eba] = "\xdb\xe8",
  [0x3ebb] = "\xb6\x4f", [0x3ebc] = "\xb6\x4b", [0x3ebd] = "\xb6\x4d",
  [0x3ebe] = "\xdb\xe9", [0x3ebf] = "\xb6\x54", [0x3ec0] = "\xb6\x50",
  [0x3ec1] = "\xb6\x4e", [0x3ec2] = "\xb8\xef", [0x3ec3] = "\xb8\xee",
  [0x3ec4] = "\xb8\xec", [0x3ec5] = "\xb8\xf0", [0x3ec7] = "\xb8\xea",
  [0x3ec8] = "\xb8\xeb", [0x3eca] = "\xb8\xe9", [0x3ecc] = "\xe0\x5b",
  [0x3ecd] = "\x9e\x48", [0x3ecf] = "\xe4\x54", [0x3ed1] = "\xbb\xac",
  [0x3ed2] = "\xbb\xad", [0x3ed3] = "\xbb\xab", [0x3ed4] = "\x99\xdb",
  [0x3ed5] = "\xe4\x53", [0x3ed6] = "\x8f\xf3", [0x3ed7] = "\xe4\x55",
  [0x3ed9] = "\xe7\xea", [0x3eda] = "\xe7\xec", [0x3edb] = "\x8f\xf9",
  [0x3edc] = "\xbd\xe7", [0x3edd] = "\xe7\xed", [0x3ede] = "\xbd\xe0",
  [0x3edf] = "\xe7\xe9", [0x3ee0] = "\xbd\xdf", [0x3ee1] = "\xbd\xe9",
  [0x3ee2] = "\xbd\xe5", [0x3ee3] = "\xbd\xe6", [0x3ee4] = "\xbd\xe2",
  [0x3ee5] = "\xe7\xe8", [0x3ee6] = "\xbd\xe1", [0x3ee7] = "\xe7\xee",
  [0x3ee8] = "\xe7\xeb", [0x3ee9] = "\x95\xc1", [0x3eea] = "\xbd\xe8",
  [0x3eeb] = "\xa0\x4e", [0x3eec] = "\xbd\xe3", [0x3eed] = "\xbd\xe4",
  [0x3eee] = "\xeb\xb5", [0x3ef0] = "\xeb\xb7", [0x3ef1] = "\xeb\xb6",
  [0x3ef2] = "\x99\xdc", [0x3ef3] = "\xeb\xb8", [0x3ef4] = "\xbf\xe0",
  [0x3ef5] = "\xeb\xb4", [0x3ef7] = "\xa0\x64", [0x3ef8] = "\xc1\xcb",
  [0x3ef9] = "\xee\xb8", [0x3efa] = "\xc1\xc8", [0x3efb] = "\xc1\xcc",
  [0x3efc] = "\xc1\xca", [0x3efd] = "\xc1\xc9", [0x3efe] = "\xf0\xf3",
  [0x3f00] = "\xf0\xf6", [0x3f02] = "\xf0\xf5", [0x3f03] = "\x8f\xf7",
  [0x3f04] = "\xf0\xf4", [0x3f05] = "\xc2\xd8", [0x3f06] = "\xf3\x48",
  [0x3f07] = "\xf3\x49", [0x3f08] = "\xc3\xd8", [0x3f09] = "\xf3\x4a",
  [0x3f0a] = "\xc3\xd9", [0x3f0b] = "\x89\xb0", [0x3f0c] = "\xa0\x48",
  [0x3f0d] = "\xc4\xba", [0x3f0f] = "\xc4\xb9", [0x3f10] = "\xf6\x52",
  [0x3f11] = "\x8f\xfb", [0x3f12] = "\x8f\xf6", [0x3f13] = "\xc5\x42",
  [0x3f14] = "\xf6\x53", [0x3f15] = "\xf7\x5c", [0x3f16] = "\xc5\xab",
  [0x3f17] = "\xc5\xac", [0x3f18] = "\x9d\xdc", [0x3f19] = "\xf8\x45",
  [0x3f1b] = "\xc6\x42", [0x3f1c] = "\x99\xdd", [0x3f1d] = "\x8b\xe8",
  [0x3f64] = "\xa8\xaa", [0x3f66] = "\xb3\x6a", [0x3f67] = "\xb3\x69",
  [0x3f68] = "\xe0\x5c", [0x3f69] = "\xe0\x5d", [0x3f6b] = "\xbb\xae",
  [0x3f6c] = "\xeb\xb9", [0x3f6d] = "\xbd\xea", [0x3f6e] = "\xeb\xba",
  [0x3f6f] = "\xee\xb9", [0x3f70] = "\xa8\xab", [0x3f72] = "\xd0\xb2",
  [0x3f73] = "\xad\x76", [0x3f74] = "\xad\x75", [0x3f76] = "\xd3\xe3",
  [0x3f77] = "\xb0\x5f", [0x3f78] = "\xd3\xe4", [0x3f79] = "\xd7\xd5",
  [0x3f7a] = "\x92\xc1", [0x3f7b] = "\xd7\xd4", [0x3f7d] = "\xd7\xd3",
  [0x3f80] = "\xdb\xee", [0x3f81] = "\xb6\x58", [0x3f82] = "\x9f\xd6",
  [0x3f84] = "\xdb\xed", [0x3f85] = "\xb6\x57", [0x3f89] = "\xdb\xef",
  [0x3f8a] = "\xb6\x56", [0x3f8c] = "\xe0\x5f", [0x3f8d] = "\xe0\x62",
  [0x3f8e] = "\xe0\x60", [0x3f8f] = "\xe0\x61", [0x3f90] = "\xe0\x65",
  [0x3f91] = "\xe0\x5e", [0x3f92] = "\xe0\x66", [0x3f93] = "\xe0\x63",
  [0x3f94] = "\xe0\x64", [0x3f95] = "\xbb\xb0", [0x3f96] = "\xe4\x56",
  [0x3f99] = "\xbb\xaf", [0x3f9b] = "\xe7\xf2", [0x3f9c] = "\xe7\xf0",
  [0x3f9f] = "\xbd\xeb", [0x3fa0] = "\xe7\xef", [0x3fa1] = "\xe7\xf1",
  [0x3fa3] = "\xbd\xec", [0x3fa5] = "\xeb\xbb", [0x3fa6] = "\xa0\xd2",
  [0x3fa7] = "\xeb\xbc", [0x3fa8] = "\xc1\xcd", [0x3fa9] = "\x90\x40",
  [0x3faa] = "\xf3\x4c", [0x3fab] = "\xf3\x4e", [0x3fac] = "\xf3\x4b",
  [0x3fad] = "\xf3\x4d", [0x3fae] = "\xf4\xd6", [0x3faf] = "\xf6\x54",
  [0x3fb2] = "\xf9\x6f", [0x3fb3] = "\xa8\xac", [0x3fb4] = "\xad\x77",
  [0x3fb5] = "\xd3\xe5", [0x3fb6] = "\xd3\xe7", [0x3fb7] = "\xd3\xe6",
  [0x3fb9] = "\xd7\xd8", [0x3fba] = "\xb3\x6c", [0x3fbc] = "\xd7\xd6",
  [0x3fbe] = "\xb3\x6b", [0x3fbf] = "\xd7\xd9", [0x3fc0] = "\x8a\xc4",
  [0x3fc1] = "\xd7\xda", [0x3fc2] = "\xd7\xd7", [0x3fc3] = "\x99\xe0",
  [0x3fc5] = "\xdb\xfb", [0x3fc6] = "\xb6\x60", [0x3fc7] = "\xdb\xf3",
  [0x3fc8] = "\xdb\xf9", [0x3fcb] = "\xb6\x5b", [0x3fcc] = "\xb6\x5e",
  [0x3fcd] = "\xdb\xf2", [0x3fce] = "\xb6\x59", [0x3fcf] = "\xdb\xf6",
  [0x3fd0] = "\xe0\x6c", [0x3fd1] = "\xb6\x5d", [0x3fd3] = "\xdb\xf1",
  [0x3fd4] = "\x9f\xf0", [0x3fd5] = "\xdb\xf7", [0x3fd6] = "\xdb\xf4",
  [0x3fd7] = "\xdb\xfa", [0x3fd8] = "\xdb\xf0", [0x3fd9] = "\xdb\xf8",
  [0x3fda] = "\xb6\x5c", [0x3fdb] = "\xb6\x5f", [0x3fdc] = "\xdb\xf5",
  [0x3fdd] = "\xb6\x5a", [0x3fdf] = "\xb8\xf2", [0x3fe0] = "\xe0\x68",
  [0x3fe1] = "\xb8\xf1", [0x3fe2] = "\xe0\x6f", [0x3fe3] = "\xe0\x6e",
  [0x3fe4] = "\xb8\xf8", [0x3fe6] = "\xb8\xf9", [0x3fe7] = "\xe0\x70",
  [0x3fe8] = "\xb8\xf3", [0x3fe9] = "\xe0\x6d", [0x3fea] = "\xb8\xf7",
  [0x3feb] = "\xe0\x72", [0x3fec] = "\xe0\x69", [0x3fee] = "\xe0\x6b",
  [0x3fef] = "\xb8\xf4", [0x3ff0] = "\xe0\x67", [0x3ff1] = "\xe0\x6a",
  [0x3ff2] = "\xe0\x71", [0x3ff3] = "\xb8\xf5", [0x3ff4] = "\xe0\x73",
  [0x3ffa] = "\xb8\xf6", [0x3ffc] = "\xbb\xb1", [0x3ffd] = "\xe4\x5b",
  [0x3ffe] = "\xe4\x61", [0x3fff] = "\xe4\x59", [0x4000] = "\xe4\x62",
  [0x4001] = "\x9f\xf3", [0x4002] = "\xe4\x58", [0x4003] = "\xe4\x5d",
  [0x4004] = "\xe4\x63", [0x4005] = "\xe4\x60", [0x4006] = "\xe4\x5f",
  [0x4007] = "\xe4\x5e", [0x4009] = "\xe4\x57", [0x400a] = "\xe4\x5c",
  [0x400d] = "\xe4\x5a", [0x400e] = "\x9d\xbf", [0x400f] = "\xbd\xf1",
  [0x4010] = "\xbd\xee", [0x4011] = "\xe7\xfb", [0x4012] = "\xe8\x41",
  [0x4013] = "\xe8\x43", [0x4014] = "\xe8\x40", [0x4015] = "\xe7\xf8",
  [0x4016] = "\xe7\xfa", [0x4017] = "\xe8\x45", [0x4018] = "\xe8\x42",
  [0x4019] = "\xe7\xfc", [0x401a] = "\xe8\x46", [0x401b] = "\xe7\xf9",
  [0x401c] = "\xe8\x44", [0x401d] = "\xbd\xef", [0x401e] = "\xbd\xf5",
  [0x401f] = "\xbd\xf3", [0x4020] = "\xe7\xf3", [0x4021] = "\xbd\xf4",
  [0x4022] = "\xbd\xf0", [0x4023] = "\xe7\xf4", [0x4024] = "\xe7\xf6",
  [0x4025] = "\xe7\xf5", [0x4026] = "\xe7\xfd", [0x4027] = "\xe7\xfe",
  [0x4028] = "\x9f\xf6", [0x4029] = "\xbd\xf2", [0x402a] = "\x95\xc8",
  [0x402b] = "\xbd\xed", [0x402d] = "\x9e\x5a", [0x402e] = "\xe7\xf7",
  [0x4030] = "\xeb\xc6", [0x4031] = "\xbf\xe2", [0x4033] = "\xeb\xbd",
  [0x4034] = "\xbf\xe3", [0x4035] = "\xbf\xe6", [0x4036] = "\xeb\xc2",
  [0x4038] = "\xeb\xbf", [0x4039] = "\xbf\xe5", [0x403a] = "\x99\xe3",
  [0x403c] = "\xeb\xc3", [0x403d] = "\xeb\xc4", [0x403e] = "\xeb\xbe",
  [0x403f] = "\xeb\xc7", [0x4040] = "\xeb\xc0", [0x4041] = "\xeb\xc5",
  [0x4042] = "\xbf\xe4", [0x4044] = "\xbf\xe1", [0x4045] = "\xeb\xc1",
  [0x4046] = "\x8a\x4a", [0x4047] = "\xee\xbf", [0x4048] = "\xc1\xd0",
  [0x4049] = "\xc1\xce", [0x404a] = "\xc1\xd1", [0x404b] = "\xc1\xcf",
  [0x404c] = "\xee\xbe", [0x404d] = "\xee\xbb", [0x404e] = "\xee\xba",
  [0x404f] = "\x9f\xf1", [0x4050] = "\xee\xbd", [0x4053] = "\xee\xbc",
  [0x4054] = "\xf1\x45", [0x4055] = "\xc2\xde", [0x4056] = "\xf0\xfb",
  [0x4057] = "\xf0\xfa", [0x4059] = "\xc2\xd9", [0x405a] = "\xf1\x41",
  [0x405b] = "\xf1\x40", [0x405c] = "\xf0\xf7", [0x405d] = "\xf1\x43",
  [0x405e] = "\xf0\xfc", [0x405f] = "\xc2\xdd", [0x4060] = "\xf0\xf9",
  [0x4061] = "\xf1\x42", [0x4062] = "\xf0\xf8", [0x4063] = "\xc2\xda",
  [0x4064] = "\xc2\xdc", [0x4065] = "\xf0\xfd", [0x4066] = "\xc2\xdb",
  [0x4067] = "\xf0\xfe", [0x4068] = "\x8a\xa7", [0x4069] = "\xf1\x44",
  [0x406a] = "\xf3\x52", [0x406c] = "\xc3\xde", [0x406d] = "\xf3\x4f",
  [0x406f] = "\xf3\x53", [0x4071] = "\x99\xe6", [0x4072] = "\xc3\xdb",
  [0x4073] = "\xf3\x51", [0x4074] = "\xc3\xe0", [0x4075] = "\x9f\xf7",
  [0x4076] = "\xc3\xdd", [0x4077] = "\x9f\xed", [0x4078] = "\xf3\x50",
  [0x407a] = "\xc3\xdf", [0x407b] = "\xf3\x54", [0x407c] = "\xc3\xda",
  [0x407e] = "\x8a\x5c", [0x4080] = "\x9d\xae", [0x4081] = "\xc4\xbc",
  [0x4082] = "\xc4\xbe", [0x4084] = "\xf4\xd9", [0x4085] = "\xc4\xbd",
  [0x4086] = "\xf4\xd7", [0x4087] = "\xc3\xdc", [0x4088] = "\xf4\xd8",
  [0x4089] = "\xc4\xbb", [0x408a] = "\xc5\x43", [0x408b] = "\xc5\x45",
  [0x408c] = "\xf6\x56", [0x408d] = "\xc5\x44", [0x408e] = "\xf6\x55",
  [0x4090] = "\xf7\x61", [0x4091] = "\xc5\xad", [0x4092] = "\xf7\x60",
  [0x4093] = "\xc5\xae", [0x4094] = "\xf7\x5e", [0x4095] = "\xf7\x5d",
  [0x4096] = "\xf7\x62", [0x4097] = "\xf7\x63", [0x4098] = "\xf8\x46",
  [0x409a] = "\xf7\x5f", [0x409d] = "\xf8\xc6", [0x409e] = "\xf8\xc3",
  [0x409f] = "\xf8\xc4", [0x40a0] = "\xf8\xc5", [0x40a1] = "\xc6\x5c",
  [0x40a3] = "\xf9\x51", [0x40a4] = "\xf9\x50", [0x40a5] = "\xf9\x4f",
  [0x40a6] = "\xf9\x70", [0x40a7] = "\x95\xc9", [0x40a8] = "\xf9\xbe",
  [0x40a9] = "\xf9\xab", [0x40aa] = "\xc6\x6e", [0x40ab] = "\xa8\xad",
  [0x40ac] = "\xb0\x60", [0x40ad] = "\x90\x48", [0x40b0] = "\x99\xe8",
  [0x40b2] = "\xb8\xfa", [0x40b6] = "\x90\x49", [0x40ba] = "\xbd\xf6",
  [0x40bc] = "\x90\xb1", [0x40bd] = "\xeb\xc8", [0x40c0] = "\xc2\xdf",
  [0x40c2] = "\xf3\x55", [0x40c3] = "\x90\x4a", [0x40c9] = "\xf9\xac",
  [0x40ca] = "\xa8\xae", [0x40cb] = "\xaa\xee", [0x40cc] = "\xad\x79",
  [0x40cd] = "\xad\x78", [0x40ce] = "\x99\xea", [0x40cf] = "\xb0\x63",
  [0x40d1] = "\xd3\xe8", [0x40d2] = "\xb0\x61", [0x40d3] = "\xd3\xe9",
  [0x40d4] = "\xb0\x62", [0x40d7] = "\xd7\xdf", [0x40d8] = "\xd7\xdb",
  [0x40da] = "\x9b\xd1", [0x40db] = "\xb3\x6d", [0x40dc] = "\xd7\xde",
  [0x40dd] = "\xd7\xdd", [0x40de] = "\xd7\xdc", [0x40df] = "\xb3\x6e",
  [0x40e0] = "\xd7\xe0", [0x40e1] = "\xd7\xe1", [0x40e2] = "\x99\xeb",
  [0x40e4] = "\x99\xec", [0x40e5] = "\xdc\x43", [0x40e6] = "\xdc\x41",
  [0x40e7] = "\xdc\x45", [0x40e8] = "\xdc\x46", [0x40e9] = "\xdc\x4c",
  [0x40eb] = "\xdc\x48", [0x40ec] = "\xdc\x4a", [0x40ed] = "\x99\xed",
  [0x40ee] = "\xdc\x42", [0x40ef] = "\xdb\xfc", [0x40f1] = "\xdc\x49",
  [0x40f2] = "\x99\xee", [0x40f4] = "\xdc\x4b", [0x40f5] = "\xdc\x44",
  [0x40f6] = "\xdc\x47", [0x40f7] = "\xdb\xfd", [0x40f8] = "\xb6\x62",
  [0x40f9] = "\xdc\x40", [0x40fa] = "\xdb\xfe", [0x40fb] = "\xb6\x61",
  [0x40fc] = "\xb6\x63", [0x40fe] = "\xb8\xfd", [0x40ff] = "\xe0\x75",
  [0x4100] = "\xe0\x77", [0x4101] = "\xe0\x76", [0x4102] = "\xe0\x7b",
  [0x4103] = "\xb8\xfb", [0x4105] = "\xe0\x78", [0x4106] = "\xe0\x74",
  [0x4107] = "\xe0\x79", [0x4108] = "\xe0\x7a", [0x4109] = "\xb8\xfc",
  [0x410a] = "\xb8\xfe", [0x410b] = "\xe0\x7c", [0x410d] = "\xe4\x67",
  [0x410e] = "\xe4\x66", [0x4110] = "\xe4\x64", [0x4111] = "\xe4\x65",
  [0x4112] = "\xbb\xb3", [0x4113] = "\xbb\xb5", [0x4114] = "\xbb\xb2",
  [0x4115] = "\xbb\xb4", [0x4116] = "\xe8\x4d", [0x4117] = "\xe8\x4e",
  [0x4118] = "\xe8\x49", [0x4119] = "\x90\x4c", [0x411a] = "\xe8\x4a",
  [0x411b] = "\xbd\xf8", [0x411c] = "\xbd\xfd", [0x411d] = "\xbd\xf7",
  [0x411e] = "\xbd\xfe", [0x411f] = "\xbd\xf9", [0x4120] = "\xe8\x4b",
  [0x4123] = "\xe8\x4c", [0x4124] = "\xe8\x48", [0x4125] = "\xbe\x40",
  [0x4126] = "\xbd\xfb", [0x4129] = "\xbd\xfa", [0x412a] = "\xbd\xfc",
  [0x412c] = "\xe8\x47", [0x412d] = "\x90\x4d", [0x412e] = "\xeb\xca",
  [0x412f] = "\xbf\xe8", [0x4130] = "\x95\xcb", [0x4132] = "\xeb\xcc",
  [0x4133] = "\xbf\xea", [0x4134] = "\xeb\xcf", [0x4135] = "\xeb\xcb",
  [0x4136] = "\xeb\xc9", [0x4137] = "\xeb\xce", [0x4138] = "\xbf\xe9",
  [0x4139] = "\xeb\xcd", [0x413b] = "\xbf\xe7", [0x413e] = "\xc1\xd3",
  [0x413f] = "\xc1\xd6", [0x4140] = "\xee\xc1", [0x4141] = "\x97\xe2",
  [0x4142] = "\xc1\xd4", [0x4143] = "\xee\xc0", [0x4144] = "\xc1\xd2",
  [0x4145] = "\xc1\xd5", [0x4146] = "\xf1\x46", [0x4147] = "\xf1\x47",
  [0x4148] = "\xf1\x48", [0x4149] = "\xc2\xe0", [0x414a] = "\x95\xcc",
  [0x414b] = "\xf1\x49", [0x414d] = "\xc2\xe1", [0x414e] = "\xc3\xe2",
  [0x414f] = "\xf3\x58", [0x4150] = "\xf3\x59", [0x4151] = "\xf3\x57",
  [0x4152] = "\xf3\x56", [0x4153] = "\xf3\x5a", [0x4154] = "\xc3\xe1",
  [0x4155] = "\xf4\xdd", [0x4156] = "\xf4\xdb", [0x4157] = "\xf4\xdc",
  [0x4158] = "\xf4\xde", [0x4159] = "\xf4\xda", [0x415a] = "\xf4\xdf",
  [0x415b] = "\xf6\x58", [0x415c] = "\x9f\x78", [0x415d] = "\xf6\x59",
  [0x415e] = "\xf6\x57", [0x415f] = "\xc5\x46", [0x4160] = "\xf7\x64",
  [0x4161] = "\xc5\xaf", [0x4162] = "\xf7\x65", [0x4163] = "\xf8\x48",
  [0x4164] = "\xf8\x47", [0x4166] = "\x89\x7c", [0x4167] = "\x89\x7d",
  [0x416e] = "\x89\x7e", [0x4193] = "\x99\x5d", [0x419b] = "\xa8\xaf",
  [0x419c] = "\xb6\x64", [0x419f] = "\xb9\x40", [0x41a0] = "\x9b\x5a",
  [0x41a3] = "\xbb\xb6", [0x41a5] = "\x90\x50", [0x41a6] = "\xbf\xec",
  [0x41a8] = "\xbf\xeb", [0x41ad] = "\xc3\xe3", [0x41ae] = "\xc4\x7c",
  [0x41af] = "\xc5\x47", [0x41b0] = "\xa8\xb0", [0x41b1] = "\xb0\x64",
  [0x41b2] = "\xb9\x41", [0x41b3] = "\x90\x54", [0x41b4] = "\xf3\x5b",
  [0x41b5] = "\xc6\xd6", [0x41b6] = "\x9a\xa8", [0x41b7] = "\x99\xef",
  [0x41b8] = "\xfe\xeb", [0x41b9] = "\x9d\xa3", [0x41ba] = "\x9d\xa1",
  [0x41bb] = "\x99\x43", [0x41bc] = "\x99\x45", [0x41be] = "\x9d\x7d",
  [0x41bf] = "\xcb\xa6", [0x41c1] = "\x99\xf0", [0x41c2] = "\xa8\xb1",
  [0x41c4] = "\xa8\xb4", [0x41c5] = "\xa8\xb3", [0x41c6] = "\xa8\xb2",
  [0x41c9] = "\xcb\xa5", [0x41ca] = "\x99\xf1", [0x41cb] = "\xcd\xcd",
  [0x41cc] = "\x99\xf2", [0x41cd] = "\xcd\xcf", [0x41ce] = "\xaa\xef",
  [0x41d0] = "\x9d\x60", [0x41d1] = "\xaa\xf1", [0x41d2] = "\xcd\xcc",
  [0x41d3] = "\xcd\xce", [0x41d4] = "\xaa\xf0", [0x41d5] = "\xcd\xd1",
  [0x41d6] = "\xcd\xd0", [0x41d7] = "\xcd\xd2", [0x41da] = "\xa0\xa3",
  [0x41e0] = "\xd0\xb6", [0x41e1] = "\xd0\xb4", [0x41e2] = "\xad\x7c",
  [0x41e3] = "\xd0\xb3", [0x41e4] = "\xad\xa3", [0x41e5] = "\xad\x7e",
  [0x41e6] = "\xad\x7b", [0x41e8] = "\xad\xa4", [0x41ea] = "\xad\x7d",
  [0x41eb] = "\xad\xa2", [0x41ed] = "\xad\xa1", [0x41ee] = "\xd0\xb5",
  [0x41f0] = "\xad\x7a", [0x41f4] = "\xb0\x6a", [0x41f5] = "\xd3\xeb",
  [0x41f6] = "\xd3\xf1", [0x41f7] = "\xb0\x67", [0x41f8] = "\xb0\x6e",
  [0x41f9] = "\x90\x5b", [0x41fa] = "\xb0\x69", [0x41fb] = "\xd3\xee",
  [0x41fc] = "\xd3\xf0", [0x41fd] = "\xb0\x6c", [0x41fe] = "\xd3\xea",
  [0x41ff] = "\xd3\xed", [0x4200] = "\xb0\x68", [0x4201] = "\xb0\x65",
  [0x4202] = "\xd3\xec", [0x4203] = "\xb0\x6b", [0x4204] = "\xd3\xef",
  [0x4205] = "\xb0\x6d", [0x4206] = "\xb0\x66", [0x4208] = "\x9e\xdb",
  [0x420b] = "\xd7\xe3", [0x420c] = "\xd7\xe6", [0x420d] = "\xb3\x70",
  [0x420f] = "\xb3\x7a", [0x4210] = "\xb3\x76", [0x4211] = "\xd7\xe4",
  [0x4212] = "\x9d\x79", [0x4214] = "\xb3\x7e", [0x4215] = "\xb3\x77",
  [0x4216] = "\xb3\x7c", [0x4217] = "\xb3\x72", [0x4219] = "\xb3\x6f",
  [0x421a] = "\xb3\x71", [0x421b] = "\xb3\x7d", [0x421c] = "\xd7\xe5",
  [0x421d] = "\xb3\x75", [0x421e] = "\xb3\x78", [0x421f] = "\xb3\x74",
  [0x4220] = "\xb3\x79", [0x4221] = "\xd7\xe7", [0x4222] = "\xb3\x7b",
  [0x4223] = "\xb3\x73", [0x4224] = "\xd7\xe2", [0x422d] = "\xdc\x4d",
  [0x422e] = "\xb6\x65", [0x422f] = "\xdc\x4f", [0x4231] = "\xb6\x67",
  [0x4232] = "\xb6\x69", [0x4233] = "\x99\xf3", [0x4234] = "\xdc\x4e",
  [0x4235] = "\xb6\x66", [0x4236] = "\xb6\x6a", [0x4237] = "\x90\x62",
  [0x4238] = "\xb6\x68", [0x423c] = "\xb9\x47", [0x423d] = "\xe0\xa3",
  [0x423e] = "\xb9\x4f", [0x423f] = "\xe0\x7e", [0x4241] = "\xb9\x50",
  [0x4242] = "\xb9\x45", [0x4244] = "\xe0\xa1", [0x4247] = "\xb9\x4a",
  [0x4249] = "\xe0\xa2", [0x424a] = "\xb9\x43", [0x424b] = "\xb9\x42",
  [0x424c] = "\x9f\x55", [0x424d] = "\xb9\x4d", [0x424e] = "\xb9\x4c",
  [0x424f] = "\xb9\x4b", [0x4250] = "\xb9\x49", [0x4251] = "\xb9\x4e",
  [0x4252] = "\xe0\x7d", [0x4253] = "\xb9\x44", [0x4254] = "\xb9\x46",
  [0x4255] = "\xb9\x48", [0x4256] = "\x9b\xf9", [0x4258] = "\xbb\xb8",
  [0x4259] = "\xbb\xbb", [0x425b] = "\xbb\xbf", [0x425c] = "\xbb\xb9",
  [0x425d] = "\xbb\xbe", [0x425e] = "\xbb\xbc", [0x4260] = "\xbb\xb7",
  [0x4261] = "\x90\x65", [0x4262] = "\xbb\xbd", [0x4263] = "\xbb\xba",
  [0x4264] = "\x96\xe0", [0x4267] = "\xe8\x52", [0x4268] = "\xbe\x43",
  [0x4269] = "\xbe\x41", [0x426b] = "\xe8\x53", [0x426c] = "\x98\xbe",
  [0x426d] = "\xbe\x44", [0x426e] = "\xbe\x42", [0x426f] = "\xe8\x51",
  [0x4270] = "\xe8\x50", [0x4272] = "\xbf\xf0", [0x4273] = "\xe8\x4f",
  [0x4274] = "\xbf\xee", [0x4275] = "\xbf\xed", [0x4276] = "\xeb\xd0",
  [0x4277] = "\xbe\x45", [0x4278] = "\xbf\xef", [0x4279] = "\xeb\xd1",
  [0x427a] = "\xbf\xf2", [0x427b] = "\xeb\xd2", [0x427c] = "\xbf\xf1",
  [0x427d] = "\xc1\xd8", [0x427e] = "\xee\xc3", [0x427f] = "\xc1\xd7",
  [0x4280] = "\xc1\xdc", [0x4281] = "\xc1\xda", [0x4282] = "\xc1\xdb",
  [0x4283] = "\xc2\xe3", [0x4284] = "\xc1\xd9", [0x4285] = "\xee\xc2",
  [0x4286] = "\xeb\xd3", [0x4287] = "\xc2\xe2", [0x4288] = "\xc2\xe4",
  [0x428a] = "\xc3\xe4", [0x428b] = "\xc3\xe5", [0x428d] = "\xf4\xe0",
  [0x428f] = "\xc5\xde", [0x4290] = "\xc5\xdd", [0x4291] = "\xa8\xb6",
  [0x4294] = "\xca\x55", [0x4295] = "\xb0\x6f", [0x4297] = "\xca\x52",
  [0x4298] = "\xca\x53", [0x4299] = "\xca\x51", [0x429b] = "\xca\x54",
  [0x429e] = "\xcb\xaa", [0x429f] = "\xcb\xa7", [0x42a0] = "\xcb\xac",
  [0x42a1] = "\xcb\xa8", [0x42a2] = "\xa8\xb7", [0x42a3] = "\xa8\xba",
  [0x42a5] = "\xcb\xa9", [0x42a6] = "\xa8\xb9", [0x42a7] = "\xcb\xab",
  [0x42a8] = "\x90\x68", [0x42aa] = "\xa8\xb8", [0x42ae] = "\x90\x6c",
  [0x42af] = "\xcd\xd5", [0x42b0] = "\xcd\xd7", [0x42b1] = "\xaa\xf4",
  [0x42b2] = "\xcd\xd3", [0x42b3] = "\xcd\xd6", [0x42b4] = "\xcd\xd4",
  [0x42b5] = "\xaa\xf2", [0x42b6] = "\xaa\xf5", [0x42b8] = "\xaa\xf3",
  [0x42bb] = "\x95\xd8", [0x42bd] = "\xd0\xb8", [0x42be] = "\xd0\xbc",
  [0x42bf] = "\xd0\xb9", [0x42c1] = "\xad\xa7", [0x42c3] = "\xad\xa8",
  [0x42c4] = "\x90\x6a", [0x42c5] = "\xd0\xbb", [0x42c7] = "\xd0\xbd",
  [0x42c8] = "\xd0\xbf", [0x42ca] = "\xad\xa5", [0x42cb] = "\xd0\xbe",
  [0x42ce] = "\xad\xa6", [0x42d4] = "\xd7\xee", [0x42d5] = "\xd0\xba",
  [0x42d6] = "\xd3\xf2", [0x42d7] = "\xd3\xfb", [0x42d8] = "\xd3\xf9",
  [0x42d9] = "\xd3\xf4", [0x42da] = "\xd3\xf5", [0x42db] = "\xd3\xfa",
  [0x42dc] = "\xd3\xfc", [0x42dd] = "\xb0\x71", [0x42df] = "\xd3\xf7",
  [0x42e0] = "\xd3\xf3", [0x42e1] = "\xb0\x70", [0x42e2] = "\xb0\x72",
  [0x42e3] = "\xd3\xf6", [0x42e4] = "\xd3\xfd", [0x42e5] = "\xd3\xf8",
  [0x42e8] = "\xb3\xa1", [0x42e9] = "\xd7\xf1", [0x42ea] = "\xd7\xe9",
  [0x42eb] = "\xd7\xef", [0x42ec] = "\xd7\xf0", [0x42ed] = "\xb3\xa2",
  [0x42ef] = "\xd7\xe8", [0x42f0] = "\xd7\xea", [0x42f1] = "\xd0\xb7",
  [0x42f2] = "\xd7\xec", [0x42f3] = "\xd7\xed", [0x42f4] = "\xd7\xeb",
  [0x42f5] = "\xb6\x6c", [0x42f9] = "\xdc\x56", [0x42fa] = "\xeb\xd4",
  [0x42fb] = "\xdc\x57", [0x42fc] = "\xdc\x54", [0x42fd] = "\xb3\xa3",
  [0x42fe] = "\xb6\x6e", [0x42ff] = "\xdc\x53", [0x4300] = "\xdc\x59",
  [0x4301] = "\xdc\x58", [0x4302] = "\xb6\x6b", [0x4303] = "\xdc\x5c",
  [0x4304] = "\xdc\x52", [0x4305] = "\xdc\x5b", [0x4306] = "\xdc\x50",
  [0x4307] = "\xdc\x5a", [0x4308] = "\xdc\x55", [0x4309] = "\xb6\x6d",
  [0x430b] = "\xe0\xaa", [0x430d] = "\xe0\xa5", [0x430e] = "\xe0\xab",
  [0x430f] = "\xe0\xa6", [0x4310] = "\xe0\xa4", [0x4311] = "\xe0\xa7",
  [0x4312] = "\xb9\x51", [0x4314] = "\xe0\xa9", [0x4316] = "\xe0\xa8",
  [0x4317] = "\xb9\x52", [0x4318] = "\xbb\xc1", [0x4319] = "\xbb\xc0",
  [0x431a] = "\xe4\x6e", [0x431b] = "\xe4\x71", [0x431c] = "\xe4\x69",
  [0x431d] = "\xe4\x6d", [0x431e] = "\xbb\xc2", [0x431f] = "\xe4\x6c",
  [0x4320] = "\xe4\x6a", [0x4321] = "\xe4\x70", [0x4322] = "\xe4\x6b",
  [0x4323] = "\xe4\x68", [0x4324] = "\xe4\x6f", [0x4326] = "\xe8\x59",
  [0x4327] = "\xbe\x48", [0x4328] = "\xf1\x4a", [0x4329] = "\xe8\x56",
  [0x432a] = "\xe8\x57", [0x432b] = "\xe8\x55", [0x432c] = "\xdc\x51",
  [0x432d] = "\xbe\x47", [0x432e] = "\xe8\x5a", [0x432f] = "\xe8\x54",
  [0x4330] = "\xbe\x46", [0x4331] = "\xbe\x49", [0x4332] = "\xe8\x58",
  [0x4333] = "\xeb\xd5", [0x4334] = "\xbf\xf3", [0x4335] = "\xeb\xd6",
  [0x4336] = "\xeb\xd7", [0x4338] = "\xee\xc4", [0x4339] = "\xc1\xdd",
  [0x433a] = "\xf1\x4b", [0x433b] = "\xf1\x4c", [0x433e] = "\xf1\x4d",
  [0x433f] = "\xf3\x5d", [0x4340] = "\xf3\x5c", [0x4341] = "\xf4\xe2",
  [0x4343] = "\xf4\xe1", [0x4344] = "\xf6\x5b", [0x4345] = "\xf6\x5c",
  [0x4346] = "\xf6\x5a", [0x4347] = "\xf7\x66", [0x4348] = "\xc5\xb0",
  [0x4349] = "\xa8\xbb", [0x434a] = "\xad\xaa", [0x434b] = "\xad\xa9",
  [0x434c] = "\xb0\x75", [0x434d] = "\xb0\x74", [0x434e] = "\xd4\x40",
  [0x434f] = "\xd4\x41", [0x4350] = "\xd3\xfe", [0x4351] = "\x9f\xb2",
  [0x4352] = "\xb0\x73", [0x4353] = "\xd7\xf5", [0x4355] = "\xd7\xf6",
  [0x4356] = "\xd7\xf2", [0x4357] = "\xb3\xa4", [0x4358] = "\xd7\xf3",
  [0x4359] = "\x9f\xae", [0x435a] = "\xd7\xf4", [0x435c] = "\x9f\xb0",
  [0x435e] = "\x89\xad", [0x435f] = "\xdc\x5f", [0x4360] = "\xdc\x61",
  [0x4361] = "\xdc\x5d", [0x4362] = "\xdc\x60", [0x4363] = "\xb6\x6f",
  [0x4364] = "\xdc\x5e", [0x4365] = "\xb6\x70", [0x4367] = "\x90\x6e",
  [0x4368] = "\xdd\x73", [0x4369] = "\xb9\x55", [0x436a] = "\xb9\x54",
  [0x436c] = "\xb9\x53", [0x436e] = "\xe0\xac", [0x436f] = "\xe0\xad",
  [0x4370] = "\x9e\x71", [0x4372] = "\xe4\x73", [0x4373] = "\xe4\x75",
  [0x4374] = "\xbb\xc6", [0x4375] = "\xbb\xc3", [0x4376] = "\x9e\x4a",
  [0x4377] = "\xbb\xc5", [0x4378] = "\xbb\xc4", [0x4379] = "\xe4\x74",
  [0x437a] = "\xe4\x72", [0x437c] = "\x9f\xdc", [0x4380] = "\xe8\x61",
  [0x4381] = "\xe8\x5e", [0x4382] = "\xe8\x5f", [0x4383] = "\xbe\x4d",
  [0x4384] = "\xe8\x60", [0x4385] = "\xe8\x5b", [0x4386] = "\xe8\x5c",
  [0x4387] = "\xbe\x4a", [0x4389] = "\xbe\x4b", [0x438a] = "\xe8\x5d",
  [0x438b] = "\xbe\x4c", [0x438c] = "\x89\xab", [0x438d] = "\xeb\xdb",
  [0x438e] = "\x9f\xb8", [0x438f] = "\xeb\xdc", [0x4390] = "\xeb\xd9",
  [0x4391] = "\xeb\xda", [0x4392] = "\xbf\xf4", [0x4393] = "\xeb\xd8",
  [0x4399] = "\xee\xc8", [0x439a] = "\xee\xc5", [0x439b] = "\xee\xc7",
  [0x439c] = "\xc1\xe0", [0x439d] = "\xee\xcb", [0x439e] = "\xc1\xdf",
  [0x439f] = "\xee\xc9", [0x43a0] = "\xee\xcc", [0x43a1] = "\xee\xca",
  [0x43a2] = "\xee\xc6", [0x43a3] = "\xc1\xde", [0x43a5] = "\xf1\x4f",
  [0x43a7] = "\xf1\x50", [0x43a8] = "\xf1\x4e", [0x43a9] = "\x90\x70",
  [0x43aa] = "\xf1\x52", [0x43ab] = "\xc2\xe5", [0x43ac] = "\xc2\xe6",
  [0x43ad] = "\xf3\x5f", [0x43ae] = "\xc3\xe7", [0x43af] = "\xf1\x51",
  [0x43b0] = "\xf3\x5e", [0x43b1] = "\xc3\xe6", [0x43b2] = "\xf4\xe5",
  [0x43b3] = "\xf4\xe6", [0x43b4] = "\xc4\xbf", [0x43b5] = "\xf4\xe4",
  [0x43b6] = "\x8b\x63", [0x43b7] = "\xf4\xe3", [0x43b9] = "\xf6\x5d",
  [0x43ba] = "\xc5\x48", [0x43bb] = "\x95\xdc", [0x43bc] = "\xf8\x49",
  [0x43bd] = "\xf8\xc8", [0x43be] = "\xf8\xc7", [0x43c0] = "\xc6\x43",
  [0x43c1] = "\xc6\x5d", [0x43c2] = "\xf8\xc9", [0x43c3] = "\xf9\x71",
  [0x43c4] = "\x90\x71", [0x43c5] = "\xc6\x6f", [0x43c6] = "\xa8\xbc",
  [0x43c7] = "\xaa\xf6", [0x43c9] = "\xb9\x56", [0x43cb] = "\xc4\xc0",
  [0x43cc] = "\xa8\xbd", [0x43cd] = "\xad\xab", [0x43ce] = "\xb3\xa5",
  [0x43cf] = "\xb6\x71", [0x43d0] = "\xc2\xe7", [0x43d1] = "\xaa\xf7",
  [0x43d3] = "\xd0\xc1", [0x43d4] = "\xd0\xc0", [0x43d5] = "\xd4\x42",
  [0x43d6] = "\xfc\x5e", [0x43d7] = "\xb0\x78", [0x43d8] = "\xb0\x76",
  [0x43d9] = "\xb0\x7a", [0x43da] = "\xd4\x44", [0x43dc] = "\xb0\x79",
  [0x43dd] = "\xb0\x77", [0x43df] = "\x89\x49", [0x43e2] = "\xd4\x43",
  [0x43e3] = "\xb3\xa8", [0x43e4] = "\xd7\xfc", [0x43e5] = "\x96\x5b",
  [0x43e6] = "\xb3\xa7", [0x43e7] = "\xb3\xa9", [0x43e8] = "\xd8\x42",
  [0x43e9] = "\xb3\xab", [0x43ea] = "\xd7\xfe", [0x43eb] = "\xd8\x40",
  [0x43ec] = "\xd7\xf7", [0x43ed] = "\xb3\xaa", [0x43ee] = "\xd8\x43",
  [0x43f1] = "\xd7\xf9", [0x43f3] = "\xd7\xfa", [0x43f4] = "\xd7\xf8",
  [0x43f5] = "\xb3\xa6", [0x43f7] = "\xd8\x41", [0x43f8] = "\xd7\xfb",
  [0x43f9] = "\xd7\xfd", [0x43fa] = "\x94\xa6", [0x43fd] = "\xdc\x6d",
  [0x43fe] = "\x8f\xd5", [0x43ff] = "\xdc\x6c", [0x4400] = "\xdc\x6a",
  [0x4401] = "\xdc\x62", [0x4402] = "\xdc\x71", [0x4403] = "\xdc\x65",
  [0x4404] = "\xdc\x6f", [0x4405] = "\xdc\x76", [0x4406] = "\xdc\x6e",
  [0x4407] = "\xb6\x79", [0x4408] = "\x9e\x73", [0x4409] = "\xb6\x75",
  [0x440a] = "\xdc\x63", [0x440c] = "\xdc\x69", [0x440d] = "\xb6\x77",
  [0x440e] = "\x90\x75", [0x440f] = "\xdc\x68", [0x4410] = "\xb6\x78",
  [0x4411] = "\xb6\x7a", [0x4412] = "\xdc\x6b", [0x4413] = "\x99\xf7",
  [0x4414] = "\xb6\x72", [0x4415] = "\xb6\x73", [0x4416] = "\xdc\x77",
  [0x4417] = "\xdc\x75", [0x4419] = "\xdc\x74", [0x441a] = "\xdc\x66",
  [0x441c] = "\xdc\x72", [0x441e] = "\xb6\x76", [0x4423] = "\xb6\x74",
  [0x4424] = "\xdc\x73", [0x4425] = "\xdc\x64", [0x4426] = "\xdc\x67",
  [0x4427] = "\xdc\x70", [0x4428] = "\x99\xf9", [0x442a] = "\x96\x63",
  [0x442b] = "\x95\xb9", [0x442d] = "\xe4\xba", [0x442e] = "\xe0\xb7",
  [0x4430] = "\xe0\xb0", [0x4431] = "\xe0\xc3", [0x4432] = "\xe0\xcc",
  [0x4433] = "\xe0\xb3", [0x4434] = "\xb9\x61", [0x4435] = "\x94\xd4",
  [0x4436] = "\xe0\xc0", [0x4437] = "\xb9\x57", [0x4438] = "\xb9\x59",
  [0x4439] = "\xb9\x65", [0x443a] = "\xe0\xb1", [0x443c] = "\xfc\xfa",
  [0x443d] = "\xb9\x5a", [0x443e] = "\xb9\x5c", [0x443f] = "\xb9\x66",
  [0x4440] = "\xb9\x5b", [0x4441] = "\x90\x77", [0x4444] = "\x90\xab",
  [0x4445] = "\xb9\x64", [0x4446] = "\xe0\xb9", [0x4448] = "\xe0\xae",
  [0x4449] = "\xb9\x62", [0x444a] = "\xe0\xb8", [0x444b] = "\xb9\x5e",
  [0x444c] = "\xe0\xca", [0x444d] = "\xb9\x63", [0x444e] = "\xe0\xc8",
  [0x444f] = "\xe0\xbc", [0x4450] = "\xe0\xc6", [0x4451] = "\xb9\x60",
  [0x4452] = "\xe0\xaf", [0x4453] = "\xe0\xc9", [0x4454] = "\xe0\xc4",
  [0x4455] = "\x9d\x4d", [0x4456] = "\xe0\xcb", [0x4457] = "\xb9\x58",
  [0x4458] = "\x99\xfa", [0x445a] = "\xb9\x67", [0x445b] = "\xb9\x5d",
  [0x445d] = "\x92\xe3", [0x445e] = "\xe0\xb5", [0x445f] = "\x97\xbb",
  [0x4460] = "\xe0\xbd", [0x4461] = "\xe0\xc1", [0x4462] = "\x90\x78",
  [0x4463] = "\xe0\xc5", [0x4464] = "\xb9\x5f", [0x4465] = "\xe0\xb4",
  [0x4466] = "\xe0\xb2", [0x4467] = "\xe0\xbe", [0x446b] = "\x99\xfb",
  [0x446c] = "\xe0\xbb", [0x446d] = "\xe0\xba", [0x446e] = "\x97\xe0",
  [0x446f] = "\xe0\xbf", [0x4470] = "\xe0\xc2", [0x4472] = "\xe0\xc7",
  [0x4476] = "\xe4\x78", [0x4477] = "\x96\xdc", [0x4478] = "\xbb\xc7",
  [0x4479] = "\xe4\xa4", [0x447a] = "\xe4\x7a", [0x447b] = "\xbb\xcc",
  [0x447c] = "\xbb\xd0", [0x447d] = "\xe4\xad", [0x447e] = "\xe4\xb5",
  [0x447f] = "\xe4\xa6", [0x4480] = "\xbb\xc8", [0x4481] = "\x9c\xa8",
  [0x4482] = "\xe4\xaa", [0x4483] = "\xe0\xb6", [0x4484] = "\x97\x72",
  [0x4485] = "\xbb\xc9", [0x4486] = "\xe4\xb1", [0x4487] = "\xe4\xb6",
  [0x4488] = "\xe4\xae", [0x4489] = "\x94\x40", [0x448a] = "\xe4\xb0",
  [0x448b] = "\xe4\xb9", [0x448c] = "\xe4\xb2", [0x448d] = "\xe4\x7e",
  [0x448e] = "\xe4\xa9", [0x448f] = "\x92\xf2", [0x4491] = "\xbb\xd1",
  [0x4493] = "\xbb\xcd", [0x4494] = "\xe4\x7c", [0x4495] = "\xe4\xab",
  [0x4496] = "\xbb\xcb", [0x4497] = "\xe4\xa5", [0x4498] = "\xbb\xca",
  [0x4499] = "\xe4\xb3", [0x449a] = "\xe4\xa2", [0x449b] = "\xe4\x79",
  [0x449c] = "\xbb\xce", [0x449d] = "\xe4\xb8", [0x44a0] = "\xe4\x7b",
  [0x44a1] = "\xe4\xaf", [0x44a2] = "\xe4\xac", [0x44a3] = "\xe4\xa7",
  [0x44a4] = "\xe4\x77", [0x44a5] = "\xe4\x76", [0x44a6] = "\xe4\xa1",
  [0x44a7] = "\xe4\xb4", [0x44a8] = "\xbb\xcf", [0x44a9] = "\xe4\xb7",
  [0x44aa] = "\xe4\x7d", [0x44ab] = "\xe4\xa3", [0x44ac] = "\xbe\x52",
  [0x44ae] = "\x99\xfd", [0x44b1] = "\x99\xfc", [0x44b2] = "\xbe\x5a",
  [0x44b3] = "\xbe\x55", [0x44b4] = "\xe8\xa4", [0x44b5] = "\xe8\xa1",
  [0x44b6] = "\xe8\x67", [0x44b7] = "\xbe\x50", [0x44b9] = "\xf9\xd7",
  [0x44ba] = "\x96\x4a", [0x44bb] = "\xbe\x4f", [0x44bc] = "\xbe\x56",
  [0x44be] = "\x96\xd8", [0x44bf] = "\x99\xfe", [0x44c0] = "\xe8\x65",
  [0x44c1] = "\xbe\x54", [0x44c2] = "\xe8\x71", [0x44c3] = "\xe8\x63",
  [0x44c4] = "\xe8\x64", [0x44c5] = "\xbe\x4e", [0x44c6] = "\xe8\xa3",
  [0x44c7] = "\xbe\x58", [0x44c8] = "\xe8\x74", [0x44c9] = "\xe8\x79",
  [0x44ca] = "\xe8\x73", [0x44cb] = "\xeb\xee", [0x44cc] = "\xe8\x6f",
  [0x44cd] = "\xe8\x77", [0x44ce] = "\xe8\x75", [0x44cf] = "\xe8\x68",
  [0x44d0] = "\xe8\x62", [0x44d1] = "\xe8\x7d", [0x44d2] = "\xbe\x57",
  [0x44d3] = "\xe8\x7e", [0x44d4] = "\x90\x4b", [0x44d5] = "\xe8\x78",
  [0x44d7] = "\xe8\x6d", [0x44d8] = "\xe8\x6b", [0x44d9] = "\xe8\x66",
  [0x44db] = "\xfa\x41", [0x44dd] = "\xe8\x6e", [0x44de] = "\xe8\x7b",
  [0x44df] = "\xe8\x6a", [0x44e0] = "\xe8\x7a", [0x44e1] = "\xe8\xa2",
  [0x44e3] = "\x9a\x40", [0x44e4] = "\xbe\x53", [0x44e5] = "\x97\x5b",
  [0x44e6] = "\xe8\x76", [0x44e7] = "\xe8\x7c", [0x44e8] = "\xe8\x72",
  [0x44e9] = "\xe8\x6c", [0x44ea] = "\xbe\x51", [0x44eb] = "\x9a\x41",
  [0x44ec] = "\x91\xdd", [0x44ee] = "\xe4\xa8", [0x44ef] = "\xe8\x70",
  [0x44f0] = "\xbe\x59", [0x44f1] = "\xe8\x69", [0x44f2] = "\x93\xfc",
  [0x44f3] = "\x9a\x42", [0x44f4] = "\x9a\x43", [0x44f6] = "\x96\x59",
  [0x44f7] = "\xeb\xf4", [0x44f8] = "\xbf\xf7", [0x44f9] = "\xeb\xf3",
  [0x44fa] = "\xeb\xf0", [0x44fb] = "\xec\x44", [0x44fc] = "\xbf\xfb",
  [0x44fd] = "\x9a\x44", [0x44fe] = "\xec\x41", [0x44ff] = "\xeb\xf8",
  [0x4500] = "\xec\x43", [0x4501] = "\xeb\xe9", [0x4502] = "\xeb\xf6",
  [0x4503] = "\x90\x51", [0x4504] = "\xbf\xfd", [0x4506] = "\xeb\xe1",
  [0x4507] = "\x94\xbf", [0x4508] = "\xeb\xdf", [0x4509] = "\xec\x42",
  [0x450b] = "\xec\x40", [0x450c] = "\xeb\xfe", [0x450d] = "\xeb\xed",
  [0x450e] = "\xeb\xec", [0x450f] = "\xeb\xe2", [0x4510] = "\xc0\x40",
  [0x4512] = "\xeb\xe8", [0x4513] = "\xeb\xf2", [0x4514] = "\xeb\xfd",
  [0x4515] = "\xc0\x43", [0x4516] = "\xec\x45", [0x4518] = "\xc1\xe8",
  [0x4519] = "\xc0\x45", [0x451a] = "\xbf\xfe", [0x451b] = "\xeb\xe6",
  [0x451d] = "\xeb\xef", [0x451e] = "\xeb\xde", [0x451f] = "\xeb\xe0",
  [0x4520] = "\xbf\xf5", [0x4521] = "\xc0\x42", [0x4522] = "\xbf\xfa",
  [0x4523] = "\xeb\xe7", [0x4524] = "\xeb\xf7", [0x4525] = "\xeb\xf1",
  [0x4526] = "\xc0\x41", [0x4527] = "\xeb\xdd", [0x4528] = "\xc1\xe3",
  [0x4529] = "\xeb\xf9", [0x452a] = "\xeb\xfc", [0x452b] = "\xbf\xfc",
  [0x452c] = "\x90\xa2", [0x452d] = "\xeb\xeb", [0x452e] = "\xc0\x44",
  [0x452f] = "\xbf\xf9", [0x4530] = "\x9c\xab", [0x4531] = "\x97\x76",
  [0x4533] = "\xbf\xf8", [0x4534] = "\xeb\xf5", [0x4535] = "\xeb\xfb",
  [0x4536] = "\xbf\xf6", [0x4538] = "\xeb\xe4", [0x4539] = "\xeb\xfa",
  [0x453c] = "\xeb\xe5", [0x4540] = "\xfc\x55", [0x4541] = "\xfe\x45",
  [0x4542] = "\x94\xa8", [0x4543] = "\x9a\x45", [0x4544] = "\xfa\x4b",
  [0x4545] = "\x9d\xe1", [0x4546] = "\xeb\xea", [0x4547] = "\xee\xd2",
  [0x4548] = "\x96\xd9", [0x4549] = "\xee\xd7", [0x454a] = "\xc1\xe5",
  [0x454b] = "\xc1\xe7", [0x454c] = "\xee\xdd", [0x454d] = "\xc1\xe1",
  [0x454e] = "\xee\xec", [0x454f] = "\xee\xe3", [0x4550] = "\xee\xd8",
  [0x4551] = "\xee\xd9", [0x4552] = "\xee\xe2", [0x4554] = "\xc1\xee",
  [0x4555] = "\xee\xe1", [0x4556] = "\xee\xd1", [0x4557] = "\xee\xe0",
  [0x4558] = "\xee\xd4", [0x4559] = "\xee\xed", [0x455a] = "\xc1\xed",
  [0x455b] = "\xc1\xeb", [0x455c] = "\xee\xd5", [0x455e] = "\xee\xe8",
  [0x455f] = "\x97\x74", [0x4560] = "\xee\xda", [0x4561] = "\xee\xe7",
  [0x4562] = "\xfd\xf5", [0x4563] = "\xee\xe9", [0x4564] = "\xee\xd0",
  [0x4565] = "\xc1\xe6", [0x4566] = "\x92\xe5", [0x4567] = "\xee\xea",
  [0x4568] = "\x96\x45", [0x4569] = "\x91\xda", [0x456a] = "\xee\xde",
  [0x456b] = "\x90\xa3", [0x456c] = "\xc1\xea", [0x456d] = "\xee\xdb",
  [0x456e] = "\xa0\x5f", [0x4570] = "\xc1\xec", [0x4571] = "\xee\xe4",
  [0x4573] = "\x90\xaf", [0x4574] = "\x97\xbf", [0x4575] = "\xc1\xe4",
  [0x4576] = "\xee\xd6", [0x4577] = "\xee\xe5", [0x4578] = "\x91\x4c",
  [0x4579] = "\xee\xdf", [0x457a] = "\xeb\xe3", [0x457b] = "\xee\xe6",
  [0x457c] = "\xee\xd3", [0x457d] = "\x96\x7a", [0x457e] = "\xc1\xe9",
  [0x4580] = "\xee\xeb", [0x4581] = "\x91\xde", [0x4582] = "\xc1\xe2",
  [0x4583] = "\xee\xce", [0x4584] = "\x9a\x46", [0x4585] = "\xfe\xb0",
  [0x4586] = "\x97\x79", [0x4587] = "\x94\x6c", [0x4588] = "\xf1\x60",
  [0x4589] = "\xf1\x59", [0x458a] = "\xc2\xe9", [0x458c] = "\xf1\x54",
  [0x458d] = "\xf1\x63", [0x458e] = "\xf1\x5b", [0x458f] = "\xee\xdc",
  [0x4590] = "\x98\x58", [0x4591] = "\xf1\x65", [0x4592] = "\xf1\x55",
  [0x4594] = "\xc2\xe8", [0x4595] = "\xf1\x5f", [0x4596] = "\xc2\xea",
  [0x4597] = "\xc2\xf2", [0x4598] = "\xc2\xf0", [0x4599] = "\xf1\x61",
  [0x459a] = "\xc2\xf1", [0x459b] = "\xf1\x57", [0x459c] = "\x92\x66",
  [0x459d] = "\xf1\x58", [0x459e] = "\xf1\x5d", [0x459f] = "\xf1\x62",
  [0x45a0] = "\x93\xfb", [0x45a1] = "\xee\xcd", [0x45a2] = "\xc2\xeb",
  [0x45a3] = "\xf1\x6a", [0x45a4] = "\xf1\x67", [0x45a5] = "\xf1\x6b",
  [0x45a6] = "\xf1\x5e", [0x45a7] = "\xf1\x5a", [0x45a8] = "\xf1\x68",
  [0x45a9] = "\xf3\x6a", [0x45aa] = "\xf1\x5c", [0x45ac] = "\xc2\xee",
  [0x45ad] = "\x9a\x47", [0x45ae] = "\xc2\xed", [0x45af] = "\xee\xcf",
  [0x45b0] = "\xc2\xef", [0x45b1] = "\xf1\x64", [0x45b2] = "\xf1\x66",
  [0x45b3] = "\xc2\xec", [0x45b4] = "\xf1\x69", [0x45b5] = "\xf1\x53",
  [0x45b7] = "\xf1\x56", [0x45b8] = "\x97\x49", [0x45bb] = "\x97\x48",
  [0x45bd] = "\x93\x4a", [0x45bf] = "\x9c\xe2", [0x45c0] = "\xf3\x73",
  [0x45c2] = "\xf3\x63", [0x45c3] = "\xc3\xeb", [0x45c4] = "\xf3\x71",
  [0x45c6] = "\x92\x64", [0x45c7] = "\xf3\x61", [0x45c8] = "\xc3\xec",
  [0x45ca] = "\xf3\x6c", [0x45cb] = "\x91\xdf", [0x45cc] = "\xf3\x68",
  [0x45cd] = "\xc3\xf1", [0x45ce] = "\xf3\x72", [0x45cf] = "\xf3\x62",
  [0x45d0] = "\xf3\x65", [0x45d1] = "\xc3\xe9", [0x45d2] = "\xf3\x74",
  [0x45d3] = "\xfb\x79", [0x45d4] = "\xf3\x6d", [0x45d5] = "\xf3\x70",
  [0x45d6] = "\xc3\xef", [0x45d7] = "\xc3\xf4", [0x45d8] = "\xc3\xf2",
  [0x45d9] = "\xf3\x69", [0x45da] = "\xf3\x64", [0x45db] = "\x96\xd7",
  [0x45dc] = "\xc3\xed", [0x45dd] = "\xc3\xee", [0x45de] = "\xf3\x60",
  [0x45df] = "\xc3\xea", [0x45e0] = "\x93\x43", [0x45e1] = "\xc3\xe8",
  [0x45e2] = "\xc3\xf0", [0x45e3] = "\xf3\x6f", [0x45e4] = "\xc3\xf3",
  [0x45e6] = "\xf3\x6b", [0x45e7] = "\xf3\x75", [0x45e8] = "\xc3\xf5",
  [0x45ec] = "\xf3\x67", [0x45ee] = "\xf3\x6e", [0x45f0] = "\xfd\xcb",
  [0x45f1] = "\xfe\x7a", [0x45f3] = "\x91\xdb", [0x45f5] = "\xf4\xf3",
  [0x45f6] = "\xf5\x42", [0x45f7] = "\xf4\xf5", [0x45f8] = "\xf4\xfc",
  [0x45f9] = "\xf3\x66", [0x45fa] = "\xf4\xfa", [0x45fb] = "\xf4\xe9",
  [0x45fc] = "\xf5\x40", [0x45fd] = "\xc4\xc3", [0x45fe] = "\xf4\xed",
  [0x45ff] = "\xf4\xfe", [0x4600] = "\xf4\xf4", [0x4601] = "\x97\xaf",
  [0x4603] = "\xc4\xc2", [0x4604] = "\x95\xdd", [0x4606] = "\xf5\x44",
  [0x4607] = "\xf4\xf6", [0x4608] = "\x93\x48", [0x4609] = "\xf4\xfb",
  [0x460a] = "\xf4\xfd", [0x460b] = "\xf4\xe7", [0x460c] = "\xf5\x41",
  [0x460d] = "\xf4\xf2", [0x460e] = "\xf4\xf7", [0x460f] = "\xf4\xeb",
  [0x4610] = "\xf4\xef", [0x4611] = "\xf5\x43", [0x4612] = "\xf4\xf9",
  [0x4613] = "\xf4\xe8", [0x4614] = "\xf4\xec", [0x4615] = "\xf4\xee",
  [0x4616] = "\xf4\xf8", [0x4617] = "\x9a\x4b", [0x4618] = "\xc4\xc1",
  [0x4619] = "\xf4\xf1", [0x461b] = "\xfc\x45", [0x461d] = "\x9a\x4d",
  [0x4620] = "\xf4\xea", [0x4624] = "\x91\xbc", [0x4625] = "\x90\xe2",
  [0x4626] = "\x90\xb4", [0x4627] = "\x95\xe1", [0x4628] = "\xf4\xf0",
  [0x4629] = "\xf6\x61", [0x462a] = "\xf6\x66", [0x462b] = "\xc5\x4f",
  [0x462c] = "\xf6\x68", [0x462d] = "\x9a\x4e", [0x462e] = "\xc5\x49",
  [0x4630] = "\xf6\x64", [0x4631] = "\xf6\x6a", [0x4632] = "\xc5\x4e",
  [0x4633] = "\xc5\x4a", [0x4635] = "\xc5\x4b", [0x4636] = "\xf6\x60",
  [0x4637] = "\xf6\x67", [0x4638] = "\xc5\x4d", [0x4639] = "\xf6\x65",
  [0x463a] = "\xc5\x4c", [0x463b] = "\xf6\x5f", [0x463c] = "\xf6\x63",
  [0x463d] = "\xf6\x62", [0x463e] = "\x9a\x4f", [0x463f] = "\xf6\x5e",
  [0x4640] = "\xf6\x69", [0x4642] = "\xfe\x40", [0x4643] = "\xfe\x43",
  [0x4644] = "\xc5\xb1", [0x4645] = "\xf7\x6d", [0x4646] = "\xf7\x70",
  [0x4647] = "\xf7\x6c", [0x4648] = "\xf7\x6e", [0x4649] = "\xf7\x6f",
  [0x464a] = "\xf7\x69", [0x464b] = "\xf7\x6a", [0x464c] = "\xf7\x67",
  [0x464d] = "\x96\xdd", [0x464f] = "\xf7\x6b", [0x4650] = "\xf7\x68",
  [0x4651] = "\xc5\xb2", [0x4652] = "\xc5\xb3", [0x4654] = "\x9a\x51",
  [0x4655] = "\xf8\x4b", [0x4657] = "\xf8\x4d", [0x4658] = "\x96\xa7",
  [0x465b] = "\x90\xb0", [0x465d] = "\xf8\x4c", [0x465e] = "\xf8\x4e",
  [0x4660] = "\xc5\xe0", [0x4662] = "\xf8\x4a", [0x4663] = "\xc5\xdf",
  [0x4664] = "\xc5\xe1", [0x4665] = "\x9c\x4e", [0x4667] = "\x94\x43",
  [0x4668] = "\xf8\xcb", [0x4669] = "\xf8\xcc", [0x466a] = "\xc6\x44",
  [0x466b] = "\xf8\xca", [0x466c] = "\x8e\xba", [0x466d] = "\xf9\x53",
  [0x466e] = "\xf9\x52", [0x466f] = "\xf9\x54", [0x4670] = "\xc6\x5f",
  [0x4671] = "\xf9\x55", [0x4672] = "\xc6\x5e", [0x4673] = "\xf9\x56",
  [0x4674] = "\xf9\x72", [0x4675] = "\xf9\x75", [0x4676] = "\xf9\x74",
  [0x4677] = "\xc6\x68", [0x4678] = "\xf9\x73", [0x4679] = "\x9a\x52",
  [0x467b] = "\xfc\xc1", [0x467c] = "\xc6\x72", [0x467d] = "\xc6\x70",
  [0x467e] = "\xc6\x71", [0x467f] = "\xc6\x77", [0x4680] = "\xf9\xc0",
  [0x4681] = "\xf9\xc1", [0x4682] = "\xf9\xbf", [0x4683] = "\xf9\xc9",
  [0x4685] = "\x8b\xe9", [0x469f] = "\x9c\xaf", [0x46a2] = "\x8b\xfd",
  [0x46c1] = "\x9a\xbc", [0x46c3] = "\x9a\xb8", [0x46dc] = "\x9a\xae",
  [0x46f6] = "\x9a\xa7", [0x472d] = "\x9a\x53", [0x4747] = "\x9d\x74",
  [0x4777] = "\xaa\xf8", [0x4778] = "\x8b\xea", [0x477a] = "\xd8\x44",
  [0x477b] = "\xdc\x78", [0x477c] = "\xe8\xa5", [0x477d] = "\xf3\x76",
  [0x477f] = "\x8b\xeb", [0x4780] = "\xaa\xf9", [0x4782] = "\xad\xac",
  [0x4783] = "\xb0\x7b", [0x4785] = "\x90\xb2", [0x4786] = "\xd8\x45",
  [0x4788] = "\xd8\x46", [0x4789] = "\xb3\xac", [0x478b] = "\xb6\x7d",
  [0x478c] = "\xdc\x7a", [0x478d] = "\xdc\x79", [0x478e] = "\xb6\xa3",
  [0x478f] = "\xb6\x7c", [0x4790] = "\xdc\x7b", [0x4791] = "\xb6\x7e",
  [0x4792] = "\xb6\xa2", [0x4793] = "\xb6\xa1", [0x4794] = "\xb6\x7b",
  [0x4796] = "\x95\xe9", [0x4797] = "\x95\xe8", [0x4798] = "\xb9\x68",
  [0x4799] = "\x95\xe6", [0x479b] = "\xe0\xd0", [0x479c] = "\xe0\xce",
  [0x479e] = "\xe0\xcf", [0x479f] = "\xe0\xcd", [0x47a0] = "\x90\xb5",
  [0x47a1] = "\xbb\xd2", [0x47a2] = "\x9a\x54", [0x47a3] = "\xbb\xd5",
  [0x47a4] = "\xbb\xd7", [0x47a5] = "\xbb\xd6", [0x47a6] = "\x90\xb3",
  [0x47a7] = "\x95\xe7", [0x47a8] = "\xbb\xd3", [0x47a9] = "\xbb\xd4",
  [0x47aa] = "\x8b\x50", [0x47ab] = "\xe8\xa7", [0x47ac] = "\xe8\xa6",
  [0x47ad] = "\xbe\x5b", [0x47ae] = "\xe8\xa8", [0x47b0] = "\xe8\xa9",
  [0x47b1] = "\xbe\x5c", [0x47b5] = "\xec\x4d", [0x47b6] = "\xec\x4b",
  [0x47b7] = "\xee\xf3", [0x47b9] = "\xec\x49", [0x47ba] = "\xec\x4a",
  [0x47bb] = "\xc0\x46", [0x47bc] = "\xec\x46", [0x47bd] = "\xec\x4e",
  [0x47be] = "\xec\x48", [0x47bf] = "\xec\x4c", [0x47c0] = "\xee\xef",
  [0x47c3] = "\xee\xf1", [0x47c5] = "\xee\xf2", [0x47c6] = "\xc1\xf3",
  [0x47c7] = "\xee\xee", [0x47c8] = "\xc1\xf2", [0x47c9] = "\xee\xf0",
  [0x47ca] = "\xc1\xef", [0x47cb] = "\xc1\xf0", [0x47cc] = "\xc1\xf1",
  [0x47cd] = "\xec\x47", [0x47d0] = "\xc2\xf5", [0x47d1] = "\xf1\x6e",
  [0x47d2] = "\xf1\x6c", [0x47d3] = "\xf1\x6d", [0x47d4] = "\xc2\xf3",
  [0x47d5] = "\xc2\xf6", [0x47d6] = "\xc2\xf4", [0x47da] = "\xf3\x77",
  [0x47db] = "\xf3\x78", [0x47dc] = "\xc3\xf6", [0x47de] = "\xf5\x45",
  [0x47df] = "\xf5\x47", [0x47e0] = "\xf5\x46", [0x47e1] = "\xc4\xc4",
  [0x47e2] = "\xc5\x50", [0x47e3] = "\xf6\x6d", [0x47e4] = "\xf6\x6c",
  [0x47e5] = "\xf6\x6b", [0x47e8] = "\x8b\xec", [0x47f4] = "\x9a\x56",
  [0x481c] = "\xaa\xfa", [0x481d] = "\x8b\xfb", [0x481e] = "\xc9\xaa",
  [0x4820] = "\xca\x58", [0x4821] = "\xa6\xe9", [0x4822] = "\xca\x56",
  [0x4823] = "\xca\x59", [0x4824] = "\xca\x57", [0x4828] = "\xcb\xae",
  [0x482a] = "\xa8\xc1", [0x482c] = "\xa8\xc2", [0x482d] = "\xcb\xb0",
  [0x482e] = "\xa8\xbf", [0x482f] = "\xcb\xaf", [0x4830] = "\xcb\xad",
  [0x4831] = "\xa8\xc0", [0x4832] = "\xa8\xbe", [0x4833] = "\x9a\x57",
  [0x4838] = "\xa0\xaa", [0x4839] = "\xcd\xd8", [0x483a] = "\xcd\xdb",
  [0x483b] = "\xaa\xfd", [0x483c] = "\xcd\xda", [0x483d] = "\xcd\xd9",
  [0x483f] = "\xaa\xfc", [0x4840] = "\xaa\xfb", [0x4841] = "\x9f\xa6",
  [0x4842] = "\xab\x40", [0x4843] = "\xcd\xdc", [0x4844] = "\xaa\xfe",
  [0x4845] = "\x99\xcc", [0x484a] = "\xd0\xc6", [0x484b] = "\xad\xae",
  [0x484c] = "\xad\xaf", [0x484d] = "\xad\xb0", [0x484e] = "\xd0\xc7",
  [0x484f] = "\xd0\xc3", [0x4850] = "\xad\xad", [0x4851] = "\xd0\xc4",
  [0x4853] = "\xd0\xc5", [0x4854] = "\xd0\xc2", [0x4856] = "\x9c\x59",
  [0x4858] = "\xb0\xa4", [0x485b] = "\xb0\xa1", [0x485c] = "\xd4\x45",
  [0x485d] = "\xb0\xa2", [0x485e] = "\xb0\xa5", [0x485f] = "\xd4\x46",
  [0x4861] = "\xb0\x7e", [0x4862] = "\xb0\x7c", [0x4863] = "\xb0\x7d",
  [0x4864] = "\xb0\xa3", [0x4869] = "\x99\xb5", [0x486a] = "\xb3\xad",
  [0x486b] = "\xd8\x49", [0x486c] = "\xb3\xb5", [0x486d] = "\xd8\x48",
  [0x486f] = "\xd8\x4b", [0x4870] = "\xb3\xb1", [0x4871] = "\xd8\x4a",
  [0x4872] = "\xb6\xab", [0x4873] = "\xb3\xaf", [0x4874] = "\xb3\xb2",
  [0x4875] = "\xb3\xae", [0x4876] = "\xb3\xb3", [0x4877] = "\xb3\xb4",
  [0x4878] = "\xb3\xb0", [0x487b] = "\x90\xbe", [0x487c] = "\xd8\x47",
  [0x487d] = "\xb6\xa7", [0x487e] = "\xdc\x7d", [0x4880] = "\xdc\xa3",
  [0x4881] = "\x9f\xaf", [0x4883] = "\xdc\xa2", [0x4884] = "\xb6\xac",
  [0x4885] = "\xb6\xa8", [0x4886] = "\xb6\xa9", [0x4887] = "\xdc\x7c",
  [0x4888] = "\xdc\x7e", [0x4889] = "\xdc\xa1", [0x488a] = "\xb6\xa4",
  [0x488b] = "\xb6\xa6", [0x488d] = "\xb6\xaa", [0x488e] = "\xb6\xa5",
  [0x488f] = "\x95\xf2", [0x4891] = "\xe0\xd3", [0x4892] = "\xe0\xd1",
  [0x4893] = "\xe0\xd2", [0x4894] = "\xb9\x6a", [0x4895] = "\xb9\x6b",
  [0x4896] = "\x90\xbf", [0x4897] = "\xe0\xd4", [0x4898] = "\xb9\x69",
  [0x4899] = "\xbb\xd8", [0x489b] = "\xbb\xda", [0x489c] = "\xbb\xd9",
  [0x489e] = "\xe4\xbb", [0x48a1] = "\xe4\xbc", [0x48a2] = "\xe8\xab",
  [0x48a3] = "\x90\xc1", [0x48a4] = "\xe8\xaa", [0x48a5] = "\xfe\xe4",
  [0x48a7] = "\xc0\x47", [0x48a8] = "\xc0\x48", [0x48a9] = "\xec\x4f",
  [0x48aa] = "\xc0\x49", [0x48ac] = "\xee\xf6", [0x48ae] = "\xee\xf4",
  [0x48b0] = "\xee\xf5", [0x48b1] = "\xc1\xf4", [0x48b3] = "\xf1\x6f",
  [0x48b4] = "\xc3\xf7", [0x48b6] = "\x90\xc4", [0x48b8] = "\xc1\xf5",
  [0x48b9] = "\xab\x41", [0x48bb] = "\xb0\xa6", [0x48bc] = "\xd4\x47",
  [0x48bd] = "\x90\xc7", [0x48bf] = "\xd8\x4c", [0x48c0] = "\xb3\xb6",
  [0x48c1] = "\xb6\xad", [0x48c2] = "\xdc\xa4", [0x48c3] = "\xdc\xa6",
  [0x48c4] = "\xb6\xaf", [0x48c5] = "\xb6\xae", [0x48c6] = "\xb6\xb0",
  [0x48c7] = "\xb6\xb1", [0x48c8] = "\xdc\xa5", [0x48c9] = "\xb9\x6e",
  [0x48ca] = "\xb9\x6f", [0x48cb] = "\xb9\x6d", [0x48cc] = "\xbb\xdb",
  [0x48cd] = "\xb9\x6c", [0x48ce] = "\xe0\xd5", [0x48d2] = "\xbb\xdc",
  [0x48d3] = "\xe8\xac", [0x48d4] = "\xec\x50", [0x48d5] = "\xc0\x4a",
  [0x48d6] = "\xc1\xf6", [0x48d7] = "\xf1\x70", [0x48d8] = "\xf1\x74",
  [0x48d9] = "\xc2\xf9", [0x48da] = "\xf1\x71", [0x48db] = "\xc2\xfa",
  [0x48dc] = "\xc2\xf8", [0x48dd] = "\xf1\x75", [0x48de] = "\xc2\xfb",
  [0x48df] = "\xf1\x73", [0x48e1] = "\xf3\x79", [0x48e2] = "\xc2\xf7",
  [0x48e3] = "\xc3\xf8", [0x48e5] = "\xf8\xcd", [0x48e8] = "\xab\x42",
  [0x48e9] = "\xb3\xb8", [0x48ea] = "\xb3\xb7", [0x48ef] = "\xb6\xb2",
  [0x48f0] = "\xdc\xa8", [0x48f1] = "\xdc\xa7", [0x48f2] = "\xb6\xb3",
  [0x48f4] = "\x92\xe4", [0x48f5] = "\xe0\xd9", [0x48f6] = "\xb9\x73",
  [0x48f7] = "\xb9\x70", [0x48f8] = "\xe0\xd8", [0x48f9] = "\xb9\x72",
  [0x48fa] = "\xe0\xd6", [0x48fb] = "\xb9\x71", [0x48fd] = "\xe0\xd7",
  [0x48ff] = "\xe4\xbd", [0x4900] = "\xbb\xdd", [0x4902] = "\xe8\xaf",
  [0x4903] = "\x9f\x52", [0x4904] = "\xbe\x5d", [0x4905] = "\xe8\xad",
  [0x4906] = "\xbe\x5e", [0x4907] = "\xbe\x5f", [0x4908] = "\xe8\xae",
  [0x4909] = "\xbe\x60", [0x490b] = "\xec\x51", [0x490d] = "\xc0\x4e",
  [0x490e] = "\xc0\x4b", [0x490f] = "\xc0\x50", [0x4910] = "\xec\x53",
  [0x4911] = "\xc0\x4c", [0x4912] = "\xec\x52", [0x4913] = "\xc0\x4f",
  [0x4916] = "\xc0\x4d", [0x4918] = "\xee\xf9", [0x4919] = "\xee\xfb",
  [0x491b] = "\x90\xdb", [0x491c] = "\xc1\xf7", [0x491d] = "\xee\xfa",
  [0x491e] = "\xc1\xf8", [0x491f] = "\xee\xf8", [0x4920] = "\xee\xf7",
  [0x4921] = "\xa0\x66", [0x4922] = "\xf1\x77", [0x4923] = "\xf1\x76",
  [0x4924] = "\xc2\xfc", [0x4925] = "\xf1\x78", [0x4926] = "\xf3\x7e",
  [0x4927] = "\xc3\xfa", [0x4928] = "\xf3\x7d", [0x4929] = "\xf3\x7a",
  [0x492a] = "\xc3\xf9", [0x492b] = "\xf3\x7b", [0x492c] = "\xf3\x7c",
  [0x492e] = "\xf5\x48", [0x492f] = "\xf5\x49", [0x4930] = "\xc4\xc5",
  [0x4931] = "\x90\xd2", [0x4932] = "\xc5\x53", [0x4935] = "\xf6\x6e",
  [0x4936] = "\x90\xd4", [0x4938] = "\xc5\x51", [0x4939] = "\xc5\x52",
  [0x493a] = "\xf6\x6f", [0x493d] = "\xc5\xb4", [0x493e] = "\xc5\xb5",
  [0x493f] = "\xf7\x71", [0x4940] = "\x9a\x5b", [0x4941] = "\x95\xfd",
  [0x4942] = "\xc6\x45", [0x4943] = "\xf8\xcf", [0x4944] = "\xc6\x47",
  [0x4946] = "\xf8\xce", [0x4947] = "\xf8\xd0", [0x4948] = "\xc6\x46",
  [0x4949] = "\xf9\x57", [0x494b] = "\xf9\xad", [0x4951] = "\x8b\xc4",
  [0x4952] = "\xab\x43", [0x4956] = "\xb9\x74", [0x4957] = "\x90\xde",
  [0x4958] = "\xe4\xbe", [0x495a] = "\xe8\xb0", [0x495b] = "\xc0\x51",
  [0x495c] = "\xc0\x52", [0x495d] = "\x9c\xe4", [0x495e] = "\xab\x44",
  [0x495f] = "\x90\xe1", [0x4960] = "\xbe\x61", [0x4961] = "\xc3\xfb",
  [0x4962] = "\xad\xb1", [0x4966] = "\xc0\x53", [0x4968] = "\xc5\xe2",
  [0x4969] = "\xad\xb2", [0x496a] = "\xd8\x4d", [0x496c] = "\xdc\xa9",
  [0x496d] = "\x9e\x46", [0x496e] = "\xdc\xab", [0x4970] = "\xdc\xaa",
  [0x4971] = "\x96\x51", [0x4972] = "\xe0\xdd", [0x4973] = "\xe0\xda",
  [0x4974] = "\xb9\x75", [0x4976] = "\xb9\x76", [0x4977] = "\xe0\xdb",
  [0x4978] = "\xe0\xdc", [0x497a] = "\xe4\xc0", [0x497b] = "\xe4\xc5",
  [0x497c] = "\xbb\xde", [0x497d] = "\xe4\xbf", [0x497e] = "\xe4\xc1",
  [0x497f] = "\xe4\xc8", [0x4980] = "\xe4\xc3", [0x4981] = "\xe4\xc7",
  [0x4982] = "\xe4\xc4", [0x4983] = "\xe4\xc2", [0x4984] = "\xe4\xc6",
  [0x4985] = "\xbb\xdf", [0x4987] = "\xfb\x58", [0x4988] = "\xe8\xb3",
  [0x4989] = "\x90\xe6", [0x498a] = "\xe8\xb1", [0x498b] = "\xbe\x63",
  [0x498d] = "\xbe\x62", [0x498e] = "\xe8\xb2", [0x498f] = "\xbe\x64",
  [0x4994] = "\xec\x56", [0x4997] = "\xec\x55", [0x4998] = "\xc0\x54",
  [0x4999] = "\xec\x54", [0x499a] = "\xee\xfc", [0x499b] = "\x96\x50",
  [0x499c] = "\xee\xfe", [0x499d] = "\xef\x41", [0x499e] = "\xef\x40",
  [0x499f] = "\x90\xe7", [0x49a0] = "\xc1\xf9", [0x49a1] = "\xee\xfd",
  [0x49a2] = "\xf1\xa1", [0x49a3] = "\xc2\xfd", [0x49a4] = "\xf1\x7d",
  [0x49a5] = "\xf1\xa2", [0x49a6] = "\xc2\xfe", [0x49a8] = "\xf1\x7b",
  [0x49aa] = "\xf1\x7e", [0x49ab] = "\xf1\x7c", [0x49ac] = "\xf1\x79",
  [0x49ad] = "\xc3\x40", [0x49ae] = "\xf1\x7a", [0x49b1] = "\x90\xe8",
  [0x49b2] = "\x9a\x5d", [0x49b3] = "\xf3\xa1", [0x49b4] = "\x9f\x7a",
  [0x49b6] = "\xf3\xa3", [0x49b7] = "\xf3\xa2", [0x49b8] = "\x9b\x5c",
  [0x49b9] = "\xf5\x4a", [0x49ba] = "\x9f\x7c", [0x49bb] = "\xf5\x4b",
  [0x49bd] = "\xfc\x52", [0x49be] = "\x90\xe9", [0x49bf] = "\xf6\x70",
  [0x49c0] = "\x90\xea", [0x49c1] = "\xc5\xb7", [0x49c2] = "\x9a\x5e",
  [0x49c3] = "\xc5\xb6", [0x49c4] = "\xf8\x4f", [0x49c5] = "\xf8\x50",
  [0x49c6] = "\xc6\x48", [0x49c7] = "\xf8\xd1", [0x49c8] = "\x9f\x76",
  [0x49c9] = "\xc6\x69", [0x49cb] = "\xad\xb3", [0x49cc] = "\xb6\xb4",
  [0x49cd] = "\xe4\xca", [0x49ce] = "\xe4\xc9", [0x49cf] = "\xe8\xb5",
  [0x49d0] = "\xe8\xb4", [0x49d2] = "\x90\xeb", [0x49d3] = "\xc1\xfa",
  [0x49d4] = "\xef\x43", [0x49d5] = "\xef\x42", [0x49d6] = "\xf1\xa5",
  [0x49d7] = "\xf1\xa3", [0x49d8] = "\xf1\xa6", [0x49d9] = "\xf1\xa4",
  [0x49dc] = "\xc3\xfc", [0x49dd] = "\xf3\xa4", [0x49de] = "\xf3\xa5",
  [0x49df] = "\xf3\xa6", [0x49e0] = "\x90\xec", [0x49e1] = "\xf6\x71",
  [0x49e3] = "\xf7\x72", [0x49e5] = "\xf8\xd2", [0x49e6] = "\x8b\xee",
  [0x49ed] = "\xad\xb4", [0x49ee] = "\x90\xee", [0x49f0] = "\xec\x57",
  [0x49f1] = "\xef\x44", [0x49f2] = "\x91\xc6", [0x49f3] = "\xad\xb5",
  [0x49f5] = "\x90\xf2", [0x49f6] = "\xbb\xe0", [0x49f8] = "\xec\x58",
  [0x49f9] = "\xc3\x41", [0x49fa] = "\xf1\xa7", [0x49fb] = "\xc3\xfd",
  [0x49fd] = "\xf5\x4c", [0x49fe] = "\xf5\x4d", [0x49ff] = "\xc5\x54",
  [0x4a00] = "\xf8\x51", [0x4a01] = "\xad\xb6", [0x4a02] = "\xb3\xbb",
  [0x4a03] = "\xb3\xbc", [0x4a04] = "\xd8\x4e", [0x4a05] = "\xb6\xb5",
  [0x4a06] = "\xb6\xb6", [0x4a07] = "\xdc\xac", [0x4a08] = "\xb6\xb7",
  [0x4a0a] = "\xb9\x7a", [0x4a0c] = "\xb9\x7c", [0x4a0d] = "\xe0\xdf",
  [0x4a0e] = "\xe0\xe0", [0x4a0f] = "\xe0\xde", [0x4a10] = "\xb9\x77",
  [0x4a11] = "\xb9\x78", [0x4a12] = "\xb9\x7b", [0x4a13] = "\xb9\x79",
  [0x4a14] = "\xfc\xbc", [0x4a15] = "\x8a\x74", [0x4a16] = "\xe4\xcb",
  [0x4a17] = "\xbb\xe1", [0x4a18] = "\xbb\xe2", [0x4a1b] = "\xe8\xbc",
  [0x4a1c] = "\xbe\x67", [0x4a1d] = "\xe8\xb7", [0x4a1e] = "\xe8\xb6",
  [0x4a1f] = "\x96\x57", [0x4a20] = "\xe8\xbb", [0x4a21] = "\xbe\x65",
  [0x4a23] = "\x9c\xef", [0x4a24] = "\xc0\x5b", [0x4a26] = "\xe8\xb8",
  [0x4a27] = "\xe8\xbd", [0x4a28] = "\xe8\xba", [0x4a29] = "\xe8\xb9",
  [0x4a2b] = "\xbe\x66", [0x4a2d] = "\xc0\x59", [0x4a2e] = "\x9f\xdf",
  [0x4a2f] = "\xec\x5a", [0x4a30] = "\xc0\x55", [0x4a32] = "\xec\x5b",
  [0x4a33] = "\x90\xf7", [0x4a34] = "\x90\xf6", [0x4a35] = "\xec\x59",
  [0x4a37] = "\xc0\x58", [0x4a38] = "\xc0\x56", [0x4a39] = "\xc0\x5a",
  [0x4a3b] = "\xc0\x57", [0x4a41] = "\xef\x45", [0x4a43] = "\xef\x4a",
  [0x4a44] = "\xef\x46", [0x4a45] = "\xef\x49", [0x4a46] = "\xc1\xfb",
  [0x4a47] = "\x9b\x5e", [0x4a48] = "\xed\xd4", [0x4a49] = "\xef\x48",
  [0x4a4a] = "\xef\x47", [0x4a4b] = "\x90\xf8", [0x4a4c] = "\xc3\x44",
  [0x4a4d] = "\xc3\x42", [0x4a4e] = "\xc3\x45", [0x4a4f] = "\xc3\x43",
  [0x4a50] = "\xf1\xa8", [0x4a51] = "\xf1\xa9", [0x4a52] = "\xf1\xaa",
  [0x4a53] = "\xc3\x46", [0x4a57] = "\xf3\xaa", [0x4a58] = "\xc4\x40",
  [0x4a59] = "\xf3\xa8", [0x4a5b] = "\xc4\x41", [0x4a5c] = "\xf3\xa7",
  [0x4a5d] = "\xf3\xa9", [0x4a5e] = "\xc3\xfe", [0x4a5f] = "\xf5\x51",
  [0x4a60] = "\xf5\x4e", [0x4a62] = "\xf5\x4f", [0x4a63] = "\xf5\x50",
  [0x4a64] = "\xf6\x72", [0x4a65] = "\xc5\x56", [0x4a66] = "\x90\xf9",
  [0x4a67] = "\xc5\x55", [0x4a69] = "\xf7\x74", [0x4a6a] = "\xf7\x73",
  [0x4a6b] = "\xc5\xb8", [0x4a6c] = "\xfa\x6a", [0x4a6f] = "\xc5\xe3",
  [0x4a70] = "\xc6\x49", [0x4a71] = "\xc6\x60", [0x4a72] = "\xf9\x58",
  [0x4a73] = "\xf9\xae", [0x4a74] = "\xf9\xaf", [0x4a75] = "\x8b\xef",
  [0x4aa8] = "\xad\xb7", [0x4aa9] = "\xdc\xad", [0x4aac] = "\xe0\xe1",
  [0x4aad] = "\xe4\xcc", [0x4aae] = "\xe4\xcd", [0x4aaf] = "\xbb\xe3",
  [0x4ab1] = "\xbb\xe4", [0x4ab2] = "\xe8\xbe", [0x4ab3] = "\xbe\x68",
  [0x4ab4] = "\x9f\xe0", [0x4ab6] = "\xc1\xfc", [0x4ab7] = "\x91\x42",
  [0x4ab8] = "\xf1\xab", [0x4ab9] = "\x9a\x62", [0x4aba] = "\xc3\x47",
  [0x4abb] = "\xf3\xad", [0x4abc] = "\xc4\x42", [0x4abd] = "\xf3\xac",
  [0x4abe] = "\xf3\xae", [0x4abf] = "\xf3\xab", [0x4ac0] = "\xf6\x75",
  [0x4ac1] = "\xf5\x52", [0x4ac2] = "\xf5\x53", [0x4ac3] = "\x95\x69",
  [0x4ac4] = "\xc4\xc6", [0x4ac6] = "\xf6\x74", [0x4ac7] = "\x91\x44",
  [0x4ac8] = "\x91\x43", [0x4ac9] = "\xf6\x73", [0x4aca] = "\x91\x41",
  [0x4acb] = "\xf7\x75", [0x4acc] = "\xf9\xb0", [0x4ace] = "\x8b\xf0",
  [0x4adb] = "\xad\xb8", [0x4adc] = "\x96\x60", [0x4ade] = "\x8b\xf1",
  [0x4adf] = "\xad\xb9", [0x4ae0] = "\x99\xf6", [0x4ae1] = "\x91\x49",
  [0x4ae2] = "\xb0\xa7", [0x4ae3] = "\xd4\x48", [0x4ae5] = "\xd8\x4f",
  [0x4ae6] = "\x91\x4a", [0x4ae7] = "\xb6\xb8", [0x4ae9] = "\xb6\xbb",
  [0x4aea] = "\xb6\xb9", [0x4aeb] = "\xdc\xae", [0x4aec] = "\x91\x4b",
  [0x4aed] = "\xb6\xbd", [0x4aef] = "\xb6\xba", [0x4af1] = "\x9a\x64",
  [0x4af2] = "\xb6\xbc", [0x4af4] = "\xb9\x7e", [0x4af5] = "\x8a\xbf",
  [0x4af6] = "\xe0\xe2", [0x4af9] = "\xe0\xe3", [0x4afa] = "\xe8\xc0",
  [0x4afc] = "\xb9\x7d", [0x4afd] = "\xb9\xa1", [0x4afe] = "\xb9\xa2",
  [0x4b00] = "\xe4\xcf", [0x4b02] = "\xe4\xce", [0x4b03] = "\xbb\xe5",
  [0x4b05] = "\xbb\xe6", [0x4b07] = "\xe4\xd0", [0x4b08] = "\xe8\xbf",
  [0x4b09] = "\xbb\xe8", [0x4b0a] = "\xbe\x69", [0x4b0c] = "\xbb\xe7",
  [0x4b0e] = "\x9a\x66", [0x4b10] = "\xc0\x5c", [0x4b11] = "\xe8\xc1",
  [0x4b12] = "\xbe\x6b", [0x4b13] = "\xbe\x6a", [0x4b14] = "\xe8\xc2",
  [0x4b15] = "\xe8\xc5", [0x4b16] = "\xe8\xc3", [0x4b17] = "\xe8\xc4",
  [0x4b18] = "\xbe\x6c", [0x4b19] = "\x9a\x67", [0x4b1a] = "\xc0\x61",
  [0x4b1b] = "\xc0\x5f", [0x4b1c] = "\x9a\x69", [0x4b1e] = "\xc0\x5e",
  [0x4b1f] = "\xec\x5d", [0x4b21] = "\xc0\x60", [0x4b24] = "\xec\x5c",
  [0x4b25] = "\xef\x4b", [0x4b27] = "\xec\x5e", [0x4b28] = "\xc0\x5d",
  [0x4b29] = "\xec\x5f", [0x4b2a] = "\xef\x4e", [0x4b2b] = "\xef\x4c",
  [0x4b2c] = "\xef\x4d", [0x4b2d] = "\xef\x52", [0x4b2e] = "\xc3\x4b",
  [0x4b2f] = "\xef\x51", [0x4b30] = "\xef\x54", [0x4b31] = "\xef\x53",
  [0x4b32] = "\xef\x50", [0x4b33] = "\xef\x4f", [0x4b35] = "\xc1\xfd",
  [0x4b37] = "\x9a\x6a", [0x4b38] = "\x96\x52", [0x4b39] = "\x91\x4d",
  [0x4b3a] = "\xf1\xae", [0x4b3b] = "\x96\x66", [0x4b3c] = "\xf1\xad",
  [0x4b3d] = "\xc3\x4a", [0x4b3e] = "\xc3\x48", [0x4b3f] = "\xc3\x49",
  [0x4b40] = "\x9f\x7b", [0x4b41] = "\xf1\xac", [0x4b42] = "\x9a\x6b",
  [0x4b43] = "\xf3\xb1", [0x4b45] = "\xc4\x43", [0x4b47] = "\xf3\xb0",
  [0x4b48] = "\xf3\xaf", [0x4b49] = "\xc4\x44", [0x4b4a] = "\xa0\x6c",
  [0x4b4b] = "\xf5\x58", [0x4b4c] = "\xf5\x57", [0x4b4d] = "\x96\x67",
  [0x4b4e] = "\xf5\x55", [0x4b50] = "\xf5\x54", [0x4b51] = "\xc4\xc8",
  [0x4b52] = "\xc4\xc7", [0x4b53] = "\xf5\x59", [0x4b54] = "\xf7\x76",
  [0x4b55] = "\xc5\xb9", [0x4b56] = "\xf6\x77", [0x4b57] = "\xc5\x57",
  [0x4b58] = "\xf6\x76", [0x4b59] = "\xf5\x56", [0x4b5b] = "\xf7\x77",
  [0x4b5c] = "\xc5\xe4", [0x4b5d] = "\x9a\x6c", [0x4b5e] = "\xc6\x61",
  [0x4b5f] = "\xf9\x59", [0x4b61] = "\xf9\xb1", [0x4b62] = "\x9a\x6d",
  [0x4b63] = "\x8b\xf2", [0x4b96] = "\xad\xba", [0x4b97] = "\xd8\x50",
  [0x4b98] = "\xef\x55", [0x4b99] = "\xad\xbb", [0x4b9b] = "\x96\x6a",
  [0x4b9c] = "\xe4\xd2", [0x4b9d] = "\xe4\xd1", [0x4b9e] = "\xec\x60",
  [0x4ba1] = "\xef\x57", [0x4ba3] = "\xef\x56", [0x4ba4] = "\xfc\xea",
  [0x4ba5] = "\xc3\x4c", [0x4ba6] = "\xf3\xb2", [0x4ba7] = "\xf3\xb3",
  [0x4ba8] = "\xc4\xc9", [0x4baa] = "\x96\x6c", [0x4bab] = "\xf9\xb2",
  [0x4bac] = "\xb0\xa8", [0x4bad] = "\xb6\xbf", [0x4bae] = "\xb6\xbe",
  [0x4baf] = "\xe0\xe4", [0x4bb0] = "\xe0\xe6", [0x4bb1] = "\xb9\xa4",
  [0x4bb2] = "\xe0\xe5", [0x4bb3] = "\xb9\xa3", [0x4bb4] = "\xb9\xa5",
  [0x4bb5] = "\xe0\xe7", [0x4bb8] = "\x91\xc4", [0x4bb9] = "\xe4\xd4",
  [0x4bba] = "\xe4\xd6", [0x4bbb] = "\xe4\xd5", [0x4bbc] = "\x96\x77",
  [0x4bbd] = "\xe4\xd8", [0x4bc1] = "\xbb\xe9", [0x4bc2] = "\xe4\xd7",
  [0x4bc3] = "\xe4\xd3", [0x4bc4] = "\x99\xf4", [0x4bc5] = "\x9a\x6f",
  [0x4bc7] = "\xe4\xd9", [0x4bc9] = "\xe8\xcc", [0x4bcb] = "\xe8\xcf",
  [0x4bcc] = "\xe8\xd1", [0x4bcd] = "\xe8\xc7", [0x4bce] = "\xe8\xcb",
  [0x4bcf] = "\xe8\xc8", [0x4bd0] = "\xbe\x6e", [0x4bd1] = "\xbe\x71",
  [0x4bd2] = "\xbe\x73", [0x4bd3] = "\xe8\xc9", [0x4bd4] = "\xe8\xca",
  [0x4bd5] = "\xbe\x72", [0x4bd6] = "\xe8\xcd", [0x4bd7] = "\xe8\xd0",
  [0x4bd8] = "\xe8\xce", [0x4bd9] = "\xbe\x74", [0x4bda] = "\x9f\xab",
  [0x4bdb] = "\xbe\x70", [0x4bdc] = "\xe8\xc6", [0x4bdd] = "\xbe\x6d",
  [0x4bdf] = "\xbe\x6f", [0x4be1] = "\x8e\xc1", [0x4be2] = "\xc0\x63",
  [0x4be3] = "\xec\x66", [0x4be4] = "\xec\x64", [0x4be5] = "\xec\x63",
  [0x4be6] = "\x95\x55", [0x4be7] = "\xec\x69", [0x4be9] = "\xec\x68",
  [0x4bea] = "\xec\x67", [0x4bec] = "\xec\x62", [0x4bed] = "\xc0\x62",
  [0x4bee] = "\xec\x61", [0x4bf0] = "\xec\x65", [0x4bf1] = "\xc0\x64",
  [0x4bf4] = "\xef\x5a", [0x4bf5] = "\x91\x52", [0x4bf6] = "\xef\x5e",
  [0x4bf7] = "\xef\x5b", [0x4bf8] = "\xef\x5d", [0x4bf9] = "\xef\x5c",
  [0x4bfa] = "\xef\x59", [0x4bfb] = "\xef\x5f", [0x4bfc] = "\xef\x62",
  [0x4bfd] = "\xef\x60", [0x4bfe] = "\xef\x61", [0x4bff] = "\xc2\x40",
  [0x4c01] = "\xc1\xfe", [0x4c02] = "\xef\x58", [0x4c03] = "\xef\x63",
  [0x4c04] = "\xf1\xb3", [0x4c05] = "\xf1\xb6", [0x4c06] = "\xf1\xb8",
  [0x4c07] = "\xf1\xb7", [0x4c09] = "\xf1\xb1", [0x4c0a] = "\xf1\xb5",
  [0x4c0b] = "\xf1\xb0", [0x4c0c] = "\x91\x53", [0x4c0d] = "\xf1\xb2",
  [0x4c0e] = "\xc3\x4d", [0x4c0f] = "\xf1\xaf", [0x4c10] = "\x91\x55",
  [0x4c11] = "\xf1\xb4", [0x4c14] = "\xf3\xc0", [0x4c15] = "\xf3\xb5",
  [0x4c16] = "\xc4\x45", [0x4c19] = "\xc4\x46", [0x4c1a] = "\xf3\xb4",
  [0x4c1b] = "\xf3\xb9", [0x4c1c] = "\xf3\xbf", [0x4c1d] = "\xf3\xb7",
  [0x4c1e] = "\xf3\xbe", [0x4c1f] = "\x95\x5d", [0x4c20] = "\xf3\xbb",
  [0x4c21] = "\x96\x71", [0x4c22] = "\xf3\xba", [0x4c23] = "\xf3\xbd",
  [0x4c24] = "\xf3\xb8", [0x4c25] = "\xf3\xb6", [0x4c26] = "\x9c\x6d",
  [0x4c27] = "\xf3\xbc", [0x4c29] = "\xf5\x60", [0x4c2a] = "\xf5\x5e",
  [0x4c2b] = "\xc4\xca", [0x4c2c] = "\xf5\x5d", [0x4c2d] = "\xf5\x63",
  [0x4c2e] = "\xf5\x61", [0x4c2f] = "\x96\x73", [0x4c30] = "\xc4\xcb",
  [0x4c31] = "\xf5\x5c", [0x4c32] = "\xf5\x5a", [0x4c34] = "\xf5\x5b",
  [0x4c35] = "\xc4\xcd", [0x4c36] = "\xf5\x5f", [0x4c37] = "\xc4\xcc",
  [0x4c38] = "\xf5\x62", [0x4c39] = "\xf6\x78", [0x4c3a] = "\xf6\x7e",
  [0x4c3b] = "\x91\x54", [0x4c3c] = "\x9a\x71", [0x4c3d] = "\xf6\x79",
  [0x4c3e] = "\xc5\x5b", [0x4c3f] = "\xf6\xa1", [0x4c40] = "\xc5\x5a",
  [0x4c41] = "\xf6\x7d", [0x4c42] = "\xf6\x7c", [0x4c43] = "\xc5\x59",
  [0x4c44] = "\xf6\x7b", [0x4c45] = "\xc5\x58", [0x4c46] = "\xf6\x7a",
  [0x4c48] = "\xf7\x7d", [0x4c49] = "\xf7\xa1", [0x4c4a] = "\xf7\x7e",
  [0x4c4c] = "\xf7\x7b", [0x4c4d] = "\xc5\xbb", [0x4c4e] = "\xf7\x78",
  [0x4c4f] = "\xf7\x7c", [0x4c50] = "\xf7\xa3", [0x4c52] = "\xf7\xa2",
  [0x4c53] = "\xf7\x79", [0x4c54] = "\xf7\x7a", [0x4c55] = "\xc5\xba",
  [0x4c56] = "\xf8\x52", [0x4c57] = "\xc5\xe7", [0x4c58] = "\x91\x56",
  [0x4c59] = "\xf8\x53", [0x4c5a] = "\xc5\xe5", [0x4c5b] = "\xc5\xe6",
  [0x4c5c] = "\x96\x6d", [0x4c5e] = "\xf8\xd3", [0x4c5f] = "\xc6\x4a",
  [0x4c60] = "\xf9\x76", [0x4c62] = "\xc6\x6a", [0x4c63] = "\x95\x57",
  [0x4c64] = "\xf9\xb3", [0x4c65] = "\xc6\x6b", [0x4c66] = "\xf9\xb4",
  [0x4c67] = "\xf9\xb5", [0x4c68] = "\xf9\xc3", [0x4c69] = "\xf9\xc2",
  [0x4c6a] = "\xc6\x7a", [0x4c6b] = "\xf9\xcd", [0x4c6c] = "\x89\xc6",
  [0x4c8f] = "\x89\xc7", [0x4ca8] = "\xb0\xa9", [0x4cab] = "\xe0\xe9",
  [0x4cad] = "\xe0\xe8", [0x4caf] = "\xbb\xea", [0x4cb0] = "\xbb\xeb",
  [0x4cb1] = "\xe4\xda", [0x4cb2] = "\x8a\x6a", [0x4cb3] = "\xe8\xd2",
  [0x4cb4] = "\xec\x6c", [0x4cb6] = "\x8b\x57", [0x4cb7] = "\xbe\x75",
  [0x4cb8] = "\xc0\x65", [0x4cb9] = "\xec\x6a", [0x4cba] = "\x9f\xe1",
  [0x4cbb] = "\xec\x6d", [0x4cbc] = "\xc0\x66", [0x4cbd] = "\x9b\x5f",
  [0x4cbe] = "\xef\x64", [0x4cbf] = "\xec\x6b", [0x4cc0] = "\xf1\xb9",
  [0x4cc1] = "\xc3\x4e", [0x4cc2] = "\xf3\xc1", [0x4cc6] = "\xf5\x66",
  [0x4cc7] = "\xf5\x64", [0x4cca] = "\xf5\x65", [0x4ccd] = "\xf6\xa2",
  [0x4ccf] = "\xc5\x5c", [0x4cd0] = "\xf7\xa4", [0x4cd1] = "\xc5\xea",
  [0x4cd2] = "\xc5\xbc", [0x4cd3] = "\xc5\xe8", [0x4cd4] = "\xc5\xe9",
  [0x4cd5] = "\xf8\xd4", [0x4cd6] = "\xc6\x62", [0x4cd7] = "\xa0\x5d",
  [0x4cd8] = "\xb0\xaa", [0x4cdc] = "\xf1\xba", [0x4cdf] = "\xd4\x49",
  [0x4ce0] = "\x91\x5b", [0x4ce1] = "\xb9\xa6", [0x4ce2] = "\x91\x5c",
  [0x4ce3] = "\xe4\xdb", [0x4ce6] = "\xbb\xec", [0x4ce7] = "\xe4\xdc",
  [0x4ceb] = "\xe8\xd4", [0x4cec] = "\xe8\xd3", [0x4ced] = "\xc0\x68",
  [0x4cee] = "\xbe\x76", [0x4cef] = "\xbe\x77", [0x4cf1] = "\xe8\xd7",
  [0x4cf2] = "\xe8\xd6", [0x4cf3] = "\xe8\xd5", [0x4cf4] = "\x91\x5e",
  [0x4cf6] = "\xec\x6e", [0x4cf7] = "\xec\x71", [0x4cf9] = "\xec\x70",
  [0x4cfa] = "\xec\x6f", [0x4cfb] = "\xc0\x67", [0x4cfc] = "\xef\x68",
  [0x4cfd] = "\xef\x66", [0x4cfe] = "\xef\x65", [0x4cff] = "\x9f\x5c",
  [0x4d01] = "\xef\x67", [0x4d02] = "\x9f\x57", [0x4d03] = "\xc3\x4f",
  [0x4d04] = "\xf1\xbc", [0x4d05] = "\xf1\xbd", [0x4d06] = "\xc3\x50",
  [0x4d08] = "\xf1\xbb", [0x4d09] = "\x9f\x65", [0x4d0a] = "\xf3\xc3",
  [0x4d0b] = "\xf3\xc2", [0x4d0c] = "\xf3\xc5", [0x4d0d] = "\xc4\x47",
  [0x4d0e] = "\xf3\xc4", [0x4d0f] = "\x9a\x72", [0x4d10] = "\xf5\x67",
  [0x4d11] = "\xf5\x69", [0x4d12] = "\xf5\x68", [0x4d14] = "\x91\x60",
  [0x4d15] = "\xf6\xa3", [0x4d16] = "\xf6\xa6", [0x4d17] = "\xf6\xa4",
  [0x4d18] = "\xf6\xa5", [0x4d19] = "\xf7\xa5", [0x4d1a] = "\xc5\xbd",
  [0x4d1e] = "\xf8\x54", [0x4d1f] = "\xf8\x55", [0x4d20] = "\xf8\x56",
  [0x4d22] = "\xc6\x4b", [0x4d23] = "\xc6\x63", [0x4d24] = "\xf9\xb6",
  [0x4d25] = "\xb0\xab", [0x4d27] = "\xbe\x78", [0x4d28] = "\xc0\x69",
  [0x4d29] = "\xf1\xbe", [0x4d2a] = "\x9f\x5e", [0x4d2b] = "\xf7\xa6",
  [0x4d2d] = "\x91\x61", [0x4d2e] = "\xf9\xc4", [0x4d2f] = "\xd4\x4a",
  [0x4d31] = "\xc6\x7b", [0x4d32] = "\xb0\xac", [0x4d33] = "\xec\x72",
  [0x4d34] = "\x91\x64", [0x4d35] = "\xf1\xbf", [0x4d37] = "\xf3\xc6",
  [0x4d39] = "\x9f\x41", [0x4d3a] = "\xf6\xa7", [0x4d3b] = "\xf7\xa7",
  [0x4d3c] = "\xb0\xad", [0x4d3e] = "\xe4\xdd", [0x4d3f] = "\xe4\xde",
  [0x4d40] = "\x91\x69", [0x4d41] = "\xbb\xed", [0x4d42] = "\xbb\xee",
  [0x4d43] = "\xe8\xd9", [0x4d44] = "\xbe\x7a", [0x4d45] = "\xbe\x79",
  [0x4d46] = "\xe8\xd8", [0x4d48] = "\xef\x69", [0x4d4a] = "\xf1\xc0",
  [0x4d4b] = "\xf1\xc2", [0x4d4c] = "\xf1\xc1", [0x4d4d] = "\xc3\x53",
  [0x4d4e] = "\xc3\x52", [0x4d4f] = "\xc3\x51", [0x4d50] = "\x91\x68",
  [0x4d51] = "\xc5\x5e", [0x4d52] = "\xf6\xa8", [0x4d54] = "\xc5\x5d",
  [0x4d55] = "\xf7\xa9", [0x4d56] = "\xf7\xa8", [0x4d58] = "\xc6\x4c",
  [0x4d59] = "\xf8\xd5", [0x4d5a] = "\xb3\xbd", [0x4d5b] = "\xe0\xea",
  [0x4d5f] = "\xe4\xe1", [0x4d60] = "\xe4\xdf", [0x4d61] = "\xe4\xe0",
  [0x4d64] = "\xe8\xe2", [0x4d66] = "\xe8\xdd", [0x4d67] = "\xe8\xda",
  [0x4d68] = "\xe8\xe1", [0x4d69] = "\x9a\x74", [0x4d6c] = "\xe8\xe3",
  [0x4d6f] = "\xbe\x7c", [0x4d70] = "\xe8\xe0", [0x4d71] = "\xe8\xdc",
  [0x4d74] = "\xe8\xdb", [0x4d75] = "\xe8\xdf", [0x4d76] = "\xe8\xde",
  [0x4d77] = "\xbe\x7b", [0x4d7a] = "\xec\x7d", [0x4d7b] = "\xec\x78",
  [0x4d7c] = "\xec\x76", [0x4d7d] = "\xec\xa1", [0x4d7e] = "\xec\x77",
  [0x4d7f] = "\x96\xb2", [0x4d80] = "\xec\x73", [0x4d81] = "\x9a\x75",
  [0x4d82] = "\xec\x79", [0x4d83] = "\xfd\xa5", [0x4d85] = "\xec\x74",
  [0x4d86] = "\xef\x72", [0x4d87] = "\xec\x75", [0x4d88] = "\xec\xa2",
  [0x4d8b] = "\x9e\xe9", [0x4d8d] = "\x8b\xba", [0x4d8e] = "\x91\x6d",
  [0x4d8f] = "\xa0\x60", [0x4d90] = "\xec\x7c", [0x4d91] = "\xc0\x6a",
  [0x4d92] = "\xec\x7b", [0x4d93] = "\xec\x7a", [0x4d95] = "\xec\x7e",
  [0x4d97] = "\x9f\xde", [0x4d9a] = "\xef\x6a", [0x4d9b] = "\xef\x6d",
  [0x4d9d] = "\x9f\xc3", [0x4d9e] = "\xef\x6c", [0x4d9f] = "\x96\xb5",
  [0x4da0] = "\xef\x74", [0x4da1] = "\xef\x6f", [0x4da2] = "\xef\x73",
  [0x4da4] = "\xef\x71", [0x4da5] = "\xef\x70", [0x4da6] = "\xef\x6e",
  [0x4da8] = "\xef\x6b", [0x4daa] = "\xc2\x43", [0x4dab] = "\xc2\x42",
  [0x4dad] = "\xc2\x44", [0x4dae] = "\xc2\x41", [0x4daf] = "\xef\x75",
  [0x4db0] = "\xa0\x67", [0x4db5] = "\xf1\xc8", [0x4db6] = "\xf1\xcb",
  [0x4db8] = "\xf1\xc9", [0x4db9] = "\xf1\xcd", [0x4dbd] = "\xf1\xce",
  [0x4dbf] = "\xf1\xc6", [0x4dc0] = "\xc3\x58", [0x4dc1] = "\xf1\xc7",
  [0x4dc3] = "\xf1\xc5", [0x4dc4] = "\xf1\xcc", [0x4dc6] = "\xf1\xc4",
  [0x4dc7] = "\xf1\xc3", [0x4dc8] = "\xc3\x57", [0x4dc9] = "\xc3\x55",
  [0x4dca] = "\xc3\x54", [0x4dcf] = "\x96\xb3", [0x4dd3] = "\xf1\xca",
  [0x4dd4] = "\xf3\xcf", [0x4dd5] = "\xf3\xd5", [0x4dd6] = "\xc4\x4a",
  [0x4dd7] = "\xf3\xd0", [0x4dd9] = "\xf3\xd3", [0x4dda] = "\xf3\xd7",
  [0x4ddb] = "\xc4\x4b", [0x4ddc] = "\xf3\xd2", [0x4ddd] = "\x9a\x76",
  [0x4dde] = "\xf3\xca", [0x4de0] = "\xf3\xc9", [0x4de1] = "\xf3\xd6",
  [0x4de2] = "\xf3\xcd", [0x4de4] = "\xf3\xcb", [0x4de5] = "\xf3\xd4",
  [0x4de6] = "\xf3\xcc", [0x4de7] = "\xc4\x49", [0x4de8] = "\xc4\x48",
  [0x4de9] = "\x95\xd5", [0x4dea] = "\xf3\xc7", [0x4deb] = "\xf3\xc8",
  [0x4dec] = "\xf3\xd1", [0x4ded] = "\x9e\xca", [0x4df0] = "\xf3\xce",
  [0x4df1] = "\x9a\x77", [0x4df4] = "\x9a\x78", [0x4df7] = "\xf5\x6c",
  [0x4df8] = "\xf5\x6f", [0x4dfd] = "\xc3\x56", [0x4dff] = "\x91\x70",
  [0x4e02] = "\x91\x6f", [0x4e05] = "\xf5\x6d", [0x4e06] = "\xf5\x73",
  [0x4e07] = "\xf5\x71", [0x4e08] = "\xf5\x6b", [0x4e09] = "\xf5\x76",
  [0x4e0a] = "\x9f\xa3", [0x4e0b] = "\xf5\x6a", [0x4e0c] = "\x91\x71",
  [0x4e0d] = "\xc4\xcf", [0x4e0e] = "\xf5\x72", [0x4e10] = "\x96\xb1",
  [0x4e12] = "\xf5\x6e", [0x4e13] = "\xc4\xce", [0x4e14] = "\xf5\x75",
  [0x4e15] = "\x9f\x63", [0x4e17] = "\xf5\x74", [0x4e1b] = "\x9f\x67",
  [0x4e1c] = "\xf6\xab", [0x4e1d] = "\xf6\xaa", [0x4e1f] = "\x8b\xb9",
  [0x4e20] = "\x9a\x7a", [0x4e21] = "\xf6\xb1", [0x4e23] = "\xf6\xad",
  [0x4e24] = "\xf6\xb0", [0x4e25] = "\xc5\x60", [0x4e26] = "\x8b\x56",
  [0x4e28] = "\xf6\xae", [0x4e29] = "\xf6\xaf", [0x4e2b] = "\xf6\xa9",
  [0x4e2c] = "\xf6\xac", [0x4e2d] = "\xc5\x5f", [0x4e2f] = "\x9a\xda",
  [0x4e31] = "\xc5\xbf", [0x4e32] = "\xf7\xb4", [0x4e33] = "\xf7\xaf",
  [0x4e34] = "\xf7\xb3", [0x4e35] = "\x96\xb0", [0x4e36] = "\xf7\xb6",
  [0x4e37] = "\xf7\xb2", [0x4e39] = "\xf7\xae", [0x4e3a] = "\x9a\x7e",
  [0x4e3b] = "\xc5\xc1", [0x4e3c] = "\xf7\xb1", [0x4e3d] = "\xf7\xb5",
  [0x4e3e] = "\xc5\xc0", [0x4e3f] = "\xf7\xac", [0x4e40] = "\xf5\x70",
  [0x4e41] = "\xf7\xb0", [0x4e44] = "\xf7\xad", [0x4e45] = "\x9d\xde",
  [0x4e46] = "\xf7\xaa", [0x4e48] = "\xf7\xab", [0x4e49] = "\xc5\xbe",
  [0x4e4a] = "\xf8\x5a", [0x4e4b] = "\xf8\x5c", [0x4e4c] = "\xf8\x5f",
  [0x4e4d] = "\xf8\x5b", [0x4e4e] = "\xf8\x60", [0x4e4f] = "\x96\xad",
  [0x4e50] = "\xf8\x59", [0x4e52] = "\xf8\x57", [0x4e53] = "\x96\xae",
  [0x4e54] = "\xc5\xeb", [0x4e55] = "\xf8\x5d", [0x4e56] = "\xc5\xed",
  [0x4e57] = "\xc5\xec", [0x4e58] = "\xf8\x58", [0x4e59] = "\xf8\x5e",
  [0x4e5d] = "\x9e\xa1", [0x4e5e] = "\xf8\xda", [0x4e5f] = "\xc6\x4d",
  [0x4e60] = "\xf8\xdb", [0x4e62] = "\xf8\xd9", [0x4e63] = "\xf8\xd6",
  [0x4e66] = "\xf8\xd8", [0x4e67] = "\xf8\xd7", [0x4e68] = "\xf9\x5a",
  [0x4e6d] = "\xf9\x5c", [0x4e6e] = "\xf9\x5b", [0x4e71] = "\xf9\x79",
  [0x4e72] = "\x9e\x50", [0x4e73] = "\xf9\x78", [0x4e74] = "\xf9\x77",
  [0x4e75] = "\xf9\x7a", [0x4e77] = "\xc6\x73", [0x4e78] = "\xc6\x74",
  [0x4e79] = "\xf9\xca", [0x4e7a] = "\xf9\xce", [0x4e7b] = "\x96\xaf",
  [0x4e7c] = "\x8b\xf4", [0x4ee5] = "\xb3\xbe", [0x4ee6] = "\xdc\xaf",
  [0x4ee7] = "\xe0\xed", [0x4ee9] = "\xb9\xa7", [0x4eea] = "\xe0\xeb",
  [0x4eed] = "\xe0\xec", [0x4ef1] = "\xe4\xe2", [0x4ef2] = "\xe4\xe3",
  [0x4ef3] = "\xbb\xf1", [0x4ef4] = "\xbb\xef", [0x4ef5] = "\xe4\xe4",
  [0x4ef6] = "\xbb\xf0", [0x4ef7] = "\xe8\xe8", [0x4ef9] = "\xe8\xeb",
  [0x4efa] = "\xe8\xe5", [0x4efb] = "\xe8\xec", [0x4efc] = "\xe8\xe4",
  [0x4efd] = "\xe8\xe6", [0x4eff] = "\xe8\xe7", [0x4f00] = "\xe8\xea",
  [0x4f02] = "\x9f\xa4", [0x4f03] = "\xbe\xa1", [0x4f04] = "\xe8\xef",
  [0x4f05] = "\xe8\xee", [0x4f06] = "\xbe\x7d", [0x4f07] = "\xe8\xe9",
  [0x4f08] = "\xe8\xed", [0x4f09] = "\xbe\x7e", [0x4f0c] = "\x96\xbd",
  [0x4f10] = "\xec\xac", [0x4f12] = "\xc0\x6f", [0x4f14] = "\xec\xa7",
  [0x4f15] = "\xc0\x6b", [0x4f16] = "\x96\xf4", [0x4f17] = "\xec\xa4",
  [0x4f18] = "\xec\xaa", [0x4f19] = "\xec\xad", [0x4f1b] = "\xc0\x70",
  [0x4f1d] = "\xec\xa9", [0x4f1e] = "\xec\xa6", [0x4f1f] = "\xec\xae",
  [0x4f20] = "\xec\xa5", [0x4f21] = "\x96\xb8", [0x4f22] = "\xec\xab",
  [0x4f23] = "\xc0\x6c", [0x4f25] = "\xec\xa3", [0x4f26] = "\xc0\x6d",
  [0x4f28] = "\xc0\x6e", [0x4f29] = "\xec\xa8", [0x4f2d] = "\xef\xa9",
  [0x4f2e] = "\xef\x7a", [0x4f2f] = "\xef\x7b", [0x4f30] = "\xef\x7e",
  [0x4f31] = "\xef\x7c", [0x4f33] = "\xef\x76", [0x4f34] = "\xfa\xa1",
  [0x4f36] = "\xef\x79", [0x4f37] = "\xef\xa5", [0x4f38] = "\xef\x7d",
  [0x4f39] = "\x91\xa7", [0x4f3b] = "\xc2\x45", [0x4f3d] = "\xef\xa7",
  [0x4f3e] = "\xef\xa4", [0x4f3f] = "\xc2\x46", [0x4f40] = "\xef\xa6",
  [0x4f41] = "\xef\x77", [0x4f42] = "\xef\xa2", [0x4f43] = "\xef\xa3",
  [0x4f44] = "\xa0\x5e", [0x4f45] = "\xef\xa1", [0x4f49] = "\x9a\x7d",
  [0x4f4a] = "\xf1\xd2", [0x4f4b] = "\xf1\xd4", [0x4f4c] = "\xf1\xd7",
  [0x4f4e] = "\x89\x48", [0x4f4f] = "\xf1\xd1", [0x4f50] = "\x9e\xb1",
  [0x4f51] = "\xc3\x59", [0x4f52] = "\xf1\xd9", [0x4f53] = "\xf1\xd0",
  [0x4f54] = "\xf1\xda", [0x4f56] = "\xf1\xd6", [0x4f57] = "\xf1\xd8",
  [0x4f58] = "\xf1\xdc", [0x4f59] = "\xf1\xd5", [0x4f5a] = "\xf1\xdd",
  [0x4f5b] = "\xf1\xd3", [0x4f5c] = "\xf1\xcf", [0x4f5d] = "\xc3\x5a",
  [0x4f5e] = "\x9d\xdb", [0x4f5f] = "\xf1\xdb", [0x4f60] = "\xc3\x5b",
  [0x4f61] = "\xc4\x4d", [0x4f67] = "\xef\x78", [0x4f68] = "\xf3\xf1",
  [0x4f69] = "\xf3\xe8", [0x4f6a] = "\xc4\x4f", [0x4f6b] = "\xf3\xe4",
  [0x4f6c] = "\xc4\x50", [0x4f6d] = "\x95\xbf", [0x4f6e] = "\x8a\x73",
  [0x4f6f] = "\xf3\xed", [0x4f70] = "\xf3\xe7", [0x4f71] = "\xf3\xdd",
  [0x4f72] = "\xc4\x4e", [0x4f73] = "\xf3\xea", [0x4f74] = "\xf3\xe5",
  [0x4f75] = "\xf3\xe6", [0x4f77] = "\xf3\xd8", [0x4f78] = "\xf3\xdf",
  [0x4f79] = "\xf3\xee", [0x4f7b] = "\xf3\xeb", [0x4f7c] = "\x9e\xfe",
  [0x4f7d] = "\xf3\xe3", [0x4f7e] = "\x91\x7a", [0x4f7f] = "\xf3\xef",
  [0x4f80] = "\xf3\xde", [0x4f81] = "\xf3\xd9", [0x4f82] = "\xf3\xec",
  [0x4f83] = "\x91\x7b", [0x4f84] = "\xf3\xdb", [0x4f85] = "\xf3\xe9",
  [0x4f86] = "\xf3\xe0", [0x4f87] = "\xf3\xf0", [0x4f88] = "\xf3\xdc",
  [0x4f89] = "\xc4\x4c", [0x4f8a] = "\xf3\xda", [0x4f8b] = "\xf3\xe1",
  [0x4f8c] = "\xf3\xe2", [0x4f90] = "\xf5\x7d", [0x4f92] = "\xf5\x7b",
  [0x4f93] = "\x9a\xa3", [0x4f94] = "\xf5\xa2", [0x4f96] = "\xf5\xae",
  [0x4f97] = "\xf5\xa5", [0x4f98] = "\xf5\x7c", [0x4f99] = "\xf5\x78",
  [0x4f9a] = "\xf5\xa7", [0x4f9b] = "\xf5\x7e", [0x4f9c] = "\xf5\xa3",
  [0x4f9d] = "\xf5\x7a", [0x4f9e] = "\xf5\xaa", [0x4f9f] = "\xf5\x77",
  [0x4fa0] = "\xf5\xa1", [0x4fa1] = "\xf5\xa6", [0x4fa2] = "\xf5\xa8",
  [0x4fa3] = "\xf5\xab", [0x4fa4] = "\xf5\x79", [0x4fa5] = "\x96\xc2",
  [0x4fa6] = "\xf5\xaf", [0x4fa7] = "\xf5\xb0", [0x4fa8] = "\xf5\xa9",
  [0x4fa9] = "\xf5\xad", [0x4faa] = "\xf5\xa4", [0x4fab] = "\x9f\x77",
  [0x4fac] = "\xf6\xc1", [0x4fad] = "\xf6\xc4", [0x4faf] = "\xc5\x61",
  [0x4fb1] = "\xf6\xc3", [0x4fb2] = "\xf6\xc8", [0x4fb3] = "\xf6\xc6",
  [0x4fb4] = "\xc5\x62", [0x4fb5] = "\xf6\xbd", [0x4fb6] = "\xf6\xb3",
  [0x4fb7] = "\xf6\xb2", [0x4fb8] = "\xc5\x64", [0x4fb9] = "\xf6\xbf",
  [0x4fba] = "\xf6\xc0", [0x4fbb] = "\xf6\xbc", [0x4fbc] = "\xf6\xb4",
  [0x4fbd] = "\x9a\xa4", [0x4fbe] = "\xf6\xb9", [0x4fbf] = "\xf5\xac",
  [0x4fc0] = "\x9a\xa5", [0x4fc1] = "\xf6\xb5", [0x4fc2] = "\xc5\x63",
  [0x4fc3] = "\xf6\xbb", [0x4fc4] = "\x91\xa1", [0x4fc5] = "\xf6\xba",
  [0x4fc7] = "\xf6\xb6", [0x4fc8] = "\xf6\xc2", [0x4fc9] = "\x89\xb8",
  [0x4fca] = "\xf6\xb7", [0x4fcb] = "\xf7\xbb", [0x4fcc] = "\xf6\xc5",
  [0x4fcd] = "\xf6\xc7", [0x4fce] = "\xf6\xbe", [0x4fcf] = "\xf6\xb8",
  [0x4fd0] = "\xf7\xbc", [0x4fd1] = "\xf7\xbe", [0x4fd2] = "\xf7\xb8",
  [0x4fd3] = "\xc5\xc2", [0x4fd4] = "\x91\x73", [0x4fd5] = "\xf7\xc5",
  [0x4fd6] = "\xf7\xc3", [0x4fd7] = "\xc5\xc3", [0x4fd8] = "\xf7\xc2",
  [0x4fd9] = "\xf7\xc1", [0x4fda] = "\xf7\xba", [0x4fdb] = "\xf7\xb7",
  [0x4fdc] = "\xf7\xbd", [0x4fdd] = "\xf7\xc6", [0x4fde] = "\xf7\xb9",
  [0x4fdf] = "\xf7\xbf", [0x4fe1] = "\xf8\x69", [0x4fe2] = "\xf8\x6e",
  [0x4fe3] = "\xf8\x64", [0x4fe4] = "\xf8\x67", [0x4fe5] = "\xc5\xee",
  [0x4fe6] = "\xf8\x6b", [0x4fe8] = "\xf8\x72", [0x4fe9] = "\xf7\xc0",
  [0x4feb] = "\xf8\x65", [0x4fec] = "\xf8\x6f", [0x4fed] = "\xf8\x73",
  [0x4fee] = "\xf8\x6a", [0x4fef] = "\xf8\x63", [0x4ff0] = "\xf8\x6d",
  [0x4ff2] = "\xf8\x6c", [0x4ff3] = "\xf8\x71", [0x4ff4] = "\xf8\x70",
  [0x4ff5] = "\xf7\xc4", [0x4ff6] = "\xf8\x68", [0x4ff7] = "\xf8\x62",
  [0x4ff8] = "\xf8\x66", [0x4ff9] = "\xc6\x4e", [0x4ffa] = "\xc6\x4f",
  [0x4ffb] = "\xf8\x61", [0x4ffc] = "\x9a\xa6", [0x4ffd] = "\xf8\xe6",
  [0x4ffe] = "\xf8\xdd", [0x4fff] = "\xf8\xe5", [0x5000] = "\xf8\xe2",
  [0x5001] = "\xf8\xe3", [0x5002] = "\xf8\xdc", [0x5003] = "\xf8\xdf",
  [0x5004] = "\xf8\xe7", [0x5005] = "\xf8\xe1", [0x5006] = "\xf8\xe0",
  [0x5007] = "\xf8\xde", [0x5009] = "\xf8\xe4", [0x500a] = "\x89\xbd",
  [0x500b] = "\xf9\x5d", [0x500c] = "\x89\xb9", [0x500d] = "\xf9\x5e",
  [0x500e] = "\x91\x7d", [0x500f] = "\xf9\x60", [0x5010] = "\xf9\x5f",
  [0x5011] = "\xf9\x62", [0x5012] = "\xf9\x61", [0x5013] = "\xf9\x7c",
  [0x5014] = "\xf9\x7b", [0x5015] = "\xf9\xb7", [0x5017] = "\xf9\xb8",
  [0x5018] = "\x96\xbb", [0x5019] = "\xf9\xc5", [0x501a] = "\xc6\x78",
  [0x501b] = "\xc6\x7c", [0x501c] = "\x9f\xf2", [0x501d] = "\xf9\xcf",
  [0x501e] = "\xc6\x7d", [0x501f] = "\x8b\xf5", [0x5075] = "\xb3\xbf",
  [0x5079] = "\xc4\xd0", [0x507a] = "\xf6\xc9", [0x507b] = "\x9a\xa9",
  [0x507c] = "\xc6\x50", [0x507d] = "\xc6\x51", [0x507f] = "\xb3\xc0",
  [0x5080] = "\xe0\xee", [0x5081] = "\x9f\x54", [0x5082] = "\xb9\xa8",
  [0x5083] = "\xe8\xf0", [0x5084] = "\x9f\xe3", [0x5085] = "\x9e\xed",
  [0x5086] = "\xec\xb0", [0x5087] = "\xec\xb1", [0x5088] = "\xec\xaf",
  [0x5089] = "\xef\xab", [0x508a] = "\xef\xaa", [0x508b] = "\xc2\x47",
  [0x508c] = "\xf1\xdf", [0x508d] = "\xef\xac", [0x508e] = "\xf1\xde",
  [0x5090] = "\x91\xaa", [0x5091] = "\xf3\xf3", [0x5092] = "\xc4\x51",
  [0x5093] = "\xc4\x53", [0x5094] = "\xf3\xf2", [0x5095] = "\x91\xab",
  [0x5096] = "\xa0\x70", [0x5097] = "\xc4\x52", [0x5098] = "\x9f\x6d",
  [0x5099] = "\xf5\xb1", [0x509a] = "\xf5\xb3", [0x509b] = "\xf5\xb2",
  [0x509c] = "\xf6\xca", [0x509d] = "\xc5\x65", [0x509e] = "\x91\xac",
  [0x509f] = "\xc5\xef", [0x50a0] = "\xf8\xe8", [0x50a1] = "\xf9\x63",
  [0x50a2] = "\x91\xad", [0x50a4] = "\xf9\xd2", [0x50a5] = "\xb3\xc1",
  [0x50a6] = "\xa0\xfd", [0x50a7] = "\xe4\xe5", [0x50a8] = "\x9f\xe2",
  [0x50a9] = "\xbe\xa2", [0x50aa] = "\x91\xaf", [0x50ab] = "\x9e\x41",
  [0x50ac] = "\x9a\xaa", [0x50ad] = "\xec\xb3", [0x50ae] = "\xec\xb2",
  [0x50af] = "\x91\xb0", [0x50b0] = "\xef\xad", [0x50b1] = "\x9a\xab",
  [0x50b4] = "\xc4\x54", [0x50b5] = "\xc4\xd1", [0x50b6] = "\xf7\xc7",
  [0x50b7] = "\xf9\xcb", [0x50bb] = "\xb3\xc2", [0x50bc] = "\xbb\xf2",
  [0x50bd] = "\x9a\xac", [0x50be] = "\xbe\xa3", [0x50bf] = "\x9a\x4a",
  [0x50c0] = "\xf3\xf4", [0x50c1] = "\x91\xb2", [0x50c2] = "\xf8\x74",
  [0x50c3] = "\xb6\xc0", [0x50c4] = "\x8b\xf6", [0x50c6] = "\x9a\xad",
  [0x50c7] = "\x89\xb6", [0x50c8] = "\xef\xae", [0x50cc] = "\xc6\x64",
  [0x50cd] = "\xb6\xc1", [0x50ce] = "\xbe\xa4", [0x50cf] = "\xc2\x48",
  [0x50d0] = "\xf8\x75", [0x50d1] = "\xb6\xc2", [0x50d3] = "\xe8\xf1",
  [0x50d4] = "\xc0\x72", [0x50d5] = "\xec\xb4", [0x50d6] = "\xec\xb5",
  [0x50d8] = "\xc0\x71", [0x50da] = "\xef\xaf", [0x50db] = "\xc2\x4c",
  [0x50dc] = "\xc2\x4a", [0x50dd] = "\xc2\x4b", [0x50de] = "\xc2\x49",
  [0x50df] = "\xf1\xe0", [0x50e0] = "\xc3\x5c", [0x50e2] = "\x9a\xaf",
  [0x50e4] = "\xf5\xb5", [0x50e5] = "\xf5\xb4", [0x50e6] = "\xf5\xb7",
  [0x50e7] = "\xf5\xb6", [0x50e8] = "\xc4\xd2", [0x50eb] = "\xf6\xcb",
  [0x50ed] = "\xf6\xcd", [0x50ee] = "\xf6\xcc", [0x50ef] = "\xc5\x66",
  [0x50f0] = "\xf7\xc8", [0x50f1] = "\x9a\xb0", [0x50f2] = "\xf8\x76",
  [0x50f3] = "\xf8\x77", [0x50f4] = "\xc5\xf0", [0x50f5] = "\xf9\x64",
  [0x50f6] = "\xf9\x7d", [0x50f7] = "\xc6\x75", [0x50f8] = "\x9a\xb1",
  [0x50f9] = "\xdc\xb0", [0x50fa] = "\xec\xb6", [0x50fb] = "\xef\xb0",
  [0x50fc] = "\xf3\xf5", [0x50fd] = "\xe0\xef", [0x50fe] = "\x9a\xa1",
  [0x50ff] = "\xef\xb1", [0x5100] = "\xf1\xe2", [0x5101] = "\xf1\xe1",
  [0x5102] = "\x91\xb9", [0x5106] = "\xf8\x78", [0x5107] = "\xc6\x52",
  [0x5108] = "\x91\xba", [0x5109] = "\xf9\x65", [0x510a] = "\xf9\x7e",
  [0x510e] = "\xb9\xa9", [0x510f] = "\xe8\xf2", [0x5110] = "\xe8\xf3",
  [0x5112] = "\xec\xb7", [0x5113] = "\xb9\xaa", [0x5115] = "\xc3\x5d",
  [0x5116] = "\xf1\xe3", [0x5117] = "\x91\xbe", [0x5118] = "\xf6\xcf",
  [0x5119] = "\xc5\x67", [0x511a] = "\xf6\xd0", [0x511b] = "\xf6\xce",
  [0x511c] = "\xf8\x79", [0x511e] = "\xf8\xe9", [0x5120] = "\xb9\xab",
  [0x5122] = "\xef\xb4", [0x5123] = "\xef\xb3", [0x5124] = "\xef\xb2",
  [0x5125] = "\xf1\xe4", [0x5126] = "\xa0\x41", [0x5127] = "\x8b\xb7",
  [0x5128] = "\xf1\xe8", [0x5129] = "\xf1\xe7", [0x512a] = "\xf1\xe6",
  [0x512b] = "\xf1\xe5", [0x512c] = "\xc3\x5e", [0x512d] = "\xf3\xf6",
  [0x512e] = "\xf5\xb9", [0x512f] = "\xc4\xd3", [0x5130] = "\xf5\xb8",
  [0x5131] = "\xf6\xd1", [0x5132] = "\xf7\xcb", [0x5133] = "\xf7\xca",
  [0x5134] = "\xc5\xc4", [0x5135] = "\xf7\xc9", [0x5136] = "\xf8\x7c",
  [0x5137] = "\xf8\x7b", [0x5138] = "\xf8\x7a", [0x5139] = "\x91\xc0",
  [0x513b] = "\xbb\xf3", [0x513d] = "\xec\xb8", [0x513e] = "\xc2\x4d",
  [0x5140] = "\xf3\xf7", [0x5141] = "\xf3\xf8", [0x5142] = "\xf7\xcc",
  [0x5143] = "\xf8\x7d", [0x5144] = "\x9a\xb3", [0x5145] = "\x91\xc3",
  [0x5146] = "\xf8\xea", [0x5147] = "\xf9\x66", [0x5148] = "\xf9\xb9",
  [0x5149] = "\xf9\xd4", [0x514a] = "\xbb\xf4", [0x514b] = "\xc2\x4e",
  [0x514c] = "\xf1\xe9", [0x514d] = "\xf3\xf9", [0x514e] = "\xf6\xd2",
  [0x514f] = "\xf8\x7e", [0x5150] = "\xa0\xfc", [0x5152] = "\xbe\xa6",
  [0x5153] = "\x9f\xee", [0x5154] = "\xef\xb5", [0x5155] = "\xf1\xea",
  [0x5156] = "\xf3\xfa", [0x5157] = "\xf3\xfb", [0x5158] = "\xf3\xfc",
  [0x5159] = "\xf5\xbe", [0x515a] = "\x9f\x69", [0x515b] = "\xf5\xba",
  [0x515c] = "\xc5\x68", [0x515d] = "\xf5\xbd", [0x515e] = "\xf5\xbc",
  [0x515f] = "\xc4\xd4", [0x5160] = "\xf5\xbb", [0x5161] = "\xc4\xd6",
  [0x5162] = "\x91\xc8", [0x5163] = "\xc4\xd5", [0x5164] = "\xf6\xd4",
  [0x5165] = "\xf6\xd3", [0x5166] = "\xc5\x69", [0x5167] = "\xc5\x6a",
  [0x5169] = "\x91\xc9", [0x516a] = "\xc5\xc6", [0x516b] = "\xf7\xcd",
  [0x516c] = "\xc5\xc5", [0x516e] = "\xf8\xa3", [0x516f] = "\xf8\xa4",
  [0x5170] = "\xf8\xa2", [0x5171] = "\xf8\xa1", [0x5172] = "\xc6\x54",
  [0x5174] = "\xf8\xeb", [0x5175] = "\xf8\xec", [0x5176] = "\xf8\xed",
  [0x5177] = "\xc6\x53", [0x5178] = "\xf9\x67", [0x5179] = "\xf9\x6a",
  [0x517a] = "\xf9\x69", [0x517b] = "\xf9\x68", [0x517e] = "\xf9\xd3",
  [0x517f] = "\x8d\xe6", [0x518d] = "\xc0\x73", [0x518e] = "\x91\xcb",
  [0x5190] = "\xc3\x65", [0x5191] = "\xf5\xbf", [0x5192] = "\xf6\xd5",
  [0x5194] = "\xc5\xc7", [0x5195] = "\xf7\xce", [0x5198] = "\xf9\xd5",
  [0x5199] = "\x89\xc8", [0x519c] = "\x8b\xf8", [0x519f] = "\x8d\xaa",
  [0x51a0] = "\xef\xb6", [0x51a2] = "\xf7\xcf", [0x51a4] = "\xf9\xa1",
};

/* CJK compatibility forms.
   The table can be created using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0xfe' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0xfe30,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab12[][2] =
{
  [0x0000] = "\xa1\x4a", [0x0001] = "\xa1\x57", [0x0003] = "\xa1\x59",
  [0x0004] = "\xa1\x5b", [0x0005] = "\xa1\x5f", [0x0006] = "\xa1\x60",
  [0x0007] = "\xa1\x63", [0x0008] = "\xa1\x64", [0x0009] = "\xa1\x67",
  [0x000a] = "\xa1\x68", [0x000b] = "\xa1\x6b", [0x000c] = "\xa1\x6c",
  [0x000d] = "\xa1\x6f", [0x000e] = "\xa1\x70", [0x000f] = "\xa1\x73",
  [0x0010] = "\xa1\x74", [0x0011] = "\xa1\x77", [0x0012] = "\xa1\x78",
  [0x0013] = "\xa1\x7b", [0x0014] = "\xa1\x7c", [0x0019] = "\xa1\xc6",
  [0x001a] = "\xa1\xc7", [0x001b] = "\xa1\xca", [0x001c] = "\xa1\xcb",
  [0x001d] = "\xa1\xc8", [0x001e] = "\xa1\xc9", [0x001f] = "\xa1\x5c",
  [0x0020] = "\xa1\x4d", [0x0022] = "\xa1\x4f", [0x0024] = "\xa1\x51",
  [0x0025] = "\xa1\x52", [0x0026] = "\xa1\x53", [0x0027] = "\xa1\x54",
  [0x0029] = "\xa1\x7d", [0x002a] = "\xa1\x7e", [0x002b] = "\xa1\xa1",
  [0x002c] = "\xa1\xa2", [0x002d] = "\xa1\xa3", [0x002e] = "\xa1\xa4",
  [0x002f] = "\xa1\xcc", [0x0030] = "\xa1\xcd", [0x0031] = "\xa1\xce",
  [0x0032] = "\xa1\xde", [0x0033] = "\xa1\xdf", [0x0034] = "\xa1\xe0",
  [0x0035] = "\xa1\xe1", [0x0036] = "\xa1\xe2", [0x0039] = "\xa2\x4c",
  [0x003a] = "\xa2\x4d", [0x003b] = "\xa2\x4e",
};

/* Fullwidth ASCII.
   The table can be created using

   egrep '^0x'  /mnt/cdrom/unix/mappings/eastasia/other/big5.txt |
   grep -v 'NO MAPPING' | awk '{ print $2, $1 }' | sort |
   egrep '^0xff' | perl tab.pl

   where tab.pl is

   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   $n=0;
   while (<>) {
     local($ucs4, $big5, %rest) = split;
     local($u)=hex($ucs4);
     local($b)=hex($big5);
     printf ("\n ") if ($n % 3 == 0);
     ++$n;
     printf (" [0x%04x] = \"\\x%02x\\x%02x\",", $u - 0xff01,
	     $b < 0x100 ? $b : int($b/256), $b < 0x100 ? 0 : $b&255);
   }
   printf ("\n");
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
*/
static const char from_ucs4_tab13[][2] =
{
  [0x0000] = "\xa1\x49", [0x0002] = "\xa1\xad", [0x0003] = "\xa2\x43",
  [0x0004] = "\xa2\x48", [0x0005] = "\xa1\xae", [0x0007] = "\xa1\x5d",
  [0x0008] = "\xa1\x5e", [0x0009] = "\xa1\xaf", [0x000a] = "\xa1\xcf",
  [0x000b] = "\xa1\x41", [0x000c] = "\xa1\xd0", [0x000d] = "\xa1\x44",
  [0x000e] = "\xa2\x41", [0x000f] = "\xa2\xaf", [0x0010] = "\xa2\xb0",
  [0x0011] = "\xa2\xb1", [0x0012] = "\xa2\xb2", [0x0013] = "\xa2\xb3",
  [0x0014] = "\xa2\xb4", [0x0015] = "\xa2\xb5", [0x0016] = "\xa2\xb6",
  [0x0017] = "\xa2\xb7", [0x0018] = "\xa2\xb8", [0x0019] = "\xa1\x47",
  [0x001a] = "\xa1\x46", [0x001b] = "\xa1\xd5", [0x001c] = "\xa1\xd7",
  [0x001d] = "\xa1\xd6", [0x001e] = "\xa1\x48", [0x001f] = "\xa2\x49",
  [0x0020] = "\xa2\xcf", [0x0021] = "\xa2\xd0", [0x0022] = "\xa2\xd1",
  [0x0023] = "\xa2\xd2", [0x0024] = "\xa2\xd3", [0x0025] = "\xa2\xd4",
  [0x0026] = "\xa2\xd5", [0x0027] = "\xa2\xd6", [0x0028] = "\xa2\xd7",
  [0x0029] = "\xa2\xd8", [0x002a] = "\xa2\xd9", [0x002b] = "\xa2\xda",
  [0x002c] = "\xa2\xdb", [0x002d] = "\xa2\xdc", [0x002e] = "\xa2\xdd",
  [0x002f] = "\xa2\xde", [0x0030] = "\xa2\xdf", [0x0031] = "\xa2\xe0",
  [0x0032] = "\xa2\xe1", [0x0033] = "\xa2\xe2", [0x0034] = "\xa2\xe3",
  [0x0035] = "\xa2\xe4", [0x0036] = "\xa2\xe5", [0x0037] = "\xa2\xe6",
  [0x0038] = "\xa2\xe7", [0x0039] = "\xa2\xe8", [0x003b] = "\xa2\x42",
  [0x003e] = "\xa1\xc4", [0x0040] = "\xa2\xe9", [0x0041] = "\xa2\xea",
  [0x0042] = "\xa2\xeb", [0x0043] = "\xa2\xec", [0x0044] = "\xa2\xed",
  [0x0045] = "\xa2\xee", [0x0046] = "\xa2\xef", [0x0047] = "\xa2\xf0",
  [0x0048] = "\xa2\xf1", [0x0049] = "\xa2\xf2", [0x004a] = "\xa2\xf3",
  [0x004b] = "\xa2\xf4", [0x004c] = "\xa2\xf5", [0x004d] = "\xa2\xf6",
  [0x004e] = "\xa2\xf7", [0x004f] = "\xa2\xf8", [0x0050] = "\xa2\xf9",
  [0x0051] = "\xa2\xfa", [0x0052] = "\xa2\xfb", [0x0053] = "\xa2\xfc",
  [0x0054] = "\xa2\xfd", [0x0055] = "\xa2\xfe", [0x0056] = "\xa3\x40",
  [0x0057] = "\xa3\x41", [0x0058] = "\xa3\x42", [0x0059] = "\xa3\x43",
  [0x005a] = "\xa1\x61", [0x005b] = "\xa1\x55", [0x005c] = "\xa1\x62",
  [0x0063] = "\xa1\x4e",
};

/* ISO10646 PUA back to Big5-HKSCS, including compatibility code points.
   Generated from the table available at
   http://www.digital21.gov.hk/eng/hkscs/download.html

   Roger So <spacehunt@e-fever.org>  07/01/2000
*/
static const char from_ucs4_tab14[][2] =
{
  [0x0000] = "\xfa\x40", [0x0001] = "\xfa\x41", [0x0002] = "\xfa\x42",
  [0x0003] = "\xfa\x43", [0x0004] = "\xfa\x44", [0x0005] = "\xfa\x45",
  [0x0006] = "\xfa\x46", [0x0007] = "\xfa\x47", [0x0008] = "\xfa\x48",
  [0x0009] = "\xfa\x49", [0x000a] = "\xfa\x4a", [0x000b] = "\xfa\x4b",
  [0x000c] = "\xfa\x4c", [0x000d] = "\xfa\x4d", [0x000e] = "\xfa\x4e",
  [0x000f] = "\xfa\x4f", [0x0010] = "\xfa\x50", [0x0011] = "\xfa\x51",
  [0x0012] = "\xfa\x52", [0x0013] = "\xfa\x53", [0x0014] = "\xfa\x54",
  [0x0015] = "\xfa\x55", [0x0016] = "\xfa\x56", [0x0017] = "\xfa\x57",
  [0x0018] = "\xfa\x58", [0x0019] = "\xfa\x59", [0x001a] = "\xfa\x5a",
  [0x001b] = "\xfa\x5b", [0x001c] = "\xfa\x5c", [0x001d] = "\xfa\x5d",
  [0x001e] = "\xfa\x5e", [0x001f] = "\xfa\x5f", [0x0020] = "\xfa\x60",
  [0x0021] = "\xfa\x61", [0x0022] = "\xfa\x62", [0x0023] = "\xfa\x63",
  [0x0024] = "\xfa\x64", [0x0025] = "\xfa\x65", [0x0026] = "\xfa\x66",
  [0x0027] = "\xfa\x67", [0x0028] = "\xfa\x68", [0x0029] = "\xfa\x69",
  [0x002a] = "\xfa\x6a", [0x002b] = "\xfa\x6b", [0x002c] = "\xfa\x6c",
  [0x002d] = "\xfa\x6d", [0x002e] = "\xfa\x6e", [0x002f] = "\xfa\x6f",
  [0x0030] = "\xfa\x70", [0x0031] = "\xfa\x71", [0x0032] = "\xfa\x72",
  [0x0033] = "\xfa\x73", [0x0034] = "\xfa\x74", [0x0035] = "\xfa\x75",
  [0x0036] = "\xfa\x76", [0x0037] = "\xfa\x77", [0x0038] = "\xfa\x78",
  [0x0039] = "\xfa\x79", [0x003a] = "\xfa\x7a", [0x003b] = "\xfa\x7b",
  [0x003c] = "\xfa\x7c", [0x003d] = "\xfa\x7d", [0x003e] = "\xfa\x7e",
  [0x003f] = "\xfa\xa1", [0x0040] = "\xfa\xa2", [0x0041] = "\xfa\xa3",
  [0x0042] = "\xfa\xa4", [0x0043] = "\xfa\xa5", [0x0044] = "\xfa\xa6",
  [0x0045] = "\xfa\xa7", [0x0046] = "\xfa\xa8", [0x0047] = "\xfa\xa9",
  [0x0048] = "\xfa\xaa", [0x0049] = "\xfa\xab", [0x004a] = "\xfa\xac",
  [0x004b] = "\xfa\xad", [0x004c] = "\xfa\xae", [0x004d] = "\xfa\xaf",
  [0x004e] = "\xfa\xb0", [0x004f] = "\xfa\xb1", [0x0050] = "\xfa\xb2",
  [0x0051] = "\xfa\xb3", [0x0052] = "\xfa\xb4", [0x0053] = "\xfa\xb5",
  [0x0054] = "\xfa\xb6", [0x0055] = "\xfa\xb7", [0x0056] = "\xfa\xb8",
  [0x0057] = "\xfa\xb9", [0x0058] = "\xfa\xba", [0x0059] = "\xfa\xbb",
  [0x005a] = "\xfa\xbc", [0x005b] = "\xfa\xbd", [0x005c] = "\xfa\xbe",
  [0x005d] = "\xfa\xbf", [0x005e] = "\xfa\xc0", [0x005f] = "\xfa\xc1",
  [0x0060] = "\xfa\xc2", [0x0061] = "\xfa\xc3", [0x0062] = "\xfa\xc4",
  [0x0063] = "\xfa\xc5", [0x0064] = "\xfa\xc6", [0x0065] = "\xfa\xc7",
  [0x0066] = "\xfa\xc8", [0x0067] = "\xfa\xc9", [0x0068] = "\xfa\xca",
  [0x0069] = "\xfa\xcb", [0x006a] = "\xfa\xcc", [0x006b] = "\xfa\xcd",
  [0x006c] = "\xfa\xce", [0x006d] = "\xfa\xcf", [0x006e] = "\xfa\xd0",
  [0x006f] = "\xfa\xd1", [0x0070] = "\xfa\xd2", [0x0071] = "\xfa\xd3",
  [0x0072] = "\xfa\xd4", [0x0073] = "\xfa\xd5", [0x0074] = "\xfa\xd6",
  [0x0075] = "\xfa\xd7", [0x0076] = "\xfa\xd8", [0x0077] = "\xfa\xd9",
  [0x0078] = "\xfa\xda", [0x0079] = "\xfa\xdb", [0x007a] = "\xfa\xdc",
  [0x007b] = "\xfa\xdd", [0x007c] = "\xfa\xde", [0x007d] = "\xfa\xdf",
  [0x007e] = "\xfa\xe0", [0x007f] = "\xfa\xe1", [0x0080] = "\xfa\xe2",
  [0x0081] = "\xfa\xe3", [0x0082] = "\xfa\xe4", [0x0083] = "\xfa\xe5",
  [0x0084] = "\xfa\xe6", [0x0085] = "\xfa\xe7", [0x0086] = "\xfa\xe8",
  [0x0087] = "\xfa\xe9", [0x0088] = "\xfa\xea", [0x0089] = "\xfa\xeb",
  [0x008a] = "\xfa\xec", [0x008b] = "\xfa\xed", [0x008c] = "\xfa\xee",
  [0x008d] = "\xfa\xef", [0x008e] = "\xfa\xf0", [0x008f] = "\xfa\xf1",
  [0x0090] = "\xfa\xf2", [0x0091] = "\xfa\xf3", [0x0092] = "\xfa\xf4",
  [0x0093] = "\xfa\xf5", [0x0094] = "\xfa\xf6", [0x0095] = "\xfa\xf7",
  [0x0096] = "\xfa\xf8", [0x0097] = "\xfa\xf9", [0x0098] = "\xfa\xfa",
  [0x0099] = "\xfa\xfb", [0x009a] = "\xfa\xfc", [0x009b] = "\xfa\xfd",
  [0x009c] = "\xfa\xfe", [0x009d] = "\xfb\x40", [0x009e] = "\xfb\x41",
  [0x009f] = "\xfb\x42", [0x00a0] = "\xfb\x43", [0x00a1] = "\xfb\x44",
  [0x00a2] = "\xfb\x45", [0x00a3] = "\xfb\x46", [0x00a4] = "\xfb\x47",
  [0x00a5] = "\xfb\x48", [0x00a6] = "\xfb\x49", [0x00a7] = "\xfb\x4a",
  [0x00a8] = "\xfb\x4b", [0x00a9] = "\xfb\x4c", [0x00aa] = "\xfb\x4d",
  [0x00ab] = "\xfb\x4e", [0x00ac] = "\xfb\x4f", [0x00ad] = "\xfb\x50",
  [0x00ae] = "\xfb\x51", [0x00af] = "\xfb\x52", [0x00b0] = "\xfb\x53",
  [0x00b1] = "\xfb\x54", [0x00b2] = "\xfb\x55", [0x00b3] = "\xfb\x56",
  [0x00b4] = "\xfb\x57", [0x00b5] = "\xfb\x58", [0x00b6] = "\xfb\x59",
  [0x00b7] = "\xfb\x5a", [0x00b8] = "\xfb\x5b", [0x00b9] = "\xfb\x5c",
  [0x00ba] = "\xfb\x5d", [0x00bb] = "\xfb\x5e", [0x00bc] = "\xfb\x5f",
  [0x00bd] = "\xfb\x60", [0x00be] = "\xfb\x61", [0x00bf] = "\xfb\x62",
  [0x00c0] = "\xfb\x63", [0x00c1] = "\xfb\x64", [0x00c2] = "\xfb\x65",
  [0x00c3] = "\xfb\x66", [0x00c4] = "\xfb\x67", [0x00c5] = "\xfb\x68",
  [0x00c6] = "\xfb\x69", [0x00c7] = "\xfb\x6a", [0x00c8] = "\xfb\x6b",
  [0x00c9] = "\xfb\x6c", [0x00ca] = "\xfb\x6d", [0x00cb] = "\xfb\x6e",
  [0x00cc] = "\xfb\x6f", [0x00cd] = "\xfb\x70", [0x00ce] = "\xfb\x71",
  [0x00cf] = "\xfb\x72", [0x00d0] = "\xfb\x73", [0x00d1] = "\xfb\x74",
  [0x00d2] = "\xfb\x75", [0x00d3] = "\xfb\x76", [0x00d4] = "\xfb\x77",
  [0x00d5] = "\xfb\x78", [0x00d6] = "\xfb\x79", [0x00d7] = "\xfb\x7a",
  [0x00d8] = "\xfb\x7b", [0x00d9] = "\xfb\x7c", [0x00da] = "\xfb\x7d",
  [0x00db] = "\xfb\x7e", [0x00dc] = "\xfb\xa1", [0x00dd] = "\xfb\xa2",
  [0x00de] = "\xfb\xa3", [0x00df] = "\xfb\xa4", [0x00e0] = "\xfb\xa5",
  [0x00e1] = "\xfb\xa6", [0x00e2] = "\xfb\xa7", [0x00e3] = "\xfb\xa8",
  [0x00e4] = "\xfb\xa9", [0x00e5] = "\xfb\xaa", [0x00e6] = "\xfb\xab",
  [0x00e7] = "\xfb\xac", [0x00e8] = "\xfb\xad", [0x00e9] = "\xfb\xae",
  [0x00ea] = "\xfb\xaf", [0x00eb] = "\xfb\xb0", [0x00ec] = "\xfb\xb1",
  [0x00ed] = "\xfb\xb2", [0x00ee] = "\xfb\xb3", [0x00ef] = "\xfb\xb4",
  [0x00f0] = "\xfb\xb5", [0x00f1] = "\xfb\xb6", [0x00f2] = "\xfb\xb7",
  [0x00f3] = "\xfb\xb8", [0x00f4] = "\xfb\xb9", [0x00f5] = "\xfb\xba",
  [0x00f6] = "\xfb\xbb", [0x00f7] = "\xfb\xbc", [0x00f8] = "\xfb\xbd",
  [0x00f9] = "\xfb\xbe", [0x00fa] = "\xfb\xbf", [0x00fb] = "\xfb\xc0",
  [0x00fc] = "\xfb\xc1", [0x00fd] = "\xfb\xc2", [0x00fe] = "\xfb\xc3",
  [0x00ff] = "\xfb\xc4", [0x0100] = "\xfb\xc5", [0x0101] = "\xfb\xc6",
  [0x0102] = "\xfb\xc7", [0x0103] = "\xfb\xc8", [0x0104] = "\xfb\xc9",
  [0x0105] = "\xfb\xca", [0x0106] = "\xfb\xcb", [0x0107] = "\xfb\xcc",
  [0x0108] = "\xfb\xcd", [0x0109] = "\xfb\xce", [0x010a] = "\xfb\xcf",
  [0x010b] = "\xfb\xd0", [0x010c] = "\xfb\xd1", [0x010d] = "\xfb\xd2",
  [0x010e] = "\xfb\xd3", [0x010f] = "\xfb\xd4", [0x0110] = "\xfb\xd5",
  [0x0111] = "\xfb\xd6", [0x0112] = "\xfb\xd7", [0x0113] = "\xfb\xd8",
  [0x0114] = "\xfb\xd9", [0x0115] = "\xfb\xda", [0x0116] = "\xfb\xdb",
  [0x0117] = "\xfb\xdc", [0x0118] = "\xfb\xdd", [0x0119] = "\xfb\xde",
  [0x011a] = "\xfb\xdf", [0x011b] = "\xfb\xe0", [0x011c] = "\xfb\xe1",
  [0x011d] = "\xfb\xe2", [0x011e] = "\xfb\xe3", [0x011f] = "\xfb\xe4",
  [0x0120] = "\xfb\xe5", [0x0121] = "\xfb\xe6", [0x0122] = "\xfb\xe7",
  [0x0123] = "\xfb\xe8", [0x0124] = "\xfb\xe9", [0x0125] = "\xfb\xea",
  [0x0126] = "\xfb\xeb", [0x0127] = "\xfb\xec", [0x0128] = "\xfb\xed",
  [0x0129] = "\xfb\xee", [0x012a] = "\xfb\xef", [0x012b] = "\xfb\xf0",
  [0x012c] = "\xfb\xf1", [0x012d] = "\xfb\xf2", [0x012e] = "\xfb\xf3",
  [0x012f] = "\xfb\xf4", [0x0130] = "\xfb\xf5", [0x0131] = "\xfb\xf6",
  [0x0132] = "\xfb\xf7", [0x0133] = "\xfb\xf8", [0x0134] = "\xfb\xf9",
  [0x0135] = "\xfb\xfa", [0x0136] = "\xfb\xfb", [0x0137] = "\xfb\xfc",
  [0x0138] = "\xfb\xfd", [0x0139] = "\xfb\xfe", [0x013a] = "\xfc\x40",
  [0x013b] = "\xfc\x41", [0x013c] = "\xfc\x42", [0x013d] = "\xfc\x43",
  [0x013e] = "\xfc\x44", [0x013f] = "\xfc\x45", [0x0140] = "\xfc\x46",
  [0x0141] = "\xfc\x47", [0x0142] = "\xfc\x48", [0x0143] = "\xfc\x49",
  [0x0144] = "\xfc\x4a", [0x0145] = "\xfc\x4b", [0x0146] = "\xfc\x4c",
  [0x0147] = "\xfc\x4d", [0x0148] = "\xfc\x4e", [0x0149] = "\xfc\x4f",
  [0x014a] = "\xfc\x50", [0x014b] = "\xfc\x51", [0x014c] = "\xfc\x52",
  [0x014d] = "\xfc\x53", [0x014e] = "\xfc\x54", [0x014f] = "\xfc\x55",
  [0x0150] = "\xfc\x56", [0x0151] = "\xfc\x57", [0x0152] = "\xfc\x58",
  [0x0153] = "\xfc\x59", [0x0154] = "\xfc\x5a", [0x0155] = "\xfc\x5b",
  [0x0156] = "\xfc\x5c", [0x0157] = "\xfc\x5d", [0x0158] = "\xfc\x5e",
  [0x0159] = "\xfc\x5f", [0x015a] = "\xfc\x60", [0x015b] = "\xfc\x61",
  [0x015c] = "\xfc\x62", [0x015d] = "\xfc\x63", [0x015e] = "\xfc\x64",
  [0x015f] = "\xfc\x65", [0x0160] = "\xfc\x66", [0x0161] = "\xfc\x67",
  [0x0162] = "\xfc\x68", [0x0163] = "\xfc\x69", [0x0164] = "\xfc\x6a",
  [0x0165] = "\xfc\x6b", [0x0166] = "\xfc\x6c", [0x0167] = "\xfc\x6d",
  [0x0168] = "\xfc\x6e", [0x0169] = "\xfc\x6f", [0x016a] = "\xfc\x70",
  [0x016b] = "\xfc\x71", [0x016c] = "\xfc\x72", [0x016d] = "\xfc\x73",
  [0x016e] = "\xfc\x74", [0x016f] = "\xfc\x75", [0x0170] = "\xfc\x76",
  [0x0171] = "\xfc\x77", [0x0172] = "\xfc\x78", [0x0173] = "\xfc\x79",
  [0x0174] = "\xfc\x7a", [0x0175] = "\xfc\x7b", [0x0176] = "\xfc\x7c",
  [0x0177] = "\xfc\x7d", [0x0178] = "\xfc\x7e", [0x0179] = "\xfc\xa1",
  [0x017a] = "\xfc\xa2", [0x017b] = "\xfc\xa3", [0x017c] = "\xfc\xa4",
  [0x017d] = "\xfc\xa5", [0x017e] = "\xfc\xa6", [0x017f] = "\xfc\xa7",
  [0x0180] = "\xfc\xa8", [0x0181] = "\xfc\xa9", [0x0182] = "\xfc\xaa",
  [0x0183] = "\xfc\xab", [0x0184] = "\xfc\xac", [0x0185] = "\xfc\xad",
  [0x0186] = "\xfc\xae", [0x0187] = "\xfc\xaf", [0x0188] = "\xfc\xb0",
  [0x0189] = "\xfc\xb1", [0x018a] = "\xfc\xb2", [0x018b] = "\xfc\xb3",
  [0x018c] = "\xfc\xb4", [0x018d] = "\xfc\xb5", [0x018e] = "\xfc\xb6",
  [0x018f] = "\xfc\xb7", [0x0190] = "\xfc\xb8", [0x0191] = "\xfc\xb9",
  [0x0192] = "\xfc\xba", [0x0193] = "\xfc\xbb", [0x0194] = "\xfc\xbc",
  [0x0195] = "\xfc\xbd", [0x0196] = "\xfc\xbe", [0x0197] = "\xfc\xbf",
  [0x0198] = "\xfc\xc0", [0x0199] = "\xfc\xc1", [0x019a] = "\xfc\xc2",
  [0x019b] = "\xfc\xc3", [0x019c] = "\xfc\xc4", [0x019d] = "\xfc\xc5",
  [0x019e] = "\xfc\xc6", [0x019f] = "\xfc\xc7", [0x01a0] = "\xfc\xc8",
  [0x01a1] = "\xfc\xc9", [0x01a2] = "\xfc\xca", [0x01a3] = "\xfc\xcb",
  [0x01a4] = "\xfc\xcc", [0x01a5] = "\xfc\xcd", [0x01a6] = "\xfc\xce",
  [0x01a7] = "\xfc\xcf", [0x01a8] = "\xfc\xd0", [0x01a9] = "\xfc\xd1",
  [0x01aa] = "\xfc\xd2", [0x01ab] = "\xfc\xd3", [0x01ac] = "\xfc\xd4",
  [0x01ad] = "\xfc\xd5", [0x01ae] = "\xfc\xd6", [0x01af] = "\xfc\xd7",
  [0x01b0] = "\xfc\xd8", [0x01b1] = "\xfc\xd9", [0x01b2] = "\xfc\xda",
  [0x01b3] = "\xfc\xdb", [0x01b4] = "\xfc\xdc", [0x01b5] = "\xfc\xdd",
  [0x01b6] = "\xfc\xde", [0x01b7] = "\xfc\xdf", [0x01b8] = "\xfc\xe0",
  [0x01b9] = "\xfc\xe1", [0x01ba] = "\xfc\xe2", [0x01bb] = "\xfc\xe3",
  [0x01bc] = "\xfc\xe4", [0x01bd] = "\xfc\xe5", [0x01be] = "\xfc\xe6",
  [0x01bf] = "\xfc\xe7", [0x01c0] = "\xfc\xe8", [0x01c1] = "\xfc\xe9",
  [0x01c2] = "\xfc\xea", [0x01c3] = "\xfc\xeb", [0x01c4] = "\xfc\xec",
  [0x01c5] = "\xfc\xed", [0x01c6] = "\xfc\xee", [0x01c7] = "\xfc\xef",
  [0x01c8] = "\xfc\xf0", [0x01c9] = "\xfc\xf1", [0x01ca] = "\xfc\xf2",
  [0x01cb] = "\xfc\xf3", [0x01cc] = "\xfc\xf4", [0x01cd] = "\xfc\xf5",
  [0x01ce] = "\xfc\xf6", [0x01cf] = "\xfc\xf7", [0x01d0] = "\xfc\xf8",
  [0x01d1] = "\xfc\xf9", [0x01d2] = "\xfc\xfa", [0x01d3] = "\xfc\xfb",
  [0x01d4] = "\xfc\xfc", [0x01d5] = "\xfc\xfd", [0x01d6] = "\xfc\xfe",
  [0x01d7] = "\xfd\x40", [0x01d8] = "\xfd\x41", [0x01d9] = "\xfd\x42",
  [0x01da] = "\xfd\x43", [0x01db] = "\xfd\x44", [0x01dc] = "\xfd\x45",
  [0x01dd] = "\xfd\x46", [0x01de] = "\xfd\x47", [0x01df] = "\xfd\x48",
  [0x01e0] = "\xfd\x49", [0x01e1] = "\xfd\x4a", [0x01e2] = "\xfd\x4b",
  [0x01e3] = "\xfd\x4c", [0x01e4] = "\xfd\x4d", [0x01e5] = "\xfd\x4e",
  [0x01e6] = "\xfd\x4f", [0x01e7] = "\xfd\x50", [0x01e8] = "\xfd\x51",
  [0x01e9] = "\xfd\x52", [0x01ea] = "\xfd\x53", [0x01eb] = "\xfd\x54",
  [0x01ec] = "\xfd\x55", [0x01ed] = "\xfd\x56", [0x01ee] = "\xfd\x57",
  [0x01ef] = "\xfd\x58", [0x01f0] = "\xfd\x59", [0x01f1] = "\xfd\x5a",
  [0x01f2] = "\xfd\x5b", [0x01f3] = "\xfd\x5c", [0x01f4] = "\xfd\x5d",
  [0x01f5] = "\xfd\x5e", [0x01f6] = "\xfd\x5f", [0x01f7] = "\xfd\x60",
  [0x01f8] = "\xfd\x61", [0x01f9] = "\xfd\x62", [0x01fa] = "\xfd\x63",
  [0x01fb] = "\xfd\x64", [0x01fc] = "\xfd\x65", [0x01fd] = "\xfd\x66",
  [0x01fe] = "\xfd\x67", [0x01ff] = "\xfd\x68", [0x0200] = "\xfd\x69",
  [0x0201] = "\xfd\x6a", [0x0202] = "\xfd\x6b", [0x0203] = "\xfd\x6c",
  [0x0204] = "\xfd\x6d", [0x0205] = "\xfd\x6e", [0x0206] = "\xfd\x6f",
  [0x0207] = "\xfd\x70", [0x0208] = "\xfd\x71", [0x0209] = "\xfd\x72",
  [0x020a] = "\xfd\x73", [0x020b] = "\xfd\x74", [0x020c] = "\xfd\x75",
  [0x020d] = "\xfd\x76", [0x020e] = "\xfd\x77", [0x020f] = "\xfd\x78",
  [0x0210] = "\xfd\x79", [0x0211] = "\xfd\x7a", [0x0212] = "\xfd\x7b",
  [0x0213] = "\xfd\x7c", [0x0214] = "\xfd\x7d", [0x0215] = "\xfd\x7e",
  [0x0216] = "\xfd\xa1", [0x0217] = "\xfd\xa2", [0x0218] = "\xfd\xa3",
  [0x0219] = "\xfd\xa4", [0x021a] = "\xfd\xa5", [0x021b] = "\xfd\xa6",
  [0x021c] = "\xfd\xa7", [0x021d] = "\xfd\xa8", [0x021e] = "\xfd\xa9",
  [0x021f] = "\xfd\xaa", [0x0220] = "\xfd\xab", [0x0221] = "\xfd\xac",
  [0x0222] = "\xfd\xad", [0x0223] = "\xfd\xae", [0x0224] = "\xfd\xaf",
  [0x0225] = "\xfd\xb0", [0x0226] = "\xfd\xb1", [0x0227] = "\xfd\xb2",
  [0x0228] = "\xfd\xb3", [0x0229] = "\xfd\xb4", [0x022a] = "\xfd\xb5",
  [0x022b] = "\xfd\xb6", [0x022c] = "\xfd\xb7", [0x022d] = "\xfd\xb8",
  [0x022e] = "\xfd\xb9", [0x022f] = "\xfd\xba", [0x0230] = "\xfd\xbb",
  [0x0231] = "\xfd\xbc", [0x0232] = "\xfd\xbd", [0x0233] = "\xfd\xbe",
  [0x0234] = "\xfd\xbf", [0x0235] = "\xfd\xc0", [0x0236] = "\xfd\xc1",
  [0x0237] = "\xfd\xc2", [0x0238] = "\xfd\xc3", [0x0239] = "\xfd\xc4",
  [0x023a] = "\xfd\xc5", [0x023b] = "\xfd\xc6", [0x023c] = "\xfd\xc7",
  [0x023d] = "\xfd\xc8", [0x023e] = "\xfd\xc9", [0x023f] = "\xfd\xca",
  [0x0240] = "\xfd\xcb", [0x0241] = "\xfd\xcc", [0x0242] = "\xfd\xcd",
  [0x0243] = "\xfd\xce", [0x0244] = "\xfd\xcf", [0x0245] = "\xfd\xd0",
  [0x0246] = "\xfd\xd1", [0x0247] = "\xfd\xd2", [0x0248] = "\xfd\xd3",
  [0x0249] = "\xfd\xd4", [0x024a] = "\xfd\xd5", [0x024b] = "\xfd\xd6",
  [0x024c] = "\xfd\xd7", [0x024d] = "\xfd\xd8", [0x024e] = "\xfd\xd9",
  [0x024f] = "\xfd\xda", [0x0250] = "\xfd\xdb", [0x0251] = "\xfd\xdc",
  [0x0252] = "\xfd\xdd", [0x0253] = "\xfd\xde", [0x0254] = "\xfd\xdf",
  [0x0255] = "\xfd\xe0", [0x0256] = "\xfd\xe1", [0x0257] = "\xfd\xe2",
  [0x0258] = "\xfd\xe3", [0x0259] = "\xfd\xe4", [0x025a] = "\xfd\xe5",
  [0x025b] = "\xfd\xe6", [0x025c] = "\xfd\xe7", [0x025d] = "\xfd\xe8",
  [0x025e] = "\xfd\xe9", [0x025f] = "\xfd\xea", [0x0260] = "\xfd\xeb",
  [0x0261] = "\xfd\xec", [0x0262] = "\xfd\xed", [0x0263] = "\xfd\xee",
  [0x0264] = "\xfd\xef", [0x0265] = "\xfd\xf0", [0x0266] = "\xfd\xf1",
  [0x0267] = "\xfd\xf2", [0x0268] = "\xfd\xf3", [0x0269] = "\xfd\xf4",
  [0x026a] = "\xfd\xf5", [0x026b] = "\xfd\xf6", [0x026c] = "\xfd\xf7",
  [0x026d] = "\xfd\xf8", [0x026e] = "\xfd\xf9", [0x026f] = "\xfd\xfa",
  [0x0270] = "\xfd\xfb", [0x0271] = "\xfd\xfc", [0x0272] = "\xfd\xfd",
  [0x0273] = "\xfd\xfe", [0x0274] = "\xfe\x40", [0x0275] = "\xfe\x41",
  [0x0276] = "\xfe\x42", [0x0277] = "\xfe\x43", [0x0278] = "\xfe\x44",
  [0x0279] = "\xfe\x45", [0x027a] = "\xfe\x46", [0x027b] = "\xfe\x47",
  [0x027c] = "\xfe\x48", [0x027d] = "\xfe\x49", [0x027e] = "\xfe\x4a",
  [0x027f] = "\xfe\x4b", [0x0280] = "\xfe\x4c", [0x0281] = "\xfe\x4d",
  [0x0282] = "\xfe\x4e", [0x0283] = "\xfe\x4f", [0x0284] = "\xfe\x50",
  [0x0285] = "\xfe\x51", [0x0286] = "\xfe\x52", [0x0287] = "\xfe\x53",
  [0x0288] = "\xfe\x54", [0x0289] = "\xfe\x55", [0x028a] = "\xfe\x56",
  [0x028b] = "\xfe\x57", [0x028c] = "\xfe\x58", [0x028d] = "\xfe\x59",
  [0x028e] = "\xfe\x5a", [0x028f] = "\xfe\x5b", [0x0290] = "\xfe\x5c",
  [0x0291] = "\xfe\x5d", [0x0292] = "\xfe\x5e", [0x0293] = "\xfe\x5f",
  [0x0294] = "\xfe\x60", [0x0295] = "\xfe\x61", [0x0296] = "\xfe\x62",
  [0x0297] = "\xfe\x63", [0x0298] = "\xfe\x64", [0x0299] = "\xfe\x65",
  [0x029a] = "\xfe\x66", [0x029b] = "\xfe\x67", [0x029c] = "\xfe\x68",
  [0x029d] = "\xfe\x69", [0x029e] = "\xfe\x6a", [0x029f] = "\xfe\x6b",
  [0x02a0] = "\xfe\x6c", [0x02a1] = "\xfe\x6d", [0x02a2] = "\xfe\x6e",
  [0x02a3] = "\xfe\x6f", [0x02a4] = "\xfe\x70", [0x02a5] = "\xfe\x71",
  [0x02a6] = "\xfe\x72", [0x02a7] = "\xfe\x73", [0x02a8] = "\xfe\x74",
  [0x02a9] = "\xfe\x75", [0x02aa] = "\xfe\x76", [0x02ab] = "\xfe\x77",
  [0x02ac] = "\xfe\x78", [0x02ad] = "\xfe\x79", [0x02ae] = "\xfe\x7a",
  [0x02af] = "\xfe\x7b", [0x02b0] = "\xfe\x7c", [0x02b1] = "\xfe\x7d",
  [0x02b2] = "\xfe\x7e", [0x02b3] = "\xfe\xa1", [0x02b4] = "\xfe\xa2",
  [0x02b5] = "\xfe\xa3", [0x02b6] = "\xfe\xa4", [0x02b7] = "\xfe\xa5",
  [0x02b8] = "\xfe\xa6", [0x02b9] = "\xfe\xa7", [0x02ba] = "\xfe\xa8",
  [0x02bb] = "\xfe\xa9", [0x02bc] = "\xfe\xaa", [0x02bd] = "\xfe\xab",
  [0x02be] = "\xfe\xac", [0x02bf] = "\xfe\xad", [0x02c0] = "\xfe\xae",
  [0x02c1] = "\xfe\xaf", [0x02c2] = "\xfe\xb0", [0x02c3] = "\xfe\xb1",
  [0x02c4] = "\xfe\xb2", [0x02c5] = "\xfe\xb3", [0x02c6] = "\xfe\xb4",
  [0x02c7] = "\xfe\xb5", [0x02c8] = "\xfe\xb6", [0x02c9] = "\xfe\xb7",
  [0x02ca] = "\xfe\xb8", [0x02cb] = "\xfe\xb9", [0x02cc] = "\xfe\xba",
  [0x02cd] = "\xfe\xbb", [0x02ce] = "\xfe\xbc", [0x02cf] = "\xfe\xbd",
  [0x02d0] = "\xfe\xbe", [0x02d1] = "\xfe\xbf", [0x02d2] = "\xfe\xc0",
  [0x02d3] = "\xfe\xc1", [0x02d4] = "\xfe\xc2", [0x02d5] = "\xfe\xc3",
  [0x02d6] = "\xfe\xc4", [0x02d7] = "\xfe\xc5", [0x02d8] = "\xfe\xc6",
  [0x02d9] = "\xfe\xc7", [0x02da] = "\xfe\xc8", [0x02db] = "\xfe\xc9",
  [0x02dc] = "\xfe\xca", [0x02dd] = "\xfe\xcb", [0x02de] = "\xfe\xcc",
  [0x02df] = "\xfe\xcd", [0x02e0] = "\xfe\xce", [0x02e1] = "\xfe\xcf",
  [0x02e2] = "\xfe\xd0", [0x02e3] = "\xfe\xd1", [0x02e4] = "\xfe\xd2",
  [0x02e5] = "\xfe\xd3", [0x02e6] = "\xfe\xd4", [0x02e7] = "\xfe\xd5",
  [0x02e8] = "\xfe\xd6", [0x02e9] = "\xfe\xd7", [0x02ea] = "\xfe\xd8",
  [0x02eb] = "\xfe\xd9", [0x02ec] = "\xfe\xda", [0x02ed] = "\xfe\xdb",
  [0x02ee] = "\xfe\xdc", [0x02ef] = "\xfe\xdd", [0x02f0] = "\xfe\xde",
  [0x02f1] = "\xfe\xdf", [0x02f2] = "\xfe\xe0", [0x02f3] = "\xfe\xe1",
  [0x02f4] = "\xfe\xe2", [0x02f5] = "\xfe\xe3", [0x02f6] = "\xfe\xe4",
  [0x02f7] = "\xfe\xe5", [0x02f8] = "\xfe\xe6", [0x02f9] = "\xfe\xe7",
  [0x02fa] = "\xfe\xe8", [0x02fb] = "\xfe\xe9", [0x02fc] = "\xfe\xea",
  [0x02fd] = "\xfe\xeb", [0x02fe] = "\xfe\xec", [0x02ff] = "\xfe\xed",
  [0x0300] = "\xfe\xee", [0x0301] = "\xfe\xef", [0x0302] = "\xfe\xf0",
  [0x0303] = "\xfe\xf1", [0x0304] = "\xfe\xf2", [0x0305] = "\xfe\xf3",
  [0x0306] = "\xfe\xf4", [0x0307] = "\xfe\xf5", [0x0308] = "\xfe\xf6",
  [0x0309] = "\xfe\xf7", [0x030a] = "\xfe\xf8", [0x030b] = "\xfe\xf9",
  [0x030c] = "\xfe\xfa", [0x030d] = "\xfe\xfb", [0x030e] = "\xfe\xfc",
  [0x030f] = "\xfe\xfd", [0x0310] = "\xfe\xfe", [0x0311] = "\x8e\x40",
  [0x0312] = "\x8e\x41", [0x0313] = "\x8e\x42", [0x0314] = "\x8e\x43",
  [0x0315] = "\x8e\x44", [0x0316] = "\x8e\x45", [0x0317] = "\x8e\x46",
  [0x0318] = "\x8e\x47", [0x0319] = "\x8e\x48", [0x031a] = "\x8e\x49",
  [0x031b] = "\x8e\x4a", [0x031c] = "\x8e\x4b", [0x031d] = "\x8e\x4c",
  [0x031e] = "\x8e\x4d", [0x031f] = "\x8e\x4e", [0x0320] = "\x8e\x4f",
  [0x0321] = "\x8e\x50", [0x0322] = "\x8e\x51", [0x0323] = "\x8e\x52",
  [0x0324] = "\x8e\x53", [0x0325] = "\x8e\x54", [0x0326] = "\x8e\x55",
  [0x0327] = "\x8e\x56", [0x0328] = "\x8e\x57", [0x0329] = "\x8e\x58",
  [0x032a] = "\x8e\x59", [0x032b] = "\x8e\x5a", [0x032c] = "\x8e\x5b",
  [0x032d] = "\x8e\x5c", [0x032e] = "\x8e\x5d", [0x032f] = "\x8e\x5e",
  [0x0330] = "\x8e\x5f", [0x0331] = "\x8e\x60", [0x0332] = "\x8e\x61",
  [0x0333] = "\x8e\x62", [0x0334] = "\x8e\x63", [0x0335] = "\x8e\x64",
  [0x0336] = "\x8e\x65", [0x0337] = "\x8e\x66", [0x0338] = "\x8e\x67",
  [0x0339] = "\x8e\x68", [0x033a] = "\x8e\x69", [0x033b] = "\x8e\x6a",
  [0x033c] = "\x8e\x6b", [0x033d] = "\x8e\x6c", [0x033e] = "\x8e\x6d",
  [0x033f] = "\x8e\x6e", [0x0340] = "\x8e\x6f", [0x0341] = "\x8e\x70",
  [0x0342] = "\x8e\x71", [0x0343] = "\x8e\x72", [0x0344] = "\x8e\x73",
  [0x0345] = "\x8e\x74", [0x0346] = "\x8e\x75", [0x0347] = "\x8e\x76",
  [0x0348] = "\x8e\x77", [0x0349] = "\x8e\x78", [0x034a] = "\x8e\x79",
  [0x034b] = "\x8e\x7a", [0x034c] = "\x8e\x7b", [0x034d] = "\x8e\x7c",
  [0x034e] = "\x8e\x7d", [0x034f] = "\x8e\x7e", [0x0350] = "\x8e\xa1",
  [0x0351] = "\x8e\xa2", [0x0352] = "\x8e\xa3", [0x0353] = "\x8e\xa4",
  [0x0354] = "\x8e\xa5", [0x0355] = "\x8e\xa6", [0x0356] = "\x8e\xa7",
  [0x0357] = "\x8e\xa8", [0x0358] = "\x8e\xa9", [0x0359] = "\x8e\xaa",
  [0x035a] = "\x8e\xab", [0x035b] = "\x8e\xac", [0x035c] = "\x8e\xad",
  [0x035d] = "\x8e\xae", [0x035e] = "\x8e\xaf", [0x035f] = "\x8e\xb0",
  [0x0360] = "\x8e\xb1", [0x0361] = "\x8e\xb2", [0x0362] = "\x8e\xb3",
  [0x0363] = "\x8e\xb4", [0x0364] = "\x8e\xb5", [0x0365] = "\x8e\xb6",
  [0x0366] = "\x8e\xb7", [0x0367] = "\x8e\xb8", [0x0368] = "\x8e\xb9",
  [0x0369] = "\x8e\xba", [0x036a] = "\x8e\xbb", [0x036b] = "\x8e\xbc",
  [0x036c] = "\x8e\xbd", [0x036d] = "\x8e\xbe", [0x036e] = "\x8e\xbf",
  [0x036f] = "\x8e\xc0", [0x0370] = "\x8e\xc1", [0x0371] = "\x8e\xc2",
  [0x0372] = "\x8e\xc3", [0x0373] = "\x8e\xc4", [0x0374] = "\x8e\xc5",
  [0x0375] = "\x8e\xc6", [0x0376] = "\x8e\xc7", [0x0377] = "\x8e\xc8",
  [0x0378] = "\x8e\xc9", [0x0379] = "\x8e\xca", [0x037a] = "\x8e\xcb",
  [0x037b] = "\x8e\xcc", [0x037c] = "\x8e\xcd", [0x037d] = "\x8e\xce",
  [0x037e] = "\x8e\xcf", [0x037f] = "\x8e\xd0", [0x0380] = "\x8e\xd1",
  [0x0381] = "\x8e\xd2", [0x0382] = "\x8e\xd3", [0x0383] = "\x8e\xd4",
  [0x0384] = "\x8e\xd5", [0x0385] = "\x8e\xd6", [0x0386] = "\x8e\xd7",
  [0x0387] = "\x8e\xd8", [0x0388] = "\x8e\xd9", [0x0389] = "\x8e\xda",
  [0x038a] = "\x8e\xdb", [0x038b] = "\x8e\xdc", [0x038c] = "\x8e\xdd",
  [0x038d] = "\x8e\xde", [0x038e] = "\x8e\xdf", [0x038f] = "\x8e\xe0",
  [0x0390] = "\x8e\xe1", [0x0391] = "\x8e\xe2", [0x0392] = "\x8e\xe3",
  [0x0393] = "\x8e\xe4", [0x0394] = "\x8e\xe5", [0x0395] = "\x8e\xe6",
  [0x0396] = "\x8e\xe7", [0x0397] = "\x8e\xe8", [0x0398] = "\x8e\xe9",
  [0x0399] = "\x8e\xea", [0x039a] = "\x8e\xeb", [0x039b] = "\x8e\xec",
  [0x039c] = "\x8e\xed", [0x039d] = "\x8e\xee", [0x039e] = "\x8e\xef",
  [0x039f] = "\x8e\xf0", [0x03a0] = "\x8e\xf1", [0x03a1] = "\x8e\xf2",
  [0x03a2] = "\x8e\xf3", [0x03a3] = "\x8e\xf4", [0x03a4] = "\x8e\xf5",
  [0x03a5] = "\x8e\xf6", [0x03a6] = "\x8e\xf7", [0x03a7] = "\x8e\xf8",
  [0x03a8] = "\x8e\xf9", [0x03a9] = "\x8e\xfa", [0x03aa] = "\x8e\xfb",
  [0x03ab] = "\x8e\xfc", [0x03ac] = "\x8e\xfd", [0x03ad] = "\x8e\xfe",
  [0x03ae] = "\x8f\x40", [0x03af] = "\x8f\x41", [0x03b0] = "\x8f\x42",
  [0x03b1] = "\x8f\x43", [0x03b2] = "\x8f\x44", [0x03b3] = "\x8f\x45",
  [0x03b4] = "\x8f\x46", [0x03b5] = "\x8f\x47", [0x03b6] = "\x8f\x48",
  [0x03b7] = "\x8f\x49", [0x03b8] = "\x8f\x4a", [0x03b9] = "\x8f\x4b",
  [0x03ba] = "\x8f\x4c", [0x03bb] = "\x8f\x4d", [0x03bc] = "\x8f\x4e",
  [0x03bd] = "\x8f\x4f", [0x03be] = "\x8f\x50", [0x03bf] = "\x8f\x51",
  [0x03c0] = "\x8f\x52", [0x03c1] = "\x8f\x53", [0x03c2] = "\x8f\x54",
  [0x03c3] = "\x8f\x55", [0x03c4] = "\x8f\x56", [0x03c5] = "\x8f\x57",
  [0x03c6] = "\x8f\x58", [0x03c7] = "\x8f\x59", [0x03c8] = "\x8f\x5a",
  [0x03c9] = "\x8f\x5b", [0x03ca] = "\x8f\x5c", [0x03cb] = "\x8f\x5d",
  [0x03cc] = "\x8f\x5e", [0x03cd] = "\x8f\x5f", [0x03ce] = "\x8f\x60",
  [0x03cf] = "\x8f\x61", [0x03d0] = "\x8f\x62", [0x03d1] = "\x8f\x63",
  [0x03d2] = "\x8f\x64", [0x03d3] = "\x8f\x65", [0x03d4] = "\x8f\x66",
  [0x03d5] = "\x8f\x67", [0x03d6] = "\x8f\x68", [0x03d7] = "\x8f\x69",
  [0x03d8] = "\x8f\x6a", [0x03d9] = "\x8f\x6b", [0x03da] = "\x8f\x6c",
  [0x03db] = "\x8f\x6d", [0x03dc] = "\x8f\x6e", [0x03dd] = "\x8f\x6f",
  [0x03de] = "\x8f\x70", [0x03df] = "\x8f\x71", [0x03e0] = "\x8f\x72",
  [0x03e1] = "\x8f\x73", [0x03e2] = "\x8f\x74", [0x03e3] = "\x8f\x75",
  [0x03e4] = "\x8f\x76", [0x03e5] = "\x8f\x77", [0x03e6] = "\x8f\x78",
  [0x03e7] = "\x8f\x79", [0x03e8] = "\x8f\x7a", [0x03e9] = "\x8f\x7b",
  [0x03ea] = "\x8f\x7c", [0x03eb] = "\x8f\x7d", [0x03ec] = "\x8f\x7e",
  [0x03ed] = "\x8f\xa1", [0x03ee] = "\x8f\xa2", [0x03ef] = "\x8f\xa3",
  [0x03f0] = "\x8f\xa4", [0x03f1] = "\x8f\xa5", [0x03f2] = "\x8f\xa6",
  [0x03f3] = "\x8f\xa7", [0x03f4] = "\x8f\xa8", [0x03f5] = "\x8f\xa9",
  [0x03f6] = "\x8f\xaa", [0x03f7] = "\x8f\xab", [0x03f8] = "\x8f\xac",
  [0x03f9] = "\x8f\xad", [0x03fa] = "\x8f\xae", [0x03fb] = "\x8f\xaf",
  [0x03fc] = "\x8f\xb0", [0x03fd] = "\x8f\xb1", [0x03fe] = "\x8f\xb2",
  [0x03ff] = "\x8f\xb3", [0x0400] = "\x8f\xb4", [0x0401] = "\x8f\xb5",
  [0x0402] = "\x8f\xb6", [0x0403] = "\x8f\xb7", [0x0404] = "\x8f\xb8",
  [0x0405] = "\x8f\xb9", [0x0406] = "\x8f\xba", [0x0407] = "\x8f\xbb",
  [0x0408] = "\x8f\xbc", [0x0409] = "\x8f\xbd", [0x040a] = "\x8f\xbe",
  [0x040b] = "\x8f\xbf", [0x040c] = "\x8f\xc0", [0x040d] = "\x8f\xc1",
  [0x040e] = "\x8f\xc2", [0x040f] = "\x8f\xc3", [0x0410] = "\x8f\xc4",
  [0x0411] = "\x8f\xc5", [0x0412] = "\x8f\xc6", [0x0413] = "\x8f\xc7",
  [0x0414] = "\x8f\xc8", [0x0415] = "\x8f\xc9", [0x0416] = "\x8f\xca",
  [0x0417] = "\x8f\xcb", [0x0418] = "\x8f\xcc", [0x0419] = "\x8f\xcd",
  [0x041a] = "\x8f\xce", [0x041b] = "\x8f\xcf", [0x041c] = "\x8f\xd0",
  [0x041d] = "\x8f\xd1", [0x041e] = "\x8f\xd2", [0x041f] = "\x8f\xd3",
  [0x0420] = "\x8f\xd4", [0x0421] = "\x8f\xd5", [0x0422] = "\x8f\xd6",
  [0x0423] = "\x8f\xd7", [0x0424] = "\x8f\xd8", [0x0425] = "\x8f\xd9",
  [0x0426] = "\x8f\xda", [0x0427] = "\x8f\xdb", [0x0428] = "\x8f\xdc",
  [0x0429] = "\x8f\xdd", [0x042a] = "\x8f\xde", [0x042b] = "\x8f\xdf",
  [0x042c] = "\x8f\xe0", [0x042d] = "\x8f\xe1", [0x042e] = "\x8f\xe2",
  [0x042f] = "\x8f\xe3", [0x0430] = "\x8f\xe4", [0x0431] = "\x8f\xe5",
  [0x0432] = "\x8f\xe6", [0x0433] = "\x8f\xe7", [0x0434] = "\x8f\xe8",
  [0x0435] = "\x8f\xe9", [0x0436] = "\x8f\xea", [0x0437] = "\x8f\xeb",
  [0x0438] = "\x8f\xec", [0x0439] = "\x8f\xed", [0x043a] = "\x8f\xee",
  [0x043b] = "\x8f\xef", [0x043c] = "\x8f\xf0", [0x043d] = "\x8f\xf1",
  [0x043e] = "\x8f\xf2", [0x043f] = "\x8f\xf3", [0x0440] = "\x8f\xf4",
  [0x0441] = "\x8f\xf5", [0x0442] = "\x8f\xf6", [0x0443] = "\x8f\xf7",
  [0x0444] = "\x8f\xf8", [0x0445] = "\x8f\xf9", [0x0446] = "\x8f\xfa",
  [0x0447] = "\x8f\xfb", [0x0448] = "\x8f\xfc", [0x0449] = "\x8f\xfd",
  [0x044a] = "\x8f\xfe", [0x044b] = "\x90\x40", [0x044c] = "\x90\x41",
  [0x044d] = "\x90\x42", [0x044e] = "\x90\x43", [0x044f] = "\x90\x44",
  [0x0450] = "\x90\x45", [0x0451] = "\x90\x46", [0x0452] = "\x90\x47",
  [0x0453] = "\x90\x48", [0x0454] = "\x90\x49", [0x0455] = "\x90\x4a",
  [0x0456] = "\x90\x4b", [0x0457] = "\x90\x4c", [0x0458] = "\x90\x4d",
  [0x0459] = "\x90\x4e", [0x045a] = "\x90\x4f", [0x045b] = "\x90\x50",
  [0x045c] = "\x90\x51", [0x045d] = "\x90\x52", [0x045e] = "\x90\x53",
  [0x045f] = "\x90\x54", [0x0460] = "\x90\x55", [0x0461] = "\x90\x56",
  [0x0462] = "\x90\x57", [0x0463] = "\x90\x58", [0x0464] = "\x90\x59",
  [0x0465] = "\x90\x5a", [0x0466] = "\x90\x5b", [0x0467] = "\x90\x5c",
  [0x0468] = "\x90\x5d", [0x0469] = "\x90\x5e", [0x046a] = "\x90\x5f",
  [0x046b] = "\x90\x60", [0x046c] = "\x90\x61", [0x046d] = "\x90\x62",
  [0x046e] = "\x90\x63", [0x046f] = "\x90\x64", [0x0470] = "\x90\x65",
  [0x0471] = "\x90\x66", [0x0472] = "\x90\x67", [0x0473] = "\x90\x68",
  [0x0474] = "\x90\x69", [0x0475] = "\x90\x6a", [0x0476] = "\x90\x6b",
  [0x0477] = "\x90\x6c", [0x0478] = "\x90\x6d", [0x0479] = "\x90\x6e",
  [0x047a] = "\x90\x6f", [0x047b] = "\x90\x70", [0x047c] = "\x90\x71",
  [0x047d] = "\x90\x72", [0x047e] = "\x90\x73", [0x047f] = "\x90\x74",
  [0x0480] = "\x90\x75", [0x0481] = "\x90\x76", [0x0482] = "\x90\x77",
  [0x0483] = "\x90\x78", [0x0484] = "\x90\x79", [0x0485] = "\x90\x7a",
  [0x0486] = "\x90\x7b", [0x0487] = "\x90\x7c", [0x0488] = "\x90\x7d",
  [0x0489] = "\x90\x7e", [0x048a] = "\x90\xa1", [0x048b] = "\x90\xa2",
  [0x048c] = "\x90\xa3", [0x048d] = "\x90\xa4", [0x048e] = "\x90\xa5",
  [0x048f] = "\x90\xa6", [0x0490] = "\x90\xa7", [0x0491] = "\x90\xa8",
  [0x0492] = "\x90\xa9", [0x0493] = "\x90\xaa", [0x0494] = "\x90\xab",
  [0x0495] = "\x90\xac", [0x0496] = "\x90\xad", [0x0497] = "\x90\xae",
  [0x0498] = "\x90\xaf", [0x0499] = "\x90\xb0", [0x049a] = "\x90\xb1",
  [0x049b] = "\x90\xb2", [0x049c] = "\x90\xb3", [0x049d] = "\x90\xb4",
  [0x049e] = "\x90\xb5", [0x049f] = "\x90\xb6", [0x04a0] = "\x90\xb7",
  [0x04a1] = "\x90\xb8", [0x04a2] = "\x90\xb9", [0x04a3] = "\x90\xba",
  [0x04a4] = "\x90\xbb", [0x04a5] = "\x90\xbc", [0x04a6] = "\x90\xbd",
  [0x04a7] = "\x90\xbe", [0x04a8] = "\x90\xbf", [0x04a9] = "\x90\xc0",
  [0x04aa] = "\x90\xc1", [0x04ab] = "\x90\xc2", [0x04ac] = "\x90\xc3",
  [0x04ad] = "\x90\xc4", [0x04ae] = "\x90\xc5", [0x04af] = "\x90\xc6",
  [0x04b0] = "\x90\xc7", [0x04b1] = "\x90\xc8", [0x04b2] = "\x90\xc9",
  [0x04b3] = "\x90\xca", [0x04b4] = "\x90\xcb", [0x04b5] = "\x90\xcc",
  [0x04b6] = "\x90\xcd", [0x04b7] = "\x90\xce", [0x04b8] = "\x90\xcf",
  [0x04b9] = "\x90\xd0", [0x04ba] = "\x90\xd1", [0x04bb] = "\x90\xd2",
  [0x04bc] = "\x90\xd3", [0x04bd] = "\x90\xd4", [0x04be] = "\x90\xd5",
  [0x04bf] = "\x90\xd6", [0x04c0] = "\x90\xd7", [0x04c1] = "\x90\xd8",
  [0x04c2] = "\x90\xd9", [0x04c3] = "\x90\xda", [0x04c4] = "\x90\xdb",
  [0x04c5] = "\x90\xdc", [0x04c6] = "\x90\xdd", [0x04c7] = "\x90\xde",
  [0x04c8] = "\x90\xdf", [0x04c9] = "\x90\xe0", [0x04ca] = "\x90\xe1",
  [0x04cb] = "\x90\xe2", [0x04cc] = "\x90\xe3", [0x04cd] = "\x90\xe4",
  [0x04ce] = "\x90\xe5", [0x04cf] = "\x90\xe6", [0x04d0] = "\x90\xe7",
  [0x04d1] = "\x90\xe8", [0x04d2] = "\x90\xe9", [0x04d3] = "\x90\xea",
  [0x04d4] = "\x90\xeb", [0x04d5] = "\x90\xec", [0x04d6] = "\x90\xed",
  [0x04d7] = "\x90\xee", [0x04d8] = "\x90\xef", [0x04d9] = "\x90\xf0",
  [0x04da] = "\x90\xf1", [0x04db] = "\x90\xf2", [0x04dc] = "\x90\xf3",
  [0x04dd] = "\x90\xf4", [0x04de] = "\x90\xf5", [0x04df] = "\x90\xf6",
  [0x04e0] = "\x90\xf7", [0x04e1] = "\x90\xf8", [0x04e2] = "\x90\xf9",
  [0x04e3] = "\x90\xfa", [0x04e4] = "\x90\xfb", [0x04e5] = "\x90\xfc",
  [0x04e6] = "\x90\xfd", [0x04e7] = "\x90\xfe", [0x04e8] = "\x91\x40",
  [0x04e9] = "\x91\x41", [0x04ea] = "\x91\x42", [0x04eb] = "\x91\x43",
  [0x04ec] = "\x91\x44", [0x04ed] = "\x91\x45", [0x04ee] = "\x91\x46",
  [0x04ef] = "\x91\x47", [0x04f0] = "\x91\x48", [0x04f1] = "\x91\x49",
  [0x04f2] = "\x91\x4a", [0x04f3] = "\x91\x4b", [0x04f4] = "\x91\x4c",
  [0x04f5] = "\x91\x4d", [0x04f6] = "\x91\x4e", [0x04f7] = "\x91\x4f",
  [0x04f8] = "\x91\x50", [0x04f9] = "\x91\x51", [0x04fa] = "\x91\x52",
  [0x04fb] = "\x91\x53", [0x04fc] = "\x91\x54", [0x04fd] = "\x91\x55",
  [0x04fe] = "\x91\x56", [0x04ff] = "\x91\x57", [0x0500] = "\x91\x58",
  [0x0501] = "\x91\x59", [0x0502] = "\x91\x5a", [0x0503] = "\x91\x5b",
  [0x0504] = "\x91\x5c", [0x0505] = "\x91\x5d", [0x0506] = "\x91\x5e",
  [0x0507] = "\x91\x5f", [0x0508] = "\x91\x60", [0x0509] = "\x91\x61",
  [0x050a] = "\x91\x62", [0x050b] = "\x91\x63", [0x050c] = "\x91\x64",
  [0x050d] = "\x91\x65", [0x050e] = "\x91\x66", [0x050f] = "\x91\x67",
  [0x0510] = "\x91\x68", [0x0511] = "\x91\x69", [0x0512] = "\x91\x6a",
  [0x0513] = "\x91\x6b", [0x0514] = "\x91\x6c", [0x0515] = "\x91\x6d",
  [0x0516] = "\x91\x6e", [0x0517] = "\x91\x6f", [0x0518] = "\x91\x70",
  [0x0519] = "\x91\x71", [0x051a] = "\x91\x72", [0x051b] = "\x91\x73",
  [0x051c] = "\x91\x74", [0x051d] = "\x91\x75", [0x051e] = "\x91\x76",
  [0x051f] = "\x91\x77", [0x0520] = "\x91\x78", [0x0521] = "\x91\x79",
  [0x0522] = "\x91\x7a", [0x0523] = "\x91\x7b", [0x0524] = "\x91\x7c",
  [0x0525] = "\x91\x7d", [0x0526] = "\x91\x7e", [0x0527] = "\x91\xa1",
  [0x0528] = "\x91\xa2", [0x0529] = "\x91\xa3", [0x052a] = "\x91\xa4",
  [0x052b] = "\x91\xa5", [0x052c] = "\x91\xa6", [0x052d] = "\x91\xa7",
  [0x052e] = "\x91\xa8", [0x052f] = "\x91\xa9", [0x0530] = "\x91\xaa",
  [0x0531] = "\x91\xab", [0x0532] = "\x91\xac", [0x0533] = "\x91\xad",
  [0x0534] = "\x91\xae", [0x0535] = "\x91\xaf", [0x0536] = "\x91\xb0",
  [0x0537] = "\x91\xb1", [0x0538] = "\x91\xb2", [0x0539] = "\x91\xb3",
  [0x053a] = "\x91\xb4", [0x053b] = "\x91\xb5", [0x053c] = "\x91\xb6",
  [0x053d] = "\x91\xb7", [0x053e] = "\x91\xb8", [0x053f] = "\x91\xb9",
  [0x0540] = "\x91\xba", [0x0541] = "\x91\xbb", [0x0542] = "\x91\xbc",
  [0x0543] = "\x91\xbd", [0x0544] = "\x91\xbe", [0x0545] = "\x91\xbf",
  [0x0546] = "\x91\xc0", [0x0547] = "\x91\xc1", [0x0548] = "\x91\xc2",
  [0x0549] = "\x91\xc3", [0x054a] = "\x91\xc4", [0x054b] = "\x91\xc5",
  [0x054c] = "\x91\xc6", [0x054d] = "\x91\xc7", [0x054e] = "\x91\xc8",
  [0x054f] = "\x91\xc9", [0x0550] = "\x91\xca", [0x0551] = "\x91\xcb",
  [0x0552] = "\x91\xcc", [0x0553] = "\x91\xcd", [0x0554] = "\x91\xce",
  [0x0555] = "\x91\xcf", [0x0556] = "\x91\xd0", [0x0557] = "\x91\xd1",
  [0x0558] = "\x91\xd2", [0x0559] = "\x91\xd3", [0x055a] = "\x91\xd4",
  [0x055b] = "\x91\xd5", [0x055c] = "\x91\xd6", [0x055d] = "\x91\xd7",
  [0x055e] = "\x91\xd8", [0x055f] = "\x91\xd9", [0x0560] = "\x91\xda",
  [0x0561] = "\x91\xdb", [0x0562] = "\x91\xdc", [0x0563] = "\x91\xdd",
  [0x0564] = "\x91\xde", [0x0565] = "\x91\xdf", [0x0566] = "\x91\xe0",
  [0x0567] = "\x91\xe1", [0x0568] = "\x91\xe2", [0x0569] = "\x91\xe3",
  [0x056a] = "\x91\xe4", [0x056b] = "\x91\xe5", [0x056c] = "\x91\xe6",
  [0x056d] = "\x91\xe7", [0x056e] = "\x91\xe8", [0x056f] = "\x91\xe9",
  [0x0570] = "\x91\xea", [0x0571] = "\x91\xeb", [0x0572] = "\x91\xec",
  [0x0573] = "\x91\xed", [0x0574] = "\x91\xee", [0x0575] = "\x91\xef",
  [0x0576] = "\x91\xf0", [0x0577] = "\x91\xf1", [0x0578] = "\x91\xf2",
  [0x0579] = "\x91\xf3", [0x057a] = "\x91\xf4", [0x057b] = "\x91\xf5",
  [0x057c] = "\x91\xf6", [0x057d] = "\x91\xf7", [0x057e] = "\x91\xf8",
  [0x057f] = "\x91\xf9", [0x0580] = "\x91\xfa", [0x0581] = "\x91\xfb",
  [0x0582] = "\x91\xfc", [0x0583] = "\x91\xfd", [0x0584] = "\x91\xfe",
  [0x0585] = "\x92\x40", [0x0586] = "\x92\x41", [0x0587] = "\x92\x42",
  [0x0588] = "\x92\x43", [0x0589] = "\x92\x44", [0x058a] = "\x92\x45",
  [0x058b] = "\x92\x46", [0x058c] = "\x92\x47", [0x058d] = "\x92\x48",
  [0x058e] = "\x92\x49", [0x058f] = "\x92\x4a", [0x0590] = "\x92\x4b",
  [0x0591] = "\x92\x4c", [0x0592] = "\x92\x4d", [0x0593] = "\x92\x4e",
  [0x0594] = "\x92\x4f", [0x0595] = "\x92\x50", [0x0596] = "\x92\x51",
  [0x0597] = "\x92\x52", [0x0598] = "\x92\x53", [0x0599] = "\x92\x54",
  [0x059a] = "\x92\x55", [0x059b] = "\x92\x56", [0x059c] = "\x92\x57",
  [0x059d] = "\x92\x58", [0x059e] = "\x92\x59", [0x059f] = "\x92\x5a",
  [0x05a0] = "\x92\x5b", [0x05a1] = "\x92\x5c", [0x05a2] = "\x92\x5d",
  [0x05a3] = "\x92\x5e", [0x05a4] = "\x92\x5f", [0x05a5] = "\x92\x60",
  [0x05a6] = "\x92\x61", [0x05a7] = "\x92\x62", [0x05a8] = "\x92\x63",
  [0x05a9] = "\x92\x64", [0x05aa] = "\x92\x65", [0x05ab] = "\x92\x66",
  [0x05ac] = "\x92\x67", [0x05ad] = "\x92\x68", [0x05ae] = "\x92\x69",
  [0x05af] = "\x92\x6a", [0x05b0] = "\x92\x6b", [0x05b1] = "\x92\x6c",
  [0x05b2] = "\x92\x6d", [0x05b3] = "\x92\x6e", [0x05b4] = "\x92\x6f",
  [0x05b5] = "\x92\x70", [0x05b6] = "\x92\x71", [0x05b7] = "\x92\x72",
  [0x05b8] = "\x92\x73", [0x05b9] = "\x92\x74", [0x05ba] = "\x92\x75",
  [0x05bb] = "\x92\x76", [0x05bc] = "\x92\x77", [0x05bd] = "\x92\x78",
  [0x05be] = "\x92\x79", [0x05bf] = "\x92\x7a", [0x05c0] = "\x92\x7b",
  [0x05c1] = "\x92\x7c", [0x05c2] = "\x92\x7d", [0x05c3] = "\x92\x7e",
  [0x05c4] = "\x92\xa1", [0x05c5] = "\x92\xa2", [0x05c6] = "\x92\xa3",
  [0x05c7] = "\x92\xa4", [0x05c8] = "\x92\xa5", [0x05c9] = "\x92\xa6",
  [0x05ca] = "\x92\xa7", [0x05cb] = "\x92\xa8", [0x05cc] = "\x92\xa9",
  [0x05cd] = "\x92\xaa", [0x05ce] = "\x92\xab", [0x05cf] = "\x92\xac",
  [0x05d0] = "\x92\xad", [0x05d1] = "\x92\xae", [0x05d2] = "\x92\xaf",
  [0x05d3] = "\x92\xb0", [0x05d4] = "\x92\xb1", [0x05d5] = "\x92\xb2",
  [0x05d6] = "\x92\xb3", [0x05d7] = "\x92\xb4", [0x05d8] = "\x92\xb5",
  [0x05d9] = "\x92\xb6", [0x05da] = "\x92\xb7", [0x05db] = "\x92\xb8",
  [0x05dc] = "\x92\xb9", [0x05dd] = "\x92\xba", [0x05de] = "\x92\xbb",
  [0x05df] = "\x92\xbc", [0x05e0] = "\x92\xbd", [0x05e1] = "\x92\xbe",
  [0x05e2] = "\x92\xbf", [0x05e3] = "\x92\xc0", [0x05e4] = "\x92\xc1",
  [0x05e5] = "\x92\xc2", [0x05e6] = "\x92\xc3", [0x05e7] = "\x92\xc4",
  [0x05e8] = "\x92\xc5", [0x05e9] = "\x92\xc6", [0x05ea] = "\x92\xc7",
  [0x05eb] = "\x92\xc8", [0x05ec] = "\x92\xc9", [0x05ed] = "\x92\xca",
  [0x05ee] = "\x92\xcb", [0x05ef] = "\x92\xcc", [0x05f0] = "\x92\xcd",
  [0x05f1] = "\x92\xce", [0x05f2] = "\x92\xcf", [0x05f3] = "\x92\xd0",
  [0x05f4] = "\x92\xd1", [0x05f5] = "\x92\xd2", [0x05f6] = "\x92\xd3",
  [0x05f7] = "\x92\xd4", [0x05f8] = "\x92\xd5", [0x05f9] = "\x92\xd6",
  [0x05fa] = "\x92\xd7", [0x05fb] = "\x92\xd8", [0x05fc] = "\x92\xd9",
  [0x05fd] = "\x92\xda", [0x05fe] = "\x92\xdb", [0x05ff] = "\x92\xdc",
  [0x0600] = "\x92\xdd", [0x0601] = "\x92\xde", [0x0602] = "\x92\xdf",
  [0x0603] = "\x92\xe0", [0x0604] = "\x92\xe1", [0x0605] = "\x92\xe2",
  [0x0606] = "\x92\xe3", [0x0607] = "\x92\xe4", [0x0608] = "\x92\xe5",
  [0x0609] = "\x92\xe6", [0x060a] = "\x92\xe7", [0x060b] = "\x92\xe8",
  [0x060c] = "\x92\xe9", [0x060d] = "\x92\xea", [0x060e] = "\x92\xeb",
  [0x060f] = "\x92\xec", [0x0610] = "\x92\xed", [0x0611] = "\x92\xee",
  [0x0612] = "\x92\xef", [0x0613] = "\x92\xf0", [0x0614] = "\x92\xf1",
  [0x0615] = "\x92\xf2", [0x0616] = "\x92\xf3", [0x0617] = "\x92\xf4",
  [0x0618] = "\x92\xf5", [0x0619] = "\x92\xf6", [0x061a] = "\x92\xf7",
  [0x061b] = "\x92\xf8", [0x061c] = "\x92\xf9", [0x061d] = "\x92\xfa",
  [0x061e] = "\x92\xfb", [0x061f] = "\x92\xfc", [0x0620] = "\x92\xfd",
  [0x0621] = "\x92\xfe", [0x0622] = "\x93\x40", [0x0623] = "\x93\x41",
  [0x0624] = "\x93\x42", [0x0625] = "\x93\x43", [0x0626] = "\x93\x44",
  [0x0627] = "\x93\x45", [0x0628] = "\x93\x46", [0x0629] = "\x93\x47",
  [0x062a] = "\x93\x48", [0x062b] = "\x93\x49", [0x062c] = "\x93\x4a",
  [0x062d] = "\x93\x4b", [0x062e] = "\x93\x4c", [0x062f] = "\x93\x4d",
  [0x0630] = "\x93\x4e", [0x0631] = "\x93\x4f", [0x0632] = "\x93\x50",
  [0x0633] = "\x93\x51", [0x0634] = "\x93\x52", [0x0635] = "\x93\x53",
  [0x0636] = "\x93\x54", [0x0637] = "\x93\x55", [0x0638] = "\x93\x56",
  [0x0639] = "\x93\x57", [0x063a] = "\x93\x58", [0x063b] = "\x93\x59",
  [0x063c] = "\x93\x5a", [0x063d] = "\x93\x5b", [0x063e] = "\x93\x5c",
  [0x063f] = "\x93\x5d", [0x0640] = "\x93\x5e", [0x0641] = "\x93\x5f",
  [0x0642] = "\x93\x60", [0x0643] = "\x93\x61", [0x0644] = "\x93\x62",
  [0x0645] = "\x93\x63", [0x0646] = "\x93\x64", [0x0647] = "\x93\x65",
  [0x0648] = "\x93\x66", [0x0649] = "\x93\x67", [0x064a] = "\x93\x68",
  [0x064b] = "\x93\x69", [0x064c] = "\x93\x6a", [0x064d] = "\x93\x6b",
  [0x064e] = "\x93\x6c", [0x064f] = "\x93\x6d", [0x0650] = "\x93\x6e",
  [0x0651] = "\x93\x6f", [0x0652] = "\x93\x70", [0x0653] = "\x93\x71",
  [0x0654] = "\x93\x72", [0x0655] = "\x93\x73", [0x0656] = "\x93\x74",
  [0x0657] = "\x93\x75", [0x0658] = "\x93\x76", [0x0659] = "\x93\x77",
  [0x065a] = "\x93\x78", [0x065b] = "\x93\x79", [0x065c] = "\x93\x7a",
  [0x065d] = "\x93\x7b", [0x065e] = "\x93\x7c", [0x065f] = "\x93\x7d",
  [0x0660] = "\x93\x7e", [0x0661] = "\x93\xa1", [0x0662] = "\x93\xa2",
  [0x0663] = "\x93\xa3", [0x0664] = "\x93\xa4", [0x0665] = "\x93\xa5",
  [0x0666] = "\x93\xa6", [0x0667] = "\x93\xa7", [0x0668] = "\x93\xa8",
  [0x0669] = "\x93\xa9", [0x066a] = "\x93\xaa", [0x066b] = "\x93\xab",
  [0x066c] = "\x93\xac", [0x066d] = "\x93\xad", [0x066e] = "\x93\xae",
  [0x066f] = "\x93\xaf", [0x0670] = "\x93\xb0", [0x0671] = "\x93\xb1",
  [0x0672] = "\x93\xb2", [0x0673] = "\x93\xb3", [0x0674] = "\x93\xb4",
  [0x0675] = "\x93\xb5", [0x0676] = "\x93\xb6", [0x0677] = "\x93\xb7",
  [0x0678] = "\x93\xb8", [0x0679] = "\x93\xb9", [0x067a] = "\x93\xba",
  [0x067b] = "\x93\xbb", [0x067c] = "\x93\xbc", [0x067d] = "\x93\xbd",
  [0x067e] = "\x93\xbe", [0x067f] = "\x93\xbf", [0x0680] = "\x93\xc0",
  [0x0681] = "\x93\xc1", [0x0682] = "\x93\xc2", [0x0683] = "\x93\xc3",
  [0x0684] = "\x93\xc4", [0x0685] = "\x93\xc5", [0x0686] = "\x93\xc6",
  [0x0687] = "\x93\xc7", [0x0688] = "\x93\xc8", [0x0689] = "\x93\xc9",
  [0x068a] = "\x93\xca", [0x068b] = "\x93\xcb", [0x068c] = "\x93\xcc",
  [0x068d] = "\x93\xcd", [0x068e] = "\x93\xce", [0x068f] = "\x93\xcf",
  [0x0690] = "\x93\xd0", [0x0691] = "\x93\xd1", [0x0692] = "\x93\xd2",
  [0x0693] = "\x93\xd3", [0x0694] = "\x93\xd4", [0x0695] = "\x93\xd5",
  [0x0696] = "\x93\xd6", [0x0697] = "\x93\xd7", [0x0698] = "\x93\xd8",
  [0x0699] = "\x93\xd9", [0x069a] = "\x93\xda", [0x069b] = "\x93\xdb",
  [0x069c] = "\x93\xdc", [0x069d] = "\x93\xdd", [0x069e] = "\x93\xde",
  [0x069f] = "\x93\xdf", [0x06a0] = "\x93\xe0", [0x06a1] = "\x93\xe1",
  [0x06a2] = "\x93\xe2", [0x06a3] = "\x93\xe3", [0x06a4] = "\x93\xe4",
  [0x06a5] = "\x93\xe5", [0x06a6] = "\x93\xe6", [0x06a7] = "\x93\xe7",
  [0x06a8] = "\x93\xe8", [0x06a9] = "\x93\xe9", [0x06aa] = "\x93\xea",
  [0x06ab] = "\x93\xeb", [0x06ac] = "\x93\xec", [0x06ad] = "\x93\xed",
  [0x06ae] = "\x93\xee", [0x06af] = "\x93\xef", [0x06b0] = "\x93\xf0",
  [0x06b1] = "\x93\xf1", [0x06b2] = "\x93\xf2", [0x06b3] = "\x93\xf3",
  [0x06b4] = "\x93\xf4", [0x06b5] = "\x93\xf5", [0x06b6] = "\x93\xf6",
  [0x06b7] = "\x93\xf7", [0x06b8] = "\x93\xf8", [0x06b9] = "\x93\xf9",
  [0x06ba] = "\x93\xfa", [0x06bb] = "\x93\xfb", [0x06bc] = "\x93\xfc",
  [0x06bd] = "\x93\xfd", [0x06be] = "\x93\xfe", [0x06bf] = "\x94\x40",
  [0x06c0] = "\x94\x41", [0x06c1] = "\x94\x42", [0x06c2] = "\x94\x43",
  [0x06c3] = "\x94\x44", [0x06c4] = "\x94\x45", [0x06c5] = "\x94\x46",
  [0x06c6] = "\x94\x47", [0x06c7] = "\x94\x48", [0x06c8] = "\x94\x49",
  [0x06c9] = "\x94\x4a", [0x06ca] = "\x94\x4b", [0x06cb] = "\x94\x4c",
  [0x06cc] = "\x94\x4d", [0x06cd] = "\x94\x4e", [0x06ce] = "\x94\x4f",
  [0x06cf] = "\x94\x50", [0x06d0] = "\x94\x51", [0x06d1] = "\x94\x52",
  [0x06d2] = "\x94\x53", [0x06d3] = "\x94\x54", [0x06d4] = "\x94\x55",
  [0x06d5] = "\x94\x56", [0x06d6] = "\x94\x57", [0x06d7] = "\x94\x58",
  [0x06d8] = "\x94\x59", [0x06d9] = "\x94\x5a", [0x06da] = "\x94\x5b",
  [0x06db] = "\x94\x5c", [0x06dc] = "\x94\x5d", [0x06dd] = "\x94\x5e",
  [0x06de] = "\x94\x5f", [0x06df] = "\x94\x60", [0x06e0] = "\x94\x61",
  [0x06e1] = "\x94\x62", [0x06e2] = "\x94\x63", [0x06e3] = "\x94\x64",
  [0x06e4] = "\x94\x65", [0x06e5] = "\x94\x66", [0x06e6] = "\x94\x67",
  [0x06e7] = "\x94\x68", [0x06e8] = "\x94\x69", [0x06e9] = "\x94\x6a",
  [0x06ea] = "\x94\x6b", [0x06eb] = "\x94\x6c", [0x06ec] = "\x94\x6d",
  [0x06ed] = "\x94\x6e", [0x06ee] = "\x94\x6f", [0x06ef] = "\x94\x70",
  [0x06f0] = "\x94\x71", [0x06f1] = "\x94\x72", [0x06f2] = "\x94\x73",
  [0x06f3] = "\x94\x74", [0x06f4] = "\x94\x75", [0x06f5] = "\x94\x76",
  [0x06f6] = "\x94\x77", [0x06f7] = "\x94\x78", [0x06f8] = "\x94\x79",
  [0x06f9] = "\x94\x7a", [0x06fa] = "\x94\x7b", [0x06fb] = "\x94\x7c",
  [0x06fc] = "\x94\x7d", [0x06fd] = "\x94\x7e", [0x06fe] = "\x94\xa1",
  [0x06ff] = "\x94\xa2", [0x0700] = "\x94\xa3", [0x0701] = "\x94\xa4",
  [0x0702] = "\x94\xa5", [0x0703] = "\x94\xa6", [0x0704] = "\x94\xa7",
  [0x0705] = "\x94\xa8", [0x0706] = "\x94\xa9", [0x0707] = "\x94\xaa",
  [0x0708] = "\x94\xab", [0x0709] = "\x94\xac", [0x070a] = "\x94\xad",
  [0x070b] = "\x94\xae", [0x070c] = "\x94\xaf", [0x070d] = "\x94\xb0",
  [0x070e] = "\x94\xb1", [0x070f] = "\x94\xb2", [0x0710] = "\x94\xb3",
  [0x0711] = "\x94\xb4", [0x0712] = "\x94\xb5", [0x0713] = "\x94\xb6",
  [0x0714] = "\x94\xb7", [0x0715] = "\x94\xb8", [0x0716] = "\x94\xb9",
  [0x0717] = "\x94\xba", [0x0718] = "\x94\xbb", [0x0719] = "\x94\xbc",
  [0x071a] = "\x94\xbd", [0x071b] = "\x94\xbe", [0x071c] = "\x94\xbf",
  [0x071d] = "\x94\xc0", [0x071e] = "\x94\xc1", [0x071f] = "\x94\xc2",
  [0x0720] = "\x94\xc3", [0x0721] = "\x94\xc4", [0x0722] = "\x94\xc5",
  [0x0723] = "\x94\xc6", [0x0724] = "\x94\xc7", [0x0725] = "\x94\xc8",
  [0x0726] = "\x94\xc9", [0x0727] = "\x94\xca", [0x0728] = "\x94\xcb",
  [0x0729] = "\x94\xcc", [0x072a] = "\x94\xcd", [0x072b] = "\x94\xce",
  [0x072c] = "\x94\xcf", [0x072d] = "\x94\xd0", [0x072e] = "\x94\xd1",
  [0x072f] = "\x94\xd2", [0x0730] = "\x94\xd3", [0x0731] = "\x94\xd4",
  [0x0732] = "\x94\xd5", [0x0733] = "\x94\xd6", [0x0734] = "\x94\xd7",
  [0x0735] = "\x94\xd8", [0x0736] = "\x94\xd9", [0x0737] = "\x94\xda",
  [0x0738] = "\x94\xdb", [0x0739] = "\x94\xdc", [0x073a] = "\x94\xdd",
  [0x073b] = "\x94\xde", [0x073c] = "\x94\xdf", [0x073d] = "\x94\xe0",
  [0x073e] = "\x94\xe1", [0x073f] = "\x94\xe2", [0x0740] = "\x94\xe3",
  [0x0741] = "\x94\xe4", [0x0742] = "\x94\xe5", [0x0743] = "\x94\xe6",
  [0x0744] = "\x94\xe7", [0x0745] = "\x94\xe8", [0x0746] = "\x94\xe9",
  [0x0747] = "\x94\xea", [0x0748] = "\x94\xeb", [0x0749] = "\x94\xec",
  [0x074a] = "\x94\xed", [0x074b] = "\x94\xee", [0x074c] = "\x94\xef",
  [0x074d] = "\x94\xf0", [0x074e] = "\x94\xf1", [0x074f] = "\x94\xf2",
  [0x0750] = "\x94\xf3", [0x0751] = "\x94\xf4", [0x0752] = "\x94\xf5",
  [0x0753] = "\x94\xf6", [0x0754] = "\x94\xf7", [0x0755] = "\x94\xf8",
  [0x0756] = "\x94\xf9", [0x0757] = "\x94\xfa", [0x0758] = "\x94\xfb",
  [0x0759] = "\x94\xfc", [0x075a] = "\x94\xfd", [0x075b] = "\x94\xfe",
  [0x075c] = "\x95\x40", [0x075d] = "\x95\x41", [0x075e] = "\x95\x42",
  [0x075f] = "\x95\x43", [0x0760] = "\x95\x44", [0x0761] = "\x95\x45",
  [0x0762] = "\x95\x46", [0x0763] = "\x95\x47", [0x0764] = "\x95\x48",
  [0x0765] = "\x95\x49", [0x0766] = "\x95\x4a", [0x0767] = "\x95\x4b",
  [0x0768] = "\x95\x4c", [0x0769] = "\x95\x4d", [0x076a] = "\x95\x4e",
  [0x076b] = "\x95\x4f", [0x076c] = "\x95\x50", [0x076d] = "\x95\x51",
  [0x076e] = "\x95\x52", [0x076f] = "\x95\x53", [0x0770] = "\x95\x54",
  [0x0771] = "\x95\x55", [0x0772] = "\x95\x56", [0x0773] = "\x95\x57",
  [0x0774] = "\x95\x58", [0x0775] = "\x95\x59", [0x0776] = "\x95\x5a",
  [0x0777] = "\x95\x5b", [0x0778] = "\x95\x5c", [0x0779] = "\x95\x5d",
  [0x077a] = "\x95\x5e", [0x077b] = "\x95\x5f", [0x077c] = "\x95\x60",
  [0x077d] = "\x95\x61", [0x077e] = "\x95\x62", [0x077f] = "\x95\x63",
  [0x0780] = "\x95\x64", [0x0781] = "\x95\x65", [0x0782] = "\x95\x66",
  [0x0783] = "\x95\x67", [0x0784] = "\x95\x68", [0x0785] = "\x95\x69",
  [0x0786] = "\x95\x6a", [0x0787] = "\x95\x6b", [0x0788] = "\x95\x6c",
  [0x0789] = "\x95\x6d", [0x078a] = "\x95\x6e", [0x078b] = "\x95\x6f",
  [0x078c] = "\x95\x70", [0x078d] = "\x95\x71", [0x078e] = "\x95\x72",
  [0x078f] = "\x95\x73", [0x0790] = "\x95\x74", [0x0791] = "\x95\x75",
  [0x0792] = "\x95\x76", [0x0793] = "\x95\x77", [0x0794] = "\x95\x78",
  [0x0795] = "\x95\x79", [0x0796] = "\x95\x7a", [0x0797] = "\x95\x7b",
  [0x0798] = "\x95\x7c", [0x0799] = "\x95\x7d", [0x079a] = "\x95\x7e",
  [0x079b] = "\x95\xa1", [0x079c] = "\x95\xa2", [0x079d] = "\x95\xa3",
  [0x079e] = "\x95\xa4", [0x079f] = "\x95\xa5", [0x07a0] = "\x95\xa6",
  [0x07a1] = "\x95\xa7", [0x07a2] = "\x95\xa8", [0x07a3] = "\x95\xa9",
  [0x07a4] = "\x95\xaa", [0x07a5] = "\x95\xab", [0x07a6] = "\x95\xac",
  [0x07a7] = "\x95\xad", [0x07a8] = "\x95\xae", [0x07a9] = "\x95\xaf",
  [0x07aa] = "\x95\xb0", [0x07ab] = "\x95\xb1", [0x07ac] = "\x95\xb2",
  [0x07ad] = "\x95\xb3", [0x07ae] = "\x95\xb4", [0x07af] = "\x95\xb5",
  [0x07b0] = "\x95\xb6", [0x07b1] = "\x95\xb7", [0x07b2] = "\x95\xb8",
  [0x07b3] = "\x95\xb9", [0x07b4] = "\x95\xba", [0x07b5] = "\x95\xbb",
  [0x07b6] = "\x95\xbc", [0x07b7] = "\x95\xbd", [0x07b8] = "\x95\xbe",
  [0x07b9] = "\x95\xbf", [0x07ba] = "\x95\xc0", [0x07bb] = "\x95\xc1",
  [0x07bc] = "\x95\xc2", [0x07bd] = "\x95\xc3", [0x07be] = "\x95\xc4",
  [0x07bf] = "\x95\xc5", [0x07c0] = "\x95\xc6", [0x07c1] = "\x95\xc7",
  [0x07c2] = "\x95\xc8", [0x07c3] = "\x95\xc9", [0x07c4] = "\x95\xca",
  [0x07c5] = "\x95\xcb", [0x07c6] = "\x95\xcc", [0x07c7] = "\x95\xcd",
  [0x07c8] = "\x95\xce", [0x07c9] = "\x95\xcf", [0x07ca] = "\x95\xd0",
  [0x07cb] = "\x95\xd1", [0x07cc] = "\x95\xd2", [0x07cd] = "\x95\xd3",
  [0x07ce] = "\x95\xd4", [0x07cf] = "\x95\xd5", [0x07d0] = "\x95\xd6",
  [0x07d1] = "\x95\xd7", [0x07d2] = "\x95\xd8", [0x07d3] = "\x95\xd9",
  [0x07d4] = "\x95\xda", [0x07d5] = "\x95\xdb", [0x07d6] = "\x95\xdc",
  [0x07d7] = "\x95\xdd", [0x07d8] = "\x95\xde", [0x07d9] = "\x95\xdf",
  [0x07da] = "\x95\xe0", [0x07db] = "\x95\xe1", [0x07dc] = "\x95\xe2",
  [0x07dd] = "\x95\xe3", [0x07de] = "\x95\xe4", [0x07df] = "\x95\xe5",
  [0x07e0] = "\x95\xe6", [0x07e1] = "\x95\xe7", [0x07e2] = "\x95\xe8",
  [0x07e3] = "\x95\xe9", [0x07e4] = "\x95\xea", [0x07e5] = "\x95\xeb",
  [0x07e6] = "\x95\xec", [0x07e7] = "\x95\xed", [0x07e8] = "\x95\xee",
  [0x07e9] = "\x95\xef", [0x07ea] = "\x95\xf0", [0x07eb] = "\x95\xf1",
  [0x07ec] = "\x95\xf2", [0x07ed] = "\x95\xf3", [0x07ee] = "\x95\xf4",
  [0x07ef] = "\x95\xf5", [0x07f0] = "\x95\xf6", [0x07f1] = "\x95\xf7",
  [0x07f2] = "\x95\xf8", [0x07f3] = "\x95\xf9", [0x07f4] = "\x95\xfa",
  [0x07f5] = "\x95\xfb", [0x07f6] = "\x95\xfc", [0x07f7] = "\x95\xfd",
  [0x07f8] = "\x95\xfe", [0x07f9] = "\x96\x40", [0x07fa] = "\x96\x41",
  [0x07fb] = "\x96\x42", [0x07fc] = "\x96\x43", [0x07fd] = "\x96\x44",
  [0x07fe] = "\x96\x45", [0x07ff] = "\x96\x46", [0x0800] = "\x96\x47",
  [0x0801] = "\x96\x48", [0x0802] = "\x96\x49", [0x0803] = "\x96\x4a",
  [0x0804] = "\x96\x4b", [0x0805] = "\x96\x4c", [0x0806] = "\x96\x4d",
  [0x0807] = "\x96\x4e", [0x0808] = "\x96\x4f", [0x0809] = "\x96\x50",
  [0x080a] = "\x96\x51", [0x080b] = "\x96\x52", [0x080c] = "\x96\x53",
  [0x080d] = "\x96\x54", [0x080e] = "\x96\x55", [0x080f] = "\x96\x56",
  [0x0810] = "\x96\x57", [0x0811] = "\x96\x58", [0x0812] = "\x96\x59",
  [0x0813] = "\x96\x5a", [0x0814] = "\x96\x5b", [0x0815] = "\x96\x5c",
  [0x0816] = "\x96\x5d", [0x0817] = "\x96\x5e", [0x0818] = "\x96\x5f",
  [0x0819] = "\x96\x60", [0x081a] = "\x96\x61", [0x081b] = "\x96\x62",
  [0x081c] = "\x96\x63", [0x081d] = "\x96\x64", [0x081e] = "\x96\x65",
  [0x081f] = "\x96\x66", [0x0820] = "\x96\x67", [0x0821] = "\x96\x68",
  [0x0822] = "\x96\x69", [0x0823] = "\x96\x6a", [0x0824] = "\x96\x6b",
  [0x0825] = "\x96\x6c", [0x0826] = "\x96\x6d", [0x0827] = "\x96\x6e",
  [0x0828] = "\x96\x6f", [0x0829] = "\x96\x70", [0x082a] = "\x96\x71",
  [0x082b] = "\x96\x72", [0x082c] = "\x96\x73", [0x082d] = "\x96\x74",
  [0x082e] = "\x96\x75", [0x082f] = "\x96\x76", [0x0830] = "\x96\x77",
  [0x0831] = "\x96\x78", [0x0832] = "\x96\x79", [0x0833] = "\x96\x7a",
  [0x0834] = "\x96\x7b", [0x0835] = "\x96\x7c", [0x0836] = "\x96\x7d",
  [0x0837] = "\x96\x7e", [0x0838] = "\x96\xa1", [0x0839] = "\x96\xa2",
  [0x083a] = "\x96\xa3", [0x083b] = "\x96\xa4", [0x083c] = "\x96\xa5",
  [0x083d] = "\x96\xa6", [0x083e] = "\x96\xa7", [0x083f] = "\x96\xa8",
  [0x0840] = "\x96\xa9", [0x0841] = "\x96\xaa", [0x0842] = "\x96\xab",
  [0x0843] = "\x96\xac", [0x0844] = "\x96\xad", [0x0845] = "\x96\xae",
  [0x0846] = "\x96\xaf", [0x0847] = "\x96\xb0", [0x0848] = "\x96\xb1",
  [0x0849] = "\x96\xb2", [0x084a] = "\x96\xb3", [0x084b] = "\x96\xb4",
  [0x084c] = "\x96\xb5", [0x084d] = "\x96\xb6", [0x084e] = "\x96\xb7",
  [0x084f] = "\x96\xb8", [0x0850] = "\x96\xb9", [0x0851] = "\x96\xba",
  [0x0852] = "\x96\xbb", [0x0853] = "\x96\xbc", [0x0854] = "\x96\xbd",
  [0x0855] = "\x96\xbe", [0x0856] = "\x96\xbf", [0x0857] = "\x96\xc0",
  [0x0858] = "\x96\xc1", [0x0859] = "\x96\xc2", [0x085a] = "\x96\xc3",
  [0x085b] = "\x96\xc4", [0x085c] = "\x96\xc5", [0x085d] = "\x96\xc6",
  [0x085e] = "\x96\xc7", [0x085f] = "\x96\xc8", [0x0860] = "\x96\xc9",
  [0x0861] = "\x96\xca", [0x0862] = "\x96\xcb", [0x0863] = "\x96\xcc",
  [0x0864] = "\x96\xcd", [0x0865] = "\x96\xce", [0x0866] = "\x96\xcf",
  [0x0867] = "\x96\xd0", [0x0868] = "\x96\xd1", [0x0869] = "\x96\xd2",
  [0x086a] = "\x96\xd3", [0x086b] = "\x96\xd4", [0x086c] = "\x96\xd5",
  [0x086d] = "\x96\xd6", [0x086e] = "\x96\xd7", [0x086f] = "\x96\xd8",
  [0x0870] = "\x96\xd9", [0x0871] = "\x96\xda", [0x0872] = "\x96\xdb",
  [0x0873] = "\x96\xdc", [0x0874] = "\x96\xdd", [0x0875] = "\x96\xde",
  [0x0876] = "\x96\xdf", [0x0877] = "\x96\xe0", [0x0878] = "\x96\xe1",
  [0x0879] = "\x96\xe2", [0x087a] = "\x96\xe3", [0x087b] = "\x96\xe4",
  [0x087c] = "\x96\xe5", [0x087d] = "\x96\xe6", [0x087e] = "\x96\xe7",
  [0x087f] = "\x96\xe8", [0x0880] = "\x96\xe9", [0x0881] = "\x96\xea",
  [0x0882] = "\x96\xeb", [0x0883] = "\x96\xec", [0x0884] = "\x96\xed",
  [0x0885] = "\x96\xee", [0x0886] = "\x96\xef", [0x0887] = "\x96\xf0",
  [0x0888] = "\x96\xf1", [0x0889] = "\x96\xf2", [0x088a] = "\x96\xf3",
  [0x088b] = "\x96\xf4", [0x088c] = "\x96\xf5", [0x088d] = "\x96\xf6",
  [0x088e] = "\x96\xf7", [0x088f] = "\x96\xf8", [0x0890] = "\x96\xf9",
  [0x0891] = "\x96\xfa", [0x0892] = "\x96\xfb", [0x0893] = "\x96\xfc",
  [0x0894] = "\x96\xfd", [0x0895] = "\x96\xfe", [0x0896] = "\x97\x40",
  [0x0897] = "\x97\x41", [0x0898] = "\x97\x42", [0x0899] = "\x97\x43",
  [0x089a] = "\x97\x44", [0x089b] = "\x97\x45", [0x089c] = "\x97\x46",
  [0x089d] = "\x97\x47", [0x089e] = "\x97\x48", [0x089f] = "\x97\x49",
  [0x08a0] = "\x97\x4a", [0x08a1] = "\x97\x4b", [0x08a2] = "\x97\x4c",
  [0x08a3] = "\x97\x4d", [0x08a4] = "\x97\x4e", [0x08a5] = "\x97\x4f",
  [0x08a6] = "\x97\x50", [0x08a7] = "\x97\x51", [0x08a8] = "\x97\x52",
  [0x08a9] = "\x97\x53", [0x08aa] = "\x97\x54", [0x08ab] = "\x97\x55",
  [0x08ac] = "\x97\x56", [0x08ad] = "\x97\x57", [0x08ae] = "\x97\x58",
  [0x08af] = "\x97\x59", [0x08b0] = "\x97\x5a", [0x08b1] = "\x97\x5b",
  [0x08b2] = "\x97\x5c", [0x08b3] = "\x97\x5d", [0x08b4] = "\x97\x5e",
  [0x08b5] = "\x97\x5f", [0x08b6] = "\x97\x60", [0x08b7] = "\x97\x61",
  [0x08b8] = "\x97\x62", [0x08b9] = "\x97\x63", [0x08ba] = "\x97\x64",
  [0x08bb] = "\x97\x65", [0x08bc] = "\x97\x66", [0x08bd] = "\x97\x67",
  [0x08be] = "\x97\x68", [0x08bf] = "\x97\x69", [0x08c0] = "\x97\x6a",
  [0x08c1] = "\x97\x6b", [0x08c2] = "\x97\x6c", [0x08c3] = "\x97\x6d",
  [0x08c4] = "\x97\x6e", [0x08c5] = "\x97\x6f", [0x08c6] = "\x97\x70",
  [0x08c7] = "\x97\x71", [0x08c8] = "\x97\x72", [0x08c9] = "\x97\x73",
  [0x08ca] = "\x97\x74", [0x08cb] = "\x97\x75", [0x08cc] = "\x97\x76",
  [0x08cd] = "\x97\x77", [0x08ce] = "\x97\x78", [0x08cf] = "\x97\x79",
  [0x08d0] = "\x97\x7a", [0x08d1] = "\x97\x7b", [0x08d2] = "\x97\x7c",
  [0x08d3] = "\x97\x7d", [0x08d4] = "\x97\x7e", [0x08d5] = "\x97\xa1",
  [0x08d6] = "\x97\xa2", [0x08d7] = "\x97\xa3", [0x08d8] = "\x97\xa4",
  [0x08d9] = "\x97\xa5", [0x08da] = "\x97\xa6", [0x08db] = "\x97\xa7",
  [0x08dc] = "\x97\xa8", [0x08dd] = "\x97\xa9", [0x08de] = "\x97\xaa",
  [0x08df] = "\x97\xab", [0x08e0] = "\x97\xac", [0x08e1] = "\x97\xad",
  [0x08e2] = "\x97\xae", [0x08e3] = "\x97\xaf", [0x08e4] = "\x97\xb0",
  [0x08e5] = "\x97\xb1", [0x08e6] = "\x97\xb2", [0x08e7] = "\x97\xb3",
  [0x08e8] = "\x97\xb4", [0x08e9] = "\x97\xb5", [0x08ea] = "\x97\xb6",
  [0x08eb] = "\x97\xb7", [0x08ec] = "\x97\xb8", [0x08ed] = "\x97\xb9",
  [0x08ee] = "\x97\xba", [0x08ef] = "\x97\xbb", [0x08f0] = "\x97\xbc",
  [0x08f1] = "\x97\xbd", [0x08f2] = "\x97\xbe", [0x08f3] = "\x97\xbf",
  [0x08f4] = "\x97\xc0", [0x08f5] = "\x97\xc1", [0x08f6] = "\x97\xc2",
  [0x08f7] = "\x97\xc3", [0x08f8] = "\x97\xc4", [0x08f9] = "\x97\xc5",
  [0x08fa] = "\x97\xc6", [0x08fb] = "\x97\xc7", [0x08fc] = "\x97\xc8",
  [0x08fd] = "\x97\xc9", [0x08fe] = "\x97\xca", [0x08ff] = "\x97\xcb",
  [0x0900] = "\x97\xcc", [0x0901] = "\x97\xcd", [0x0902] = "\x97\xce",
  [0x0903] = "\x97\xcf", [0x0904] = "\x97\xd0", [0x0905] = "\x97\xd1",
  [0x0906] = "\x97\xd2", [0x0907] = "\x97\xd3", [0x0908] = "\x97\xd4",
  [0x0909] = "\x97\xd5", [0x090a] = "\x97\xd6", [0x090b] = "\x97\xd7",
  [0x090c] = "\x97\xd8", [0x090d] = "\x97\xd9", [0x090e] = "\x97\xda",
  [0x090f] = "\x97\xdb", [0x0910] = "\x97\xdc", [0x0911] = "\x97\xdd",
  [0x0912] = "\x97\xde", [0x0913] = "\x97\xdf", [0x0914] = "\x97\xe0",
  [0x0915] = "\x97\xe1", [0x0916] = "\x97\xe2", [0x0917] = "\x97\xe3",
  [0x0918] = "\x97\xe4", [0x0919] = "\x97\xe5", [0x091a] = "\x97\xe6",
  [0x091b] = "\x97\xe7", [0x091c] = "\x97\xe8", [0x091d] = "\x97\xe9",
  [0x091e] = "\x97\xea", [0x091f] = "\x97\xeb", [0x0920] = "\x97\xec",
  [0x0921] = "\x97\xed", [0x0922] = "\x97\xee", [0x0923] = "\x97\xef",
  [0x0924] = "\x97\xf0", [0x0925] = "\x97\xf1", [0x0926] = "\x97\xf2",
  [0x0927] = "\x97\xf3", [0x0928] = "\x97\xf4", [0x0929] = "\x97\xf5",
  [0x092a] = "\x97\xf6", [0x092b] = "\x97\xf7", [0x092c] = "\x97\xf8",
  [0x092d] = "\x97\xf9", [0x092e] = "\x97\xfa", [0x092f] = "\x97\xfb",
  [0x0930] = "\x97\xfc", [0x0931] = "\x97\xfd", [0x0932] = "\x97\xfe",
  [0x0933] = "\x98\x40", [0x0934] = "\x98\x41", [0x0935] = "\x98\x42",
  [0x0936] = "\x98\x43", [0x0937] = "\x98\x44", [0x0938] = "\x98\x45",
  [0x0939] = "\x98\x46", [0x093a] = "\x98\x47", [0x093b] = "\x98\x48",
  [0x093c] = "\x98\x49", [0x093d] = "\x98\x4a", [0x093e] = "\x98\x4b",
  [0x093f] = "\x98\x4c", [0x0940] = "\x98\x4d", [0x0941] = "\x98\x4e",
  [0x0942] = "\x98\x4f", [0x0943] = "\x98\x50", [0x0944] = "\x98\x51",
  [0x0945] = "\x98\x52", [0x0946] = "\x98\x53", [0x0947] = "\x98\x54",
  [0x0948] = "\x98\x55", [0x0949] = "\x98\x56", [0x094a] = "\x98\x57",
  [0x094b] = "\x98\x58", [0x094c] = "\x98\x59", [0x094d] = "\x98\x5a",
  [0x094e] = "\x98\x5b", [0x094f] = "\x98\x5c", [0x0950] = "\x98\x5d",
  [0x0951] = "\x98\x5e", [0x0952] = "\x98\x5f", [0x0953] = "\x98\x60",
  [0x0954] = "\x98\x61", [0x0955] = "\x98\x62", [0x0956] = "\x98\x63",
  [0x0957] = "\x98\x64", [0x0958] = "\x98\x65", [0x0959] = "\x98\x66",
  [0x095a] = "\x98\x67", [0x095b] = "\x98\x68", [0x095c] = "\x98\x69",
  [0x095d] = "\x98\x6a", [0x095e] = "\x98\x6b", [0x095f] = "\x98\x6c",
  [0x0960] = "\x98\x6d", [0x0961] = "\x98\x6e", [0x0962] = "\x98\x6f",
  [0x0963] = "\x98\x70", [0x0964] = "\x98\x71", [0x0965] = "\x98\x72",
  [0x0966] = "\x98\x73", [0x0967] = "\x98\x74", [0x0968] = "\x98\x75",
  [0x0969] = "\x98\x76", [0x096a] = "\x98\x77", [0x096b] = "\x98\x78",
  [0x096c] = "\x98\x79", [0x096d] = "\x98\x7a", [0x096e] = "\x98\x7b",
  [0x096f] = "\x98\x7c", [0x0970] = "\x98\x7d", [0x0971] = "\x98\x7e",
  [0x0972] = "\x98\xa1", [0x0973] = "\x98\xa2", [0x0974] = "\x98\xa3",
  [0x0975] = "\x98\xa4", [0x0976] = "\x98\xa5", [0x0977] = "\x98\xa6",
  [0x0978] = "\x98\xa7", [0x0979] = "\x98\xa8", [0x097a] = "\x98\xa9",
  [0x097b] = "\x98\xaa", [0x097c] = "\x98\xab", [0x097d] = "\x98\xac",
  [0x097e] = "\x98\xad", [0x097f] = "\x98\xae", [0x0980] = "\x98\xaf",
  [0x0981] = "\x98\xb0", [0x0982] = "\x98\xb1", [0x0983] = "\x98\xb2",
  [0x0984] = "\x98\xb3", [0x0985] = "\x98\xb4", [0x0986] = "\x98\xb5",
  [0x0987] = "\x98\xb6", [0x0988] = "\x98\xb7", [0x0989] = "\x98\xb8",
  [0x098a] = "\x98\xb9", [0x098b] = "\x98\xba", [0x098c] = "\x98\xbb",
  [0x098d] = "\x98\xbc", [0x098e] = "\x98\xbd", [0x098f] = "\x98\xbe",
  [0x0990] = "\x98\xbf", [0x0991] = "\x98\xc0", [0x0992] = "\x98\xc1",
  [0x0993] = "\x98\xc2", [0x0994] = "\x98\xc3", [0x0995] = "\x98\xc4",
  [0x0996] = "\x98\xc5", [0x0997] = "\x98\xc6", [0x0998] = "\x98\xc7",
  [0x0999] = "\x98\xc8", [0x099a] = "\x98\xc9", [0x099b] = "\x98\xca",
  [0x099c] = "\x98\xcb", [0x099d] = "\x98\xcc", [0x099e] = "\x98\xcd",
  [0x099f] = "\x98\xce", [0x09a0] = "\x98\xcf", [0x09a1] = "\x98\xd0",
  [0x09a2] = "\x98\xd1", [0x09a3] = "\x98\xd2", [0x09a4] = "\x98\xd3",
  [0x09a5] = "\x98\xd4", [0x09a6] = "\x98\xd5", [0x09a7] = "\x98\xd6",
  [0x09a8] = "\x98\xd7", [0x09a9] = "\x98\xd8", [0x09aa] = "\x98\xd9",
  [0x09ab] = "\x98\xda", [0x09ac] = "\x98\xdb", [0x09ad] = "\x98\xdc",
  [0x09ae] = "\x98\xdd", [0x09af] = "\x98\xde", [0x09b0] = "\x98\xdf",
  [0x09b1] = "\x98\xe0", [0x09b2] = "\x98\xe1", [0x09b3] = "\x98\xe2",
  [0x09b4] = "\x98\xe3", [0x09b5] = "\x98\xe4", [0x09b6] = "\x98\xe5",
  [0x09b7] = "\x98\xe6", [0x09b8] = "\x98\xe7", [0x09b9] = "\x98\xe8",
  [0x09ba] = "\x98\xe9", [0x09bb] = "\x98\xea", [0x09bc] = "\x98\xeb",
  [0x09bd] = "\x98\xec", [0x09be] = "\x98\xed", [0x09bf] = "\x98\xee",
  [0x09c0] = "\x98\xef", [0x09c1] = "\x98\xf0", [0x09c2] = "\x98\xf1",
  [0x09c3] = "\x98\xf2", [0x09c4] = "\x98\xf3", [0x09c5] = "\x98\xf4",
  [0x09c6] = "\x98\xf5", [0x09c7] = "\x98\xf6", [0x09c8] = "\x98\xf7",
  [0x09c9] = "\x98\xf8", [0x09ca] = "\x98\xf9", [0x09cb] = "\x98\xfa",
  [0x09cc] = "\x98\xfb", [0x09cd] = "\x98\xfc", [0x09ce] = "\x98\xfd",
  [0x09cf] = "\x98\xfe", [0x09d0] = "\x99\x40", [0x09d1] = "\x99\x41",
  [0x09d2] = "\x99\x42", [0x09d3] = "\x99\x43", [0x09d4] = "\x99\x44",
  [0x09d5] = "\x99\x45", [0x09d6] = "\x99\x46", [0x09d7] = "\x99\x47",
  [0x09d8] = "\x99\x48", [0x09d9] = "\x99\x49", [0x09da] = "\x99\x4a",
  [0x09db] = "\x99\x4b", [0x09dc] = "\x99\x4c", [0x09dd] = "\x99\x4d",
  [0x09de] = "\x99\x4e", [0x09df] = "\x99\x4f", [0x09e0] = "\x99\x50",
  [0x09e1] = "\x99\x51", [0x09e2] = "\x99\x52", [0x09e3] = "\x99\x53",
  [0x09e4] = "\x99\x54", [0x09e5] = "\x99\x55", [0x09e6] = "\x99\x56",
  [0x09e7] = "\x99\x57", [0x09e8] = "\x99\x58", [0x09e9] = "\x99\x59",
  [0x09ea] = "\x99\x5a", [0x09eb] = "\x99\x5b", [0x09ec] = "\x99\x5c",
  [0x09ed] = "\x99\x5d", [0x09ee] = "\x99\x5e", [0x09ef] = "\x99\x5f",
  [0x09f0] = "\x99\x60", [0x09f1] = "\x99\x61", [0x09f2] = "\x99\x62",
  [0x09f3] = "\x99\x63", [0x09f4] = "\x99\x64", [0x09f5] = "\x99\x65",
  [0x09f6] = "\x99\x66", [0x09f7] = "\x99\x67", [0x09f8] = "\x99\x68",
  [0x09f9] = "\x99\x69", [0x09fa] = "\x99\x6a", [0x09fb] = "\x99\x6b",
  [0x09fc] = "\x99\x6c", [0x09fd] = "\x99\x6d", [0x09fe] = "\x99\x6e",
  [0x09ff] = "\x99\x6f", [0x0a00] = "\x99\x70", [0x0a01] = "\x99\x71",
  [0x0a02] = "\x99\x72", [0x0a03] = "\x99\x73", [0x0a04] = "\x99\x74",
  [0x0a05] = "\x99\x75", [0x0a06] = "\x99\x76", [0x0a07] = "\x99\x77",
  [0x0a08] = "\x99\x78", [0x0a09] = "\x99\x79", [0x0a0a] = "\x99\x7a",
  [0x0a0b] = "\x99\x7b", [0x0a0c] = "\x99\x7c", [0x0a0d] = "\x99\x7d",
  [0x0a0e] = "\x99\x7e", [0x0a0f] = "\x99\xa1", [0x0a10] = "\x99\xa2",
  [0x0a11] = "\x99\xa3", [0x0a12] = "\x99\xa4", [0x0a13] = "\x99\xa5",
  [0x0a14] = "\x99\xa6", [0x0a15] = "\x99\xa7", [0x0a16] = "\x99\xa8",
  [0x0a17] = "\x99\xa9", [0x0a18] = "\x99\xaa", [0x0a19] = "\x99\xab",
  [0x0a1a] = "\x99\xac", [0x0a1b] = "\x99\xad", [0x0a1c] = "\x99\xae",
  [0x0a1d] = "\x99\xaf", [0x0a1e] = "\x99\xb0", [0x0a1f] = "\x99\xb1",
  [0x0a20] = "\x99\xb2", [0x0a21] = "\x99\xb3", [0x0a22] = "\x99\xb4",
  [0x0a23] = "\x99\xb5", [0x0a24] = "\x99\xb6", [0x0a25] = "\x99\xb7",
  [0x0a26] = "\x99\xb8", [0x0a27] = "\x99\xb9", [0x0a28] = "\x99\xba",
  [0x0a29] = "\x99\xbb", [0x0a2a] = "\x99\xbc", [0x0a2b] = "\x99\xbd",
  [0x0a2c] = "\x99\xbe", [0x0a2d] = "\x99\xbf", [0x0a2e] = "\x99\xc0",
  [0x0a2f] = "\x99\xc1", [0x0a30] = "\x99\xc2", [0x0a31] = "\x99\xc3",
  [0x0a32] = "\x99\xc4", [0x0a33] = "\x99\xc5", [0x0a34] = "\x99\xc6",
  [0x0a35] = "\x99\xc7", [0x0a36] = "\x99\xc8", [0x0a37] = "\x99\xc9",
  [0x0a38] = "\x99\xca", [0x0a39] = "\x99\xcb", [0x0a3a] = "\x99\xcc",
  [0x0a3b] = "\x99\xcd", [0x0a3c] = "\x99\xce", [0x0a3d] = "\x99\xcf",
  [0x0a3e] = "\x99\xd0", [0x0a3f] = "\x99\xd1", [0x0a40] = "\x99\xd2",
  [0x0a41] = "\x99\xd3", [0x0a42] = "\x99\xd4", [0x0a43] = "\x99\xd5",
  [0x0a44] = "\x99\xd6", [0x0a45] = "\x99\xd7", [0x0a46] = "\x99\xd8",
  [0x0a47] = "\x99\xd9", [0x0a48] = "\x99\xda", [0x0a49] = "\x99\xdb",
  [0x0a4a] = "\x99\xdc", [0x0a4b] = "\x99\xdd", [0x0a4c] = "\x99\xde",
  [0x0a4d] = "\x99\xdf", [0x0a4e] = "\x99\xe0", [0x0a4f] = "\x99\xe1",
  [0x0a50] = "\x99\xe2", [0x0a51] = "\x99\xe3", [0x0a52] = "\x99\xe4",
  [0x0a53] = "\x99\xe5", [0x0a54] = "\x99\xe6", [0x0a55] = "\x99\xe7",
  [0x0a56] = "\x99\xe8", [0x0a57] = "\x99\xe9", [0x0a58] = "\x99\xea",
  [0x0a59] = "\x99\xeb", [0x0a5a] = "\x99\xec", [0x0a5b] = "\x99\xed",
  [0x0a5c] = "\x99\xee", [0x0a5d] = "\x99\xef", [0x0a5e] = "\x99\xf0",
  [0x0a5f] = "\x99\xf1", [0x0a60] = "\x99\xf2", [0x0a61] = "\x99\xf3",
  [0x0a62] = "\x99\xf4", [0x0a63] = "\x99\xf5", [0x0a64] = "\x99\xf6",
  [0x0a65] = "\x99\xf7", [0x0a66] = "\x99\xf8", [0x0a67] = "\x99\xf9",
  [0x0a68] = "\x99\xfa", [0x0a69] = "\x99\xfb", [0x0a6a] = "\x99\xfc",
  [0x0a6b] = "\x99\xfd", [0x0a6c] = "\x99\xfe", [0x0a6d] = "\x9a\x40",
  [0x0a6e] = "\x9a\x41", [0x0a6f] = "\x9a\x42", [0x0a70] = "\x9a\x43",
  [0x0a71] = "\x9a\x44", [0x0a72] = "\x9a\x45", [0x0a73] = "\x9a\x46",
  [0x0a74] = "\x9a\x47", [0x0a75] = "\x9a\x48", [0x0a76] = "\x9a\x49",
  [0x0a77] = "\x9a\x4a", [0x0a78] = "\x9a\x4b", [0x0a79] = "\x9a\x4c",
  [0x0a7a] = "\x9a\x4d", [0x0a7b] = "\x9a\x4e", [0x0a7c] = "\x9a\x4f",
  [0x0a7d] = "\x9a\x50", [0x0a7e] = "\x9a\x51", [0x0a7f] = "\x9a\x52",
  [0x0a80] = "\x9a\x53", [0x0a81] = "\x9a\x54", [0x0a82] = "\x9a\x55",
  [0x0a83] = "\x9a\x56", [0x0a84] = "\x9a\x57", [0x0a85] = "\x9a\x58",
  [0x0a86] = "\x9a\x59", [0x0a87] = "\x9a\x5a", [0x0a88] = "\x9a\x5b",
  [0x0a89] = "\x9a\x5c", [0x0a8a] = "\x9a\x5d", [0x0a8b] = "\x9a\x5e",
  [0x0a8c] = "\x9a\x5f", [0x0a8d] = "\x9a\x60", [0x0a8e] = "\x9a\x61",
  [0x0a8f] = "\x9a\x62", [0x0a90] = "\x9a\x63", [0x0a91] = "\x9a\x64",
  [0x0a92] = "\x9a\x65", [0x0a93] = "\x9a\x66", [0x0a94] = "\x9a\x67",
  [0x0a95] = "\x9a\x68", [0x0a96] = "\x9a\x69", [0x0a97] = "\x9a\x6a",
  [0x0a98] = "\x9a\x6b", [0x0a99] = "\x9a\x6c", [0x0a9a] = "\x9a\x6d",
  [0x0a9b] = "\x9a\x6e", [0x0a9c] = "\x9a\x6f", [0x0a9d] = "\x9a\x70",
  [0x0a9e] = "\x9a\x71", [0x0a9f] = "\x9a\x72", [0x0aa0] = "\x9a\x73",
  [0x0aa1] = "\x9a\x74", [0x0aa2] = "\x9a\x75", [0x0aa3] = "\x9a\x76",
  [0x0aa4] = "\x9a\x77", [0x0aa5] = "\x9a\x78", [0x0aa6] = "\x9a\x79",
  [0x0aa7] = "\x9a\x7a", [0x0aa8] = "\x9a\x7b", [0x0aa9] = "\x9a\x7c",
  [0x0aaa] = "\x9a\x7d", [0x0aab] = "\x9a\x7e", [0x0aac] = "\x9a\xa1",
  [0x0aad] = "\x9a\xa2", [0x0aae] = "\x9a\xa3", [0x0aaf] = "\x9a\xa4",
  [0x0ab0] = "\x9a\xa5", [0x0ab1] = "\x9a\xa6", [0x0ab2] = "\x9a\xa7",
  [0x0ab3] = "\x9a\xa8", [0x0ab4] = "\x9a\xa9", [0x0ab5] = "\x9a\xaa",
  [0x0ab6] = "\x9a\xab", [0x0ab7] = "\x9a\xac", [0x0ab8] = "\x9a\xad",
  [0x0ab9] = "\x9a\xae", [0x0aba] = "\x9a\xaf", [0x0abb] = "\x9a\xb0",
  [0x0abc] = "\x9a\xb1", [0x0abd] = "\x9a\xb2", [0x0abe] = "\x9a\xb3",
  [0x0abf] = "\x9a\xb4", [0x0ac0] = "\x9a\xb5", [0x0ac1] = "\x9a\xb6",
  [0x0ac2] = "\x9a\xb7", [0x0ac3] = "\x9a\xb8", [0x0ac4] = "\x9a\xb9",
  [0x0ac5] = "\x9a\xba", [0x0ac6] = "\x9a\xbb", [0x0ac7] = "\x9a\xbc",
  [0x0ac8] = "\x9a\xbd", [0x0ac9] = "\x9a\xbe", [0x0aca] = "\x9a\xbf",
  [0x0acb] = "\x9a\xc0", [0x0acc] = "\x9a\xc1", [0x0acd] = "\x9a\xc2",
  [0x0ace] = "\x9a\xc3", [0x0acf] = "\x9a\xc4", [0x0ad0] = "\x9a\xc5",
  [0x0ad1] = "\x9a\xc6", [0x0ad2] = "\x9a\xc7", [0x0ad3] = "\x9a\xc8",
  [0x0ad4] = "\x9a\xc9", [0x0ad5] = "\x9a\xca", [0x0ad6] = "\x9a\xcb",
  [0x0ad7] = "\x9a\xcc", [0x0ad8] = "\x9a\xcd", [0x0ad9] = "\x9a\xce",
  [0x0ada] = "\x9a\xcf", [0x0adb] = "\x9a\xd0", [0x0adc] = "\x9a\xd1",
  [0x0add] = "\x9a\xd2", [0x0ade] = "\x9a\xd3", [0x0adf] = "\x9a\xd4",
  [0x0ae0] = "\x9a\xd5", [0x0ae1] = "\x9a\xd6", [0x0ae2] = "\x9a\xd7",
  [0x0ae3] = "\x9a\xd8", [0x0ae4] = "\x9a\xd9", [0x0ae5] = "\x9a\xda",
  [0x0ae6] = "\x9a\xdb", [0x0ae7] = "\x9a\xdc", [0x0ae8] = "\x9a\xdd",
  [0x0ae9] = "\x9a\xde", [0x0aea] = "\x9a\xdf", [0x0aeb] = "\x9a\xe0",
  [0x0aec] = "\x9a\xe1", [0x0aed] = "\x9a\xe2", [0x0aee] = "\x9a\xe3",
  [0x0aef] = "\x9a\xe4", [0x0af0] = "\x9a\xe5", [0x0af1] = "\x9a\xe6",
  [0x0af2] = "\x9a\xe7", [0x0af3] = "\x9a\xe8", [0x0af4] = "\x9a\xe9",
  [0x0af5] = "\x9a\xea", [0x0af6] = "\x9a\xeb", [0x0af7] = "\x9a\xec",
  [0x0af8] = "\x9a\xed", [0x0af9] = "\x9a\xee", [0x0afa] = "\x9a\xef",
  [0x0afb] = "\x9a\xf0", [0x0afc] = "\x9a\xf1", [0x0afd] = "\x9a\xf2",
  [0x0afe] = "\x9a\xf3", [0x0aff] = "\x9a\xf4", [0x0b00] = "\x9a\xf5",
  [0x0b01] = "\x9a\xf6", [0x0b02] = "\x9a\xf7", [0x0b03] = "\x9a\xf8",
  [0x0b04] = "\x9a\xf9", [0x0b05] = "\x9a\xfa", [0x0b06] = "\x9a\xfb",
  [0x0b07] = "\x9a\xfc", [0x0b08] = "\x9a\xfd", [0x0b09] = "\x9a\xfe",
  [0x0b0a] = "\x9b\x40", [0x0b0b] = "\x9b\x41", [0x0b0c] = "\x9b\x42",
  [0x0b0d] = "\x9b\x43", [0x0b0e] = "\x9b\x44", [0x0b0f] = "\x9b\x45",
  [0x0b10] = "\x9b\x46", [0x0b11] = "\x9b\x47", [0x0b12] = "\x9b\x48",
  [0x0b13] = "\x9b\x49", [0x0b14] = "\x9b\x4a", [0x0b15] = "\x9b\x4b",
  [0x0b16] = "\x9b\x4c", [0x0b17] = "\x9b\x4d", [0x0b18] = "\x9b\x4e",
  [0x0b19] = "\x9b\x4f", [0x0b1a] = "\x9b\x50", [0x0b1b] = "\x9b\x51",
  [0x0b1c] = "\x9b\x52", [0x0b1d] = "\x9b\x53", [0x0b1e] = "\x9b\x54",
  [0x0b1f] = "\x9b\x55", [0x0b20] = "\x9b\x56", [0x0b21] = "\x9b\x57",
  [0x0b22] = "\x9b\x58", [0x0b23] = "\x9b\x59", [0x0b24] = "\x9b\x5a",
  [0x0b25] = "\x9b\x5b", [0x0b26] = "\x9b\x5c", [0x0b27] = "\x9b\x5d",
  [0x0b28] = "\x9b\x5e", [0x0b29] = "\x9b\x5f", [0x0b2a] = "\x9b\x60",
  [0x0b2b] = "\x9b\x61", [0x0b2c] = "\x9b\x62", [0x0b2d] = "\x9b\x63",
  [0x0b2e] = "\x9b\x64", [0x0b2f] = "\x9b\x65", [0x0b30] = "\x9b\x66",
  [0x0b31] = "\x9b\x67", [0x0b32] = "\x9b\x68", [0x0b33] = "\x9b\x69",
  [0x0b34] = "\x9b\x6a", [0x0b35] = "\x9b\x6b", [0x0b36] = "\x9b\x6c",
  [0x0b37] = "\x9b\x6d", [0x0b38] = "\x9b\x6e", [0x0b39] = "\x9b\x6f",
  [0x0b3a] = "\x9b\x70", [0x0b3b] = "\x9b\x71", [0x0b3c] = "\x9b\x72",
  [0x0b3d] = "\x9b\x73", [0x0b3e] = "\x9b\x74", [0x0b3f] = "\x9b\x75",
  [0x0b40] = "\x9b\x76", [0x0b41] = "\x9b\x77", [0x0b42] = "\x9b\x78",
  [0x0b43] = "\x9b\x79", [0x0b44] = "\x9b\x7a", [0x0b45] = "\x9b\x7b",
  [0x0b46] = "\x9b\x7c", [0x0b47] = "\x9b\x7d", [0x0b48] = "\x9b\x7e",
  [0x0b49] = "\x9b\xa1", [0x0b4a] = "\x9b\xa2", [0x0b4b] = "\x9b\xa3",
  [0x0b4c] = "\x9b\xa4", [0x0b4d] = "\x9b\xa5", [0x0b4e] = "\x9b\xa6",
  [0x0b4f] = "\x9b\xa7", [0x0b50] = "\x9b\xa8", [0x0b51] = "\x9b\xa9",
  [0x0b52] = "\x9b\xaa", [0x0b53] = "\x9b\xab", [0x0b54] = "\x9b\xac",
  [0x0b55] = "\x9b\xad", [0x0b56] = "\x9b\xae", [0x0b57] = "\x9b\xaf",
  [0x0b58] = "\x9b\xb0", [0x0b59] = "\x9b\xb1", [0x0b5a] = "\x9b\xb2",
  [0x0b5b] = "\x9b\xb3", [0x0b5c] = "\x9b\xb4", [0x0b5d] = "\x9b\xb5",
  [0x0b5e] = "\x9b\xb6", [0x0b5f] = "\x9b\xb7", [0x0b60] = "\x9b\xb8",
  [0x0b61] = "\x9b\xb9", [0x0b62] = "\x9b\xba", [0x0b63] = "\x9b\xbb",
  [0x0b64] = "\x9b\xbc", [0x0b65] = "\x9b\xbd", [0x0b66] = "\x9b\xbe",
  [0x0b67] = "\x9b\xbf", [0x0b68] = "\x9b\xc0", [0x0b69] = "\x9b\xc1",
  [0x0b6a] = "\x9b\xc2", [0x0b6b] = "\x9b\xc3", [0x0b6c] = "\x9b\xc4",
  [0x0b6d] = "\x9b\xc5", [0x0b6e] = "\x9b\xc6", [0x0b6f] = "\x9b\xc7",
  [0x0b70] = "\x9b\xc8", [0x0b71] = "\x9b\xc9", [0x0b72] = "\x9b\xca",
  [0x0b73] = "\x9b\xcb", [0x0b74] = "\x9b\xcc", [0x0b75] = "\x9b\xcd",
  [0x0b76] = "\x9b\xce", [0x0b77] = "\x9b\xcf", [0x0b78] = "\x9b\xd0",
  [0x0b79] = "\x9b\xd1", [0x0b7a] = "\x9b\xd2", [0x0b7b] = "\x9b\xd3",
  [0x0b7c] = "\x9b\xd4", [0x0b7d] = "\x9b\xd5", [0x0b7e] = "\x9b\xd6",
  [0x0b7f] = "\x9b\xd7", [0x0b80] = "\x9b\xd8", [0x0b81] = "\x9b\xd9",
  [0x0b82] = "\x9b\xda", [0x0b83] = "\x9b\xdb", [0x0b84] = "\x9b\xdc",
  [0x0b85] = "\x9b\xdd", [0x0b86] = "\x9b\xde", [0x0b87] = "\x9b\xdf",
  [0x0b88] = "\x9b\xe0", [0x0b89] = "\x9b\xe1", [0x0b8a] = "\x9b\xe2",
  [0x0b8b] = "\x9b\xe3", [0x0b8c] = "\x9b\xe4", [0x0b8d] = "\x9b\xe5",
  [0x0b8e] = "\x9b\xe6", [0x0b8f] = "\x9b\xe7", [0x0b90] = "\x9b\xe8",
  [0x0b91] = "\x9b\xe9", [0x0b92] = "\x9b\xea", [0x0b93] = "\x9b\xeb",
  [0x0b94] = "\x9b\xec", [0x0b95] = "\x9b\xed", [0x0b96] = "\x9b\xee",
  [0x0b97] = "\x9b\xef", [0x0b98] = "\x9b\xf0", [0x0b99] = "\x9b\xf1",
  [0x0b9a] = "\x9b\xf2", [0x0b9b] = "\x9b\xf3", [0x0b9c] = "\x9b\xf4",
  [0x0b9d] = "\x9b\xf5", [0x0b9e] = "\x9b\xf6", [0x0b9f] = "\x9b\xf7",
  [0x0ba0] = "\x9b\xf8", [0x0ba1] = "\x9b\xf9", [0x0ba2] = "\x9b\xfa",
  [0x0ba3] = "\x9b\xfb", [0x0ba4] = "\x9b\xfc", [0x0ba5] = "\x9b\xfd",
  [0x0ba6] = "\x9b\xfe", [0x0ba7] = "\x9c\x40", [0x0ba8] = "\x9c\x41",
  [0x0ba9] = "\x9c\x42", [0x0baa] = "\x9c\x43", [0x0bab] = "\x9c\x44",
  [0x0bac] = "\x9c\x45", [0x0bad] = "\x9c\x46", [0x0bae] = "\x9c\x47",
  [0x0baf] = "\x9c\x48", [0x0bb0] = "\x9c\x49", [0x0bb1] = "\x9c\x4a",
  [0x0bb2] = "\x9c\x4b", [0x0bb3] = "\x9c\x4c", [0x0bb4] = "\x9c\x4d",
  [0x0bb5] = "\x9c\x4e", [0x0bb6] = "\x9c\x4f", [0x0bb7] = "\x9c\x50",
  [0x0bb8] = "\x9c\x51", [0x0bb9] = "\x9c\x52", [0x0bba] = "\x9c\x53",
  [0x0bbb] = "\x9c\x54", [0x0bbc] = "\x9c\x55", [0x0bbd] = "\x9c\x56",
  [0x0bbe] = "\x9c\x57", [0x0bbf] = "\x9c\x58", [0x0bc0] = "\x9c\x59",
  [0x0bc1] = "\x9c\x5a", [0x0bc2] = "\x9c\x5b", [0x0bc3] = "\x9c\x5c",
  [0x0bc4] = "\x9c\x5d", [0x0bc5] = "\x9c\x5e", [0x0bc6] = "\x9c\x5f",
  [0x0bc7] = "\x9c\x60", [0x0bc8] = "\x9c\x61", [0x0bc9] = "\x9c\x62",
  [0x0bca] = "\x9c\x63", [0x0bcb] = "\x9c\x64", [0x0bcc] = "\x9c\x65",
  [0x0bcd] = "\x9c\x66", [0x0bce] = "\x9c\x67", [0x0bcf] = "\x9c\x68",
  [0x0bd0] = "\x9c\x69", [0x0bd1] = "\x9c\x6a", [0x0bd2] = "\x9c\x6b",
  [0x0bd3] = "\x9c\x6c", [0x0bd4] = "\x9c\x6d", [0x0bd5] = "\x9c\x6e",
  [0x0bd6] = "\x9c\x6f", [0x0bd7] = "\x9c\x70", [0x0bd8] = "\x9c\x71",
  [0x0bd9] = "\x9c\x72", [0x0bda] = "\x9c\x73", [0x0bdb] = "\x9c\x74",
  [0x0bdc] = "\x9c\x75", [0x0bdd] = "\x9c\x76", [0x0bde] = "\x9c\x77",
  [0x0bdf] = "\x9c\x78", [0x0be0] = "\x9c\x79", [0x0be1] = "\x9c\x7a",
  [0x0be2] = "\x9c\x7b", [0x0be3] = "\x9c\x7c", [0x0be4] = "\x9c\x7d",
  [0x0be5] = "\x9c\x7e", [0x0be6] = "\x9c\xa1", [0x0be7] = "\x9c\xa2",
  [0x0be8] = "\x9c\xa3", [0x0be9] = "\x9c\xa4", [0x0bea] = "\x9c\xa5",
  [0x0beb] = "\x9c\xa6", [0x0bec] = "\x9c\xa7", [0x0bed] = "\x9c\xa8",
  [0x0bee] = "\x9c\xa9", [0x0bef] = "\x9c\xaa", [0x0bf0] = "\x9c\xab",
  [0x0bf1] = "\x9c\xac", [0x0bf2] = "\x9c\xad", [0x0bf3] = "\x9c\xae",
  [0x0bf4] = "\x9c\xaf", [0x0bf5] = "\x9c\xb0", [0x0bf6] = "\x9c\xb1",
  [0x0bf7] = "\x9c\xb2", [0x0bf8] = "\x9c\xb3", [0x0bf9] = "\x9c\xb4",
  [0x0bfa] = "\x9c\xb5", [0x0bfb] = "\x9c\xb6", [0x0bfc] = "\x9c\xb7",
  [0x0bfd] = "\x9c\xb8", [0x0bfe] = "\x9c\xb9", [0x0bff] = "\x9c\xba",
  [0x0c00] = "\x9c\xbb", [0x0c01] = "\x9c\xbc", [0x0c02] = "\x9c\xbd",
  [0x0c03] = "\x9c\xbe", [0x0c04] = "\x9c\xbf", [0x0c05] = "\x9c\xc0",
  [0x0c06] = "\x9c\xc1", [0x0c07] = "\x9c\xc2", [0x0c08] = "\x9c\xc3",
  [0x0c09] = "\x9c\xc4", [0x0c0a] = "\x9c\xc5", [0x0c0b] = "\x9c\xc6",
  [0x0c0c] = "\x9c\xc7", [0x0c0d] = "\x9c\xc8", [0x0c0e] = "\x9c\xc9",
  [0x0c0f] = "\x9c\xca", [0x0c10] = "\x9c\xcb", [0x0c11] = "\x9c\xcc",
  [0x0c12] = "\x9c\xcd", [0x0c13] = "\x9c\xce", [0x0c14] = "\x9c\xcf",
  [0x0c15] = "\x9c\xd0", [0x0c16] = "\x9c\xd1", [0x0c17] = "\x9c\xd2",
  [0x0c18] = "\x9c\xd3", [0x0c19] = "\x9c\xd4", [0x0c1a] = "\x9c\xd5",
  [0x0c1b] = "\x9c\xd6", [0x0c1c] = "\x9c\xd7", [0x0c1d] = "\x9c\xd8",
  [0x0c1e] = "\x9c\xd9", [0x0c1f] = "\x9c\xda", [0x0c20] = "\x9c\xdb",
  [0x0c21] = "\x9c\xdc", [0x0c22] = "\x9c\xdd", [0x0c23] = "\x9c\xde",
  [0x0c24] = "\x9c\xdf", [0x0c25] = "\x9c\xe0", [0x0c26] = "\x9c\xe1",
  [0x0c27] = "\x9c\xe2", [0x0c28] = "\x9c\xe3", [0x0c29] = "\x9c\xe4",
  [0x0c2a] = "\x9c\xe5", [0x0c2b] = "\x9c\xe6", [0x0c2c] = "\x9c\xe7",
  [0x0c2d] = "\x9c\xe8", [0x0c2e] = "\x9c\xe9", [0x0c2f] = "\x9c\xea",
  [0x0c30] = "\x9c\xeb", [0x0c31] = "\x9c\xec", [0x0c32] = "\x9c\xed",
  [0x0c33] = "\x9c\xee", [0x0c34] = "\x9c\xef", [0x0c35] = "\x9c\xf0",
  [0x0c36] = "\x9c\xf1", [0x0c37] = "\x9c\xf2", [0x0c38] = "\x9c\xf3",
  [0x0c39] = "\x9c\xf4", [0x0c3a] = "\x9c\xf5", [0x0c3b] = "\x9c\xf6",
  [0x0c3c] = "\x9c\xf7", [0x0c3d] = "\x9c\xf8", [0x0c3e] = "\x9c\xf9",
  [0x0c3f] = "\x9c\xfa", [0x0c40] = "\x9c\xfb", [0x0c41] = "\x9c\xfc",
  [0x0c42] = "\x9c\xfd", [0x0c43] = "\x9c\xfe", [0x0c44] = "\x9d\x40",
  [0x0c45] = "\x9d\x41", [0x0c46] = "\x9d\x42", [0x0c47] = "\x9d\x43",
  [0x0c48] = "\x9d\x44", [0x0c49] = "\x9d\x45", [0x0c4a] = "\x9d\x46",
  [0x0c4b] = "\x9d\x47", [0x0c4c] = "\x9d\x48", [0x0c4d] = "\x9d\x49",
  [0x0c4e] = "\x9d\x4a", [0x0c4f] = "\x9d\x4b", [0x0c50] = "\x9d\x4c",
  [0x0c51] = "\x9d\x4d", [0x0c52] = "\x9d\x4e", [0x0c53] = "\x9d\x4f",
  [0x0c54] = "\x9d\x50", [0x0c55] = "\x9d\x51", [0x0c56] = "\x9d\x52",
  [0x0c57] = "\x9d\x53", [0x0c58] = "\x9d\x54", [0x0c59] = "\x9d\x55",
  [0x0c5a] = "\x9d\x56", [0x0c5b] = "\x9d\x57", [0x0c5c] = "\x9d\x58",
  [0x0c5d] = "\x9d\x59", [0x0c5e] = "\x9d\x5a", [0x0c5f] = "\x9d\x5b",
  [0x0c60] = "\x9d\x5c", [0x0c61] = "\x9d\x5d", [0x0c62] = "\x9d\x5e",
  [0x0c63] = "\x9d\x5f", [0x0c64] = "\x9d\x60", [0x0c65] = "\x9d\x61",
  [0x0c66] = "\x9d\x62", [0x0c67] = "\x9d\x63", [0x0c68] = "\x9d\x64",
  [0x0c69] = "\x9d\x65", [0x0c6a] = "\x9d\x66", [0x0c6b] = "\x9d\x67",
  [0x0c6c] = "\x9d\x68", [0x0c6d] = "\x9d\x69", [0x0c6e] = "\x9d\x6a",
  [0x0c6f] = "\x9d\x6b", [0x0c70] = "\x9d\x6c", [0x0c71] = "\x9d\x6d",
  [0x0c72] = "\x9d\x6e", [0x0c73] = "\x9d\x6f", [0x0c74] = "\x9d\x70",
  [0x0c75] = "\x9d\x71", [0x0c76] = "\x9d\x72", [0x0c77] = "\x9d\x73",
  [0x0c78] = "\x9d\x74", [0x0c79] = "\x9d\x75", [0x0c7a] = "\x9d\x76",
  [0x0c7b] = "\x9d\x77", [0x0c7c] = "\x9d\x78", [0x0c7d] = "\x9d\x79",
  [0x0c7e] = "\x9d\x7a", [0x0c7f] = "\x9d\x7b", [0x0c80] = "\x9d\x7c",
  [0x0c81] = "\x9d\x7d", [0x0c82] = "\x9d\x7e", [0x0c83] = "\x9d\xa1",
  [0x0c84] = "\x9d\xa2", [0x0c85] = "\x9d\xa3", [0x0c86] = "\x9d\xa4",
  [0x0c87] = "\x9d\xa5", [0x0c88] = "\x9d\xa6", [0x0c89] = "\x9d\xa7",
  [0x0c8a] = "\x9d\xa8", [0x0c8b] = "\x9d\xa9", [0x0c8c] = "\x9d\xaa",
  [0x0c8d] = "\x9d\xab", [0x0c8e] = "\x9d\xac", [0x0c8f] = "\x9d\xad",
  [0x0c90] = "\x9d\xae", [0x0c91] = "\x9d\xaf", [0x0c92] = "\x9d\xb0",
  [0x0c93] = "\x9d\xb1", [0x0c94] = "\x9d\xb2", [0x0c95] = "\x9d\xb3",
  [0x0c96] = "\x9d\xb4", [0x0c97] = "\x9d\xb5", [0x0c98] = "\x9d\xb6",
  [0x0c99] = "\x9d\xb7", [0x0c9a] = "\x9d\xb8", [0x0c9b] = "\x9d\xb9",
  [0x0c9c] = "\x9d\xba", [0x0c9d] = "\x9d\xbb", [0x0c9e] = "\x9d\xbc",
  [0x0c9f] = "\x9d\xbd", [0x0ca0] = "\x9d\xbe", [0x0ca1] = "\x9d\xbf",
  [0x0ca2] = "\x9d\xc0", [0x0ca3] = "\x9d\xc1", [0x0ca4] = "\x9d\xc2",
  [0x0ca5] = "\x9d\xc3", [0x0ca6] = "\x9d\xc4", [0x0ca7] = "\x9d\xc5",
  [0x0ca8] = "\x9d\xc6", [0x0ca9] = "\x9d\xc7", [0x0caa] = "\x9d\xc8",
  [0x0cab] = "\x9d\xc9", [0x0cac] = "\x9d\xca", [0x0cad] = "\x9d\xcb",
  [0x0cae] = "\x9d\xcc", [0x0caf] = "\x9d\xcd", [0x0cb0] = "\x9d\xce",
  [0x0cb1] = "\x9d\xcf", [0x0cb2] = "\x9d\xd0", [0x0cb3] = "\x9d\xd1",
  [0x0cb4] = "\x9d\xd2", [0x0cb5] = "\x9d\xd3", [0x0cb6] = "\x9d\xd4",
  [0x0cb7] = "\x9d\xd5", [0x0cb8] = "\x9d\xd6", [0x0cb9] = "\x9d\xd7",
  [0x0cba] = "\x9d\xd8", [0x0cbb] = "\x9d\xd9", [0x0cbc] = "\x9d\xda",
  [0x0cbd] = "\x9d\xdb", [0x0cbe] = "\x9d\xdc", [0x0cbf] = "\x9d\xdd",
  [0x0cc0] = "\x9d\xde", [0x0cc1] = "\x9d\xdf", [0x0cc2] = "\x9d\xe0",
  [0x0cc3] = "\x9d\xe1", [0x0cc4] = "\x9d\xe2", [0x0cc5] = "\x9d\xe3",
  [0x0cc6] = "\x9d\xe4", [0x0cc7] = "\x9d\xe5", [0x0cc8] = "\x9d\xe6",
  [0x0cc9] = "\x9d\xe7", [0x0cca] = "\x9d\xe8", [0x0ccb] = "\x9d\xe9",
  [0x0ccc] = "\x9d\xea", [0x0ccd] = "\x9d\xeb", [0x0cce] = "\x9d\xec",
  [0x0ccf] = "\x9d\xed", [0x0cd0] = "\x9d\xee", [0x0cd1] = "\x9d\xef",
  [0x0cd2] = "\x9d\xf0", [0x0cd3] = "\x9d\xf1", [0x0cd4] = "\x9d\xf2",
  [0x0cd5] = "\x9d\xf3", [0x0cd6] = "\x9d\xf4", [0x0cd7] = "\x9d\xf5",
  [0x0cd8] = "\x9d\xf6", [0x0cd9] = "\x9d\xf7", [0x0cda] = "\x9d\xf8",
  [0x0cdb] = "\x9d\xf9", [0x0cdc] = "\x9d\xfa", [0x0cdd] = "\x9d\xfb",
  [0x0cde] = "\x9d\xfc", [0x0cdf] = "\x9d\xfd", [0x0ce0] = "\x9d\xfe",
  [0x0ce1] = "\x9e\x40", [0x0ce2] = "\x9e\x41", [0x0ce3] = "\x9e\x42",
  [0x0ce4] = "\x9e\x43", [0x0ce5] = "\x9e\x44", [0x0ce6] = "\x9e\x45",
  [0x0ce7] = "\x9e\x46", [0x0ce8] = "\x9e\x47", [0x0ce9] = "\x9e\x48",
  [0x0cea] = "\x9e\x49", [0x0ceb] = "\x9e\x4a", [0x0cec] = "\x9e\x4b",
  [0x0ced] = "\x9e\x4c", [0x0cee] = "\x9e\x4d", [0x0cef] = "\x9e\x4e",
  [0x0cf0] = "\x9e\x4f", [0x0cf1] = "\x9e\x50", [0x0cf2] = "\x9e\x51",
  [0x0cf3] = "\x9e\x52", [0x0cf4] = "\x9e\x53", [0x0cf5] = "\x9e\x54",
  [0x0cf6] = "\x9e\x55", [0x0cf7] = "\x9e\x56", [0x0cf8] = "\x9e\x57",
  [0x0cf9] = "\x9e\x58", [0x0cfa] = "\x9e\x59", [0x0cfb] = "\x9e\x5a",
  [0x0cfc] = "\x9e\x5b", [0x0cfd] = "\x9e\x5c", [0x0cfe] = "\x9e\x5d",
  [0x0cff] = "\x9e\x5e", [0x0d00] = "\x9e\x5f", [0x0d01] = "\x9e\x60",
  [0x0d02] = "\x9e\x61", [0x0d03] = "\x9e\x62", [0x0d04] = "\x9e\x63",
  [0x0d05] = "\x9e\x64", [0x0d06] = "\x9e\x65", [0x0d07] = "\x9e\x66",
  [0x0d08] = "\x9e\x67", [0x0d09] = "\x9e\x68", [0x0d0a] = "\x9e\x69",
  [0x0d0b] = "\x9e\x6a", [0x0d0c] = "\x9e\x6b", [0x0d0d] = "\x9e\x6c",
  [0x0d0e] = "\x9e\x6d", [0x0d0f] = "\x9e\x6e", [0x0d10] = "\x9e\x6f",
  [0x0d11] = "\x9e\x70", [0x0d12] = "\x9e\x71", [0x0d13] = "\x9e\x72",
  [0x0d14] = "\x9e\x73", [0x0d15] = "\x9e\x74", [0x0d16] = "\x9e\x75",
  [0x0d17] = "\x9e\x76", [0x0d18] = "\x9e\x77", [0x0d19] = "\x9e\x78",
  [0x0d1a] = "\x9e\x79", [0x0d1b] = "\x9e\x7a", [0x0d1c] = "\x9e\x7b",
  [0x0d1d] = "\x9e\x7c", [0x0d1e] = "\x9e\x7d", [0x0d1f] = "\x9e\x7e",
  [0x0d20] = "\x9e\xa1", [0x0d21] = "\x9e\xa2", [0x0d22] = "\x9e\xa3",
  [0x0d23] = "\x9e\xa4", [0x0d24] = "\x9e\xa5", [0x0d25] = "\x9e\xa6",
  [0x0d26] = "\x9e\xa7", [0x0d27] = "\x9e\xa8", [0x0d28] = "\x9e\xa9",
  [0x0d29] = "\x9e\xaa", [0x0d2a] = "\x9e\xab", [0x0d2b] = "\x9e\xac",
  [0x0d2c] = "\x9e\xad", [0x0d2d] = "\x9e\xae", [0x0d2e] = "\x9e\xaf",
  [0x0d2f] = "\x9e\xb0", [0x0d30] = "\x9e\xb1", [0x0d31] = "\x9e\xb2",
  [0x0d32] = "\x9e\xb3", [0x0d33] = "\x9e\xb4", [0x0d34] = "\x9e\xb5",
  [0x0d35] = "\x9e\xb6", [0x0d36] = "\x9e\xb7", [0x0d37] = "\x9e\xb8",
  [0x0d38] = "\x9e\xb9", [0x0d39] = "\x9e\xba", [0x0d3a] = "\x9e\xbb",
  [0x0d3b] = "\x9e\xbc", [0x0d3c] = "\x9e\xbd", [0x0d3d] = "\x9e\xbe",
  [0x0d3e] = "\x9e\xbf", [0x0d3f] = "\x9e\xc0", [0x0d40] = "\x9e\xc1",
  [0x0d41] = "\x9e\xc2", [0x0d42] = "\x9e\xc3", [0x0d43] = "\x9e\xc4",
  [0x0d44] = "\x9e\xc5", [0x0d45] = "\x9e\xc6", [0x0d46] = "\x9e\xc7",
  [0x0d47] = "\x9e\xc8", [0x0d48] = "\x9e\xc9", [0x0d49] = "\x9e\xca",
  [0x0d4a] = "\x9e\xcb", [0x0d4b] = "\x9e\xcc", [0x0d4c] = "\x9e\xcd",
  [0x0d4d] = "\x9e\xce", [0x0d4e] = "\x9e\xcf", [0x0d4f] = "\x9e\xd0",
  [0x0d50] = "\x9e\xd1", [0x0d51] = "\x9e\xd2", [0x0d52] = "\x9e\xd3",
  [0x0d53] = "\x9e\xd4", [0x0d54] = "\x9e\xd5", [0x0d55] = "\x9e\xd6",
  [0x0d56] = "\x9e\xd7", [0x0d57] = "\x9e\xd8", [0x0d58] = "\x9e\xd9",
  [0x0d59] = "\x9e\xda", [0x0d5a] = "\x9e\xdb", [0x0d5b] = "\x9e\xdc",
  [0x0d5c] = "\x9e\xdd", [0x0d5d] = "\x9e\xde", [0x0d5e] = "\x9e\xdf",
  [0x0d5f] = "\x9e\xe0", [0x0d60] = "\x9e\xe1", [0x0d61] = "\x9e\xe2",
  [0x0d62] = "\x9e\xe3", [0x0d63] = "\x9e\xe4", [0x0d64] = "\x9e\xe5",
  [0x0d65] = "\x9e\xe6", [0x0d66] = "\x9e\xe7", [0x0d67] = "\x9e\xe8",
  [0x0d68] = "\x9e\xe9", [0x0d69] = "\x9e\xea", [0x0d6a] = "\x9e\xeb",
  [0x0d6b] = "\x9e\xec", [0x0d6c] = "\x9e\xed", [0x0d6d] = "\x9e\xee",
  [0x0d6e] = "\x9e\xef", [0x0d6f] = "\x9e\xf0", [0x0d70] = "\x9e\xf1",
  [0x0d71] = "\x9e\xf2", [0x0d72] = "\x9e\xf3", [0x0d73] = "\x9e\xf4",
  [0x0d74] = "\x9e\xf5", [0x0d75] = "\x9e\xf6", [0x0d76] = "\x9e\xf7",
  [0x0d77] = "\x9e\xf8", [0x0d78] = "\x9e\xf9", [0x0d79] = "\x9e\xfa",
  [0x0d7a] = "\x9e\xfb", [0x0d7b] = "\x9e\xfc", [0x0d7c] = "\x9e\xfd",
  [0x0d7d] = "\x9e\xfe", [0x0d7e] = "\x9f\x40", [0x0d7f] = "\x9f\x41",
  [0x0d80] = "\x9f\x42", [0x0d81] = "\x9f\x43", [0x0d82] = "\x9f\x44",
  [0x0d83] = "\x9f\x45", [0x0d84] = "\x9f\x46", [0x0d85] = "\x9f\x47",
  [0x0d86] = "\x9f\x48", [0x0d87] = "\x9f\x49", [0x0d88] = "\x9f\x4a",
  [0x0d89] = "\x9f\x4b", [0x0d8a] = "\x9f\x4c", [0x0d8b] = "\x9f\x4d",
  [0x0d8c] = "\x9f\x4e", [0x0d8d] = "\x9f\x4f", [0x0d8e] = "\x9f\x50",
  [0x0d8f] = "\x9f\x51", [0x0d90] = "\x9f\x52", [0x0d91] = "\x9f\x53",
  [0x0d92] = "\x9f\x54", [0x0d93] = "\x9f\x55", [0x0d94] = "\x9f\x56",
  [0x0d95] = "\x9f\x57", [0x0d96] = "\x9f\x58", [0x0d97] = "\x9f\x59",
  [0x0d98] = "\x9f\x5a", [0x0d99] = "\x9f\x5b", [0x0d9a] = "\x9f\x5c",
  [0x0d9b] = "\x9f\x5d", [0x0d9c] = "\x9f\x5e", [0x0d9d] = "\x9f\x5f",
  [0x0d9e] = "\x9f\x60", [0x0d9f] = "\x9f\x61", [0x0da0] = "\x9f\x62",
  [0x0da1] = "\x9f\x63", [0x0da2] = "\x9f\x64", [0x0da3] = "\x9f\x65",
  [0x0da4] = "\x9f\x66", [0x0da5] = "\x9f\x67", [0x0da6] = "\x9f\x68",
  [0x0da7] = "\x9f\x69", [0x0da8] = "\x9f\x6a", [0x0da9] = "\x9f\x6b",
  [0x0daa] = "\x9f\x6c", [0x0dab] = "\x9f\x6d", [0x0dac] = "\x9f\x6e",
  [0x0dad] = "\x9f\x6f", [0x0dae] = "\x9f\x70", [0x0daf] = "\x9f\x71",
  [0x0db0] = "\x9f\x72", [0x0db1] = "\x9f\x73", [0x0db2] = "\x9f\x74",
  [0x0db3] = "\x9f\x75", [0x0db4] = "\x9f\x76", [0x0db5] = "\x9f\x77",
  [0x0db6] = "\x9f\x78", [0x0db7] = "\x9f\x79", [0x0db8] = "\x9f\x7a",
  [0x0db9] = "\x9f\x7b", [0x0dba] = "\x9f\x7c", [0x0dbb] = "\x9f\x7d",
  [0x0dbc] = "\x9f\x7e", [0x0dbd] = "\x9f\xa1", [0x0dbe] = "\x9f\xa2",
  [0x0dbf] = "\x9f\xa3", [0x0dc0] = "\x9f\xa4", [0x0dc1] = "\x9f\xa5",
  [0x0dc2] = "\x9f\xa6", [0x0dc3] = "\x9f\xa7", [0x0dc4] = "\x9f\xa8",
  [0x0dc5] = "\x9f\xa9", [0x0dc6] = "\x9f\xaa", [0x0dc7] = "\x9f\xab",
  [0x0dc8] = "\x9f\xac", [0x0dc9] = "\x9f\xad", [0x0dca] = "\x9f\xae",
  [0x0dcb] = "\x9f\xaf", [0x0dcc] = "\x9f\xb0", [0x0dcd] = "\x9f\xb1",
  [0x0dce] = "\x9f\xb2", [0x0dcf] = "\x9f\xb3", [0x0dd0] = "\x9f\xb4",
  [0x0dd1] = "\x9f\xb5", [0x0dd2] = "\x9f\xb6", [0x0dd3] = "\x9f\xb7",
  [0x0dd4] = "\x9f\xb8", [0x0dd5] = "\x9f\xb9", [0x0dd6] = "\x9f\xba",
  [0x0dd7] = "\x9f\xbb", [0x0dd8] = "\x9f\xbc", [0x0dd9] = "\x9f\xbd",
  [0x0dda] = "\x9f\xbe", [0x0ddb] = "\x9f\xbf", [0x0ddc] = "\x9f\xc0",
  [0x0ddd] = "\x9f\xc1", [0x0dde] = "\x9f\xc2", [0x0ddf] = "\x9f\xc3",
  [0x0de0] = "\x9f\xc4", [0x0de1] = "\x9f\xc5", [0x0de2] = "\x9f\xc6",
  [0x0de3] = "\x9f\xc7", [0x0de4] = "\x9f\xc8", [0x0de5] = "\x9f\xc9",
  [0x0de6] = "\x9f\xca", [0x0de7] = "\x9f\xcb", [0x0de8] = "\x9f\xcc",
  [0x0de9] = "\x9f\xcd", [0x0dea] = "\x9f\xce", [0x0deb] = "\x9f\xcf",
  [0x0dec] = "\x9f\xd0", [0x0ded] = "\x9f\xd1", [0x0dee] = "\x9f\xd2",
  [0x0def] = "\x9f\xd3", [0x0df0] = "\x9f\xd4", [0x0df1] = "\x9f\xd5",
  [0x0df2] = "\x9f\xd6", [0x0df3] = "\x9f\xd7", [0x0df4] = "\x9f\xd8",
  [0x0df5] = "\x9f\xd9", [0x0df6] = "\x9f\xda", [0x0df7] = "\x9f\xdb",
  [0x0df8] = "\x9f\xdc", [0x0df9] = "\x9f\xdd", [0x0dfa] = "\x9f\xde",
  [0x0dfb] = "\x9f\xdf", [0x0dfc] = "\x9f\xe0", [0x0dfd] = "\x9f\xe1",
  [0x0dfe] = "\x9f\xe2", [0x0dff] = "\x9f\xe3", [0x0e00] = "\x9f\xe4",
  [0x0e01] = "\x9f\xe5", [0x0e02] = "\x9f\xe6", [0x0e03] = "\x9f\xe7",
  [0x0e04] = "\x9f\xe8", [0x0e05] = "\x9f\xe9", [0x0e06] = "\x9f\xea",
  [0x0e07] = "\x9f\xeb", [0x0e08] = "\x9f\xec", [0x0e09] = "\x9f\xed",
  [0x0e0a] = "\x9f\xee", [0x0e0b] = "\x9f\xef", [0x0e0c] = "\x9f\xf0",
  [0x0e0d] = "\x9f\xf1", [0x0e0e] = "\x9f\xf2", [0x0e0f] = "\x9f\xf3",
  [0x0e10] = "\x9f\xf4", [0x0e11] = "\x9f\xf5", [0x0e12] = "\x9f\xf6",
  [0x0e13] = "\x9f\xf7", [0x0e14] = "\x9f\xf8", [0x0e15] = "\x9f\xf9",
  [0x0e16] = "\x9f\xfa", [0x0e17] = "\x9f\xfb", [0x0e18] = "\x9f\xfc",
  [0x0e19] = "\x9f\xfd", [0x0e1a] = "\x9f\xfe", [0x0e1b] = "\xa0\x40",
  [0x0e1c] = "\xa0\x41", [0x0e1d] = "\xa0\x42", [0x0e1e] = "\xa0\x43",
  [0x0e1f] = "\xa0\x44", [0x0e20] = "\xa0\x45", [0x0e21] = "\xa0\x46",
  [0x0e22] = "\xa0\x47", [0x0e23] = "\xa0\x48", [0x0e24] = "\xa0\x49",
  [0x0e25] = "\xa0\x4a", [0x0e26] = "\xa0\x4b", [0x0e27] = "\xa0\x4c",
  [0x0e28] = "\xa0\x4d", [0x0e29] = "\xa0\x4e", [0x0e2a] = "\xa0\x4f",
  [0x0e2b] = "\xa0\x50", [0x0e2c] = "\xa0\x51", [0x0e2d] = "\xa0\x52",
  [0x0e2e] = "\xa0\x53", [0x0e2f] = "\xa0\x54", [0x0e30] = "\xa0\x55",
  [0x0e31] = "\xa0\x56", [0x0e32] = "\xa0\x57", [0x0e33] = "\xa0\x58",
  [0x0e34] = "\xa0\x59", [0x0e35] = "\xa0\x5a", [0x0e36] = "\xa0\x5b",
  [0x0e37] = "\xa0\x5c", [0x0e38] = "\xa0\x5d", [0x0e39] = "\xa0\x5e",
  [0x0e3a] = "\xa0\x5f", [0x0e3b] = "\xa0\x60", [0x0e3c] = "\xa0\x61",
  [0x0e3d] = "\xa0\x62", [0x0e3e] = "\xa0\x63", [0x0e3f] = "\xa0\x64",
  [0x0e40] = "\xa0\x65", [0x0e41] = "\xa0\x66", [0x0e42] = "\xa0\x67",
  [0x0e43] = "\xa0\x68", [0x0e44] = "\xa0\x69", [0x0e45] = "\xa0\x6a",
  [0x0e46] = "\xa0\x6b", [0x0e47] = "\xa0\x6c", [0x0e48] = "\xa0\x6d",
  [0x0e49] = "\xa0\x6e", [0x0e4a] = "\xa0\x6f", [0x0e4b] = "\xa0\x70",
  [0x0e4c] = "\xa0\x71", [0x0e4d] = "\xa0\x72", [0x0e4e] = "\xa0\x73",
  [0x0e4f] = "\xa0\x74", [0x0e50] = "\xa0\x75", [0x0e51] = "\xa0\x76",
  [0x0e52] = "\xa0\x77", [0x0e53] = "\xa0\x78", [0x0e54] = "\xa0\x79",
  [0x0e55] = "\xa0\x7a", [0x0e56] = "\xa0\x7b", [0x0e57] = "\xa0\x7c",
  [0x0e58] = "\xa0\x7d", [0x0e59] = "\xa0\x7e", [0x0e5a] = "\xa0\xa1",
  [0x0e5b] = "\xa0\xa2", [0x0e5c] = "\xa0\xa3", [0x0e5d] = "\xa0\xa4",
  [0x0e5e] = "\xa0\xa5", [0x0e5f] = "\xa0\xa6", [0x0e60] = "\xa0\xa7",
  [0x0e61] = "\xa0\xa8", [0x0e62] = "\xa0\xa9", [0x0e63] = "\xa0\xaa",
  [0x0e64] = "\xa0\xab", [0x0e65] = "\xa0\xac", [0x0e66] = "\xa0\xad",
  [0x0e67] = "\xa0\xae", [0x0e68] = "\xa0\xaf", [0x0e69] = "\xa0\xb0",
  [0x0e6a] = "\xa0\xb1", [0x0e6b] = "\xa0\xb2", [0x0e6c] = "\xa0\xb3",
  [0x0e6d] = "\xa0\xb4", [0x0e6e] = "\xa0\xb5", [0x0e6f] = "\xa0\xb6",
  [0x0e70] = "\xa0\xb7", [0x0e71] = "\xa0\xb8", [0x0e72] = "\xa0\xb9",
  [0x0e73] = "\xa0\xba", [0x0e74] = "\xa0\xbb", [0x0e75] = "\xa0\xbc",
  [0x0e76] = "\xa0\xbd", [0x0e77] = "\xa0\xbe", [0x0e78] = "\xa0\xbf",
  [0x0e79] = "\xa0\xc0", [0x0e7a] = "\xa0\xc1", [0x0e7b] = "\xa0\xc2",
  [0x0e7c] = "\xa0\xc3", [0x0e7d] = "\xa0\xc4", [0x0e7e] = "\xa0\xc5",
  [0x0e7f] = "\xa0\xc6", [0x0e80] = "\xa0\xc7", [0x0e81] = "\xa0\xc8",
  [0x0e82] = "\xa0\xc9", [0x0e83] = "\xa0\xca", [0x0e84] = "\xa0\xcb",
  [0x0e85] = "\xa0\xcc", [0x0e86] = "\xa0\xcd", [0x0e87] = "\xa0\xce",
  [0x0e88] = "\xa0\xcf", [0x0e89] = "\xa0\xd0", [0x0e8a] = "\xa0\xd1",
  [0x0e8b] = "\xa0\xd2", [0x0e8c] = "\xa0\xd3", [0x0e8d] = "\xa0\xd4",
  [0x0e8e] = "\xa0\xd5", [0x0e8f] = "\xa0\xd6", [0x0e90] = "\xa0\xd7",
  [0x0e91] = "\xa0\xd8", [0x0e92] = "\xa0\xd9", [0x0e93] = "\xa0\xda",
  [0x0e94] = "\xa0\xdb", [0x0e95] = "\xa0\xdc", [0x0e96] = "\xa0\xdd",
  [0x0e97] = "\xa0\xde", [0x0e98] = "\xa0\xdf", [0x0e99] = "\xa0\xe0",
  [0x0e9a] = "\xa0\xe1", [0x0e9b] = "\xa0\xe2", [0x0e9c] = "\xa0\xe3",
  [0x0e9d] = "\xa0\xe4", [0x0e9e] = "\xa0\xe5", [0x0e9f] = "\xa0\xe6",
  [0x0ea0] = "\xa0\xe7", [0x0ea1] = "\xa0\xe8", [0x0ea2] = "\xa0\xe9",
  [0x0ea3] = "\xa0\xea", [0x0ea4] = "\xa0\xeb", [0x0ea5] = "\xa0\xec",
  [0x0ea6] = "\xa0\xed", [0x0ea7] = "\xa0\xee", [0x0ea8] = "\xa0\xef",
  [0x0ea9] = "\xa0\xf0", [0x0eaa] = "\xa0\xf1", [0x0eab] = "\xa0\xf2",
  [0x0eac] = "\xa0\xf3", [0x0ead] = "\xa0\xf4", [0x0eae] = "\xa0\xf5",
  [0x0eaf] = "\xa0\xf6", [0x0eb0] = "\xa0\xf7", [0x0eb1] = "\xa0\xf8",
  [0x0eb2] = "\xa0\xf9", [0x0eb3] = "\xa0\xfa", [0x0eb4] = "\xa0\xfb",
  [0x0eb5] = "\xa0\xfc", [0x0eb6] = "\xa0\xfd", [0x0eb7] = "\xa0\xfe",
  [0x0eb8] = "\x81\x40", [0x0eb9] = "\x81\x41", [0x0eba] = "\x81\x42",
  [0x0ebb] = "\x81\x43", [0x0ebc] = "\x81\x44", [0x0ebd] = "\x81\x45",
  [0x0ebe] = "\x81\x46", [0x0ebf] = "\x81\x47", [0x0ec0] = "\x81\x48",
  [0x0ec1] = "\x81\x49", [0x0ec2] = "\x81\x4a", [0x0ec3] = "\x81\x4b",
  [0x0ec4] = "\x81\x4c", [0x0ec5] = "\x81\x4d", [0x0ec6] = "\x81\x4e",
  [0x0ec7] = "\x81\x4f", [0x0ec8] = "\x81\x50", [0x0ec9] = "\x81\x51",
  [0x0eca] = "\x81\x52", [0x0ecb] = "\x81\x53", [0x0ecc] = "\x81\x54",
  [0x0ecd] = "\x81\x55", [0x0ece] = "\x81\x56", [0x0ecf] = "\x81\x57",
  [0x0ed0] = "\x81\x58", [0x0ed1] = "\x81\x59", [0x0ed2] = "\x81\x5a",
  [0x0ed3] = "\x81\x5b", [0x0ed4] = "\x81\x5c", [0x0ed5] = "\x81\x5d",
  [0x0ed6] = "\x81\x5e", [0x0ed7] = "\x81\x5f", [0x0ed8] = "\x81\x60",
  [0x0ed9] = "\x81\x61", [0x0eda] = "\x81\x62", [0x0edb] = "\x81\x63",
  [0x0edc] = "\x81\x64", [0x0edd] = "\x81\x65", [0x0ede] = "\x81\x66",
  [0x0edf] = "\x81\x67", [0x0ee0] = "\x81\x68", [0x0ee1] = "\x81\x69",
  [0x0ee2] = "\x81\x6a", [0x0ee3] = "\x81\x6b", [0x0ee4] = "\x81\x6c",
  [0x0ee5] = "\x81\x6d", [0x0ee6] = "\x81\x6e", [0x0ee7] = "\x81\x6f",
  [0x0ee8] = "\x81\x70", [0x0ee9] = "\x81\x71", [0x0eea] = "\x81\x72",
  [0x0eeb] = "\x81\x73", [0x0eec] = "\x81\x74", [0x0eed] = "\x81\x75",
  [0x0eee] = "\x81\x76", [0x0eef] = "\x81\x77", [0x0ef0] = "\x81\x78",
  [0x0ef1] = "\x81\x79", [0x0ef2] = "\x81\x7a", [0x0ef3] = "\x81\x7b",
  [0x0ef4] = "\x81\x7c", [0x0ef5] = "\x81\x7d", [0x0ef6] = "\x81\x7e",
  [0x0ef7] = "\x81\xa1", [0x0ef8] = "\x81\xa2", [0x0ef9] = "\x81\xa3",
  [0x0efa] = "\x81\xa4", [0x0efb] = "\x81\xa5", [0x0efc] = "\x81\xa6",
  [0x0efd] = "\x81\xa7", [0x0efe] = "\x81\xa8", [0x0eff] = "\x81\xa9",
  [0x0f00] = "\x81\xaa", [0x0f01] = "\x81\xab", [0x0f02] = "\x81\xac",
  [0x0f03] = "\x81\xad", [0x0f04] = "\x81\xae", [0x0f05] = "\x81\xaf",
  [0x0f06] = "\x81\xb0", [0x0f07] = "\x81\xb1", [0x0f08] = "\x81\xb2",
  [0x0f09] = "\x81\xb3", [0x0f0a] = "\x81\xb4", [0x0f0b] = "\x81\xb5",
  [0x0f0c] = "\x81\xb6", [0x0f0d] = "\x81\xb7", [0x0f0e] = "\x81\xb8",
  [0x0f0f] = "\x81\xb9", [0x0f10] = "\x81\xba", [0x0f11] = "\x81\xbb",
  [0x0f12] = "\x81\xbc", [0x0f13] = "\x81\xbd", [0x0f14] = "\x81\xbe",
  [0x0f15] = "\x81\xbf", [0x0f16] = "\x81\xc0", [0x0f17] = "\x81\xc1",
  [0x0f18] = "\x81\xc2", [0x0f19] = "\x81\xc3", [0x0f1a] = "\x81\xc4",
  [0x0f1b] = "\x81\xc5", [0x0f1c] = "\x81\xc6", [0x0f1d] = "\x81\xc7",
  [0x0f1e] = "\x81\xc8", [0x0f1f] = "\x81\xc9", [0x0f20] = "\x81\xca",
  [0x0f21] = "\x81\xcb", [0x0f22] = "\x81\xcc", [0x0f23] = "\x81\xcd",
  [0x0f24] = "\x81\xce", [0x0f25] = "\x81\xcf", [0x0f26] = "\x81\xd0",
  [0x0f27] = "\x81\xd1", [0x0f28] = "\x81\xd2", [0x0f29] = "\x81\xd3",
  [0x0f2a] = "\x81\xd4", [0x0f2b] = "\x81\xd5", [0x0f2c] = "\x81\xd6",
  [0x0f2d] = "\x81\xd7", [0x0f2e] = "\x81\xd8", [0x0f2f] = "\x81\xd9",
  [0x0f30] = "\x81\xda", [0x0f31] = "\x81\xdb", [0x0f32] = "\x81\xdc",
  [0x0f33] = "\x81\xdd", [0x0f34] = "\x81\xde", [0x0f35] = "\x81\xdf",
  [0x0f36] = "\x81\xe0", [0x0f37] = "\x81\xe1", [0x0f38] = "\x81\xe2",
  [0x0f39] = "\x81\xe3", [0x0f3a] = "\x81\xe4", [0x0f3b] = "\x81\xe5",
  [0x0f3c] = "\x81\xe6", [0x0f3d] = "\x81\xe7", [0x0f3e] = "\x81\xe8",
  [0x0f3f] = "\x81\xe9", [0x0f40] = "\x81\xea", [0x0f41] = "\x81\xeb",
  [0x0f42] = "\x81\xec", [0x0f43] = "\x81\xed", [0x0f44] = "\x81\xee",
  [0x0f45] = "\x81\xef", [0x0f46] = "\x81\xf0", [0x0f47] = "\x81\xf1",
  [0x0f48] = "\x81\xf2", [0x0f49] = "\x81\xf3", [0x0f4a] = "\x81\xf4",
  [0x0f4b] = "\x81\xf5", [0x0f4c] = "\x81\xf6", [0x0f4d] = "\x81\xf7",
  [0x0f4e] = "\x81\xf8", [0x0f4f] = "\x81\xf9", [0x0f50] = "\x81\xfa",
  [0x0f51] = "\x81\xfb", [0x0f52] = "\x81\xfc", [0x0f53] = "\x81\xfd",
  [0x0f54] = "\x81\xfe", [0x0f55] = "\x82\x40", [0x0f56] = "\x82\x41",
  [0x0f57] = "\x82\x42", [0x0f58] = "\x82\x43", [0x0f59] = "\x82\x44",
  [0x0f5a] = "\x82\x45", [0x0f5b] = "\x82\x46", [0x0f5c] = "\x82\x47",
  [0x0f5d] = "\x82\x48", [0x0f5e] = "\x82\x49", [0x0f5f] = "\x82\x4a",
  [0x0f60] = "\x82\x4b", [0x0f61] = "\x82\x4c", [0x0f62] = "\x82\x4d",
  [0x0f63] = "\x82\x4e", [0x0f64] = "\x82\x4f", [0x0f65] = "\x82\x50",
  [0x0f66] = "\x82\x51", [0x0f67] = "\x82\x52", [0x0f68] = "\x82\x53",
  [0x0f69] = "\x82\x54", [0x0f6a] = "\x82\x55", [0x0f6b] = "\x82\x56",
  [0x0f6c] = "\x82\x57", [0x0f6d] = "\x82\x58", [0x0f6e] = "\x82\x59",
  [0x0f6f] = "\x82\x5a", [0x0f70] = "\x82\x5b", [0x0f71] = "\x82\x5c",
  [0x0f72] = "\x82\x5d", [0x0f73] = "\x82\x5e", [0x0f74] = "\x82\x5f",
  [0x0f75] = "\x82\x60", [0x0f76] = "\x82\x61", [0x0f77] = "\x82\x62",
  [0x0f78] = "\x82\x63", [0x0f79] = "\x82\x64", [0x0f7a] = "\x82\x65",
  [0x0f7b] = "\x82\x66", [0x0f7c] = "\x82\x67", [0x0f7d] = "\x82\x68",
  [0x0f7e] = "\x82\x69", [0x0f7f] = "\x82\x6a", [0x0f80] = "\x82\x6b",
  [0x0f81] = "\x82\x6c", [0x0f82] = "\x82\x6d", [0x0f83] = "\x82\x6e",
  [0x0f84] = "\x82\x6f", [0x0f85] = "\x82\x70", [0x0f86] = "\x82\x71",
  [0x0f87] = "\x82\x72", [0x0f88] = "\x82\x73", [0x0f89] = "\x82\x74",
  [0x0f8a] = "\x82\x75", [0x0f8b] = "\x82\x76", [0x0f8c] = "\x82\x77",
  [0x0f8d] = "\x82\x78", [0x0f8e] = "\x82\x79", [0x0f8f] = "\x82\x7a",
  [0x0f90] = "\x82\x7b", [0x0f91] = "\x82\x7c", [0x0f92] = "\x82\x7d",
  [0x0f93] = "\x82\x7e", [0x0f94] = "\x82\xa1", [0x0f95] = "\x82\xa2",
  [0x0f96] = "\x82\xa3", [0x0f97] = "\x82\xa4", [0x0f98] = "\x82\xa5",
  [0x0f99] = "\x82\xa6", [0x0f9a] = "\x82\xa7", [0x0f9b] = "\x82\xa8",
  [0x0f9c] = "\x82\xa9", [0x0f9d] = "\x82\xaa", [0x0f9e] = "\x82\xab",
  [0x0f9f] = "\x82\xac", [0x0fa0] = "\x82\xad", [0x0fa1] = "\x82\xae",
  [0x0fa2] = "\x82\xaf", [0x0fa3] = "\x82\xb0", [0x0fa4] = "\x82\xb1",
  [0x0fa5] = "\x82\xb2", [0x0fa6] = "\x82\xb3", [0x0fa7] = "\x82\xb4",
  [0x0fa8] = "\x82\xb5", [0x0fa9] = "\x82\xb6", [0x0faa] = "\x82\xb7",
  [0x0fab] = "\x82\xb8", [0x0fac] = "\x82\xb9", [0x0fad] = "\x82\xba",
  [0x0fae] = "\x82\xbb", [0x0faf] = "\x82\xbc", [0x0fb0] = "\x82\xbd",
  [0x0fb1] = "\x82\xbe", [0x0fb2] = "\x82\xbf", [0x0fb3] = "\x82\xc0",
  [0x0fb4] = "\x82\xc1", [0x0fb5] = "\x82\xc2", [0x0fb6] = "\x82\xc3",
  [0x0fb7] = "\x82\xc4", [0x0fb8] = "\x82\xc5", [0x0fb9] = "\x82\xc6",
  [0x0fba] = "\x82\xc7", [0x0fbb] = "\x82\xc8", [0x0fbc] = "\x82\xc9",
  [0x0fbd] = "\x82\xca", [0x0fbe] = "\x82\xcb", [0x0fbf] = "\x82\xcc",
  [0x0fc0] = "\x82\xcd", [0x0fc1] = "\x82\xce", [0x0fc2] = "\x82\xcf",
  [0x0fc3] = "\x82\xd0", [0x0fc4] = "\x82\xd1", [0x0fc5] = "\x82\xd2",
  [0x0fc6] = "\x82\xd3", [0x0fc7] = "\x82\xd4", [0x0fc8] = "\x82\xd5",
  [0x0fc9] = "\x82\xd6", [0x0fca] = "\x82\xd7", [0x0fcb] = "\x82\xd8",
  [0x0fcc] = "\x82\xd9", [0x0fcd] = "\x82\xda", [0x0fce] = "\x82\xdb",
  [0x0fcf] = "\x82\xdc", [0x0fd0] = "\x82\xdd", [0x0fd1] = "\x82\xde",
  [0x0fd2] = "\x82\xdf", [0x0fd3] = "\x82\xe0", [0x0fd4] = "\x82\xe1",
  [0x0fd5] = "\x82\xe2", [0x0fd6] = "\x82\xe3", [0x0fd7] = "\x82\xe4",
  [0x0fd8] = "\x82\xe5", [0x0fd9] = "\x82\xe6", [0x0fda] = "\x82\xe7",
  [0x0fdb] = "\x82\xe8", [0x0fdc] = "\x82\xe9", [0x0fdd] = "\x82\xea",
  [0x0fde] = "\x82\xeb", [0x0fdf] = "\x82\xec", [0x0fe0] = "\x82\xed",
  [0x0fe1] = "\x82\xee", [0x0fe2] = "\x82\xef", [0x0fe3] = "\x82\xf0",
  [0x0fe4] = "\x82\xf1", [0x0fe5] = "\x82\xf2", [0x0fe6] = "\x82\xf3",
  [0x0fe7] = "\x82\xf4", [0x0fe8] = "\x82\xf5", [0x0fe9] = "\x82\xf6",
  [0x0fea] = "\x82\xf7", [0x0feb] = "\x82\xf8", [0x0fec] = "\x82\xf9",
  [0x0fed] = "\x82\xfa", [0x0fee] = "\x82\xfb", [0x0fef] = "\x82\xfc",
  [0x0ff0] = "\x82\xfd", [0x0ff1] = "\x82\xfe", [0x0ff2] = "\x83\x40",
  [0x0ff3] = "\x83\x41", [0x0ff4] = "\x83\x42", [0x0ff5] = "\x83\x43",
  [0x0ff6] = "\x83\x44", [0x0ff7] = "\x83\x45", [0x0ff8] = "\x83\x46",
  [0x0ff9] = "\x83\x47", [0x0ffa] = "\x83\x48", [0x0ffb] = "\x83\x49",
  [0x0ffc] = "\x83\x4a", [0x0ffd] = "\x83\x4b", [0x0ffe] = "\x83\x4c",
  [0x0fff] = "\x83\x4d", [0x1000] = "\x83\x4e", [0x1001] = "\x83\x4f",
  [0x1002] = "\x83\x50", [0x1003] = "\x83\x51", [0x1004] = "\x83\x52",
  [0x1005] = "\x83\x53", [0x1006] = "\x83\x54", [0x1007] = "\x83\x55",
  [0x1008] = "\x83\x56", [0x1009] = "\x83\x57", [0x100a] = "\x83\x58",
  [0x100b] = "\x83\x59", [0x100c] = "\x83\x5a", [0x100d] = "\x83\x5b",
  [0x100e] = "\x83\x5c", [0x100f] = "\x83\x5d", [0x1010] = "\x83\x5e",
  [0x1011] = "\x83\x5f", [0x1012] = "\x83\x60", [0x1013] = "\x83\x61",
  [0x1014] = "\x83\x62", [0x1015] = "\x83\x63", [0x1016] = "\x83\x64",
  [0x1017] = "\x83\x65", [0x1018] = "\x83\x66", [0x1019] = "\x83\x67",
  [0x101a] = "\x83\x68", [0x101b] = "\x83\x69", [0x101c] = "\x83\x6a",
  [0x101d] = "\x83\x6b", [0x101e] = "\x83\x6c", [0x101f] = "\x83\x6d",
  [0x1020] = "\x83\x6e", [0x1021] = "\x83\x6f", [0x1022] = "\x83\x70",
  [0x1023] = "\x83\x71", [0x1024] = "\x83\x72", [0x1025] = "\x83\x73",
  [0x1026] = "\x83\x74", [0x1027] = "\x83\x75", [0x1028] = "\x83\x76",
  [0x1029] = "\x83\x77", [0x102a] = "\x83\x78", [0x102b] = "\x83\x79",
  [0x102c] = "\x83\x7a", [0x102d] = "\x83\x7b", [0x102e] = "\x83\x7c",
  [0x102f] = "\x83\x7d", [0x1030] = "\x83\x7e", [0x1031] = "\x83\xa1",
  [0x1032] = "\x83\xa2", [0x1033] = "\x83\xa3", [0x1034] = "\x83\xa4",
  [0x1035] = "\x83\xa5", [0x1036] = "\x83\xa6", [0x1037] = "\x83\xa7",
  [0x1038] = "\x83\xa8", [0x1039] = "\x83\xa9", [0x103a] = "\x83\xaa",
  [0x103b] = "\x83\xab", [0x103c] = "\x83\xac", [0x103d] = "\x83\xad",
  [0x103e] = "\x83\xae", [0x103f] = "\x83\xaf", [0x1040] = "\x83\xb0",
  [0x1041] = "\x83\xb1", [0x1042] = "\x83\xb2", [0x1043] = "\x83\xb3",
  [0x1044] = "\x83\xb4", [0x1045] = "\x83\xb5", [0x1046] = "\x83\xb6",
  [0x1047] = "\x83\xb7", [0x1048] = "\x83\xb8", [0x1049] = "\x83\xb9",
  [0x104a] = "\x83\xba", [0x104b] = "\x83\xbb", [0x104c] = "\x83\xbc",
  [0x104d] = "\x83\xbd", [0x104e] = "\x83\xbe", [0x104f] = "\x83\xbf",
  [0x1050] = "\x83\xc0", [0x1051] = "\x83\xc1", [0x1052] = "\x83\xc2",
  [0x1053] = "\x83\xc3", [0x1054] = "\x83\xc4", [0x1055] = "\x83\xc5",
  [0x1056] = "\x83\xc6", [0x1057] = "\x83\xc7", [0x1058] = "\x83\xc8",
  [0x1059] = "\x83\xc9", [0x105a] = "\x83\xca", [0x105b] = "\x83\xcb",
  [0x105c] = "\x83\xcc", [0x105d] = "\x83\xcd", [0x105e] = "\x83\xce",
  [0x105f] = "\x83\xcf", [0x1060] = "\x83\xd0", [0x1061] = "\x83\xd1",
  [0x1062] = "\x83\xd2", [0x1063] = "\x83\xd3", [0x1064] = "\x83\xd4",
  [0x1065] = "\x83\xd5", [0x1066] = "\x83\xd6", [0x1067] = "\x83\xd7",
  [0x1068] = "\x83\xd8", [0x1069] = "\x83\xd9", [0x106a] = "\x83\xda",
  [0x106b] = "\x83\xdb", [0x106c] = "\x83\xdc", [0x106d] = "\x83\xdd",
  [0x106e] = "\x83\xde", [0x106f] = "\x83\xdf", [0x1070] = "\x83\xe0",
  [0x1071] = "\x83\xe1", [0x1072] = "\x83\xe2", [0x1073] = "\x83\xe3",
  [0x1074] = "\x83\xe4", [0x1075] = "\x83\xe5", [0x1076] = "\x83\xe6",
  [0x1077] = "\x83\xe7", [0x1078] = "\x83\xe8", [0x1079] = "\x83\xe9",
  [0x107a] = "\x83\xea", [0x107b] = "\x83\xeb", [0x107c] = "\x83\xec",
  [0x107d] = "\x83\xed", [0x107e] = "\x83\xee", [0x107f] = "\x83\xef",
  [0x1080] = "\x83\xf0", [0x1081] = "\x83\xf1", [0x1082] = "\x83\xf2",
  [0x1083] = "\x83\xf3", [0x1084] = "\x83\xf4", [0x1085] = "\x83\xf5",
  [0x1086] = "\x83\xf6", [0x1087] = "\x83\xf7", [0x1088] = "\x83\xf8",
  [0x1089] = "\x83\xf9", [0x108a] = "\x83\xfa", [0x108b] = "\x83\xfb",
  [0x108c] = "\x83\xfc", [0x108d] = "\x83\xfd", [0x108e] = "\x83\xfe",
  [0x108f] = "\x84\x40", [0x1090] = "\x84\x41", [0x1091] = "\x84\x42",
  [0x1092] = "\x84\x43", [0x1093] = "\x84\x44", [0x1094] = "\x84\x45",
  [0x1095] = "\x84\x46", [0x1096] = "\x84\x47", [0x1097] = "\x84\x48",
  [0x1098] = "\x84\x49", [0x1099] = "\x84\x4a", [0x109a] = "\x84\x4b",
  [0x109b] = "\x84\x4c", [0x109c] = "\x84\x4d", [0x109d] = "\x84\x4e",
  [0x109e] = "\x84\x4f", [0x109f] = "\x84\x50", [0x10a0] = "\x84\x51",
  [0x10a1] = "\x84\x52", [0x10a2] = "\x84\x53", [0x10a3] = "\x84\x54",
  [0x10a4] = "\x84\x55", [0x10a5] = "\x84\x56", [0x10a6] = "\x84\x57",
  [0x10a7] = "\x84\x58", [0x10a8] = "\x84\x59", [0x10a9] = "\x84\x5a",
  [0x10aa] = "\x84\x5b", [0x10ab] = "\x84\x5c", [0x10ac] = "\x84\x5d",
  [0x10ad] = "\x84\x5e", [0x10ae] = "\x84\x5f", [0x10af] = "\x84\x60",
  [0x10b0] = "\x84\x61", [0x10b1] = "\x84\x62", [0x10b2] = "\x84\x63",
  [0x10b3] = "\x84\x64", [0x10b4] = "\x84\x65", [0x10b5] = "\x84\x66",
  [0x10b6] = "\x84\x67", [0x10b7] = "\x84\x68", [0x10b8] = "\x84\x69",
  [0x10b9] = "\x84\x6a", [0x10ba] = "\x84\x6b", [0x10bb] = "\x84\x6c",
  [0x10bc] = "\x84\x6d", [0x10bd] = "\x84\x6e", [0x10be] = "\x84\x6f",
  [0x10bf] = "\x84\x70", [0x10c0] = "\x84\x71", [0x10c1] = "\x84\x72",
  [0x10c2] = "\x84\x73", [0x10c3] = "\x84\x74", [0x10c4] = "\x84\x75",
  [0x10c5] = "\x84\x76", [0x10c6] = "\x84\x77", [0x10c7] = "\x84\x78",
  [0x10c8] = "\x84\x79", [0x10c9] = "\x84\x7a", [0x10ca] = "\x84\x7b",
  [0x10cb] = "\x84\x7c", [0x10cc] = "\x84\x7d", [0x10cd] = "\x84\x7e",
  [0x10ce] = "\x84\xa1", [0x10cf] = "\x84\xa2", [0x10d0] = "\x84\xa3",
  [0x10d1] = "\x84\xa4", [0x10d2] = "\x84\xa5", [0x10d3] = "\x84\xa6",
  [0x10d4] = "\x84\xa7", [0x10d5] = "\x84\xa8", [0x10d6] = "\x84\xa9",
  [0x10d7] = "\x84\xaa", [0x10d8] = "\x84\xab", [0x10d9] = "\x84\xac",
  [0x10da] = "\x84\xad", [0x10db] = "\x84\xae", [0x10dc] = "\x84\xaf",
  [0x10dd] = "\x84\xb0", [0x10de] = "\x84\xb1", [0x10df] = "\x84\xb2",
  [0x10e0] = "\x84\xb3", [0x10e1] = "\x84\xb4", [0x10e2] = "\x84\xb5",
  [0x10e3] = "\x84\xb6", [0x10e4] = "\x84\xb7", [0x10e5] = "\x84\xb8",
  [0x10e6] = "\x84\xb9", [0x10e7] = "\x84\xba", [0x10e8] = "\x84\xbb",
  [0x10e9] = "\x84\xbc", [0x10ea] = "\x84\xbd", [0x10eb] = "\x84\xbe",
  [0x10ec] = "\x84\xbf", [0x10ed] = "\x84\xc0", [0x10ee] = "\x84\xc1",
  [0x10ef] = "\x84\xc2", [0x10f0] = "\x84\xc3", [0x10f1] = "\x84\xc4",
  [0x10f2] = "\x84\xc5", [0x10f3] = "\x84\xc6", [0x10f4] = "\x84\xc7",
  [0x10f5] = "\x84\xc8", [0x10f6] = "\x84\xc9", [0x10f7] = "\x84\xca",
  [0x10f8] = "\x84\xcb", [0x10f9] = "\x84\xcc", [0x10fa] = "\x84\xcd",
  [0x10fb] = "\x84\xce", [0x10fc] = "\x84\xcf", [0x10fd] = "\x84\xd0",
  [0x10fe] = "\x84\xd1", [0x10ff] = "\x84\xd2", [0x1100] = "\x84\xd3",
  [0x1101] = "\x84\xd4", [0x1102] = "\x84\xd5", [0x1103] = "\x84\xd6",
  [0x1104] = "\x84\xd7", [0x1105] = "\x84\xd8", [0x1106] = "\x84\xd9",
  [0x1107] = "\x84\xda", [0x1108] = "\x84\xdb", [0x1109] = "\x84\xdc",
  [0x110a] = "\x84\xdd", [0x110b] = "\x84\xde", [0x110c] = "\x84\xdf",
  [0x110d] = "\x84\xe0", [0x110e] = "\x84\xe1", [0x110f] = "\x84\xe2",
  [0x1110] = "\x84\xe3", [0x1111] = "\x84\xe4", [0x1112] = "\x84\xe5",
  [0x1113] = "\x84\xe6", [0x1114] = "\x84\xe7", [0x1115] = "\x84\xe8",
  [0x1116] = "\x84\xe9", [0x1117] = "\x84\xea", [0x1118] = "\x84\xeb",
  [0x1119] = "\x84\xec", [0x111a] = "\x84\xed", [0x111b] = "\x84\xee",
  [0x111c] = "\x84\xef", [0x111d] = "\x84\xf0", [0x111e] = "\x84\xf1",
  [0x111f] = "\x84\xf2", [0x1120] = "\x84\xf3", [0x1121] = "\x84\xf4",
  [0x1122] = "\x84\xf5", [0x1123] = "\x84\xf6", [0x1124] = "\x84\xf7",
  [0x1125] = "\x84\xf8", [0x1126] = "\x84\xf9", [0x1127] = "\x84\xfa",
  [0x1128] = "\x84\xfb", [0x1129] = "\x84\xfc", [0x112a] = "\x84\xfd",
  [0x112b] = "\x84\xfe", [0x112c] = "\x85\x40", [0x112d] = "\x85\x41",
  [0x112e] = "\x85\x42", [0x112f] = "\x85\x43", [0x1130] = "\x85\x44",
  [0x1131] = "\x85\x45", [0x1132] = "\x85\x46", [0x1133] = "\x85\x47",
  [0x1134] = "\x85\x48", [0x1135] = "\x85\x49", [0x1136] = "\x85\x4a",
  [0x1137] = "\x85\x4b", [0x1138] = "\x85\x4c", [0x1139] = "\x85\x4d",
  [0x113a] = "\x85\x4e", [0x113b] = "\x85\x4f", [0x113c] = "\x85\x50",
  [0x113d] = "\x85\x51", [0x113e] = "\x85\x52", [0x113f] = "\x85\x53",
  [0x1140] = "\x85\x54", [0x1141] = "\x85\x55", [0x1142] = "\x85\x56",
  [0x1143] = "\x85\x57", [0x1144] = "\x85\x58", [0x1145] = "\x85\x59",
  [0x1146] = "\x85\x5a", [0x1147] = "\x85\x5b", [0x1148] = "\x85\x5c",
  [0x1149] = "\x85\x5d", [0x114a] = "\x85\x5e", [0x114b] = "\x85\x5f",
  [0x114c] = "\x85\x60", [0x114d] = "\x85\x61", [0x114e] = "\x85\x62",
  [0x114f] = "\x85\x63", [0x1150] = "\x85\x64", [0x1151] = "\x85\x65",
  [0x1152] = "\x85\x66", [0x1153] = "\x85\x67", [0x1154] = "\x85\x68",
  [0x1155] = "\x85\x69", [0x1156] = "\x85\x6a", [0x1157] = "\x85\x6b",
  [0x1158] = "\x85\x6c", [0x1159] = "\x85\x6d", [0x115a] = "\x85\x6e",
  [0x115b] = "\x85\x6f", [0x115c] = "\x85\x70", [0x115d] = "\x85\x71",
  [0x115e] = "\x85\x72", [0x115f] = "\x85\x73", [0x1160] = "\x85\x74",
  [0x1161] = "\x85\x75", [0x1162] = "\x85\x76", [0x1163] = "\x85\x77",
  [0x1164] = "\x85\x78", [0x1165] = "\x85\x79", [0x1166] = "\x85\x7a",
  [0x1167] = "\x85\x7b", [0x1168] = "\x85\x7c", [0x1169] = "\x85\x7d",
  [0x116a] = "\x85\x7e", [0x116b] = "\x85\xa1", [0x116c] = "\x85\xa2",
  [0x116d] = "\x85\xa3", [0x116e] = "\x85\xa4", [0x116f] = "\x85\xa5",
  [0x1170] = "\x85\xa6", [0x1171] = "\x85\xa7", [0x1172] = "\x85\xa8",
  [0x1173] = "\x85\xa9", [0x1174] = "\x85\xaa", [0x1175] = "\x85\xab",
  [0x1176] = "\x85\xac", [0x1177] = "\x85\xad", [0x1178] = "\x85\xae",
  [0x1179] = "\x85\xaf", [0x117a] = "\x85\xb0", [0x117b] = "\x85\xb1",
  [0x117c] = "\x85\xb2", [0x117d] = "\x85\xb3", [0x117e] = "\x85\xb4",
  [0x117f] = "\x85\xb5", [0x1180] = "\x85\xb6", [0x1181] = "\x85\xb7",
  [0x1182] = "\x85\xb8", [0x1183] = "\x85\xb9", [0x1184] = "\x85\xba",
  [0x1185] = "\x85\xbb", [0x1186] = "\x85\xbc", [0x1187] = "\x85\xbd",
  [0x1188] = "\x85\xbe", [0x1189] = "\x85\xbf", [0x118a] = "\x85\xc0",
  [0x118b] = "\x85\xc1", [0x118c] = "\x85\xc2", [0x118d] = "\x85\xc3",
  [0x118e] = "\x85\xc4", [0x118f] = "\x85\xc5", [0x1190] = "\x85\xc6",
  [0x1191] = "\x85\xc7", [0x1192] = "\x85\xc8", [0x1193] = "\x85\xc9",
  [0x1194] = "\x85\xca", [0x1195] = "\x85\xcb", [0x1196] = "\x85\xcc",
  [0x1197] = "\x85\xcd", [0x1198] = "\x85\xce", [0x1199] = "\x85\xcf",
  [0x119a] = "\x85\xd0", [0x119b] = "\x85\xd1", [0x119c] = "\x85\xd2",
  [0x119d] = "\x85\xd3", [0x119e] = "\x85\xd4", [0x119f] = "\x85\xd5",
  [0x11a0] = "\x85\xd6", [0x11a1] = "\x85\xd7", [0x11a2] = "\x85\xd8",
  [0x11a3] = "\x85\xd9", [0x11a4] = "\x85\xda", [0x11a5] = "\x85\xdb",
  [0x11a6] = "\x85\xdc", [0x11a7] = "\x85\xdd", [0x11a8] = "\x85\xde",
  [0x11a9] = "\x85\xdf", [0x11aa] = "\x85\xe0", [0x11ab] = "\x85\xe1",
  [0x11ac] = "\x85\xe2", [0x11ad] = "\x85\xe3", [0x11ae] = "\x85\xe4",
  [0x11af] = "\x85\xe5", [0x11b0] = "\x85\xe6", [0x11b1] = "\x85\xe7",
  [0x11b2] = "\x85\xe8", [0x11b3] = "\x85\xe9", [0x11b4] = "\x85\xea",
  [0x11b5] = "\x85\xeb", [0x11b6] = "\x85\xec", [0x11b7] = "\x85\xed",
  [0x11b8] = "\x85\xee", [0x11b9] = "\x85\xef", [0x11ba] = "\x85\xf0",
  [0x11bb] = "\x85\xf1", [0x11bc] = "\x85\xf2", [0x11bd] = "\x85\xf3",
  [0x11be] = "\x85\xf4", [0x11bf] = "\x85\xf5", [0x11c0] = "\x85\xf6",
  [0x11c1] = "\x85\xf7", [0x11c2] = "\x85\xf8", [0x11c3] = "\x85\xf9",
  [0x11c4] = "\x85\xfa", [0x11c5] = "\x85\xfb", [0x11c6] = "\x85\xfc",
  [0x11c7] = "\x85\xfd", [0x11c8] = "\x85\xfe", [0x11c9] = "\x86\x40",
  [0x11ca] = "\x86\x41", [0x11cb] = "\x86\x42", [0x11cc] = "\x86\x43",
  [0x11cd] = "\x86\x44", [0x11ce] = "\x86\x45", [0x11cf] = "\x86\x46",
  [0x11d0] = "\x86\x47", [0x11d1] = "\x86\x48", [0x11d2] = "\x86\x49",
  [0x11d3] = "\x86\x4a", [0x11d4] = "\x86\x4b", [0x11d5] = "\x86\x4c",
  [0x11d6] = "\x86\x4d", [0x11d7] = "\x86\x4e", [0x11d8] = "\x86\x4f",
  [0x11d9] = "\x86\x50", [0x11da] = "\x86\x51", [0x11db] = "\x86\x52",
  [0x11dc] = "\x86\x53", [0x11dd] = "\x86\x54", [0x11de] = "\x86\x55",
  [0x11df] = "\x86\x56", [0x11e0] = "\x86\x57", [0x11e1] = "\x86\x58",
  [0x11e2] = "\x86\x59", [0x11e3] = "\x86\x5a", [0x11e4] = "\x86\x5b",
  [0x11e5] = "\x86\x5c", [0x11e6] = "\x86\x5d", [0x11e7] = "\x86\x5e",
  [0x11e8] = "\x86\x5f", [0x11e9] = "\x86\x60", [0x11ea] = "\x86\x61",
  [0x11eb] = "\x86\x62", [0x11ec] = "\x86\x63", [0x11ed] = "\x86\x64",
  [0x11ee] = "\x86\x65", [0x11ef] = "\x86\x66", [0x11f0] = "\x86\x67",
  [0x11f1] = "\x86\x68", [0x11f2] = "\x86\x69", [0x11f3] = "\x86\x6a",
  [0x11f4] = "\x86\x6b", [0x11f5] = "\x86\x6c", [0x11f6] = "\x86\x6d",
  [0x11f7] = "\x86\x6e", [0x11f8] = "\x86\x6f", [0x11f9] = "\x86\x70",
  [0x11fa] = "\x86\x71", [0x11fb] = "\x86\x72", [0x11fc] = "\x86\x73",
  [0x11fd] = "\x86\x74", [0x11fe] = "\x86\x75", [0x11ff] = "\x86\x76",
  [0x1200] = "\x86\x77", [0x1201] = "\x86\x78", [0x1202] = "\x86\x79",
  [0x1203] = "\x86\x7a", [0x1204] = "\x86\x7b", [0x1205] = "\x86\x7c",
  [0x1206] = "\x86\x7d", [0x1207] = "\x86\x7e", [0x1208] = "\x86\xa1",
  [0x1209] = "\x86\xa2", [0x120a] = "\x86\xa3", [0x120b] = "\x86\xa4",
  [0x120c] = "\x86\xa5", [0x120d] = "\x86\xa6", [0x120e] = "\x86\xa7",
  [0x120f] = "\x86\xa8", [0x1210] = "\x86\xa9", [0x1211] = "\x86\xaa",
  [0x1212] = "\x86\xab", [0x1213] = "\x86\xac", [0x1214] = "\x86\xad",
  [0x1215] = "\x86\xae", [0x1216] = "\x86\xaf", [0x1217] = "\x86\xb0",
  [0x1218] = "\x86\xb1", [0x1219] = "\x86\xb2", [0x121a] = "\x86\xb3",
  [0x121b] = "\x86\xb4", [0x121c] = "\x86\xb5", [0x121d] = "\x86\xb6",
  [0x121e] = "\x86\xb7", [0x121f] = "\x86\xb8", [0x1220] = "\x86\xb9",
  [0x1221] = "\x86\xba", [0x1222] = "\x86\xbb", [0x1223] = "\x86\xbc",
  [0x1224] = "\x86\xbd", [0x1225] = "\x86\xbe", [0x1226] = "\x86\xbf",
  [0x1227] = "\x86\xc0", [0x1228] = "\x86\xc1", [0x1229] = "\x86\xc2",
  [0x122a] = "\x86\xc3", [0x122b] = "\x86\xc4", [0x122c] = "\x86\xc5",
  [0x122d] = "\x86\xc6", [0x122e] = "\x86\xc7", [0x122f] = "\x86\xc8",
  [0x1230] = "\x86\xc9", [0x1231] = "\x86\xca", [0x1232] = "\x86\xcb",
  [0x1233] = "\x86\xcc", [0x1234] = "\x86\xcd", [0x1235] = "\x86\xce",
  [0x1236] = "\x86\xcf", [0x1237] = "\x86\xd0", [0x1238] = "\x86\xd1",
  [0x1239] = "\x86\xd2", [0x123a] = "\x86\xd3", [0x123b] = "\x86\xd4",
  [0x123c] = "\x86\xd5", [0x123d] = "\x86\xd6", [0x123e] = "\x86\xd7",
  [0x123f] = "\x86\xd8", [0x1240] = "\x86\xd9", [0x1241] = "\x86\xda",
  [0x1242] = "\x86\xdb", [0x1243] = "\x86\xdc", [0x1244] = "\x86\xdd",
  [0x1245] = "\x86\xde", [0x1246] = "\x86\xdf", [0x1247] = "\x86\xe0",
  [0x1248] = "\x86\xe1", [0x1249] = "\x86\xe2", [0x124a] = "\x86\xe3",
  [0x124b] = "\x86\xe4", [0x124c] = "\x86\xe5", [0x124d] = "\x86\xe6",
  [0x124e] = "\x86\xe7", [0x124f] = "\x86\xe8", [0x1250] = "\x86\xe9",
  [0x1251] = "\x86\xea", [0x1252] = "\x86\xeb", [0x1253] = "\x86\xec",
  [0x1254] = "\x86\xed", [0x1255] = "\x86\xee", [0x1256] = "\x86\xef",
  [0x1257] = "\x86\xf0", [0x1258] = "\x86\xf1", [0x1259] = "\x86\xf2",
  [0x125a] = "\x86\xf3", [0x125b] = "\x86\xf4", [0x125c] = "\x86\xf5",
  [0x125d] = "\x86\xf6", [0x125e] = "\x86\xf7", [0x125f] = "\x86\xf8",
  [0x1260] = "\x86\xf9", [0x1261] = "\x86\xfa", [0x1262] = "\x86\xfb",
  [0x1263] = "\x86\xfc", [0x1264] = "\x86\xfd", [0x1265] = "\x86\xfe",
  [0x1266] = "\x87\x40", [0x1267] = "\x87\x41", [0x1268] = "\x87\x42",
  [0x1269] = "\x87\x43", [0x126a] = "\x87\x44", [0x126b] = "\x87\x45",
  [0x126c] = "\x87\x46", [0x126d] = "\x87\x47", [0x126e] = "\x87\x48",
  [0x126f] = "\x87\x49", [0x1270] = "\x87\x4a", [0x1271] = "\x87\x4b",
  [0x1272] = "\x87\x4c", [0x1273] = "\x87\x4d", [0x1274] = "\x87\x4e",
  [0x1275] = "\x87\x4f", [0x1276] = "\x87\x50", [0x1277] = "\x87\x51",
  [0x1278] = "\x87\x52", [0x1279] = "\x87\x53", [0x127a] = "\x87\x54",
  [0x127b] = "\x87\x55", [0x127c] = "\x87\x56", [0x127d] = "\x87\x57",
  [0x127e] = "\x87\x58", [0x127f] = "\x87\x59", [0x1280] = "\x87\x5a",
  [0x1281] = "\x87\x5b", [0x1282] = "\x87\x5c", [0x1283] = "\x87\x5d",
  [0x1284] = "\x87\x5e", [0x1285] = "\x87\x5f", [0x1286] = "\x87\x60",
  [0x1287] = "\x87\x61", [0x1288] = "\x87\x62", [0x1289] = "\x87\x63",
  [0x128a] = "\x87\x64", [0x128b] = "\x87\x65", [0x128c] = "\x87\x66",
  [0x128d] = "\x87\x67", [0x128e] = "\x87\x68", [0x128f] = "\x87\x69",
  [0x1290] = "\x87\x6a", [0x1291] = "\x87\x6b", [0x1292] = "\x87\x6c",
  [0x1293] = "\x87\x6d", [0x1294] = "\x87\x6e", [0x1295] = "\x87\x6f",
  [0x1296] = "\x87\x70", [0x1297] = "\x87\x71", [0x1298] = "\x87\x72",
  [0x1299] = "\x87\x73", [0x129a] = "\x87\x74", [0x129b] = "\x87\x75",
  [0x129c] = "\x87\x76", [0x129d] = "\x87\x77", [0x129e] = "\x87\x78",
  [0x129f] = "\x87\x79", [0x12a0] = "\x87\x7a", [0x12a1] = "\x87\x7b",
  [0x12a2] = "\x87\x7c", [0x12a3] = "\x87\x7d", [0x12a4] = "\x87\x7e",
  [0x12a5] = "\x87\xa1", [0x12a6] = "\x87\xa2", [0x12a7] = "\x87\xa3",
  [0x12a8] = "\x87\xa4", [0x12a9] = "\x87\xa5", [0x12aa] = "\x87\xa6",
  [0x12ab] = "\x87\xa7", [0x12ac] = "\x87\xa8", [0x12ad] = "\x87\xa9",
  [0x12ae] = "\x87\xaa", [0x12af] = "\x87\xab", [0x12b0] = "\x87\xac",
  [0x12b1] = "\x87\xad", [0x12b2] = "\x87\xae", [0x12b3] = "\x87\xaf",
  [0x12b4] = "\x87\xb0", [0x12b5] = "\x87\xb1", [0x12b6] = "\x87\xb2",
  [0x12b7] = "\x87\xb3", [0x12b8] = "\x87\xb4", [0x12b9] = "\x87\xb5",
  [0x12ba] = "\x87\xb6", [0x12bb] = "\x87\xb7", [0x12bc] = "\x87\xb8",
  [0x12bd] = "\x87\xb9", [0x12be] = "\x87\xba", [0x12bf] = "\x87\xbb",
  [0x12c0] = "\x87\xbc", [0x12c1] = "\x87\xbd", [0x12c2] = "\x87\xbe",
  [0x12c3] = "\x87\xbf", [0x12c4] = "\x87\xc0", [0x12c5] = "\x87\xc1",
  [0x12c6] = "\x87\xc2", [0x12c7] = "\x87\xc3", [0x12c8] = "\x87\xc4",
  [0x12c9] = "\x87\xc5", [0x12ca] = "\x87\xc6", [0x12cb] = "\x87\xc7",
  [0x12cc] = "\x87\xc8", [0x12cd] = "\x87\xc9", [0x12ce] = "\x87\xca",
  [0x12cf] = "\x87\xcb", [0x12d0] = "\x87\xcc", [0x12d1] = "\x87\xcd",
  [0x12d2] = "\x87\xce", [0x12d3] = "\x87\xcf", [0x12d4] = "\x87\xd0",
  [0x12d5] = "\x87\xd1", [0x12d6] = "\x87\xd2", [0x12d7] = "\x87\xd3",
  [0x12d8] = "\x87\xd4", [0x12d9] = "\x87\xd5", [0x12da] = "\x87\xd6",
  [0x12db] = "\x87\xd7", [0x12dc] = "\x87\xd8", [0x12dd] = "\x87\xd9",
  [0x12de] = "\x87\xda", [0x12df] = "\x87\xdb", [0x12e0] = "\x87\xdc",
  [0x12e1] = "\x87\xdd", [0x12e2] = "\x87\xde", [0x12e3] = "\x87\xdf",
  [0x12e4] = "\x87\xe0", [0x12e5] = "\x87\xe1", [0x12e6] = "\x87\xe2",
  [0x12e7] = "\x87\xe3", [0x12e8] = "\x87\xe4", [0x12e9] = "\x87\xe5",
  [0x12ea] = "\x87\xe6", [0x12eb] = "\x87\xe7", [0x12ec] = "\x87\xe8",
  [0x12ed] = "\x87\xe9", [0x12ee] = "\x87\xea", [0x12ef] = "\x87\xeb",
  [0x12f0] = "\x87\xec", [0x12f1] = "\x87\xed", [0x12f2] = "\x87\xee",
  [0x12f3] = "\x87\xef", [0x12f4] = "\x87\xf0", [0x12f5] = "\x87\xf1",
  [0x12f6] = "\x87\xf2", [0x12f7] = "\x87\xf3", [0x12f8] = "\x87\xf4",
  [0x12f9] = "\x87\xf5", [0x12fa] = "\x87\xf6", [0x12fb] = "\x87\xf7",
  [0x12fc] = "\x87\xf8", [0x12fd] = "\x87\xf9", [0x12fe] = "\x87\xfa",
  [0x12ff] = "\x87\xfb", [0x1300] = "\x87\xfc", [0x1301] = "\x87\xfd",
  [0x1302] = "\x87\xfe", [0x1303] = "\x88\x40", [0x1304] = "\x88\x41",
  [0x1305] = "\x88\x42", [0x1306] = "\x88\x43", [0x1307] = "\x88\x44",
  [0x1308] = "\x88\x45", [0x1309] = "\x88\x46", [0x130a] = "\x88\x47",
  [0x130b] = "\x88\x48", [0x130c] = "\x88\x49", [0x130d] = "\x88\x4a",
  [0x130e] = "\x88\x4b", [0x130f] = "\x88\x4c", [0x1310] = "\x88\x4d",
  [0x1311] = "\x88\x4e", [0x1312] = "\x88\x4f", [0x1313] = "\x88\x50",
  [0x1314] = "\x88\x51", [0x1315] = "\x88\x52", [0x1316] = "\x88\x53",
  [0x1317] = "\x88\x54", [0x1318] = "\x88\x55", [0x1319] = "\x88\x56",
  [0x131a] = "\x88\x57", [0x131b] = "\x88\x58", [0x131c] = "\x88\x59",
  [0x131d] = "\x88\x5a", [0x131e] = "\x88\x5b", [0x131f] = "\x88\x5c",
  [0x1320] = "\x88\x5d", [0x1321] = "\x88\x5e", [0x1322] = "\x88\x5f",
  [0x1323] = "\x88\x60", [0x1324] = "\x88\x61", [0x1325] = "\x88\x62",
  [0x1326] = "\x88\x63", [0x1327] = "\x88\x64", [0x1328] = "\x88\x65",
  [0x1329] = "\x88\x66", [0x132a] = "\x88\x67", [0x132b] = "\x88\x68",
  [0x132c] = "\x88\x69", [0x132d] = "\x88\x6a", [0x132e] = "\x88\x6b",
  [0x132f] = "\x88\x6c", [0x1330] = "\x88\x6d", [0x1331] = "\x88\x6e",
  [0x1332] = "\x88\x6f", [0x1333] = "\x88\x70", [0x1334] = "\x88\x71",
  [0x1335] = "\x88\x72", [0x1336] = "\x88\x73", [0x1337] = "\x88\x74",
  [0x1338] = "\x88\x75", [0x1339] = "\x88\x76", [0x133a] = "\x88\x77",
  [0x133b] = "\x88\x78", [0x133c] = "\x88\x79", [0x133d] = "\x88\x7a",
  [0x133e] = "\x88\x7b", [0x133f] = "\x88\x7c", [0x1340] = "\x88\x7d",
  [0x1341] = "\x88\x7e", [0x1342] = "\x88\xa1", [0x1343] = "\x88\xa2",
  [0x1344] = "\x88\xa3", [0x1345] = "\x88\xa4", [0x1346] = "\x88\xa5",
  [0x1347] = "\x88\xa6", [0x1348] = "\x88\xa7", [0x1349] = "\x88\xa8",
  [0x134a] = "\x88\xa9", [0x134b] = "\x88\xaa", [0x134c] = "\x88\xab",
  [0x134d] = "\x88\xac", [0x134e] = "\x88\xad", [0x134f] = "\x88\xae",
  [0x1350] = "\x88\xaf", [0x1351] = "\x88\xb0", [0x1352] = "\x88\xb1",
  [0x1353] = "\x88\xb2", [0x1354] = "\x88\xb3", [0x1355] = "\x88\xb4",
  [0x1356] = "\x88\xb5", [0x1357] = "\x88\xb6", [0x1358] = "\x88\xb7",
  [0x1359] = "\x88\xb8", [0x135a] = "\x88\xb9", [0x135b] = "\x88\xba",
  [0x135c] = "\x88\xbb", [0x135d] = "\x88\xbc", [0x135e] = "\x88\xbd",
  [0x135f] = "\x88\xbe", [0x1360] = "\x88\xbf", [0x1361] = "\x88\xc0",
  [0x1362] = "\x88\xc1", [0x1363] = "\x88\xc2", [0x1364] = "\x88\xc3",
  [0x1365] = "\x88\xc4", [0x1366] = "\x88\xc5", [0x1367] = "\x88\xc6",
  [0x1368] = "\x88\xc7", [0x1369] = "\x88\xc8", [0x136a] = "\x88\xc9",
  [0x136b] = "\x88\xca", [0x136c] = "\x88\xcb", [0x136d] = "\x88\xcc",
  [0x136e] = "\x88\xcd", [0x136f] = "\x88\xce", [0x1370] = "\x88\xcf",
  [0x1371] = "\x88\xd0", [0x1372] = "\x88\xd1", [0x1373] = "\x88\xd2",
  [0x1374] = "\x88\xd3", [0x1375] = "\x88\xd4", [0x1376] = "\x88\xd5",
  [0x1377] = "\x88\xd6", [0x1378] = "\x88\xd7", [0x1379] = "\x88\xd8",
  [0x137a] = "\x88\xd9", [0x137b] = "\x88\xda", [0x137c] = "\x88\xdb",
  [0x137d] = "\x88\xdc", [0x137e] = "\x88\xdd", [0x137f] = "\x88\xde",
  [0x1380] = "\x88\xdf", [0x1381] = "\x88\xe0", [0x1382] = "\x88\xe1",
  [0x1383] = "\x88\xe2", [0x1384] = "\x88\xe3", [0x1385] = "\x88\xe4",
  [0x1386] = "\x88\xe5", [0x1387] = "\x88\xe6", [0x1388] = "\x88\xe7",
  [0x1389] = "\x88\xe8", [0x138a] = "\x88\xe9", [0x138b] = "\x88\xea",
  [0x138c] = "\x88\xeb", [0x138d] = "\x88\xec", [0x138e] = "\x88\xed",
  [0x138f] = "\x88\xee", [0x1390] = "\x88\xef", [0x1391] = "\x88\xf0",
  [0x1392] = "\x88\xf1", [0x1393] = "\x88\xf2", [0x1394] = "\x88\xf3",
  [0x1395] = "\x88\xf4", [0x1396] = "\x88\xf5", [0x1397] = "\x88\xf6",
  [0x1398] = "\x88\xf7", [0x1399] = "\x88\xf8", [0x139a] = "\x88\xf9",
  [0x139b] = "\x88\xfa", [0x139c] = "\x88\xfb", [0x139d] = "\x88\xfc",
  [0x139e] = "\x88\xfd", [0x139f] = "\x88\xfe", [0x13a0] = "\x89\x40",
  [0x13a1] = "\x89\x41", [0x13a2] = "\x89\x42", [0x13a3] = "\x89\x43",
  [0x13a4] = "\x89\x44", [0x13a5] = "\x89\x45", [0x13a6] = "\x89\x46",
  [0x13a7] = "\x89\x47", [0x13a8] = "\x89\x48", [0x13a9] = "\x89\x49",
  [0x13aa] = "\x89\x4a", [0x13ab] = "\x89\x4b", [0x13ac] = "\x89\x4c",
  [0x13ad] = "\x89\x4d", [0x13ae] = "\x89\x4e", [0x13af] = "\x89\x4f",
  [0x13b0] = "\x89\x50", [0x13b1] = "\x89\x51", [0x13b2] = "\x89\x52",
  [0x13b3] = "\x89\x53", [0x13b4] = "\x89\x54", [0x13b5] = "\x89\x55",
  [0x13b6] = "\x89\x56", [0x13b7] = "\x89\x57", [0x13b8] = "\x89\x58",
  [0x13b9] = "\x89\x59", [0x13ba] = "\x89\x5a", [0x13bb] = "\x89\x5b",
  [0x13bc] = "\x89\x5c", [0x13bd] = "\x89\x5d", [0x13be] = "\x89\x5e",
  [0x13bf] = "\x89\x5f", [0x13c0] = "\x89\x60", [0x13c1] = "\x89\x61",
  [0x13c2] = "\x89\x62", [0x13c3] = "\x89\x63", [0x13c4] = "\x89\x64",
  [0x13c5] = "\x89\x65", [0x13c6] = "\x89\x66", [0x13c7] = "\x89\x67",
  [0x13c8] = "\x89\x68", [0x13c9] = "\x89\x69", [0x13ca] = "\x89\x6a",
  [0x13cb] = "\x89\x6b", [0x13cc] = "\x89\x6c", [0x13cd] = "\x89\x6d",
  [0x13ce] = "\x89\x6e", [0x13cf] = "\x89\x6f", [0x13d0] = "\x89\x70",
  [0x13d1] = "\x89\x71", [0x13d2] = "\x89\x72", [0x13d3] = "\x89\x73",
  [0x13d4] = "\x89\x74", [0x13d5] = "\x89\x75", [0x13d6] = "\x89\x76",
  [0x13d7] = "\x89\x77", [0x13d8] = "\x89\x78", [0x13d9] = "\x89\x79",
  [0x13da] = "\x89\x7a", [0x13db] = "\x89\x7b", [0x13dc] = "\x89\x7c",
  [0x13dd] = "\x89\x7d", [0x13de] = "\x89\x7e", [0x13df] = "\x89\xa1",
  [0x13e0] = "\x89\xa2", [0x13e1] = "\x89\xa3", [0x13e2] = "\x89\xa4",
  [0x13e3] = "\x89\xa5", [0x13e4] = "\x89\xa6", [0x13e5] = "\x89\xa7",
  [0x13e6] = "\x89\xa8", [0x13e7] = "\x89\xa9", [0x13e8] = "\x89\xaa",
  [0x13e9] = "\x89\xab", [0x13ea] = "\x89\xac", [0x13eb] = "\x89\xad",
  [0x13ec] = "\x89\xae", [0x13ed] = "\x89\xaf", [0x13ee] = "\x89\xb0",
  [0x13ef] = "\x89\xb1", [0x13f0] = "\x89\xb2", [0x13f1] = "\x89\xb3",
  [0x13f2] = "\x89\xb4", [0x13f3] = "\x89\xb5", [0x13f4] = "\x89\xb6",
  [0x13f5] = "\x89\xb7", [0x13f6] = "\x89\xb8", [0x13f7] = "\x89\xb9",
  [0x13f8] = "\x89\xba", [0x13f9] = "\x89\xbb", [0x13fa] = "\x89\xbc",
  [0x13fb] = "\x89\xbd", [0x13fc] = "\x89\xbe", [0x13fd] = "\x89\xbf",
  [0x13fe] = "\x89\xc0", [0x13ff] = "\x89\xc1", [0x1400] = "\x89\xc2",
  [0x1401] = "\x89\xc3", [0x1402] = "\x89\xc4", [0x1403] = "\x89\xc5",
  [0x1404] = "\x89\xc6", [0x1405] = "\x89\xc7", [0x1406] = "\x89\xc8",
  [0x1407] = "\x89\xc9", [0x1408] = "\x89\xca", [0x1409] = "\x89\xcb",
  [0x140a] = "\x89\xcc", [0x140b] = "\x89\xcd", [0x140c] = "\x89\xce",
  [0x140d] = "\x89\xcf", [0x140e] = "\x89\xd0", [0x140f] = "\x89\xd1",
  [0x1410] = "\x89\xd2", [0x1411] = "\x89\xd3", [0x1412] = "\x89\xd4",
  [0x1413] = "\x89\xd5", [0x1414] = "\x89\xd6", [0x1415] = "\x89\xd7",
  [0x1416] = "\x89\xd8", [0x1417] = "\x89\xd9", [0x1418] = "\x89\xda",
  [0x1419] = "\x89\xdb", [0x141a] = "\x89\xdc", [0x141b] = "\x89\xdd",
  [0x141c] = "\x89\xde", [0x141d] = "\x89\xdf", [0x141e] = "\x89\xe0",
  [0x141f] = "\x89\xe1", [0x1420] = "\x89\xe2", [0x1421] = "\x89\xe3",
  [0x1422] = "\x89\xe4", [0x1423] = "\x89\xe5", [0x1424] = "\x89\xe6",
  [0x1425] = "\x89\xe7", [0x1426] = "\x89\xe8", [0x1427] = "\x89\xe9",
  [0x1428] = "\x89\xea", [0x1429] = "\x89\xeb", [0x142a] = "\x89\xec",
  [0x142b] = "\x89\xed", [0x142c] = "\x89\xee", [0x142d] = "\x89\xef",
  [0x142e] = "\x89\xf0", [0x142f] = "\x89\xf1", [0x1430] = "\x89\xf2",
  [0x1431] = "\x89\xf3", [0x1432] = "\x89\xf4", [0x1433] = "\x89\xf5",
  [0x1434] = "\x89\xf6", [0x1435] = "\x89\xf7", [0x1436] = "\x89\xf8",
  [0x1437] = "\x89\xf9", [0x1438] = "\x89\xfa", [0x1439] = "\x89\xfb",
  [0x143a] = "\x89\xfc", [0x143b] = "\x89\xfd", [0x143c] = "\x89\xfe",
  [0x143d] = "\x8a\x40", [0x143e] = "\x8a\x41", [0x143f] = "\x8a\x42",
  [0x1440] = "\x8a\x43", [0x1441] = "\x8a\x44", [0x1442] = "\x8a\x45",
  [0x1443] = "\x8a\x46", [0x1444] = "\x8a\x47", [0x1445] = "\x8a\x48",
  [0x1446] = "\x8a\x49", [0x1447] = "\x8a\x4a", [0x1448] = "\x8a\x4b",
  [0x1449] = "\x8a\x4c", [0x144a] = "\x8a\x4d", [0x144b] = "\x8a\x4e",
  [0x144c] = "\x8a\x4f", [0x144d] = "\x8a\x50", [0x144e] = "\x8a\x51",
  [0x144f] = "\x8a\x52", [0x1450] = "\x8a\x53", [0x1451] = "\x8a\x54",
  [0x1452] = "\x8a\x55", [0x1453] = "\x8a\x56", [0x1454] = "\x8a\x57",
  [0x1455] = "\x8a\x58", [0x1456] = "\x8a\x59", [0x1457] = "\x8a\x5a",
  [0x1458] = "\x8a\x5b", [0x1459] = "\x8a\x5c", [0x145a] = "\x8a\x5d",
  [0x145b] = "\x8a\x5e", [0x145c] = "\x8a\x5f", [0x145d] = "\x8a\x60",
  [0x145e] = "\x8a\x61", [0x145f] = "\x8a\x62", [0x1460] = "\x8a\x63",
  [0x1461] = "\x8a\x64", [0x1462] = "\x8a\x65", [0x1463] = "\x8a\x66",
  [0x1464] = "\x8a\x67", [0x1465] = "\x8a\x68", [0x1466] = "\x8a\x69",
  [0x1467] = "\x8a\x6a", [0x1468] = "\x8a\x6b", [0x1469] = "\x8a\x6c",
  [0x146a] = "\x8a\x6d", [0x146b] = "\x8a\x6e", [0x146c] = "\x8a\x6f",
  [0x146d] = "\x8a\x70", [0x146e] = "\x8a\x71", [0x146f] = "\x8a\x72",
  [0x1470] = "\x8a\x73", [0x1471] = "\x8a\x74", [0x1472] = "\x8a\x75",
  [0x1473] = "\x8a\x76", [0x1474] = "\x8a\x77", [0x1475] = "\x8a\x78",
  [0x1476] = "\x8a\x79", [0x1477] = "\x8a\x7a", [0x1478] = "\x8a\x7b",
  [0x1479] = "\x8a\x7c", [0x147a] = "\x8a\x7d", [0x147b] = "\x8a\x7e",
  [0x147c] = "\x8a\xa1", [0x147d] = "\x8a\xa2", [0x147e] = "\x8a\xa3",
  [0x147f] = "\x8a\xa4", [0x1480] = "\x8a\xa5", [0x1481] = "\x8a\xa6",
  [0x1482] = "\x8a\xa7", [0x1483] = "\x8a\xa8", [0x1484] = "\x8a\xa9",
  [0x1485] = "\x8a\xaa", [0x1486] = "\x8a\xab", [0x1487] = "\x8a\xac",
  [0x1488] = "\x8a\xad", [0x1489] = "\x8a\xae", [0x148a] = "\x8a\xaf",
  [0x148b] = "\x8a\xb0", [0x148c] = "\x8a\xb1", [0x148d] = "\x8a\xb2",
  [0x148e] = "\x8a\xb3", [0x148f] = "\x8a\xb4", [0x1490] = "\x8a\xb5",
  [0x1491] = "\x8a\xb6", [0x1492] = "\x8a\xb7", [0x1493] = "\x8a\xb8",
  [0x1494] = "\x8a\xb9", [0x1495] = "\x8a\xba", [0x1496] = "\x8a\xbb",
  [0x1497] = "\x8a\xbc", [0x1498] = "\x8a\xbd", [0x1499] = "\x8a\xbe",
  [0x149a] = "\x8a\xbf", [0x149b] = "\x8a\xc0", [0x149c] = "\x8a\xc1",
  [0x149d] = "\x8a\xc2", [0x149e] = "\x8a\xc3", [0x149f] = "\x8a\xc4",
  [0x14a0] = "\x8a\xc5", [0x14a1] = "\x8a\xc6", [0x14a2] = "\x8a\xc7",
  [0x14a3] = "\x8a\xc8", [0x14a4] = "\x8a\xc9", [0x14a5] = "\x8a\xca",
  [0x14a6] = "\x8a\xcb", [0x14a7] = "\x8a\xcc", [0x14a8] = "\x8a\xcd",
  [0x14a9] = "\x8a\xce", [0x14aa] = "\x8a\xcf", [0x14ab] = "\x8a\xd0",
  [0x14ac] = "\x8a\xd1", [0x14ad] = "\x8a\xd2", [0x14ae] = "\x8a\xd3",
  [0x14af] = "\x8a\xd4", [0x14b0] = "\x8a\xd5", [0x14b1] = "\x8a\xd6",
  [0x14b2] = "\x8a\xd7", [0x14b3] = "\x8a\xd8", [0x14b4] = "\x8a\xd9",
  [0x14b5] = "\x8a\xda", [0x14b6] = "\x8a\xdb", [0x14b7] = "\x8a\xdc",
  [0x14b8] = "\x8a\xdd", [0x14b9] = "\x8a\xde", [0x14ba] = "\x8a\xdf",
  [0x14bb] = "\x8a\xe0", [0x14bc] = "\x8a\xe1", [0x14bd] = "\x8a\xe2",
  [0x14be] = "\x8a\xe3", [0x14bf] = "\x8a\xe4", [0x14c0] = "\x8a\xe5",
  [0x14c1] = "\x8a\xe6", [0x14c2] = "\x8a\xe7", [0x14c3] = "\x8a\xe8",
  [0x14c4] = "\x8a\xe9", [0x14c5] = "\x8a\xea", [0x14c6] = "\x8a\xeb",
  [0x14c7] = "\x8a\xec", [0x14c8] = "\x8a\xed", [0x14c9] = "\x8a\xee",
  [0x14ca] = "\x8a\xef", [0x14cb] = "\x8a\xf0", [0x14cc] = "\x8a\xf1",
  [0x14cd] = "\x8a\xf2", [0x14ce] = "\x8a\xf3", [0x14cf] = "\x8a\xf4",
  [0x14d0] = "\x8a\xf5", [0x14d1] = "\x8a\xf6", [0x14d2] = "\x8a\xf7",
  [0x14d3] = "\x8a\xf8", [0x14d4] = "\x8a\xf9", [0x14d5] = "\x8a\xfa",
  [0x14d6] = "\x8a\xfb", [0x14d7] = "\x8a\xfc", [0x14d8] = "\x8a\xfd",
  [0x14d9] = "\x8a\xfe", [0x14da] = "\x8b\x40", [0x14db] = "\x8b\x41",
  [0x14dc] = "\x8b\x42", [0x14dd] = "\x8b\x43", [0x14de] = "\x8b\x44",
  [0x14df] = "\x8b\x45", [0x14e0] = "\x8b\x46", [0x14e1] = "\x8b\x47",
  [0x14e2] = "\x8b\x48", [0x14e3] = "\x8b\x49", [0x14e4] = "\x8b\x4a",
  [0x14e5] = "\x8b\x4b", [0x14e6] = "\x8b\x4c", [0x14e7] = "\x8b\x4d",
  [0x14e8] = "\x8b\x4e", [0x14e9] = "\x8b\x4f", [0x14ea] = "\x8b\x50",
  [0x14eb] = "\x8b\x51", [0x14ec] = "\x8b\x52", [0x14ed] = "\x8b\x53",
  [0x14ee] = "\x8b\x54", [0x14ef] = "\x8b\x55", [0x14f0] = "\x8b\x56",
  [0x14f1] = "\x8b\x57", [0x14f2] = "\x8b\x58", [0x14f3] = "\x8b\x59",
  [0x14f4] = "\x8b\x5a", [0x14f5] = "\x8b\x5b", [0x14f6] = "\x8b\x5c",
  [0x14f7] = "\x8b\x5d", [0x14f8] = "\x8b\x5e", [0x14f9] = "\x8b\x5f",
  [0x14fa] = "\x8b\x60", [0x14fb] = "\x8b\x61", [0x14fc] = "\x8b\x62",
  [0x14fd] = "\x8b\x63", [0x14fe] = "\x8b\x64", [0x14ff] = "\x8b\x65",
  [0x1500] = "\x8b\x66", [0x1501] = "\x8b\x67", [0x1502] = "\x8b\x68",
  [0x1503] = "\x8b\x69", [0x1504] = "\x8b\x6a", [0x1505] = "\x8b\x6b",
  [0x1506] = "\x8b\x6c", [0x1507] = "\x8b\x6d", [0x1508] = "\x8b\x6e",
  [0x1509] = "\x8b\x6f", [0x150a] = "\x8b\x70", [0x150b] = "\x8b\x71",
  [0x150c] = "\x8b\x72", [0x150d] = "\x8b\x73", [0x150e] = "\x8b\x74",
  [0x150f] = "\x8b\x75", [0x1510] = "\x8b\x76", [0x1511] = "\x8b\x77",
  [0x1512] = "\x8b\x78", [0x1513] = "\x8b\x79", [0x1514] = "\x8b\x7a",
  [0x1515] = "\x8b\x7b", [0x1516] = "\x8b\x7c", [0x1517] = "\x8b\x7d",
  [0x1518] = "\x8b\x7e", [0x1519] = "\x8b\xa1", [0x151a] = "\x8b\xa2",
  [0x151b] = "\x8b\xa3", [0x151c] = "\x8b\xa4", [0x151d] = "\x8b\xa5",
  [0x151e] = "\x8b\xa6", [0x151f] = "\x8b\xa7", [0x1520] = "\x8b\xa8",
  [0x1521] = "\x8b\xa9", [0x1522] = "\x8b\xaa", [0x1523] = "\x8b\xab",
  [0x1524] = "\x8b\xac", [0x1525] = "\x8b\xad", [0x1526] = "\x8b\xae",
  [0x1527] = "\x8b\xaf", [0x1528] = "\x8b\xb0", [0x1529] = "\x8b\xb1",
  [0x152a] = "\x8b\xb2", [0x152b] = "\x8b\xb3", [0x152c] = "\x8b\xb4",
  [0x152d] = "\x8b\xb5", [0x152e] = "\x8b\xb6", [0x152f] = "\x8b\xb7",
  [0x1530] = "\x8b\xb8", [0x1531] = "\x8b\xb9", [0x1532] = "\x8b\xba",
  [0x1533] = "\x8b\xbb", [0x1534] = "\x8b\xbc", [0x1535] = "\x8b\xbd",
  [0x1536] = "\x8b\xbe", [0x1537] = "\x8b\xbf", [0x1538] = "\x8b\xc0",
  [0x1539] = "\x8b\xc1", [0x153a] = "\x8b\xc2", [0x153b] = "\x8b\xc3",
  [0x153c] = "\x8b\xc4", [0x153d] = "\x8b\xc5", [0x153e] = "\x8b\xc6",
  [0x153f] = "\x8b\xc7", [0x1540] = "\x8b\xc8", [0x1541] = "\x8b\xc9",
  [0x1542] = "\x8b\xca", [0x1543] = "\x8b\xcb", [0x1544] = "\x8b\xcc",
  [0x1545] = "\x8b\xcd", [0x1546] = "\x8b\xce", [0x1547] = "\x8b\xcf",
  [0x1548] = "\x8b\xd0", [0x1549] = "\x8b\xd1", [0x154a] = "\x8b\xd2",
  [0x154b] = "\x8b\xd3", [0x154c] = "\x8b\xd4", [0x154d] = "\x8b\xd5",
  [0x154e] = "\x8b\xd6", [0x154f] = "\x8b\xd7", [0x1550] = "\x8b\xd8",
  [0x1551] = "\x8b\xd9", [0x1552] = "\x8b\xda", [0x1553] = "\x8b\xdb",
  [0x1554] = "\x8b\xdc", [0x1555] = "\x8b\xdd", [0x1556] = "\x8b\xde",
  [0x1557] = "\x8b\xdf", [0x1558] = "\x8b\xe0", [0x1559] = "\x8b\xe1",
  [0x155a] = "\x8b\xe2", [0x155b] = "\x8b\xe3", [0x155c] = "\x8b\xe4",
  [0x155d] = "\x8b\xe5", [0x155e] = "\x8b\xe6", [0x155f] = "\x8b\xe7",
  [0x1560] = "\x8b\xe8", [0x1561] = "\x8b\xe9", [0x1562] = "\x8b\xea",
  [0x1563] = "\x8b\xeb", [0x1564] = "\x8b\xec", [0x1565] = "\x8b\xed",
  [0x1566] = "\x8b\xee", [0x1567] = "\x8b\xef", [0x1568] = "\x8b\xf0",
  [0x1569] = "\x8b\xf1", [0x156a] = "\x8b\xf2", [0x156b] = "\x8b\xf3",
  [0x156c] = "\x8b\xf4", [0x156d] = "\x8b\xf5", [0x156e] = "\x8b\xf6",
  [0x156f] = "\x8b\xf7", [0x1570] = "\x8b\xf8", [0x1571] = "\x8b\xf9",
  [0x1572] = "\x8b\xfa", [0x1573] = "\x8b\xfb", [0x1574] = "\x8b\xfc",
  [0x1575] = "\x8b\xfd", [0x1576] = "\x8b\xfe", [0x1577] = "\x8c\x40",
  [0x1578] = "\x8c\x41", [0x1579] = "\x8c\x42", [0x157a] = "\x8c\x43",
  [0x157b] = "\x8c\x44", [0x157c] = "\x8c\x45", [0x157d] = "\x8c\x46",
  [0x157e] = "\x8c\x47", [0x157f] = "\x8c\x48", [0x1580] = "\x8c\x49",
  [0x1581] = "\x8c\x4a", [0x1582] = "\x8c\x4b", [0x1583] = "\x8c\x4c",
  [0x1584] = "\x8c\x4d", [0x1585] = "\x8c\x4e", [0x1586] = "\x8c\x4f",
  [0x1587] = "\x8c\x50", [0x1588] = "\x8c\x51", [0x1589] = "\x8c\x52",
  [0x158a] = "\x8c\x53", [0x158b] = "\x8c\x54", [0x158c] = "\x8c\x55",
  [0x158d] = "\x8c\x56", [0x158e] = "\x8c\x57", [0x158f] = "\x8c\x58",
  [0x1590] = "\x8c\x59", [0x1591] = "\x8c\x5a", [0x1592] = "\x8c\x5b",
  [0x1593] = "\x8c\x5c", [0x1594] = "\x8c\x5d", [0x1595] = "\x8c\x5e",
  [0x1596] = "\x8c\x5f", [0x1597] = "\x8c\x60", [0x1598] = "\x8c\x61",
  [0x1599] = "\x8c\x62", [0x159a] = "\x8c\x63", [0x159b] = "\x8c\x64",
  [0x159c] = "\x8c\x65", [0x159d] = "\x8c\x66", [0x159e] = "\x8c\x67",
  [0x159f] = "\x8c\x68", [0x15a0] = "\x8c\x69", [0x15a1] = "\x8c\x6a",
  [0x15a2] = "\x8c\x6b", [0x15a3] = "\x8c\x6c", [0x15a4] = "\x8c\x6d",
  [0x15a5] = "\x8c\x6e", [0x15a6] = "\x8c\x6f", [0x15a7] = "\x8c\x70",
  [0x15a8] = "\x8c\x71", [0x15a9] = "\x8c\x72", [0x15aa] = "\x8c\x73",
  [0x15ab] = "\x8c\x74", [0x15ac] = "\x8c\x75", [0x15ad] = "\x8c\x76",
  [0x15ae] = "\x8c\x77", [0x15af] = "\x8c\x78", [0x15b0] = "\x8c\x79",
  [0x15b1] = "\x8c\x7a", [0x15b2] = "\x8c\x7b", [0x15b3] = "\x8c\x7c",
  [0x15b4] = "\x8c\x7d", [0x15b5] = "\x8c\x7e", [0x15b6] = "\x8c\xa1",
  [0x15b7] = "\x8c\xa2", [0x15b8] = "\x8c\xa3", [0x15b9] = "\x8c\xa4",
  [0x15ba] = "\x8c\xa5", [0x15bb] = "\x8c\xa6", [0x15bc] = "\x8c\xa7",
  [0x15bd] = "\x8c\xa8", [0x15be] = "\x8c\xa9", [0x15bf] = "\x8c\xaa",
  [0x15c0] = "\x8c\xab", [0x15c1] = "\x8c\xac", [0x15c2] = "\x8c\xad",
  [0x15c3] = "\x8c\xae", [0x15c4] = "\x8c\xaf", [0x15c5] = "\x8c\xb0",
  [0x15c6] = "\x8c\xb1", [0x15c7] = "\x8c\xb2", [0x15c8] = "\x8c\xb3",
  [0x15c9] = "\x8c\xb4", [0x15ca] = "\x8c\xb5", [0x15cb] = "\x8c\xb6",
  [0x15cc] = "\x8c\xb7", [0x15cd] = "\x8c\xb8", [0x15ce] = "\x8c\xb9",
  [0x15cf] = "\x8c\xba", [0x15d0] = "\x8c\xbb", [0x15d1] = "\x8c\xbc",
  [0x15d2] = "\x8c\xbd", [0x15d3] = "\x8c\xbe", [0x15d4] = "\x8c\xbf",
  [0x15d5] = "\x8c\xc0", [0x15d6] = "\x8c\xc1", [0x15d7] = "\x8c\xc2",
  [0x15d8] = "\x8c\xc3", [0x15d9] = "\x8c\xc4", [0x15da] = "\x8c\xc5",
  [0x15db] = "\x8c\xc6", [0x15dc] = "\x8c\xc7", [0x15dd] = "\x8c\xc8",
  [0x15de] = "\x8c\xc9", [0x15df] = "\x8c\xca", [0x15e0] = "\x8c\xcb",
  [0x15e1] = "\x8c\xcc", [0x15e2] = "\x8c\xcd", [0x15e3] = "\x8c\xce",
  [0x15e4] = "\x8c\xcf", [0x15e5] = "\x8c\xd0", [0x15e6] = "\x8c\xd1",
  [0x15e7] = "\x8c\xd2", [0x15e8] = "\x8c\xd3", [0x15e9] = "\x8c\xd4",
  [0x15ea] = "\x8c\xd5", [0x15eb] = "\x8c\xd6", [0x15ec] = "\x8c\xd7",
  [0x15ed] = "\x8c\xd8", [0x15ee] = "\x8c\xd9", [0x15ef] = "\x8c\xda",
  [0x15f0] = "\x8c\xdb", [0x15f1] = "\x8c\xdc", [0x15f2] = "\x8c\xdd",
  [0x15f3] = "\x8c\xde", [0x15f4] = "\x8c\xdf", [0x15f5] = "\x8c\xe0",
  [0x15f6] = "\x8c\xe1", [0x15f7] = "\x8c\xe2", [0x15f8] = "\x8c\xe3",
  [0x15f9] = "\x8c\xe4", [0x15fa] = "\x8c\xe5", [0x15fb] = "\x8c\xe6",
  [0x15fc] = "\x8c\xe7", [0x15fd] = "\x8c\xe8", [0x15fe] = "\x8c\xe9",
  [0x15ff] = "\x8c\xea", [0x1600] = "\x8c\xeb", [0x1601] = "\x8c\xec",
  [0x1602] = "\x8c\xed", [0x1603] = "\x8c\xee", [0x1604] = "\x8c\xef",
  [0x1605] = "\x8c\xf0", [0x1606] = "\x8c\xf1", [0x1607] = "\x8c\xf2",
  [0x1608] = "\x8c\xf3", [0x1609] = "\x8c\xf4", [0x160a] = "\x8c\xf5",
  [0x160b] = "\x8c\xf6", [0x160c] = "\x8c\xf7", [0x160d] = "\x8c\xf8",
  [0x160e] = "\x8c\xf9", [0x160f] = "\x8c\xfa", [0x1610] = "\x8c\xfb",
  [0x1611] = "\x8c\xfc", [0x1612] = "\x8c\xfd", [0x1613] = "\x8c\xfe",
  [0x1614] = "\x8d\x40", [0x1615] = "\x8d\x41", [0x1616] = "\x8d\x42",
  [0x1617] = "\x8d\x43", [0x1618] = "\x8d\x44", [0x1619] = "\x8d\x45",
  [0x161a] = "\x8d\x46", [0x161b] = "\x8d\x47", [0x161c] = "\x8d\x48",
  [0x161d] = "\x8d\x49", [0x161e] = "\x8d\x4a", [0x161f] = "\x8d\x4b",
  [0x1620] = "\x8d\x4c", [0x1621] = "\x8d\x4d", [0x1622] = "\x8d\x4e",
  [0x1623] = "\x8d\x4f", [0x1624] = "\x8d\x50", [0x1625] = "\x8d\x51",
  [0x1626] = "\x8d\x52", [0x1627] = "\x8d\x53", [0x1628] = "\x8d\x54",
  [0x1629] = "\x8d\x55", [0x162a] = "\x8d\x56", [0x162b] = "\x8d\x57",
  [0x162c] = "\x8d\x58", [0x162d] = "\x8d\x59", [0x162e] = "\x8d\x5a",
  [0x162f] = "\x8d\x5b", [0x1630] = "\x8d\x5c", [0x1631] = "\x8d\x5d",
  [0x1632] = "\x8d\x5e", [0x1633] = "\x8d\x5f", [0x1634] = "\x8d\x60",
  [0x1635] = "\x8d\x61", [0x1636] = "\x8d\x62", [0x1637] = "\x8d\x63",
  [0x1638] = "\x8d\x64", [0x1639] = "\x8d\x65", [0x163a] = "\x8d\x66",
  [0x163b] = "\x8d\x67", [0x163c] = "\x8d\x68", [0x163d] = "\x8d\x69",
  [0x163e] = "\x8d\x6a", [0x163f] = "\x8d\x6b", [0x1640] = "\x8d\x6c",
  [0x1641] = "\x8d\x6d", [0x1642] = "\x8d\x6e", [0x1643] = "\x8d\x6f",
  [0x1644] = "\x8d\x70", [0x1645] = "\x8d\x71", [0x1646] = "\x8d\x72",
  [0x1647] = "\x8d\x73", [0x1648] = "\x8d\x74", [0x1649] = "\x8d\x75",
  [0x164a] = "\x8d\x76", [0x164b] = "\x8d\x77", [0x164c] = "\x8d\x78",
  [0x164d] = "\x8d\x79", [0x164e] = "\x8d\x7a", [0x164f] = "\x8d\x7b",
  [0x1650] = "\x8d\x7c", [0x1651] = "\x8d\x7d", [0x1652] = "\x8d\x7e",
  [0x1653] = "\x8d\xa1", [0x1654] = "\x8d\xa2", [0x1655] = "\x8d\xa3",
  [0x1656] = "\x8d\xa4", [0x1657] = "\x8d\xa5", [0x1658] = "\x8d\xa6",
  [0x1659] = "\x8d\xa7", [0x165a] = "\x8d\xa8", [0x165b] = "\x8d\xa9",
  [0x165c] = "\x8d\xaa", [0x165d] = "\x8d\xab", [0x165e] = "\x8d\xac",
  [0x165f] = "\x8d\xad", [0x1660] = "\x8d\xae", [0x1661] = "\x8d\xaf",
  [0x1662] = "\x8d\xb0", [0x1663] = "\x8d\xb1", [0x1664] = "\x8d\xb2",
  [0x1665] = "\x8d\xb3", [0x1666] = "\x8d\xb4", [0x1667] = "\x8d\xb5",
  [0x1668] = "\x8d\xb6", [0x1669] = "\x8d\xb7", [0x166a] = "\x8d\xb8",
  [0x166b] = "\x8d\xb9", [0x166c] = "\x8d\xba", [0x166d] = "\x8d\xbb",
  [0x166e] = "\x8d\xbc", [0x166f] = "\x8d\xbd", [0x1670] = "\x8d\xbe",
  [0x1671] = "\x8d\xbf", [0x1672] = "\x8d\xc0", [0x1673] = "\x8d\xc1",
  [0x1674] = "\x8d\xc2", [0x1675] = "\x8d\xc3", [0x1676] = "\x8d\xc4",
  [0x1677] = "\x8d\xc5", [0x1678] = "\x8d\xc6", [0x1679] = "\x8d\xc7",
  [0x167a] = "\x8d\xc8", [0x167b] = "\x8d\xc9", [0x167c] = "\x8d\xca",
  [0x167d] = "\x8d\xcb", [0x167e] = "\x8d\xcc", [0x167f] = "\x8d\xcd",
  [0x1680] = "\x8d\xce", [0x1681] = "\x8d\xcf", [0x1682] = "\x8d\xd0",
  [0x1683] = "\x8d\xd1", [0x1684] = "\x8d\xd2", [0x1685] = "\x8d\xd3",
  [0x1686] = "\x8d\xd4", [0x1687] = "\x8d\xd5", [0x1688] = "\x8d\xd6",
  [0x1689] = "\x8d\xd7", [0x168a] = "\x8d\xd8", [0x168b] = "\x8d\xd9",
  [0x168c] = "\x8d\xda", [0x168d] = "\x8d\xdb", [0x168e] = "\x8d\xdc",
  [0x168f] = "\x8d\xdd", [0x1690] = "\x8d\xde", [0x1691] = "\x8d\xdf",
  [0x1692] = "\x8d\xe0", [0x1693] = "\x8d\xe1", [0x1694] = "\x8d\xe2",
  [0x1695] = "\x8d\xe3", [0x1696] = "\x8d\xe4", [0x1697] = "\x8d\xe5",
  [0x1698] = "\x8d\xe6", [0x1699] = "\x8d\xe7", [0x169a] = "\x8d\xe8",
  [0x169b] = "\x8d\xe9", [0x169c] = "\x8d\xea", [0x169d] = "\x8d\xeb",
  [0x169e] = "\x8d\xec", [0x169f] = "\x8d\xed", [0x16a0] = "\x8d\xee",
  [0x16a1] = "\x8d\xef", [0x16a2] = "\x8d\xf0", [0x16a3] = "\x8d\xf1",
  [0x16a4] = "\x8d\xf2", [0x16a5] = "\x8d\xf3", [0x16a6] = "\x8d\xf4",
  [0x16a7] = "\x8d\xf5", [0x16a8] = "\x8d\xf6", [0x16a9] = "\x8d\xf7",
  [0x16aa] = "\x8d\xf8", [0x16ab] = "\x8d\xf9", [0x16ac] = "\x8d\xfa",
  [0x16ad] = "\x8d\xfb", [0x16ae] = "\x8d\xfc", [0x16af] = "\x8d\xfd",
  [0x16b0] = "\x8d\xfe",
};

/* Definitions used in the body of the `gconv' function.  */
#define CHARSET_NAME		"BIG5HKSCS//"
#define FROM_LOOP		from_big5
#define TO_LOOP			to_big5
#define DEFINE_INIT		1
#define DEFINE_FINI		1
#define MIN_NEEDED_FROM		1
#define MAX_NEEDED_FROM		2
#define MIN_NEEDED_TO		4


/* First define the conversion function from Big5 to UCS4.  */
#define MIN_NEEDED_INPUT	MIN_NEEDED_FROM
#define MAX_NEEDED_INPUT	MAX_NEEDED_FROM
#define MIN_NEEDED_OUTPUT	MIN_NEEDED_TO
#define LOOPFCT			FROM_LOOP
#define BODY \
  {									      \
    uint32_t ch = *inptr;						      \
									      \
    if (ch >= 0x81 && ch <= 0xfe)					      \
      {									      \
	/* Two-byte character.  First test whether the next character	      \
	   is also available.  */					      \
	uint32_t ch2;							      \
	int idx;							      \
									      \
	if (__builtin_expect (inptr + 1 >= inend, 0))			      \
	  {								      \
	    /* The second character is not available.  */		      \
	    result = __GCONV_INCOMPLETE_INPUT;				      \
	    break;							      \
	  }								      \
									      \
	idx = (ch - 0x81) * 157;					      \
	ch2 = inptr[1];							      \
	/* See whether the second byte is in the correct range.  */	      \
	if (ch2 >= 0x40 && ch2 <= 0x7e)					      \
	  idx += ch2 - 0x40;						      \
	else if (__builtin_expect (ch2, 0xa1) >= 0xa1			      \
		 && __builtin_expect (ch2, 0xa1) <= 0xfe)		      \
	  idx += 0x3f + (ch2 - 0xa1);					      \
	else								      \
	  {								      \
	    /* This is illegal.  */					      \
	    if (! ignore_errors_p ())					      \
	      {								      \
		result = __GCONV_ILLEGAL_INPUT;				      \
		break;							      \
	      }								      \
									      \
	    ++inptr;							      \
	    ++*irreversible;						      \
	    continue;							      \
	  }								      \
									      \
	/* Get the value from the table.  */				      \
	ch = big5_to_ucs[idx];						      \
									      \
	/* Is this character defined?  */				      \
	if (__builtin_expect (ch, 1) == 0 && *inptr != '\0')		      \
	  {								      \
	    /* This is an illegal character.  */			      \
	    if (! ignore_errors_p ())					      \
	      {								      \
		result = __GCONV_ILLEGAL_INPUT;				      \
		break;							      \
	      }								      \
									      \
	    inptr += 2;							      \
	    ++*irreversible;						      \
	    continue;							      \
	  }								      \
									      \
	inptr += 2;							      \
      }									      \
    else								      \
      ++inptr;								      \
									      \
    put32 (outptr, ch);							      \
    outptr += 4;							      \
  }
#define LOOP_NEED_FLAGS
#include <iconv/loop.c>


/* Next, define the other direction.  */
#define MIN_NEEDED_INPUT	MIN_NEEDED_TO
#define MIN_NEEDED_OUTPUT	MIN_NEEDED_FROM
#define MAX_NEEDED_OUTPUT	MAX_NEEDED_FROM
#define LOOPFCT			TO_LOOP
#define BODY \
  {									      \
    uint32_t ch = get32 (inptr);					      \
    char buf[2];							      \
    const char *cp;							      \
									      \
    if (__builtin_expect (ch, 0)					      \
	>= sizeof (from_ucs4_tab1) / sizeof (from_ucs4_tab1[0]))	      \
      switch (ch)							      \
	{								      \
        case 0x2c7 ... 0x2d9:						      \
	  cp = from_ucs4_tab2[ch - 0x2c7];				      \
	  break;							      \
	case 0x391 ... 0x451:						      \
	  cp = from_ucs4_tab3[ch - 0x391];				      \
	  break;							      \
	case 0x2013 ... 0x203e:						      \
	  cp = from_ucs4_tab4[ch - 0x2013];				      \
	  break;							      \
	case 0x2103:							      \
	  cp = "\xa2\x4a";						      \
	  break;							      \
	case 0x2105:							      \
	  cp = "\xa1\xc1";						      \
	  break;							      \
	case 0x2109:							      \
	  cp = "\xa2\x4b";						      \
	  break;							      \
	case 0x2160 ... 0x2169:						      \
	  {								      \
	    buf[0] = '\xa2';						      \
	    buf[1] = '\xb9' + (ch - 0x2160);				      \
	    cp = buf;							      \
	  }								      \
	  break;							      \
	case 0x2190 ... 0x2199:						      \
	  cp = from_ucs4_tab5[ch - 0x2190];				      \
	  break;							      \
	case 0x221a ...	0x22bf:						      \
	  cp = from_ucs4_tab6[ch - 0x221a];				      \
	  break;							      \
	case 0x2460 ... 0x247d:						      \
	  cp = from_ucs4_tab7[ch - 0x2460];				      \
	  break;							      \
	case 0x2500 ... 0x2642:						      \
	  cp = from_ucs4_tab8[ch - 0x2500];				      \
	  break;							      \
	case 0x3000 ... 0x3129:						      \
	  cp = from_ucs4_tab9[ch - 0x3000];				      \
	  break;							      \
	case 0x32a3:							      \
	  cp = "\xa1\xc0";						      \
	  break;							      \
	case 0x338e ... 0x33d5:						      \
	  cp = from_ucs4_tab10[ch - 0x338e];				      \
	  break;							      \
	case 0x4e00 ... 0x9fa4:						      \
	  cp = from_ucs4_tab11[ch - 0x4e00];				      \
	  break;							      \
	case 0xe000 ... 0xf6b0:						      \
	  cp = from_ucs4_tab14[ch - 0xe000];				      \
	  break;							      \
	case 0xfa0c:							      \
	  cp = "\xc9\x4a";						      \
	  break;							      \
	case 0xfa0d:							      \
	  cp = "\xdd\xfc";						      \
	  break;							      \
	case 0xfe30 ... 0xfe6b:						      \
	  cp = from_ucs4_tab12[ch - 0xfe30];				      \
	  break;							      \
	case 0xff01 ... 0xff64:						      \
	  cp = from_ucs4_tab13[ch - 0xff01];				      \
	  break;							      \
	default:							      \
	  /* Illegal character.  */					      \
	  cp = "";							      \
	  break;							      \
	}								      \
    else								      \
      cp = from_ucs4_tab1[ch];						      \
									      \
    if (__builtin_expect (cp[0], '\1') == '\0' && ch != 0)		      \
      {									      \
	/* Illegal character.  */					      \
	STANDARD_ERR_HANDLER (4);					      \
      }									      \
    else								      \
      {									      \
	/* See whether there is enough room for the second byte we write.  */ \
	if (__builtin_expect (cp[1], '\1') != '\0'			      \
	    && __builtin_expect (outptr + 1 >= outend, 0))		      \
	  {								      \
	    /* We have not enough room.  */				      \
	    result = __GCONV_FULL_OUTPUT;				      \
	    break;							      \
	  }								      \
									      \
	*outptr++ = cp[0];						      \
	if (cp[1] != '\0')						      \
	  *outptr++ = cp[1];						      \
      }									      \
									      \
    inptr += 4;								      \
  }
#define LOOP_NEED_FLAGS
#include <iconv/loop.c>


/* Now define the toplevel functions.  */
#include <iconv/skeleton.c>
